import astropy.io.fits as pyfits
import numpy as np
import os
import copy
from .image import ccdproc
from ..utils import macro
from ..utils import messager
from . import m2fsConfigs
import sys
import ipdb


def m2fsImageMerge(c1, c2, c3, c4):
    c00 = c4
    c01 = np.fliplr(c3)
    c10 = np.flipud(c1)
    c11 = np.fliplr(np.flipud(c2))
    result = np.concatenate((\
        np.concatenate((c00, c01), axis=1),
        np.concatenate((c10, c11), axis=1)
        ),axis=0)
    return result

def amp(inputFile, configs={}):#<==
    """merge 4 amp images and perform bias subtraction

    Parameters
    ---------------
    inputFile : str
        one file name
    configs : configs about amp
        rerun : whether to rerun this process and overwrite the old result

    """#==>
    outputFileName = os.path.basename(inputFile)+".fits"

    configs.update({"config2show":["rerun", "subDirection"]})
    global logger
    exec(messager.initLoggerMacro)
    global outputPath, outFile, rerun, toReturn
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return

    c1name= str(inputFile) + 'c1.fits'
    c2name= str(inputFile) + 'c2.fits'
    c3name= str(inputFile) + 'c3.fits'
    c4name= str(inputFile) + 'c4.fits'
    # do overscan
    configs.update({"makeZeroExt":True})
    hdulist1 = ccdproc.overscanAndTrim(c1name, configs)
    hdulist2 = ccdproc.overscanAndTrim(c2name, configs)
    hdulist3 = ccdproc.overscanAndTrim(c3name, configs)
    hdulist4 = ccdproc.overscanAndTrim(c4name, configs)
    # adu to electron and calculate
    configs.update({'errType':'error'})
    hdulist1 = ccdproc.adu2electro(hdulist1, configs)
    hdulist2 = ccdproc.adu2electro(hdulist2, configs)
    hdulist3 = ccdproc.adu2electro(hdulist3, configs)
    hdulist4 = ccdproc.adu2electro(hdulist4, configs)
    c1, e1 = hdulist1["SCI"].data, hdulist1["STD"].data
    c2, e2 = hdulist2["SCI"].data, hdulist2["STD"].data
    c3, e3 = hdulist3["SCI"].data, hdulist3["STD"].data
    c4, e4 = hdulist4["SCI"].data, hdulist4["STD"].data
    h1, h2, h3, h4 =\
            hdulist1[0].header,hdulist2[0].header,hdulist3[0].header,hdulist4[0].header
    en1 = np.ones(c1.shape)*h1["ENOISE"]
    en2 = np.ones(c2.shape)*h2["ENOISE"]
    en3 = np.ones(c3.shape)*h3["ENOISE"]
    en4 = np.ones(c4.shape)*h4["ENOISE"]
    # merge 4 frame
    data = m2fsImageMerge(c1, c2, c3, c4)
    error = m2fsImageMerge(e1, e2, e3, e4)
    enoise = m2fsImageMerge(en1, en2, en3, en4)
    #update header
    header = h1
    oldObjName = header["OBJECT"]
    sheader = hdulist1["SCI"].header
    eheader = hdulist1["STD"].header
    enheader = pyfits.Header()
    enheader["extname"] = "ENOISE"
    header['BUNIT']= 'E-/PIXEL'
    del header['EGAIN']
    header.append(('ENOISE',
        str([header['ENOISE'], h2['ENOISE'], h3['ENOISE'], h4['ENOISE']]),
        "ENOISE of 4 amplifier"))
    del header['BIASSEC']
    assert data.shape==(2056, 2048), data.shape #!! m2fs config22 image size
    header['DATASEC']='[1:2048,1:2056]'
    header['TRIMSEC']='[1:2048,1:2056]'
    #format(configs.get("object", "?"),
    objStr = "{} @ {} @ {}".\
            format(header["OBJECT"],
                   header.get("EXPTIME", "?"),
                   "T".join([header.get("UT-DATE","?"),
                             header.get("UT-TIME")]))
    header['OLDOBJ'] =oldObjName
    sheader['OLDOBJ']=oldObjName
    eheader['OLDOBJ']=oldObjName
    enheader['OLDOBJ']=oldObjName
    header['object'] =objStr
    sheader['object']=objStr
    eheader['object']=objStr
    enheader['object']=objStr
    header['FILENAME']= str(inputFile)
    header.add_history('amp: Merged 4 amplifiers into a single frame')
    header.add_history('amp: calculate std error and save it at frame 2')
    #save merged image into fits file
    primaryHDU = pyfits.PrimaryHDU(None, header = header)
    outData = pyfits.ImageHDU(data, header=sheader)
    outError= pyfits.ImageHDU(error, header=eheader)
    outEn= pyfits.ImageHDU(enoise, header=enheader)
    fits = pyfits.HDUList([primaryHDU, outData, outError, outEn])
    fits.writeto(outFile, clobber='True')
    return outFile, objStr.split("@")[-1]
