import astropy.io.fits as pyfits
import numpy as np
import os
import copy
from .image import ccdproc
from ..utils import macro
from ..utils import messager
from . import m2fsConfigs
import sys
import ipdb

def lacosmic(inputFile, configs={}):#<==
    """Do lacosmic for all input files

    Parameters
    ---------------
    inputFile : str
        one file name
    configs : configs about cosmic
        rerun : whether to rerun this process and overwrite the old result

    """#==>
    outputFileName = "c" + os.path.basename(inputFile)
    configs.update({"config2show":["rerun", "Ngrow",
                                   "sigclip", "maxiter"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return

    thisConfig = configs
    configs["logger"] = logger

    fits = ccdproc.LACosmic(inputFile, configs=thisConfig)

    objStr = fits[0].header["object"]
    fits.writeto(outFile, clobber='True')
    return outFile, objStr.split("@")[-1]

def avecosmic(inputFileList, configs={}):#<==
    """
    Parameters
    ---------------
    inputFile : str
    configs : configs about cosmic
    """#==>
    #name = configs.get("name", "????")
    name = "_".join(("".join(inputFileList)).split("outputs/"))
    combineFile = os.path.join(configs["outputPath"],
                               "aveCosmicCombine_{}.fits".format(name))
    rerun = configs.get("rerun", False)
    exec(messager.initLoggerMacro)

    fileName = globals().get("__file__","")
    if fileName:
        fileName = os.path.basename(fileName)
        fileName = fileName.split(".py")[0]
        fileName += "."
    funcName = sys._getframe().f_code.co_name
    logger.p("{}{}({}, {})".format(fileName, funcName, [os.path.basename(each) for each in inputFileList], {"rerun": rerun}))

    thisConfig = configs
    thisConfig["logger"] = logger
    if rerun or not os.path.exists(combineFile):
        logger.p("\tcombine({}, {})".format(inputFileList,
            {"hsigrej":thisConfig.get("hsigrej"), "lsigrej":thisConfig.get("lsigrej") }))
        combineFits = ccdproc.combine(inputFileList, configs=thisConfig)
        combineFits.writeto(combineFile, clobber='True')
    else:
        combineFits = pyfits.open(combineFile)
    N = len(inputFileList)
    maskIndexs = (np.arange(N)+2) * 3

    #ipdb.set_trace()
    #!! make new masks
    result = []
    for index, inputFile in zip(maskIndexs,inputFileList):
        outFile = os.path.join(configs["outputPath"], "c"+os.path.basename(inputFile))
        if not rerun and os.path.exists(outFile):
            logger.p("\t {} already exists, skip".format(outFile))
            result.append((outFile, pyfits.open(outFile)[0].header["object"].split("@")[-1]))
            continue
        else:
            logger.p("\tavecosmic({})".format(outFile))
        fits = ccdproc.growByMask(inputFile, combineFits[index].data, {"logger":logger})
        objStr = fits[0].header["object"]
        fits.writeto(outFile, clobber='True')
        result.append((outFile, objStr.split("@")[-1]))
    return result, combineFile

def fixpixel(inputFile, configs={}):
    outputFileName = "masterFlat.fits"
    configs.update({"config2show":["interpDirection"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return

    if configs.get("interDirection") is not None:
        fixpixelConfig = {"direction":thisConfig.get("interDirection")}
    else:
        fixpixelConfig = {}
    fixpixelConfig["logger"] = logger
    fits = ccdproc.fixpixel(inputFile, configs=fixpixelConfig)

    objStr = fits[0].header["object"]
    fits.writeto(outFile, clobber='True')
    return outFile, objStr.split("@")[-1]
