import astropy.io.fits as pyfits
import numpy as np
import os
import copy
from .image import ccdproc
from ..utils import macro
from ..utils import messager
from . import m2fsConfigs
import sys
import pdb

def masterDark(inputFile, configs={}):#<==
    """Do make master Dark

    Parameters
    ---------------
    inputFiles : str
        dark file list
    configs : configs about cosmic
        rerun : whether to rerun this process and overwrite the old result

    """#==>
    outputFileName = "masterDark_notclean.fits"
    configs.update({"config2show":["method", "maxiter", "sigrej", "maxRej"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return

    thisConfig = configs
    thisConfig["logger"] = logger

    fits = ccdproc.combine(inputFile, configs=thisConfig)

    objStr = fits[0].header["object"]
    fits.writeto(outFile, clobber='True')
    return outFile, objStr.split("@")[-1]

def fixpixel(inputFile, configs={}):
    outputFileName = "masterDark.fits"
    configs.update({"config2show":["interpDirection"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return

    if configs.get("interDirection") is not None:
        fixpixelConfig = {"direction":thisConfig.get("interDirection")}
    else:
        fixpixelConfig = {}
    fixpixelConfig["logger"] = logger
    fits = ccdproc.fixpixel(inputFile, configs=fixpixelConfig)

    objStr = fits[0].header["object"]
    fits.writeto(outFile, clobber='True')
    return outFile, objStr.split("@")[-1]

def subDark(inputFileList, configs={}):#<==
    """subtract master Dark

    Parameters
    ---------------
    inputFiles :
    configs :

    """#==>
    masterDarkFile = os.path.join(configs["outputPath"], "masterDark.fits")
    rerun = configs.get("rerun", False)
    exec(messager.initLoggerMacro)

    fileName = globals().get("__file__","")
    if fileName:
        fileName = os.path.basename(fileName)
        fileName = fileName.split(".py")[0]
        fileName += "."
    funcName = sys._getframe().f_code.co_name
    logger.p("{}{}({}, {})".format( fileName, funcName, [os.path.basename(each) for each in inputFileList], {"rerun": rerun}))

    thisConfig = configs
    thisConfig["logger"] = logger
    objM = pyfits.open(masterDarkFile)
    mheader = objM[0].header
    mtime = mheader["EXPTIME"]

    result = []
    for inputFile in inputFileList:
        outFile = os.path.join(configs["outputPath"], "d"+os.path.basename(inputFile))
        if os.path.exists(outFile):
            logger.p("\t {} already exists, skip".format(outFile))
            result.append((outFile, pyfits.open(outFile)[0].header["object"].split("@")[-1]))
            continue
        else:
            logger.p("\tsubdark({})".format(outFile))

        obj = pyfits.open(inputFile)
        header = obj[0].header
        sheader = obj[1].header
        eheader = obj[2].header
        time = header["EXPTIME"]

        data = obj["SCI"].data
        error = obj["STD"].data
        dataM = objM["SCI"].data / mtime*time
        errorM = objM["STD"].data / mtime*time

        data, error = ccdproc.subtract(data, error, dataM, errorM)

        obj["SCI"].data = data
        obj["STD"].data = error
        header.add_history("Subtract masterDark")
        sheader.add_history("Subtract masterDark")
        eheader.add_history("Subtract masterDark")
        fits = obj

        objStr = fits[0].header["object"]
        fits.writeto(outFile, clobber='True')
        result.append((outFile, objStr.split("@")[-1]))
    return result
