from __future__ import (absolute_import, division, print_function, unicode_literals)
import numpy as np
from scipy.optimize import curve_fit
from matplotlib import cm
import matplotlib as mpl
import matplotlib.pyplot as plt
import ipdb
import pickle
import os
import sys

#!! some bugs when do parallel
# http://stackoverflow.com/questions/16254191/python-rpy2-and-matplotlib-conflict-when-using-multiprocessing
#mpl.use(mpl.rcParamsDefault["backend"])
# Add backend: Agg to ~/.matplotlib/matplotlibrc !! solve it


# Define model function to be used to fit to the data above:
def gauss(x, *p):
    A, mu, sigma = p
    diff = (x-mu)/sigma
    return A/(np.sqrt(2*np.pi)*np.abs(sigma))*np.exp(-diff**2 / 2.)

def gauss3(x, *p):
    A, mu, sigma = p
    diff = (x-mu)/sigma
    return A/(2.575798634*np.abs(sigma))*np.exp(-np.abs(diff)**3 / 3.)

def fitGauss(x, y, p0):
    coeff, var_matrix = curve_fit(gauss, x, y, p0=p0)
    model = gauss(x, *coeff)
    return coeff, model

def fitPlot(x, y, p0):
    xmin = x.min()
    xmax = x.max()
    coeff, var_matrix = curve_fit(gauss, x, y, p0=p0)
    xx = np.linspace(xmin, xmax, 100)
    model = gauss(x, *coeff)
    mmodel = gauss(xx, *coeff)
    ymin = mmodel.min()
    ymax = mmodel.max()*1.3

    print("init: {:12.5f}, {:12.5f}, {:12.5f}".format(*p0))
    print(" fit: {:12.5f}, {:12.5f}, {:12.5f}".format(*coeff))
    fig = plt.figure()
    plt.plot(x, model)
    plt.plot(xx, mmodel)
    plt.plot(np.array([p0[1], p0[1]]), np.array([ymin, ymax]), 'r--')
    plt.plot(np.array([coeff[1], coeff[1]]), np.array([ymin, ymax]), 'g--')
    plt.scatter(x,y, s=2)
    plt.show(0)

def fitPlotWithSigma(x, y, sigma, p0):
    xmin = x.min()
    xmax = x.max()
    coeff, var_matrix = curve_fit(gauss, x, y, p0=p0, sigma=sigma)
    xx = np.linspace(xmin, xmax, 100)
    model = gauss(x, *coeff)
    mmodel = gauss(xx, *coeff)
    ymin = mmodel.min()
    ymax = mmodel.max()*1.3

    print("init: {:12.5f}, {:12.5f}, {:12.5f}".format(*p0))
    print(" fit: {:12.5f}, {:12.5f}, {:12.5f}".format(*coeff))
    fig = plt.figure()
    plt.plot(x, model)
    plt.plot(xx, mmodel)
    plt.plot(np.array([p0[1], p0[1]]), np.array([ymin, ymax]), 'r--')
    plt.plot(np.array([coeff[1], coeff[1]]), np.array([ymin, ymax]), 'g--')
    plt.scatter(x,y, s=5)
    plt.show(0)


def fitPlotWithSigma3(x, y, sigma, p0):
    xmin = x.min()
    xmax = x.max()
    coeff, var_matrix = curve_fit(gauss3, x, y, p0=p0, sigma=sigma)
    xx = np.linspace(xmin, xmax, 100)
    model = gauss(x, *coeff)
    mmodel = gauss(xx, *coeff)
    ymin = mmodel.min()
    ymax = mmodel.max()*1.3

    print("init: {:12.5f}, {:12.5f}, {:12.5f}".format(*p0))
    print(" fit: {:12.5f}, {:12.5f}, {:12.5f}".format(*coeff))
    fig = plt.figure()
    plt.plot(x, model)
    plt.plot(xx, mmodel)
    plt.plot(np.array([p0[1], p0[1]]), np.array([ymin, ymax]), 'r--')
    plt.plot(np.array([coeff[1], coeff[1]]), np.array([ymin, ymax]), 'g--')
    plt.scatter(x,y, s=5)
    plt.show(0)

p = plt.plot
s = plt.show
f = plt.figure
sc= plt.scatter
text = plt.text
gca=plt.gca
gcf=plt.gcf




def d(fig, filename, configs={}):
    xscale = configs.get("xscale")
    if xscale is not None:
        ax = fig.get_axes()[0]
        xlim = ax.get_xlim()
        xstart = xlim[0]
        xend = xstart + (xlim[1]-xlim[0])*xscale
        ax.set_xlim([xstart, xend])
    dirname = os.path.dirname(filename)
    if dirname:
        safe_mkdir(dirname)
    with open(filename, "wb") as f:
        pickle.dump(fig, f)

def l(filename):
    with open(filename, "rb") as f:
        return pickle.load(f)

def safe_mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)

def piecePlotMulti(input, outDir="./Debug", prefix="", N=100):
    """input should have (x, y, type, mask) or (x,y)"""
    safe_mkdir(outDir)
    print("output {} debug figures in {}".format(N, outDir))
    globalMin =  np.inf
    globalMax = -np.inf
    max2dsize=0
    todo = []
    for each in input:
        x = each[0]
        y = each[1]
        t = each[2]
        if len(each)==3:
            if len(x.shape)==1:
                mask=np.isfinite(y)
                x = x[mask]
                y = y[mask]
            elif len(x.shape)==2:
                max2dsize = np.max([max2dsize, x.shape[0]])
                #randomColor = np.random.rand(max2dsize,3)
                randomColor = cm.get_cmap("gist_rainbow")\
                    (np.linspace(0,1,max2dsize))
        elif len(each)==4:
            mask = each[3]
            x = x[mask]
            y = y[mask]
        else:
            raise Exception("")
        try:
            globalMin = np.fmin(x[:].min(), globalMin)
            globalMax = np.fmax(x[:].max(), globalMax)
        except:
            print("stop here")
        todo.append([x, y, t])
    width = (globalMax-globalMin)/N
    nodes = np.linspace(globalMin, globalMax, N+1)
    starts = nodes[:-1]-width*0.1
    ends = nodes[1:]+width*0.1
    fig = plt.figure()
    ax = fig.add_subplot(111)
    if max2dsize>0:
        for i in range(max2dsize):
            ax.scatter(0,0,s=100, label=str(i), c=randomColor[i])
        plt.legend(loc=2)
        plt.savefig(os.path.join(outDir,"-ColorIndex.png"), bbox_inches='tight')
        ax.clear()
    for i in range(N):
        sys.stdout.write('output figure {}\r'.format(i))
        sys.stdout.flush()
        for each in todo:
            if len(each[0].shape)==1:
                x = each[0]
                y = each[1]
                t = each[2]
                inRange = np.logical_and(x>starts[i], x<ends[i])
                thex = x[inRange]
                they = y[inRange]
                if type(t)==list:
                    if t[0]=="plot":
                        ax.plot(thex, they, **t[1])
                    elif t[0]=="scatter":
                        ax.scatter(thex, they, **t[1])
                    else:
                        raise Exception("unknown type {}".format(t))
                elif type(t)==str:
                    if t=="plot":
                        ax.plot(thex, they)
                    elif t=="scatter":
                        ax.scatter(thex, they)
                    else:
                        raise Exception("unknown type {}".format(t))
            elif len(each[0].shape)==2:
                t = each[2]
                n=-1
                for x, y in zip(each[0], each[1]):
                    n+=1
                    inRange = np.logical_and(x>starts[i], x<ends[i])
                    thex = x[inRange]
                    they = y[inRange]
                    if type(t)==list:
                        if t[0]=="plot":
                            ax.plot(thex, they, **t[1])
                        elif t[0]=="scatter":
                            ax.scatter(thex, they, **t[1])
                        else:
                            raise Exception("unknown type {}".format(t))
                    elif type(t)==str:
                        if "plot" in t:
                            try:
                                s=int(t.split("plot")[1])
                            except ValueError:
                                ax.plot(thex, they, c=randomColor[n])
                            else:
                                ax.plot(thex, they, c=randomColor[n], s=s)
                        elif "scatter" in t:
                            try:
                                s=int(t.split("scatter")[1])
                            except ValueError:
                                ax.scatter(thex, they, c=randomColor[n], edgecolors="None")
                            else:
                                ax.scatter(thex, they,
                                           c=randomColor[n], edgecolors="None", s=s)
                        else:
                            raise Exception("unknown type {}".format(t))
            else:
                raise Exception("x.shape = {} is unknown".format(x.shape))
        ax.set_xlim(starts[i], ends[i])
        fig.savefig(os.path.join(outDir,"{}{}.png".format(prefix, i)))
        ax.clear()
    plt.close(fig)


def dp(plotList, title="", filename="./debug.fig"):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    for each in plotList:
        x = each[0]
        y = each[1]
        if len(each)==3:
            kwargs = each[2]
        else:
            kwargs = {}
        if kwargs.get("type") is None:
            ax.plot(x,y,**kwargs)
        elif kwargs.get("type") =="plot":
            ax.plot(x,y,**kwargs)
        elif kwargs.get("type") =="scatter":
            ax.scatter(x,y,**kwargs)
        else:
            raise Exception("unknown plot type:{}".format(kwargs.get("type")))
    d(fig, filename)
    plt.close(fig)
