from __future__ import (absolute_import, division, print_function, unicode_literals)
import numpy as np
import collections
OD = collections.OrderedDict
import os
import sys
import matplotlib.pyplot as plt
import numpy
from numpy.polynomial import Legendre
from . import de
from .image import ccdproc
from .image import specLib
from ..utils import macro
from ..utils import messager
from ..utils import configTools
import scipy.interpolate as interp
from scipy.optimize import curve_fit
import glob
import multiprocessing
import copy
from ..DS10 import ds9
import ipdb

from . import m2fsConfigs
import shutil
import astropy.io.fits as pyfits
from astropy import wcs
from astropy.modeling import models, fitting
from astropy.modeling.models import Voigt1D

regStrH="""# Region file format: DS9 version 4.1
global color={color} dashlist=8 3 width=1 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1
physical
"""
regStrP = "circle {0} {1} {2}i #color={3} move=0\n"
regStrT = "#text({0},{1}) text={{{2}}} color={3}\n"

def makeTraceSet(inputFile, configs={}):
    #region: setup
    outputFileName= 'traceset.fits'
    configs.update({"config2show":["rerun", "threshFactor"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return

    traceDebugFits = os.path.join(outputPath, 'ysetDebug.fits')
    traceDebugFig = os.path.join(outputPath, 'ysetDebug.fig')

    fiberInfo = configs["fiberInfo"]
    badFiber = fiberInfo["type"]=="I"
    makeRegs = configs.get("makeRegs", True)
    infile = configs.get("infile", "tracing")
    blockSize = configs.get("blockSize", 16)

    waveDirection = configs.get('waveDirection', 'row')
    if waveDirection not in ['row', 'column']:
        raise logger.Exception('invalid waveDirection!')

    debug = configs.get('debug',1)
    traceConfig = {'infile':infile, "logger":logger,
                   "threshFactor":configs.get("threshFactor")}
    obj = pyfits.open(inputFile)
    data = obj["sci"].data; sheader = obj["sci"].header
    error = obj["std"].data; eheader = obj["std"].header
    # later we call the wave direction the "x" direction, the space direction the "y" direction
    if waveDirection=='column':
        data = data.T.copy() # copy is necessary because we call c function latter
        error = error.T.copy()
    ny, nx = data.shape
    # trace for center
    thisConfig = configs
    thisConfig.update(traceConfig)
    #endregion: setup
    #region: tracing
    #!!                     indexs are good indexs
    initX, yPeaksCenterRaw, indexs = traceCenter(data, error, configs = thisConfig)
    nFiberRaw = yPeaksCenterRaw.shape[0] # should be 128
    badIndexs = np.array(list(set(range(nFiberRaw)) - set(indexs.tolist())))
    yPeaksCenter = yPeaksCenterRaw[indexs]
    nFiber = yPeaksCenter.size # number of fibers that detected by peak finder
    traceRawResult=os.path.join(outputPath, 'traceRawResult.fits')
    # trace for all
    thisConfig = {'initX':initX, 'yPeaks':yPeaksCenter}
    yset = np.ones((nFiberRaw, data.shape[1]))*np.nan
    yerror = np.ones((nFiberRaw, data.shape[1]))*np.inf
    yset_r, yerror_r = traceImageByFlux(data, error, configs = thisConfig)
    yset[indexs, :] = yset_r[:,:]
    yerror[indexs, :] = yerror_r[:,:]
    for eachBadIndex in badIndexs: #!! fill in the bad fiber
        yset[eachBadIndex, :] = guessForMissingFiber(eachBadIndex,
                             indexs, np.arange(nFiberRaw), blockSize, yset)
        yerror[eachBadIndex, :] = guessForMissingFiber(eachBadIndex,
                             indexs, np.arange(nFiberRaw), blockSize, yerror)
    ccdproc.toFits(traceRawResult, [(yset, None), (yerror, None)])
    # fit it to be a leg
    xx = np.arange(nx)
    ysetFit = np.zeros(yset.shape)

    fig = plt.figure()
    ax = fig.add_subplot(111)
    for i in range(nFiberRaw):
        try:
            leg = Legendre.fit(xx, yset[i,:], 5, w=1/yerror[i,:]**2)
        except Exception as e:
            print("i={}".format(i))
            print(e)
            ipdb.set_trace()
        ysetFit[i,:] = leg(xx)
        ax.plot(xx[::10], yset[i,::10], '--.', xx, ysetFit[i,:])
    de.d(fig, traceDebugFig)
    #plt.show()
    fig.clear()
    plt.close()

    diff = np.abs(ysetFit - yset)
    badFib = np.sum(np.logical_and(diff>0.5,~np.isfinite(yerror)), axis=1)>len(yset)*0.1
    badMask = np.tile(badFib.astype(np.int), (nx,1)).T
    badFib = np.logical_or(badFib, badFiber) #!! if set to badmask, mask all pixels
    thisConfig = {'kind':'linear', 'axis':"row", "logger":logger}
    ysetFitFit = ccdproc.interpImage(ysetFit, badMask, configs=thisConfig)
    yerror[badIndexs, :] = np.inf
    badFib[badIndexs] = 1

    if debug:
        ccdproc.toFits(traceDebugFits,
                [(yset, pyfits.Header([("OBJECT","yset")])),
                 (ysetFit, pyfits.Header([("OBJECT","ysetFit")])),
                 (ysetFitFit, pyfits.Header([("OBJECT","ysetFitFit")])),
                 (np.abs(ysetFitFit-ysetFit), pyfits.Header([("OBJECT","fitfit-fit")])),
                 (np.abs(ysetFit-yset), pyfits.Header([("OBJECT","fit - yset")])),
                 (np.abs(ysetFit-yset)/yset, pyfits.Header([("OBJECT","(fit-yset)/yset")]))])

    #!! cut blocks
    blockNumber = int(nFiberRaw / blockSize)
    boundary = np.ones((blockNumber, 2))
    for i in range(blockNumber):
        temp = ysetFitFit[i*blockSize:(i+1)*blockSize]
        boundary[i][0] = temp.min() - 30
        boundary[i][1] = temp.max() + 30
    data1 = pyfits.PrimaryHDU(ysetFitFit, pyfits.Header([("OBJECT","yset fit and fixpixel")]))
    data2 = pyfits.ImageHDU(yerror, pyfits.Header([("OBJECT","yset error")]))
    data3 = pyfits.ImageHDU(badFib.astype(np.int32), pyfits.Header([("OBJECT","bad fiber mask")]))
    data4 = pyfits.ImageHDU(boundary.astype(np.int32), pyfits.Header([("OBJECT","block boundary")]))
    fits  = pyfits.HDUList([data1, data2, data3, data4])
    fits.writeto(outFile, clobber='True')

    #!! make reg file for masterFlat
    if makeRegs:
        outfileR = os.path.join(outputPath, "masterFlat.reg")
        genTracingMarkers(outfileR, fiberInfo, ysetFitFit, badFib, np.where(~badFib)[0])
    #endregion: tracing

#!! badFiberMask.size=128, allGoodIndex is good fiber that have good wavelength
def genTracingMarkers(filename, fiberInfo, yset, badFiberMask, allGoodIndex):
    xx = np.arange(yset.shape[1])
    if fiberInfo is not None:
        fiberID = [each['fiber'] + b'_' + each['type'] + b"_" + each["id"] for each in fiberInfo]
    else:
        ipdb.set_trace()
    with open(filename, "w") as f:
        f.write(regStrH)
        ind = -1
        indf = -1
        for i in range(badFiberMask.size):
            try:
                name = fiberID[i].decode("utf8")
            except:
                name = fiberID[i]
            if (not badFiberMask[i]) and (i in allGoodIndex):
                ind += 1
                indf += 1
                color = ["red", "green"][i%2]
                text  = "{:3} {:3} {:3} {:<15} ".format(str(i), str(ind), str(indf), name)
                f.write(regStrT.format(-20, yset[i][0]+1, text, color))
                for x in xx[::10]:
                    f.write(regStrP.format(x+1, yset[i][x]+1, 0.5, color))
            elif (not badFiberMask[i]):
                ind += 1
                text  = "{:3} {:3} {:<15} ".format(str(i), str(ind), name)
                color = "blue"
                f.write(regStrT.format(-20, yset[i][0]+1, text, color))
                for x in xx[::10]:
                    f.write(regStrP.format(x+1, yset[i][x]+1, 0.5, color))
            else:
                text  = "{:3}      {:<15} ".format(str(i), name)
                f.write(regStrT.format(-20, yset[i][0]+1, text, "black"))

def genTracingMarkersByNames(filename, names, yset):
    xx = np.arange(yset.shape[1])
    with open(filename, "w") as f:
        f.write(regStrH)
        ind = -1
        for i, eachname in enumerate(names):
            ind += 1
            color = ["red", "green"][i%2]
            text  = "{:<15} ".format(eachname)
            f.write(regStrT.format(-20, yset[i][0], text, color))
            for x in xx[::10]:
                f.write(regStrP.format(x+1, yset[i][x], 0.5, color))
def genTracingMarkersByNamesSimple(filename, names, yset):
    xx = np.arange(yset.shape[1])
    with open(filename, "w") as f:
        f.write(regStrH)
        ind = -1
        for i, eachname in enumerate(names):
            ind += 1
            color = ["red", "green"][i%2]
            text  = "{:<15} ".format(eachname)
            f.write(regStrT.format(-20, yset[i][0], text, color))


def traceImageByFlux(image, error, configs={}):
    """ do tracing """
    initX  = configs['initX']
    yPeaks = configs['yPeaks']
    maxerr = configs.get('maxerr', 0.2)
    #maxshift = configs.get('maxshift', 0.5)
    maxshift = configs.get('maxshift', 1.0)
    ny, nx = image.shape
    nTrace = yPeaks.size
    ivar = 1./error**2
    # box mean along nx
    imgCov  = np.apply_along_axis(boxMean, 1, image*ivar, 5)
    ivarCov = np.apply_along_axis(boxMean, 1, ivar, 5)
    smoothImage = imgCov/ivarCov;
    #imageUtils.toFits('ivarCov.fits', [(imgCov0/ivarCov0, None),(smoothImage, None)])
    # y center at each xind along one fiber
    yset   = np.zeros((nTrace, nx))
    yerror = np.zeros((nTrace, nx))
#===============this two function is the python document for
#                    the c_traceAllFiber function(if i have made it...)
#def traceAllFiberPython():
    for i in range(nTrace):
        yPeak = yPeaks[i]
        #!! get accurate  center for one fiber(profile)
        yFit, yErr = traceOneCenterByFlux(image, ivar, initX, yPeak)
        print()
        print(i, yPeak, yFit, yErr)
        shift = yFit - yPeak;
        if -maxshift<shift<maxshift and yErr < maxerr:
            yset[i, initX] = yFit
            yerror[i, initX]= yErr
        else:
            yset[i, initX] = yPeak
            yerror[i, initX]=np.inf
        for x in range(initX-1, -1, -1):# trace for initX-1 ~ 0
            yPeak = yset[i, x+1]
            yFit, yErr = traceOneCenterByFlux(image, ivar, x, yPeak)
            shift = yFit - yPeak;
            if -maxshift<shift<maxshift and yErr < maxerr:
                yset[i, x] = yFit
                yerror[i, x]= yErr
            else:
                yset[i, x] = yPeak
                yerror[i, x]=np.inf
            sys.stdout.write("tracing fiber:{}/{} x:{}\r".format(i,nTrace,x))
            sys.stdout.flush()
        for x in range(initX+1, nx):# trace for initX+1 ~ nx-1
            yPeak = yset[i, x-1]
            yFit, yErr = traceOneCenterByFlux(image, ivar, x, yPeak)
            shift = yFit - yPeak;
            if -maxshift<shift<maxshift and yErr < maxerr:
                yset[i, x] = yFit
                yerror[i, x]= yErr
            else:
                yset[i, x] = yPeak
                yerror[i, x]=np.inf
            sys.stdout.write("tracing fiber:{}/{} x:{}\r".format(i,nTrace,x))
            sys.stdout.flush()
    return yset, yerror

def traceOneCenterByFlux(image, ivar, x, ycen, radius=3, configs={}):
    ny, nx = image.shape
    y1 = ycen - radius + 0.5
    y2 = ycen + radius + 0.5
    iy1 = int(np.floor(y1))
    iy2 = int(np.floor(y2))

    fullpix = int(iy2-iy1)
    iiLen   = fullpix+3
    ii     = np.arange(iiLen) #!! array 0~iiLen-1
    qbad   = np.zeros(iiLen) #!! array 0 * iiLen

    ih     = iy1 - 1 + ii; #!! array
    assert np.all(np.logical_and(0<ih, ih<ny))
    ydiff  = ih - ycen
    wt     = (radius - abs(ydiff) + 0.5); wt[wt<0] = 0; wt[wt>1] = 1
    sumw   = image[ih, x] * wt
    sumwt  = wt
    sumxw  = image[ih, x] * wt * ydiff
    varTerm= wt**2 / (ivar[ih, x] + (ivar[ih, x] < 1e-20))
    sumsx2 = varTerm
    sumsx1 = ydiff**2 * varTerm
    qbad   = ivar[ih, x] < 1e-20
    #print sumw; print sumwt; print sumxw; print varTerm; print sumsx2; print sumsx1;
    sumw  = sumw.sum()
    sumwt = sumwt.sum()
    sumxw = sumxw.sum()
    sumsx2= sumsx2.sum()
    sumsx1= sumsx1.sum()
    qbad  = qbad.sum()
    yNew  = ycen
    bad = np.abs(yNew-ycen)>radius+0.5 or ycen < radius-0.5 or ycen > ny-0.5-radius
    if sumw>0 and qbad==0 and not bad:# good
        delta = sumxw/sumw
        yNew = delta + yNew
        yErr = np.sqrt(sumsx1 + sumsx2 * delta**2)/sumw
        return yNew, yErr
    else:
        return ycen, np.inf
#=======================================================

def traceCenter(image, error, configs={}):
    #region: setup
    assert image.shape==error.shape
    ny, nx = image.shape
    initX = configs.get('initX', int(nx/2)) #!! start from middle column
    nFiber = configs.get('nFiber', 128)
    threshFactor = configs.get('threshFactor', 60)
    nMedian= configs.get('nMedian', 7)
    threshPercent  = float(configs.get('threshPercent', 50))
    infile = configs['infile']
    logger = configs["logger"]
    traceTemplate = configs['traceTemplate']
    outputPath = configs.get("outputPath", ".")
    #!! use the middle "band" to get the center positions
    xLow = initX-int(nMedian/2)
    xHigh= initX+int(nMedian/2)
    subivar= 1/error[:, xLow:xHigh]**2
    #endregion: setup
    #region: getYpeak
    kernel = np.array([-0.25,-0.25,0.25,0.50,0.25,-0.25,-0.25])
    yValues             = np.nanpercentile(image[:,xLow:xHigh], threshPercent, axis=1)
    assert np.all(np.logical_and(-2000<yValues,yValues<65535))
    #!! thresh
    thresh              = np.nanpercentile(yValues, threshFactor)
    yValueSmooth        = np.convolve(yValues, kernel)[3:-3]
    yValueUpValue       = np.zeros(ny);
    yValueUpValue[0]    = -1; yValueUpValue[1:] = (yValueSmooth[1:]-yValueSmooth[0:-1])
    yValueUp            = yValueUpValue>0
    yValueDownValue     = np.zeros(ny);
    yValueDownValue[-1] = -1; yValueDownValue[:-1] = (yValueSmooth[:-1]-yValueSmooth[1:])
    yValueDown          = yValueDownValue>0
    #!! local max and > thresh
    yPeaksMask          = (yValueUp & yValueDown & (yValueSmooth>thresh))
    yPeaksArray = np.where(yPeaksMask)[0]
    diffs      = - 0.5 + yValueUpValue[yPeaksArray] / (yValueUpValue[yPeaksArray]+yValueDownValue[yPeaksArray])

    #de.plt.switch_backend("Agg")
    fig = de.f();
    de.p(yValueSmooth, "r");
    de.p(yValues, "g");
    de.p([0, yValueSmooth.size], [thresh, thresh])
    de.gca().set_title("extract information")
    de.d(fig, os.path.join(outputPath, "tracing.fig"))
    fig.clear()
    de.plt.close(fig)
    #de.plt.switch_backend(configs["plotBackend"])

    yPeaksArray = yPeaksArray + diffs
    #yPeaksArrayValue = (yValueSmooth[np.cile(yPeaksArraky).asdtype(np.int)] * (0.5+diffs) +
    #                    yValueSmooth[np.floor(yPeaksArraky).asdtype(np.int)] * (0.5-diffs))
    #endregion: getYpeak
    #region: test traceTemplate
    inputDir = outputPath
    tracingInfoDir = os.path.join(inputDir, 'tracingInfo')
    tracingInfoFile = os.path.join(tracingInfoDir, 'tracingInfoFile.txt')
    #!! output debug pictures
    safe_mkdir(tracingInfoDir)
    fig = plt.figure('')
    ax  = plt.subplot(1,1,1)
    ax.set_title('initX={initX}, nMedian={nMedian}, threshPercent={threshPercent}\n threshFactor={threshFactor}'.format(initX=initX, nMedian=nMedian, threshPercent=threshPercent, threshFactor=threshFactor))
    ax.set_xlabel('y pixel')
    ax.plot(yValues,      color='red', label="yValues")
    ax.plot(yValueSmooth, color='blue', label="smoothed yValues")
    ax.plot([0,nx],[thresh,thresh], '--', label="thresh")
    count = 0
    f = open(tracingInfoFile,'w')
    f.write('# fiberNum={}\n'.format(nFiber))
    f.write('[yPeaks]\n')
    f.write('yIndex  refId\n')
    for i, value in enumerate(yPeaksArray):
        ax.text(value, yValueSmooth[int(value)]*1.1, str(i+1), color = 'black')
        f.write('{:5}    {:5}\n'.format(value, i+1))
    f.close()
    nImg = 20
    limitLength = ny/float(nImg)
    for i in range(nImg):
        ax.set_xlim(i*limitLength*0.9, (i+1)*limitLength*1.1)
        outFigName = os.path.join(tracingInfoDir,
                          os.path.basename(infile) + '_' +str(i) + '.png')
        fig.savefig(outFigName)
    #!! raise error if template file not exists
    if not os.path.exists(traceTemplate):
        raise logger.Exception("It must be the first time that you trace the 2D spectrum for the instrument\nSo you should Modify the {}, copy it to {} and run the pipeline again!".format(tracingInfoFile, traceTemplate))
    #endregion: test traceTemplate
    #!! now we have the traceTemplate, align it using np.correlate
    ypeaksTemp = configTools.fieldLoad(traceTemplate)['yPeaks'][1:]
    if len(ypeaksTemp) != nFiber:
        raise logger.Exception("{} ypeaks in {}, not {}".format(len(ypeaksTemp), tracingInfoFile, nFiber))
    yPeaksArrayTemp = np.array(ypeaksTemp,
            dtype=[(str('yIndex'), np.float), (str('refId'), "S10")])['yIndex']
    #!! integer diff, very rough
    diff = crudeDiff(yPeaksArray, yPeaksArrayTemp, ny)
    yPeaksArrayTemp = yPeaksArrayTemp + diff # tobe debug for the sign!!!
    try:
        peakA, peakB, indexs = makePairs(yPeaksArray, yPeaksArrayTemp)
    except Exception as e:
        logger.p(str(e))
        raise logger.Exception("check {} and see what happend, maybe you should only change some parameters in the config files".format(tracingInfoDir))

    #!! get offset as a whole
    offset = np.mean(peakA-peakB)
    yPeaksArrayTemp += offset
    assert len(yPeaksArrayTemp)==nFiber
    return initX, yPeaksArrayTemp, indexs

#region: functions for tracing
def safe_mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)

def guessForMissingFiber(thisIndex, goodIndexs, allIndexs, blockSize, rawArray):
    if thisIndex in goodIndexs:#!! no need to guess
        return rawArray[thisIndex]
    xx = np.arange(blockSize)
    blockNum = int(thisIndex/blockSize)
    blockIndexs = blockSize * blockNum + xx
    goodIndexs = goodIndexs[np.logical_and(goodIndexs>=blockIndexs[0], goodIndexs<=blockIndexs[-1])]
    if len(goodIndexs)<2:
        raise Exception("have less than 2 good fiber in one block.... need debug")
    #!! first try to find 2 upper fiber
    upper = goodIndexs[goodIndexs > thisIndex]
    lower = goodIndexs[goodIndexs < thisIndex]
    if len(upper):
        upper_distance = upper[0] - thisIndex
    else:
        upper_distance = blockSize
    if len(lower):
        lower_distance = thisIndex - lower[0]
    else:
        lower_distance = blockSize
    if (upper_distance<=lower_distance) and (upper.size>=2):#!! use upper
        step = np.abs(np.mean(rawArray[upper[1]]-rawArray[upper[0]])/(upper[1]-upper[0]))
        return rawArray[upper[0]] + step * (thisIndex - upper[0])
    elif (upper_distance>=lower_distance) and (lower.size>=2):#!! use lower
        step = np.abs(np.mean(rawArray[lower[1]]-rawArray[lower[0]])/(lower[1]-lower[0]))
        return rawArray[lower[0]] + step * (thisIndex - lower[0])
    else:
        raise Exception("need debug here..")

def readTraceSet(outputPath, configs={}):
    filename = os.path.join(outputPath, 'traceset.fits')
    obj = pyfits.open(filename)
    yset = obj[0].data
    badFiber = obj[2].data.astype(np.bool)
    boundary = obj[3].data
    return yset, badFiber, boundary

def readFlux(filename, configs={}):
    obj = pyfits.open(filename)
    flux = obj["SCI"].data
    std = obj["std"].data
    #yset = obj["xcen"].data
    yset = obj["xcenSmooth"].data
    return flux, std, yset

def boxMean(image, nave):
    assert nave%2==1
    kernel = np.ones(nave)
    tail = int((nave-1)/2)
    imgCov  = np.convolve(image, kernel)[tail:-tail]
    return imgCov

def crudeDiff(peaksA, peaksB, ny):
    """ offset """
    maxLag = 20
    npad = 20
    dlag = 1
    nlag = int(maxLag/dlag)
    lagRange = np.linspace(0, 2*maxLag, 2*nlag + 1) - maxLag
    xx = np.arange(0,ny)
    #modA = specLib.buildLamp(peaksA, np.ones(peaksA.shape), ny, npad)
    #modB = specLib.buildLamp(peaksB, np.ones(peaksB.shape), ny, npad)
    modA = specLib.buildSmoothLamp(peaksA, np.ones(peaksA.shape), xx, 1.6)
    modB = specLib.buildSmoothLamp(peaksB, np.ones(peaksB.shape), xx, 1.6)
    corrval, lag = specLib.maxCorrelate(modA, modB, lagRange)
    #if np.abs(lag)>5:
        #print "too large {} offset!! check the warnings!!!".format(lag)
        #corrval, lag = specLib.maxCorrelate(modA, modB, lagRange, ind=-2)
        #print "use {} instead..".format(lag)
    return lag

def makePairs(todo, template):
    """ get index of each exist fiber"""
    N = todo.size
    todo = todo.reshape((N,1))
    template = np.tile(template,(N,1))
    diff = todo - template
    ind = np.argmin(np.abs(diff), axis=1)
    if len(np.unique(ind)) != len(ind):
        raise Exception("len(np.unique(ind)) != len(ind)")
    todo = todo.flatten(); template = template.flatten()
    value = template[ind]
    return todo.flatten(), value.flatten(), ind

def findPeaks(image, configs={}):
    n = image.size
    threshFactor = configs.get('threshFactor', 3)
    threshPercent= configs.get('threshPercent', 50)
    #region: get peaks
    #kernel = np.array([-0.25,-0.25,0.25,0.50,0.25,-0.25,-0.25])
    kernel = np.array([-0.25,-0.25,0.25,0.50,0.25,-0.25,-0.25])
    yValues             = image;
    thresh              = np.percentile(yValues, threshPercent) * threshFactor
    yValueSmooth        = np.convolve(yValues, kernel)[3:-3]
    yValueUpValue       = np.zeros(n);
    yValueUpValue[0]    = -1; yValueUpValue[1:] = (yValueSmooth[1:]-yValueSmooth[0:-1])
    yValueUp            = yValueUpValue>0
    yValueDownValue     = np.zeros(n);
    yValueDownValue[-1] = -1; yValueDownValue[:-1] = (yValueSmooth[:-1]-yValueSmooth[1:])
    yValueDown          = yValueDownValue>0
    yPeaksMask          = (yValueUp & yValueDown & (yValueSmooth>thresh))
    PeakValues          = image[yPeaksMask]
    yPeaksArray = np.where(yPeaksMask)[0]
    diffs      = - 0.5 + yValueUpValue[yPeaksArray] / (yValueUpValue[yPeaksArray]+yValueDownValue[yPeaksArray])
    yPeaksArray = yPeaksArray + diffs
    return yPeaksArray, PeakValues
    #endregion: get peaks
#endregion: functions for tracing

def imageExtract(inputFile, configs={}):
    #region: setup
    outputPath = configs["outputPath"]
    outputFileName = os.path.basename(inputFile.split(".fits")[0]) + "Flux.fits"
    outfileM       = inputFile.split(".fits")[0] + "ExtractModel.fits"
    outfile2d      = inputFile.split(".fits")[0] + "Map2D.fits"
    outfileR       = inputFile.split(".fits")[0] + "ExtractModel.reg"
    outfileRO      = inputFile.split(".fits")[0] + ".reg"

    configs.update({"config2show":["rerun", 'maxMuDiff', "offset"]})
    genReg = configs.get("genReg", True)
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return outfile2d

    #!! gen region file for different type
    filename = os.path.basename(outfileRO)
    if filename[0]=="d":
        outfilecb = os.path.join(outputPath, os.path.basename(outfileRO)[1:])
    elif filename[0]=="c":
        outfilecb = os.path.join(outputPath, os.path.basename(outfileRO)[2:])
    elif "masterFlat" in filename or "final" in filename:
        outfilecb = os.path.join(outputPath, os.path.basename(outfileRO))
    else:
        raise logger.Exception("unknown file type for {}".format(outfileRO))
    #shutil.copy(os.path.join(outputPath, "masterFlat.reg"), outfileR)

    fiberInfo = configs.get("fiberInfo", None)
    fiberHalfWidth = configs.get("fiberHalfWidth", 5)
    makeRegs = configs.get("makeRegs", 1)
    extractSigma = configs.get("extractSigma", 2)
    waveDirection = configs.get('waveDirection', 'row')
    blockSize = configs.get("blockSize", 16)
    if waveDirection not in ['row', 'column']:
        raise Exception('invalid waveDirection!')

    obj = pyfits.open(inputFile)
    if configs.get("sci") is not None:
        data = configs["sci"].data; dataHeader = configs["sci"].header
        error = configs["std"].data; errorHeader = configs["std"].header
    else:
        data = obj["sci"].data; dataHeader = obj["sci"].header
        error = obj["std"].data; errorHeader = obj["std"].header
    nanErrorMask = np.isnan(error)
    error[nanErrorMask] = np.inf
    data[nanErrorMask] = 0
    zeroStdMask = error==0
    error[zeroStdMask] = np.inf
    data[zeroStdMask] = 0

    if waveDirection=='column':
        data = data.T.copy() # copy is necessary because we call c function latter
        error = error.T.copy()

    ny, nx = data.shape
    #endregion: setup
    #region: extract image
    thisConfig  = configs

    map2DLength = 5*data.shape[0]#!! Caution here!
    thisConfig["map2DLength"] = map2DLength
    if configs.get("yset") is not None:
        yset = configs["yset"]
        badFiber = configs["badFiber"]
    else:
        yset, badFiber, blockBoundary = readTraceSet(outputPath)

    badMask = (badFiber.astype(np.bool))
    blockNum = int(badMask.shape[0]/blockSize)
    goodNum = (~badMask).sum()
    extractSigma = configs.get("extractSigma", 2)
    maxMuDiff = configs.get("maxMuDiff", 0.6)
    offset = configs.get("offset", 0.0)

    #obj = pyfits.open(inputFile)
    #data = obj["sci"].data; dataHeader = obj["sci"].header
    #error = obj["std"].data; errorHeader = obj["std"].header
    originConfig = copy.copy(thisConfig)
    xcen = yset[~badMask]
    ysetForReg = np.zeros(yset.shape)

    objName = dataHeader["OBJECT"]
    if os.path.basename(inputFile)=="masterFlat.fits" or "twilight" in objName.lower():
        offset = 0.0

    xcen = xcen + offset
    thisConfig.update({"proftype":2, "map2DMaxHalfWidth":0, "maxMuDiff":maxMuDiff,
                       "prefix":"first", "file":os.path.basename(inputFile).split(".fits")[0]})

    info = extract(data, 1/(error**2), xcen, extractSigma, configs=thisConfig)
    mu = info["xcen"]
    tinvvar = info["invvar"]
    newXcen = np.zeros(xcen.shape)

    xx = np.arange(xcen.shape[1])
    for i in range(xcen.shape[0]):
        thisFit = Legendre.fit(xx, mu[i]-xcen[i], 4, w=tinvvar[i])
        newXcen[i,:] = xcen[i] + thisFit(xx)
    logger.p("\tfirst {} offset: {} + {}".format(os.path.basename(inputFile), (mu-xcen).mean(), offset), "ykb")
    logger.warn2file("\tfirst {} offset: {} + {}".format(os.path.basename(inputFile), (mu-xcen).mean(), offset))

    originConfig.update({"prefix":"second", "file":os.path.basename(inputFile).split(".fits")[0]})
    info = extract(data, 1/(error**2), newXcen, extractSigma, configs=originConfig)
    amod = info["amod"]
    flux = info["flux"]
    ivar = info["invvar"]
    sumFlux = info["sumFlux"]
    sumVar = info["sumVar"]
    bmod = info["bmod"]
    xcen = info["xcen"]
    newXcen = np.zeros(xcen.shape)
    xx = np.arange(xcen.shape[1])
    for i in range(xcen.shape[0]):
        thisFit = Legendre.fit(xx, mu[i]-xcen[i], 4, w=tinvvar[i])
        newXcen[i,:] = xcen[i] + thisFit(xx)
    logger.p("\tsecond {} offset: {} + {}".format(os.path.basename(inputFile), (mu-xcen).mean(), offset), "ykb")
    logger.warn2file("\tsecond {} offset: {} + {}".format(os.path.basename(inputFile), (mu-xcen).mean(), offset))

    ysetForReg[~badMask, :] = newXcen[:,:]
    if genReg:
        genTracingMarkers(outfileR, fiberInfo, ysetForReg, badFiber, np.where(~badFiber)[0])
        shutil.copy(outfileR, outfileRO)
        shutil.copy(outfileR, outfilecb)

    #fimageModel = info["fimageModel"]
    profileModel = info["profileModel"]
    fimageModelOrigin = info["fimageModelOrigin"]
    fimageModelFixWidth = info["fimageModelFixWidth"]
    #fimageModelHybird = info["fimageModelHybird"]
    map2D = info["map2D"]
    map2DOrigin = info["map2DOrigin"]
    reject2D = info["reject2D"]
    maxLength = info["maxLength"]

    map2d = map2D[:,:maxLength+1]
    map2dOrigin = map2DOrigin[:,:maxLength+1]
    #endregion: extract image
    #region: save files
    #!! save flux
    sheader = pyfits.Header()
    eheader = pyfits.Header()
    cheader = pyfits.Header()
    ccheader = pyfits.Header()
    rheader = pyfits.Header()
    aheader = pyfits.Header()
    iheader = pyfits.Header()
    sheader["OBJECT"]="extract flux"
    sheader["EXTNAME"] = "sumFlux"
    eheader["OBJECT"]="extract flux std"
    eheader["EXTNAME"] = "sumStd"
    cheader["OBJECT"]="extract xcen"
    cheader["EXTNAME"] = "xcen"
    ccheader["OBJECT"]="extract xcenSmooth"
    ccheader["EXTNAME"] = "xcenSmooth"
    rheader["OBJECT"]="extract reject"
    rheader["EXTNAME"] = "rej"
    aheader["OBJECT"]="extract A"
    aheader["EXTNAME"] = "SCI"
    iheader["OBJECT"]="extract std A"
    iheader["EXTNAME"] = "STD"
    data1 = pyfits.PrimaryHDU(flux, header=aheader)
    data2 = pyfits.ImageHDU(1/np.sqrt(ivar), header=iheader)
    data3 = pyfits.ImageHDU(xcen, header=cheader)
    data35 = pyfits.ImageHDU(newXcen, header=ccheader)
    data4 = pyfits.ImageHDU(reject2D, header=rheader)
    data5 = pyfits.ImageHDU(sumFlux, header=sheader)
    data6 = pyfits.ImageHDU(np.sqrt(sumVar), header=eheader)
    fits  = pyfits.HDUList([data1, data2, data3, data35, data4, data5, data6])
    fits.writeto(outFile, clobber='True')
    #!! save extract info
    #header1 = pyfits.Header()
    header2 = pyfits.Header()
    #header3 = pyfits.Header()
    #header4 = pyfits.Header()
    #header5 = pyfits.Header()
    header55 = pyfits.Header()
    header6 = pyfits.Header()
    header65 = pyfits.Header()
    header7 = pyfits.Header()
    header8 = pyfits.Header()
    header9 = pyfits.Header()
    #header10 = pyfits.Header()
    #header11 = pyfits.Header()
    #header12 = pyfits.Header()
    #header13 = pyfits.Header()
    #header1["OBJECT"] = "fimageModel"
    header2["OBJECT"] = "data"
    #header3["OBJECT"] = "fimageModel-data"
    #header4["OBJECT"] = "(fimageModel-data)/sigma"
    #header5["OBJECT"] = "(fimageModel-data)/amod*100."
    header6["OBJECT"] = "fimageModelOrigin"
    header65["OBJECT"] = "fimageModelFixWidth"
    header65["EXTNAME"] = "FIXEDIMAGE"
    header7["OBJECT"] = "fimageModelOrigin-data"
    header8["OBJECT"] = "(fimageModelOrigin-data)/sigma"
    header9["OBJECT"] = "(fimageModelOrigin-data)/amod*100."
    header55["OBJECT"] = "profileModel"
    #header10["OBJECT"] = "fimageModelHybird"
    #header11["OBJECT"] =  "fimageModelHybird-data"
    #header12["OBJECT"] = "(fimageModelHybird-data)/sigma"
    #header13["OBJECT"] = "(fimageModelHybird-data)/amod*100."
    #data1 = pyfits.PrimaryHDU(fimageModel, header=header1)
    data2 = pyfits.PrimaryHDU(data, header=header2)
    #data3 = pyfits.ImageHDU((fimageModel-data), header=header3)
    #data4 = pyfits.ImageHDU((fimageModel-data)/error, header=header4)
    #data5 = pyfits.ImageHDU((fimageModel-data)/data*100., header=header5)
    data6 = pyfits.ImageHDU(fimageModelOrigin, header=header6)
    data65 = pyfits.ImageHDU(fimageModelFixWidth, header=header65)
    data7 = pyfits.ImageHDU((fimageModelOrigin-data), header=header7)
    data8 = pyfits.ImageHDU((fimageModelOrigin-data)/error, header=header8)
    data9 = pyfits.ImageHDU((fimageModelOrigin-data)/data*100., header=header9)
    data55 = pyfits.ImageHDU(profileModel, header=header55)
    #data10 =  pyfits.ImageHDU(fimageModelHybird, header=header10)
    #data11 = pyfits.ImageHDU((fimageModelHybird-data), header=header11)
    #data12 = pyfits.ImageHDU((fimageModelHybird-data)/error, header=header12)
    #data13 = pyfits.ImageHDU((fimageModelHybird-data)/data*100., header=header13)
    #fits  = pyfits.HDUList([data1, data2, data3, data4, data5, data55, data6,
                            #data7, data8, data9, data10, data11, data12, data13])
    fits  = pyfits.HDUList([data2, data6, data65, data7, data8, data9, data55])
    fits.writeto(outfileM, clobber='True')
    #!! save 2dDebug
    header1 = pyfits.Header()
    header2 = pyfits.Header()
    header3 = pyfits.Header()
    header1["OBJECT"] =  "MAP2D"
    header1["EXTNAME"] = "MAP2D"
    header2["OBJECT"] =  "bmod"
    header2["EXTNAME"] = "bmod"
    header3["OBJECT"] =  "MAP2DOrigin"
    header3["EXTNAME"] =  "MAP2DO"
    #!! fimageModelOrigin = map2DModel(bmod, map2DOrigin, flux, ivar)
    data1 = pyfits.PrimaryHDU(map2d, header=header1)
    data2 = pyfits.ImageHDU(bmod, header=header2)
    data3 = pyfits.ImageHDU(map2dOrigin, header=header3)
    fits  = pyfits.HDUList([data1, data2, data3])
    fits.writeto(outfile2d, clobber='True')
    #endregion: save files
    return outfile2d

def extract(fimage, invvar, xcen, sigma, configs={}):
    nFiber, nx = xcen.shape
    xcenShape = nFiber, nx
    flux = np.zeros(xcenShape)
    ivar = np.zeros(xcenShape)
    sumFlux = np.zeros(xcenShape)
    sumBack = np.zeros(xcenShape)
    sumVar = np.zeros(xcenShape)
    reject2D = np.zeros(xcenShape)
    mu = np.zeros(xcenShape)
    amod = np.zeros(fimage.shape)
    bmod = np.zeros(fimage.shape)
    map2DLength = configs.get("map2DLength", 5*fimage.shape[1])#!! Caution here!

    configs["map2DLength"] = map2DLength
    map2D = np.zeros((nx, map2DLength))
    map2DOrigin = np.zeros((nx, map2DLength))
    objHeader = configs.get("header")
    debugIndex = configs.get("debugIndex", [])
    prefix = configs.get("prefix", "????")
    filename = configs.get("file", "?????")

    np.set_printoptions(precision=5)
    for xInd in range(nx):
        # amod only contains A and background
        configs["xInd"] = xInd
        #!! Here to change extract function!!
        info=\
            specLib.extract(fimage[:, xInd], invvar[:, xInd],
                    xcen[:,xInd], sigma, configs=configs)
        amod[:,xInd] = info["amod"][:]
        flux[:,xInd] = info["flux"][:]
        ivar[:,xInd] = info["invvar"][:]
        bmod[:,xInd] = info["bmod"][:]
        mu[:,xInd] = info["xcen"][:]
        sumFlux[:,xInd] = info["sumFlux"][:]
        sumBack[:,xInd] = info["sumFluxBack"][:]
        sumVar[:,xInd] = info["sumVar"][:]
        reject2D[:,xInd] = info["reject2D"][:]
        map2D[xInd,:] = info["map2D"][:]
        map2DOrigin[xInd,:] = info["map2DOrigin"][:]
        reducedStr = "".join(str(np.array(info["reducedChiOut"])).split("\n"))
        toPrint = "\t{} extract for {:15} xIndex:{:5} niter:{:3} reducedChi:{}".format(
                prefix, filename, xInd, info["niter"], reducedStr)
        sys.stdout.write(toPrint+'\r')
        sys.stdout.flush()
        #print(toPrint)
    print()
    # the struct of map2d
    # 0: total length
    maxLength = int(np.max(map2D[:,0]))
    #fimageModel = map2DModel(bmod, map2D, flux, ivar) #!! old fimageModel build
    sumFlux[~np.isfinite(sumVar)] = 0.0
    #fimageModel = map2DModel(bmod, inMap2D, sumFlux, 1/sumVar)
    fimageModel       = map2DModel(bmod, map2D, sumFlux, 1/sumVar)
    fimageModelOrigin = map2DModel(bmod, map2DOrigin, flux, ivar)
    #fimageModelHybird = map2DModel(bmod, map2D, flux, ivar)
    fimageModelFixWidth = map2DModelFixNaNWidth(bmod, map2DOrigin, flux, ivar, xcen)
    profileModel = map2DModel(np.zeros(bmod.shape), map2DOrigin, np.ones(flux.shape), np.ones(flux.shape))

    #return amod, flux, ivar, bmod, fimage, fimageModel, map2D, maxLength #!! old return
    # "fimageModelHybird":fimageModelHybird,
    return {"amod":amod, "sumFlux":sumFlux, "sumVar":sumVar, "flux":flux, "profileModel":profileModel,
            "bmod":bmod, "fimageModel":fimageModel, "map2D":map2D, "fimageModelOrigin":fimageModelOrigin,
            "map2DOrigin":map2DOrigin, "fimageModelFixWidth": fimageModelFixWidth,
            "reject2D":reject2D, "maxLength":maxLength, "invvar":ivar, "xcen":mu}

def map2DModel(bmod, map2D, flux, invvar, badValue=0, badRange=3, width=0):
    """invvar is only used as mask"""
    ny, nx = bmod.shape
    result = np.zeros(bmod.shape)
    for xInd in range(nx):
        result[:, xInd] =\
            specLib.map2DSlice(bmod[:, xInd],
                    map2D[xInd, :], flux[:, xInd], invvar[:, xInd])#!! caution!! what is the meaning of width???
    return result

def map2DModelFixNaNWidth(bmod, map2D, flux, invvar, yset, badValue=np.nan, halfwidth=5):
    """invvar is only used as mask"""
    ny, nx = bmod.shape
    result = np.zeros(bmod.shape)
    for xInd in range(nx):
        result[:, xInd] =\
            specLib.map2DSliceFixNaNWidth(bmod[:, xInd],
                    map2D[xInd, :], flux[:, xInd], invvar[:, xInd], yset[:, xInd],
                    badValue, halfwidth)
    return result


#region: backup function
    #def map2DFixWidthRead(yset, flux, width):
    #   ny, nx = flux.shape
    #   nTrace = yset.shape[0]
    #   nB = 2*width+1
    #   result = np.zeros((nTrace*nB,nx))
    #   for xInd in range(nx):
    #       result[:, xInd] = \
    #           specLib.map2DFixWidthRead(nTrace, yset[:, xInd], flux[:, xInd], width)
    #   return result
    #
    #def map2DFixWidthWrite(yset, flux, result, width):
    #   ny, nx = result.shape
    #   nTrace = yset.shape[0]
    #   nB = 2*width+1
    #   assert flux.size == nx*nTrace*nB
    #   result = result.copy()
    #   for xInd in range(nx):
    #       result[:, xInd] = \
    #           specLib.map2DFixWidthWrite(nTrace, yset[:, xInd], flux[:, xInd], result[:, xInd], width)
    #   return result
    #
    #def buildMasterProfile(fileList, outputfileNameRaw, configs={}):
    #   debugDir = outputfileNameRaw+"-debug"
    #   outputFileName= outputfileNameRaw + '.fits'
    #   configs.update({"name":outputfileNameRaw})
    #   configs.update({"config2show":["rerun", "badFactor", "parallel", "name", "goodSNRMean"]})
    #   exec(messager.initLoggerMacro)
    #   exec(m2fsConfigs.singleFileMacro)
    #   if toReturn: return
    #
    #   safe_mkdir(os.path.join(outputPath, debugDir))
    #   arrayList = [pyfits.open(each)[0].data for each in fileList]
    #   masterProfile = specLib.buildMasterProfile(arrayList, configs=configs)
    #
    #   data1 = pyfits.PrimaryHDU(masterProfile, pyfits.Header([("OBJECT","master profile")]))
    #   fits  = pyfits.HDUList([data1])
    #   fits.writeto(outFile, clobber='True')
    #   assert np.all(np.isfinite(masterProfile))
    #
    #   configs["outputfileNameRaw"] = outputfileNameRaw
    #   configs.update({"debugPath":os.path.join(outputPath, debugDir)})
    #   obj = fitForProfile(masterProfile, configs=configs)
    #   pieceDump(obj, outputPath, outputfileNameRaw)
    #
    #def buildMasterProfileModel(modelFile, outputfileNameRaw, otherData, configs={}):
    #   outputPath = configs["outputPath"]
    #   outputFileName = outputfileNameRaw + "Model.fits"
    #   configs.update({"config2show":["rerun", "badFactor"]})
    #   exec(messager.initLoggerMacro)
    #   exec(m2fsConfigs.singleFileMacro)
    #   if toReturn: return
    #
    #   otherHDUList = [pyfits.open(each)["SCI"] for each in otherData]
    #   outfileR = outFile.split(".fits")[0] + ".reg"
    #   shutil.copy(os.path.join(outputPath, "masterFlat.reg"), outfileR)
    #   obj = pyfits.open(modelFile.split(".fits")[0] + "DirectFlux.fits")
    #   flux = obj["SCI"].data
    #   flux = np.ones(flux.shape)
    #   std = np.ones(flux.shape)
    #   map2DFile = modelFile.split(".fits")[0] + "DirectMap2D.fits"
    #   obj = pyfits.open(map2DFile)
    #   inMap2D = obj[0].data
    #   bmod = obj[1].data
    #   sys.stdout.write("loading master profile.....");sys.stdout.flush()
    #   masterProfile = pieceLoad(outputPath, outputfileNameRaw)
    #   print("load master profile ok")
    #   assert np.all(np.diff(inMap2D[:,1])==0)
    #   nTrace = int(inMap2D[0,1])
    #   nx = int(inMap2D.shape[0])
    #   #!! build map2d
    #   for x in range(nx):
    #       tIndex = 2
    #       for i in range(nTrace):
    #           thisLen = int(inMap2D[x, tIndex])
    #           thisCen = inMap2D[x, tIndex+1]
    #           xx = np.arange(int(inMap2D[x, tIndex+2]), int(inMap2D[x, tIndex+3])+1)
    #           if masterProfile[x, i] is None:
    #               todoProfile = np.zeros(xx.shape)
    #           else:
    #               todoProfile = masterProfile[x, i](xx-thisCen)[:]
    #           todoProfile[todoProfile<0]=0
    #           if np.any(~np.isfinite(todoProfile)):
    #               print("!!!!!!!!!!!!!!!!!!!!!!!!!!")
    #           inMap2D[x,tIndex+8:tIndex+8+thisLen] = todoProfile[:]
    #           tIndex += 8+thisLen
    #   fimageModel = map2DModel(bmod, inMap2D, flux, 1/(std**2))
    #
    #   header = pyfits.Header()
    #   header["OBJECT"]="model profile"
    #   data = [pyfits.PrimaryHDU(fimageModel, header=header)]
    #   if len(otherHDUList):
    #       data += otherHDUList
    #   fits = pyfits.HDUList(data)
    #   fits.writeto(outFile, clobber='True')
    #
    #def getMap2DInfo(inMap2D):
    #   assert np.all(np.diff(inMap2D[:,1])==0)
    #   nTrace = int(inMap2D[0,1])
    #   nx = int(inMap2D.shape[0])
    #
    #
    #   length = np.zeros((nTrace, nx))
    #   xmin = np.zeros((nTrace, nx))
    #   xmax = np.zeros((nTrace, nx))
    #   xcen = np.zeros((nTrace, nx))
    #
    #   for x in range(nx):
    #       tIndex = 2
    #       for i in range(nTrace):
    #           length[i,x] = thisLen = int(inMap2D[x, tIndex])
    #           xcen[i,x] = thisCen = inMap2D[x, tIndex+1]
    #           xmin[i,x] = int(inMap2D[x, tIndex+2])
    #           xmax[i,x] = int(inMap2D[x, tIndex+3])
    #           tIndex += 8+thisLen
    #   return length, xmin, xmax, xcen
    #
    #def masterProfileInfo(masterProfile):
    #   r = dict()
    #   r["A"] = [each.A]
    #
    #class myProfile():
    #   def __init__(self, mu=0, sigma=1.4, A=0.1,
    #                      sigma3=1.4, A3=0.1):
    #       self.A = A
    #       self.mu = mu
    #       self.sigma = sigma
    #       self.A3 = A3
    #       self.sigma3 = sigma3
    #       self.c = np.sqrt(2*np.pi)
    #
    #   def f(self, x, A, mu, sigma, A3, sigma3):
    #       newX = np.tile(x, (5,1)) + np.array([-0.4, -0.2, 0, 0.2, 0.4]).reshape((5,1))
    #       diff = (newX-mu)/sigma
    #       diff3 = (newX-mu)/sigma3
    #       diffabs = np.abs(diff3)
    #       epow = diff*diff
    #       epow3 = diff3*diff3*diffabs
    #       r = A/(self.c*sigma) * np.exp(-epow/2)
    #       r3 = A3/(2.575798634 * sigma3) * np.exp(-epow3/3)
    #       return r.sum(axis=0) + r3.sum(axis=0)
    #
    #   def fit(self, x, y, w):
    #       popt, pcov = curve_fit(
    #               self.f, x, y, sigma=w,
    #               p0=[self.A, self.mu, self.sigma,
    #                   self.A3, self.sigma3],
    #               bounds=[(0, -0.2, 1.0, 0, 1.0),
    #                       (1,  0.2, 2.5, 1, 2.5)])
    #       self.A, self.mu, self.sigma, self.A3, self.sigma3 = popt
    #   def __call__(self, x):
    #       return self.f(x, self.A, self.mu, self.sigma, self.A3, self.sigma3)
    #
    #def fitForProfile(masterProfile, configs={}):
    #   ny = int(masterProfile.shape[0])
    #   nx = int(masterProfile[0,1])
    #   badFactor = configs.get("badFactor", 5e-5)
    #   debug = configs.get("debug", False)
    #   outputfileNameRaw = configs["outputfileNameRaw"]
    #
    #   parallel = configs.get("parallel", False)
    #   logger = configs.pop("logger")
    #   temp = configs.pop("teeStderr")
    #   de.plt.switch_backend("Agg")
    #   if parallel:
    #       N = ny
    #       parallel = configs.get("parallel", False)
    #       parallelN= configs.get("parallelN", 6)
    #       logger.p("\tparallel fit master profile use {} process".format(parallelN))
    #       num = int(N/parallelN)
    #       start = list(range(0, N, num))
    #       allSlices = [slice(start[i], start[i+1]) for i in range(len(start)-1)]
    #       if start[-1]<N:
    #           allSlices += [slice(start[-1], None)]
    #       cacheDir = configs.get("cacheDir","./cacheDir")
    #       safe_mkdir(cacheDir)
    #       outFiles = [os.path.join(cacheDir, "{}-{}.pkl".format(each, N)) for each in start]
    #       logger.p("\tallSlices:{}".format(allSlices))
    #       logger.p("\toutFiles:{}".format(outFiles))
    #
    #       count = parallelN
    #       pool = multiprocessing.Pool(processes=count)
    #       jobList = []
    #       for i in range(len(allSlices)):
    #           thisConfig = copy.deepcopy(configs)
    #           thisLogger = {"warningFile":thisConfig["warningFile"],
    #                     "errorFile":thisConfig["errorFile"],
    #                     "name":outputfileNameRaw + os.path.basename(outFiles[i]),
    #                     "initErrorLog":False,
    #                     "logFile":[os.path.join(thisConfig["logPath"], "parallel-log",
    #                                             outputfileNameRaw +
    #                                                 os.path.basename(outFiles[i])+"-profile.log"),
    #                                thisConfig["logFile"]],
    #                     "initWarningLog":False,"initLog":False,
    #                     "renewLog":[True,False], "renewError":False, "renewWarning":False,
    #                     }
    #
    #           thisProfile = masterProfile[allSlices[i]]
    #           thisDict = copy.copy(configs)
    #           thisDict["logger"] = thisLogger
    #           thisDict.update({"total":ny, "offset":start[i], "cacheFile": outFiles[i]})
    #           jobList.append(pool.apply_async(_fitForProfile, (thisProfile, thisDict)))
    #       pool.close()
    #       pool.join()
    #       print(jobList)
    #       result = [each.get() for each in jobList]
    #       allObjs = [np.load(each) for each in outFiles]
    #       profileObj = np.concatenate(allObjs)
    #       assert profileObj.shape[0]==ny
    #   else:
    #       configs.update({"total":ny})
    #       profileObj = _fitForProfile(masterProfile, configs=configs)
    #   #de.plt.switch_backend("MacOsX")
    #   de.plt.switch_backend(configs["plotBackend"])
    #
    #   return profileObj
    #
    #def _fitForProfile(masterProfile, configs={}):
    #   exec(messager.initLoggerMacro)
    #   ny = int(masterProfile.shape[0])
    #   nx = int(masterProfile[0,1])
    #   np.set_printoptions(precision=4)
    #   profileObj = np.empty((ny, nx), dtype=np.object)
    #   badFactor = configs.get("badFactor", 5e-5)
    #   badFactorPlot = configs.get("badFactorPlot", 5e-5)
    #   goodSNRMean = configs.get("goodSNRMean", 25)
    #   forceDoFit = configs.get("forceDoFit", False)
    #   debugPath = configs.get("debugPath")
    #   debug = configs.get("debug", False)
    #
    #   offset = configs.get("offset", 0)
    #   total  = configs.get("total", 0)
    #   cacheFile = configs.get("cacheFile", "")
    #   logger.p("working for {} with forceDoFit:{}".\
    #           format(cacheFile, forceDoFit))
    #   warnFlag = False
    #
    #   for y in range(ny):
    #       todo = masterProfile[y]
    #       p = 2;
    #       for x in range(nx):
    #           thisLen = int(todo[p])
    #           totalLen = thisLen + 40
    #           indStart = 20
    #           indEnd = 20 + thisLen
    #           xx = todo[p+1:p+1+thisLen]
    #           yy = todo[p+1+thisLen:p+1+thisLen*2]
    #           snr = todo[p+1+thisLen*2:p+1+thisLen*3]
    #           xmin=xx.min(); xmax=xx.max()
    #           step = 0.8
    #
    #           xxx = np.zeros(totalLen)
    #           yyy = np.zeros(totalLen)
    #           ssnnrr = np.ones(totalLen)*snr.max()
    #           xxx[:indStart] = np.linspace(xmin-indStart*step, xmin-step, indStart)
    #           xxx[indStart:indEnd] = xx[:]
    #           xxx[indEnd:] = np.linspace(xmax+step, xmax+(totalLen-indEnd)*step, (totalLen-indEnd))
    #           yyy[indStart:indEnd] = yy[:]
    #           ssnnrr[indStart:indEnd] = snr[:]
    #           snrMax = snr.max()
    #           if snrMax>goodSNRMean or forceDoFit:
    #               obj = myProfile()
    #               obj.snrmean = snrMax
    #               try:
    #                   obj.fit(xxx,yyy,1/ssnnrr)
    #               except Exception as e:
    #                   print(e)
    #                   ss = np.unique(ssnnrr.copy());ss.sort();ss=ss[::-1]
    #                   plotFitForProfile(xx, yy, obj, xInd=x, yInd=y+offset,
    #                         badness="bad-fit", outputPath=debugPath, snr=ss)
    #               goodmask = snr>0
    #               badness = ((obj(xx)[goodmask]-yy[goodmask])**2/snr[goodmask]).sum()
    #               if snrMax<goodSNRMean:
    #                   ss = np.unique(ssnnrr.copy());ss.sort();ss=ss[::-1]
    #                   #!! TODO: to use logger
    #                   if not warnFlag:
    #                       logger.warn("       fitting {:5}/{:5} {:5}/{:5} {:5}/{:5} bad snrMax:{}, caution..".format(y+offset, total, x, nx, y, ny, snrMax))
    #                       warnFlag = True
    #                   plotFitForProfile(xx, yy, obj, xInd=x, yInd=y+offset,
    #                         badness="bad-snr", outputPath=debugPath, snr=ss)
    #               if debug>1:
    #                   ss = np.unique(ssnnrr.copy());ss.sort();ss=ss[::-1]
    #                   plotFitForProfile(xx, yy, obj, xInd=x, yInd=y+offset,
    #                         badness=badness, outputPath=debugPath, snr=ss)
    #               elif debug>0:
    #                   ss = np.unique(ssnnrr.copy());ss.sort();ss=ss[::-1]
    #                   if not np.isfinite(badness) or badness>badFactorPlot:
    #                       plotFitForProfile(xx, yy, obj, xInd=x, yInd=y+offset,
    #                             badness=badness, outputPath=debugPath, snr=ss)
    #               obj.badness = badness
    #               profileObj[y,x] = obj
    #               p += 1+thisLen*3
    #           else:
    #               obj=None
    #           sys.stdout.write("       fitting {:5}/{:5} {:5}/{:5} {:5}/{:5}\r".format(y+offset, total, x, nx, y, ny))
    #           sys.stdout.flush()
    #   print()
    #
    #   logger.p("working done for "+cacheFile+" "*20)
    #   if cacheFile:
    #       profileObj.dump(cacheFile)
    #   else:
    #       return profileObj
    #
    #def plotFitForProfile(xx, yy, obj, xInd="?", yInd="?",
    #                     badness="-1", outputPath="./", snr=np.array([0])):
    #   xmin=xx.min()
    #   xmax=xx.max()
    #   fig = de.f()
    #   ax = fig.add_subplot(111)
    #   ax.scatter(xx,yy)
    #   xxx = np.linspace(xmin,xmax,100)
    #   ax.plot(xxx, obj(xxx))
    #   de.plt.title("[{}, {}] : {}\n{:10f} ({:10f},{:10f}) ({:10f}, {:10f})\nsnr:{}".format(\
    #           xInd, yInd, badness,
    #           obj.mu,
    #           obj.A, obj.sigma,
    #           obj.A3, obj.sigma3, snr))
    #   fig.tight_layout()
    #   ax.set_xlim([-10,10])
    #   if isinstance(badness, list):
    #       badness = badness[1]
    #   #!! 9999999999 means fit failed
    #   if isinstance(badness, float):
    #       fig.savefig(os.path.join(outputPath, "{}-{}-{:_^7.2g}.png".format(xInd, yInd, badness)))
    #   else:
    #       fig.savefig(os.path.join(outputPath, "{}-{}-{:}.png".format(xInd, yInd, badness)))
    #
    #   fig.clear()
    #   de.plt.close(fig)
    #
    #def imageExtract23(inputFiles, masterProfileName, masterProfileGood,configs={}):
    #   exec(messager.initLoggerMacro)
    #   rerun = configs.get("rerun")
    #   eachFileNames = [os.path.basename(each) for each in inputFiles]
    #   if not os.path.exists(inputFiles[0].split("fits")[0]+"Flux.fits") or rerun:
    #       logger.p("imageExtractUsingProfile:\n\t{}".format(eachFileNames))
    #
    #   outputPath = configs.get("outputPath")
    #   sys.stdout.write("loading master profile...");sys.stdout.flush()
    #   masterProfile = pieceLoad(outputPath, masterProfileName)
    #   sys.stdout.write("load master profile ok...");sys.stdout.flush()
    #   if masterProfileName!=masterProfileGood:
    #       sys.stdout.write("loading good master profile...");sys.stdout.flush()
    #       masterProfileGood = pieceLoad(outputPath, masterProfileName)
    #       sys.stdout.write("load good master profile ok...");sys.stdout.flush()
    #       assert masterProfile.shape==masterProfileGood.shape
    #       sys.stdout.write("refining master profile...");sys.stdout.flush()
    #       nx,ny = masterProfile.shape
    #       for x in range(nx):
    #           for y in range(ny):
    #               if masterProfile[x,y] is None:
    #                   masterProfile[x,y] = masterProfileGood[x,y]
    #       sys.stdout.write("refining master profile ok...");sys.stdout.flush()
    #   print()
    #
    #   yset, badFiber, blockBoundary = readTraceSet(outputPath)
    #   badMask = (badFiber.astype(np.bool))
    #   yset = yset[~badMask]
    #   nTrace, nx = yset.shape
    #   offset = np.zeros((nTrace, nx))
    #   sigma= np.zeros((nTrace, nx))
    #   sf   = np.zeros((nTrace, nx))
    #   oA   = np.zeros((nTrace, nx))
    #   oA3  = np.zeros((nTrace, nx))
    #   for x in range(nx):
    #       for nt in range(nTrace):
    #           obj = masterProfile[x, nt]
    #           offset[nt, x] = obj.mu
    #           sigma[nt, x] = obj.sigma
    #           sf[nt, x] = obj.sigma3/obj.sigma
    #           oA[nt, x] = obj.A
    #           oA3[nt, x] = obj.A3
    #
    #   in2DMapFiles = [each.split(".fits")[0]+"DirectMap2D.fits" for each in inputFiles]
    #   for eachFile,in2DMapFile in zip(inputFiles, in2DMapFiles):
    #       print("\t {} ~ {}".format(eachFile, in2DMapFile))
    #       inMap2D = pyfits.open(in2DMapFile)[0].data
    #       assert np.all(np.diff(inMap2D[:,1])==0)
    #       nTrace = int(inMap2D[0,1])
    #       nx = int(inMap2D.shape[0])
    #
    #       xcen = np.zeros((nTrace, nx))
    #       sys.stdout.write("get profile info...");sys.stdout.flush()
    #       for x in range(nx):
    #           tIndex = 2
    #           for i in range(nTrace):
    #               thisLen = int(inMap2D[x, tIndex])
    #               xcen[i,x] = inMap2D[x, tIndex+1]
    #               tIndex += 8+thisLen
    #       sys.stdout.write("all done\n");sys.stdout.flush()
    #       _imageExtract23(eachFile, xcen+offset, sigma, sf, oA, oA3, configs=configs)
    #
    #def _imageExtract23(inputFile, xcen, sigma, sf, oA, oA3, configs={}):
    #   #region: setup
    #   outputPath = configs["outputPath"]
    #   outputFileName = os.path.basename(inputFile.split(".fits")[0]) + "Flux.fits"
    #   outfileM       = inputFile.split(".fits")[0] + "ExtractModel.fits"
    #   outfile2d      = inputFile.split(".fits")[0] + "Map2D.fits"
    #   outfileR       = inputFile.split(".fits")[0] + "ExtractModel.reg"
    #   outfileR2d     = outfile2d.split(".fits")[0] + ".reg"
    #
    #   configs.update({"config2show":["rerun"]})
    #   exec(messager.initLoggerMacro)
    #   exec(m2fsConfigs.singleFileMacro)
    #   if toReturn: return outfile2d
    #
    #   shutil.copy(os.path.join(outputPath, "masterFlat.reg"), outfileR)
    #
    #   fiberInfo = configs.get("fiberInfo", None)
    #   fiberHalfWidth = configs.get("fiberHalfWidth", 5)
    #   makeRegs = configs.get("makeRegs", 1)
    #   extractSigma = configs.get("extractSigma", 2)
    #   waveDirection = configs.get('waveDirection', 'row')
    #   blockSize = configs.get("blockSize", 16)
    #   if waveDirection not in ['row', 'column']:
    #       raise Exception('invalid waveDirection!')
    #
    #   obj = pyfits.open(inputFile)
    #   data = obj["sci"].data; dataHeader = obj["sci"].header
    #   error = obj["std"].data; errorHeader = obj["std"].header
    #   if waveDirection=='column':
    #       data = data.T.copy() # copy is necessary because we call c function latter
    #       error = error.T.copy()
    #
    #   ny, nx = data.shape
    #   #endregion: setup
    #   #region: extract image
    #   thisConfig  = configs
    #
    #   map2DLength = 5*data.shape[0]#!! Caution here!
    #   thisConfig["map2DLength"] = map2DLength
    #   safe_mkdir(os.path.join(inputFile.split(".fits")[0] + "extractDebug"))
    #   thisConfig["path"] = os.path.join(inputFile.split(".fits")[0] + "extractDebug")
    #
    #   ind = 0
    #   info = extract23(data, 1/(error**2), xcen, sigma, sf, oA, oA3, configs=thisConfig)
    #   amod = info["amod"]
    #   flux = info["flux"]
    #   ivar = info["ivar"]
    #   sumFlux = info["sumFlux"]
    #   sumVar = info["sumVar"]
    #   bmod = info["bmod"]
    #   bmod = info["bmod"]
    #   fimageModel = info["fimageModel"]
    #   fimageModelOrigin = info["fimageModelOrigin"]
    #   fimageModelHybird = info["fimageModelHybird"]
    #   map2D = info["map2D"]
    #   map2DOrigin = info["map2DOrigin"]
    #   reject2D = info["reject2D"]
    #   maxLength = info["maxLength"]
    #
    #   map2d = map2D[:,:maxLength+1]
    #   map2dOrigin = map2DOrigin[:,:maxLength+1]
    #   #endregion: extract image
    #   #region: save files
    #   #!! save flux
    #   sheader = pyfits.Header()
    #   eheader = pyfits.Header()
    #   cheader = pyfits.Header()
    #   rheader = pyfits.Header()
    #   aheader = pyfits.Header()
    #   iheader = pyfits.Header()
    #   sheader["OBJECT"]="extract flux"
    #   sheader["EXTNAME"] = "SCI"
    #   eheader["OBJECT"]="extract flux std"
    #   eheader["EXTNAME"] = "STD"
    #   cheader["OBJECT"]="extract xcen"
    #   cheader["EXTNAME"] = "cen"
    #   rheader["OBJECT"]="extract reject"
    #   rheader["EXTNAME"] = "rej"
    #   aheader["OBJECT"]="extract A"
    #   aheader["EXTNAME"] = "A"
    #   iheader["OBJECT"]="extract std A"
    #   iheader["EXTNAME"] = "STDA"
    #   data1 = pyfits.PrimaryHDU(sumFlux, header=sheader)
    #   data2 = pyfits.ImageHDU(np.sqrt(sumVar), header=eheader)
    #   data3 = pyfits.ImageHDU(xcen, header=cheader)
    #   data4 = pyfits.ImageHDU(reject2D, header=rheader)
    #   data5 = pyfits.ImageHDU(flux, header=aheader)
    #   data6 = pyfits.ImageHDU(1/np.sqrt(ivar), header=iheader)
    #   fits  = pyfits.HDUList([data1, data2, data3, data4, data5, data6])
    #   fits.writeto(outFile, clobber='True')
    #   #!! save extract info
    #   header1 = pyfits.Header()
    #   header2 = pyfits.Header()
    #   header3 = pyfits.Header()
    #   header4 = pyfits.Header()
    #   header5 = pyfits.Header()
    #   header6 = pyfits.Header()
    #   header7 = pyfits.Header()
    #   header8 = pyfits.Header()
    #   header9 = pyfits.Header()
    #   header10 = pyfits.Header()
    #   header11 = pyfits.Header()
    #   header12 = pyfits.Header()
    #   header13 = pyfits.Header()
    #   header1["OBJECT"] = "fimageModel"
    #   header2["OBJECT"] = "data"
    #   header3["OBJECT"] = "fimageModel-data"
    #   header4["OBJECT"] = "(fimageModel-data)/sigma"
    #   header5["OBJECT"] = "(fimageModel-data)/amod*100."
    #   header6["OBJECT"] = "fimageModelOrigin"
    #   header7["OBJECT"] = "fimageModelOrigin-data"
    #   header8["OBJECT"] = "(fimageModelOrigin-data)/sigma"
    #   header9["OBJECT"] = "(fimageModelOrigin-data)/amod*100."
    #   header10["OBJECT"] = "fimageModelHybird"
    #   header11["OBJECT"] =  "fimageModelHybird-data"
    #   header12["OBJECT"] = "(fimageModelHybird-data)/sigma"
    #   header13["OBJECT"] = "(fimageModelHybird-data)/amod*100."
    #   data1 = pyfits.PrimaryHDU(fimageModel, header=header1)
    #   data2 = pyfits.ImageHDU(data, header=header2)
    #   data3 = pyfits.ImageHDU((fimageModel-data), header=header3)
    #   data4 = pyfits.ImageHDU((fimageModel-data)/error, header=header4)
    #   data5 = pyfits.ImageHDU((fimageModel-data)/data*100., header=header5)
    #   data6 = pyfits.ImageHDU(fimageModelOrigin, header=header6)
    #   data7 = pyfits.ImageHDU((fimageModelOrigin-data), header=header7)
    #   data8 = pyfits.ImageHDU((fimageModelOrigin-data)/error, header=header8)
    #   data9 = pyfits.ImageHDU((fimageModelOrigin-data)/data*100., header=header9)
    #   data10 =  pyfits.ImageHDU(fimageModelHybird, header=header10)
    #   data11 = pyfits.ImageHDU((fimageModelHybird-data), header=header11)
    #   data12 = pyfits.ImageHDU((fimageModelHybird-data)/error, header=header12)
    #   data13 = pyfits.ImageHDU((fimageModelHybird-data)/data*100., header=header13)
    #   fits  = pyfits.HDUList([data1, data2, data3, data4, data5, data6,
    #                           data7, data8, data9, data10, data11, data12, data13])
    #   fits.writeto(outfileM, clobber='True')
    #   #!! save 2dDebug
    #   header1 = pyfits.Header()
    #   header2 = pyfits.Header()
    #   header3 = pyfits.Header()
    #   header1["OBJECT"] =  "MAP2D"
    #   header1["EXTNAME"] = "MAP2D"
    #   header2["OBJECT"] =  "bmod"
    #   header2["EXTNAME"] = "bmod"
    #   header3["OBJECT"] =  "MAP2DOrigin"
    #   header3["EXTNAME"] =  "MAP2DO"
    #   data1 = pyfits.PrimaryHDU(map2d, header=header1)
    #   data2 = pyfits.ImageHDU(bmod, header=header2)
    #   data3 = pyfits.ImageHDU(map2dOrigin, header=header3)
    #   fits  = pyfits.HDUList([data1, data2, data3])
    #   fits.writeto(outfile2d, clobber='True')
    #   #endregion: save files
    #   return outfile2d
    #
    #def extract23(fimage, invvar, xcen, sigma, sf, oA, oA3, configs={}):
    #   nFiber, nx = xcen.shape
    #   xcenShape = nFiber, nx
    #   flux = np.zeros(xcenShape)
    #   ivar = np.zeros(xcenShape)
    #   sumFlux = np.zeros(xcenShape)
    #   sumBack = np.zeros(xcenShape)
    #   sumVar = np.zeros(xcenShape)
    #   reject2D = np.zeros(xcenShape)
    #   mu = np.zeros(xcenShape)
    #   amod = np.zeros(fimage.shape)
    #   bmod = np.zeros(fimage.shape)
    #   map2DLength = configs.get("map2DLength", 5*fimage.shape[1])#!! Caution here!
    #
    #   configs["map2DLength"] = map2DLength
    #   map2D = np.zeros((nx, map2DLength))
    #   map2DOrigin = np.zeros((nx, map2DLength))
    #   objHeader = configs.get("header")
    #   debugIndex = configs.get("debugIndex", [])
    #   name = configs.get("path", "?????")
    #   if len(debugIndex):
    #       logger.p("\tdebug extract using profile output for {} at index {}".\
    #               format(os.path.basename(name), debugIndex))
    #
    #   for xInd in range(nx):
    #       # amod only contains A and background
    #       configs["xInd"] = xInd
    #       if xInd in debugIndex:
    #           configs["doDebug"] = True
    #       else:
    #           configs["doDebug"] = False
    #       #!! Here to change extract function!!
    #       amod[:,xInd], flux[:,xInd], ivar[:,xInd], info=\
    #           specLib.extractDirect23(fimage[:, xInd], invvar[:, xInd],
    #                   xcen[:,xInd], sigma[:,xInd], sf[:,xInd], oA[:,xInd], oA3[:,xInd], configs=configs)
    #       bmod[:,xInd] = info["bmod"][:]
    #       sumFlux[:,xInd] = info["sumFlux"][:]
    #       sumBack[:,xInd] = info["sumFluxBack"][:]
    #       sumVar[:,xInd] = info["sumVar"][:]
    #       reject2D[:,xInd] = info["reject2D"][:]
    #       map2D[xInd,:] = info["map2D"][:]
    #       map2DOrigin[xInd,:] = info["map2DOrigin"][:]
    #       toPrint = "xIndex:{:5} niter:{:3} reducedChi:{:15.10}".format(xInd, info["niter"], info["reducedChi"])
    #       sys.stdout.write(toPrint+'\r')
    #       sys.stdout.flush()
    #       #print(toPrint)
    #   print()
    #   # the struct of map2d
    #   # 0: total length
    #   maxLength = int(np.max(map2D[:,0]))
    #   #fimageModel = map2DModel(bmod, map2D, flux, ivar) #!! old fimageModel build
    #   sumFlux[~np.isfinite(sumVar)] = 0.0
    #   #fimageModel = map2DModel(bmod, inMap2D, sumFlux, 1/sumVar)
    #   fimageModel       = map2DModel(bmod, map2D, sumFlux, 1/sumVar)
    #   fimageModelOrigin = map2DModel(bmod, map2DOrigin, flux, ivar)
    #   fimageModelHybird = map2DModel(bmod, map2D, flux, ivar)
    #
    #   #return amod, flux, ivar, bmod, fimage, fimageModel, map2D, maxLength #!! old return
    #   return {"amod":amod, "sumFlux":sumFlux, "sumVar":sumVar, "flux":flux,
    #           "bmod":bmod, "fimageModel":fimageModel, "fimageModelHybird":fimageModelHybird,
    #           "map2D":map2D, "fimageModelOrigin":fimageModelOrigin, "map2DOrigin":map2DOrigin,
    #           "reject2D":reject2D, "maxLength":maxLength, "ivar":ivar}
    #
    #def imageExtractUsingProfile(inputFiles, masterProfileName, masterProfileGood,configs={}):
    #   exec(messager.initLoggerMacro)
    #   rerun = configs.get("rerun")
    #   eachFileNames = [os.path.basename(each) for each in inputFiles]
    #   if not os.path.exists(inputFiles[0].split("fits")[0]+"Flux.fits") or rerun:
    #       logger.p("imageExtractUsingProfile:\n\t{}".format(eachFileNames))
    #
    #   outputPath = configs.get("outputPath")
    #   sys.stdout.write("loading master profile...");sys.stdout.flush()
    #   masterProfile = pieceLoad(outputPath, masterProfileName)
    #   sys.stdout.write("load master profile ok...");sys.stdout.flush()
    #   if masterProfileName!=masterProfileGood:
    #       sys.stdout.write("loading good master profile...");sys.stdout.flush()
    #       masterProfileGood = pieceLoad(outputPath, masterProfileName)
    #       sys.stdout.write("load good master profile ok...");sys.stdout.flush()
    #       assert masterProfile.shape==masterProfileGood.shape
    #       sys.stdout.write("refining master profile...");sys.stdout.flush()
    #       nx,ny = masterProfile.shape
    #       for x in range(nx):
    #           for y in range(ny):
    #               if masterProfile[x,y] is None:
    #                   masterProfile[x,y] = masterProfileGood[x,y]
    #       sys.stdout.write("refining master profile ok...");sys.stdout.flush()
    #   print()
    #
    #   in2DMapFiles = [each.split(".fits")[0]+"DirectMap2D.fits" for each in inputFiles]
    #   #region: second extract
    #   for eachFile,in2DMapFile in zip(inputFiles, in2DMapFiles):
    #       print("\t {} ~ {}".format(eachFile, in2DMapFile))
    #       #region: build inMap2D
    #       inMap2D = pyfits.open(in2DMapFile)[0].data
    #       assert np.all(np.diff(inMap2D[:,1])==0)
    #       nTrace = int(inMap2D[0,1])
    #       nx = int(inMap2D.shape[0])
    #
    #       length = np.zeros((nTrace, nx))
    #       xmin = np.zeros((nTrace, nx))
    #       xmax = np.zeros((nTrace, nx))
    #       xcen = np.zeros((nTrace, nx))
    #
    #       sys.stdout.write("get profile info...");sys.stdout.flush()
    #       for x in range(nx):
    #           tIndex = 2
    #           for i in range(nTrace):
    #               length[i,x] = thisLen = int(inMap2D[x, tIndex])
    #               xcen[i,x] = thisCen = inMap2D[x, tIndex+1]
    #               xmin[i,x] = thisxmin = int(inMap2D[x, tIndex+2])
    #               xmax[i,x] = thisxmax = int(inMap2D[x, tIndex+3])
    #               xx = np.arange(thisxmin, thisxmax+1)
    #               todoProfile = masterProfile[x, i](-xx+thisCen)
    #               todoProfile[todoProfile<0]=0
    #               if np.any(~np.isfinite(todoProfile)):
    #                   print("!!!!!!!!!!!!!!!!!!!!!!!!!!")
    #               inMap2D[x,tIndex+8:tIndex+8+thisLen] = todoProfile[:]
    #               tIndex += 8+thisLen
    #
    #       yset, badFiber, blockBoundary = readTraceSet(outputPath)
    #       badMask = (badFiber.astype(np.bool))
    #       yset = yset[~badMask]
    #       assert yset.shape[0]==nTrace
    #       newXcen = yset+((xcen-yset).mean(axis=1).reshape(nTrace, 1))
    #
    #       #sys.stdout.write("smooth x center...");sys.stdout.flush()
    #       #for x in range(nx):
    #       #    tIndex = 2
    #       #    for i in range(nTrace):
    #       #        thisLen = int(inMap2D[x, tIndex])
    #       #        thisxcen = inMap2D[x, tIndex+1]
    #       #        thisxmin = int(inMap2D[x, tIndex+2])
    #       #        thisxmax = int(inMap2D[x, tIndex+3])
    #       #        diff = newXcen[i, x] - thisxcen
    #       #        inMap2D[x, tIndex+1] = newXcen[i, x]
    #       #        inMap2D[x, tIndex+2] = int(thisxmin-diff)
    #       #        inMap2D[x, tIndex+3] = int(thisxmax-diff)
    #       #        tIndex += 8+thisLen
    #       sys.stdout.write("all done\n");sys.stdout.flush()
    #       #endregion: build inMap2D
    #       _imageExtractUsingProfile(eachFile, inMap2D, masterProfile, xcen, configs=configs)
    #   #endregion: second extract
    #
    #def _imageExtractUsingProfile(inputFile, inMap2D, masterProfile, xcen, configs={}):
    #   #region: setup
    #   outputPath = configs["outputPath"]
    #   outputFileName = os.path.basename(inputFile.split(".fits")[0]) + "Flux.fits"
    #   outfileM       = inputFile.split(".fits")[0] + "ExtractModel.fits"
    #   outfile2d      = inputFile.split(".fits")[0] + "Map2D.fits"
    #   outfileR       = inputFile.split(".fits")[0] + "ExtractModel.reg"
    #   outfileR2d     = outfile2d.split(".fits")[0] + ".reg"
    #
    #   configs.update({"config2show":["rerun"]})
    #   exec(messager.initLoggerMacro)
    #   exec(m2fsConfigs.singleFileMacro)
    #   if toReturn: return outfile2d
    #
    #   shutil.copy(os.path.join(outputPath, "masterFlat.reg"), outfileR)
    #
    #   fiberInfo = configs.get("fiberInfo", None)
    #   fiberHalfWidth = configs.get("fiberHalfWidth", 5)
    #   makeRegs = configs.get("makeRegs", 1)
    #   extractSigma = configs.get("extractSigma", 2)
    #   waveDirection = configs.get('waveDirection', 'row')
    #   blockSize = configs.get("blockSize", 16)
    #   if waveDirection not in ['row', 'column']:
    #       raise Exception('invalid waveDirection!')
    #   yset, badFiber, blockBoundary = readTraceSet(outputPath)
    #   badMask = (badFiber.astype(np.bool))
    #   blockNum = int(badMask.shape[0]/blockSize)
    #   goodNum = (~badMask).sum()
    #
    #   obj = pyfits.open(inputFile)
    #   data = obj["sci"].data; dataHeader = obj["sci"].header
    #   error = obj["std"].data; errorHeader = obj["std"].header
    #   if waveDirection=='column':
    #       data = data.T.copy() # copy is necessary because we call c function latter
    #       error = error.T.copy()
    #
    #   ny, nx = data.shape
    #   #endregion: setup
    #   #region: extract image
    #   thisConfig  = configs
    #
    #   map2DLength = 5*data.shape[0]#!! Caution here!
    #   thisConfig["map2DLength"] = map2DLength
    #   safe_mkdir(os.path.join(inputFile.split(".fits")[0] + "extractDebug"))
    #   thisConfig["path"] = os.path.join(inputFile.split(".fits")[0] + "extractDebug")
    #
    #   ind = 0
    #   info = extractUsingProfile(data, 1/(error**2), inMap2D, configs=thisConfig)
    #   amod = info["amod"]
    #   flux = info["flux"]
    #   ivar = info["ivar"]
    #   sumFlux = info["sumFlux"]
    #   sumVar = info["sumVar"]
    #   bmod = info["bmod"]
    #   bmod = info["bmod"]
    #   fimageModel = info["fimageModel"]
    #   fimageModelOrigin = info["fimageModelOrigin"]
    #   fimageModelHybird = info["fimageModelHybird"]
    #   map2D = info["map2D"]
    #   reject2D = info["reject2D"]
    #   maxLength = info["maxLength"]
    #
    #   map2d = map2D[:,:maxLength+1]
    #   #endregion: extract image
    #   #region: save files
    #   #!! save flux
    #   sheader = pyfits.Header()
    #   eheader = pyfits.Header()
    #   cheader = pyfits.Header()
    #   rheader = pyfits.Header()
    #   aheader = pyfits.Header()
    #   iheader = pyfits.Header()
    #   sheader["OBJECT"]="extract flux"
    #   sheader["EXTNAME"] = "SCI"
    #   eheader["OBJECT"]="extract flux std"
    #   eheader["EXTNAME"] = "STD"
    #   cheader["OBJECT"]="extract xcen"
    #   cheader["EXTNAME"] = "cen"
    #   rheader["OBJECT"]="extract reject"
    #   rheader["EXTNAME"] = "rej"
    #   aheader["OBJECT"]="extract A"
    #   aheader["EXTNAME"] = "A"
    #   iheader["OBJECT"]="extract std A"
    #   iheader["EXTNAME"] = "STDA"
    #   data1 = pyfits.PrimaryHDU(sumFlux, header=sheader)
    #   data2 = pyfits.ImageHDU(np.sqrt(sumVar), header=eheader)
    #   data3 = pyfits.ImageHDU(xcen, header=cheader)
    #   data4 = pyfits.ImageHDU(reject2D, header=rheader)
    #   data5 = pyfits.ImageHDU(flux, header=aheader)
    #   data6 = pyfits.ImageHDU(1/np.sqrt(ivar), header=iheader)
    #   fits  = pyfits.HDUList([data1, data2, data3, data4, data5, data6])
    #   fits.writeto(outFile, clobber='True')
    #   #!! save extract info
    #   header1 = pyfits.Header()
    #   header2 = pyfits.Header()
    #   header3 = pyfits.Header()
    #   header4 = pyfits.Header()
    #   header5 = pyfits.Header()
    #   header6 = pyfits.Header()
    #   header7 = pyfits.Header()
    #   header8 = pyfits.Header()
    #   header9 = pyfits.Header()
    #   header10 = pyfits.Header()
    #   header11 = pyfits.Header()
    #   header12 = pyfits.Header()
    #   header13 = pyfits.Header()
    #   header1["OBJECT"] = "fimageModel"
    #   header2["OBJECT"] = "data"
    #   header3["OBJECT"] = "fimageModel-data"
    #   header4["OBJECT"] = "(fimageModel-data)/sigma"
    #   header5["OBJECT"] = "(fimageModel-data)/amod*100."
    #   header6["OBJECT"] = "fimageModelOrigin"
    #   header7["OBJECT"] = "fimageModelOrigin-data"
    #   header8["OBJECT"] = "(fimageModelOrigin-data)/sigma"
    #   header9["OBJECT"] = "(fimageModelOrigin-data)/amod*100."
    #   header10["OBJECT"] = "fimageModelHybird"
    #   header11["OBJECT"] =  "fimageModelHybird-data"
    #   header12["OBJECT"] = "(fimageModelHybird-data)/sigma"
    #   header13["OBJECT"] = "(fimageModelHybird-data)/amod*100."
    #   data1 = pyfits.PrimaryHDU(fimageModel, header=header1)
    #   data2 = pyfits.ImageHDU(data, header=header2)
    #   data3 = pyfits.ImageHDU((fimageModel-data), header=header3)
    #   data4 = pyfits.ImageHDU((fimageModel-data)/error, header=header4)
    #   data5 = pyfits.ImageHDU((fimageModel-data)/data*100., header=header5)
    #   data6 = pyfits.ImageHDU(fimageModelOrigin, header=header6)
    #   data7 = pyfits.ImageHDU((fimageModelOrigin-data), header=header7)
    #   data8 = pyfits.ImageHDU((fimageModelOrigin-data)/error, header=header8)
    #   data9 = pyfits.ImageHDU((fimageModelOrigin-data)/data*100., header=header9)
    #   data10 =  pyfits.ImageHDU(fimageModelHybird, header=header10)
    #   data11 = pyfits.ImageHDU((fimageModelHybird-data), header=header11)
    #   data12 = pyfits.ImageHDU((fimageModelHybird-data)/error, header=header12)
    #   data13 = pyfits.ImageHDU((fimageModelHybird-data)/data*100., header=header13)
    #   fits  = pyfits.HDUList([data1, data2, data3, data4, data5, data6,
    #                           data7, data8, data9, data10, data11, data12, data13])
    #   fits.writeto(outfileM, clobber='True')
    #   #!! save 2dDebug
    #   header1 = pyfits.Header()
    #   header2 = pyfits.Header()
    #   header3 = pyfits.Header()
    #   header1["OBJECT"] =  "MAP2D"
    #   header1["EXTNAME"] = "MAP2D"
    #   header2["OBJECT"] =  "bmod"
    #   header2["EXTNAME"] = "bmod"
    #   header3["OBJECT"] =  "INMAP2D"
    #   header3["EXTNAME"] =  "INMAP2D"
    #   data1 = pyfits.PrimaryHDU(map2d, header=header1)
    #   data2 = pyfits.ImageHDU(bmod, header=header2)
    #   data3 = pyfits.ImageHDU(inMap2D, header=header3)
    #   fits  = pyfits.HDUList([data1, data2, data3])
    #   fits.writeto(outfile2d, clobber='True')
    #   #endregion: save files
    #   return outfile2d
    #
    #def extractUsingProfile(fimage, invvar, inMap2D, configs={}):
    #   assert np.all(np.diff(inMap2D[:,1])==0)
    #   nFiber = inMap2D[0,1]
    #   nx = fimage.shape[1]
    #   xcenShape = nFiber, nx
    #   flux = np.zeros(xcenShape)
    #   ivar = np.zeros(xcenShape)
    #   sumFlux = np.zeros(xcenShape)
    #   sumBack = np.zeros(xcenShape)
    #   sumVar = np.zeros(xcenShape)
    #   reject2D = np.zeros(xcenShape)
    #   mu = np.zeros(xcenShape)
    #   amod = np.zeros(fimage.shape)
    #   bmod = np.zeros(fimage.shape)
    #   map2DLength = configs.get("map2DLength", 5*fimage.shape[1])#!! Caution here!
    #
    #   configs["map2DLength"] = map2DLength
    #   map2D = np.zeros((nx, map2DLength))
    #   objHeader = configs.get("header")
    #   debugIndex = configs.get("debugIndex", [])
    #   name = configs.get("path", "?????")
    #   if len(debugIndex):
    #       logger.p("\tdebug extract using profile output for {} at index {}".\
    #               format(os.path.basename(name), debugIndex))
    #
    #   for xInd in range(nx):
    #       # amod only contains A and background
    #       configs["xInd"] = xInd
    #       if xInd in debugIndex:
    #           configs["doDebug"] = True
    #       else:
    #           configs["doDebug"] = False
    #       #!! Here to change extract function!!
    #       amod[:,xInd], flux[:,xInd], ivar[:,xInd], info=\
    #           specLib.extractUsingProfile(fimage[:, xInd], invvar[:, xInd],
    #                              inMap2D[xInd], configs=configs)
    #       bmod[:,xInd] = info["bmod"][:]
    #       sumFlux[:,xInd] = info["sumFlux"][:]
    #       sumBack[:,xInd] = info["sumFluxBack"][:]
    #       sumVar[:,xInd] = info["sumVar"][:]
    #       reject2D[:,xInd] = info["reject2D"][:]
    #       thisMap2d = info["map2D"]
    #       map2D[xInd,:] = thisMap2d[:]
    #       toPrint = "xIndex:{:5} niter:{:3} reducedChi:{:15.10}".format(xInd, info["niter"], info["reducedChi"])
    #       sys.stdout.write(toPrint+'\r')
    #       sys.stdout.flush()
    #       #print(toPrint)
    #   print()
    #   # the struct of map2d
    #   # 0: total length
    #   maxLength = int(np.max(map2D[:,0]))
    #   #fimageModel = map2DModel(bmod, map2D, flux, ivar) #!! old fimageModel build
    #   sumFlux[~np.isfinite(sumVar)] = 0.0
    #   #fimageModel = map2DModel(bmod, inMap2D, sumFlux, 1/sumVar)
    #   fimageModel = map2DModel(bmod, map2D, sumFlux, 1/sumVar)
    #   fimageModelOrigin = map2DModel(bmod, inMap2D, flux, ivar)
    #   fimageModelHybird = map2DModel(bmod, map2D, flux, ivar)
    #
    #   #return amod, flux, ivar, bmod, fimage, fimageModel, map2D, maxLength #!! old return
    #   return {"amod":amod, "sumFlux":sumFlux, "sumVar":sumVar, "flux":flux,
    #           "fimageModelHybird":fimageModelHybird, "bmod":bmod, "fimageModel":fimageModel,
    #           "map2D":map2D, "fimageModelOrigin":fimageModelOrigin,
    #           "reject2D":reject2D, "maxLength":maxLength, "ivar":ivar}
    #
    #def extractDirect(fimage, invvar, xcen, sigma, configs={}):
    #   map2DLength = configs.get("map2DLength", 5*fimage.shape[1])#!! Caution here!
    #   #region: first extract
    #   assert fimage.shape[1]==xcen.shape[1]
    #   nFiber, nx = xcen.shape
    #   mu = np.zeros(xcen.shape)
    #   tinvvar = np.zeros(xcen.shape)
    #
    #   configs["map2DLength"] = 0
    #   map2D = np.zeros((nx, map2DLength))
    #   objHeader = configs.get("header")
    #   debugIndex = configs.get("debugIndex", [])
    #   name = configs.get("path", "?????")
    #   if len(debugIndex):
    #       logger.p("\tdebug output for {} at index {}".\
    #               format(os.path.basename(name), debugIndex))
    #
    #   configs["doDebug"] = False
    #   configs["xInd"] = []
    #   for xInd in range(nx):
    #       # amod only contains A and background
    #       #!! Here to change extract function!!
    #       temp1, temp2, tinvvar[:,xInd], info=\
    #           specLib.extractDirect(fimage[:, xInd], invvar[:, xInd],
    #                              xcen[:, xInd], sigma, configs=configs)
    #       mu[:,xInd] = info["xcen"][:]
    #       #toPrint = "xIndex:{:5} niter:{:3} reducedChi:{:15.10}".format(xInd, info["niter"], info["reducedChi"])
    #       toPrint = "\tfirst extract: xIndex:{:5} niter:{:3} reducedChi:{:15.10} for {}".format(\
    #               xInd, info["niter"], info["reducedChi"],
    #               os.path.basename(configs["path"].split("DirectExtractDebug")[0]))
    #       sys.stdout.write(toPrint+'\r')
    #       sys.stdout.flush()
    #       #print(toPrint)
    #   print()
    #   #endregion: first extract
    #   #region: second extract
    #   #newXcen = xcen+((mu-xcen).mean(axis=1).reshape(nFiber, 1))
    #   newXcen = xcen.copy()
    #   xx = np.arange(xcen.shape[1])
    #   for i in range(xcen.shape[0]):
    #       thisFit = Legendre.fit(xx, mu[i]-xcen[i], 4, w=tinvvar[i])
    #       newXcen[i,:] = xcen[i] + thisFit(xx)
    #
    #   de.plt.switch_backend("Agg")
    #   path = configs.get("path",".")
    #   logger.p("\t{} offset: {}".format(name, (mu-xcen).mean()), "ykb")
    #   fig = de.f();
    #   de.p(mu.T);
    #   de.p(newXcen.T, "--");
    #   de.gca().set_title(path)
    #   de.d(fig, os.path.join(path, "allTrace.fig"))
    #   fig.clear()
    #   de.plt.close(fig)
    #   #de.plt.switch_backend("MacOsX")
    #   de.plt.switch_backend(configs["plotBackend"])
    #
    #   flux = np.zeros(xcen.shape)
    #   ivar = np.zeros(xcen.shape)
    #   sumFlux = np.zeros(xcen.shape)
    #   sumBack = np.zeros(xcen.shape)
    #   sumVar = np.zeros(xcen.shape)
    #   reject2D = np.zeros(xcen.shape)
    #   mu = np.zeros(xcen.shape)
    #   amod = np.zeros(fimage.shape)
    #   bmod = np.zeros(fimage.shape)
    #
    #   configs["map2DLength"] = map2DLength
    #   map2D = np.zeros((nx, map2DLength))
    #   objHeader = configs.get("header")
    #   debugIndex = configs.get("debugIndex", [])
    #   name = configs.get("path", "?????")
    #   if len(debugIndex):
    #       logger.p("\tdebug output for {} at index {}".\
    #               format(os.path.basename(name), debugIndex))
    #
    #   for xInd in range(nx):
    #       # amod only contains A and background
    #       configs["fitmask"]  = np.array([1, 1, 0])#!! only fit for sigma and A
    #       configs["xInd"] = xInd
    #       if xInd in debugIndex:
    #           configs["doDebug"] = True
    #       else:
    #           configs["doDebug"] = False
    #       #!! Here to change extract function!!
    #       amod[:,xInd], flux[:,xInd], ivar[:,xInd], info=\
    #           specLib.extractDirect(fimage[:, xInd], invvar[:, xInd],
    #                              newXcen[:, xInd], sigma, configs=configs)
    #       bmod[:,xInd] = info["bmod"][:]
    #       sumFlux[:,xInd] = info["sumFlux"][:]
    #       sumBack[:,xInd] = info["sumFluxBack"][:]
    #       sumVar[:,xInd] = info["sumVar"][:]
    #       reject2D[:,xInd] = info["reject2D"][:]
    #       mu[:,xInd] = info["xcen"][:]
    #       thisMap2d = info["map2D"]#!! this xcen had been smoothed
    #       map2D[xInd,:] = thisMap2d[:]
    #       toPrint = "\tsecond extract: xIndex:{:5} niter:{:3} reducedChi:{:15.10} for {}".format(\
    #               xInd, info["niter"], info["reducedChi"],
    #               os.path.basename(configs["path"].split("DirectExtractDebug")[0]))
    #       sys.stdout.write(toPrint+'\r')
    #       sys.stdout.flush()
    #       #print(toPrint)
    #   print()
    #   #endregion: second extract
    #   # the struct of map2d
    #   # 0: total length
    #   maxLength = int(np.max(map2D[:,0]))
    #   #fimageModel = map2DModel(bmod, map2D, flux, ivar) #!! old fimageModel build
    #   fimageModel = map2DModel(bmod, map2D, sumFlux, 1/sumVar)
    #
    #   #return amod, flux, ivar, bmod, fimage, fimageModel, map2D, maxLength #!! old return
    #   #return amod, sumFlux, sumVar, bmod, fimageModel, map2D, reject2D, maxLength, newXcen
    #   return {"amod": amod, "sumFlux":sumFlux, "sumVar":sumVar, "bmod":bmod,
    #           "fimageModel":fimageModel, "map2D":map2D, "flux":flux, "ivar":ivar,
    #           "reject2D":reject2D, "maxLength":maxLength, "xcen":newXcen}
    #
    #def imageExtractDirect(inputFile, configs={}):
    #   #region: setup
    #   outputPath = configs["outputPath"]
    #   outputFileName = os.path.basename(inputFile.split(".fits")[0]) + "DirectFlux.fits"
    #   outfileM       = inputFile.split(".fits")[0] + "DirectExtractModel.fits"
    #   outfile2d      = inputFile.split(".fits")[0] + "DirectMap2D.fits"
    #   outfileR       = inputFile.split(".fits")[0] + "DirectExtractModel.reg"
    #   outfileR2d     = outfile2d.split(".fits")[0] + ".reg"
    #
    #   configs.update({"config2show":["rerun"]})
    #   exec(messager.initLoggerMacro)
    #   exec(m2fsConfigs.singleFileMacro)
    #   if toReturn: return outfile2d
    #
    #   shutil.copy(os.path.join(outputPath, "masterFlat.reg"), outfileR)
    #
    #   fiberInfo = configs.get("fiberInfo", None)
    #   fiberHalfWidth = configs.get("fiberHalfWidth", 5)
    #   makeRegs = configs.get("makeRegs", 1)
    #   extractSigma = configs.get("extractSigma", 2)
    #   waveDirection = configs.get('waveDirection', 'row')
    #   blockSize = configs.get("blockSize", 16)
    #   if waveDirection not in ['row', 'column']:
    #       raise Exception('invalid waveDirection!')
    #   yset, badFiber, blockBoundary = readTraceSet(outputPath)
    #   badMask = (badFiber.astype(np.bool))
    #   blockNum = int(badMask.shape[0]/blockSize)
    #   goodNum = (~badMask).sum()
    #
    #   obj = pyfits.open(inputFile)
    #   data = obj["sci"].data; dataHeader = obj["sci"].header
    #   error = obj["std"].data; errorHeader = obj["std"].header
    #   if waveDirection=='column':
    #       data = data.T.copy() # copy is necessary because we call c function latter
    #       error = error.T.copy()
    #
    #   ny, nx = data.shape
    #   #endregion: setup
    #   #region: extract image
    #   thisConfig  = configs
    #
    #   map2DLength = 5*data.shape[0]#!! Caution here!
    #   thisConfig["map2DLength"] = map2DLength
    #   safe_mkdir(os.path.join(inputFile.split(".fits")[0] + "DirectExtractDebug"))
    #   thisConfig["path"] = os.path.join(inputFile.split(".fits")[0] + "DirectExtractDebug")
    #
    #   ind = 0
    #   try:
    #       info = extractDirect(data, 1/(error**2), yset[~badMask], extractSigma, configs=thisConfig)
    #   except Exception as e:
    #       raise logger.error("extract error in: {}".format(inputFile))
    #   amod = info["amod"]
    #   flux = info["flux"]
    #   ivar = info["ivar"]
    #   sumFlux = info["sumFlux"]
    #   sumVar = info["sumVar"]
    #   bmod = info["bmod"]
    #   fimageModel = info["fimageModel"]
    #   map2D = info["map2D"]
    #   reject2D = info["reject2D"]
    #   maxLength = info["maxLength"]
    #   xcen = info["xcen"]
    #
    #   map2d = map2D[:,:maxLength+1]
    #   #endregion: extract image
    #   #region: save files
    #   #!! save flux
    #   sheader = pyfits.Header()
    #   eheader = pyfits.Header()
    #   cheader = pyfits.Header()
    #   rheader = pyfits.Header()
    #   aheader = pyfits.Header()
    #   iheader = pyfits.Header()
    #   sheader["OBJECT"]="extract flux"
    #   sheader["EXTNAME"] = "SCI"
    #   eheader["OBJECT"]="extract flux std"
    #   eheader["EXTNAME"] = "STD"
    #   cheader["OBJECT"]="extract xcen"
    #   cheader["EXTNAME"] = "cen"
    #   rheader["OBJECT"]="extract reject"
    #   rheader["EXTNAME"] = "rej"
    #   aheader["OBJECT"]="extract A"
    #   aheader["EXTNAME"] = "A"
    #   iheader["OBJECT"]="extract std A"
    #   iheader["EXTNAME"] = "STDA"
    #   data1 = pyfits.PrimaryHDU(sumFlux, header=sheader)
    #   data2 = pyfits.ImageHDU(np.sqrt(sumVar), header=eheader)
    #   data3 = pyfits.ImageHDU(xcen, header=cheader)
    #   data4 = pyfits.ImageHDU(reject2D, header=rheader)
    #   data5 = pyfits.ImageHDU(flux, header=aheader)
    #   data6 = pyfits.ImageHDU(1/np.sqrt(ivar), header=iheader)
    #   fits  = pyfits.HDUList([data1, data2, data3, data4, data5, data6])
    #   fits.writeto(outFile, clobber='True')
    #   #!! save extract info
    #   header1 = pyfits.Header()
    #   header2 = pyfits.Header()
    #   header3 = pyfits.Header()
    #   header4 = pyfits.Header()
    #   header5 = pyfits.Header()
    #   header1["OBJECT"] = "fimageModel"
    #   header2["OBJECT"] = "data"
    #   header3["OBJECT"] = "fimageModel-data"
    #   header4["OBJECT"] = "(fimageModel-data)/sigma"
    #   header5["OBJECT"] = "(fimageModel-data)/amod*100."
    #   data1 = pyfits.PrimaryHDU(fimageModel, header=header1)
    #   data2 = pyfits.ImageHDU(data, header=header2)
    #   data3 = pyfits.ImageHDU((fimageModel-data), header=header3)
    #   data4 = pyfits.ImageHDU((fimageModel-data)/error, header=header4)
    #   data5 = pyfits.ImageHDU((fimageModel-data)/data*100., header=header5)
    #   fits  = pyfits.HDUList([data1, data2, data3, data4, data5])
    #   fits.writeto(outfileM, clobber='True')
    #   #!! save 2dDebug
    #   header1 = pyfits.Header()
    #   header2 = pyfits.Header()
    #   header1["OBJECT"] =  "MAP2D"
    #   header1["EXTNAME"] =  "MAP2D"
    #   header2["OBJECT"] =  "bmod"
    #   header2["EXTNAME"] =  "bmod"
    #   data1 = pyfits.PrimaryHDU(map2d, header=header1)
    #   data2 = pyfits.ImageHDU(bmod, header=header2)
    #   fits  = pyfits.HDUList([data1, data2])
    #   fits.writeto(outfile2d, clobber='True')
    #   #endregion: save files
    #   return outfile2d
    #
    #def pieceDump(obj, outputPath, name, num=500):
    #   N = obj.shape[0]
    #   start = list(range(0,N, num))
    #   allSlices = [slice(start[i], start[i+1]) for i in range(len(start)-1)]
    #   if start[-1]<N:
    #       allSlices += [slice(start[-1], None)]
    #   basePath = os.path.join(outputPath, name)
    #   for i, eachSlice in enumerate(allSlices):
    #       obj[eachSlice].dump(basePath+".pkl{:0>5}".format(i))
    #
    #def pieceLoad(outputPath, name):
    #   basePath = os.path.join(outputPath, name) + ".pkl"
    #   allFiles = glob.glob(basePath+"*")
    #   allFiles.sort()
    #   result = np.concatenate([np.load(each) for each in allFiles])
    #   return result
    #
    ##==================================================================================================
    ## old imageExtract
    #def imageExtractOld(inputFile, outfile=None, configs={}):
    #   #region: setup
    #   outputPath = configs["outputPath"]
    #   if outfile is None:
    #       outputFileName = inputFile.split(".fits")[0] + "Flux.fits"
    #       outfileM= inputFile.split(".fits")[0] + "ExtractModel.fits"
    #       outfile2d= inputFile.split(".fits")[0] + "Map2D.fits"
    #       outfileR= inputFile.split(".fits")[0] + "ExtractModel.reg"
    #       outfileR2d= outfile2d.split(".fits")[0] + ".reg"
    #   else:
    #       outputFileName = os.path.join(outputPath, outputFileName+"Flux.fits")
    #       outfileM= os.path.join(outputPath, outputFileName+"ExtractModel.fits")
    #       outfile2d= os.path.join(outputPath, outputFileName+"Map2D.fits")
    #       outfileR= inputFile.split(".fits")[0] + "ExtractModel.reg"
    #       outfileR2d= outfile2d.split(".fits")[0] + ".reg"
    #
    #   configs.update({"config2show":["rerun"]})
    #   exec(messager.initLoggerMacro)
    #   exec(m2fsConfigs.singleFileMacro)
    #   if toReturn: return
    #
    #   shutil.copy(os.path.join(outputPath, "masterFlat.reg"), outfileR)
    #   shutil.copy(os.path.join(outputPath, "masterFlat.reg"), outfileR2d)
    #
    #   fiberInfo = configs.get("fiberInfo", None)
    #   fiberHalfWidth = configs.get("fiberHalfWidth", 5)
    #   makeRegs = configs.get("makeRegs", 1)
    #   extractSigma = configs.get("extractSigma", 2)
    #   waveDirection = configs.get('waveDirection', 'row')
    #   blockSize = configs.get("blockSize", 16)
    #   if waveDirection not in ['row', 'column']:
    #       raise Exception('invalid waveDirection!')
    #   yset, badFiber, blockBoundary = readTraceSet(outputPath)
    #   badMask = (badFiber.astype(np.bool))
    #   blockNum = int(badMask.shape[0]/blockSize)
    #   goodNum = (~badMask).sum()
    #
    #   obj = pyfits.open(inputFile)
    #   data = obj["sci"].data; dataHeader = obj["sci"].header
    #   error = obj["std"].data; errorHeader = obj["std"].header
    #   if waveDirection=='column':
    #       data = data.T.copy() # copy is necessary because we call c function latter
    #       error = error.T.copy()
    #
    #   ny, nx = data.shape
    #   #endregion: setup
    #   #region: extract image
    #   thisConfig  = configs
    #
    #   map2DLength = 5*data.shape[0]#!! Caution here!
    #   thisConfig["map2DLength"] = map2DLength
    #
    #   amod = np.zeros((ny, nx))
    #   bmod = np.zeros((ny, nx))
    #   fimage = np.zeros((ny, nx))
    #   fimageModel = np.zeros((ny, nx))
    #   flux = np.zeros((goodNum, nx))
    #   ivar = np.zeros((goodNum, nx))
    #   map2d= np.zeros((goodNum, nx, map2DLength))
    #   allMaxLen = np.zeros((goodNum,))
    #
    #   ind = 0
    #   for i in range(blockNum):
    #       lower = blockBoundary[i][0]
    #       upper = blockBoundary[i][1]
    #       thisGoodMask = np.logical_not(badMask[i*blockSize:(i+1)*blockSize])
    #       thisGoodNum = thisGoodMask.sum()
    #       thisConfig["fiberNum"] = i
    #      # amod,  flux,  ivar,  bmod,  fimage,  fimageModel,  map2D,  maxLength
    #       tamod, tflux, tivar, tbmod, tfimage, tfimageModel, tmap2D, tmaxLength=\
    #               extractRow(data[lower:upper+1],
    #                       1/(error**2)[lower:upper+1],
    #                       yset[i*blockSize:(i+1)*blockSize][thisGoodMask]-lower,
    #                       extractSigma, configs=thisConfig)
    #       amod[lower:upper+1] = tamod[:]
    #       bmod[lower:upper+1] = tbmod[:]
    #       flux[ind:ind+thisGoodNum] = tflux[:]
    #       ivar[ind:ind+thisGoodNum] = tivar[:]
    #       fimage[lower:upper+1, :] = tfimage[:,:]
    #       fimageModel[lower:upper+1, :] = tfimageModel[:,:]
    #       map2d[i,:,:] = tmap2D[:,:]
    #       allMaxLen[i] = tmaxLength
    #       ind += thisGoodNum
    #   maxLength2D = allMaxLen.max()
    #   map2d = map2d[:,:,:maxLength2D+1]
    #   #endregion: extract image
    #   #region: save files
    #   #!! save flux
    #   header = pyfits.Header()
    #   header["comment"]="[0] flux"
    #   header["comment"]="[1] ivar"
    #   data1 = pyfits.PrimaryHDU(flux, header=header)
    #   data2 = pyfits.ImageHDU(ivar, header=header)
    #   fits  = pyfits.HDUList([data1, data2])
    #   fits.writeto(outputFileName, clobber='True')
    #   #!! save extract info
    #   header = pyfits.Header()
    #   header["comment"]="[0] amod"
    #   header["comment"]="[1] data"
    #   header["comment"]="[2] amod-data"
    #   header["comment"]="[3] (amod-data)/sigma"
    #   header["comment"]="[4] (amod-data)/amod*100."
    #   data1 = pyfits.PrimaryHDU(amod, header=header)
    #   data2 = pyfits.ImageHDU(data, header=header)
    #   data3 = pyfits.ImageHDU((amod-data), header=header)
    #   data4 = pyfits.ImageHDU((amod-data)/error, header=header)
    #   data5 = pyfits.ImageHDU((amod-data)/amod*100., header=header)
    #   fits  = pyfits.HDUList([data1, data2, data3, data4, data5])
    #   fits.writeto(outfileM, clobber='True')
    #   #!! save 2dDebug
    #   header = pyfits.Header()
    #   header["comment"]="[0] Map2D"
    #   header["comment"]="[1] bmod"
    #   header["comment"]="[2] fimage"
    #   header["comment"]="[3] fimageModel"
    #   header["comment"]="[4] fimageModel-fimage"
    #   header["OBJECT"] =  "MAP2D"
    #   data1 = pyfits.PrimaryHDU(map2d, header=header)
    #   header["OBJECT"] =  "bmod"
    #   data2 = pyfits.ImageHDU(bmod, header=header)
    #   header["OBJECT"] =  "fimage"
    #   data3 = pyfits.ImageHDU(fimage, header=header)
    #   header["OBJECT"] =  "fimageModel"
    #   data4 = pyfits.ImageHDU(fimageModel, header=header)
    #   header["OBJECT"] =  "fimageModel - fimage"
    #   data5 = pyfits.ImageHDU(fimageModel-fimage, header=header)
    #   fits  = pyfits.HDUList([data1, data2, data3, data4, data5])
    #   fits.writeto(outfile2d, clobber='True')
    #   #endregion: save files
    #   return
    ## old extract
    #def extractRow(fimage, invvar, xcen, sigma, configs={}):
    #   assert fimage.shape[1]==xcen.shape[1]
    #   nFiber, nx = xcen.shape
    #   flux = np.zeros(xcen.shape)
    #   ivar = np.zeros(xcen.shape)
    #   amod = np.zeros(fimage.shape)
    #   bmod = np.zeros(fimage.shape)
    #   #map2DLength = int(2 + fimage.shape[0] + 5*nFiber)
    #   map2DLength = configs.get("map2DLength", 5*fimage.shape[1])#!! Caution here!
    #
    #   configs["map2DLength"] = map2DLength
    #   map2D = np.zeros((nx, map2DLength))
    #   objHeader = configs.get("header")
    #
    #   for xInd in range(nx):
    #       # amod only contains A and background
    #       configs["xInd"] = xInd
    #       #!! Here to change extract function!!
    #       amod[:,xInd], flux[:,xInd], ivar[:,xInd], info=\
    #           specLib.extract(fimage[:, xInd], invvar[:, xInd],
    #                              xcen[:, xInd], sigma, configs=configs)
    #       bmod[:,xInd] = info["bmod"][:]
    #       thisMap2d = info["map2D"]
    #       map2D[xInd,:] = thisMap2d[:]
    #       toPrint = "{:5} {:3} {:15.10}".format(xInd, info["niter"], info["reducedChi"])
    #       sys.stdout.write(toPrint+'\r')
    #       sys.stdout.flush()
    #       #print(toPrint)
    #   print()
    #   # the struct of map2d
    #   # 0: total length
    #   maxLength = int(np.max(map2D[:,0]))
    #   #map2D = map2D[:,:maxLength+1]
    #   fimageModel = map2DModel(bmod, map2D, flux, ivar)
    #
    #   return amod, flux, ivar, bmod, fimage, fimageModel, map2D, maxLength
    #
    #def _fitForProfileDirect(masterProfile, configs={}):
    #   ny = int(masterProfile.shape[0])
    #   nx = int(masterProfile[0,1])
    #   profileObj = np.empty((ny, nx), dtype=np.object)
    #   badFactor = configs.get("badFactor", 5e-5)
    #   outputPath = configs.get("outputPath")
    #   debug = configs.get("debug", False)
    #   if outputPath is None:
    #       outputPath="./"
    #   else:
    #       outputPath = os.path.join(outputPath, "masterProfile-debug")
    #
    #   parallel = configs.get("parallel", False)
    #   parallelN= configs.get("parallelN", 5)
    #
    #   de.plt.switch_backend("Agg")
    #   for y in range(ny):
    #       todo = masterProfile[y]
    #       p = 2;
    #       for x in range(nx):
    #           thisLen = int(todo[p])
    #           totalLen = thisLen + 40
    #           indStart = 20
    #           indEnd = 20 + thisLen
    #           xx = todo[p+1:p+1+thisLen]
    #           yy = todo[p+1+thisLen:p+1+thisLen*2]
    #           snr = todo[p+1+thisLen*2:p+1+thisLen*3]
    #           xmin=xx.min(); xmax=xx.max()
    #           step = 0.8
    #
    #           xxx = np.zeros(totalLen)
    #           yyy = np.zeros(totalLen)
    #           ssnnrr = np.ones(totalLen)*snr.max()
    #           xxx[:indStart] = np.linspace(xmin-indStart*step, xmin-step, indStart)
    #           xxx[indStart:indEnd] = xx[:]
    #           xxx[indEnd:] = np.linspace(xmax+step, xmax+(totalLen-indEnd)*step, (totalLen-indEnd))
    #           yyy[indStart:indEnd] = yy[:]
    #           ssnnrr[indStart:indEnd] = snr[:]
    #           obj = myProfile()
    #           try:
    #               obj.fit(xxx,yyy,1/ssnnrr)
    #           except Exception as e:
    #               print(e)
    #               plotFitForProfile(xx, yy, obj, xInd=x, yInd=y,
    #                     badness="bad-fit", outputPath=outputPath)
    #           badness = ((obj(xx)-yy)**2/snr).sum()
    #           if debug>1:
    #               if not np.isfinite(badness) or badness>badFactor:
    #                   plotFitForProfile(xx, yy, obj, xInd=x, yInd=y,
    #                         badness=badness, outputPath=outputPath)
    #           #obj = interp.UnivariateSpline(xxx, yyy, ssnnrr, s=20)
    #           #badness = ((obj(xx)-yy)**2/snr).sum()
    #           #if badness>badFactor or not np.isfinite(badness):
    #           #    #m_init = Voigt1D(x_0=0., amplitude_L=1, fwhm_L=0.8, fwhm_G=0.8)
    #           #    #fit_m = fitting.LevMarLSQFitter()
    #           #    #obj = fit_m(m_init, xxx, yyy, weights=ssnnrr)
    #           #    obj = myProfile()
    #           #    if not np.isfinite(badness):
    #           #        yyy[~np.isfinite(yyy)]=0
    #           #        ssnnrr[~np.isfinite(ssnnrr)]=0
    #           #        toPrint = "bad at fitting {:4}/{:4} {:4}/{:4}\r".format(y, ny, x, nx)
    #           #        sys.stdout.write(toPrint+'\r')
    #           #        sys.stdout.flush()
    #           #    obj.fit(xxx,yyy,ssnnrr)
    #           #    badness = [badness, ((obj(xx)-yy)**2/snr).sum()]
    #
    #           #    #if not np.all(np.isfinite(badness)):
    #           #        #print("bad profiles{:4}/{:4} {:4}/{:4}@:\n\t{}\n\t{}\n\t{}\n".format(y, ny, x, nx, xx, yy, snr))
    #           #if debug>1:
    #           #    if not isinstance(badness, list):
    #           #        if badness>5e-5:
    #           #            plotFitForProfile(xx, yy, obj, xInd=x, yInd=y,
    #           #                          badness=badness, outputPath=outputPath)
    #           #    else:
    #           #        plotFitForProfile(xx, yy, obj, xInd=x, yInd=y,
    #           #                          badness=badness, outputPath=outputPath)
    #           profileObj[y,x] = obj
    #           p += 1+thisLen*3
    #           sys.stdout.write("       fitting {:4}/{:4} {:4}/{:4}\r".format(y, ny, x, nx))
    #           sys.stdout.flush()
    #   print()
    #   #de.plt.switch_backend("MacOsX")
    #   de.plt.switch_backend(configs["plotBackend"])
    #   return profileObj
#endregion: backup function
