#region: imports
from __future__ import (absolute_import, division, print_function, unicode_literals)
import numpy as np
import os
import stat
import sys
import matplotlib.pyplot as plt
import numpy
from numpy.polynomial import Legendre
from . import de
from . import extract
from . import wavecalib
from .image import ccdproc
from .image import specLib
from ..utils import macro
from ..utils import messager
from ..utils import configTools
import scipy.interpolate as interp
from scipy.optimize import curve_fit
import glob
import multiprocessing
import copy
import scipy
from ..DS10 import ds9

from . import m2fsConfigs
import shutil
import astropy.io.fits as pyfits
import pdb
from numpy.polynomial import Legendre
from scipy import ndimage
import signal
import time
import pickle
import multiprocessing
import ipdb


class TimeoutException(Exception):   # Custom exception class
    pass
def TimeoutHandler(signum, frame):   # Custom signal handler
    raise TimeoutException
    #raise Exception("Timeout")
OriginalHandler = signal.signal(signal.SIGALRM,TimeoutHandler)
#signal.signal(signal.SIGALRM,OriginalHandler)
#endregion: imports

def makeFlatField(inputFile, configs={}):
    #region: setups
    inputFileName = os.path.basename(inputFile).split(".fits")[0]
    outputFileName= inputFileName + '-flatfield.fits'
    configs.update({"config2show":["rerun", "ncoeff", "initRow", "debug", "nodeFactor"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return
    configs["logger"] = logger
    #endregion: setups
    #de.plt.switch_backend("Agg")
    fiberFlatSmooth, allFactor, fiberIndex = flatField(inputFile, configs=configs)
    #de.plt.switch_backend(configs["plotBackend"])
    #region: save files
    h1 = pyfits.Header()
    h2 = pyfits.Header()
    h3 = pyfits.Header()
    h1["OBJECT"] = "flat"
    h1["EXTNAME"] = "flat"
    h2["OBJECT"] = "factor"
    h2["EXTNAME"] = "factor"
    h3["OBJECT"] = "fiberIndex"
    h3["EXTNAME"] = "fiberIndex"
    data1 = pyfits.PrimaryHDU(fiberFlatSmooth, header=h1)
    data2 = pyfits.ImageHDU(allFactor, header=h2)
    data3 = pyfits.ImageHDU(fiberIndex, header=h3)
    fits  = pyfits.HDUList([data1, data2, data3])
    fits.writeto(outFile, clobber='True')
    #endregion: save files

#region: notes for flatfield
"""
    debug plots:
        -origin:
            all flat points
        -knots:
            red points: old knots
            green points: new knots
            pink points: all flat
            black points: flat points used to fit
    to scale flux:
        flux[i] / fiberFlatSmooth[i] * allFactor[i]
"""
#endregion: notes for flatfield
def flatField(infile, configs={}):
    #region: setups
    outputPath = configs["outputPath"]
    name = os.path.basename(infile).split(".fits")[0]
    modelDir = os.path.join(outputPath, name+"FlatDebug")
    skyFlatFile = os.path.join(outputPath, name+"-skyflat.dat")
    debug = configs.get("debug", 1)
    rerun = configs.get("rerun", 0)
    logger = configs.get("logger")
    smallFluxPer = configs.get("smallFluxPer", 1)
    smallFluxFactor = configs.get("smallFluxFactor", 3)
    rejectFactor = float(configs.get("rejectFactor", 3))
    maxReject = configs.get("maxReject", 300)
    wFactor = configs.get("wFactor",10)
    scaleFactor = configs.get("scaleFactor", 0.99)
    #endregion: setups
    #region: get waveset
    #!! this yset if from traceset
    waveset, coeffs, yset, badFiber, debugInfo = wavecalib.readWaveSet(outputPath, debugInfo=1)
    rms = debugInfo["rms"].flatten()
    finalGoodFiber = np.isfinite(rms)
    waveset = waveset[finalGoodFiber]
    coeffs = coeffs[finalGoodFiber]

    goodFiber = badFiber<1
    logWave = np.log10(waveset)

    fiberIndex = np.where(goodFiber)[0]
    fiberIndex = fiberIndex[finalGoodFiber]
    #endregion: get waveset
    #region: important parameters
    nodeFactor = configs.get("nodeFactor", 1)#!! the factor that node different from the default value
    minDNode  = configs.get("minDNode",0.0001/np.log(10)) * nodeFactor
    smoothDNode  = configs.get("smoothDNode",0.002/np.log(10))
    # magic number to overfit the features in the flat
    # then we can make the fit more smooth to represent the flat...
    nmed = configs.get("nmed", 60)
    #endregion: important parameters
    #region: find the best fiber to be scaled to
    infile = infile.split(".fits")[0] + "Flux.fits"
    thisData, thisStd, thisYset = extract.readFlux(infile)
    thisData = thisData[finalGoodFiber]
    thisStd = thisStd[finalGoodFiber]
    thisYset = thisYset[finalGoodFiber]

    smallStd = np.nanpercentile(thisStd, smallFluxPer)
    bugPoints = thisStd<smallStd/smallFluxFactor
    if bugPoints.sum():
        logger.warn("there are some bug points(sigma toooooo small), filter them\n{}\n{}".\
                format(np.where(bugPoints)[0], np.where(bugPoints)[1]))
    thisStd[bugPoints] = np.inf

    yset = thisYset
    rms = rms[finalGoodFiber]
    thisIvar = 1/thisStd**2
    thisBad   = thisIvar==0
    nFib, nx = thisData.shape
    assert nFib==fiberIndex.size, "nFib:{}, fiberIndex:{}".format(nFib, fiberIndex.size)
    # setup all array
    allBad = np.zeros(thisData.shape, dtype=np.bool)
    allBad[:,:] = thisBad
    allDataFluxSum = np.nansum(thisData, axis=1)
    assert np.all(np.isfinite(allDataFluxSum))
    dataFluxSumSort = np.argsort(allDataFluxSum)
    #rmsSort = np.argsort(rms)
    #goodN = int(nFib/10)
    #good = np.logical_and((nFib-1-dataFluxSumSort)<goodN, rmsSort<goodN)
    #goodIndex = np.where(good)[0]
    #goodRms = rms[goodIndex]
    #indToScale = goodIndex[np.argmin(goodRms)]
    indToScale = np.argsort(rms/allDataFluxSum)[0]
    #indToScale = np.argmin(rms)
    this1dData = thisData[indToScale]
    #endregion: find the best fiber to be scaled to
    #!! (the one that have the latest rms in lamp fit)
    #region: scale all the fibers to the best one
    med = np.nanmedian(this1dData)
    first1dData = thisData[indToScale] / med
    first1dIvar = thisIvar[indToScale] * med**2
    firstLogWave = logWave[indToScale]
    #allFlux, allIvar, allFactor =\
    #    scaleAllToTheFirst(logWave, thisData, thisIvar, 1./med)
    #!! scaleByMinChi2
    allFlux, allIvar, allFactor =\
        scaleAll(logWave, thisData, thisIvar, firstLogWave, first1dData, first1dIvar)
    assert np.all(np.isfinite(allFactor))
    #!! make super flat
    if not os.path.exists(skyFlatFile) or rerun:
        thisConfig={"minDNode":minDNode, "debug":debug, "logger":logger,
                "outputPath":outputPath, "name":name, "maxReject":maxReject,
                "wFactor":wFactor, "scaleFactor":scaleFactor}
        waveF, fluxF, ivarF, badF, w,\
              workwaveF, workfluxF, workivarF, workbadF, workw, \
              fit, newFit, knots, newKnots =\
            superFlat(logWave, allFlux, allIvar, allBad, rejectFactor, configs=thisConfig)
        with open(skyFlatFile, "wb") as f:
            pickle.dump(newFit, f)
        skyFit = newFit
        #!! debug plots
        allFlux[allBad]=np.nan
        if debug>0:
            de.piecePlotMulti( \
               [[logWave, allFlux, "plot"],
                [waveF, fluxF, ["scatter", {"s":6, "c":"black","edgecolors":"None"}], ~badF]],
                N=100, outDir=modelDir+"-origin")
            de.piecePlotMulti(\
                [[waveF, fluxF, ["scatter", {"s":4, "c":"pink","edgecolors":"None"}], ~badF],
                 [workwaveF, workfluxF, ["scatter", {"s":4, "c":"black","edgecolors":"None"}], ~workbadF],
                 [waveF, newFit(waveF), "plot"],
                 [knots, fit(knots), ["scatter",{"s":10, "c":"red", "edgecolors":"None"}]],
                 [newKnots, newFit(newKnots), ["scatter",{"s":20, "c":"green", "edgecolors":"None"}]]],
                 N=100, outDir=modelDir+"-knots")
    else:
        with open(skyFlatFile, "rb") as f:
            skyFit = newFit = pickle.load(f)
    skyMod = newFit(logWave) #!! the fitting model
    for i in range(skyMod.shape[0]):
        if not np.all(skyMod[i]>0):
            badIndex = np.where(np.logical_not(skyMod[i]>0))[0]
            if not (np.all((badIndex<5)|(badIndex>skyMod[i].shape[0]-5))):
                ipdb.set_trace()
                raise Exception("bad flat!!")
            else:
                if not np.all(badIndex<5):
                    skyMod[i, badIndex[badIndex<5]] = np.median(skyMod[i][:10])
                if not np.all(badIndex>skyMod[i].shape[0]-5):
                    skyMod[i, badIndex[badIndex>skyMod[i].shape[0]-5]] = np.median(skyMod[i][-10:])
                logger.warn("bad tail at fiber {}".format(i))

    #endregion: scale all the fibers to the best one
    #region: redo the process use skyMod as the reference
    sflatInit = allFlux / skyMod #!! all scale variation
    sflat = np.zeros(sflatInit.shape)
    fflat = np.zeros(sflat.shape)
    fivar = np.zeros(sflat.shape)
    xx = np.arange(sflat.shape[1])
    fiberFlat = np.zeros(sflat.shape)
    fiberFlatSmooth = np.zeros(sflat.shape)
    logger.p("calculate fiberFlat for each fiber...")
    nsmooth=20
    nKernel = 2*nsmooth + 1
    gausskern = np.exp(- ((nKernel-1)/2. - np.arange(nKernel))**2 / nsmooth**2)
    gausskern = gausskern / gausskern.sum()
    for i in range(sflat.shape[0]):
        #!! large scale variation
        badMask = np.logical_not(np.isfinite(sflatInit[i]))
        try:
            sflatInit[i][:] = scipy.interpolate.interp1d(xx[~badMask], sflatInit[i][~badMask], fill_value="extrapolate")(xx)[:]
        except Exception as e:
            print(e)
            ipdb.set_trace()
        todo = np.concatenate(([sflatInit[i][0]]*nsmooth,
                                sflatInit[i],
                               [sflatInit[i][-1]]*nsmooth))
        fflat[i,:] = np.convolve(todo, gausskern, 'same')[nsmooth:-nsmooth]
        fivar[i,:] = allIvar[i]*(fflat[i]/sflatInit[i])**2
        fflat[i,badMask]=0
        fivar[i,badMask]=0
        #!! fit large scale variation
        fit = scipy.interpolate.UnivariateSpline(\
            logWave[i], fflat[i,:], w=fivar[i], s=0)
        fiberFlat[i,:] = fit(logWave[i])
        knots = fit.get_knots()
        newKnots = smoothNode(knots, smoothDNode)
        #!! use smooth knots and again fit large scale variation
        newFit = scipy.interpolate.LSQUnivariateSpline(\
            logWave[i], fflat[i,:], newKnots[1:-1], w=fivar[i])
        fiberFlatSmooth[i,:] = newFit(logWave[i])
        if not np.all(fiberFlatSmooth[i,:]>0):
            ipdb.set_trace()
        fflat[i,badMask]=np.nan
        fivar[i,badMask]=np.nan
        #!! small scale variation
        sflat[i,:] = sflatInit[i,:]/fflat[i,:]
        if debug>1:
            de.p(logWave[i], fflat[i], c="red")
            de.p(logWave[i], fiberFlat[i], c="blue")
            de.p(logWave[i], fiberFlatSmooth[i], c="green")
    logger.p("recalculate factors")
    #!! correct for large scale variance and do scale again
    skyFlatenedOrigin = thisData/fiberFlatSmooth
    skyFlatenedIvarOrigin = thisIvar*fiberFlatSmooth**2
    skyFlatened, skyFlatenedIvar, allFactor =\
       scaleAllSameSize(skyFlatenedOrigin, skyFlatenedIvarOrigin, skyMod, np.isfinite(skyMod))
    #endregion: redo the process use skyMod as the reference
    #ipdb.set_trace()
    if debug>0:
        #de.piecePlotMulti([[logWave, sflat, "plot"]],outDir=modelDir + "-sflat")
        de.piecePlotMulti([[logWave, fiberFlat, "plot"]], outDir=modelDir + "-fiberFlat")
        de.piecePlotMulti([[logWave, fiberFlatSmooth, "plot"]], outDir=modelDir + "-fiberFlatSmooth")
        ivarF = allIvar.flatten()
        badF  =  ivarF==0
        waveF = logWave.flatten()
        sortInd = np.argsort(waveF)
        waveF = waveF[sortInd]
        fluxF = allFlux.flatten()
        fluxF = fluxF[sortInd]
        newfluxF = skyFlatened.flatten()
        newfluxF = newfluxF[sortInd]
        badF  =  badF[sortInd]
        #de.piecePlotMulti( \
            #[[waveF, fluxF, ["scatter", {"s":4, "c":"black","edgecolors":"None"}], ~badF],
             #[waveF, newfluxF, ["scatter", {"s":4, "c":"green","edgecolors":"None"}], ~badF],
             #[waveF, skyFit(waveF), "plot"],
             #[waveF[::100], skyFit(waveF[::100]), ["scatter",{"s":20, "c":"green", "edgecolors":"None"}]]],
            #N=100, outDir=modelDir+"-Flattened")
        factor = 1 / fiberFlatSmooth * allFactor.reshape((allFactor.size, 1))
        flux = thisData * factor
        bad = flux==0
        de.piecePlotMulti( \
           [[waveset, flux, "plot"],
            [waveset, flux, ["scatter", {"s":6, "c":"black","edgecolors":"None"}], ~bad]],
            N=100, outDir=modelDir+"-scaled-result")

    #!! debug note: i=10;de.p(thisData[i]*allFactor[i],"b");de.p(thisData[i]*allFactor[i]/fiberFlatSmooth[i],"r");de.p(skyMod[i],"g")
    #!! first is large scale variance, second
    if not np.all(fiberFlatSmooth>0):
        ipdb.set_trace()
    assert np.all(fiberFlatSmooth>0)
    assert np.all(allFactor>0)
    return fiberFlatSmooth, allFactor, fiberIndex

#region: aux functions
def safe_mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)

def trySplineFit(args, kwargs):
    scipy.interpolate.UnivariateSpline(*args, **kwargs)

def saveSplineFit(args, kwargs, cacheFilePrefix):
    fit = scipy.interpolate.UnivariateSpline(*args, **kwargs)
    with open(cacheFilePrefix+"splineFit.dat", "wb") as f:
        pickle.dump(fit, f)

def loadSplineFit(cacheFilePrefix):
    with open(cacheFilePrefix+"splineFit.dat", "rb") as f:
        result = pickle.load(f)
    return result
#endregion: aux functions

def scaleAll(logWave, thisData, thisIvar, firstLogWave, first1dData, first1dIvar, configs={}):
    nFib, nx = logWave.shape
    globalMin = np.max(logWave[:,0])
    globalMax = np.min(logWave[:,-1])
    allFlux = np.zeros((nFib, nx))
    allIvar = np.zeros((nFib, nx))
    allFactor = np.ones(nFib) * -1
    firstGood = first1dIvar>0
    firstLogWave = firstLogWave.copy()[firstGood]
    first1dData = first1dData.copy()[firstGood]
    first1dIvar = first1dIvar.copy()[firstGood]
    # scale all other fibers
    for i in range(nFib):
        this1dData = thisData[i].copy()
        this1dIvar = thisIvar[i].copy()
        thisGood = this1dIvar>0
        factor,error = scaleByMinChi2(globalMin, globalMax,
                          firstLogWave, first1dData, first1dIvar,
                      logWave[i][thisGood], this1dData[thisGood], this1dIvar[thisGood], configs={})
        allFlux[i,:] = this1dData[:] * factor
        allIvar[i,:] = this1dIvar[:] / factor**2
        allFactor[i] = factor
        #medianFactor = 1/np.nanmedian(this1dData)
        #print "{:3} scale:{:10.5e} difference: {:>+5.1e}%".format(i, factor, (factor-medianFactor)/medianFactor*100)
    return allFlux, allIvar, allFactor

def scaleByMinChi2(globalMin, globalMax, logWave0, data0, ivar0, logWave1, data1, ivar1, configs={}):
    xx = np.linspace(globalMin, globalMax, data0.size*4)
    inRange0 = np.logical_and(logWave0>globalMin, logWave0<globalMax)
    inRange1 = np.logical_and(logWave1>globalMin, logWave1<globalMax)
    goodMask0 = ivar0[inRange0]>0
    goodMask1 = ivar1[inRange1]>0
    logWave0 = logWave0[inRange0][goodMask0]
    logWave1 = logWave1[inRange1][goodMask1]
    data0 = data0[inRange0][goodMask0]
    data1 = data1[inRange1][goodMask1]
    ivar0 = ivar0[inRange0][goodMask0]
    ivar1 = ivar1[inRange1][goodMask1]
    fit0d = scipy.interpolate.interp1d(logWave0, data0, bounds_error=0)
    fit0i = scipy.interpolate.interp1d(logWave0, ivar0, bounds_error=0)
    fit1d = scipy.interpolate.interp1d(logWave1, data1, bounds_error=0)
    fit1i = scipy.interpolate.interp1d(logWave1, ivar1, bounds_error=0)
    mod0d = fit0d(xx)
    mod0i = fit0i(xx) # use witch?
    mod1d = fit1d(xx)
    mod1i = fit1i(xx)
    factor = np.nansum(mod1d * mod0d * mod1i)/np.nansum(mod1d**2 * mod1i)
    error = 1/np.sqrt(np.nansum(mod1d**2*mod1i))
        #ivar = 1/(1/mod0i + 1/mod1i*factor**2)
        #factor = np.nansum(mod1d * mod0d * ivar)/np.nansum(mod1d**2 * ivar)
        #ivar = 1/(1/fit0i(logWave1) + 1/fit1i(logWave1)*factor**2)
        #diff = (fit0d(logWave1) - factor*data1)**2 * (ivar)
        #goodMask = diff<20
        #logWave1 = logWave1[goodMask]
        #data1 = data1[goodMask]
        #ivar1 = ivar1[goodMask]
        #fit1d = scipy.interpolate.interp1d(logWave1, data1, bounds_error=0)
        #fit1i = scipy.interpolate.interp1d(logWave1, ivar1, bounds_error=0)
        #mod1d = fit1d(xx)
        #mod1i = fit1i(xx)
        #ivar = 1/(1/mod0i + 1/mod1i)
        #factor = np.nansum(mod1d * mod0d * mod0i)/np.nansum(mod1d**2 * mod0i)
        #ivar = 1/(1/mod0i + 1/mod1i*factor**2)
        #factor = np.nansum(mod1d * mod0d * ivar)/np.nansum(mod1d**2 * ivar)
    return factor, error

def smoothNode(node, minD):
    """i forgot how it works...."""
    node = node.copy()
    diff = np.diff(node)
    starts, ends = findBlock(diff<minD)
    for i in range(starts.size):
        start = node[starts[i]]
        end = node[ends[i]]
        N = int((end-start)/minD)+1 # N points in this range
        if N>1:
            node[starts[i]:starts[i]+N] = np.linspace(start, end, N)
        node[starts[i]+N:ends[i]+1]=np.nan
    node = node[np.isfinite(node)]
    #node[0] = node[0] + (node[1]-node[0])*1e-10
    #node[-1] = node[-1] - (node[-1] - node[-2])*1e-10
    return node

def findBlock(diff):
    N = diff.size
    ddiff = np.diff(diff.astype(np.int))
    diffNonZero = ddiff[(ddiff!=0)]
    starts = np.where(ddiff>0)[0]+1
    ends = np.where(ddiff<0)[0]+1
    if starts.size==0:
        starts = np.insert(starts, 0, 0)
        if ends.size==0:
            ends = np.append(ends, N)
    else:
        if diffNonZero[0]==-1:
            starts = np.insert(starts, 0, 0)
        if diffNonZero[-1]==1:
            ends = np.append(ends, N)
    if starts.size!=ends.size:
        ipdb.set_trace()
    return starts, ends

def selectBestFlats(fileList, configs={}):
    #region: setups
    configs["rerun"] = configs.get("rerun", False)
    outputFileName= "masterFlatField" + '.fits'
    configs.update({"config2show":["rerun","manually"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)

    manually = configs.get("manually", "")
    outputPath = configs["outputPath"]
    allFiles = [os.path.join(outputPath, os.path.basename(each).\
                split(".fits")[0]+"-flatfield.fits")
                    for each in fileList]
    allFileNames = [os.path.basename(each) for each in allFiles]
    allFileListNames = [os.path.basename(each) for each in fileList]
    f = "ds10"
    for each1,each2 in zip(allFileListNames, allFileNames):
        f += " {} -f1 --regions='masterFlat.reg' {} -f0 ".format(each1, each2)
    #endregion: setups

    if manually:
        bestFlatNames = list(filter(lambda x:manually in x, allFiles))
        if len(bestFlatNames)!=1:
            raise logger.Exception("manually file not unique: {}".format(bestFlatNames))
        else:
            bestFlatName = bestFlatNames[0]
    else:
        allObject = [pyfits.open(each) for each in allFiles]
        allFlat = np.array([each[0].data for each in allObject])
        allFactor = np.array([each[1].data for each in allObject])
        #!! debug notes: de.p((allFactor/allFactor.sum(axis=1).reshape((allFactor.shape[0],1))).T
        #!! debug notes: i=50;de.p(allFactor[:,i].T)
        factorSum = allFactor.sum(axis=1)
        if not np.all(np.isfinite(factorSum)):
            badMask = ~np.isfinite(factorSum)
            badFiles = [each for i,each in enumerate(allFiles) if badMask[i]]
            logger.warn("bug found in file:{}\nfrom {}".format(badFiles, allFiles))
        bestInd = np.nanargmax(factorSum)
        bestFlatName = allFiles[bestInd]
    if toReturn: return f
    logger.p("select best flat as {} from {}".format(bestFlatName, allFiles))
    logger.warn2file("select best flat as {} from {}".format(bestFlatName, allFiles))
    shutil.copy(bestFlatName, outFile)
    return f


#region: notes for superFlat
""" notes for the code
    workfluxF, workivarF, workbadF, workwaveF, workw all flatten
    fit is the fitting object
    1. let s=workw.size*wFactor
       and then enlarge wFactor to make the fitting
            finish in firstTryTimeout
       get wFactor
    2. let s=workw.size*wFactor, caleFactor = 0.99, timeout = 30
           res = (fluxMod-workfluxF)*workw
           std = np.std(res)
       wFactor = (std**2)*scaleFactor
       s is getting smaller until fitting time out.
    3. smooth the node and make a last fit
        minDNode = configs.get("minDNode", 0.0001)
        this is the rough step for the node...
"""
#endregion: notes
def superFlat(logWave, allFlux, allIvar, allBad, rejectFactor, configs={}):
    minDNode = configs.get("minDNode", 0.0001)
    debug = configs.get("debug",0)
    logger = configs["logger"]
    name = configs.get("name",str(time.time()))
    outputPath = configs["outputPath"]
    maxReject = configs.get("maxReject", 300)
    #region: init all outputs as flatten array
    ivarF = allIvar.flatten()
    badF  =  allBad.flatten()
    waveF = logWave.flatten()
    fluxF = allFlux.flatten()
    sortInd = np.argsort(waveF)
    fluxF = fluxF[sortInd]
    ivarF = ivarF[sortInd]
    badF  =  badF[sortInd]
    waveF = waveF[sortInd]
    w = np.sqrt(ivarF*(1-badF))
    workfluxF = fluxF.copy()
    workivarF = ivarF.copy()
    workbadF  = badF.copy()
    workwaveF = waveF.copy()
    workw = w.copy()
    workfluxF[workbadF]=-65535
    #endregion: init all outputs
    #region: to make wFactor large enough to be calculated at the first time
    #wFactor = 0.00001
    wFactor = configs["wFactor"]
    tempN = 0
    maxTempN = 100
    firstTryTimeout = 3
    while True:
        tempN+=1
        args = [workwaveF, workfluxF]
        kwargs = {"w":workw, "s":workw.size*wFactor}
        p = multiprocessing.Process(target=trySplineFit,
                        args=(args, kwargs))
        p.start()
        p.join(firstTryTimeout)
        if p.is_alive():#!! need more time to fit
            p.terminate()
            logger.p("First try Time out with wFactor {}".format(wFactor))
            wFactor = wFactor*1.1
            if tempN>maxTempN:
                raise Exception("all {} first try failed with final wFactor {}...". \
                            format(maxTempN, wFactor))
        else:
            logger.p("First try ok with wFactor {}".format(wFactor))
            break
    #endregion: to make wFactor large enough to be calculated at the first time
    #region: begin eain fitting loop
    maxiter = 599
    i = 0
    timeout = 30
    scaleFactor = configs.get("scaleFactor", 0.99)
    cacheFilePrefix=os.path.join(outputPath, name+"-")
    # iterate fitting
    logger.p("\t\tbegin iteration fitting for the flat")
    while True:
        i+=1
        if i>maxiter:
            raise Exception("reach maxiter, there must be some problems")
        workfluxF[workbadF]=-65535
        # use signal
        args = [workwaveF, workfluxF]
        kwargs = {"w":workw, "s":workw.size*wFactor}
        p = multiprocessing.Process(target=saveSplineFit,
                                    args=(args, kwargs, cacheFilePrefix))
        theTime = time.time()
        p.start()
        p.join(timeout)
        theTime = time.time()-theTime
        if p.is_alive():
            p.terminate()
            logger.p("Time out break with wFactor {:10.5f} with time {:5.1f}".\
                format(wFactor, theTime))
            break
        else:
            fit = loadSplineFit(cacheFilePrefix)

        fluxMod = fit(workwaveF)
        res = (fluxMod-workfluxF)*workw
        std = np.std(res)
        if not np.isfinite(std):
            ipdb.set_trace()
        badMask = np.abs(res)>rejectFactor*std
        if badMask.sum()>maxReject:
            bad2good = np.argsort(np.abs(res)[badMask])[maxReject:]
            badMaskIndex = np.where(badMask)[0]
            badMask[badMaskIndex[bad2good]] = False
        goodMask = np.logical_not(badMask)

        badCount = (~goodMask).sum()
        logger.p("reject {:5d} from {:7d} with std {:10.6f} running:{:3.1f}s".\
            format(badCount, workw.size, std, theTime))
        lastGoodWorkfluxF = workfluxF
        lastGoodWorkwaveF = workwaveF
        lastGoodFit       = fit
        lastGoodW         = workw
        lastworkbadF      = workbadF
        if badCount==0:
            break
        workfluxF = workfluxF[goodMask]
        workivarF = workivarF[goodMask]
        workbadF  =  workbadF[goodMask]
        workwaveF = workwaveF[goodMask]
        workw = np.sqrt(workivarF*(1-workbadF))
        wFactor = (std**2)*scaleFactor
        if False:
            fig=plt.figure()
            ax=fig.add_subplot(111)
            ax.scatter(waveF, fluxF, edgecolors="None", s=3, c="black")
            ax.set_xlim([3.881, 3.883])
            ax.plot(waveF, fit(waveF), c="red")
            plt.show()
        #fluxMod = fit(workwaveF)
    #endregion: begin eain fitting loop
    #region: now smooth the node and make a last fit
    workfluxF = lastGoodWorkfluxF
    workwaveF = lastGoodWorkwaveF
    workw     = lastGoodW
    workbadF  = lastworkbadF
    fit       = lastGoodFit
    knots = fit.get_knots()
    #!! smooth the too-detailed fit by setting all node with a minimum know distance
    newKnots = smoothNode(knots, minDNode)
    #de.piecePlotMulti( \
    #    [[knots, fit(knots), ["scatter",{"s":30, "c":"red", "edgecolors":"None"}]],
    #     [newKnots, fit(newKnots), ["scatter",{"s":8, "c":"green", "edgecolors":"None"}]]], N=100)
    #last fit with new knots
    newFit = scipy.interpolate.LSQUnivariateSpline(\
        workwaveF, workfluxF, newKnots[1:-1], w=workw)
    #endregion: now smooth the node and make a last fit

    return waveF, fluxF, ivarF, badF, w,\
           workwaveF, workfluxF, workivarF, workbadF, workw,\
           fit, newFit, knots, newKnots

def readFiberFlat(outputPath):
    file = os.path.join(outputPath, 'masterFlatFlux.fits')
    obj = pyfits.open(file)
    fiberFlat = obj[0].data
    factors = obj[1].data
    return fiberFlat, factors

def scaleAllSameSize(thisData, thisIvar, firstData, firstIvar, configs={}):
    assert thisData.shape==firstData.shape
    nFib, nx = thisData.shape
    allFlux = np.zeros((nFib, nx))
    allIvar = np.zeros((nFib, nx))
    allFactor = np.ones(nFib) * -1
    # scale all other fibers
    for i in range(nFib):
        this1dData = thisData[i].copy()
        this1dIvar = thisIvar[i].copy()
        first1dData = firstData[i].copy()
        first1dIvar = firstIvar[i].copy()
        good1 = this1dIvar>0
        good2 = first1dIvar>0
        good = np.logical_and(good1, good2)
        factor,error = scaleByMinChi2SameSize(\
                this1dData[good], this1dIvar[good],
                first1dData[good], configs={})
        allFlux[i,:] = this1dData[:] * factor
        allIvar[i,:] = this1dIvar[:] / factor**2
        allFactor[i] = factor
        #medianFactor = 1/np.nanmedian(this1dData)
        #print "{:3} scale:{:10.5e} difference: {:>+5.1e}%".format(i, factor, (factor-medianFactor)/medianFactor*100)
    return allFlux, allIvar, allFactor

#region: backup codes
# ....
def smoothNodeBySNR(x, SNR, N=4000):
    assert x.shape == SNR.shape
    SNRAll = SNR.sum()
    dSNR = SNRAll / N
    result = np.zeros(x.size)
    ind = 1
    sum = 0
    result[0] = x[0]
    for i in range(1, x.size-1):
        sum += SNR[i]
        if sum>=dSNR:
            result[ind] = x[i]
            sum = 0
            ind += 1
    result[ind] = x[-1]
    result = result[:ind+1]
    return result

def superFlatWithKnots(logWave, allFlux, allIvar, allBad, newKnots):
    ivarF = allIvar.flatten()
    badF  =  allBad.flatten()
    waveF = logWave.flatten()
    fluxF = allFlux.flatten()
    sortInd = np.argsort(waveF)
    fluxF = fluxF[sortInd]
    ivarF = ivarF[sortInd]
    badF  =  badF[sortInd]
    waveF = waveF[sortInd]
    w = np.sqrt(ivarF*(1-badF))
    workfluxF = fluxF.copy()
    workivarF = ivarF.copy()
    workbadF  = badF.copy()
    workwaveF = waveF.copy()
    workw = w.copy()
    workfluxF[workbadF]=-65535

    fit = scipy.interpolate.LSQUnivariateSpline( \
        workwaveF, workfluxF, newKnots[1:-1], w=workw)

    return waveF, fluxF, ivarF, badF, w, \
           workwaveF, workfluxF, workivarF, workbadF, workw, \
           fit

def masterFlat(infileList, outputPath, configs={}):
    infile = [os.path.join(outputPath, each+'md.fits') for each in infileList]
    outfile= os.path.join(outputPath, 'masterFlatFlux.fits')
    rerun = int(configs.get('rerun', 0))
    if not rerun and os.path.exists(outfile):
        print("in masterFlat: {} already exists, skip".format(outfile))
        return
    print("masterFlat({})".format(infileList))
    fileToUse = configs.get("fileToUseAsFlat", None)
    if fileToUse is None:
        maxcount = 0
        maxcountInd = -1
        for i, eachFile in enumerate(infile):
            obj = pyfits.open(eachFile)
            eachData = obj[0].data
            eachError= obj[1].data
            theCount = eachData[np.isfinite(eachError)].sum()
            if theCount>maxcount:
                maxcount = theCount
                maxcountInd = i
        fileToUse = infile[maxcountInd]
        print("\tin masterFlat: use {} to build the master flat".format(fileToUse))
        extract.imageExtract(fileToUse, outputPath, configs=configs.pop("flatExtract"))
        configs["modelDir"]=fileToUse.split("md.fits")[0]+"FlatModel"

    fiberFlatSmooth, allFactor = masterFlatOneFile(fileToUse.split("md.fits")[0],
                                                   outputPath, configs=configs)
    data1 = pyfits.PrimaryHDU(fiberFlatSmooth, header=None)
    data2 = pyfits.ImageHDU(allFactor, header=None)
    fits  = pyfits.HDUList([data1, data2])
    fits.writeto(outfile, clobber='True')

def scaleByMinChi2SameSize(data0, ivar0, data1, configs={}):
    #fit1i = scipy.interpolate.interp1d(logWave1, ivar1, bounds_error=0)
    mod0d = data0
    mod0i = ivar0
    mod1d = data1
    #mod1i = fit1i(xx)
    factor = np.nansum(mod1d * mod0d * mod0i)/np.nansum(mod0d**2 * mod0i)
    error = 1/np.sqrt(np.nansum(mod1d**2*mod0i))
        #ivar = 1/(1/mod0i + 1/mod1i*factor**2)
        #factor = np.nansum(mod1d * mod0d * ivar)/np.nansum(mod1d**2 * ivar)
        #ivar = 1/(1/fit0i(logWave1) + 1/fit1i(logWave1)*factor**2)
        #diff = (fit0d(logWave1) - factor*data1)**2 * (ivar)
        #goodMask = diff<20
        #logWave1 = logWave1[goodMask]
        #data1 = data1[goodMask]
        #ivar1 = ivar1[goodMask]
        #fit1d = scipy.interpolate.interp1d(logWave1, data1, bounds_error=0)
        #fit1i = scipy.interpolate.interp1d(logWave1, ivar1, bounds_error=0)
        #mod1d = fit1d(xx)
        #mod1i = fit1i(xx)
        #ivar = 1/(1/mod0i + 1/mod1i)
        #factor = np.nansum(mod1d * mod0d * mod0i)/np.nansum(mod1d**2 * mod0i)
        #ivar = 1/(1/mod0i + 1/mod1i*factor**2)
        #factor = np.nansum(mod1d * mod0d * ivar)/np.nansum(mod1d**2 * ivar)
    return factor, error

def UnivariateSplineWithReject(xx, yy, w=None, reject=3, s=None, name="?", configs={}):
    debug = configs.get("debug", 1)
    outputPath = configs.get("outputPath", "./debugSplineRejectFit")
    rerun = configs.get("rerun", 0)
    newFitFile = os.path.join(outputPath, name+"newFit.dat")
    newFitSigmaFile = os.path.join(outputPath, name+"newFitSigma2.dat")
    if os.path.exists(newFitFile) and os.path.exists(newFitSigmaFile) and not rerun:
        with open(newFitFile, "rb") as f:
            newFit = pickle.load(f)
        with open(newFitSigmaFile, "rb") as f:
            newFitSigma = pickle.load(f)
        return newFit, newFitSigma

    obj = [0,1][xx.dtype==np.object]
    assert xx.shape == yy.shape
    N = xx.shape[0]
    if obj:
        size = 0
        for each in xx:
            size += each.size
        xxFlatten = np.zeros(size)
        yyFlatten = np.zeros(size)
        wFlatten  = np.ones(size)
        ind=0
        if w is not None:
            for i in range(xx.size):
                thisSize = xx[i].size
                xxFlatten[ind:ind+thisSize] = xx[i][:]
                yyFlatten[ind:ind+thisSize] = yy[i][:]
                wFlatten[ind:ind+thisSize]  =  w[i][:]
                ind += thisSize
        else:
            for i in range(xx.size):
                thisSize = xx[i].size
                xxFlatten[ind:ind+thisSize] = xx[i][:]
                yyFlatten[ind:ind+thisSize] = yy[i][:]
                ind += thisSize

        sortInd = np.argsort(xxFlatten)
        xxFlatten = xxFlatten[sortInd]
        yyFlatten = yyFlatten[sortInd]
        wFlatten =  wFlatten[sortInd]
        yyFlatten[wFlatten==0]=-65535
    else:
        xxFlatten = xx.flatten()
        yyFlatten = yy.flatten()
        sortInd = np.argsort(xxFlatten)
        xxFlatten = xxFlatten[sortInd]
        yyFlatten = yyFlatten[sortInd]
        if w is not None:
            wFlatten = w.flatten()
            wFlatten = wFlatten[sortInd]
            yyFlatten[wFlatten==0]=-65535
        else:
            wFlatten = np.ones(xxFlatten.size)

    workxx = xxFlatten.copy()
    workyy = yyFlatten.copy()
    workw  = wFlatten
    workxx = workxx[workw>0]
    workyy = workyy[workw>0]
    workw  =  workw[workw>0]

    wFactor = configs.get("wFactor", 0.01)
    firstTryTimeout = configs.get("firstTryTimeout", 3)
    maxTempN = configs.get("maxTemmpN", 100)# max first try num
    tempN=0
    while True:
        tempN += 1
        args = [workxx, workyy]
        kwargs = {"w":workw, "s":workxx.size * wFactor}
        p = multiprocessing.Process(target=trySplineFit,
                                    args=(args, kwargs))
        p.start()
        p.join(firstTryTimeout)
        if p.is_alive():
            p.terminate()
            print("First try Time out with wFactor {}".format(wFactor))
            wFactor = wFactor*1.1
            if tempN>maxTempN:
                raise Exception("all {} first try failed with final wFactor {}...". \
                                format(maxTempN, wFactor))
        else:
            print("First try ok with wFactor {}".format(wFactor))
            break

    i = 0
    timeout = configs.get("timeout", 30)
    scaleFactor = configs.get("scaleFactor", 0.99)
    cacheFilePrefix = configs.get("cacheFilePrefix","./")
    maxiter = configs.get("maxiter", 200)
    # iterate fitting
    print("\t\tbegin iteration fitting for {}".format(name))
    while True:
        i+=1
        if i>maxiter:
            raise Exception("reach maxiter, there must be some problems")
        # use signal
        args = [workxx, workyy]
        kwargs = {"w":workw, "s":workxx.size*wFactor}
        p = multiprocessing.Process(target=saveSplineFit,
                                    args=(args, kwargs, cacheFilePrefix))
        theTime = time.time()
        p.start()
        p.join(timeout)
        theTime = time.time()-theTime
        if p.is_alive():
            p.terminate()
            print("Time out break with wFactor {:10.5f} with time {:5.1f}". \
                format(wFactor, theTime))
            break
        else:
            fit = loadSplineFit(cacheFilePrefix)

        fluxMod = fit(workxx)
        res = (fluxMod-workyy)*workw
        std = np.std(res)
        goodMask = np.abs(res)<reject*std
        badCount = (~goodMask).sum()
        print("reject {:5d} from {:7d} with std {:10.6f} running:{:3.1f}s". \
            format(badCount, workw.size, std, theTime))
        lastGoodyy = workyy
        lastGoodxx = workxx
        lastGoodFit  = fit
        lastGoodw    = workw
        if badCount==0:
            break
        workyy = workyy[goodMask]
        workxx = workxx[goodMask]
        workw  = workw[goodMask]
        wFactor = (std**2)*scaleFactor

    fit = lastGoodFit
    goodxx = lastGoodxx
    goodyy = lastGoodyy
    goodw  = lastGoodw

    #knots  = fit.get_knots()
    #minDNode = configs.get("minDNode", 0.3)
    #newKnots = smoothNode(knots, minDNode)
    N = configs.get("N", 2048)
    newKnots = smoothNodeBySNR(goodxx, np.sqrt(goodyy*goodw),N=N)
    newKnots = smoothNode(newKnots, 1)

    newFit = scipy.interpolate.LSQUnivariateSpline( \
        goodxx, goodyy, newKnots[1:-1], w=goodw)
    newMod = newFit(goodxx)
    sigma = np.abs(newMod - goodyy)

    newFitSigma = scipy.interpolate.LSQUnivariateSpline( \
        goodxx, sigma, newKnots[1:-1])

    with open(newFitFile, "wb") as f:
        pickle.dump(newFit, f)
    with open(newFitSigmaFile, "wb") as f:
        pickle.dump(newFitSigma, f)

    if debug>0:
        knotsMod = newFit(newKnots)
        sigmaMod = newFitSigma(goodxx)
        sigmaNodeMod = newFitSigma(newKnots)
        de.piecePlotMulti(\
            [[xx, yy, ["scatter", {"s":6, "c":"black", "edgecolors":"None"}], w>0],
             [goodxx, goodyy, ["scatter", {"s":6, "c":"red", "edgecolors":"None"}]],
             [goodxx, newMod, "plot"],
             [newKnots, knotsMod, ["scatter", {"s":6, "c":"green", "edgecolors":"None"}]],
             [goodxx, sigma, ["scatter", {"s":6, "c":"black", "edgecolors":"None"}]],
             [goodxx, sigmaMod, "plot"],
             [newKnots, sigmaNodeMod, ["scatter", {"s":6, "c":"black", "edgecolors":"None"}]]],
              N=100, outDir=outputPath, prefix=name)
    return newFit, newFitSigma

#endregion: backup codes


