# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function, unicode_literals)

from ctypes import *
import numpy as np
from numpy.polynomial import Legendre
import os
import sys
from .. import de
import ipdb
import warnings
warnings.filterwarnings("ignore", message="overflow encountered in exp")
warnings.filterwarnings("error", message="invalid value encountered in sqrt")

#region: load dynamic library
#so = CDLL(os.path.join(os.path.dirname(__file__),"pySpecLib.so"))
so = CDLL(os.path.join(os.path.dirname(__file__),"lib","extract_row.so"))
extract_row_c = so.extract_row
extract_wapper_c = so.extract_wapper
extractDirect_c = so.extractDirect
extractDirect23_c = so.extractDirect23
extractOnly23_c = so.extractOnly23
profile23_c = so.profile23
profile233_c = so.profile233
profileGaussian_c = so.profileGaussian
extractUsingProfileWapper_c = so.extractUsingProfileWapper
map2DCutFixWidth_c = so.map2DCutFixWidth
remapFlux_c = so.remapFlux
#buildMasterProfile_c = so.buildMasterProfile


extract_row23_c = so.extract_row23
map2DSlice_c  = so.map2DSlice
map2DSliceFixNaNWidth_c  = so.map2DSliceFixNaNWidth
#map2DFixWidthRead_c = so.map2DFixWidthRead
#map2DFixWidthWrite_c = so.map2DFixWidthWrite
so = CDLL(os.path.join(os.path.dirname(__file__),"lib","lampTools.so"))
correlate_c = so.correlate
buildLamp_c = so.buildLamp
buildSmoothLamp_c = so.buildSmoothLamp
getSubIndex_c = so.getSubIndex

so = CDLL(os.path.join(os.path.dirname(__file__),"lib","nrlib.so"))
gaussianFitIter_c = so.gaussianFitIter
genBlocks_c = so.genBlocks
variateDeconvolve_c = so.variateDeconvolve
variateConvolve_c = so.variateConvolve

c_int_p = POINTER(c_int)
c_long_p = POINTER(c_long)
c_double_p = POINTER(c_double)

def p3(array):
    m, n=array.shape
    ctypes_arrays = [each for each in np.ctypeslib.as_ctypes(array)]
    return (POINTER(ctypes_arrays[0]._type_)*n)(*ctypes_arrays)

#endregion: load dynamic library

#region: about lamps
# void buildLamp(double* lam, int nLam,
# double* modLam, int nx, int npad,
# double* intensity, double* model)

buildLamp_c.argtypes = [c_double_p, c_int, c_double_p, c_int, c_int, c_double_p, c_double_p]
def buildLamp(lam, intensity, modLam, npad):
    '''adlocate intensity to 2 pixel near by according to its distance to the 2 pixel
    lam and intensity is from arc data(nArc), modLam is the test modLam (nx)
     x          y                               xx'''
    nlam = lam.size
    nx = modLam.size
    lam       = lam.astype(np.float64)
    modLam    = modLam.astype(np.float64)
    intensity = intensity.astype(np.float64)
    model     = np.zeros(nx+2*npad)
    lam_c       = np.ctypeslib.as_ctypes(lam)
    modLam_c    = np.ctypeslib.as_ctypes(modLam)
    intensity_c = np.ctypeslib.as_ctypes(intensity)
    model_c     = np.ctypeslib.as_ctypes(model)
    buildLamp_c(lam_c, nlam, modLam_c, nx, npad, intensity_c, model_c)
    return model

#def buildSmoothLamp(lam, intensity, modStart, nMod, step, sigmas, sigmaLength=5):
def buildSmoothLamp(lam, intensity, modLam, sigmas, sigmaLength=5):
    """              x       y        xx
    len(kernel) = sigmaLength*2 + 1"""
    nlam = lam.size
    nMod = modLam.size
    if isinstance(sigmas, float) or isinstance(sigmas, int):
        sigmas = sigmas * np.ones(nlam)
    elif not isinstance(sigmas, np.ndarray):
        raise Exception("sigmas not scale or ndarray? Error input")
    lam       = lam.astype(np.float64)
    modLam    = modLam.astype(np.float64)
    intensity = intensity.astype(np.float64)
    model     = np.zeros(nMod)
    lam_c       = np.ctypeslib.as_ctypes(lam)
    modLam_c    = np.ctypeslib.as_ctypes(modLam)
    intensity_c = np.ctypeslib.as_ctypes(intensity)
    model_c     = np.ctypeslib.as_ctypes(model)
    sigmas_c    = np.ctypeslib.as_ctypes(sigmas)
    #void buildSmoothLamp(double* lam, int nLam, double* modLam, int nx, int subPixel,
    #                     double* intensity, double *sigmas, int sigmaLength, double* model)
    buildSmoothLamp_c(lam_c, nlam, modLam_c, nMod, 5,
                      intensity_c, sigmas_c, sigmaLength, model_c)
    return model

#void correlate(double* v1, int n1, double* v2, int n2, int* lag, int nlag, double* value)
correlate_c.argtypes = [c_double_p, c_int, c_double_p, c_int, c_int_p, c_int, c_double_p]
def correlate(v1, v2, lags):
    n1 = v1.size
    n2 = v2.size
    nlag = lags.size
    value = np.zeros(nlag)
    v1 = v1.astype(np.float64)
    v2 = v2.astype(np.float64)
    lags = lags.astype(np.int32)
    value = np.zeros(nlag, dtype=np.float64)
    v1_c    = np.ctypeslib.as_ctypes(v1)
    v2_c    = np.ctypeslib.as_ctypes(v2)
    lags_c  = np.ctypeslib.as_ctypes(lags)
    value_c = np.ctypeslib.as_ctypes(value)
    correlate_c(v1_c, n1, v2_c, n2, lags_c, nlag, value_c)
    return value

def maxCorrelate(v1, v2, lags, ind=-1):
    n1 = v1.size
    n2 = v2.size
    nlag = lags.size
    value = np.zeros(nlag)
    v1 = v1.astype(np.float64)
    v2 = v2.astype(np.float64)
    lags = lags.astype(np.int32)
    value = np.zeros(nlag, dtype=np.float64)
    v1_c    = np.ctypeslib.as_ctypes(v1)
    v2_c    = np.ctypeslib.as_ctypes(v2)
    lags_c  = np.ctypeslib.as_ctypes(lags)
    value_c = np.ctypeslib.as_ctypes(value)
    correlate_c(v1_c, n1, v2_c, n2, lags_c, nlag, value_c)
    #maxInd = np.argmax(value)
    maxInd = np.argsort(value)[ind]
    return value[maxInd], lags[maxInd]

#endregion: lamps

#region: about extract
def extract(fimage, invvar, xcen, sigma, configs={} ):
    #!! proftype is of no use
    #region: setups for all input
    fimage= fimage.copy().astype(np.float64) # copy is necessary, or we will get confused result
    invvar = invvar.copy().astype(np.float64)
    xcen   = xcen.copy().astype(np.float64)
    nTrace   = xcen.size
    lTrace   = np.arange(nTrace)
    if type(sigma)!=np.ndarray:
        sigma = np.ones(nTrace, dtype=np.float64) * sigma
    else:
        assert xcen.size==sigma.size
        sigma = sigma.astype(np.float64)

    sf   = configs.get("sf", np.ones(xcen.shape, dtype=np.float64))
    oA   = configs.get("oA", np.ones(xcen.shape, dtype=np.float64))
    oA3  = configs.get("oA3", np.zeros(xcen.shape, dtype=np.float64))
    oA21 = configs.get("oA21", np.zeros(xcen.shape, dtype=np.float64))
    oA22 = configs.get("oA22", np.zeros(xcen.shape, dtype=np.float64))
    assert fimage.shape == invvar.shape
    fimage[invvar==0]=0
    assert np.all(np.isfinite(fimage))
    assert np.all(np.isfinite(invvar))
    # if the length is too short, the c function will stop, follow the notice
    # profileMaxLength = configs.get("profileMaxLength", 15*np.max(sigma))
    niter = 0
    debug = configs.get("debug", 1)
    nx       = fimage.size
    x        = np.arange(nx, dtype=np.float64)
    sigmal   = configs.get('sigmal', 10.0)
    proftype = configs.get('proftype', 23)
    map2DReject = configs.get("map2DReject", 0.8)
    halfWidth = configs.get("halfWidth", 5)
    reducedChiOut = []

    if proftype not in [2, 3, 23, 32, 2323]:
        raise Exception("unknown proftype {}".format(proftype))
    cDebug   = configs.get('cDebug', 0)
    reject   = configs.get('reject', np.array([0.8, 0.8, 0.8])).astype(np.float64)
    if proftype in [2, 3, 2323]: #!! simple gaussian and **3 profile fitting
        fitmask = configs.get("fitmask", np.array([1,1,1,0]))
        maxMuDiff    = configs.get('maxMuDiff', 0.2)
    elif proftype in [23, 32]:
        fitmask = configs.get("fitmask", np.array([1,1,1,0]))
        if proftype==23:
            proftype = None
            info = extract(fimage, invvar, xcen, sigma,
                           {"proftype":2, "map2DMaxHalfWidth":0})
            niter += info["niter"]
            reducedChiOut.append(info["reducedChi"])
            xcen = info["xcen"]
            sigma2 = info["sigma"].astype(np.float64)
            info = extract(fimage, invvar, xcen, sigma,
                           {"proftype":3, "fitmask":np.array([1,1,0,0]), "map2DMaxHalfWidth":0 })
            niter += info["niter"]
            reducedChiOut.append(info["reducedChi"])
            sigma3 = info["sigma"].astype(np.float64)
            sf = sigma3/sigma2
            sigma = sigma2
            assert(np.all(np.isfinite(sf)))
        if proftype==32:
            proftype = None
            info = extract(fimage, invvar, xcen, sigma,
                          {"proftype":3, "map2DMaxHalfWidth":0})
            niter += info["niter"]
            reducedChiOut.append(info["reducedChi"])
            xcen = info["xcen"]
            sigma3 = info["sigma"].astype(np.float64)
            info= extract(fimage, invvar, xcen, sigma,
                          {"proftype":2, "fitmask":np.array([1,1,0,0]),"map2DMaxHalfWidth":0})
            niter += info["niter"]
            reducedChiOut.append(info["reducedChi"])
            sigma2 = info["sigma"].astype(np.float64)
            sf = sigma3/sigma2
            sigma = sigma2
            assert(np.all(np.isfinite(sf)))

    cDebug   = configs.get('cDebug', 0)
    reject   = configs.get('reject', np.array([0.8, 0.8, 0.8])).astype(np.float64)

    nCoeff = fitmask.size
    nPoly    = configs.get('nPoly', 4)
    ia       = np.concatenate((np.tile(fitmask, (nTrace,)), np.ones(nPoly))).astype(np.int64)
    calcCovar= 0
    # two parameters for converge
    mindMu   = configs.get('mindMu', 0.01)
    mindSigma= configs.get('mindSigma', 0.05)
    # contral the max step
    #maxSigmaDiff = configs.get('maxSigmaDiff', 1.0)
    sigmaMax = configs.get("sigmaMax", 2.5)
    sigmaMin = configs.get("sigmaMin", 1.0)
    #endregion: setups for all input
    #region: allocate memory for all outputs
    ma            = nTrace * nCoeff + nPoly
    fullreject    = np.zeros(nTrace, dtype=np.long)
    partialreject = np.zeros(nTrace, dtype=np.long)
    ans           = np.zeros(ma, dtype=np.float64)
    ymod          = np.zeros(nx, dtype=np.float64)
    amod          = np.zeros(nx, dtype=np.float64)
    bmod          = np.zeros(nx, dtype=np.float64)
    sumFlux       = np.zeros(nTrace, dtype=np.float64)
    sumFluxBack   = np.zeros(nTrace, dtype=np.float64)
    sumVar        = np.zeros(nTrace, dtype=np.float64)
    reject2D      = np.zeros(nTrace, dtype=np.float64)
    p             = np.zeros(ma, dtype=np.float64)
    covar         = np.zeros((ma, ma), dtype=np.float64)
    #endregion: allocate memory for all outputs
    #region: setup c function
    #long extract_wapper(
    #  //  xLen   xIndArr     yValue          invvar           xcen guess
    #    long nx, double* x, double* fimage, double* invvar, double* xcen,
    #    double* sigma, double* sf, double* oA, double* oA3, double* oA21, double* oA22,
    #  //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
    #  //    fiberNum   background    1,2 or 3
    #    long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
    #    long proftype, long calcCovar, double* reject, long* fullreject, long* partialreject,
    #    double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
    #    double* sumFlux, double* sumFluxBack, double* sumVar, double* reject2d,
    #    double* map2d, long halfWidth,
    #    double map2DReject, long debug)
    extract_wapper_c.argtypes =\
        [ c_long,  c_double_p,  c_double_p,     c_double_p,    c_double_p,
          c_double_p, c_double_p, c_double_p, c_double_p, c_double_p, c_double_p,
          c_long,    c_long,      c_long,     c_double,      c_long_p,
          c_long,       c_long,        c_double_p,        c_long_p,       c_long_p,
          c_double_p,   c_double_p,   ma*c_double_p,  c_double_p,  c_double_p, c_double_p,
          c_double_p, c_double_p, c_double_p, c_double_p,
          c_double_p, c_double_p, c_long, c_double, c_long ]
    extract_wapper_c.restype = c_long
    #endregion: setup c function
    #region: setup for iteratively
    maxiter = configs.get('maxiter', 20)
    if debug>0:
        ansDebug= np.tile(ans, (maxiter, 1))
        xcenDebug=np.zeros((maxiter+1, xcen.size))
        xcenDebug[0,:]=xcen[:]
        sigmaDebug=np.zeros((maxiter+1, sigma.size))
        sigmaDebug[0,:]=sigma[:]
        dmuDebug=np.zeros((maxiter, xcen.size))
        dsigmaDebug=np.zeros((maxiter, xcen.size))
    #!! caution here
    map2DLength = configs.get("map2DLength", 5*nx)
    if map2DLength:
        doDebug  = configs.get("doDebug", False)
        map2D = np.zeros(map2DLength, dtype=np.float64)
        map2D_c       = np.ctypeslib.as_ctypes(map2D)
        map2DOrigin   = np.zeros(map2DLength, dtype=np.float64)
        map2DOrigin_c = np.ctypeslib.as_ctypes(map2DOrigin)
    else:
        doDebug=0
        map2D = None
        map2D_c = None
        map2DOrigin = None
        map2DOrigin_c = None

    mask = invvar>1e-20; # assert mask.sum()>0
    if mask.sum()==0:
        allZeros = np.zeros(fimage.shape)
        allOnes = np.ones(fimage.shape)
        return {"niter":0, "reducedChi":np.inf, "ymod":allZeros,
                  "bmod":allZeros,"map2D":map2D, "sigma":allOnes*np.inf, "xcen":xcen,
                  "map2DOrigin":map2DOrigin, "reducedChiOut":[np.inf],
                  "sumFlux":sumFlux, "sumFluxBack": sumFluxBack, "sumVar": sumVar,
                  "reject2D":reject2D, "amod":allZeros, "flux":np.zeros(xcen.shape),
                  "invvar":np.zeros(xcen.shape)}

    #fimage = fimage * mask
    xcenOrigin    = xcen.copy()
    sigmaOrigin   = sigma.copy()
    # unchanged value
    #        double* sumFlux, double* sumFluxBack, double* sumVar,
    fimage_c        = np.ctypeslib.as_ctypes(fimage)
    x_c             = np.ctypeslib.as_ctypes(x)
    ia_c            = np.ctypeslib.as_ctypes(ia)
    reject_c        = np.ctypeslib.as_ctypes(reject)
    xFib            = np.arange(nTrace)*nCoeff
    workinvvar      = invvar * mask
    workinvvar_c    = np.ctypeslib.as_ctypes(workinvvar)
    sigma_c         = np.ctypeslib.as_ctypes(sigma)
    sf_c         = np.ctypeslib.as_ctypes(sf)
    oA_c         = np.ctypeslib.as_ctypes(oA)
    oA3_c        = np.ctypeslib.as_ctypes(oA3)
    oA21_c        = np.ctypeslib.as_ctypes(oA21)
    oA22_c        = np.ctypeslib.as_ctypes(oA22)
    xcen_c          = np.ctypeslib.as_ctypes(xcen)
    fullreject_c    = np.ctypeslib.as_ctypes(fullreject)
    partialreject_c = np.ctypeslib.as_ctypes(partialreject)
    ans_c           = np.ctypeslib.as_ctypes(ans)
    ymod_c          = np.ctypeslib.as_ctypes(ymod)
    amod_c          = np.ctypeslib.as_ctypes(amod)
    sumFlux_c       = np.ctypeslib.as_ctypes(sumFlux)
    sumFluxBack_c   = np.ctypeslib.as_ctypes(sumFluxBack)
    sumVar_c        = np.ctypeslib.as_ctypes(sumVar)
    reject2D_c      = np.ctypeslib.as_ctypes(reject2D)
    covar_c         = p3(covar)
    p_c             = np.ctypeslib.as_ctypes(p)
    bmod_c          = np.ctypeslib.as_ctypes(bmod)
    #endregion: setup for iteratively
    if proftype in [2, 3, 2323]:# for single profile
        #region: call c function iteratively
        while niter<maxiter :
            # make sure these may-be-changed variables are all 0s
            fullreject[:]=0; partialreject[:]=0; ans[:]=0;
            ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; bmod[:]=0;
            sumFlux[:]=0; sumFluxBack[:]=0; sumVar[:]=0
            #!! key function call
            result = \
            extract_wapper_c(
              c_long(nx),        x_c,       fimage_c, workinvvar_c,     xcen_c,
              sigma_c, sf_c, oA_c, oA3_c, oA21_c, oA22_c,
              c_long(nTrace),    c_long(nPoly),  c_long(nCoeff), c_double(sigmal), ia_c,
              c_long(proftype),  c_long(calcCovar),    reject_c,       fullreject_c,     partialreject_c,
              ans_c,       ymod_c,     covar_c,     p_c,   bmod_c, amod_c,
              sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
              None, None, c_long(halfWidth), c_double(map2DReject), c_long(cDebug))
            #region: test ans
            assert result==1
            assert np.all(np.isfinite(ans))
            diffs = (fimage-ymod) * np.sqrt(workinvvar)
            chisq = np.sum(diffs**2)
            count = ia.sum()
            good = workinvvar>1e-10; assert good.sum()>0
            reducedChi = chisq/(np.sum(good) - count)
            #endregion: test ans
            #region: make result and change for step
            A      = ans[xFib]#!! ans[0 3 6...]
            A[A==0]= 1
            badMask  = A<0
            dSigma = ans[xFib+1]/A
            dMu    = ans[xFib+2]/A

            sigmaHigh     = (sigma+dSigma) > sigmaMax
            sigmaLow      = (sigma+dSigma) < sigmaMin
            sigmaNormal   = np.logical_and(np.logical_not(sigmaHigh), np.logical_not(sigmaLow))
            sigma[sigmaNormal] += dSigma[sigmaNormal]
            sigma[sigmaHigh]    = sigmaMax
            sigma[sigmaLow]     = sigmaMin
            sigma[badMask]      = sigmaOrigin[badMask]

            muHigh     = (xcen+dMu-xcenOrigin) >  maxMuDiff
            muLow      = (xcen+dMu-xcenOrigin) < -maxMuDiff
            muNormal   = np.logical_and(np.logical_not(muHigh), np.logical_not(muLow))
            xcen[muNormal] += dMu[muNormal]
            xcen[muHigh]    = xcenOrigin[muHigh] + maxMuDiff
            xcen[muLow]     = xcenOrigin[muLow]  - maxMuDiff
            xcen[badMask]   = xcenOrigin[badMask]

            if debug>0:
                ansDebug[niter-1, :] = ans
                dsigmaDebug[niter,:] = dSigma[:]
                dmuDebug[niter,:] = dMu[:]
                xcenDebug[niter+1,:]  = xcen[:]
                sigmaDebug[niter+1,:] = sigma[:]
                #finish = np.all(np.abs(dSigma)<mindSigma) and np.all(np.abs(dMu)<mindMu)

            niter += 1
            #endregion: make result and change for step
        #endregion: call c function iteratively
        #region: handle last fit for single profile
        thisia = np.concatenate((np.tile(np.array([1,0,0,0]), (nTrace,)),
                              np.ones(nPoly))).astype(np.int64)
        thisia_c = np.ctypeslib.as_ctypes(thisia)
        fullreject[:]=0; partialreject[:]=0; ans[:]=0;
        ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; bmod[:]=0
        sumFlux[:]=0; sumFluxBack[:]=0; sumVar[:]=0
        result = \
            extract_wapper_c(
                c_long(nx), x_c, fimage_c, workinvvar_c, xcen_c,
                sigma_c, sf_c, oA_c, oA3_c, oA21_c, oA22_c,
                c_long(nTrace), c_long(nPoly), c_long(nCoeff), c_double(sigmal), thisia_c,
                c_long(proftype), c_long(calcCovar), reject_c, fullreject_c, partialreject_c,
                ans_c,       ymod_c, covar_c, p_c,   bmod_c, amod_c,
                sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
                map2D_c, map2DOrigin_c, c_long(halfWidth), c_double(map2DReject), c_long(cDebug))
        if not (result==1 and np.all(np.isfinite(ans))):
            raise
        diffs = (fimage-ymod) * np.sqrt(workinvvar)
        chisq = np.sum(diffs**2)
        count = ia.sum()
        good = workinvvar>1e-10; assert good.sum()>0
        reducedChi = chisq/(np.sum(good) - count)
        # make result and change for step
        #endregion: handle last fit for defferent type
        #region: gen output and return
        A   = ans[xFib]
        finvrow = p[xFib]*p[xFib]
        bad = np.where(np.logical_or(~np.isfinite(A), finvrow<1e-16, fullreject))
        finvrow[bad]=0.0
        result = {"niter":niter, "reducedChi":reducedChi, "ymod":ymod,
                  "bmod":bmod,"map2D":map2D, "sigma":sigma, "xcen":xcen,
                  "map2DOrigin":map2DOrigin, "reducedChiOut":[reducedChi],
                  "sumFlux":sumFlux, "sumFluxBack": sumFluxBack, "sumVar": sumVar,
                  "reject2D":reject2D, "amod":amod, "flux":A, "invvar":finvrow}
        if debug>0:
            result.update({"dmuDebug":dmuDebug, "dsigmaDebug":dsigmaDebug,
                "xcenDebug":xcenDebug, "sigmaDebug":sigmaDebug})
        return result
        #endregion: gen output and return
    elif proftype is None:
        #region: define A2 and A3
        thisia = np.concatenate((np.tile(np.array([1,1,0,0]), (nTrace,)),
                              np.ones(nPoly))).astype(np.int64)
        thisia_c = np.ctypeslib.as_ctypes(thisia)
        fullreject[:]=0; partialreject[:]=0; ans[:]=0;
        ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; bmod[:]=0;
        sumFlux[:]=0; sumFluxBack[:]=0; sumVar[:]=0
        #!! key function call
        result = \
        extract_wapper_c(
          c_long(nx),        x_c,       fimage_c, workinvvar_c,     xcen_c,
          sigma_c, sf_c, oA_c, oA3_c, oA21_c, oA22_c,
          c_long(nTrace),    c_long(nPoly),  c_long(nCoeff), c_double(sigmal), thisia_c,
          c_long(2233),  c_long(calcCovar),    reject_c,       fullreject_c,     partialreject_c,
          ans_c,       ymod_c,     covar_c,     p_c,   bmod_c, amod_c,
          sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
          None, None, c_long(halfWidth), c_double(map2DReject), c_long(cDebug))
        if not (result==1 and np.all(np.isfinite(ans))):
            raise
        toA = ans[xFib]
        toA3 = ans[xFib+1]
        good = np.logical_or(toA>0, toA3>0)
        oA[good] = toA[good]
        oA3[good] = toA3[good]
        #endregion: define A2 and A3
        #region: refine xcen
        info = extract(fimage, invvar, xcen, sigma,
                       {"proftype":2323, "map2DMaxHalfWidth":0,
                           "oA":oA, "oA3":oA3, "sf":sf, 'maxMuDiff':0.1 })
        xcen[:] = info["xcen"][:]
        sigma[:] = info["sigma"][:]
        niter += info["niter"]
        reducedChiOut.append(info["reducedChi"])
        #endregion: refine xcen
        #region: define oA, oA3, oA21, oA22 (do not use...)
        do2345 = False
        if do2345:#!! do not use 22334455 profile
            thisia = np.concatenate((np.tile(np.array([1,1,1,1]), (nTrace,)),
                                  np.ones(nPoly))).astype(np.int64)
            thisia_c = np.ctypeslib.as_ctypes(thisia)
            fullreject[:]=0; partialreject[:]=0; ans[:]=0;
            ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; bmod[:]=0;
            sumFlux[:]=0; sumFluxBack[:]=0; sumVar[:]=0
            result = \
            extract_wapper_c(
              c_long(nx),        x_c,       fimage_c, workinvvar_c,     xcen_c,
              sigma_c, sf_c, oA_c, oA3_c, oA21_c, oA22_c,
              c_long(nTrace),    c_long(nPoly),  c_long(nCoeff), c_double(sigmal), thisia_c,
              c_long(22334455),  c_long(calcCovar),    reject_c,       fullreject_c,     partialreject_c,
              ans_c,       ymod_c,     covar_c,     p_c,   bmod_c, amod_c,
              sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
              #None, None, c_long(halfWidth), c_double(map2DReject), c_long(cDebug))
              None, None, c_long(halfWidth), c_double(map2DReject), c_long(cDebug))
            if not (result==1 and np.all(np.isfinite(ans))):
                raise
            toA = ans[xFib]
            toA3 = ans[xFib+1]
            toA21 = ans[xFib+2]
            toA22 = ans[xFib+3]
            good = (toA + toA + toA21 + toA22)>0
            oA[good] = toA[good]
            oA3[good] = toA3[good]
            oA21[good] = toA21[good]
            oA22[good] = toA22[good]
        #endregion: define oA, oA3, oA21, oA22
        #region: handle last fit
        thisia = np.concatenate((np.tile(np.array([1,0,0,0]), (nTrace,)),
                              np.ones(nPoly))).astype(np.int64)
        thisia_c = np.ctypeslib.as_ctypes(thisia)
        fullreject[:]=0; partialreject[:]=0; ans[:]=0;
        ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; bmod[:]=0
        sumFlux[:]=0; sumFluxBack[:]=0; sumVar[:]=0
        result = \
            extract_wapper_c(
                c_long(nx), x_c, fimage_c, workinvvar_c, xcen_c,
                sigma_c, sf_c, oA_c, oA3_c, oA21_c, oA22_c,
                c_long(nTrace), c_long(nPoly), c_long(nCoeff), c_double(sigmal), thisia_c,
                c_long(2323), c_long(calcCovar), reject_c, fullreject_c, partialreject_c,
                ans_c, ymod_c, covar_c, p_c, bmod_c, amod_c,
                sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
                map2D_c, map2DOrigin_c, c_long(halfWidth), c_double(map2DReject), c_long(cDebug))
        if not (result==1 and np.all(np.isfinite(ans))):
            raise
        diffs = (fimage-ymod) * np.sqrt(workinvvar)
        chisq = np.sum(diffs**2)
        count = ia.sum()
        good = workinvvar>1e-10; assert good.sum()>0
        reducedChi = chisq/(np.sum(good) - count)
        reducedChiOut.append(reducedChi)
        # make result and change for step
        #endregion: handle last fit
        #region: gen output and return
        A   = ans[xFib]
        finvrow = p[xFib]*p[xFib]
        bad = np.where(np.logical_or(~np.isfinite(A), finvrow<1e-16, fullreject))
        finvrow[bad]=0.0
        result = {"niter":niter, "reducedChi":reducedChi, "ymod":ymod,
                  "bmod":bmod,"map2D":map2D, "sigma":sigma, "xcen":xcen,
                  "map2DOrigin":map2DOrigin, "reducedChiOut":reducedChiOut,
                  "sumFlux":sumFlux, "sumFluxBack": sumFluxBack, "sumVar": sumVar,
                  "reject2D":reject2D, "amod":amod, "flux":A, "invvar":finvrow}
        if debug>0:
            result.update({"dmuDebug":dmuDebug, "dsigmaDebug":dsigmaDebug,
                           "xcenDebug":xcenDebug, "sigmaDebug":sigmaDebug})
        return result
        #endregion: gen output and return
    else:
        raise Exception("should not be here!!!")

def map2DSlice(bmod, map2D, flux, invvar):
    assert flux.size==invvar.size
    nx = bmod.size
    bmod   = bmod.copy().astype(np.float64)
    map2D  = map2D.copy().astype(np.float64)
    result = np.zeros(bmod.shape, dtype=np.float64)
    flux   = flux.copy().astype(np.float64)
    invvar = invvar.copy().astype(np.float64)
    map2DSlice_c.argtypes =[\
        c_long, c_double_p, c_double_p, c_double_p,
        c_double_p, c_double_p]
    result_c = np.ctypeslib.as_ctypes(result)
    bmod_c   = np.ctypeslib.as_ctypes(bmod)
    map2D_c  = np.ctypeslib.as_ctypes(map2D)
    flux_c   = np.ctypeslib.as_ctypes(flux)
    invvar_c = np.ctypeslib.as_ctypes(invvar)
    map2DSlice_c(nx, result_c, bmod_c, map2D_c, flux_c, invvar_c)
    return result

#map2DSliceFixNaNWidth(long nx, double result[], double bmod[], double map2D[], double flux[],
#                      double invvar[], double yset[], double badValue, int width)
def map2DSliceFixNaNWidth(bmod, map2D, flux, invvar, yset, badValue=np.nan, halfwidth=5):
    assert flux.size==invvar.size==yset.size
    nx = bmod.size
    bmod   = bmod.copy().astype(np.float64)
    map2D  = map2D.copy().astype(np.float64)
    result = np.zeros(bmod.shape, dtype=np.float64)
    flux   = flux.copy().astype(np.float64)
    yset   = yset.copy().astype(np.float64)
    invvar = invvar.copy().astype(np.float64)
    map2DSliceFixNaNWidth_c.argtypes =[\
        c_long, c_double_p, c_double_p, c_double_p,
        c_double_p, c_double_p, c_double_p, c_double, c_int]
    result_c = np.ctypeslib.as_ctypes(result)
    bmod_c   = np.ctypeslib.as_ctypes(bmod)
    map2D_c  = np.ctypeslib.as_ctypes(map2D)
    flux_c   = np.ctypeslib.as_ctypes(flux)
    yset_c   = np.ctypeslib.as_ctypes(yset)
    invvar_c = np.ctypeslib.as_ctypes(invvar)
    map2DSliceFixNaNWidth_c(nx, result_c, bmod_c, map2D_c, flux_c, invvar_c, yset_c, c_double(badValue), c_int(halfwidth))
    return result

def map2DCutFixWidth(fimage, fstd, yset, waveset, halfwidth=5):
    assert fimage.shape == fstd.shape
    assert fimage.shape[1] == yset.shape[1]
    fimage = fimage.astype(np.float64)
    fstd   = fstd.astype(np.float64)
    yset   = yset.astype(np.float64)
    waveset = waveset.astype(np.float64)
    width  = 2*halfwidth + 1
    ny, nx = fimage.shape
    nfiber, nx = yset.shape
    cutFlux = np.zeros((nfiber*width, nx))
    cutFstd = np.zeros((nfiber*width, nx))
    cutYset = np.zeros((nfiber*width, nx))
    cutFlux_c = np.ctypeslib.as_ctypes(cutFlux)
    cutFstd_c = np.ctypeslib.as_ctypes(cutFstd)
    cutYset_c = np.ctypeslib.as_ctypes(cutYset)
    fimage_c = np.ctypeslib.as_ctypes(fimage)
    fstd_c = np.ctypeslib.as_ctypes(fstd)
    yset_c = np.ctypeslib.as_ctypes(yset)
    cutClass = cutFlux_c.__class__
    imageClass = fimage_c.__class__
    fluxClass = yset_c.__class__
    map2DCutFixWidth_c.argtypes = [c_long, c_long, c_long, c_long,
                                   imageClass, imageClass,
                                   fluxClass,
                                   cutClass, cutClass, cutClass]
    map2DCutFixWidth_c(c_long(nx), c_long(ny), c_long(nfiber), c_long(width),
                       fimage_c, fstd_c,
                       yset_c,
                       cutFlux_c, cutFstd_c, cutYset_c)
    cutWaveset = np.zeros((nfiber*width, nx))
    for i in range(width):
        cutWaveset[i::width,:] = waveset[:,:]
    result = {"cutFlux":cutFlux, "cutFstd":cutFstd, "cutYset":cutYset,
              "cutWaveset":cutWaveset, "width":width}
    return result

def map2DCutFixWidthSingle(fimage, yset, halfwidth=5):
    assert fimage.shape[1] == yset.shape[1]
    fimage = fimage.astype(np.float64)
    yset   = yset.astype(np.float64)
    width  = 2*halfwidth + 1
    ny, nx = fimage.shape
    nfiber, nx = yset.shape
    cutFlux = np.zeros((nfiber*width, nx))
    cutFstd = np.zeros((nfiber*width, nx))
    cutYset = np.zeros((nfiber*width, nx))
    cutFlux_c = np.ctypeslib.as_ctypes(cutFlux)
    cutFstd_c = np.ctypeslib.as_ctypes(cutFstd)
    cutYset_c = np.ctypeslib.as_ctypes(cutYset)
    fimage_c = np.ctypeslib.as_ctypes(fimage)
    fstd_c = fimage_c
    yset_c = np.ctypeslib.as_ctypes(yset)
    cutClass = cutFlux_c.__class__
    imageClass = fimage_c.__class__
    fluxClass = yset_c.__class__
    map2DCutFixWidth_c.argtypes = [c_long, c_long, c_long, c_long,
                                   imageClass, imageClass,
                                   fluxClass,
                                   cutClass, cutClass, cutClass]
    map2DCutFixWidth_c(c_long(nx), c_long(ny), c_long(nfiber), c_long(width),
                       fimage_c, fstd_c,
                       yset_c,
                       cutFlux_c, cutFstd_c, cutYset_c)
    result = {"cutFlux":cutFlux, "cutYset":cutYset, "width":width}
    return result


def remapFlux(cutFlux, cutFstd, cutXset, cutYset, width, reXset, reYset):
    #region: convert type and init
    cutFlux = cutFlux.astype(np.float64)
    cutYset = cutYset.astype(np.float64)
    cutXset = cutXset.astype(np.float64)
    reXset  = reXset.astype(np.float64)
    reYset  = reYset.astype(np.float64)
    bny, bnx = cutFlux.shape
    sny, snx = reXset.shape
    bN = np.zeros((bny, bnx))
    bE = np.zeros((bny, bnx))
    bW = np.zeros((bny, bnx))
    bS = np.zeros((bny, bnx))
    sN = np.zeros((sny, snx))
    sE = np.zeros((sny, snx))
    sW = np.zeros((sny, snx))
    sS = np.zeros((sny, snx))
    assert bny%width==0
    nFiber = int(bny/width)
    resultFlux = np.zeros((sny*nFiber, snx))
    resultVar = np.zeros((sny*nFiber, snx))
    sFillingFactor = np.zeros((sny*nFiber, snx))
    #endregion: convert type and init
    #region: calculate boundarys for each pixel
    bxMid = (cutXset[:,:-1]+cutXset[:,1:])/2
    bW[:,1:] = bxMid[:,:]
    bW[:,0] = cutXset[:,0]-(bxMid[:,0]-cutXset[:,0])
    bE[:,:-1] = bxMid[:,:]
    bE[:,-1] = cutXset[:,-1]+(cutXset[:,-1]-bxMid[:,-1])
    sxMid = (reXset[:,:-1]+reXset[:,1:])/2
    sW[:,1:] = sxMid[:,:]
    sW[:,0] = reXset[:,0]-(sxMid[:,0]-reXset[:,0])
    sE[:,:-1] = sxMid[:,:]
    sE[:,-1] = reXset[:,-1]+(reXset[:,-1]-sxMid[:,-1])

    syMid = (reYset[:-1,:]+reYset[1:,:])/2
    sS[1:,:] = syMid[:,:]
    sS[0,:] = reYset[0,:]-(syMid[0,:]-reYset[0,:])
    sN[:-1,:] = syMid[:,:]
    sN[-1,:] = reYset[-1,:]+(reYset[-1,:]-syMid[-1,:])
    for i in range(nFiber):
        ts = i*width
        te = (i+1)*width
        #byMid = (cutYset[:-1,:]+cutYset[1:,:])/2
        byMid = (cutYset[ts:te-1,:]+cutYset[ts+1:te,:])/2
        #bS[1:,:] = byMid[:,:]
        bS[ts+1:te,:] = byMid[:,:]
        #bS[0,:] = cutYset[0,:]-(byMid[0,:]-cutYset[0,:])
        bS[ts,:] = cutYset[ts,:]-(byMid[0,:]-cutYset[ts,:])
        #bN[:-1,:] = byMid[:,:]
        bN[ts:te-1,:] = byMid[:,:]
        #bN[-1,:] = cutYset[-1,:]+(cutYset[-1,:]-byMid[-1,:])
        bN[te-1,:] = cutYset[te-1,:]+(cutYset[te-1,:]-byMid[-1,:])
    #endregion: calculate boundarys for each pixel
    #region: make ctypes
    cutFlux_c = np.ctypeslib.as_ctypes(cutFlux)
    cutFstd_c = np.ctypeslib.as_ctypes(cutFstd)
    bW_c = np.ctypeslib.as_ctypes(bW)
    bE_c = np.ctypeslib.as_ctypes(bE)
    bN_c = np.ctypeslib.as_ctypes(bN)
    bS_c = np.ctypeslib.as_ctypes(bS)
    sW_c = np.ctypeslib.as_ctypes(sW)
    sE_c = np.ctypeslib.as_ctypes(sE)
    sN_c = np.ctypeslib.as_ctypes(sN)
    sS_c = np.ctypeslib.as_ctypes(sS)
    resultFlux_c = np.ctypeslib.as_ctypes(resultFlux)
    resultVar_c = np.ctypeslib.as_ctypes(resultVar)
    sFillingFactor_c = np.ctypeslib.as_ctypes(sFillingFactor)
    bigClass = bW_c.__class__
    smallClass = sW_c.__class__
    allSmallClass = resultFlux_c.__class__
    remapFlux_c.argtypes = [bigClass, bigClass,
                            bigClass,     bigClass,   bigClass,   bigClass, c_long, c_long, c_long,
                            smallClass, smallClass, smallClass, smallClass, c_long, c_long,
                            allSmallClass, allSmallClass, allSmallClass]
    remapFlux_c(cutFlux_c, cutFstd_c,
                bN_c, bS_c, bW_c, bE_c, c_long(bnx), c_long(bny), c_long(width),
                sN_c, sS_c, sW_c, sE_c, c_long(snx), c_long(sny),
                resultFlux_c, resultVar_c, sFillingFactor_c)
    fillingCount = np.logical_and(np.isfinite(resultVar), resultVar>0)
    return {"mapFlux":resultFlux, "mapVar":resultVar, "mapCount":fillingCount}
    #endregion: make ctypes

def drawPixel(n, s, w, e, width=11, nFiber=1, xs=0, xe=100, color="red"):
    nn = n[:width*nFiber,xs:xe].flatten()
    ss = s[:width*nFiber,xs:xe].flatten()
    ww = w[:width*nFiber,xs:xe].flatten()
    ee = e[:width*nFiber,xs:xe].flatten()
    #nw, ne, se, sw
    #nnss = np.concatenate((nn,nn,ss,ss,nn))
    #weew = np.concatenate((ww,ee,ee,ww,ww))
    nnss = np.concatenate((nn,nn,ss,ss,nn)).reshape(5,nn.size).T
    weew = np.concatenate((ww,ee,ee,ww,ww)).reshape(5,ee.size).T
    de.p(weew.T, nnss.T, c=color)


#endregion: about extract

#region: about wavecalib
#void getSubIndex(long ii, long* nsteps, long n, long* result)
getSubIndex_c.argtypes = [c_long, c_long_p, c_long, c_long_p]
def getSubIndex(ii, nsteps):
    n = nsteps.size
    result = np.zeros(n, dtype=np.int64)
    nsteps = nsteps.astype(np.int64)
    nsteps_c = np.ctypeslib.as_ctypes(nsteps)
    result_c = np.ctypeslib.as_ctypes(result)
    getSubIndex_c(ii, nsteps_c, n, result_c)
    return result

def gaussianFit(fimage, sig,
                xcen, xcenMin, xcenMax,
                sigma, sigmaMin, sigmaMax, initA=None, x=None, configs={} ):
    #region: setup for all inputs
    fimage   = fimage.copy().astype(np.float64) # copy is necessary, or we will get confused result
    sig      = sig.copy().astype(np.float64)
    xcen     = xcen.copy().astype(np.float64)
    xcenMin  = xcenMin.copy().astype(np.float64)
    xcenMax  = xcenMax.copy().astype(np.float64)
    sigma    = sigma.copy().astype(np.float64)
    sigmaMin = sigmaMin.copy().astype(np.float64)
    sigmaMax = sigmaMax.copy().astype(np.float64)
    debugStop = configs.get("debugStop", False)
    if initA is not None:
        assert np.all(initA>0);
        initLogA = np.log(initA).copy().astype(np.float64)
    else:
        defaultLogA = configs.get("logA", 8)
        initLogA = (np.ones(xcen.shape)*defaultLogA).astype(np.float64)
    assert fimage.shape == sig.shape
    assert np.all(np.isfinite(fimage))
    nx       = fimage.size
    if x is None:
        x = np.arange(nx, dtype=np.float64)
    else:
        x = x.astype(np.float64)
    nTrace   = xcen.size
    cDebug   = configs.get('cDebug', 0)
    reject   = configs.get('reject', 0.7)
    nPoly    = configs.get('nPoly' , 2)
    maxiter  = configs.get('maxiter', 1000)
    mindchi2 = configs.get('mindchi2', 0.0001)
    finishGoodCount = configs.get("finishGoodCount", 5)
    dlamda   = configs.get("dlamda", 10.0)
    sigmal   = configs.get("sigmal", 3)
    ignoreBadFit = configs.get("ignoreBadFit", 0)
    nCoeff   = 3
    tTrace   = nTrace * nCoeff
    #endregion: setup for all inputs
    #region: allocate memory for all outputs
    ma         = nTrace * nCoeff + nPoly
    gmod       = np.zeros(nx, dtype=np.float64)
    bmod       = np.zeros(nx, dtype=np.float64)
    fullreject = np.zeros(nTrace, dtype=np.int32)
    ans        = np.zeros(ma, dtype=np.float64)
    p          = np.zeros(tTrace, dtype=np.float64)
    ia         = np.ones( tTrace, dtype=np.int32)
    chi2       = np.zeros(nTrace+1, dtype=np.float64)
    niter      = np.zeros(nTrace+1, dtype=np.int32)
    bMin       = np.zeros(nTrace, dtype=np.int32)
    bMax       = np.zeros(nTrace, dtype=np.int32)
    gaussianFitIter_c.argtypes =[\
        c_int, c_double_p, c_double_p, c_double_p, c_double_p,
        c_double_p, c_double_p, c_double_p,
        c_double_p, c_double_p, c_double_p,
        c_int, c_int, c_int_p, c_double, c_double, c_int,
        c_double_p, c_double_p, c_int_p, c_int,
        c_double_p, c_double_p, c_double_p,
        c_int_p, c_int_p, c_int_p,
        c_double, c_int, c_double, c_int]
    gaussianFitIter_c.restype = c_int
    #endregion: allocate memory for all outputs
    #region: setup for iteratively
    fimage_c     = np.ctypeslib.as_ctypes(fimage)
    x_c          = np.ctypeslib.as_ctypes(x)
    sig_c        = np.ctypeslib.as_ctypes(sig)
    sigma_c      = np.ctypeslib.as_ctypes(sigma)
    sigmaMin_c   = np.ctypeslib.as_ctypes(sigmaMin)
    sigmaMax_c   = np.ctypeslib.as_ctypes(sigmaMax)
    xcen_c       = np.ctypeslib.as_ctypes(xcen)
    xcenMin_c    = np.ctypeslib.as_ctypes(xcenMin)
    xcenMax_c    = np.ctypeslib.as_ctypes(xcenMax)
    fullreject_c = np.ctypeslib.as_ctypes(fullreject)
    ia_c         = np.ctypeslib.as_ctypes(ia)
    ans_c        = np.ctypeslib.as_ctypes(ans)
    gmod_c       = np.ctypeslib.as_ctypes(gmod)
    bmod_c       = np.ctypeslib.as_ctypes(bmod)
    p_c          = np.ctypeslib.as_ctypes(p)
    chi2_c       = np.ctypeslib.as_ctypes(chi2)
    niter_c      = np.ctypeslib.as_ctypes(niter)
    initLogA_c   = np.ctypeslib.as_ctypes(initLogA)
    bMin_c       = np.ctypeslib.as_ctypes(bMin)
    bMax_c       = np.ctypeslib.as_ctypes(bMax)
    #endregion: setup for iteratively
    #region: call c function iteratively
    good = \
    gaussianFitIter_c(\
        nx, x_c, fimage_c, sig_c, initLogA_c,
        xcen_c, xcenMin_c, xcenMax_c,
        sigma_c, sigmaMin_c, sigmaMax_c,
        nTrace, nPoly, fullreject_c, reject, sigmal, ignoreBadFit,
        ans_c, p_c, ia_c, maxiter,
        gmod_c, bmod_c, chi2_c, bMin_c, bMax_c, niter_c,
        mindchi2, finishGoodCount, dlamda, cDebug)
    ans = ans[:tTrace]
    p = p[:tTrace]
    if debugStop:
        ipdb.set_trace()

    A = np.exp(ans[::3])*(1-fullreject)
    mu = ans[1::3]
    sigma = ans[2::3]
    fullreject = fullreject.astype(np.bool)
    Asig = np.zeros(fullreject.shape)
    AExpSig = np.zeros(fullreject.shape)
    muSig = np.zeros(fullreject.shape)
    sigSig = np.zeros(fullreject.shape)
    peakSNRIndex = np.round(mu).astype(np.int)
    peakSNRIndex[peakSNRIndex<0]=0
    peakSNRIndex[peakSNRIndex>nx-1]=nx-1
    peakSNR = gmod[peakSNRIndex]/bmod[peakSNRIndex]
    peakSigSNR = gmod[peakSNRIndex]/sig[peakSNRIndex]
    peakSNR[~np.isfinite(peakSNR)] = 0
    peakSNR[peakSNR<0] = 0


    #Asig = A*(np.exp(np.sqrt(p[::3]))-1);Asig[fullreject]=np.inf
    #AExpSig = np.sqrt(p[::3]);AExpSig[fullreject]=np.inf
    #muSig = np.sqrt(p[1::3]);muSig[fullreject]=np.inf
    #sigSig = np.sqrt(p[2::3]);sigSig[fullreject]=np.inf
    Ap   = p[::3]
    mup  = p[1::3]
    sigp = p[2::3]
    try:
        Asig[~fullreject] = A[~fullreject]*(np.exp(np.sqrt(Ap[~fullreject]))-1);Asig[fullreject]=np.inf
        AExpSig[~fullreject] = np.sqrt(Ap[~fullreject]);AExpSig[fullreject]=np.inf
        muSig[~fullreject] = np.sqrt(mup[~fullreject]);muSig[fullreject]=np.inf
        sigSig[~fullreject] = np.sqrt(sigp[~fullreject]);sigSig[fullreject]=np.inf
        #if not isinstance(chi2[0], int):
            #print "Warning!!!!!! chi2[0]={} size:{}".format(chi2[0], chi2.shape)
        reduceChi2 = chi2[1:int(chi2[0])+1]
    except:
        print("some bug in gaussianFit... need to fix it")
    if cDebug>0:
        print("niter:", niter)
        print("chi2:", chi2)
        print("bMin:", bMin)
        print("bMax:", bMax)
        print("p:", p.size)
        print(p)
        print("fullreject", fullreject)
    #endregion: call c function iteratively
    return A, mu, sigma, reduceChi2, {"gmod":gmod, "bmod":bmod, "AExpSig":AExpSig,
                        "niter":niter, "Asig":Asig, "muSig":muSig, "sigSig":sigSig,
                        "fullreject":fullreject, "peakSNR":peakSNR, "peakSigSNR":peakSigSNR}
#endregion: about wavecalib


def profileGaussianGen(nPoly):
    def profile(x, *opts):
        N = int((len(opts)-nPoly)/3)
        mu = np.array(opts[:N], dtype=np.float64)
        A = np.array(opts[N:2*N], dtype=np.float64)
        sigma = np.array(opts[2*N:3*N], dtype=np.float64)
        poly = opts[3*N:]

        x = x.astype(np.float64)
        n = x.size
        result = np.zeros(x.shape, dtype=np.float64)
        x_c = np.ctypeslib.as_ctypes(x)
        result_c = np.ctypeslib.as_ctypes(result)
        A_c = np.ctypeslib.as_ctypes(A)
        mu_c = np.ctypeslib.as_ctypes(mu)
        sigma_c = np.ctypeslib.as_ctypes(sigma)

        profileGaussian_c.argtypes = [
                c_int, c_double_p,
                c_double_p, c_long,
                c_double_p, c_double_p,
                c_double_p]
        profileGaussian_c(n, x_c,
                     mu_c, c_long(N),
                     A_c, sigma_c,
                     result_c)
        if len(poly):
            result += Legendre(poly, domain=[x.min(), x.max()])(x)
        return result
    return profile

def profileGen(nPoly):
    def profile(x, *opts):
        N = int((len(opts)-nPoly)/5)
        mu = np.array(opts[:N], dtype=np.float64)
        A = np.array(opts[N:2*N], dtype=np.float64)
        sigma = np.array(opts[2*N:3*N], dtype=np.float64)
        A3 = np.array(opts[3*N:4*N], dtype=np.float64)
        sigma3 = np.array(opts[4*N:5*N], dtype=np.float64)
        poly = opts[5*N:]

        x = x.astype(np.float64)
        n = x.size
        result = np.zeros(x.shape, dtype=np.float64)
        x_c = np.ctypeslib.as_ctypes(x)
        result_c = np.ctypeslib.as_ctypes(result)
        A_c = np.ctypeslib.as_ctypes(A)
        mu_c = np.ctypeslib.as_ctypes(mu)
        sigma_c = np.ctypeslib.as_ctypes(sigma)
        A3_c = np.ctypeslib.as_ctypes(A3)
        sigma3_c = np.ctypeslib.as_ctypes(sigma3)

        profile233_c.argtypes = [
                c_int, c_double_p,
                c_double_p, c_long,
                c_double_p, c_double_p,
                c_double_p, c_double_p,
                c_double_p]
        profile233_c(n, x_c,
                     mu_c, c_long(N),
                     A_c, sigma_c,
                     A3_c, sigma3_c,
                     result_c)
        if len(poly):
            result += Legendre(poly, domain=[x.min(), x.max()])(x)
        return result
    return profile

def profileGenSP(nPoly):
    def profile(x, *opts):
        N = int((len(opts)-nPoly)/5)
        mu = np.array(opts[:N], dtype=np.float64)
        A = np.array(opts[N:2*N], dtype=np.float64)
        sigma = np.array(opts[2*N:3*N], dtype=np.float64)
        A3 = np.array(opts[3*N:4*N], dtype=np.float64)
        sigma3 = np.array(opts[4*N:5*N], dtype=np.float64)
        poly = opts[5*N:]

        x = x.astype(np.float64)
        n = x.size
        result = np.zeros(x.shape, dtype=np.float64)
        x_c = np.ctypeslib.as_ctypes(x)
        result_c = np.ctypeslib.as_ctypes(result)
        A_c = np.ctypeslib.as_ctypes(A)
        mu_c = np.ctypeslib.as_ctypes(mu)
        sigma_c = np.ctypeslib.as_ctypes(sigma)
        A3_c = np.ctypeslib.as_ctypes(A3)
        sigma3_c = np.ctypeslib.as_ctypes(sigma3)

        profile233_c.argtypes = [
                c_int, c_double_p,
                c_double_p, c_long,
                c_double_p, c_double_p,
                c_double_p, c_double_p,
                c_double_p]
        profile233_c(n, x_c,
                     mu_c, c_long(N),
                     A_c, sigma_c,
                     A3_c, sigma3_c,
                     result_c)
        if len(poly):
            p = Legendre(poly, domain=[x.min(), x.max()])(x)
        else:
            p = np.zeros(x.shape)
        return result,p
    return profile

profile00 = profileGen(0)
def profile23(x, mu, A, sigma, A3, sigma3):
    xmask = (x > mu - 5*sigma)&(x < mu + 5*sigma)
    x = x[xmask]
    opts = [mu, A, sigma, A3, sigma3]
    result = profile00(x, *opts)
    return x,result

def variateConvolve(x, y, sigma, npad, configs={}):
    assert x.shape == sigma.shape
    debug = configs.get("debug", 0)

    ndata = x.size
    nB = 2*npad + 1
    x = x.astype(np.float64)
    y = y.astype(np.float64)
    sigma = sigma.astype(np.float64)
    result = np.zeros(y.shape, dtype=np.float64)
    kernel = np.zeros(nB * ndata, dtype=np.float64)

    x_c    = np.ctypeslib.as_ctypes(x)
    y_c    = np.ctypeslib.as_ctypes(y)
    sigma_c = np.ctypeslib.as_ctypes(sigma)
    result_c= np.ctypeslib.as_ctypes(result)
    kernel_c= np.ctypeslib.as_ctypes(kernel)

    variateConvolve_c(ndata, npad, x_c, y_c, sigma_c, result_c, kernel_c)
    if debug<1:
        return result
    else:
        kernels = kernel.reshape((ndata, nB))
        return result, {"kernels":kernels}


#region: backup codes
#def extractDirect(fimage, invvar, xcen, sigma, configs={} ):
#    #region: setups for all input
#    fimage= fimage.copy().astype(np.float64) # copy is necessary, or we will get confused result
#    invvar = invvar.copy().astype(np.float64)
#    xcen   = xcen.copy().astype(np.float64)
#    assert fimage.shape == invvar.shape
#    fimage[invvar==0]=0
#    assert np.all(np.isfinite(fimage))
#    assert np.all(np.isfinite(invvar))
#    # if the length is too short, the c function will stop, follow the notice
#    # profileMaxLength = configs.get("profileMaxLength", 15*np.max(sigma))
#    niter = 0
#    map2DMaxHalfWidth = configs.get("map2DMaxHalfWidth",6)
#    nx       = fimage.size
#    x        = np.arange(nx, dtype=np.float64)
#    nTrace   = xcen.size
#    lTrace   = np.arange(nTrace)
#    sigmal   = configs.get('sigmal', 10.0)
#    proftype = configs.get('proftype', 0)
#    map2DReject = configs.get("map2DReject", 0.8)
#    halfWidth = configs.get("halfWidth", 5)
#    if proftype not in [0, -3]:
#        raise Exception("unknown proftype {}".format(proftype))
#    cDebug   = configs.get('cDebug', 0)
#    reject   = configs.get('reject', np.array([0.8, 0.8, 0.8])).astype(np.float64)
#    if proftype in [0, -3]: #!! simple gaussian and **3 profile fitting
#        lastfitmask  = np.array([1, 0, 0]).astype(np.int64)
#        lastproftype = proftype
#        nCoeff       = 3
#
#    lastnCoeff= lastfitmask.size
#    #!! fit mask for A, sigma, mu?
#    fitmask  = configs.get("fitmask", np.array([1, 1, 1])).astype(np.int64)
#    nPoly    = configs.get('nPoly', 4)
#    ia       = np.concatenate((np.tile(fitmask, (nTrace,)), np.ones(nPoly))).astype(np.int64)
#    lastia   = np.concatenate((np.tile(lastfitmask, (nTrace,)), np.ones(nPoly))).astype(np.int64)
#    calcCovar= 0
#    # two parameters for converge
#    mindMu   = configs.get('mindMu', 0.05)
#    mindSigma= configs.get('mindSigma', 0.05)
#    # contral the max step
#    maxSigmaDiff = configs.get('maxSigmaDiff', 1.0)
#    maxMuDiff    = configs.get('maxMuDiff', 1.0)
#    if type(sigma)!=np.ndarray:
#        sigma = np.ones(nTrace, dtype=np.float64) * sigma
#    else:
#        assert xcen.size==sigma.size
#        sigma = sigma.astype(np.float64)
#    #endregion: setups for all input
#    #region: allocate memory for all outputs
#    ma            = nTrace * nCoeff + nPoly
#    lastma        = nTrace * lastnCoeff + nPoly
#    fullreject    = np.zeros(nTrace, dtype=np.long)
#    partialreject = np.zeros(nTrace, dtype=np.long)
#    ans           = np.zeros(ma, dtype=np.float64)
#    lastans       = np.zeros(lastma, dtype=np.float64)
#    ymod          = np.zeros(nx, dtype=np.float64)
#    amod          = np.zeros(nx, dtype=np.float64)
#    bmod          = np.zeros(nx, dtype=np.float64)
#    sumFlux       = np.zeros(nTrace, dtype=np.float64)
#    sumFluxBack   = np.zeros(nTrace, dtype=np.float64)
#    sumVar        = np.zeros(nTrace, dtype=np.float64)
#    reject2D      = np.zeros(nTrace, dtype=np.float64)
#    p             = np.zeros(ma, dtype=np.float64)
#    lastp         = np.zeros(lastma, dtype=np.float64)
#    covar         = np.zeros((ma, ma), dtype=np.float64)
#    lastcovar     = np.zeros((lastma, lastma), dtype=np.float64)
#    #endregion: allocate memory for all outputs
#    #region: setup c function
#    #long extract(
#    #      //  xLen   xIndArr     yValue          invvar           xcen guess     sigma guess
#    #        long nx, double* x, double* fimage, double* invvar, double* xcen, double* sigma,
#    #      //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
#    #      //    fiberNum   background    1,2 or 3
#    #        long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
#    #        long proftype, long calcCovar, double* reject, long* fullreject, long* partialreject,
#    #        double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
#    #        double* sumFlux, double* sumFluxBack, double* sumVar, double* reject2D,
#    #        double* map2d, long halfWidth, double map2DReject, long debug) {
#    extractDirect_c.argtypes =\
#        [ c_long,  c_double_p,  c_double_p,     c_double_p,    c_double_p,   c_double_p,
#          c_long,    c_long,      c_long,     c_double,      c_long_p,
#          c_long,       c_long,        c_double_p,        c_long_p,       c_long_p,
#          c_double_p,   c_double_p,   ma*c_double_p,  c_double_p,  c_double_p, c_double_p,
#          c_double_p, c_double_p, c_double_p, c_double_p,
#          c_double_p, c_long, c_double, c_long ]
#    extractDirect_c.restype = c_long
#    lp = c_long_p
#    dp = c_double_p
#    #endregion: setup c function
#    #region: setup for iteratively
#    maxiter = configs.get('maxiter', 20)
#    ansDebug= np.tile(ans, (maxiter, 1))
#    xcenDebug=np.zeros((maxiter+1, xcen.size))
#    xcenDebug[0,:]=xcen[:]
#    sigmaDebug=np.zeros((maxiter+1, sigma.size))
#    sigmaDebug[0,:]=sigma[:]
#    dmuDebug=np.zeros((maxiter, xcen.size))
#    dsigmaDebug=np.zeros((maxiter, xcen.size))
#    #map2DLength = int(2 + nx + 5*nTrace)
#    #!! caution here
#    map2DLength = configs.get("map2DLength", 5*nx)
#    # the structure of map2D
#    # (total data num) (total fiber num)
#    #      ((length of the block)(start index)(end index)(data...))*nTrace
#    # map2DLength is a bit large than the true data num
#    if map2DLength:
#        doDebug  = configs.get("doDebug", False)
#        map2D = np.zeros(map2DLength, dtype=np.float64)
#        map2D_c         = np.ctypeslib.as_ctypes(map2D)
#    else:
#        doDebug=0
#        map2D = None
#        map2D_c = None
#
#    mask = invvar>1e-20; assert mask.sum()>0
#    #fimage = fimage * mask
#    x             = x.astype(np.float64)
#    fimage        = fimage.astype(np.float64)
#    invvar        = invvar.astype(np.float64)
#    sigma         = sigma.astype(np.float64)
#    xcen          = xcen.astype(np.float64)
#    reject        = reject.astype(np.float64)
#    xcenOrigin    = xcen.copy()
#    sigmaOrigin   = sigma.copy()
#    # unchanged value
#    #        double* sumFlux, double* sumFluxBack, double* sumVar,
#    fimage_c        = np.ctypeslib.as_ctypes(fimage)
#    x_c             = np.ctypeslib.as_ctypes(x)
#    ia_c            = np.ctypeslib.as_ctypes(ia)
#    lastia_c        = np.ctypeslib.as_ctypes(lastia)
#    reject_c        = np.ctypeslib.as_ctypes(reject)
#    xFib            = np.arange(nTrace)*nCoeff
#    lastxFib        = np.arange(nTrace)*lastnCoeff
#    workinvvar      = invvar * mask
#    workinvvar_c    = np.ctypeslib.as_ctypes(workinvvar)
#    sigma_c         = np.ctypeslib.as_ctypes(sigma)
#    xcen_c          = np.ctypeslib.as_ctypes(xcen)
#    #fullreject[:]=0; partialreject[:]=0; ans[:]=0;
#    #ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; fscat[:]=0
#    fullreject_c    = np.ctypeslib.as_ctypes(fullreject)
#    partialreject_c = np.ctypeslib.as_ctypes(partialreject)
#    ans_c           = np.ctypeslib.as_ctypes(ans)
#    lastans_c       = np.ctypeslib.as_ctypes(lastans)
#    ymod_c          = np.ctypeslib.as_ctypes(ymod)
#    amod_c          = np.ctypeslib.as_ctypes(amod)
#    sumFlux_c       = np.ctypeslib.as_ctypes(sumFlux)
#    sumFluxBack_c   = np.ctypeslib.as_ctypes(sumFluxBack)
#    sumVar_c        = np.ctypeslib.as_ctypes(sumVar)
#    reject2D_c      = np.ctypeslib.as_ctypes(reject2D)
#    covar_c         = p3(covar)
#    lastcovar_c     = p3(lastcovar)
#    p_c             = np.ctypeslib.as_ctypes(p)
#    lastp_c         = np.ctypeslib.as_ctypes(lastp)
#    bmod_c          = np.ctypeslib.as_ctypes(bmod)
#    workmap2DMaxHalfWidth   = 0
#    #endregion: setup for iteratively
#    #region: call c function iteratively
#    while niter<maxiter :
#        # make sure these may-be-changed variables are all 0s
#        fullreject[:]=0; partialreject[:]=0; ans[:]=0;
#        ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; bmod[:]=0;
#        sumFlux[:]=0; sumFluxBack[:]=0; sumVar[:]=0
#        #!! key function call
#        result = \
#        extractDirect_c(
#          c_long(nx),        x_c,       fimage_c, workinvvar_c,     xcen_c,     sigma_c,
#          c_long(nTrace),    c_long(nPoly),  c_long(nCoeff), c_double(sigmal), ia_c,
#          c_long(proftype),  c_long(calcCovar),    reject_c,       fullreject_c,     partialreject_c,
#          ans_c,       ymod_c,     covar_c,     p_c,   bmod_c, amod_c,
#          sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
#          None, halfWidth, map2DReject, cDebug)
#        #region: test ans
#        assert result==1
#        assert np.all(np.isfinite(ans))
#        ansDebug[niter-1, :] = ans
#        diffs = (fimage-ymod) * np.sqrt(workinvvar)
#        chisq = np.sum(diffs**2)
#        count = ia.sum()
#        good = workinvvar>1e-10; assert good.sum()>0
#        reducedChi = chisq/(np.sum(good) - count)
#        #endregion: test ans
#        #region: make result and change for step
#        A      = ans[xFib]#!! ans[0 3 6...]
#        A[A==0]= 1
#        badMask  = A<0
#        dSigma = ans[xFib+1]/A
#        dMu    = ans[xFib+2]/A
#
#        dsigmaDebug[niter,:] = dSigma[:]
#        dmuDebug[niter,:] = dMu[:]
#
#        sigmaHigh     = (sigma+dSigma-sigmaOrigin) >  maxSigmaDiff
#        sigmaLow      = (sigma+dSigma-sigmaOrigin) < -maxSigmaDiff
#        sigmaNormal   = np.logical_and(np.logical_not(sigmaHigh), np.logical_not(sigmaLow))
#        sigma[sigmaNormal] += dSigma[sigmaNormal]
#        sigma[sigmaHigh]    = sigmaOrigin[sigmaHigh] + maxSigmaDiff
#        sigma[sigmaLow]     = sigmaOrigin[sigmaLow]  - maxSigmaDiff
#        sigma[badMask]      = sigmaOrigin[badMask]
#
#        muHigh     = (xcen+dMu-xcenOrigin) >  maxMuDiff
#        muLow      = (xcen+dMu-xcenOrigin) < -maxMuDiff
#        muNormal   = np.logical_and(np.logical_not(muHigh), np.logical_not(muLow))
#        xcen[muNormal] += dMu[muNormal]
#        xcen[muHigh]    = xcenOrigin[muHigh] + maxMuDiff
#        xcen[muLow]     = xcenOrigin[muLow]  - maxMuDiff
#        xcen[badMask]   = xcenOrigin[badMask]
#
#        xcenDebug[niter+1,:]  = xcen[:]
#        sigmaDebug[niter+1,:] = sigma[:]
#        finish = np.all(np.abs(dSigma)<mindSigma) and np.all(np.abs(dMu)<mindMu)
#        niter += 1
#        #endregion: make result and change for step
#    #endregion: call c function iteratively
#    #region: handle last fit for defferent type
#    fullreject[:]=0; partialreject[:]=0; lastans[:]=0;
#    ymod[:]=0; amod[:]=0; lastcovar[:]=0; lastp[:]=0; bmod[:]=0
#    # key function call
#    result = \
#        extractDirect_c(
#            c_long(nx),        x_c,       fimage_c, workinvvar_c,     xcen_c,     sigma_c,
#            c_long(nTrace),    c_long(nPoly),    c_long(lastnCoeff), c_double(sigmal), lastia_c,
#            c_long(lastproftype),  c_long(calcCovar),    reject_c,       fullreject_c,     partialreject_c,
#            lastans_c,       ymod_c, lastcovar_c, lastp_c,   bmod_c, amod_c,
#            sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
#            map2D_c, halfWidth, map2DReject, c_long(cDebug))
#
#    # test ans
#    if not (result==1 and np.all(np.isfinite(lastans))):
#        raise
#    #ansDebug[niter-1, :] = lastans
#    diffs = (fimage-ymod) * np.sqrt(workinvvar)
#    chisq = np.sum(diffs**2)
#    count = ia.sum()
#    good = workinvvar>1e-10; assert good.sum()>0
#    reducedChi = chisq/(np.sum(good) - count)
#    # make result and change for step
#    #endregion: handle last fit for defferent type
#    #region: gen output and return
#    lastA   = lastans[lastxFib]
#    finvrow = lastp[lastxFib]*lastp[lastxFib]
#    bad = np.where(np.logical_or(~np.isfinite(lastA), finvrow<1e-16, fullreject))
#    finvrow[bad]=0.0
#    #if proftype is None:
#    #    print "stop"
#    if doDebug:
#        #!! solve bug when doing parallel in MacOs
#        de.plt.switch_backend("Agg")
#        path = configs.get("path",".")
#        xInd = configs.get("xInd", "known_x_index")
#        #print("output debug figure of xInd: {} into {}".format(xInd, path))
#        fig = de.f();
#        de.p(lastA/np.sqrt(1/finvrow),"r");de.p(sumFlux/np.sqrt(sumVar),"g");
#        de.gca().set_title("{} flux xInd:{}".format(os.path.basename(path), xInd))
#        de.d(fig, os.path.join(path, str(xInd)+"-flux.fig"))
#        fig.clear()
#        de.plt.close(fig)
#
#        fig = de.f();
#        de.p(np.sqrt(1/finvrow),"r");de.p(np.sqrt(sumVar),"g");
#        de.gca().set_title("{} var xInd:{}".format(os.path.basename(path), xInd))
#        de.d(fig, os.path.join(path, str(xInd)+"-var.fig"))
#        fig.clear()
#        de.plt.close(fig)
#
#        fig = de.f();
#        plotMap2D(map2D, fimage, bmod, 1/invvar);
#        de.gca().set_title("{} profile xInd:{}".format(os.path.basename(path), xInd))
#        de.d(fig, os.path.join(path, str(xInd)+"-profile.fig"))
#        fig.clear()
#        de.plt.close(fig)
#        #de.plt.switch_backend("MacOsX")
#        de.plt.switch_backend(configs["plotBackend"])
#    return amod, lastA, finvrow, {"niter":niter, "reducedChi":reducedChi, "ymod":ymod,
#            "dmuDebug":dmuDebug, "dsigmaDebug":dsigmaDebug,
#            "xcenDebug":xcenDebug, "sigmaDebug":sigmaDebug,
#            "bmod":bmod,"map2D":map2D, "sigma":sigma, "xcen":xcen,
#            "sumFlux":sumFlux, "sumFluxBack": sumFluxBack, "sumVar": sumVar,
#            "reject2D":reject2D}
#    #!! some debug command when stop here
#    # de.f();de.p(lastA/np.sqrt(1/finvrow),"r");de.p(sumFlux/np.sqrt(sumVar),"g"); de.f();de.p(np.sqrt(1/finvrow),"r");de.p(np.sqrt(sumVar),"g"); de.f();plotMap2D(map2D, fimage, bmod, 1/invvar);
#    #endregion: gen output and return
#
#def extractDirect23(fimage, invvar, xcen, sigma, sf, oA, oA3, configs={} ):
#    #!! proftype is of no use
#    #region: setups for all input
#    fimage= fimage.copy().astype(np.float64) # copy is necessary, or we will get confused result
#    invvar = invvar.copy().astype(np.float64)
#    xcen   = xcen.copy().astype(np.float64)
#    sigma   = sigma.copy().astype(np.float64)
#    sf   = sf.copy().astype(np.float64)
#    oA   = oA.copy().astype(np.float64)
#    oA3  = oA3.copy().astype(np.float64)
#    assert fimage.shape == invvar.shape
#    fimage[invvar==0]=0
#    assert np.all(np.isfinite(fimage))
#    assert np.all(np.isfinite(invvar))
#    # if the length is too short, the c function will stop, follow the notice
#    # profileMaxLength = configs.get("profileMaxLength", 15*np.max(sigma))
#    niter = 0
#    map2DMaxHalfWidth = configs.get("map2DMaxHalfWidth",6)
#    nx       = fimage.size
#    x        = np.arange(nx, dtype=np.float64)
#    nTrace   = xcen.size
#    lTrace   = np.arange(nTrace)
#    sigmal   = configs.get('sigmal', 10.0)
#    proftype = configs.get('proftype', 0)
#    map2DReject = configs.get("map2DReject", 0.8)
#    halfWidth = configs.get("halfWidth", 5)
#    if proftype not in [0, -3]:
#        raise Exception("unknown proftype {}".format(proftype))
#    cDebug   = configs.get('cDebug', 0)
#    reject   = configs.get('reject', np.array([0.8, 0.8, 0.8])).astype(np.float64)
#
#    lastfitmask  = np.array([1, 0, 0]).astype(np.int64)
#    lastproftype = proftype
#    nCoeff       = 3
#
#    lastnCoeff= lastfitmask.size
#    #!! fit mask for A, sigma, mu?
#    fitmask  = configs.get("fitmask", np.array([1, 1, 1])).astype(np.int64)
#    lastsecondfitmask = np.array([1,1,0])
#    nPoly = configs.get('nPoly', 4)
#    ia = np.concatenate((np.tile(fitmask, (nTrace,)), np.ones(nPoly))).astype(np.int64)
#    lastia = np.concatenate((np.tile(lastfitmask, (nTrace,)), np.ones(nPoly))).astype(np.int64)
#    lastsecondia = np.concatenate((np.tile(lastsecondfitmask, (nTrace,)), np.ones(nPoly))).astype(np.int64)
#    calcCovar= 0
#    # two parameters for converge
#    mindMu   = configs.get('mindMu', 0.05)
#    mindSigma= configs.get('mindSigma', 0.05)
#    # contral the max step
#    maxSigmaDiff = configs.get('maxSigmaDiff', 1.0)
#    sigmaMax = configs.get("sigmaMax", 2.5)
#    sigmaMin = configs.get("sigmaMin", 1.0)
#    maxMuDiff    = configs.get('maxMuDiff', 0.5)
#    #endregion: setups for all input
#    #region: allocate memory for all outputs
#    ma            = nTrace * nCoeff + nPoly
#    lastma        = nTrace * lastnCoeff + nPoly
#    fullreject    = np.zeros(nTrace, dtype=np.long)
#    partialreject = np.zeros(nTrace, dtype=np.long)
#    ans           = np.zeros(ma, dtype=np.float64)
#    lastans       = np.zeros(lastma, dtype=np.float64)
#    ymod          = np.zeros(nx, dtype=np.float64)
#    amod          = np.zeros(nx, dtype=np.float64)
#    bmod          = np.zeros(nx, dtype=np.float64)
#    sumFlux       = np.zeros(nTrace, dtype=np.float64)
#    sumFluxBack   = np.zeros(nTrace, dtype=np.float64)
#    sumVar        = np.zeros(nTrace, dtype=np.float64)
#    reject2D      = np.zeros(nTrace, dtype=np.float64)
#    p             = np.zeros(ma, dtype=np.float64)
#    lastp         = np.zeros(lastma, dtype=np.float64)
#    covar         = np.zeros((ma, ma), dtype=np.float64)
#    lastcovar     = np.zeros((lastma, lastma), dtype=np.float64)
#    #endregion: allocate memory for all outputs
#    #region: setup c function
#    #long extractDirect23(
#    #  //  xLen   xIndArr     yValue          invvar           xcen guess
#    #    long nx, double* x, double* fimage, double* invvar, double* xcen,
#    #    double* sigma, double* sf, double* oA, double* oA3,
#    #  //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
#    #  //    fiberNum   background    1,2 or 3
#    #    long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
#    #    long proftype, long calcCovar, double* reject, long* fullreject, long* partialreject,
#    #    double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
#    #    double* sumFlux, double* sumFluxBack, double* sumVar, double* reject2d,
#    #    double* map2d, long halfWidth,
#    #    double map2DReject, long debug)
#    extractDirect_c.argtypes =\
#        [ c_long,  c_double_p,  c_double_p,     c_double_p,    c_double_p,
#          c_double_p, c_double_p, c_double_p, c_double_p,
#          c_long,    c_long,      c_long,     c_double,      c_long_p,
#          c_long,       c_long,        c_double_p,        c_long_p,       c_long_p,
#          c_double_p,   c_double_p,   ma*c_double_p,  c_double_p,  c_double_p, c_double_p,
#          c_double_p, c_double_p, c_double_p, c_double_p,
#          c_double_p, c_double_p, c_long, c_double, c_long ]
#    extractDirect_c.restype = c_long
#    lp = c_long_p
#    dp = c_double_p
#    #endregion: setup c function
#    #region: setup for iteratively
#    maxiter = configs.get('maxiter', 40)
#    ansDebug= np.tile(ans, (maxiter, 1))
#    xcenDebug=np.zeros((maxiter+1, xcen.size))
#    xcenDebug[0,:]=xcen[:]
#    sigmaDebug=np.zeros((maxiter+1, sigma.size))
#    sigmaDebug[0,:]=sigma[:]
#    dmuDebug=np.zeros((maxiter, xcen.size))
#    dsigmaDebug=np.zeros((maxiter, xcen.size))
#    #map2DLength = int(2 + nx + 5*nTrace)
#    #!! caution here
#    map2DLength = configs.get("map2DLength", 5*nx)
#    # the structure of map2D
#    # (total data num) (total fiber num)
#    #      ((length of the block)(start index)(end index)(data...))*nTrace
#    # map2DLength is a bit large than the true data num
#    if map2DLength:
#        doDebug  = configs.get("doDebug", False)
#        map2D = np.zeros(map2DLength, dtype=np.float64)
#        map2D_c       = np.ctypeslib.as_ctypes(map2D)
#        map2DOrigin   = np.zeros(map2DLength, dtype=np.float64)
#        map2DOrigin_c = np.ctypeslib.as_ctypes(map2DOrigin)
#    else:
#        doDebug=0
#        map2D = None
#        map2D_c = None
#        map2DOrigin = None
#        map2DOrigin_c = None
#
#    mask = invvar>1e-20; assert mask.sum()>0
#    #fimage = fimage * mask
#    xcenOrigin    = xcen.copy()
#    sigmaOrigin   = sigma.copy()
#    # unchanged value
#    #        double* sumFlux, double* sumFluxBack, double* sumVar,
#    fimage_c        = np.ctypeslib.as_ctypes(fimage)
#    x_c             = np.ctypeslib.as_ctypes(x)
#    ia_c            = np.ctypeslib.as_ctypes(ia)
#    lastia_c        = np.ctypeslib.as_ctypes(lastia)
#    lastsecondia_c  = np.ctypeslib.as_ctypes(lastsecondia)
#    reject_c        = np.ctypeslib.as_ctypes(reject)
#    xFib            = np.arange(nTrace)*nCoeff
#    lastxFib        = np.arange(nTrace)*lastnCoeff
#    workinvvar      = invvar * mask
#    workinvvar_c    = np.ctypeslib.as_ctypes(workinvvar)
#    sigma_c         = np.ctypeslib.as_ctypes(sigma)
#    sf_c         = np.ctypeslib.as_ctypes(sf)
#    oA_c         = np.ctypeslib.as_ctypes(oA)
#    oA3_c        = np.ctypeslib.as_ctypes(oA3)
#    xcen_c          = np.ctypeslib.as_ctypes(xcen)
#    #fullreject[:]=0; partialreject[:]=0; ans[:]=0;
#    #ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; fscat[:]=0
#    fullreject_c    = np.ctypeslib.as_ctypes(fullreject)
#    partialreject_c = np.ctypeslib.as_ctypes(partialreject)
#    ans_c           = np.ctypeslib.as_ctypes(ans)
#    lastans_c       = np.ctypeslib.as_ctypes(lastans)
#    ymod_c          = np.ctypeslib.as_ctypes(ymod)
#    amod_c          = np.ctypeslib.as_ctypes(amod)
#    sumFlux_c       = np.ctypeslib.as_ctypes(sumFlux)
#    sumFluxBack_c   = np.ctypeslib.as_ctypes(sumFluxBack)
#    sumVar_c        = np.ctypeslib.as_ctypes(sumVar)
#    reject2D_c      = np.ctypeslib.as_ctypes(reject2D)
#    covar_c         = p3(covar)
#    lastcovar_c     = p3(lastcovar)
#    p_c             = np.ctypeslib.as_ctypes(p)
#    lastp_c         = np.ctypeslib.as_ctypes(lastp)
#    bmod_c          = np.ctypeslib.as_ctypes(bmod)
#    workmap2DMaxHalfWidth   = 0
#    #endregion: setup for iteratively
#    #region: call c function iteratively
#    while niter<maxiter :
#        # make sure these may-be-changed variables are all 0s
#        fullreject[:]=0; partialreject[:]=0; ans[:]=0;
#        ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; bmod[:]=0;
#        sumFlux[:]=0; sumFluxBack[:]=0; sumVar[:]=0
#        #!! key function call
#        result = \
#        extractDirect23_c(
#          c_long(nx),        x_c,       fimage_c, workinvvar_c,     xcen_c,
#          sigma_c, sf_c, oA_c, oA3_c,
#          c_long(nTrace),    c_long(nPoly),  c_long(nCoeff), c_double(sigmal), ia_c,
#          c_long(proftype),  c_long(calcCovar),    reject_c,       fullreject_c,     partialreject_c,
#          ans_c,       ymod_c,     covar_c,     p_c,   bmod_c, amod_c,
#          sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
#          None, None, c_long(halfWidth), c_double(map2DReject), c_long(cDebug))
#        #region: test ans
#        assert result==1
#        assert np.all(np.isfinite(ans))
#        ansDebug[niter-1, :] = ans
#        diffs = (fimage-ymod) * np.sqrt(workinvvar)
#        chisq = np.sum(diffs**2)
#        count = ia.sum()
#        good = workinvvar>1e-10; assert good.sum()>0
#        reducedChi = chisq/(np.sum(good) - count)
#        #endregion: test ans
#        #region: make result and change for step
#        A      = ans[xFib]#!! ans[0 3 6...]
#        A[A==0]= 1
#        badMask  = A<0
#        dSigma = ans[xFib+1]/A
#        dMu    = ans[xFib+2]/A
#
#        dsigmaDebug[niter,:] = dSigma[:]
#        dmuDebug[niter,:] = dMu[:]
#
#        sigmaHigh     = (sigma+dSigma) > sigmaMax
#        sigmaLow      = (sigma+dSigma) < sigmaMin
#        sigmaNormal   = np.logical_and(np.logical_not(sigmaHigh), np.logical_not(sigmaLow))
#        sigma[sigmaNormal] += dSigma[sigmaNormal]
#        sigma[sigmaHigh]    = sigmaMax
#        sigma[sigmaLow]     = sigmaMin
#        sigma[badMask]      = sigmaOrigin[badMask]
#
#        muHigh     = (xcen+dMu-xcenOrigin) >  maxMuDiff
#        muLow      = (xcen+dMu-xcenOrigin) < -maxMuDiff
#        muNormal   = np.logical_and(np.logical_not(muHigh), np.logical_not(muLow))
#        xcen[muNormal] += dMu[muNormal]
#        xcen[muHigh]    = xcenOrigin[muHigh] + maxMuDiff
#        xcen[muLow]     = xcenOrigin[muLow]  - maxMuDiff
#        xcen[badMask]   = xcenOrigin[badMask]
#
#        xcenDebug[niter+1,:]  = xcen[:]
#        sigmaDebug[niter+1,:] = sigma[:]
#        finish = np.all(np.abs(dSigma)<mindSigma) and np.all(np.abs(dMu)<mindMu)
#
#        niter += 1
#        #endregion: make result and change for step
#    #endregion: call c function iteratively
#    #region: last second call
#    fullreject[:]=0; partialreject[:]=0; ans[:]=0;
#    ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; bmod[:]=0;
#    sumFlux[:]=0; sumFluxBack[:]=0; sumVar[:]=0
#    #!! key function call
#    result = \
#    extractDirect23_c(
#      c_long(nx),        x_c,       fimage_c, workinvvar_c,     xcen_c,
#      sigma_c, sf_c, oA_c, oA3_c,
#      c_long(nTrace),    c_long(nPoly),  c_long(nCoeff), c_double(sigmal), lastsecondia_c,
#      c_long(proftype),  c_long(calcCovar),    reject_c,       fullreject_c,     partialreject_c,
#      ans_c,       ymod_c,     covar_c,     p_c,   bmod_c, amod_c,
#      sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
#      None, None, c_long(halfWidth), c_double(map2DReject), c_long(cDebug))
#
#
#    #endregion: last second call
#    #region: handle last fit for defferent type
#    fullreject[:]=0; partialreject[:]=0; lastans[:]=0;
#    ymod[:]=0; amod[:]=0; lastcovar[:]=0; lastp[:]=0; bmod[:]=0
#    # key function call
#    result = \
#        extractDirect23_c(
#            c_long(nx),        x_c,       fimage_c, workinvvar_c,     xcen_c,
#            sigma_c, sf_c, oA_c, oA3_c,
#            c_long(nTrace),    c_long(nPoly),    c_long(lastnCoeff), c_double(sigmal), lastia_c,
#            c_long(lastproftype),  c_long(calcCovar),    reject_c,       fullreject_c,     partialreject_c,
#            lastans_c,       ymod_c, lastcovar_c, lastp_c,   bmod_c, amod_c,
#            sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
#            map2D_c, map2DOrigin_c, c_long(halfWidth), c_double(map2DReject), c_long(cDebug))
#    if not (result==1 and np.all(np.isfinite(ans))):
#        raise
#    toA = ans[xFib]
#    toA3 = ans[xFib+1]
#    good = np.logical_or(toA>0, toA3>0)
#    oA[good] = toA[good]
#    oA3[good] = toA3[good]
#    # test ans
#    if not (result==1 and np.all(np.isfinite(lastans))):
#        raise
#    #ansDebug[niter-1, :] = lastans
#    diffs = (fimage-ymod) * np.sqrt(workinvvar)
#    chisq = np.sum(diffs**2)
#    count = ia.sum()
#    good = workinvvar>1e-10; assert good.sum()>0
#    reducedChi = chisq/(np.sum(good) - count)
#    # make result and change for step
#    #endregion: handle last fit for defferent type
#    #region: gen output and return
#    lastA   = lastans[lastxFib]
#    finvrow = lastp[lastxFib]*lastp[lastxFib]
#    bad = np.where(np.logical_or(~np.isfinite(lastA), finvrow<1e-16, fullreject))
#    finvrow[bad]=0.0
#    #if proftype is None:
#    #    print "stop"
#    if doDebug:
#        #!! solve bug when doing parallel in MacOs
#        de.plt.switch_backend("Agg")
#        path = configs.get("path",".")
#        xInd = configs.get("xInd", "known_x_index")
#        #print("output debug figure of xInd: {} into {}".format(xInd, path))
#        fig = de.f();
#        de.p(lastA/np.sqrt(1/finvrow),"r");de.p(sumFlux/np.sqrt(sumVar),"g");
#        de.gca().set_title("{} flux xInd:{}".format(os.path.basename(path), xInd))
#        de.d(fig, os.path.join(path, str(xInd)+"-flux.fig"))
#        fig.clear()
#        de.plt.close(fig)
#
#        fig = de.f();
#        de.p(np.sqrt(1/finvrow),"r");de.p(np.sqrt(sumVar),"g");
#        de.gca().set_title("{} var xInd:{}".format(os.path.basename(path), xInd))
#        de.d(fig, os.path.join(path, str(xInd)+"-var.fig"))
#        fig.clear()
#        de.plt.close(fig)
#
#        fig = de.f();
#        plotMap2D(map2D, fimage, bmod, 1/invvar);
#        de.gca().set_title("{} profile xInd:{}".format(os.path.basename(path), xInd))
#        de.d(fig, os.path.join(path, str(xInd)+"-profile.fig"))
#        fig.clear()
#        de.plt.close(fig)
#        #de.plt.switch_backend("MacOsX")
#        de.plt.switch_backend(configs["plotBackend"])
#    return amod, lastA, finvrow, {"niter":niter, "reducedChi":reducedChi, "ymod":ymod,
#            "dmuDebug":dmuDebug, "dsigmaDebug":dsigmaDebug,
#            "xcenDebug":xcenDebug, "sigmaDebug":sigmaDebug,
#            "bmod":bmod,"map2D":map2D, "sigma":sigma, "xcen":xcen,
#            "map2DOrigin":map2DOrigin,
#            "sumFlux":sumFlux, "sumFluxBack": sumFluxBack, "sumVar": sumVar,
#            "reject2D":reject2D}
#    #!! some debug command when stop here
#    # de.f();de.p(lastA/np.sqrt(1/finvrow),"r");de.p(sumFlux/np.sqrt(sumVar),"g"); de.f();de.p(np.sqrt(1/finvrow),"r");de.p(np.sqrt(sumVar),"g"); de.f();plotMap2D(map2D, fimage, bmod, 1/invvar);
#    #endregion: gen output and return
#
#def extractUsingProfile(fimage, invvar, inMap2D, configs={} ):
#    #region: setups for all input
#    fimage= fimage.copy().astype(np.float64) # copy is necessary, or we will get confused result
#    invvar = invvar.copy().astype(np.float64)
#    inMap2D = inMap2D.copy().astype(np.float64)
#    fimage[invvar==0]=0
#    assert np.all(np.isfinite(fimage))
#    assert np.all(np.isfinite(invvar))
#    # if the length is too short, the c function will stop, follow the notice
#    # profileMaxLength = configs.get("profileMaxLength", 15*np.max(sigma))
#    niter = 0
#    nx       = fimage.size
#    x        = np.arange(nx, dtype=np.float64)
#    nTrace   = int(inMap2D[1])
#    lTrace   = np.arange(nTrace)
#    doDebug  = configs.get("doDebug", False)
#    cDebug   = configs.get("cDebug", False)
#
#    #!! fit mask for A, sigma, mu?
#    halfWidth = configs.get("halfWidth", 5)
#    fitmask  = configs.get("fitmask", np.array([1])).astype(np.long)
#    nCoeff   = fitmask.size
#    nPoly    = configs.get('nPoly', 4)
#    ia       = np.concatenate((np.tile(fitmask, (nTrace,)), np.ones(nPoly))).astype(np.int64)
#    reject   = configs.get('reject', np.array([0.8, 0.8, 0.8])).astype(np.float64)
#    calcCovar= 0
#    map2DReject = configs.get("map2DReject", 0.8)
#    #endregion: setups for all input
#    #region: allocate memory for all outputs
#    ma            = nTrace * nCoeff + nPoly
#    fullreject    = np.zeros(nTrace, dtype=np.long)
#    partialreject = np.zeros(nTrace, dtype=np.long)
#    ans           = np.zeros(ma, dtype=np.float64)
#    ymod          = np.zeros(nx, dtype=np.float64)
#    amod          = np.zeros(nx, dtype=np.float64)
#    bmod          = np.zeros(nx, dtype=np.float64)
#    sumFlux       = np.zeros(nTrace, dtype=np.float64)
#    sumFluxBack   = np.zeros(nTrace, dtype=np.float64)
#    sumVar        = np.zeros(nTrace, dtype=np.float64)
#    reject2D      = np.zeros(nTrace, dtype=np.float64)
#    p             = np.zeros(ma, dtype=np.float64)
#    covar         = np.zeros((ma, ma), dtype=np.float64)
#    #endregion: allocate memory for all outputs
#    #region: setup c function
#    #long extractUsingProfileWapper(
#    #     long nx, double* x, double* fimage, double* invvar,
#    #     long nTrace, long nPoly, long nCoeff, long* ia,
#    #     long calcCovar, double* reject, long* fullreject, long* partialreject,
#    #     double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
#    #     double* sumFlux, double* sumFluxBack, double* sumVar,
#    #     double* inMap2D, long halfWidth,
#    #     double* map2d, long debug)
#    extractDirect_c.argtypes = [
#            c_long, c_double_p, c_double_p, c_double_p,
#            c_long, c_long, c_long, c_long_p,
#            c_long, c_double_p, c_long_p, c_long_p,
#            c_double_p, c_double_p, ma*c_double_p, c_double_p, c_double_p, c_double_p,
#            c_double_p, c_double_p, c_double_p, c_double_p,
#            c_double_p, c_long,
#            c_double_p, c_double, c_long]
#    extractDirect_c.restype = c_long
#    lp = c_long_p
#    dp = c_double_p
#    #endregion: setup c function
#    #region: setup for input
#    #!! caution here
#    map2DLength = configs.get("map2DLength", 5*nx)
#    # the structure of map2D
#    # (total data num) (total fiber num)
#    #      ((length of the block)(xcen)(start index)(end index)(data...))*nTrace
#    # map2DLength is a bit large than the true data num
#    map2D = np.zeros(map2DLength, dtype=np.float64)
#
#    mask = invvar>1e-20; assert mask.sum()>0
#    #fimage = fimage * mask
#    x             = x.astype(np.float64)
#    fimage        = fimage.astype(np.float64)
#    invvar        = invvar.astype(np.float64)
#    reject        = reject.astype(np.float64)
#    # unchanged value
#    #        double* sumFlux, double* sumFluxBack, double* sumVar,
#    fimage_c        = np.ctypeslib.as_ctypes(fimage)
#    x_c             = np.ctypeslib.as_ctypes(x)
#    ia_c            = np.ctypeslib.as_ctypes(ia)
#    reject_c        = np.ctypeslib.as_ctypes(reject)
#    xFib            = np.arange(nTrace)*nCoeff
#    workinvvar      = invvar * mask
#    workinvvar_c    = np.ctypeslib.as_ctypes(workinvvar)
#
#    fullreject_c    = np.ctypeslib.as_ctypes(fullreject)
#    partialreject_c = np.ctypeslib.as_ctypes(partialreject)
#    ans_c           = np.ctypeslib.as_ctypes(ans)
#    ymod_c          = np.ctypeslib.as_ctypes(ymod)
#    amod_c          = np.ctypeslib.as_ctypes(amod)
#    sumFlux_c       = np.ctypeslib.as_ctypes(sumFlux)
#    sumFluxBack_c   = np.ctypeslib.as_ctypes(sumFluxBack)
#    sumVar_c        = np.ctypeslib.as_ctypes(sumVar)
#    reject2D_c      = np.ctypeslib.as_ctypes(reject2D)
#    covar_c         = p3(covar)
#    p_c             = np.ctypeslib.as_ctypes(p)
#    bmod_c          = np.ctypeslib.as_ctypes(bmod)
#    inMap2D_c       = np.ctypeslib.as_ctypes(inMap2D)
#    map2D_c         = np.ctypeslib.as_ctypes(map2D)
#    workmap2DMaxHalfWidth   = 0
#    #endregion: setup for iteratively
#
#    #region: handle last fit for defferent type
#    fullreject[:]=0; partialreject[:]=0; ans[:]=0;
#    ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; bmod[:]=0
#    # key function call
#    result = \
#        extractUsingProfileWapper_c(
#            c_long(nx),        x_c,       fimage_c, workinvvar_c,
#            c_long(nTrace),         c_long(nPoly),    c_long(nCoeff), ia_c,
#            c_long(calcCovar),    reject_c,       fullreject_c,     partialreject_c,
#            ans_c,       ymod_c, covar_c, p_c,   bmod_c, amod_c,
#            sumFlux_c, sumFluxBack_c, sumVar_c, reject2D_c,
#            inMap2D_c, halfWidth,
#            map2D_c, c_double(map2DReject), c_long(cDebug))
#
#    # test ans
#    if not (result==1 and np.all(np.isfinite(ans))):
#        raise
#    #ansDebug[niter-1, :] = lastans
#    diffs = (fimage-ymod) * np.sqrt(workinvvar)
#    chisq = np.sum(diffs**2)
#    count = ia.sum()
#    good = workinvvar>1e-10; assert good.sum()>0
#    reducedChi = chisq/(np.sum(good) - count)
#    # make result and change for step
#    #endregion: handle last fit for defferent type
#    #region: gen output and return
#    A   = ans[xFib]
#    finvrow = p[xFib]*p[xFib]
#    bad = np.where(np.logical_or(~np.isfinite(A), finvrow<1e-16, fullreject))
#    finvrow[bad]=0.0
#    #if proftype is None:
#    #    print "stop"
#    if doDebug:
#        #!! solve bug when doing parallel in MacOs
#        de.plt.switch_backend("Agg")
#        path = configs.get("path",".")
#        xInd = configs.get("xInd", "known_x_index")
#        print("output debug figure of xInd: {} into {}".format(xInd, path))
#        fig = de.f();
#        de.p(A/np.sqrt(1/finvrow),"r");de.p(sumFlux/np.sqrt(sumVar),"g");
#        de.gca().set_title("{} flux xInd:{}".format(os.path.basename(path), xInd))
#        de.d(fig, os.path.join(path, str(xInd)+"-flux.fig"))
#        fig.clear()
#        de.plt.close(fig)
#
#        fig = de.f();
#        de.p(np.sqrt(1/finvrow),"r");de.p(np.sqrt(sumVar),"g");
#        de.gca().set_title("{} var xInd:{}".format(os.path.basename(path), xInd))
#        de.d(fig, os.path.join(path, str(xInd)+"-var.fig"))
#        fig.clear()
#        de.plt.close(fig)
#
#        fig = de.f();
#        plotMap2D(map2D, fimage, bmod, 1/invvar);
#        de.gca().set_title("{} profile xInd:{}".format(os.path.basename(path), xInd))
#        de.d(fig, os.path.join(path, str(xInd)+"-profile.fig"))
#        fig.clear()
#        de.plt.close(fig)
#        #de.plt.switch_backend("MacOsX")
#        de.plt.switch_backend(configs["plotBackend"])
#    return amod, A, finvrow, {"niter":niter, "reducedChi":reducedChi, "ymod":ymod,
#            "bmod":bmod,"map2D":map2D, "inMap2D": inMap2D,
#            "sumFlux":sumFlux, "sumFluxBack": sumFluxBack, "sumVar": sumVar,
#            "reject2D": reject2D}
#    #!! some debug command when stop here
#    # de.f();de.p(lastA/np.sqrt(1/finvrow),"r");de.p(sumFlux/np.sqrt(sumVar),"g"); de.f();de.p(np.sqrt(1/finvrow),"r");de.p(np.sqrt(sumVar),"g"); de.f();plotMap2D(map2D, fimage, bmod, 1/invvar);
#    #endregion: gen output and return
#
#def plotMap2D(map2D, fimage, bmod, var):
#    map2D = map2D[:map2D[0]]
#    index = 2
#    xx = np.arange(fimage.size)
#    de.p(xx, fimage-bmod, "r-.")
#    de.p(xx, np.sqrt(var), "g-,")
#    for i in range(int(map2D[1])):
#        eachLen = map2D[index]
#        xStart = map2D[index+2]
#        xEnd = map2D[index+3]
#        xx = np.arange(xStart, xEnd+1)
#        yy = map2D[index+8:index+8+eachLen]
#        de.p(xx,yy)
#        index += eachLen + 8
#
#def buildMasterProfile(arrayList, configs={}):
#    nFile = len(arrayList)
#    nx = arrayList[0].shape[0]
#    for each in arrayList:
#        assert nx==each.shape[0]
#    nLine = np.max([each.shape[1] for each in arrayList])
#
#    origin = np.zeros((nFile, nx, nLine), dtype=np.float64)
#    result = np.zeros((nx, 3*nFile*nLine), dtype=np.float64)
#
#    lineNum = 0
#    for fileNum in range(nFile):
#        for lineNum in range(nx):
#            origin[fileNum][lineNum][0:arrayList[fileNum][lineNum][0]] = \
#                    arrayList[fileNum][lineNum][0:arrayList[fileNum][lineNum][0]]
#
#    origin_c = np.ctypeslib.as_ctypes(origin)
#    result_c = np.ctypeslib.as_ctypes(result)
#    type2 = result_c.__class__
#    type3 = origin_c.__class__
#    #void buildMasterProfile(long nx, long nFile, long nLine,
#    #                    // nFile * nx * nLine       nx * (3*nLine*nFile)
#    #                    double *origin,           double *result) {
#    buildMasterProfile_c.argtypes = [c_long, c_long, c_long, type3, type2]
#    buildMasterProfile_c(nx, nFile, nLine, origin_c, result_c)
#    maxLen = result[:,0].max()
#    result = result[:,:maxLen+1]
#    return result
#
#def plotMasterProfile(m, y, x):
#    ny, maxLen = m.shape
#    fiberNumber = m[0][1]
#    assert y<maxLen
#    assert x<fiberNumber
#    todo = m[y]
#
#    fibNum = 0
#    p = 2;
#    while fibNum<x:
#        thisLen = todo[p]
#        p += 1+thisLen*3
#        fibNum+=1
#    thisLen = todo[p]
#    xx = todo[p+1:p+1+thisLen]
#    yy = todo[p+1+thisLen:p+1+thisLen*2]
#    snr = todo[p+1+thisLen*2:p+1+thisLen*3]
#    de.sc(xx,yy)
#
#
##long extract_row (
##      //  xLen   xIndArr     yValue          invvar
##        long nx, float* x, float* fimage, float* invvar, float* xcen, float* sigma,
##      //    fiberNum   background    1,2 or 3
##        long nTrace, long nPoly,  long nCoeff, float sigmal, long* ia,
##        long proftype, long calcCovar, long* reject, long* fullreject, long* partialreject,
##        float* ans, float* ymod, float** covar, float* p, float* fscat)
#def extractRow(fimage, invvar, xcen, sigma, configs={} ):
#    #region: setups for all input
#    fimage= fimage.copy().astype(np.float64) # copy is necessary, or we will get confused result
#    invvar = invvar.copy().astype(np.float64)
#    xcen   = xcen.copy().astype(np.float64)
#    assert fimage.shape == invvar.shape
#    assert np.all(np.isfinite(fimage))
#    assert np.all(np.isfinite(invvar))
#    # if the length is too short, the c function will stop, follow the notice
#    # profileMaxLength = configs.get("profileMaxLength", 15*np.max(sigma))
#    niter = 0
#    map2DMaxHalfWidth = configs.get("map2DMaxHalfWidth",6)
#    nx       = fimage.size
#    x        = np.arange(nx, dtype=np.float64)
#    nTrace   = xcen.size
#    lTrace   = np.arange(nTrace)
#    sigmal   = configs.get('sigmal', 10.0)
#    proftype = configs.get('proftype', 0)
#    if proftype not in [0, -3, -23, -32]:
#        raise Exception("unknown proftype {}".format(proftype))
#    cDebug   = configs.get('cDebug', 0)
#    reject   = configs.get('reject', np.array([0.8, 0.8, 0.8])).astype(np.float64)
#    if proftype in [0, -3]: #!! simple gaussian and **3 profile fitting
#        lastfitmask  = np.array([1, 0, 0]).astype(np.int64)
#        lastproftype = proftype
#    elif proftype in [-23, -32]:
#        lastfitmask  = np.array([1, 1]).astype(np.int64)
#        if proftype==-23:
#            proftype = None
#            #!! use **2 fit and then...
#            y2, temp1, temp2, info= \
#                extractRow(fimage, invvar, xcen, sigma,{"proftype":0})
#            xcen = info["xcen"]
#            niter += info['niter']
#            sigma2 = info["sigma"].astype(np.float64)
#            y3, temp1, temp2, info= \
#                extractRow(fimage, invvar, xcen, sigma,{"proftype":-3, "fitmask":np.array([1,1,0])})
#            sigma3 = info["sigma"].astype(np.float64)
#            niter += info['niter']
#            print("stop")
#        if proftype==-32:
#            proftype = None
#            y3, temp1, temp2, info= \
#                extractRow(fimage, invvar, xcen, sigma,{"proftype":-3})
#            xcen = info["xcen"]
#            niter += info['niter']
#            sigma3 = info["sigma"].astype(np.float64)
#            y2, temp1, temp2, info= \
#                extractRow(fimage, invvar, xcen, sigma,{"proftype":0, "fitmask":np.array([1,1,0])})
#            sigma2 = info["sigma"].astype(np.float64)
#            niter += info['niter']
#
#    lastnCoeff= lastfitmask.size
#    nCoeff    = 3
#    #!! fit mask for A, sigma, mu?
#    fitmask  = configs.get("fitmask", np.array([1, 1, 1])).astype(np.int64)
#    nPoly    = configs.get('nPoly', 4)
#    ia       = np.concatenate((np.tile(fitmask, (nTrace,)), np.ones(nPoly))).astype(np.int64)
#    lastia   = np.concatenate((np.tile(lastfitmask, (nTrace,)), np.ones(nPoly))).astype(np.int64)
#    calcCovar= 0
#    # two parameters for converge
#    mindMu   = configs.get('mindMu', 0.05)
#    mindSigma= configs.get('mindSigma', 0.05)
#    # contral the max step
#    maxSigmaDiff = configs.get('maxSigmaDiff', 1.0)
#    maxMuDiff    = configs.get('maxMuDiff', 1.0)
#    if type(sigma)!=np.ndarray:
#        sigma = np.ones(nTrace, dtype=np.float64) * sigma
#    else:
#        assert xcen.size==sigma.size
#        sigma = sigma.astype(np.float64)
#    #endregion: setups for all input
#    #region: allocate memory for all outputs
#    ma            = nTrace * nCoeff + nPoly
#    lastma        = nTrace * lastnCoeff + nPoly
#    fullreject    = np.zeros(nTrace, dtype=np.long)
#    partialreject = np.zeros(nTrace, dtype=np.long)
#    ans           = np.zeros(ma, dtype=np.float64)
#    lastans       = np.zeros(lastma, dtype=np.float64)
#    ymod          = np.zeros(nx, dtype=np.float64)
#    amod          = np.zeros(nx, dtype=np.float64)
#    bmod          = np.zeros(nx, dtype=np.float64)
#    p             = np.zeros(ma, dtype=np.float64)
#    lastp         = np.zeros(lastma, dtype=np.float64)
#    covar         = np.zeros((ma, ma), dtype=np.float64)
#    lastcovar     = np.zeros((lastma, lastma), dtype=np.float64)
#    #endregion: allocate memory for all outputs
#    #region: setup c function
#    extract_row_c.argtypes =\
#        [ c_long,  c_double_p,  c_double_p,     c_double_p,    c_double_p,   c_double_p,
#          c_long,    c_long,      c_long,     c_double,      c_long_p,
#          c_long,       c_long,        c_double_p,        c_long_p,       c_long_p,
#          c_double_p,   c_double_p,   ma*c_double_p,  c_double_p,  c_double_p, c_double_p,
#          c_long, c_double_p, c_long ]
#    extract_row_c.restype = c_long
#    lp = c_long_p
#    dp = c_double_p
#    #endregion: setup c function
#    #region: setup for iteratively
#    maxiter = configs.get('maxiter', 20)
#    ansDebug= np.tile(ans, (maxiter, 1))
#    xcenDebug=np.zeros((maxiter+1, xcen.size))
#    xcenDebug[0,:]=xcen[:]
#    sigmaDebug=np.zeros((maxiter+1, sigma.size))
#    sigmaDebug[0,:]=sigma[:]
#    dmuDebug=np.zeros((maxiter, xcen.size))
#    dsigmaDebug=np.zeros((maxiter, xcen.size))
#    #map2DLength = int(2 + nx + 5*nTrace)
#    map2DLength = configs.get("map2DLength", 3*nx)
#    # the structure of map2D
#    # (total data num) (total fiber num)
#    #      ((length of the block)(start index)(end index)(data...))*nTrace
#    # map2DLength is a bit large than the true data num
#    map2D = np.zeros(map2DLength, dtype=np.float64)
#
#    mask = invvar>1e-20; assert mask.sum()>0
#    #fimage = fimage * mask
#    x             = x.astype(np.float64)
#    fimage        = fimage.astype(np.float64)
#    invvar        = invvar.astype(np.float64)
#    sigma         = sigma.astype(np.float64)
#    xcen          = xcen.astype(np.float64)
#    reject        = reject.astype(np.float64)
#    xcenOrigin    = xcen.copy()
#    sigmaOrigin   = sigma.copy()
#    # unchanged value
#    fimage_c        = np.ctypeslib.as_ctypes(fimage)
#    x_c             = np.ctypeslib.as_ctypes(x)
#    ia_c            = np.ctypeslib.as_ctypes(ia)
#    lastia_c        = np.ctypeslib.as_ctypes(lastia)
#    reject_c        = np.ctypeslib.as_ctypes(reject)
#    xFib            = np.arange(nTrace)*nCoeff
#    lastxFib        = np.arange(nTrace)*lastnCoeff
#    workinvvar      = invvar * mask
#    workinvvar_c    = np.ctypeslib.as_ctypes(workinvvar)
#    sigma_c         = np.ctypeslib.as_ctypes(sigma)
#    xcen_c          = np.ctypeslib.as_ctypes(xcen)
#    #fullreject[:]=0; partialreject[:]=0; ans[:]=0;
#    #ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; fscat[:]=0
#    fullreject_c    = np.ctypeslib.as_ctypes(fullreject)
#    partialreject_c = np.ctypeslib.as_ctypes(partialreject)
#    ans_c           = np.ctypeslib.as_ctypes(ans)
#    lastans_c       = np.ctypeslib.as_ctypes(lastans)
#    ymod_c          = np.ctypeslib.as_ctypes(ymod)
#    amod_c          = np.ctypeslib.as_ctypes(amod)
#    covar_c         = p3(covar)
#    lastcovar_c     = p3(lastcovar)
#    p_c             = np.ctypeslib.as_ctypes(p)
#    lastp_c         = np.ctypeslib.as_ctypes(lastp)
#    bmod_c          = np.ctypeslib.as_ctypes(bmod)
#    map2D_c         = np.ctypeslib.as_ctypes(map2D)
#    workmap2DMaxHalfWidth   = 0
#    #endregion: setup for iteratively
#    #region: call c function iteratively
#    if proftype is not None:# for single profile
#        while niter<maxiter :
#            # make sure these may-be-changed variables are all 0s
#            fullreject[:]=0; partialreject[:]=0; ans[:]=0;
#            ymod[:]=0; amod[:]=0; covar[:]=0; p[:]=0; bmod[:]=0
#            #!! key function call
#            result = \
#            extract_row_c(\
#              c_long(nx),        x_c,       fimage_c, workinvvar_c,     xcen_c,     sigma_c,
#              c_long(nTrace),    c_long(nPoly),   c_long(nCoeff), c_long(sigmal), ia_c,
#              c_long(proftype),  c_long(calcCovar),    reject_c,       fullreject_c,     partialreject_c,
#              ans_c,       ymod_c,     covar_c,     p_c,   bmod_c, amod_c,
#              workmap2DMaxHalfWidth, map2D_c, c_long(cDebug))
#            #region: test ans
#            assert result==1
#            assert np.all(np.isfinite(ans))
#            ansDebug[niter-1, :] = ans
#            diffs = (fimage-ymod) * np.sqrt(workinvvar)
#            chisq = np.sum(diffs**2)
#            count = ia.sum()
#            good = workinvvar>1e-10; assert good.sum()>0
#            reducedChi = chisq/(np.sum(good) - count)
#            #endregion: test ans
#            #region: make result and change for step
#            A      = ans[xFib]#!! ans[0 3 6...]
#            A[A==0]= 1
#            badMask  = A<0
#            dSigma = ans[xFib+1]/A
#            dMu    = ans[xFib+2]/A
#
#            dsigmaDebug[niter,:] = dSigma[:]
#            dmuDebug[niter,:] = dMu[:]
#
#            sigmaHigh     = (sigma+dSigma-sigmaOrigin) >  maxSigmaDiff
#            sigmaLow      = (sigma+dSigma-sigmaOrigin) < -maxSigmaDiff
#            sigmaNormal   = np.logical_and(np.logical_not(sigmaHigh), np.logical_not(sigmaLow))
#            sigma[sigmaNormal] += dSigma[sigmaNormal]
#            sigma[sigmaHigh]    = sigmaOrigin[sigmaHigh] + maxSigmaDiff
#            sigma[sigmaLow]     = sigmaOrigin[sigmaLow]  - maxSigmaDiff
#            sigma[badMask]      = sigmaOrigin[badMask]
#
#            muHigh     = (xcen+dMu-xcenOrigin) >  maxMuDiff
#            muLow      = (xcen+dMu-xcenOrigin) < -maxMuDiff
#            muNormal   = np.logical_and(np.logical_not(muHigh), np.logical_not(muLow))
#            xcen[muNormal] += dMu[muNormal]
#            xcen[muHigh]    = xcenOrigin[muHigh] + maxMuDiff
#            xcen[muLow]     = xcenOrigin[muLow]  - maxMuDiff
#            xcen[badMask]   = xcenOrigin[badMask]
#
#            xcenDebug[niter+1,:]  = xcen[:]
#            sigmaDebug[niter+1,:] = sigma[:]
#            finish = np.all(np.abs(dSigma)<mindSigma) and np.all(np.abs(dMu)<mindMu)
#            if finish or niter==maxiter-1:
#                # that's to calculate
#                workmap2DMaxHalfWidth=map2DMaxHalfWidth
#
#            niter += 1
#            #endregion: make result and change for step
#    #endregion: call c function iteratively
#    #region: handle last fit for defferent type
#        extract_row_c.argtypes = \
#            [ c_long,  c_double_p,  c_double_p,     c_double_p,    c_double_p,   c_double_p,
#              c_long,    c_long,      c_long,     c_double,      c_long_p,
#              c_long,       c_long,        c_double_p,        c_long_p,       c_long_p,
#              c_double_p,   c_double_p,   lastma*c_double_p,  c_double_p,  c_double_p, c_double_p,
#              c_long, c_double_p, c_long ]
#        extract_row_c.restype = c_long
#
#        fullreject[:]=0; partialreject[:]=0; lastans[:]=0;
#        ymod[:]=0; amod[:]=0; lastcovar[:]=0; lastp[:]=0; bmod[:]=0
#        # key function call
#        result = \
#            extract_row_c( \
#                nx,        x_c,       fimage_c, workinvvar_c,     xcen_c,     sigma_c,
#                nTrace,         nPoly,    lastnCoeff, sigmal, lastia_c,
#                lastproftype,      calcCovar,    reject_c,       fullreject_c,     partialreject_c,
#                lastans_c,       ymod_c, lastcovar_c, lastp_c,   bmod_c, amod_c,
#                workmap2DMaxHalfWidth, map2D_c, cDebug)
#    else: # for multi profile
#        extract_row23_c.argtypes = \
#            [ c_long,  c_double_p,  c_double_p,     c_double_p,    c_double_p,   c_double_p, c_double_p,
#              c_long,    c_long,      c_long,     c_double,      c_long_p,
#              c_long,       c_long,        c_double_p,        c_long_p,       c_long_p,
#              c_double_p,   c_double_p,   lastma*c_double_p,  c_double_p,  c_double_p, c_double_p,
#              c_long, c_double_p, c_long ]
#        extract_row_c.restype = c_long
#
#        sigma2_c = np.ctypeslib.as_ctypes(sigma2)
#        sigma3_c = np.ctypeslib.as_ctypes(sigma3)
#        fullreject[:]=0; partialreject[:]=0; lastans[:]=0;
#        ymod[:]=0; amod[:]=0; lastcovar[:]=0; lastp[:]=0; bmod[:]=0
#        # key function call
#        result = \
#            extract_row23_c( \
#                nx,        x_c,       fimage_c, workinvvar_c,     xcen_c,     sigma2_c, sigma3_c,
#                nTrace,         nPoly,    lastnCoeff, sigmal, lastia_c,
#                1,      calcCovar,    reject_c,       fullreject_c,     partialreject_c,
#                lastans_c,       ymod_c, lastcovar_c, lastp_c,   bmod_c, amod_c,
#                1 , map2D_c, cDebug)
#        #print "stop"
#
#    # test ans
#    if not (result==1 and np.all(np.isfinite(lastans))):
#        raise
#    #ansDebug[niter-1, :] = lastans
#    diffs = (fimage-ymod) * np.sqrt(workinvvar)
#    chisq = np.sum(diffs**2)
#    count = ia.sum()
#    good = workinvvar>1e-10; assert good.sum()>0
#    reducedChi = chisq/(np.sum(good) - count)
#    # make result and change for step
#    #endregion: handle last fit for defferent type
#    #region: gen output and return
#    lastA   = lastans[lastxFib]
#    finvrow = lastp[lastxFib]*lastp[lastxFib]
#    bad = np.where(np.logical_or(~np.isfinite(lastA), finvrow<1e-16, fullreject))
#    finvrow[bad]=0.0
#    #if proftype is None:
#    #    print "stop"
#    return amod, lastA, finvrow, {"niter":niter, "reducedChi":reducedChi, "ymod":ymod,
#                              "dmuDebug":dmuDebug, "dsigmaDebug":dsigmaDebug,
#                              "xcenDebug":xcenDebug, "sigmaDebug":sigmaDebug,
#                              "bmod":bmod,"map2D":map2D, "sigma":sigma, "xcen":xcen}
#    #endregion: gen output and return
#
#def map2DFixWidthRead(nTrace, yset, flux, width):
#    nB = 2*width + 1
#    yset  = yset.copy().astype(np.float64)
#    result = np.zeros(nTrace * nB, dtype=np.float64)
#    flux   = flux.copy().astype(np.float64)
#    map2DSlice_c.argtypes =[c_int, c_double_p, c_double_p, c_double_p, c_int]
#    result_c = np.ctypeslib.as_ctypes(result)
#    yset_c  = np.ctypeslib.as_ctypes(yset)
#    flux_c   = np.ctypeslib.as_ctypes(flux)
#    map2DFixWidthRead_c(nTrace, flux_c, result_c, yset_c, width)
#    return result
#
#def map2DFixWidthWrite(nTrace, yset, flux, result, width):
#    '''len(flux) = nB*nTrace, len(result) = ny'''
#    yset  = yset.copy().astype(np.float64)
#    flux   = flux.copy().astype(np.float64)
#    result = result.copy().astype(np.float64)
#    map2DSlice_c.argtypes =[ \
#        c_int, c_double_p, c_double_p, c_double_p, c_int]
#    result_c = np.ctypeslib.as_ctypes(result)
#    yset_c  = np.ctypeslib.as_ctypes(yset)
#    flux_c   = np.ctypeslib.as_ctypes(flux)
#    map2DFixWidthWrite_c(nTrace, flux_c, result_c, yset_c, width)
#    return result
#
#def profile23old(x, A, mu, sigma, A1, sigma1):
#    x = x.astype(np.float64)
#    n = x.size
#    result = np.zeros(x.shape, dtype=np.float64)
#    x_c = np.ctypeslib.as_ctypes(x)
#    result_c = np.ctypeslib.as_ctypes(result)
#    #profile23_c.argtypes = [
#            #c_int, c_double_p,
#            #c_double, c_double, c_double,
#            #c_double, c_double,
#            #c_double_p]
#    profile23_c(n, x_c,
#            A, mu, sigma,
#            A1, sigma1,
#            result_c)
#    return result
#
#def variateDeconvolve(x, y, sigma, npad, configs={}):
#    assert x.shape == sigma.shape
#    debug = configs.get("debug", 0)
#
#    ndata = x.size
#    nB = 2*npad + 1
#    x = x.astype(np.float64)
#    y = y.astype(np.float64)
#    sigma = sigma.astype(np.float64)
#    result = np.zeros(y.shape, dtype=np.float64)
#    kernel = np.zeros(nB * ndata, dtype=np.float64)
#
#    x_c    = np.ctypeslib.as_ctypes(x)
#    y_c    = np.ctypeslib.as_ctypes(y)
#    sigma_c = np.ctypeslib.as_ctypes(sigma)
#    result_c= np.ctypeslib.as_ctypes(result)
#    kernel_c= np.ctypeslib.as_ctypes(kernel)
#
#    variateDeconvolve_c(ndata, npad, x_c, y_c, sigma_c, result_c, kernel_c, debug)
#    if debug<1:
#        return result
#    else:
#        kernels = kernel.reshape((ndata, nB))
#        return result, {"kernels":kernels}
#
#endregion: backup codes
