#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <strings.h>
#include "../../../math/src/astroMath.h"
#include <string.h>
#include <assert.h>
#include <float.h>
#define MAXEXP 25.0
#define NSUBSAMP 1000
#define MAP2DREJECT 0.75

//#region: headers
long extractGivenProfile(
        long nx, double* x, double* fimage, double* invvar,
        long nTrace, long *xmin, long *xmax,
        double *aprofileIn,
        long nPoly,  long nCoeff, long* ia,
        long calcCovar, double* reject, long* fullreject, long* partialreject,
        double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
        long debug);
long extract_row (
      //  xLen   xIndArr     yValue          invvar
        long nx, double* x, double* fimage, double* invvar, double* xcen, double* sigma,
      //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
      //    fiberNum   background    1,2 or 3
        long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
        long proftype, long calcCovar, double* reject, long* fullreject, long* partialreject,
        double* ans, double* ymod, double** covar, double* p, double* fscat, double* amod,
        long maxHalfWidth, double* map2d, long debug);
void AddGaussWings(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma,
        long nCoeff, double contribution);
void AddExponentialWings(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigmal,
        double contribution);
void ProfileGauss(double *x, long ndat, double **y, double xcen, long xmin,
        long xmax, double sigma, long nCoeff);
void ProfileGaussJin(double *x, long ndat, double **y, double xcen, long xmin,
        long xmax, double sigma, long nCoeff);
void ProfileGaussStatic(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff);
void ProfileAbs3(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff);
void ProfileAbs25(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff);
void ProfileDoubleGauss(double *x, long ndat, double **y, double xcen, long xmin,
        long xmax, double sigma, long nCoeff);
void ProfileAbs3WideGauss(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff);
void ProfileAbs3HalfLorentz(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff);
void ProfileGaussHalfLorentz(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff);
void findXLimits(long *xmin, long *xmax, double *x, double *xcen,
        long nx, long nTrace, double *sigma, double sigmal);
void subtractProfile(double *y, long nx, long *xmin, long *xmax, long nTrace, long nCoeff,
        double **aprofile, long *ia, double *a);
void subtractPoly(double *y, long nx, long nPoly, double **apoly, long *ia, double *a);
void CheckRowFibers(double **abig, long *xmin, long *xmax,
        long nTrace, long nCoeff, double *a, long *ia, double *invvar,
        double reject1, double reject2, double rejectflux, long *partial, long *full);
void CheckMultiRowFibers(double **abig, long *xmin, long *xmax,
        long nTrace, long nCoeff, double *a, long *ia, double *invvar,
        long nsingle, long multirow);
void fillProfile(double **y, double *x, double *xcen, long *xmin,
        long *xmax, double *sigma, long nx, long nCoeff,
        long nTrace, long proftype, double sigmal,
        double contribution);
void fillPoly(double **y, double *x, long nx, long nPoly, double x1, double x2);
void fillCovar(double *ysub, double *invvar, long nx, double **aprofile,
        double **apoly, long nTrace, long nCoeff, long nBand,
        long nPoly, double *beta, long *ia, double **covar,
        long *xmin, long *xmax);
void cholslRow(double **a, long *ia, long nTrace, long nCoeff,
        long nBand, long nPoly, double *p, double *b, double *x);
void cholslRowCovar(double **a, long *ia, long nTrace, long nCoeff,
        long nBand, long nPoly, double *p);
int choldcRow(double **a, long *ia, long nTrace, long nCoeff,
        long nBand, long nPoly, double *p);
void chebyshevRow(double x, double *coeff, long nCoeff);
void find2DLimits(long *xInd, long *xmin, long *xmax, double *x,
                  double *xcen, double *sigma, long nx, long nTrace, long maxHalfWidth);
long extract_row23(
        long nx, double* x, double* fimage, double* invvar, double* xcen, double* sigma2, double* sigma3,
        long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
        long doLastFit, long calcCovar, double* reject, long* fullreject, long* partialreject,
        double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
        long maxHalfWidth, double* map2d, long debug);
void ProfileAbs232Jin(double *x, long ndat, double **y, double xcen,
                     long xmin, long xmax, double sigma, long nCoeff);
void ProfileAbs233Jin(double *x, long ndat, double **y, double xcen,
                     long xmin, long xmax, double sigma, long nCoeff);

//#endregion: headers

//#region: good profile
void fullProfile2323(double *x, long ndat, double **y, double xcen, long xmin,
                 long xmax, double sigma, double sf, double oA, double oA3, long nCoeff) {
    long i,j,k;
    double base2,base3;
    double diff2,diff3,frac,absdiff3;
    double epow2,epow3;
    double sigma2, sigma3, nf2, nf3;
    double t;

    sigma2 = sigma;
    sigma3 = sigma2*sf;
    nf2 = 1.0/(sqrt(6.283185307)*sigma2);
    nf3 = 1.0/(2.575798634*sigma3);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff2 = (x[i] - xcen + frac)/sigma2;
                diff3 = (x[i] - xcen + frac)/sigma3;
                absdiff3 = fabs(diff3);
                epow2 = 0.5*diff2*diff2;
                epow3 = 1./3.*diff3*diff3*absdiff3;

                if (epow2 < MAXEXP){
                    base2 = exp(-epow2) * nf2;
                    base3 = exp(-epow3) * nf3;
                } else {
                    base2 = 0.0;
                    base3 = 0.0;
                }
                t = oA+oA3;
                // ans[0] = A
                y[0][k] += oA/t*base2 + oA3/t*base3;
                // ans[1] = A * dSigma
                y[1][k] += oA/t*(2*epow2-1)/sigma2*base2 + oA3/t*(3*epow3-1)/sigma3*base3;
                // ans[2] = A * dMu
                y[2][k] += oA/t*diff2/sigma2*base2 + oA3/t*diff3*absdiff3/sigma3*base3;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void fullProfile2(double *x, long ndat, double **y, double xcen, long xmin,
                 long xmax, double sigma, long nCoeff) {
    long i,j,k;
    double base2;
    double diff2,frac,absdiff3;
    double epow2;
    double sigma2, nf2;

    sigma2 = sigma;
    nf2 = 1.0/(sqrt(6.283185307)*sigma2);
    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff2 = (x[i] - xcen + frac)/sigma2;
                epow2 = 0.5*diff2*diff2;
                if (epow2 < MAXEXP){
                    base2 = exp(-epow2) * nf2;
                } else {
                    base2 = 0.0;
                }
                // ans[0] = A
                y[0][k] += base2;
                // ans[1] = A * dSigma
                y[1][k] += (2*epow2-1)/sigma2*base2;
                // ans[2] = A * dMu
                y[2][k] += diff2/sigma2*base2;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void fullProfile3(double *x, long ndat, double **y, double xcen, long xmin,
                 long xmax, double sigma, long nCoeff) {
    long i,j,k;
    double base3;
    double diff3,frac,absdiff3;
    double epow3;
    double sigma3, nf3;
    sigma3 = sigma;
    nf3 = 1.0/(2.575798634*sigma3);
    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff3 = (x[i] - xcen + frac)/sigma3;
                absdiff3 = fabs(diff3);
                epow3 = 1./3.*diff3*diff3*absdiff3;

                if (epow3 < MAXEXP){
                    base3 = exp(-epow3) * nf3;
                } else {
                    base3 = 0.0;
                }
                // ans[0] = A
                y[0][k] += base3;
                // ans[1] = A * dSigma
                y[1][k] += (3*epow3-1)/sigma3*base3;
                // ans[2] = A * dMu
                y[2][k] += diff3*absdiff3/sigma3*base3;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void fullProfile2233(double *x, long ndat, double **y, double xcen, long xmin,
                 long xmax, double sigma, double sf, long nCoeff) {
    long i,j,k;
    double base2,base3;
    double diff2,diff3,frac,absdiff3;
    double epow2,epow3;
    double sigma2, sigma3, nf2, nf3;
    double t;

    sigma2 = sigma;
    sigma3 = sigma2*sf;
    nf2 = 1.0/(sqrt(6.283185307)*sigma2);
    nf3 = 1.0/(2.575798634*sigma3);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff2 = (x[i] - xcen + frac)/sigma2;
                diff3 = (x[i] - xcen + frac)/sigma3;
                absdiff3 = fabs(diff3);
                epow2 = 0.5*diff2*diff2;
                epow3 = 1./3.*diff3*diff3*absdiff3;

                if (epow2 < MAXEXP){
                    base2 = exp(-epow2) * nf2;
                    base3 = exp(-epow3) * nf3;
                } else {
                    base2 = 0.0;
                    base3 = 0.0;
                }
                // ans[0] = A2
                y[0][k] += base2;
                // ans[1] = A3
                y[1][k] += base3;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void fullProfile22334455(double *x, long ndat, double **y, double xcen, long xmin,
                 long xmax, double sigma, double sf, long nCoeff) {
    long i,j,k;
    double base2,base3,base21,base22;
    double diff2,diff3,diff21,diff22, frac,absdiff3;
    double epow2,epow3,epow21,epow22;
    double sigma2, sigma3, sigma21, sigma22, nf2, nf3, nf21, nf22;
    double t;

    sigma2 = sigma;
    sigma21 = sigma/2;
    sigma22 = sigma*1.3;
    sigma3 = sigma2*sf;
    nf2 = 1.0/(sqrt(6.283185307)*sigma2);
    nf21 = 1.0/(sqrt(6.283185307)*sigma21);
    nf22 = 1.0/(sqrt(6.283185307)*sigma22);
    nf3 = 1.0/(2.575798634*sigma3);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff2 = (x[i] - xcen + frac)/sigma2;
                diff21 = (x[i] - xcen + frac)/sigma21;
                diff22 = (x[i] - xcen + frac)/sigma22;
                diff3 = (x[i] - xcen + frac)/sigma3;
                absdiff3 = fabs(diff3);
                epow2  = 0.5*diff2*diff2;
                epow21 = 0.5*diff21*diff21;
                epow22 = 0.5*diff22*diff22;
                epow3  = 1./3.*diff3*diff3*absdiff3;

                if (epow2 < MAXEXP){
                    base2 = exp(-epow2) * nf2;
                    base21 = exp(-epow21) * nf21;
                    base22 = exp(-epow22) * nf22;
                    base3 = exp(-epow3) * nf3;
                } else {
                    base2 = 0.0;
                    base21 = 0.0;
                    base22 = 0.0;
                    base3 = 0.0;
                }
                // ans[0] = A2
                y[0][k] += base2;
                // ans[1] = A3
                y[1][k] += base3;
                // ans[2] = A2/2
                y[2][k] += base21;
                // ans[3] = A2*2
                y[3][k] += base22;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void fullProfile2345(double *x, long ndat, double **y, double xcen, long xmin,
                 long xmax, double sigma, double sf, double oA, double oA3,
                 double oA21, double oA22, long nCoeff) {
    long i,j,k;
    double base2,base3,base21,base22;
    double diff2,diff3,diff21,diff22, frac,absdiff3;
    double epow2,epow3,epow21,epow22;
    double sigma2, sigma3, sigma21, sigma22, nf2, nf3, nf21, nf22;
    double t;

    t = oA + oA3 + oA21 + oA22;
    sigma2 = sigma;
    sigma21 = sigma/2;
    sigma22 = sigma*1.3;
    sigma3 = sigma2*sf;
    nf2 = 1.0/(sqrt(6.283185307)*sigma2);
    nf21 = 1.0/(sqrt(6.283185307)*sigma21);
    nf22 = 1.0/(sqrt(6.283185307)*sigma22);
    nf3 = 1.0/(2.575798634*sigma3);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff2 = (x[i] - xcen + frac)/sigma2;
                diff21 = (x[i] - xcen + frac)/sigma21;
                diff22 = (x[i] - xcen + frac)/sigma22;
                diff3 = (x[i] - xcen + frac)/sigma3;
                absdiff3 = fabs(diff3);
                epow2  = 0.5*diff2*diff2;
                epow21 = 0.5*diff21*diff21;
                epow22 = 0.5*diff22*diff22;
                epow3  = 1./3.*diff3*diff3*absdiff3;

                if (epow2 < MAXEXP){
                    base2 = exp(-epow2) * nf2;
                    base21 = exp(-epow21) * nf21;
                    base22 = exp(-epow22) * nf22;
                    base3 = exp(-epow3) * nf3;
                } else {
                    base2 = 0.0;
                    base21 = 0.0;
                    base22 = 0.0;
                    base3 = 0.0;
                }
                // ans[0] = A2
                y[0][k] += (oA * base2 + oA3 * base3 + oA21 * base21 + oA22 * base22)/t;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}


void fillProfileFinal(double **y, double *x, double *xcen, long *xmin,
        long *xmax, double *sigma, double *sf,
        double *oA, double *oA3, double *oA21, double *oA22,
        long nx, long nCoeff, long nTrace, long proftype) {
    int i, j, m, length;
    long tCoeff;
    for (i=0,j=0; i<nTrace; i++,j+=nCoeff) {
        length = xmax[i] - xmin[i] + 1;
        if (length > 0 && sigma[i] > 0.0) {
            if (proftype==2) {// three profile: 2 + dsigma + dmu
                fullProfile2(x, nx, &y[j], xcen[i], xmin[i],
                    xmax[i], sigma[i], nCoeff);
            } else if (proftype==3) {// three profile: 2 + dsigma + dmu
                fullProfile3(x, nx, &y[j], xcen[i], xmin[i],
                    xmax[i], sigma[i], nCoeff);
            } else if (proftype==2233) {// two profile: 2+3
                fullProfile2233(x, nx, &y[j], xcen[i], xmin[i],
                    xmax[i], sigma[i], sf[i], nCoeff);
            } else if (proftype==2323) {// three profile: 23 + dsigma + dmu
                fullProfile2323(x, nx, &y[j], xcen[i], xmin[i],
                    xmax[i], sigma[i], sf[i], oA[i], oA3[i], nCoeff);
            } else if (proftype==22334455) {// four profile: 2 + 3 + 2/2 + 2*2
                fullProfile22334455(x, nx, &y[j], xcen[i], xmin[i],
                    xmax[i], sigma[i], sf[i], nCoeff);
            } else if (proftype==2345) {// one profile: 2345
                fullProfile2345(x, nx, &y[j], xcen[i], xmin[i],
                    xmax[i], sigma[i], sf[i], oA[i], oA3[i], oA21[i], oA22[i], nCoeff);
            } else {
                fprintf(stderr,"unknown proftype, exit\n %ld\n", proftype);
                exit(1);
            }
        }
    }
}

//#endregion: good profile

//!! in use
//#region: extract
long extract_wapper(
      //  xLen   xIndArr     yValue          invvar           xcen guess
        long nx, double* x, double* fimage, double* invvar, double* xcen,
        double* sigma, double* sf, double* oA, double* oA3, double* oA21, double* oA22,
      //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
      //    fiberNum   background    1,2 or 3
        long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
        long proftype, long calcCovar, double* reject, long* fullreject, long* partialreject,
        double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
        double* sumFlux, double* sumFluxBack, double* sumVar, double* reject2d,
        double* map2d, double* map2dOrigin, long halfWidth,
        double map2DReject, long debug) {
     // bmod is the background part, amod is the profile part
   //!! setup
    FILE *fp;
    long  good;
    long  *xmin, *xmax;
    double ** aprofile;
    double contribution = 0.0;
    double tsigma = 5.0;             /* set limits of profile influence   */
    long  i,j,k,l,length, tTrace;
    long map2dCount=2; // add count for total length and total fib number
    double *var;

    /*for (iy=0; iy < ma; iy++) covar[iy] = (double *) argv[argct]+iy*ma;*/
    xmin = (long *)malloc(sizeof(long)*nTrace);// checked, safe
    xmax = (long *)malloc(sizeof(long)*nTrace);// checked, safe
   //!! Find max and min of profile influence
    // have nothing to do with fimage
    //      tsigma is limits of profile influence       5 times sigma
    findXLimits(xmin, xmax, x, xcen, nx, nTrace, sigma, tsigma);
   //!! fill the aprofile
    tTrace = nCoeff*nTrace;
    // room for aprofile
    long aprofileLength=0;
    for(i=0; i<nTrace; i++){//Room for fiber profiles
        length = xmax[i] - xmin[i] + 1;
        if (length < 0) length = 0;
        aprofileLength += length*nCoeff;
    }
    double *aprofileIn;
    /*printf("aprofileLength: %ld\n", aprofileLength);*/
    aprofileIn = (double *)malloc(aprofileLength * sizeof(double));
    aprofile = (double **)malloc(tTrace * sizeof(double *));
    double *pAp=aprofileIn;
    long aux=0;
    for (i=0,k=0; i<nTrace; i++) {
        for (j=0; j<nCoeff; j++,k++) {
            aprofile[k] = pAp;
            pAp += xmax[i] - xmin[i] + 1;
            aux += xmax[i] - xmin[i] + 1;
        }
    }
    /*printf("aux: %ld\n", aux);*/
    // aprofile[nTrace*nCoeff][length of each fit]
    fillProfileFinal(aprofile, x, xcen, xmin, xmax, sigma, sf,
                     oA, oA3, oA21, oA22, nx, nCoeff, nTrace, proftype);
   //!! key function call
    good = extractGivenProfile(
        nx, x, fimage, invvar,
        nTrace, xmin, xmax,
        aprofileIn,
        nPoly, nCoeff, ia,
        calcCovar, reject, fullreject, partialreject,
        ans, ymod, covar, p, bmod, amod,
        debug);
    //assert(good);
   //!! make map2d
    if (map2d) {
        double *tempFlux, *tempBack, *tempVar;
        double eachOtherFlux, eachFlux, eachRatio, eachReject, eachSum;
        double* map2dOriginFirst, *map2dFirst;
        int haveBadPixel;
        //!! temp flux all
        tempFlux = (double*)malloc(nx * sizeof(double));
        tempBack = (double*)malloc(nx * sizeof(double));
        tempVar = (double*)malloc(nx * sizeof(double));
        var = (double*)malloc(nx * sizeof(double));
        long lowF, highF, lowX, highX;
        for (i=0; i<nx; i++)
            var[i] = 1/invvar[i];

        map2dFirst=map2d;
        map2dOriginFirst=map2dOrigin;
        *(map2d++) = map2dCount;// right map2dCount num in the end of the if
        *(map2dOrigin++) = map2dCount;// right map2dCount num in the end of the if
        *(map2d++) = nTrace;
        *(map2dOrigin++) = nTrace;
        //!! for each fiber
        for (i=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            haveBadPixel = 0;
            if (length < 0) length = 0;
            map2dCount += length+8; // count for profile data
            *(map2d++) = length;
            *(map2d++) = xcen[i];
            *(map2d++) = (double)xmin[i];
            *(map2d++) = (double)xmax[i];
            *(map2dOrigin++) = length;
            *(map2dOrigin++) = xcen[i];
            *(map2dOrigin++) = (double)xmin[i];
            *(map2dOrigin++) = (double)xmax[i];

            eachReject=0;
            eachSum=0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                tempFlux[j] = fimage[j]-bmod[j];
                tempBack[j] = bmod[j];
                tempVar[j] = var[j];
                eachSum += aprofile[i*nCoeff][k];
                if (isfinite(var[j]))
                    eachReject += aprofile[i*nCoeff][k];
            }
            eachReject = eachReject/eachSum;
            if (isfinite(eachReject))
                reject2d[i] = eachReject;
            else { //!! eachReject is nan?
                eachReject = 0;
                reject2d[i] = 0;
            }


            lowF = i-1>=0 ? i-1 :0;
            highF = i+1<=nTrace-1 ? i+1 : nTrace-1;
            //!! for each adjecent fiber
            for (j=lowF; j<=highF; j++) {
                if (j==i||xmax[j]<xmin[i]||xmin[j]>xmax[i])
                    continue;
                lowX = xmin[j] > xmin[i] ? xmin[j] : xmin[i];
                highX = xmax[j] < xmax[i] ? xmax[j] : xmax[i];

                //!! for each coresponsding pixel
                if (p[j*nCoeff]==0) {
                    for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                        if (fabs(l-xcen[i])>halfWidth-1){ //!! point not far away from center
                            if (tempFlux[l]>0) {
                                tempFlux[l] = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                                eachRatio = (tempFlux[l])/(fimage[l]-bmod[l])>0?(tempFlux[l])/(fimage[l]-bmod[l]):0;
                                tempVar[l] = var[l] * eachRatio;
                            }
                        }
                    }
                } else {
                    for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                        //!! only subtract the first profile
                        eachFlux = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                        eachOtherFlux = ans[j*nCoeff]*aprofile[j*nCoeff][k];
                        //!! we overestimate the error here... it's safe but not good...
                        eachRatio = (eachFlux)/(eachFlux+eachOtherFlux);
                        if (tempFlux[l]>0) {
                            tempFlux[l] -= (fimage[l] - bmod[l] > 0 ? (fimage[l] - bmod[l]) * (1 - eachRatio) : 0);
                            tempVar[l] -= var[l] * (1 - eachRatio);
                        }
                    }
                }
            }
            sumFlux[i] = 0;
            sumFluxBack[i] = 0;
            if (p[i*nCoeff]==0||eachReject<map2DReject)
                sumVar[i] = 1./0.;
            else
                sumVar[i] = 0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                if (isfinite(tempFlux[j]) && isfinite(tempVar[j])) {
                    if (!(tempFlux[j]>0 && tempVar[j]>0)) {
                        tempFlux[j] = 0;
                        tempVar[j] = 0;
                    }
                    sumFlux[i] += tempFlux[j];
                    sumFluxBack[i] += bmod[j];
                    sumVar[i] += tempVar[j];
                } else {
                    sumFluxBack[i] += bmod[j];
                    tempFlux[j] = 0;
                    tempVar[j] = 1./0.;
                }
            }

            if (eachReject==0)
                eachReject=1;
            if (sumFlux[i]==0) {
                for (k = xmin[i]; k <= xmax[i]; k++)
                    tempFlux[k] = 1;
                sumVar[i] = 1./0.;
            }

            *(map2d++) = eachReject;
            *(map2dOrigin++) = 1;
            sumFlux[i] = sumFlux[i]/eachReject;

            *(map2d++) = sumFlux[i];
            *(map2d++) = sumFluxBack[i];
            sumVar[i]  = sumVar[i]/(eachReject*eachReject);
            *(map2d++) = sumVar[i];

            *(map2dOrigin++) = ans[i*nCoeff];
            *(map2dOrigin++) = -1;
            *(map2dOrigin++) = 1/(p[i*nCoeff]*p[i*nCoeff]);
            for(k=xmin[i]; k<=xmax[i]; k++)
                *(map2d++) = tempFlux[k];// only get map for the first conponent
            for(k=xmin[i],l=0; k<=xmax[i]; k++,l++)
                *(map2dOrigin++) = aprofile[i*nCoeff][l];

            if (sumVar[i]<0||!isfinite(sumFlux[i])) {
                printf("start\n");
                for (k = xmin[i]; k <= xmax[i]; k++)
                    printf("%f ", tempFlux[k]);// only get map for the first conponent
                printf("\n");
                for (k = xmin[i]; k <= xmax[i]; k++)
                    printf("%f ", tempVar[k]);// only get map for the first conponent
                printf("\n");
                printf("stop\n");
            }
        }
        map2dFirst[0] = map2dCount;
        map2dOriginFirst[0] = map2dCount;
        free(tempFlux);
        free(tempBack);
        free(tempVar);
        free(var);
    }

   //!! Free temp memory
    free(aprofileIn);
    free(aprofile);
    free(xmin);
    free(xmax);
    return good;
}
//#endregion: extract

//#region: extractGivenProfile
// fit image with multi-profile and a background at each xcen
long extractGivenProfile(
      //  xLen   xIndArr     yValue          invvar           xcen guess
        long nx, double* x, double* fimage, double* invvar,
        long nTrace, long *xmin, long *xmax,
        double *aprofileIn,
      //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
        long nPoly,  long nCoeff, long* ia,
        long calcCovar, double* reject, long* fullreject, long* partialreject,
        double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
        long debug) {
     // bmod is the background part, amod is the profile part
   //!! setup
    FILE *fp;
    long  iy;
    long  good;
    double ** aprofile;
    double ** apoly;
    double tsigma = 5.0;             /* set limits of profile influence   */
    double contribution = 0.0;       /* set limits of profile influence   */
    double x1, x2;                   /* Top and Lower Chebyshev x limit */
    long  wPoly;                    /* Number of coefficients nPoly + whopping terms */
    double reject1, reject2;         /* profile area rejection thresholds */
    double rejectflux;               /* profile area rejection thresholds */
    long  i,j,k,l,length, mfit, coeff, tTrace, ma;
    long  nBand = 2;
    double * ysub;
    double * beta;
    double profileSum, temp;
    double *ansFactor, ansTotal, totalP;
    /*  Reject 1 and 2 and flux are the three members of a double array  */
    reject2 = reject[0]; rejectflux = reject[1]; reject1 = reject[2];
    wPoly = nPoly; // background part
    tTrace = nCoeff*nTrace;
    ma = tTrace + wPoly;
   //!! fill the aprofile and apoly
    double *pAp=aprofileIn;
    aprofile = (double **)malloc(tTrace * sizeof(double *));
    long aux=0;
    for (i=0,k=0; i<nTrace; i++) {
        for (j=0; j<nCoeff; j++,k++) {
            aprofile[k] = pAp;
            pAp += xmax[i] - xmin[i] + 1;
            aux += xmax[i] - xmin[i] + 1;
        }
    }
    /*printf("aux: %ld\n", aux);*/
        if (debug==1) {
        fp=fopen("aprofile1.dat","wt");
        for(i=0, k=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            if (length < 0) length = 0;
            for(j=0; j<nCoeff; j++, k++) {
                //printf("%5ld %5ld %5ld %5ld\n", k, xmin[i], xmax[i], length);
                fprintf(fp, "%5ld %5ld %5ld %5ld ", k, xmin[i], xmax[i], length);
                for(int x=0; x<length; x++) {
                    fprintf(fp, "%8.4g ", aprofile[k][x]); } fprintf(fp, "\n"); }}
        fclose(fp);
        }

    apoly = (double **)malloc(wPoly * sizeof(double *));
    // Room for polynomial profiles
    for(i=0; i<wPoly; i++) {
        apoly[i] = (double *)malloc(nx * sizeof(double));
    }
    // fill poly
    x2 = (double) (nx-1);
    x1 = 0.0;
    if (nPoly > 0) fillPoly(apoly, x, nx, nPoly, x1, x2);
        if (debug==1) {
            fp=fopen("apoly.dat","wt");
            for(i=0; i<nPoly; i++) {
                fprintf(fp, "%5ld ", i);
                for(int x=0; x<nx; x++) {
                    fprintf(fp, "%g ", apoly[i][x]); } fprintf(fp, "\n"); }
            fclose(fp);
        }
   //!! checkRowFibers
    CheckRowFibers(aprofile, xmin, xmax, nTrace, nCoeff, ans, ia, invvar,
                    reject1, reject2, rejectflux, partialreject, fullreject);
   //!! Subtract out fixed variables first
    ysub = (double *)malloc(nx * sizeof(double));
    for(i=0, mfit=0; i<ma; i++) if(ia[i]) mfit++; // count for mfit
    for(i=0; i<nx; i++) ysub[i] = fimage[i]; // copy for fimage
    if(mfit != ma) { // subtract out fixed variables
        //fprintf(stderr, "Subtracting fixed variables\n");
        subtractProfile(ysub, nx, xmin, xmax, nTrace, nCoeff, aprofile, ia, ans);
        subtractPoly(ysub, nx, wPoly, apoly, &ia[tTrace], &ans[tTrace]);
    }
    beta = (double *) malloc(sizeof(double)*ma);
    for(i=0; i<ma; i++)  //init covar
       bzero((void *) covar[i] , sizeof(double)*ma);
    for(i=0; i<ma; i++) beta[i] = 0.0;
    //for(i=0; i<ma; i++)
        //assert(isfinite(ans[i]));
   //!! fill covar
    // model is B = (X^T W X)^{-1} X^T W Y
    // covar should be ( X^T W X )
    // beta  should be ( X^T W Y )
    fillCovar(ysub, invvar, nx, aprofile, apoly, nTrace, nCoeff, nBand,
           wPoly, beta, ia, covar, xmin, xmax);
    if (debug==1) {
        fp=fopen("fillCovar.dat","wt");
        for (i=0;i<ma;i++){
            for (j=0;j<ma;j++){
                fprintf(fp, "%g ", covar[i][j]); } fprintf(fp, "\n"); }
        fclose(fp);
    }
   //!! do LLT decomposition using the Cholesky–Banachiewicz and Cholesky–Crout algorithms
    // p is the L_ii,
    good = choldcRow(covar, ia, nTrace, nCoeff, nBand, wPoly, p);
    // if (mfit!=ma&&(proftype==0||proftype==-3)) bad -= 9999;
    if (good < 1) {
        for(j=0;j<ma;j++) {
            p[j] = 0.0;
            ans[j] = 0.0;
        }
        return good;
    } else {
         cholslRow(covar, ia, nTrace, nCoeff, nBand, wPoly, p, beta, ans);
         if (calcCovar > 0) {
             cholslRowCovar(covar, ia, nTrace, nCoeff, nBand, wPoly, p);
         }
    }
   //!! build ymodel
    // ymod is fitting result contains all profile and background (include mu and sigma)
    // amod only contains A and background
    for(i=0;i<nx;i++) ymod[i] = 0.0;
    //!! scattered light first
    for (j=nCoeff*nTrace,k=0;k<wPoly;j++,k++) {
        for(i=0; i < nx; i++) {
            //assert(isfinite(apoly[k][i]));
            ymod[i] += ans[j]*apoly[k][i];
            bmod[i] += ans[j]*apoly[k][i];
        }
    }
    for(i=0;i<nx;i++) amod[i] = ymod[i];
    //!! add fiber model
    ansFactor = (double *)malloc(sizeof(double)*nTrace);
    for (j=0,l=0;j<nTrace;j++)
        for (coeff=0; coeff<nCoeff; coeff++,l++)
            for (i=xmin[j], k=0; i <= xmax[j]; i++, k++) {
                ymod[i] += ans[l] * aprofile[l][k];
                //assert(isfinite(aprofile[l][k]));
            }

    //!! build amod
    for (j=0; j<nTrace;j++)
        for (i=xmin[j], k=0; i <= xmax[j]; i++, k++)
            amod[i] += ans[j*nCoeff]*aprofile[j*nCoeff][k];
    for(i=0; i<ma; i++)
        assert(isfinite(ans[i]));
   //!! Free temporary memory
    for(i=0; i<wPoly; i++) free(apoly[i]);
    free(aprofile);
    free(apoly);
    free(ysub);
    free(beta);
    free(ansFactor);
    return good;
}

//#endregion: extractGivenProfile

//#region: bad codes....
//#region: extractDirect
long extractDirect(
      //  xLen   xIndArr     yValue          invvar           xcen guess     sigma guess
        long nx, double* x, double* fimage, double* invvar, double* xcen, double* sigma,
      //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
      //    fiberNum   background    1,2 or 3
        long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
        long proftype, long calcCovar, double* reject, long* fullreject, long* partialreject,
        double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
        double* sumFlux, double* sumFluxBack, double* sumVar, double* reject2d,
        double* map2d, long halfWidth,
        double map2DReject, long debug) {
     // bmod is the background part, amod is the profile part
   //!! setup
    FILE *fp;
    long  bad;
    long  *xmin, *xmax;
    double ** aprofile;
    double contribution = 0.0;
    double tsigma = 5.0;             /* set limits of profile influence   */
    long  i,j,k,l,length, tTrace;
    double *map2dFirst;
    long map2dCount=2; // add count for total length and total fib number
    double *var;

    /*for (iy=0; iy < ma; iy++) covar[iy] = (double *) argv[argct]+iy*ma;*/
    xmin = (long *)malloc(sizeof(long)*nTrace);// checked, safe
    xmax = (long *)malloc(sizeof(long)*nTrace);// checked, safe
   //!! Find max and min of profile influence
    // have nothing to do with fimage
    //      tsigma is limits of profile influence       5 times sigma
    findXLimits(xmin, xmax, x, xcen, nx, nTrace, sigma, tsigma);
        if (debug==1) {
        fp=fopen("xminmax.dat","wt");
        for(i=0, k=0; i<nTrace; i++)
            fprintf(fp, "%5ld %5ld %5ld %5ld %5ld\n", i, i*nCoeff, xmin[i], xmax[i], xmax[i]-xmin[i]+1);
        fclose(fp);
        fp=fopen("fimage.dat","wt");
        for(i=0; i<nx; i++)
            fprintf(fp, "%5f ", fimage[i]);
        fclose(fp);
        fp=fopen("invvar.dat","wt");
        for(i=0; i<nx; i++)
            fprintf(fp, "%5f ", invvar[i]);
        fclose(fp);
        }
   //!! fill the aprofile
    tTrace = nCoeff*nTrace;
    // room for aprofile
    long aprofileLength=0;
    for(i=0; i<nTrace; i++){//Room for fiber profiles
        length = xmax[i] - xmin[i] + 1;
        if (length < 0) length = 0;
        aprofileLength += length*nCoeff;
    }
    double *aprofileIn;
    /*printf("aprofileLength: %ld\n", aprofileLength);*/
    aprofileIn = (double *)malloc(aprofileLength * sizeof(double));
    aprofile = (double **)malloc(tTrace * sizeof(double *));
    double *pAp=aprofileIn;
    long aux=0;
    for (i=0,k=0; i<nTrace; i++) {
        for (j=0; j<nCoeff; j++,k++) {
            aprofile[k] = pAp;
            pAp += xmax[i] - xmin[i] + 1;
            aux += xmax[i] - xmin[i] + 1;
        }
    }
    /*printf("aux: %ld\n", aux);*/
    // aprofile[nTrace*nCoeff][length of each fit]
    fillProfile(aprofile, x, xcen, xmin, xmax, sigma, nx, nCoeff,
                nTrace, proftype, sigmal, contribution);
        if (debug==1) {
        fp=fopen("aprofile.dat","wt");
        for(i=0, k=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            if (length < 0) length = 0;
            for(j=0; j<nCoeff; j++, k++) {
                //printf("%5ld %5ld %5ld %5ld\n", k, xmin[i], xmax[i], length);
                fprintf(fp, "%5ld %5ld %5ld %5ld ", k, xmin[i], xmax[i], length);
                for(int x=0; x<length; x++) {
                    fprintf(fp, "%8.4g ", aprofile[k][x]); } fprintf(fp, "\n"); }}
        fclose(fp);
        }
   //!! key function call
    bad = extractGivenProfile(
        nx, x, fimage, invvar,
        nTrace, xmin, xmax,
        aprofileIn,
        nPoly, nCoeff, ia,
        calcCovar, reject, fullreject, partialreject,
        ans, ymod, covar, p, bmod, amod,
        debug);
   //!! make map2d
    if (map2d) {
        double *tempFlux, *tempBack, *tempVar;
        double eachOtherFlux, eachFlux, eachRatio, eachReject, eachSum;
        int haveBadPixel;
        //!! temp flux all
        tempFlux = (double*)malloc(nx * sizeof(double));
        tempBack = (double*)malloc(nx * sizeof(double));
        tempVar = (double*)malloc(nx * sizeof(double));
        var = (double*)malloc(nx * sizeof(double));
        long lowF, highF, lowX, highX;
        for (i=0; i<nx; i++)
            var[i] = 1/invvar[i];

        map2dFirst=map2d;
        *(map2d++) = map2dCount;// right map2dCount num in the end of the if
        *(map2d++) = nTrace;
        //!! for each fiber
        for (i=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            haveBadPixel = 0;
            if (length < 0) length = 0;
            map2dCount += length+8; // count for profile data
            *(map2d++) = length;
            *(map2d++) = xcen[i];
            *(map2d++) = (double)xmin[i];
            *(map2d++) = (double)xmax[i];

            eachReject=0;
            eachSum=0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                tempFlux[j] = fimage[j]-bmod[j];
                tempBack[j] = bmod[j];
                tempVar[j] = var[j];
                eachSum += aprofile[i*nCoeff][k];
                if (isfinite(var[j]))
                    eachReject += aprofile[i*nCoeff][k];
            }
            eachReject = eachReject/eachSum;
            if (isfinite(eachReject))
                reject2d[i] = eachReject;
            else { //!! eachReject is nan?
                eachReject = 0;
                reject2d[i] = 0;
            }


            lowF = i-1>=0 ? i-1 :0;
            highF = i+1<=nTrace-1 ? i+1 : nTrace-1;
            //!! for each adjecent fiber
            for (j=lowF; j<=highF; j++) {
                if (j==i||xmax[j]<xmin[i]||xmin[j]>xmax[i])
                    continue;
                lowX = xmin[j] > xmin[i] ? xmin[j] : xmin[i];
                highX = xmax[j] < xmax[i] ? xmax[j] : xmax[i];

                //!! for each coresponsding pixel
                if (p[j*nCoeff]==0) {
                    for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                        if (fabs(l-xcen[i])>halfWidth-1){ //!! point not far away from center
                            if (tempFlux[l]>0) {
                                tempFlux[l] = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                                eachRatio = (tempFlux[l])/(fimage[l]-bmod[l])>0?(tempFlux[l])/(fimage[l]-bmod[l]):0;
                                tempVar[l] = var[l] * eachRatio;
                            }
                        }
                    }
                } else {
                    for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                        //!! only subtract the first profile
                        eachFlux = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                        eachOtherFlux = ans[j*nCoeff]*aprofile[j*nCoeff][k];
                        //!! we overestimate the error here... it's safe but not good...
                        eachRatio = (eachFlux)/(eachFlux+eachOtherFlux);
                        if (tempFlux[l]>0) {
                            tempFlux[l] -= (fimage[l] - bmod[l] > 0 ? (fimage[l] - bmod[l]) * (1 - eachRatio) : 0);
                            tempVar[l] -= var[l] * (1 - eachRatio);
                        }
                    }
                }
            }
            sumFlux[i] = 0;
            sumFluxBack[i] = 0;
            if (p[i*nCoeff]==0||eachReject<map2DReject)
                sumVar[i] = 1./0.;
            else
                sumVar[i] = 0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                if (isfinite(tempFlux[j]) && isfinite(tempVar[j])) {
                    if (!(tempFlux[j]>0 && tempVar[j]>0)) {
                        tempFlux[j] = 0;
                        tempVar[j] = 0;
                    }
                    sumFlux[i] += tempFlux[j];
                    sumFluxBack[i] += bmod[j];
                    sumVar[i] += tempVar[j];
                } else {
                    sumFluxBack[i] += bmod[j];
                    tempFlux[j] = 0;
                    tempVar[j] = 1./0.;
                }
            }

            if (eachReject==0)
                eachReject=1;
            if (sumFlux[i]==0) {
                for (k = xmin[i]; k <= xmax[i]; k++)
                    tempFlux[k] = 1;
                sumVar[i] = 1./0.;
            }

            *(map2d++) = eachReject;
            sumFlux[i] = sumFlux[i]/eachReject;
            *(map2d++) = sumFlux[i];
            *(map2d++) = sumFluxBack[i];
            sumVar[i]  = sumVar[i]/(eachReject*eachReject);
            *(map2d++) = sumVar[i];
            for(k=xmin[i]; k<=xmax[i]; k++)
                *(map2d++) = tempFlux[k];// only get map for the first conponent
            if (sumVar[i]<0||!isfinite(sumFlux[i])) {
                printf("start\n");
                for (k = xmin[i]; k <= xmax[i]; k++)
                    printf("%f ", tempFlux[k]);// only get map for the first conponent
                printf("\n");
                for (k = xmin[i]; k <= xmax[i]; k++)
                    printf("%f ", tempVar[k]);// only get map for the first conponent
                printf("\n");
                printf("stop\n");
            }
        }
        map2dFirst[0] = map2dCount;
        free(tempFlux);
        free(tempBack);
        free(tempVar);
        free(var);
    }

   //!! Free temp memory
    free(aprofileIn);
    free(aprofile);
    free(xmin);
    free(xmax);
    return bad;
}
//#endregion: extract

void Profile23_2(double *x, long ndat, double **y, double xcen, long xmin,
                 long xmax, double sigma, double sf, double oA, double oA3, long nCoeff) {
    long i,j,k;
    double base2,base3;
    double diff2,diff3,frac,absdiff3;
    double epow2,epow3;
    double sigma2, sigma3, nf2, nf3;
    double t;

    sigma2 = sigma;
    sigma3 = sigma2*sf;
    nf2 = 1.0/(sqrt(6.283185307)*sigma2);
    nf3 = 1.0/(2.575798634*sigma3);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff2 = (x[i] - xcen + frac)/sigma2;
                diff3 = (x[i] - xcen + frac)/sigma3;
                absdiff3 = fabs(diff3);
                epow2 = 0.5*diff2*diff2;
                epow3 = 1./3.*diff3*diff3*absdiff3;

                if (epow2 < MAXEXP){
                    base2 = exp(-epow2) * nf2;
                    base3 = exp(-epow3) * nf3;
                } else {
                    base2 = 0.0;
                    base3 = 0.0;
                }
                t = oA+oA3;
                // ans[0] = A
                y[0][k] += oA/t*base2 + oA3/t*base3;
                // ans[1] = A * dSigma
                if(nCoeff >1) y[1][k] += oA/t*(2*epow2-1)/sigma2*base2 + oA3/t*(3*epow3-1)/sigma3*base3;
                // ans[2] = A * dMu
                if(nCoeff >2) y[2][k] += oA/t*diff2/sigma2*base2 + oA3/t*diff3*absdiff3/sigma3*base3;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void fillProfile23(double **y, double *x, double *xcen, long *xmin,
        long *xmax, double *sigma, double *sf, double *oA, double *oA3, long nx, long nCoeff,
        long nTrace) {
    int i, j, m, length;
    long tCoeff;
    for (i=0,j=0; i<nTrace; i++,j+=nCoeff) {
        length = xmax[i] - xmin[i] + 1;
        if (length > 0 && sigma[i] > 0.0) {
            Profile23_2(x, nx, &y[j], xcen[i], xmin[i],
                    xmax[i], sigma[i], sf[i], oA[i], oA3[i], nCoeff);
        }
    }
}

//#region: extractDirect23
long extractDirect23(
      //  xLen   xIndArr     yValue          invvar           xcen guess
        long nx, double* x, double* fimage, double* invvar, double* xcen,
        double* sigma, double* sf, double* oA, double* oA3,
      //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
      //    fiberNum   background    1,2 or 3
        long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
        long proftype, long calcCovar, double* reject, long* fullreject, long* partialreject,
        double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
        double* sumFlux, double* sumFluxBack, double* sumVar, double* reject2d,
        double* map2d, double* map2dOrigin, long halfWidth,
        double map2DReject, long debug) {
     // bmod is the background part, amod is the profile part
   //!! setup
    FILE *fp;
    long  bad;
    long  *xmin, *xmax;
    double ** aprofile;
    double contribution = 0.0;
    double tsigma = 5.0;             /* set limits of profile influence   */
    long  i,j,k,l,length, tTrace;
    long map2dCount=2; // add count for total length and total fib number
    double *var;

    /*for (iy=0; iy < ma; iy++) covar[iy] = (double *) argv[argct]+iy*ma;*/
    xmin = (long *)malloc(sizeof(long)*nTrace);// checked, safe
    xmax = (long *)malloc(sizeof(long)*nTrace);// checked, safe
   //!! Find max and min of profile influence
    // have nothing to do with fimage
    //      tsigma is limits of profile influence       5 times sigma
    findXLimits(xmin, xmax, x, xcen, nx, nTrace, sigma, tsigma);
   //!! fill the aprofile
    tTrace = nCoeff*nTrace;
    // room for aprofile
    long aprofileLength=0;
    for(i=0; i<nTrace; i++){//Room for fiber profiles
        length = xmax[i] - xmin[i] + 1;
        if (length < 0) length = 0;
        aprofileLength += length*nCoeff;
    }
    double *aprofileIn;
    /*printf("aprofileLength: %ld\n", aprofileLength);*/
    aprofileIn = (double *)malloc(aprofileLength * sizeof(double));
    aprofile = (double **)malloc(tTrace * sizeof(double *));
    double *pAp=aprofileIn;
    long aux=0;
    for (i=0,k=0; i<nTrace; i++) {
        for (j=0; j<nCoeff; j++,k++) {
            aprofile[k] = pAp;
            pAp += xmax[i] - xmin[i] + 1;
            aux += xmax[i] - xmin[i] + 1;
        }
    }
    /*printf("aux: %ld\n", aux);*/
    // aprofile[nTrace*nCoeff][length of each fit]
    fillProfile23(aprofile, x, xcen, xmin, xmax, sigma, sf, oA, oA3, nx, nCoeff, nTrace);
   //!! key function call
    bad = extractGivenProfile(
        nx, x, fimage, invvar,
        nTrace, xmin, xmax,
        aprofileIn,
        nPoly, nCoeff, ia,
        calcCovar, reject, fullreject, partialreject,
        ans, ymod, covar, p, bmod, amod,
        debug);
   //!! make map2d
    if (map2d) {
        double *tempFlux, *tempBack, *tempVar;
        double eachOtherFlux, eachFlux, eachRatio, eachReject, eachSum;
        double* map2dOriginFirst, *map2dFirst;
        int haveBadPixel;
        //!! temp flux all
        tempFlux = (double*)malloc(nx * sizeof(double));
        tempBack = (double*)malloc(nx * sizeof(double));
        tempVar = (double*)malloc(nx * sizeof(double));
        var = (double*)malloc(nx * sizeof(double));
        long lowF, highF, lowX, highX;
        for (i=0; i<nx; i++)
            var[i] = 1/invvar[i];

        map2dFirst=map2d;
        map2dOriginFirst=map2dOrigin;
        *(map2d++) = map2dCount;// right map2dCount num in the end of the if
        *(map2dOrigin++) = map2dCount;// right map2dCount num in the end of the if
        *(map2d++) = nTrace;
        *(map2dOrigin++) = nTrace;
        //!! for each fiber
        for (i=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            haveBadPixel = 0;
            if (length < 0) length = 0;
            map2dCount += length+8; // count for profile data
            *(map2d++) = length;
            *(map2d++) = xcen[i];
            *(map2d++) = (double)xmin[i];
            *(map2d++) = (double)xmax[i];
            *(map2dOrigin++) = length;
            *(map2dOrigin++) = xcen[i];
            *(map2dOrigin++) = (double)xmin[i];
            *(map2dOrigin++) = (double)xmax[i];

            eachReject=0;
            eachSum=0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                tempFlux[j] = fimage[j]-bmod[j];
                tempBack[j] = bmod[j];
                tempVar[j] = var[j];
                eachSum += aprofile[i*nCoeff][k];
                if (isfinite(var[j]))
                    eachReject += aprofile[i*nCoeff][k];
            }
            eachReject = eachReject/eachSum;
            if (isfinite(eachReject))
                reject2d[i] = eachReject;
            else { //!! eachReject is nan?
                eachReject = 0;
                reject2d[i] = 0;
            }


            lowF = i-1>=0 ? i-1 :0;
            highF = i+1<=nTrace-1 ? i+1 : nTrace-1;
            //!! for each adjecent fiber
            for (j=lowF; j<=highF; j++) {
                if (j==i||xmax[j]<xmin[i]||xmin[j]>xmax[i])
                    continue;
                lowX = xmin[j] > xmin[i] ? xmin[j] : xmin[i];
                highX = xmax[j] < xmax[i] ? xmax[j] : xmax[i];

                //!! for each coresponsding pixel
                if (p[j*nCoeff]==0) {
                    for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                        if (fabs(l-xcen[i])>halfWidth-1){ //!! point not far away from center
                            if (tempFlux[l]>0) {
                                tempFlux[l] = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                                eachRatio = (tempFlux[l])/(fimage[l]-bmod[l])>0?(tempFlux[l])/(fimage[l]-bmod[l]):0;
                                tempVar[l] = var[l] * eachRatio;
                            }
                        }
                    }
                } else {
                    for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                        //!! only subtract the first profile
                        eachFlux = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                        eachOtherFlux = ans[j*nCoeff]*aprofile[j*nCoeff][k];
                        //!! we overestimate the error here... it's safe but not good...
                        eachRatio = (eachFlux)/(eachFlux+eachOtherFlux);
                        if (tempFlux[l]>0) {
                            tempFlux[l] -= (fimage[l] - bmod[l] > 0 ? (fimage[l] - bmod[l]) * (1 - eachRatio) : 0);
                            tempVar[l] -= var[l] * (1 - eachRatio);
                        }
                    }
                }
            }
            sumFlux[i] = 0;
            sumFluxBack[i] = 0;
            if (p[i*nCoeff]==0||eachReject<map2DReject)
                sumVar[i] = 1./0.;
            else
                sumVar[i] = 0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                if (isfinite(tempFlux[j]) && isfinite(tempVar[j])) {
                    if (!(tempFlux[j]>0 && tempVar[j]>0)) {
                        tempFlux[j] = 0;
                        tempVar[j] = 0;
                    }
                    sumFlux[i] += tempFlux[j];
                    sumFluxBack[i] += bmod[j];
                    sumVar[i] += tempVar[j];
                } else {
                    sumFluxBack[i] += bmod[j];
                    tempFlux[j] = 0;
                    tempVar[j] = 1./0.;
                }
            }

            if (eachReject==0)
                eachReject=1;
            if (sumFlux[i]==0) {
                for (k = xmin[i]; k <= xmax[i]; k++)
                    tempFlux[k] = 1;
                sumVar[i] = 1./0.;
            }

            *(map2d++) = eachReject;
            *(map2dOrigin++) = 1;
            sumFlux[i] = sumFlux[i]/eachReject;

            *(map2d++) = sumFlux[i];
            *(map2d++) = sumFluxBack[i];
            sumVar[i]  = sumVar[i]/(eachReject*eachReject);
            *(map2d++) = sumVar[i];

            *(map2dOrigin++) = ans[i*nCoeff];
            *(map2dOrigin++) = -1;
            *(map2dOrigin++) = 1/(p[i*nCoeff]*p[i*nCoeff]);
            for(k=xmin[i]; k<=xmax[i]; k++)
                *(map2d++) = tempFlux[k];// only get map for the first conponent
            for(k=xmin[i],l=0; k<=xmax[i]; k++,l++)
                *(map2dOrigin++) = aprofile[i*nCoeff][l];

            if (sumVar[i]<0||!isfinite(sumFlux[i])) {
                printf("start\n");
                for (k = xmin[i]; k <= xmax[i]; k++)
                    printf("%f ", tempFlux[k]);// only get map for the first conponent
                printf("\n");
                for (k = xmin[i]; k <= xmax[i]; k++)
                    printf("%f ", tempVar[k]);// only get map for the first conponent
                printf("\n");
                printf("stop\n");
            }
        }
        map2dFirst[0] = map2dCount;
        map2dOriginFirst[0] = map2dCount;
        free(tempFlux);
        free(tempBack);
        free(tempVar);
        free(var);
    }

   //!! Free temp memory
    free(aprofileIn);
    free(aprofile);
    free(xmin);
    free(xmax);
    return bad;
}
//#endregion: extract

void ProfileOnly23(double *x, long ndat, double **y, double xcen, long xmin,
                 long xmax, double sigma, double sf, double oA, double oA3, long nCoeff) {
    long i,j,k;
    double base2,base3;
    double diff2,diff3,frac,absdiff3;
    double epow2,epow3;
    double sigma2, sigma3, nf2, nf3;
    double t;

    sigma2 = sigma;
    sigma3 = sigma2*sf;
    nf2 = 1.0/(sqrt(6.283185307)*sigma2);
    nf3 = 1.0/(2.575798634*sigma3);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff2 = (x[i] - xcen + frac)/sigma2;
                diff3 = (x[i] - xcen + frac)/sigma3;
                absdiff3 = fabs(diff3);
                epow2 = 0.5*diff2*diff2;
                epow3 = 1./3.*diff3*diff3*absdiff3;

                if (epow2 < MAXEXP){
                    base2 = exp(-epow2) * nf2;
                    base3 = exp(-epow3) * nf3;
                } else {
                    base2 = 0.0;
                    base3 = 0.0;
                }
                // ans[0] = A
                y[0][k] += base2;
                // ans[1] = A3
                y[1][k] += base3;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void fillProfileOnly23(double **y, double *x, double *xcen, long *xmin,
        long *xmax, double *sigma, double *sf, double *oA, double *oA3, long nx, long nCoeff,
        long nTrace) {
    int i, j, m, length;
    long tCoeff;
    for (i=0,j=0; i<nTrace; i++,j+=nCoeff) {
        length = xmax[i] - xmin[i] + 1;
        if (length > 0 && sigma[i] > 0.0) {
            ProfileOnly23(x, nx, &y[j], xcen[i], xmin[i],
                    xmax[i], sigma[i], sf[i], oA[i], oA3[i], nCoeff);
        }
    }
}

//#region: extractOnly23
//!! to use this function, set nCoeff=2, or set nCoeff=3 and let ia = ...1 1 0...
//!! oA and oA3 is of no use in this function
long extractOnly23(
      //  xLen   xIndArr     yValue          invvar           xcen guess
        long nx, double* x, double* fimage, double* invvar, double* xcen,
        double* sigma, double* sf, double* oA, double* oA3,
      //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
      //    fiberNum   background    1,2 or 3
        long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
        long proftype, long calcCovar, double* reject, long* fullreject, long* partialreject,
        double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
        double* sumFlux, double* sumFluxBack, double* sumVar, double* reject2d,
        double* map2d, double* map2dOrigin, long halfWidth,
        double map2DReject, long debug) {
     // bmod is the background part, amod is the profile part
   //!! setup
    FILE *fp;
    long  bad;
    long  *xmin, *xmax;
    double ** aprofile;
    double contribution = 0.0;
    double tsigma = 5.0;             /* set limits of profile influence   */
    long  i,j,k,l,length, tTrace;
    long map2dCount=2; // add count for total length and total fib number
    double *var;

    /*for (iy=0; iy < ma; iy++) covar[iy] = (double *) argv[argct]+iy*ma;*/
    xmin = (long *)malloc(sizeof(long)*nTrace);// checked, safe
    xmax = (long *)malloc(sizeof(long)*nTrace);// checked, safe
   //!! Find max and min of profile influence
    // have nothing to do with fimage
    //      tsigma is limits of profile influence       5 times sigma
    findXLimits(xmin, xmax, x, xcen, nx, nTrace, sigma, tsigma);
   //!! fill the aprofile
    tTrace = nCoeff*nTrace;
    // room for aprofile
    long aprofileLength=0;
    for(i=0; i<nTrace; i++){//Room for fiber profiles
        length = xmax[i] - xmin[i] + 1;
        if (length < 0) length = 0;
        aprofileLength += length*nCoeff;
    }
    double *aprofileIn;
    /*printf("aprofileLength: %ld\n", aprofileLength);*/
    aprofileIn = (double *)malloc(aprofileLength * sizeof(double));
    aprofile = (double **)malloc(tTrace * sizeof(double *));
    double *pAp=aprofileIn;
    long aux=0;
    for (i=0,k=0; i<nTrace; i++) {
        for (j=0; j<nCoeff; j++,k++) {
            aprofile[k] = pAp;
            pAp += xmax[i] - xmin[i] + 1;
            aux += xmax[i] - xmin[i] + 1;
        }
    }
    /*printf("aux: %ld\n", aux);*/
    // aprofile[nTrace*nCoeff][length of each fit]
    fillProfileOnly23(aprofile, x, xcen, xmin, xmax, sigma, sf, oA, oA3, nx, nCoeff, nTrace);
   //!! key function call
    bad = extractGivenProfile(
        nx, x, fimage, invvar,
        nTrace, xmin, xmax,
        aprofileIn,
        nPoly, nCoeff, ia,
        calcCovar, reject, fullreject, partialreject,
        ans, ymod, covar, p, bmod, amod,
        debug);
   //!! make map2d
    if (map2d) {
        double *tempFlux, *tempBack, *tempVar;
        double eachOtherFlux, eachFlux, eachRatio, eachReject, eachSum;
        double* map2dOriginFirst, *map2dFirst;
        int haveBadPixel;
        //!! temp flux all
        tempFlux = (double*)malloc(nx * sizeof(double));
        tempBack = (double*)malloc(nx * sizeof(double));
        tempVar = (double*)malloc(nx * sizeof(double));
        var = (double*)malloc(nx * sizeof(double));
        long lowF, highF, lowX, highX;
        for (i=0; i<nx; i++)
            var[i] = 1/invvar[i];

        map2dFirst=map2d;
        map2dOriginFirst=map2dOrigin;
        *(map2d++) = map2dCount;// right map2dCount num in the end of the if
        *(map2dOrigin++) = map2dCount;// right map2dCount num in the end of the if
        *(map2d++) = nTrace;
        *(map2dOrigin++) = nTrace;
        //!! for each fiber
        for (i=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            haveBadPixel = 0;
            if (length < 0) length = 0;
            map2dCount += length+8; // count for profile data
            *(map2d++) = length;
            *(map2d++) = xcen[i];
            *(map2d++) = (double)xmin[i];
            *(map2d++) = (double)xmax[i];
            *(map2dOrigin++) = length;
            *(map2dOrigin++) = xcen[i];
            *(map2dOrigin++) = (double)xmin[i];
            *(map2dOrigin++) = (double)xmax[i];

            eachReject=0;
            eachSum=0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                tempFlux[j] = fimage[j]-bmod[j];
                tempBack[j] = bmod[j];
                tempVar[j] = var[j];
                eachSum += aprofile[i*nCoeff][k];
                if (isfinite(var[j]))
                    eachReject += aprofile[i*nCoeff][k];
            }
            eachReject = eachReject/eachSum;
            if (isfinite(eachReject))
                reject2d[i] = eachReject;
            else { //!! eachReject is nan?
                eachReject = 0;
                reject2d[i] = 0;
            }


            lowF = i-1>=0 ? i-1 :0;
            highF = i+1<=nTrace-1 ? i+1 : nTrace-1;
            //!! for each adjecent fiber
            for (j=lowF; j<=highF; j++) {
                if (j==i||xmax[j]<xmin[i]||xmin[j]>xmax[i])
                    continue;
                lowX = xmin[j] > xmin[i] ? xmin[j] : xmin[i];
                highX = xmax[j] < xmax[i] ? xmax[j] : xmax[i];

                //!! for each coresponsding pixel
                if (p[j*nCoeff]==0) {
                    for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                        if (fabs(l-xcen[i])>halfWidth-1){ //!! point not far away from center
                            if (tempFlux[l]>0) {
                                tempFlux[l] = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                                eachRatio = (tempFlux[l])/(fimage[l]-bmod[l])>0?(tempFlux[l])/(fimage[l]-bmod[l]):0;
                                tempVar[l] = var[l] * eachRatio;
                            }
                        }
                    }
                } else {
                    for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                        //!! only subtract the first profile
                        eachFlux = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                        eachOtherFlux = ans[j*nCoeff]*aprofile[j*nCoeff][k];
                        //!! we overestimate the error here... it's safe but not good...
                        eachRatio = (eachFlux)/(eachFlux+eachOtherFlux);
                        if (tempFlux[l]>0) {
                            tempFlux[l] -= (fimage[l] - bmod[l] > 0 ? (fimage[l] - bmod[l]) * (1 - eachRatio) : 0);
                            tempVar[l] -= var[l] * (1 - eachRatio);
                        }
                    }
                }
            }
            sumFlux[i] = 0;
            sumFluxBack[i] = 0;
            if (p[i*nCoeff]==0||eachReject<map2DReject)
                sumVar[i] = 1./0.;
            else
                sumVar[i] = 0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                if (isfinite(tempFlux[j]) && isfinite(tempVar[j])) {
                    if (!(tempFlux[j]>0 && tempVar[j]>0)) {
                        tempFlux[j] = 0;
                        tempVar[j] = 0;
                    }
                    sumFlux[i] += tempFlux[j];
                    sumFluxBack[i] += bmod[j];
                    sumVar[i] += tempVar[j];
                } else {
                    sumFluxBack[i] += bmod[j];
                    tempFlux[j] = 0;
                    tempVar[j] = 1./0.;
                }
            }

            if (eachReject==0)
                eachReject=1;
            if (sumFlux[i]==0) {
                for (k = xmin[i]; k <= xmax[i]; k++)
                    tempFlux[k] = 1;
                sumVar[i] = 1./0.;
            }

            *(map2d++) = eachReject;
            *(map2dOrigin++) = 1;
            sumFlux[i] = sumFlux[i]/eachReject;

            *(map2d++) = sumFlux[i];
            *(map2d++) = sumFluxBack[i];
            sumVar[i]  = sumVar[i]/(eachReject*eachReject);
            *(map2d++) = sumVar[i];

            *(map2dOrigin++) = ans[i*nCoeff];
            *(map2dOrigin++) = -1;
            *(map2dOrigin++) = 1/(p[i*nCoeff]*p[i*nCoeff]);
            for(k=xmin[i]; k<=xmax[i]; k++)
                *(map2d++) = tempFlux[k];// only get map for the first conponent
            for(k=xmin[i],l=0; k<=xmax[i]; k++,l++)
                *(map2dOrigin++) = aprofile[i*nCoeff][l];

            if (sumVar[i]<0||!isfinite(sumFlux[i])) {
                printf("start\n");
                for (k = xmin[i]; k <= xmax[i]; k++)
                    printf("%f ", tempFlux[k]);// only get map for the first conponent
                printf("\n");
                for (k = xmin[i]; k <= xmax[i]; k++)
                    printf("%f ", tempVar[k]);// only get map for the first conponent
                printf("\n");
                printf("stop\n");
            }
        }
        map2dFirst[0] = map2dCount;
        map2dOriginFirst[0] = map2dCount;
        free(tempFlux);
        free(tempBack);
        free(tempVar);
        free(var);
    }

   //!! Free temp memory
    free(aprofileIn);
    free(aprofile);
    free(xmin);
    free(xmax);
    return bad;
}
//#endregion: extract

//#region: extractUsingProfileWapper
long extractUsingProfileWapper(
         long nx, double* x, double* fimage, double* invvar,
         long nTrace, long nPoly, long nCoeff, long* ia,
         long calcCovar, double* reject, long* fullreject, long* partialreject,
         double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
         double* sumFlux, double* sumFluxBack, double* sumVar, double* reject2d,
         double* inMap2D, long halfWidth,
         double* map2d, double map2DReject, long debug) {
   //!! setup
    long  bad, tTrace;
    long  *xmin, *xmax;
    double *xcen;
    double ** aprofile;
    long  i,j,k,l,length;
    double *map2dFirst;
    double *map2DPointer;
    long map2dCount=2; // add count for total length and total fib number
    double *var;

    tTrace = nCoeff*nTrace;
    xmin = (long *)malloc(sizeof(long)*nTrace);// checked, safe
    xmax = (long *)malloc(sizeof(long)*nTrace);// checked, safe
    xcen = (double *)malloc(sizeof(double)*nTrace);// checked, safe
   //!! Find max and min of profile influence
    long aprofileLength=0;
    map2DPointer = inMap2D+2;
    for (i=0; i<nTrace; i++) {
        length = (long)*(map2DPointer++);
        assert(length<9999);
        xcen[i] = *(map2DPointer++);
        xmin[i] = (long)*(map2DPointer++);
        xmax[i] = (long)*(map2DPointer++);
        map2DPointer += 4;
        map2DPointer += length;
        aprofileLength += length;
    }
   //!! fill the aprofile
    double *aprofileIn;
    aprofileIn = (double *)malloc(aprofileLength * sizeof(double));
    double *pAp=aprofileIn;
    map2DPointer = inMap2D+2;
    for (i=0; i<nTrace; i++) {
        length = (long)*(map2DPointer++);
        map2DPointer += 7;
        for (j=0; j<length; j++)
            *(pAp++) = *(map2DPointer++);
    }

    aprofile = (double **)malloc(tTrace * sizeof(double *));
    pAp=aprofileIn;
    for (i=0,k=0; i<nTrace; i++) {
        for (j=0; j<nCoeff; j++,k++) {
            aprofile[k] = pAp;
            pAp += xmax[i] - xmin[i] + 1;
        }
    }

   //!! key function call
    bad = extractGivenProfile(
        nx, x, fimage, invvar,
        nTrace, xmin, xmax,
        aprofileIn,
        nPoly, nCoeff, ia,
        calcCovar, reject, fullreject, partialreject,
        ans, ymod, covar, p, bmod, amod,
        debug);
   //!! make map2d
    if (map2d) {
        double *tempFlux, *tempBack, *tempVar, *map2dOriginFirst;
        double eachOtherFlux, eachFlux, eachRatio, eachReject, eachSum;
        int haveBadPixel;
        //!! temp flux all
        tempFlux = (double*)malloc(nx * sizeof(double));
        tempBack = (double*)malloc(nx * sizeof(double));
        tempVar = (double*)malloc(nx * sizeof(double));
        var = (double*)malloc(nx * sizeof(double));
        long lowF, highF, lowX, highX;
        for (i=0; i<nx; i++)
            var[i] = 1/invvar[i];

        map2dFirst=map2d;
        *(map2d++) = map2dCount;// right map2dCount num in the end of the if
        *(map2d++) = nTrace;
        //!! for each fiber
        for (i=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            haveBadPixel = 0;
            if (length < 0) length = 0;
            map2dCount += length+8; // count for profile data
            *(map2d++) = length;
            *(map2d++) = xcen[i];
            *(map2d++) = (double)xmin[i];
            *(map2d++) = (double)xmax[i];

            eachReject=0;
            eachSum=0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                tempFlux[j] = fimage[j]-bmod[j];
                tempBack[j] = bmod[j];
                tempVar[j] = var[j];
                eachSum += aprofile[i*nCoeff][k];
                if (isfinite(var[j]))
                    eachReject += aprofile[i*nCoeff][k];
            }
            eachReject = eachReject/eachSum;
            if (isfinite(eachReject))
                reject2d[i] = eachReject;
            else {
                eachReject = 0;
                reject2d[i] = 0;
            }


            lowF = i-1>=0 ? i-1 :0;
            highF = i+1<=nTrace-1 ? i+1 : nTrace-1;
            //!! for each adjecent fiber
            for (j=lowF; j<=highF; j++) {
                if (j==i||xmax[j]<xmin[i]||xmin[j]>xmax[i])
                    continue;
                lowX = xmin[j] > xmin[i] ? xmin[j] : xmin[i];
                highX = xmax[j] < xmax[i] ? xmax[j] : xmax[i];

                //!! for each coresponsding pixel
                if (p[j*nCoeff]==0) {
                    for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                        if (fabs(l-xcen[i])>halfWidth-1){ //!! point not far away from center
                            if (tempFlux[l]>0) {
                                tempFlux[l] = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                                eachRatio = (tempFlux[l])/(fimage[l]-bmod[l])>0?(tempFlux[l])/(fimage[l]-bmod[l]):0;
                                tempVar[l] = var[l] * eachRatio;
                            }
                        }
                    }
                } else {
                    for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                        //!! only subtract the first profile
                        eachFlux = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                        eachOtherFlux = ans[j*nCoeff]*aprofile[j*nCoeff][k];
                        //!! we overestimate the error here... it's safe but not good...
                        eachRatio = (eachFlux)/(eachFlux+eachOtherFlux);
                        if (tempFlux[l]>0) {
                            tempFlux[l] -= (fimage[l] - bmod[l] > 0 ? (fimage[l] - bmod[l]) * (1 - eachRatio) : 0);
                            tempVar[l] -= var[l] * (1 - eachRatio);
                        }
                    }
                }
            }
            sumFlux[i] = 0;
            sumFluxBack[i] = 0;
            if (p[i*nCoeff]==0||eachReject<map2DReject)
                sumVar[i] = 1./0.;
            else
                sumVar[i] = 0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                if (isfinite(tempFlux[j]) && isfinite(tempVar[j])) {
                    if (!(tempFlux[j]>0 && tempVar[j]>0)) {
                        tempFlux[j] = 0;
                        tempVar[j] = 0;
                    }
                    sumFlux[i] += tempFlux[j];
                    sumFluxBack[i] += bmod[j];
                    sumVar[i] += tempVar[j];
                } else {
                    sumFluxBack[i] += bmod[j];
                    tempFlux[j] = 0;
                    tempVar[j] = 1./0.;
                }
            }

            if (eachReject==0)
                eachReject=1;
            if (sumFlux[i]==0) {
                for (k = xmin[i]; k <= xmax[i]; k++)
                    tempFlux[k] = 1;
                sumVar[i] = 1./0.;
            }

            *(map2d++) = eachReject;
            sumFlux[i] = sumFlux[i]/eachReject;
            *(map2d++) = sumFlux[i];
            *(map2d++) = sumFluxBack[i];
            sumVar[i]  = sumVar[i]/(eachReject*eachReject);
            *(map2d++) = sumVar[i];
            for(k=xmin[i]; k<=xmax[i]; k++)
                *(map2d++) = tempFlux[k];// only get map for the first conponent
            if (sumVar[i]<0||!isfinite(sumFlux[i])) {
                printf("start\n");
                for (k = xmin[i]; k <= xmax[i]; k++)
                    printf("%f ", tempFlux[k]);// only get map for the first conponent
                printf("\n");
                for (k = xmin[i]; k <= xmax[i]; k++)
                    printf("%f ", tempVar[k]);// only get map for the first conponent
                printf("\n");
                printf("stop\n");
            }
        }
        map2dFirst[0] = map2dCount;
        free(tempFlux);
        free(tempBack);
        free(tempVar);
        free(var);
    }

   //!! Free temp memory
    free(aprofileIn);
    free(aprofile);
    free(xmin);
    free(xmax);
    free(xcen);
    return bad;

}
//#endregion: extractUsingProfileWapper

//#region: extract_row23
// begin of the functions
long extract_row23(
      //  xLen   xIndArr     yValue          invvar
        long nx, double* x, double* fimage, double* invvar, double* xcen, double* sigma2, double* sigma3,
      //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
      //    fiberNum   background    1,2 or 3
        long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
        // nCoeff should be 2 here...
        long doLastFit, long calcCovar, double* reject, long* fullreject, long* partialreject,
        // proftype is should < -10 here
        double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
        long maxHalfWidth, double* map2d, long debug) {
     // bmod is the background part, amod is the profile part
   // setup
    FILE *fp;
    long  iy;
    long  bad;
    long  *xmin, *xmax, *xInd;
    double ** aprofile;
    double ** apoly;
    double tsigma = 5.0;             /* set limits of profile influence   */
    double contribution = 0.0;       /* set limits of profile influence   */
    double x1, x2;                   /* Top and Lower Chebyshev x limit */
    long  wPoly;                    /* Number of coefficients nPoly + whopping terms */
    double reject1, reject2;         /* profile area rejection thresholds */
    double rejectflux;               /* profile area rejection thresholds */
    long  i,j,k,l,length, mfit, coeff, tTrace, ma;
    long  nBand = 2;
    double * ysub;
    double * beta;
    double profileSum, temp, *map2dFirst;
    double *ansFactor, ansTotal, totalP;
    map2dFirst = map2d;
    long map2dCount=2; // add count for total length and total fib number
    /*  Reject 1 and 2 and flux are the three members of a double array  */
    reject2 = reject[0]; rejectflux = reject[1]; reject1 = reject[2];
    wPoly = nPoly;
    tTrace = nCoeff*nTrace;
    ma = tTrace + wPoly;
    /*covar  = (double **) malloc(ma * sizeof(double *));*/
    /*for (iy=0; iy < ma; iy++) covar[iy] = (double *) argv[argct]+iy*ma;*/
    xmin = (long *)malloc(sizeof(long)*nTrace);
    xmax = (long *)malloc(sizeof(long)*nTrace);
    xInd = (long *)malloc(sizeof(long)*nTrace);
   // Find max and min of profile influence
    // have nothing to do with fimage
    //                                                  5 times sigma
    nCoeff = 2;
    findXLimits(xmin, xmax, x, xcen, nx, nTrace, sigma3, tsigma);
        if (debug==1) {
        fp=fopen("xminmax.dat","wt");
        for(i=0, k=0; i<nTrace; i++)
            fprintf(fp, "%5ld %5ld %5ld %5ld %5ld\n", i, i*nCoeff, xmin[i], xmax[i], xmax[i]-xmin[i]+1);
        fclose(fp);
        fp=fopen("fimage.dat","wt");
        for(i=0; i<nx; i++)
            fprintf(fp, "%5f ", fimage[i]);
        fclose(fp);
        fp=fopen("invvar.dat","wt");
        for(i=0; i<nx; i++)
            fprintf(fp, "%5f ", invvar[i]);
        fclose(fp);
        }
   // fill the aprofile and apoly
    // room for aprofile
    aprofile = (double **)malloc(tTrace * sizeof(double *));
    apoly = (double **)malloc(wPoly * sizeof(double *));
    for(i=0, k=0; i<nTrace; i++){//Room for fiber profiles
        length = xmax[i] - xmin[i] + 1;
        if (length < 0) length = 0;
        for(j=0; j<nCoeff; j++, k++) {
            aprofile[k] = (double *)malloc(length * sizeof(double));
        }
    }
    fillProfile(aprofile, x, xcen, xmin, xmax, sigma2, nx, nCoeff,
                nTrace, -232, sigmal, contribution);
    fillProfile(aprofile, x, xcen, xmin, xmax, sigma3, nx, nCoeff,
                nTrace, -233, sigmal, contribution);
        if (debug==1) {
        fp=fopen("aprofile.dat","wt");
        for(i=0, k=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            if (length < 0) length = 0;
            for(j=0; j<nCoeff; j++, k++) {
                //printf("%5ld %5ld %5ld %5ld\n", k, xmin[i], xmax[i], length);
                fprintf(fp, "%5ld %5ld %5ld %5ld ", k, xmin[i], xmax[i], length);
                for(int x=0; x<length; x++) {
                    fprintf(fp, "%g ", aprofile[k][x]); } fprintf(fp, "\n"); }}
        fclose(fp);
        }

    // Room for polynomial profiles
    for(i=0; i<wPoly; i++) {
        apoly[i] = (double *)malloc(nx * sizeof(double));
        /*tTrace = nCoeff*nTrace;*/
        /*ia[i+tTrace] = ia[i+tTrace];*/
        /*ans[i+tTrace] = ans[i+tTrace];*/
    }
    // fill poly
    x2 = (double) (nx-1);
    x1 = 0.0;
    if (nPoly > 0) fillPoly(apoly, x, nx, nPoly, x1, x2);
    if (debug==1) {
        fp=fopen("apoly.dat","wt");
        for(i=0; i<nPoly; i++) {
            fprintf(fp, "%5ld ", i);
            for(int x=0; x<nx; x++) {
                fprintf(fp, "%g ", apoly[i][x]); } fprintf(fp, "\n"); }
        fclose(fp);
    }
   // checkRowFibers
    CheckRowFibers(aprofile, xmin, xmax, nTrace, nCoeff, ans, ia, invvar,
                    reject1, reject2, rejectflux, partialreject, fullreject);
   // Subtract out fixed variables first
    ysub = (double *)malloc(nx * sizeof(double));
    for(i=0, mfit=0; i<ma; i++) if(ia[i]) mfit++; // count for mfit
    for(i=0; i<nx; i++) ysub[i] = fimage[i]; // copy for fimage
    if(mfit != ma) { // subtract out fixed variables 
        //fprintf(stderr, "Subtracting fixed variables\n");
        subtractProfile(ysub, nx, xmin, xmax, nTrace, nCoeff, aprofile, ia, ans);
        subtractPoly(ysub, nx, wPoly, apoly, &ia[tTrace], &ans[tTrace]);
    } 
    beta = (double *) malloc(sizeof(double)*ma);
    for(i=0; i<ma; i++)  //init covar
       bzero((void *) covar[i] , sizeof(double)*ma);  
    for(i=0; i<ma; i++) beta[i] = 0.0;
   // fill covar
    // model is B = (X^T W X)^{-1} X^T W Y
    // covar should be ( X^T W X )
    // beta  should be ( X^T W Y )
    fillCovar(ysub, invvar, nx, aprofile, apoly, nTrace, nCoeff, nBand,
           wPoly, beta, ia, covar, xmin, xmax);
    if (debug==1) {
        fp=fopen("fillCovar.dat","wt");
        for (i=0;i<ma;i++){
            for (j=0;j<ma;j++){
                fprintf(fp, "%g ", covar[i][j]); } fprintf(fp, "\n"); }
        fclose(fp);
    }
   // do LLT decomposition using the Cholesky–Banachiewicz and Cholesky–Crout algorithms
    // p is the L_ii,
    bad = choldcRow(covar, ia, nTrace, nCoeff, nBand, wPoly, p);
    // if (mfit!=ma&&(proftype==0||proftype==-3)) bad -= 9999;
    if (bad < 1) {
        for(j=0;j<ma;j++) {
            p[j] = 0.0;
            ans[j] = 0.0;
        }
    } else {
         cholslRow(covar, ia, nTrace, nCoeff, nBand, wPoly, p, beta, ans);
         if (calcCovar > 0) {
             cholslRowCovar(covar, ia, nTrace, nCoeff, nBand, wPoly, p);
         }
    }
   // build ymodel
    // ymod is fitting result contains all profile and background (include mu and sigma)
    // amod only contains A and background
    for(i=0;i<nx;i++) ymod[i] = 0.0;
    // scattered light first
    for (j=nCoeff*nTrace,k=0;k<wPoly;j++,k++) {
        for(i=0; i < nx; i++) {
            ymod[i] += ans[j]*apoly[k][i];
            bmod[i] += ans[j]*apoly[k][i];
        }
    }
    for(i=0;i<nx;i++) amod[i] = ymod[i];
    // add fiber model
    ansFactor = (double *)malloc(sizeof(double)*nTrace);
    for (j=0,l=0;j<nTrace;j++)
        for (coeff=0; coeff<nCoeff; coeff++,l++)
            for (i=xmin[j], k=0; i <= xmax[j]; i++, k++)
                ymod[i] += ans[l]*aprofile[l][k];
    // ymod and amod is the same here...
    for (j=0; j<nTrace;j++){
        for (i=xmin[j], k=0; i <= xmax[j]; i++, k++) {
            for (l=0;l<nCoeff;l++) {
                amod[i] += ans[j*nCoeff+l]*aprofile[j*nCoeff+l][k]; } } }
    
    // modify profile for output
    for (j=0; j<nTrace;j++){
        ansTotal = 0;
        totalP = 0;
        for (l=0;l<nCoeff;l++) {
            ansFactor[l] = ans[j*nCoeff+l];
            ansTotal += ans[j*nCoeff+l];
            //ans[j*nCoeff+l] = 0;
            totalP += 1.0/(p[j*nCoeff+l]*p[j*nCoeff+l]); // sum(sigma^2)
            //p[j*nCoeff+l] = 0;
        }
        ansTotal+=1e-20;
        for (l=0;l<nCoeff;l++) {
            ansFactor[l] /= ansTotal;
            ans[j*nCoeff+l] = ansFactor[l];
        }
        ans[j*nCoeff] = ansTotal;
        //totalP = 1.0/sqrt(totalP); // ivar = 1/sum(sigma^2)
        //p[j*nCoeff] = totalP;
        
        // modify the first profile
        for (i=xmin[j], k=0; i <= xmax[j]; i++, k++) {
            aprofile[j*nCoeff][k]      = ansFactor[0]*aprofile[j*nCoeff][k];
        }
        // add later profile
        for (l=1;l<nCoeff;l++) {
            for (i=xmin[j], k=0; i <= xmax[j]; i++, k++) {
                aprofile[j*nCoeff][k] += ansFactor[l]*aprofile[j*nCoeff+l][k];
            }
        }
    }
    if (maxHalfWidth>0) {
        *(map2d++) = map2dCount;// right map2dCount num in the end of the if
        *(map2d++) = nTrace;
        for (i=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            if (length < 0) length = 0;
            map2dCount += length+3; // count for profile data
            *(map2d++) = length;
            *(map2d++) = (double)xmin[i];
            *(map2d++) = (double)xmax[i];
            for(k=xmin[i], j=0; k<=xmax[i]; k++, j++)
                *(map2d++) = aprofile[i*nCoeff][j];// only get map for the first conponent
        }
        map2dFirst[0] = map2dCount;
    }
    if (debug==1) {
        fp=fopen("ymod.dat","wt");
        for(i=0; i<nx; i++)
            fprintf(fp, "%5f ", ymod[i]);
        fclose(fp);
        fp=fopen("amod.dat","wt");
        for(i=0; i<nx; i++)
            fprintf(fp, "%5f ", amod[i]);
        fclose(fp);
        fp=fopen("ans.dat","wt");
        for(i=0; i<ma; i++)
            fprintf(fp, "%5f ", ans[i]);
        fclose(fp);
    }
   // do last fit
    if (doLastFit) {
        for (i=0;i<nTrace;i++)
            ia[i*nCoeff+1] = 0;
       // checkRowFibers
        CheckRowFibers(aprofile, xmin, xmax, nTrace, nCoeff, ans, ia, invvar,
                        reject1, reject2, rejectflux, partialreject, fullreject);
       // Subtract out fixed variables first
        for(i=0, mfit=0; i<ma; i++) if(ia[i]) mfit++; // count for mfit
        for(i=0; i<nx; i++) ysub[i] = fimage[i]; // copy for fimage
        for(i=0; i<ma; i++)  //init covar
           bzero((void *) covar[i] , sizeof(double)*ma);
        for(i=0; i<ma; i++){
            beta[i] = 0.0;
            p[i] = 0.0;
            ans[i] = 0.0;
        }
       // fill covar
        // model is B = (X^T W X)^{-1} X^T W Y
        // covar should be ( X^T W X )
        // beta  should be ( X^T W Y )
        fillCovar(ysub, invvar, nx, aprofile, apoly, nTrace, nCoeff, nBand,
               wPoly, beta, ia, covar, xmin, xmax);
        if (debug==1) {
            fp=fopen("fillCovar.dat","wt");
            for (i=0;i<ma;i++){
                for (j=0;j<ma;j++){
                    fprintf(fp, "%g ", covar[i][j]); } fprintf(fp, "\n"); }
            fclose(fp);
        }
       // do LLT decomposition using the Cholesky–Banachiewicz and Cholesky–Crout algorithms
        // p is the L_ii,
        bad = choldcRow(covar, ia, nTrace, nCoeff, nBand, wPoly, p);
        if (bad < 1) {
            for(j=0;j<ma;j++) {
                p[j] = 0.0;
                ans[j] = 0.0;
            }
        } else {
             cholslRow(covar, ia, nTrace, nCoeff, nBand, wPoly, p, beta, ans);
             if (calcCovar > 0) {
                 cholslRowCovar(covar, ia, nTrace, nCoeff, nBand, wPoly, p);
             }
        }
       // build ymodel
        // ymod is fitting result contains all profile and background (include mu and sigma)
        // amod only contains A and background
        for(i=0;i<nx;i++) bmod[i] = 0.0;
        // scattered light first
        for (j=nCoeff*nTrace,k=0;k<wPoly;j++,k++) {
            for(i=0; i < nx; i++) {
                bmod[i] += ans[j]*apoly[k][i];
            }
        }
        for(i=0;i<nx;i++) amod[i] = bmod[i];
        for(i=0;i<nx;i++) ymod[i] = bmod[i];
        // add fiber model
        for (j=0,l=0;j<nTrace;j++)
            for (coeff=0; coeff<nCoeff; coeff++,l++)
                for (i=xmin[j], k=0; i <= xmax[j]; i++, k++)
                    ymod[i] += ans[l]*aprofile[l][k];
        // build a mod for Jin's profile
        for (j=0; j<nTrace;j++){
            for (i=xmin[j], k=0; i <= xmax[j]; i++, k++) {
                for (l=0;l<nCoeff;l++) {
                    amod[i] += ans[j*nCoeff+l]*aprofile[j*nCoeff+l][k]; } } }
        if (debug==1) {
            fp=fopen("ymod.dat","wt");
            for(i=0; i<nx; i++)
                fprintf(fp, "%5f ", ymod[i]);
            fclose(fp);
            fp=fopen("amod.dat","wt");
            for(i=0; i<nx; i++)
                fprintf(fp, "%5f ", amod[i]);
            fclose(fp);
            fp=fopen("ans.dat","wt");
            for(i=0; i<ma; i++)
                fprintf(fp, "%5f ", ans[i]);
            fclose(fp);
        }
    }
    /* Free temporary memory */
    for(i=0,k=0; i<nTrace; i++)
        for(j=0; j<nCoeff; j++,k++)
            free(aprofile[k]);
    for(i=0; i<wPoly; i++) free(apoly[i]);
    free(aprofile);
    free(apoly);
    free(ysub);
    free(beta);
    free(xmin);
    free(xmax);
    free(xInd);
    free(ansFactor);
    return bad;
}
//#endregion: extract_row23

//#region: extract_row
long extract_row (
      //  xLen   xIndArr     yValue          invvar           xcen guess     sigma guess         
        long nx, double* x, double* fimage, double* invvar, double* xcen, double* sigma,
      //    ma = nTrace*nCoeff + nPoly, ia is wfixarr
      //    fiberNum   background    1,2 or 3
        long nTrace, long nPoly,  long nCoeff, double sigmal, long* ia,
        long proftype, long calcCovar, double* reject, long* fullreject, long* partialreject,
        double* ans, double* ymod, double** covar, double* p, double* bmod, double* amod,
        long maxHalfWidth, double* map2d, long debug) {
     // bmod is the background part, amod is the profile part
   // setup
    FILE *fp;
    long  bad;
    long  *xmin, *xmax;
    double ** aprofile;
    double ** apoly;
    double reject1, reject2;         /* profile area rejection thresholds */
    double rejectflux;               /* profile area rejection thresholds */
    double tsigma = 5.0;             /* set limits of profile influence   */
    double contribution = 0.0;       /* set limits of profile influence   */
    double x1, x2;                   /* Top and Lower Chebyshev x limit */
    long  i,j,k,l,length, mfit, coeff, tTrace, ma;
    long  nBand = 1;
    double * ysub;
    double * beta;
    double profileSum, temp, *map2dFirst;
    double *ansFactor, ansTotal, totalP;
    map2dFirst = map2d;
    long map2dCount=2; // add count for total length and total fib number
    /*  Reject 1 and 2 and flux are the three members of a double array  */
    reject2 = reject[0]; rejectflux = reject[1]; reject1 = reject[2];
    tTrace = nCoeff*nTrace;
    ma = tTrace + nPoly;
    /*covar  = (double **) malloc(ma * sizeof(double *));*/
    /*for (iy=0; iy < ma; iy++) covar[iy] = (double *) argv[argct]+iy*ma;*/
    xmin = (long *)malloc(sizeof(long)*nTrace);
    xmax = (long *)malloc(sizeof(long)*nTrace);
   // Find max and min of profile influence
    // have nothing to do with fimage
    //      tsigma is limits of profile influence       5 times sigma
    findXLimits(xmin, xmax, x, xcen, nx, nTrace, sigma, tsigma);
        if (debug==1) {
        fp=fopen("xminmax.dat","wt");
        for(i=0, k=0; i<nTrace; i++)
            fprintf(fp, "%5ld %5ld %5ld %5ld %5ld\n", i, i*nCoeff, xmin[i], xmax[i], xmax[i]-xmin[i]+1);
        fclose(fp);
        fp=fopen("fimage.dat","wt");
        for(i=0; i<nx; i++)
            fprintf(fp, "%5f ", fimage[i]);
        fclose(fp);
        fp=fopen("invvar.dat","wt");
        for(i=0; i<nx; i++)
            fprintf(fp, "%5f ", invvar[i]);
        fclose(fp);
        }
   // fill the aprofile and apoly
    // room for aprofile
    aprofile = (double **)malloc(tTrace * sizeof(double *));
    apoly = (double **)malloc(nPoly * sizeof(double *));
    for(i=0, k=0; i<nTrace; i++){//Room for fiber profiles
        length = xmax[i] - xmin[i] + 1;
        if (length < 0) length = 0;
        for(j=0; j<nCoeff; j++, k++) {
            aprofile[k] = (double *)malloc(length * sizeof(double));
        }
    }
    /*if (map2dLen>0 && map2dCount>map2dLen && (proftype==0||proftype==-3)) {
        printf("map2D not long enough (size: %ld) to hold all %ld data, \
                please increase the array length!\n", map2dLen, map2dCount);
        exit(1);
    }*/
    // aprofile[nTrace*nCoeff][length of each fit]
    fillProfile(aprofile, x, xcen, xmin, xmax, sigma, nx, nCoeff,
                nTrace, proftype, sigmal, contribution);
        if (debug==1) {
        fp=fopen("aprofile.dat","wt");
        for(i=0, k=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            if (length < 0) length = 0;
            for(j=0; j<nCoeff; j++, k++) {
                //printf("%5ld %5ld %5ld %5ld\n", k, xmin[i], xmax[i], length);
                fprintf(fp, "%5ld %5ld %5ld %5ld ", k, xmin[i], xmax[i], length);
                for(int x=0; x<length; x++) {
                    fprintf(fp, "%g ", aprofile[k][x]); } fprintf(fp, "\n"); }}
        fclose(fp);
        }

    // Room for polynomial profiles
    for(i=0; i<nPoly; i++) {
        apoly[i] = (double *)malloc(nx * sizeof(double));
        /*tTrace = nCoeff*nTrace;*/
        /*ia[i+tTrace] = ia[i+tTrace];*/
        /*ans[i+tTrace] = ans[i+tTrace];*/
    }
    // fill poly
    x2 = (double) (nx-1);
    x1 = 0.0;
    if (nPoly > 0) fillPoly(apoly, x, nx, nPoly, x1, x2);
    if (debug==1) {
        fp=fopen("apoly.dat","wt");
        for(i=0; i<nPoly; i++) {
            fprintf(fp, "%5ld ", i);
            for(int x=0; x<nx; x++) {
                fprintf(fp, "%g ", apoly[i][x]); } fprintf(fp, "\n"); }
        fclose(fp);
    }
   // checkRowFibers
    CheckRowFibers(aprofile, xmin, xmax, nTrace, nCoeff, ans, ia, invvar,
                    reject1, reject2, rejectflux, partialreject, fullreject);
   // Subtract out fixed variables first
    ysub = (double *)malloc(nx * sizeof(double));
    for(i=0, mfit=0; i<ma; i++) if(ia[i]) mfit++; // count for mfit
    for(i=0; i<nx; i++) ysub[i] = fimage[i]; // copy for fimage
    if(mfit != ma) { // subtract out fixed variables 
        //fprintf(stderr, "Subtracting fixed variables\n");
        subtractProfile(ysub, nx, xmin, xmax, nTrace, nCoeff, aprofile, ia, ans);
        subtractPoly(ysub, nx, nPoly, apoly, &ia[tTrace], &ans[tTrace]);
    } 
    beta = (double *) malloc(sizeof(double)*ma);
    for(i=0; i<ma; i++)  //init covar
       bzero((void *) covar[i] , sizeof(double)*ma);  
    for(i=0; i<ma; i++) beta[i] = 0.0;
   // fill covar
    // model is B = (X^T W X)^{-1} X^T W Y
    // covar should be ( X^T W X )
    // beta  should be ( X^T W Y )
    fillCovar(ysub, invvar, nx, aprofile, apoly, nTrace, nCoeff, nBand,
           nPoly, beta, ia, covar, xmin, xmax);
    if (debug==1) {
        fp=fopen("fillCovar.dat","wt");
        for (i=0;i<ma;i++){
            for (j=0;j<ma;j++){
                fprintf(fp, "%g ", covar[i][j]); } fprintf(fp, "\n"); }
        fclose(fp);
    }
   // do LLT decomposition using the Cholesky–Banachiewicz and Cholesky–Crout algorithms
    // p is the L_ii,
    bad = choldcRow(covar, ia, nTrace, nCoeff, nBand, nPoly, p);
    // if (mfit!=ma&&(proftype==0||proftype==-3)) bad -= 9999;
    if (bad < 1) {
        for(j=0;j<ma;j++) {
            p[j] = 0.0;
            ans[j] = 0.0;
        }
    } else {
         cholslRow(covar, ia, nTrace, nCoeff, nBand, nPoly, p, beta, ans);
         if (calcCovar > 0) {
             cholslRowCovar(covar, ia, nTrace, nCoeff, nBand, nPoly, p);
         }
    }
   // build ymodel
    // ymod is fitting result contains all profile and background (include mu and sigma)
    // amod only contains A and background
    for(i=0;i<nx;i++) ymod[i] = 0.0;
    // scattered light first
    for (j=nCoeff*nTrace,k=0;k<nPoly;j++,k++) {
        for(i=0; i < nx; i++) {
            ymod[i] += ans[j]*apoly[k][i];
            bmod[i] += ans[j]*apoly[k][i];
        }
    }
    for(i=0;i<nx;i++) amod[i] = ymod[i];
    // add fiber model
    ansFactor = (double *)malloc(sizeof(double)*nTrace);
    for (j=0,l=0;j<nTrace;j++)
        for (coeff=0; coeff<nCoeff; coeff++,l++)
            for (i=xmin[j], k=0; i <= xmax[j]; i++, k++)
                ymod[i] += ans[l]*aprofile[l][k];
    // build amod for Jin's profile
    if (proftype==0||proftype==-3) {
        for (j=0; j<nTrace;j++)
            for (i=xmin[j], k=0; i <= xmax[j]; i++, k++)
                amod[i] += ans[j*nCoeff]*aprofile[j*nCoeff][k];
    } else {
        if (proftype<-10) {
            for (j=0; j<nTrace;j++){
                for (i=xmin[j], k=0; i <= xmax[j]; i++, k++) {
                    for (l=0;l<nCoeff;l++) {
                        amod[i] += ans[j*nCoeff+l]*aprofile[j*nCoeff+l][k];
                    }
                }
            }
            for (j=0; j<nTrace;j++){
                ansTotal = 0;
                totalP = 0;
                for (l=0;l<nCoeff;l++) {
                    ansFactor[l] = ans[j*nCoeff+l];
                    ansTotal += ans[j*nCoeff+l];
                    //ans[j*nCoeff+l] = 0;
                    totalP += 1.0/(p[j*nCoeff+l]*p[j*nCoeff+l]); // sum(sigma^2)
                    //p[j*nCoeff+l] = 0;
                }
                for (l=0;l<nCoeff;l++) {
                    ansFactor[l] /= ansTotal;
                    ans[j*nCoeff+l] = ansFactor[l];
                }
                ans[j*nCoeff] = ansTotal;
                //totalP = 1.0/sqrt(totalP); // ivar = 1/sum(sigma^2)
                //p[j*nCoeff] = totalP;

                // modify the first profile
                for (i=xmin[j], k=0; i <= xmax[j]; i++, k++) {
                    aprofile[j*nCoeff][k]      = ansFactor[0]*aprofile[j*nCoeff][k];
                }
                // add later profile
                for (l=1;l<nCoeff;l++) {
                    for (i=xmin[j], k=0; i <= xmax[j]; i++, k++) {
                        aprofile[j*nCoeff][k] += ansFactor[l]*aprofile[j*nCoeff+l][k];
                    }
                }
            }
        }
    }
    if (maxHalfWidth>0) {
        *(map2d++) = map2dCount;// right map2dCount num in the end of the if
        *(map2d++) = nTrace;
        for (i=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            if (length < 0) length = 0;
            map2dCount += length+3; // count for profile data
            *(map2d++) = length;
            *(map2d++) = (double)xmin[i];
            *(map2d++) = (double)xmax[i];
            for(k=xmin[i], j=0; k<=xmax[i]; k++, j++)
                *(map2d++) = aprofile[i*nCoeff][j];// only get map for the first conponent
        }
        map2dFirst[0] = map2dCount;
        //printf("map2dCount: %ld\n", map2dCount);
    }
    // maxHalfWidth is a redundent
    /*if (maxHalfWidth>0) {
        find2DLimits(xInd, xmin, xmax, x, xcen, sigma, nx, nTrace, maxHalfWidth);
        *(map2d++) = map2dCount;
        *(map2d++) = nTrace;
        for (i=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            if (length < 0) length = 0;
            map2dCount += length+3; // count for profile data
            *(map2d++) = length;
            *(map2d++) = (double)xmin[i];
            *(map2d++) = (double)xmax[i];
            for(k=xmin[i]; k<=xmax[i]; k++)
                *(map2d++) = fimage[k]-bmod[k];
            map2d -= length;
            profileSum=0;
            for(k=xmin[i]; k<=xmax[i]; k++)
                profileSum += *(map2d++);
            map2d -= length;
            for(k=xmin[i]; k<=xmax[i]; k++, map2d++) {
                *(map2d) /= profileSum;
            }
        }
        map2dFirst[0] = map2dCount;
        //printf("map2dCount: %ld\n", map2dCount);
    }*/
    if (debug==1) {
        fp=fopen("ymod.dat","wt");
        for(i=0; i<nx; i++)
            fprintf(fp, "%5f ", ymod[i]);
        fclose(fp);
        fp=fopen("amod.dat","wt");
        for(i=0; i<nx; i++)
            fprintf(fp, "%5f ", amod[i]);
        fclose(fp);
        fp=fopen("ans.dat","wt");
        for(i=0; i<ma; i++)
            fprintf(fp, "%5f ", ans[i]);
        fclose(fp);
    }

    /* Free temporary memory */
    for(i=0,k=0; i<nTrace; i++)
        for(j=0; j<nCoeff; j++,k++)
            free(aprofile[k]);
    for(i=0; i<nPoly; i++) free(apoly[i]);
    free(aprofile);
    free(apoly);
    free(ysub);
    free(beta);
    free(xmin);
    free(xmax);
    free(ansFactor);
    return bad;
}
//#endregion: extract_row

//#endregion: bad codes....

//#region: all kinds of profile
void AddGaussWings(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, 
        long nCoeff, double contribution) { 
    long i,k;
    double base;
    double diff, denom;
    double epow;
    /*   !!!!!!!!!! Hardwired sigma for wide gaussian 10 pixels!!!!!!! */
    double sigmal = 10.0;  
    denom = 1.0/sqrt(6.2832 * sigmal * sigmal);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        if(i >= 0 && i < ndat && nCoeff > 0) {
            diff = (xcen - x[i])/sigmal;
            epow = 0.5*diff*diff;

            if (epow < MAXEXP) base = exp(-epow) * denom;
            else base = 0.0;

            y[0][k] += base * contribution;
            y[0][k] /= (1.0 + contribution);
        }
    }
}
void AddExponentialWings(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigmal, 
        double contribution) { 
    long i,k;
    double diff;

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        if(i >= 0 && i < ndat) {
            diff = fabs(xcen - x[i])/sigmal;
            y[0][k] += contribution * 0.5*exp(-diff)/sigmal;

        }
    }
}
void ProfileGauss(double *x, long ndat, double **y, double xcen, long xmin,
        long xmax, double sigma, long nCoeff) { 
    long i,j,k;
    double base;
    double diff, denom, frac;
    double epow;

    denom = 1.0/sqrt(6.2832 * sigma * sigma);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (xcen - x[i] + frac)/sigma;
                epow = 0.5*diff*diff;

                if (epow < MAXEXP) base = exp(-epow) * denom;
                else base = 0.0;

                y[0][k] += base;
                if(nCoeff >1) y[1][k] += base*epow*2.0;
                if(nCoeff >2) y[2][k] += base*diff;

            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
/*void ProfileGaussJin(double *x, long ndat, double **y, double xcen, long xmin,
        long xmax, double sigma, long nCoeff) { 
    long i,j,k;
    double base;
    double diff, denom, frac;
    double epow;

    denom = 1.0/sqrt(6.283185307 * sigma * sigma);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            diff = (x[i] - xcen )/sigma;
            epow = 0.5*diff*diff;
            if (epow < MAXEXP) base = exp(-epow) * denom;
            else base = 0.0;
            // y[x] = A/sqrt(2pi * sigma)*Exp(-(x-mu)^2/(2*sigma^2))
            // ans[0] = A
            y[0][k] = base;
            // ans[1] = A * dSigma
            if(nCoeff >1) y[1][k] = base * (epow*2.0-1)/sigma;
            // ans[2] = A * dMu
            if(nCoeff >2) y[2][k] = base * diff / sigma;
            for (j=0;j<nCoeff;j++)
                if (fabs(y[j][k])>1e7)
                    printf("stop here!\n");
        }
    }
}*/
//backup version
void ProfileGaussJin(double *x, long ndat, double **y, double xcen, long xmin,
                     long xmax, double sigma, long nCoeff) {
    long i,j,k;
    double base;
    double diff, denom, frac;
    double epow;

    //denom = 1.0/sqrt(6.283185307 * sigma * sigma);
    denom = 1.0/(sqrt(6.283185307)*sigma);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (x[i] - xcen + frac)/sigma;
                epow = 0.5*diff*diff;

                if (epow < MAXEXP) base = exp(-epow) * denom;
                else base = 0.0;

                // y[x] = A/sqrt(2pi * sigma)*Exp(-(x-mu)^2/(2*sigma^2))
                // ans[0] = A
                y[0][k] += base;
                // ans[1] = A * dSigma
                if(nCoeff >1) y[1][k] += base * (epow*2.0-1)/sigma;
                // ans[2] = A * dMu
                if(nCoeff >2) y[2][k] += base * diff / sigma;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
            for (j=0;j<nCoeff;j++)
                if (fabs(y[j][k])>1e7)
                    printf("stop here!\n");
        }
    }
}
void ProfileAbs23Jin(double *x, long ndat, double **y, double xcen,
                     long xmin, long xmax, double sigma, long nCoeff) {
    long i,j,k;
    double base2, base3;
    double diff, diffabs, denom3, frac;
    double epow, epow2, epow3, denom2;
    denom3 = 1.0/(2.575798634 * sigma);
    denom2 = 1.0/(sqrt(6.283185307)*sigma);
    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (x[i] - xcen + frac)/sigma;
                diffabs = fabs(diff);
                epow = diff*diff;
                epow2= epow*0.5;
                epow3 = epow*diffabs/3.0;
                if (epow3 < MAXEXP) base3 = exp(-epow3) * denom3;
                else base3 = 0.0;
                if (epow2 < MAXEXP) base2 = exp(-epow2) * denom2;
                else base2 = 0.0;
                y[0][k] += base2;
                y[1][k] += base3;
                /*y[2][k] += base2 * (epow*2.0 - 1)/sigma;*/
                /*y[3][k] += base2 * diff/sigma;*/
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void ProfileAbs232Jin(double *x, long ndat, double **y, double xcen,
                     long xmin, long xmax, double sigma, long nCoeff) {
    long i,j,k;
    double base2, base3;
    double diff, diffabs, denom3, frac;
    double epow, epow2, epow3, denom2;
    /*denom3 = 1.0/(2.575798634 * sigma);*/
    denom2 = 1.0/(sqrt(6.283185307)*sigma);
    for (i=xmin,k=0; i<=xmax; i++, k++) {
        y[0][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (x[i] - xcen + frac)/sigma;
                //diffabs = fabs(diff);
                epow = diff*diff;
                epow2= epow*0.5;
                /*epow3 = epow*diffabs/3.0;*/
                /*if (epow3 < MAXEXP) base3 = exp(-epow3) * denom3;*/
                /*else base3 = 0.0;*/
                if (epow2 < MAXEXP) base2 = exp(-epow2) * denom2;
                else base2 = 0.0;
                y[0][k] += base2;
                /*y[1][k] += base3;*/
                /*y[2][k] += base2 * (epow*2.0 - 1)/sigma;*/
                /*y[3][k] += base2 * diff/sigma;*/
            }
            y[0][k] /= 5.0;
        }
    }
}
void ProfileAbs233Jin(double *x, long ndat, double **y, double xcen,
                     long xmin, long xmax, double sigma, long nCoeff) {
    long i,j,k;
    double base2, base3;
    double diff, diffabs, denom3, frac;
    double epow, epow2, epow3, denom2;
    denom3 = 1.0/(2.575798634 * sigma);
    /*denom2 = 1.0/(sqrt(6.283185307)*sigma);*/
    for (i=xmin,k=0; i<=xmax; i++, k++) {
        y[1][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (x[i] - xcen + frac)/sigma;
                diffabs = fabs(diff);
                epow = diff*diff;
                /*epow2= epow*0.5;*/
                epow3 = epow*diffabs/3.0;
                if (epow3 < MAXEXP) base3 = exp(-epow3) * denom3;
                else base3 = 0.0;
                /*if (epow2 < MAXEXP) base2 = exp(-epow2) * denom2;*/
                /*else base2 = 0.0;*/
                /*y[0][k] += base2;*/
                y[1][k] += base3;
                /*y[2][k] += base2 * (epow*2.0 - 1)/sigma;*/
                /*y[3][k] += base2 * diff/sigma;*/
            }
            y[1][k] /= 5.0;
        }
    }
}
    //void ProfileAbs32Jin(double *x, long ndat, double **y, double xcen,
    //                    long xmin, long xmax, double sigma, long nCoeff) {
    //    long i,j,k;
    //    double base2, base3;
    //    double diff, diffabs, denom3, frac;
    //    double epow, epow2, epow3, denom2;
    //    
    //    denom3 = 1.0/(2.575798634 * sigma);
    //    denom2 = 1.0/(sqrt(6.283185307)*sigma);
    //    
    //    for (i=xmin,k=0; i<=xmax; i++, k++) {
    //        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
    //        if(i >= 0 && i < ndat && nCoeff > 0) {
    //            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
    //                diff = (x[i] - xcen + frac)/sigma;
    //                diffabs = fabs(diff);
    //                epow = diff*diff;
    //                epow2= epow*0.5;
    //                epow3 = epow*diffabs/3.0;
    //                
    //                if (epow3 < MAXEXP) base3 = exp(-epow3) * denom3;
    //                else base3 = 0.0;
    //                if (epow2 < MAXEXP) base2 = exp(-epow2) * denom2;
    //                else base2 = 0.0;
    //                
    //                y[0][k] += base3;
    //                y[1][k] += base2;
    //                y[2][k] += base3 * (epow*3.0 - 1)/sigma;
    //                y[3][k] += base3 * diff*diffabs/sigma;
    //                
    //            }
    //            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
    //        }
    //    }
    //}
void ProfileAbs3Jin(double *x, long ndat, double **y, double xcen,
                    long xmin, long xmax, double sigma, long nCoeff) {
    long i,j,k;
    double base;
    double diff, diffabs, denom, frac;
    double epow;
    /*        Below is denominator for x^3    */
    denom = 1.0/(2.575798634 * sigma);
    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (x[i] - xcen + frac)/sigma;
                diffabs = fabs(diff);
                epow = diff*diff*diffabs/3.0;
                if (epow < MAXEXP) base = exp(-epow) * denom;
                else base = 0.0;
                y[0][k] += base;
                if(nCoeff >1) y[1][k] += base * (epow*3.0 - 1)/sigma;
                if(nCoeff >2) y[2][k] += base * diff*diffabs/sigma;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void ProfileGaussStatic(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff) { 
    long nm,i,j,k,backup,place;
    double base;
    double diff, denom, frac;
    double sqbase, xfake;
    double epow;

    static double oldsigma=0.0;
    static double model[NSUBSAMP][3][31];

    denom = 1.0/sqrt(6.2832 * sigma * sigma);

    if (sigma != oldsigma) {

        printf("Filling Arrays\n");
        oldsigma = sigma;
        /*       Fill static arrays    */
        for(nm=0; nm<NSUBSAMP; nm++) {
            xfake = 15.0 + ((double)nm) / NSUBSAMP;
            for (k=0; k<=30; k++) {
                for(j=0;j<3;j++) model[nm][j][k] = 0.0;
                for(frac = -0.48; frac <= 0.5; frac += 0.04)  {

                    diff = (xfake - (double)k + frac)/sigma;
                    epow = diff*diff/2.0;
                    if (epow < MAXEXP) base = exp(-epow) * denom;
                    else base = 0.0;

                    model[nm][0][k] += base;
                    sqbase = diff*diff*base;
                    model[nm][1][k] += sqbase;
                    model[nm][2][k] += diff*base;

                }
                for (j=0;j<3;j++) model[nm][j][k] /= 25.0; /* What is this ??? */
            }
        }
    }

    if (xcen > 0.0) {
        frac = xcen - (int)xcen;
        backup = (int)xcen - xmin;
    } else {
        frac = (int)xcen - xcen;
        backup = (int)xcen - xmin - 1;
    }

    nm = frac * NSUBSAMP;
    for(j=0;j<nCoeff;j++) 
        for (i=xmin,k=0,place=15-backup; i<=xmax; i++,k++,place++) 
        {
            y[j][k] = model[nm][j][place];
            if (nm >= 2000) printf("VERY BAD!!!\n");
            if (place >= 31) printf("VERY BAD!!!\n");
            if (j >= 3) printf("VERY BAD!!!\n");
        }
}
void ProfileAbs3(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff) {
    long i,j,k;
    double base;
    double diff, diffabs, denom, frac;
    double epow;

    /*        Below is denominator for x^3    */

    denom = 1.0/(2.88450 * 0.89298 * sigma);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (xcen - x[i] + frac)/sigma;
                diffabs = fabs(diff);
                epow = diff*diff*diffabs/3.0;

                if (epow < MAXEXP) base = exp(-epow) * denom;
                else base = 0.0;

                y[0][k] += base;
                if(nCoeff >1) y[1][k] += base*epow*3.0;
                if(nCoeff >2) y[2][k] += base*diff*diffabs;

            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void ProfileAbs25(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff) { 
    long i,j,k;
    double base, base3;
    double diff, diffabs, denom2, frac;
    double denom3;
    double epow, epow3;

    /*        Below is denominator for x^3    */

    denom3 = 1.0/(2.88450 * 0.89298 * sigma);
    denom2 = 1.0/sqrt(6.2832 * sigma * sigma);

    if (nCoeff == 0 || nCoeff > 3) return;

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;

        if(i >= 0 && i < ndat) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (xcen - x[i] + frac)/sigma;
                diffabs = fabs(diff);
                epow = 0.5*diff*diff;
                if (epow < MAXEXP) base = 0.3* exp(-epow) * denom2;
                else base = 0.0;

                epow3 = diff*diff*diffabs/3.0;
                if (epow3 < MAXEXP) base3 = 1.7 *exp(-epow3) * denom3;
                else base3 = 0.0;

                y[0][k] += base + base3;
                if(nCoeff >1) y[1][k] += 2.0*epow*base + 3.0*epow3*base3;
                if(nCoeff >2) y[2][k] += diff*base + diff*diffabs*base3;
            }
            /*    Divide by 10 instead of 5 to get area 1.0  */
            for (j=0;j<nCoeff;j++) y[j][k] /= 10.0;
        }
    }
}
void ProfileDoubleGauss(double *x, long ndat, double **y, double xcen, long xmin,
        long xmax, double sigma, long nCoeff) { 
    long i,j,k;
    double base;
    double diff, denom, frac;
    double sigma2, diff2, base2, denom2;
    double epow, epow2;

    sigma2 = 2.0*sigma;
    denom = 1.0/sqrt(6.2832 * sigma * sigma);
    denom2 = 1.0/sqrt(6.2832 * sigma2 * sigma2);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (xcen - x[i] + frac)/sigma;
                diff2 = (xcen - x[i] + frac)/sigma2;
                epow = 0.5*diff*diff;
                if (epow < MAXEXP) base = 0.9 * exp(-epow) * denom;
                else base = 0.0;
                epow2 = 0.5*diff2*diff2;
                if (epow < MAXEXP) base2 = 0.1 * exp(-epow2) * denom2;
                else base2 = 0.0;

                y[0][k] += base + base2;
                if(nCoeff >1) y[1][k] += 2.0*(epow*base + epow2*base2);
                if(nCoeff >2) y[2][k] += diff*base + diff2*base2;
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void ProfileAbs3WideGauss(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff) { 
    long nm,i,j,k,backup,place;
    double base, xfake;
    double diff, diffabs, denom, frac;
    double sigma2, diff2, base2, denom2;
    double sqbase;
    double epow;

    static double oldsigma=0.0;
    static double model[NSUBSAMP][4][31];

    /*        Below is denominator for x^3    */

    denom = 1.0/(2.88450 * 0.89298 * sigma);
    sigma2 = 2.0*sigma;
    denom2 = 1.0/sqrt(6.2832 * sigma2 * sigma2);

    if (sigma != oldsigma) {

        printf("Filling Arrays\n");
        oldsigma = sigma;
        /*       Fill static arrays    */
        for(nm=0; nm<NSUBSAMP; nm++) {
            xfake = 15.0 + ((double)nm) / NSUBSAMP;
            for (k=0; k<=30; k++) {
                for(j=0;j<4;j++)model[nm][j][k] = 0.0;
                for(frac = -0.4; frac <= 0.5; frac += 0.2)  {

                    diff = (xfake - (double)k + frac)/sigma;
                    diffabs = fabs(diff);
                    epow = diff*diff*diffabs/3.0;
                    if (epow < MAXEXP) base = exp(-epow) * denom;
                    else base = 0.0;

                    diff2 = (xfake - (double)k + frac)/sigma2;
                    epow = diff2*diff2/2.0;
                    if (epow < MAXEXP) base2 = exp(-epow) * denom2;
                    else base2 = 0.0;

                    model[nm][0][k] += base;
                    model[nm][1][k] += base2;
                    sqbase = diff*diff*diffabs*base;
                    model[nm][2][k] += sqbase;
                    model[nm][3][k] += diff*diffabs*base;

                }
                for (j=0;j<4;j++) model[nm][j][k] /= 5.0;
            }
        }
    }

    if (xcen > 0.0) {
        frac = xcen - (int)xcen;
        backup = (int)xcen - xmin;
    } else {
        frac = (int)xcen - xcen;
        backup = (int)xcen - xmin - 1;
    }

    nm = frac * NSUBSAMP;
    for(j=0;j<nCoeff;j++) 
        for (i=xmin,k=0,place=15-backup; i<=xmax; i++,k++,place++) 
        {
            y[j][k] = model[nm][j][place];
        }
}
void ProfileAbs3HalfLorentz(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff) { 
    long i,j,k;
    double base, mult;
    double diff, diffabs, denom, frac;
    double fwhm2, diff2, base2, denom2;
    double sqbase;
    double epow;
    /*        double total;     */


    /*        Below is denominator for x^3    */

    denom = 1.0/(2.88450 * 0.89298 * sigma);
    fwhm2 = 2.0*sigma;
    denom2 = 3.1416/(2.0 * fwhm2);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (xcen - x[i] + frac)/sigma;
                diffabs = fabs(diff);
                epow = diff*diff*diffabs/3.0;
                if (epow < MAXEXP) base = exp(-epow) * denom;
                else base = 0.0;
                diff2 = 2.0*(xcen - x[i] + frac)/fwhm2;
                base2 = denom2/(diff2*diff2 + 1.0);

                y[0][k] += base;
                if(nCoeff >1 && diff2 < 0.0) y[1][k] += base2;
                if(nCoeff >2 && diff2 >= 0.0) y[2][k] += base2;
                if(nCoeff >3) {
                    sqbase = diff*diff*diffabs*base;
                    y[3][k] += sqbase;

                    if(nCoeff >4) y[4][k] += diff*base;

                    for (j=5,mult=diff;j<nCoeff;j++,mult *= diff)
                        y[j][k] += mult*sqbase;
                }
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
}
void ProfileGaussHalfLorentz(double *x, long ndat, double **y, double xcen,
        long xmin, long xmax, double sigma, long nCoeff) { 
    long i,j,k;
    double base, mult;
    double diff, denom, frac;
    double fwhm2, diff2, base2, denom2;
    double sqbase;
    double epow;
    /*        double total;    */
    /*        Below is denominator fro x^3    */

    denom = 1.0/sqrt(6.2832 * sigma * sigma);
    fwhm2 = 2.0*sigma;
    denom2 = 5.0/(fwhm2 * 3.1416);

    for (i=xmin,k=0; i<=xmax; i++, k++) {
        for (j=0;j<nCoeff;j++) y[j][k] = 0.0;
        if(i >= 0 && i < ndat && nCoeff > 0) {
            for(frac = -0.4; frac <= 0.5; frac += 0.2)  {
                diff = (xcen - x[i] + frac)/sigma;
                epow = diff*diff/2.0;
                if (epow < MAXEXP) base = exp(-epow) * denom;
                else base = 0.0;
                diff2 = 2.0*(xcen - x[i] + frac)/fwhm2;
                base2 = denom2/(diff2*diff2 + 1.0);
                y[0][k] += base;
                if(nCoeff >1 && diff2 < 0.0) y[1][k] += base2;
                if(nCoeff >2 && diff2 >= 0.0) y[2][k] += base2;
                if(nCoeff >3) {
                    sqbase = diff*diff*base;
                    y[3][k] += sqbase;
                    if(nCoeff >4) y[4][k] += diff*base;
                    for (j=5,mult=diff;j<nCoeff;j++,mult *= diff)
                        y[j][k] += mult*sqbase;
                }
            }
            for (j=0;j<nCoeff;j++) y[j][k] /= 5.0;
        }
    }
    /*
    //    This is just for debugging profiles
    //
    //    for (j=0;j<nCoeff;j++) {
    //      for (i=xmin,k=0,total=0.0; i<=xmax; i++, k++)
    //        if(i >= 0 && i < ndat && nCoeff > 0) {
    //          total += y[j][k];
    //        }
    //         fprintf(stderr,"%d %g ",(int)j,total);
    //    }
    //         fprintf(stderr,"\n");    */
}
//#endregion: all kinds of profile

//#region: key functions
void findXLimits(long *xmin, long *xmax, double *x, double *xcen,
        long nx, long nTrace, double *sigma, double sigmal) {
    /*
    //        Find max and min of profile influence
    //    */
    long i, place, place2;
    double diff;

    place = 0;

    for(i=0; i< nTrace; i++) {
        diff = sigmal*sigma[i];
        while(x[place] < xcen[i] - diff && place < nx) place++;
        if(place >= nx) place--;

        place2 = place;
        while(x[place2] <= xcen[i] + diff && place2 < nx) place2++;
        if(place2 >= nx) place2--;

        xmin[i] = place;
        xmax[i] = place2 - 1;

        /*printf("i:%5ld, xmin:%5ld, xmax:%5ld\n", i, xmin[i], xmax[i]);*/
        if(xmax[i] == xmin[i])  xmax[i]--;
        if(xmax[i] <= xmin[i]) {
            fprintf(stderr," Fiber # %d has no influence %g, xcen: %g, nx %ld\n", (int) i, diff, xcen[i], nx);
            exit(1);
        }
           /*fprintf(stderr," %d %g   %d %d \n", (int) i, xcen[i], xmin[i], xmax[i]);*/

        place = xmin[i];
    }
}
void find2DLimits(long *xInd, long *xmin, long *xmax, double *x, double *xcen, double *sigma, long nx, long nTrace, long maxHalfWidth) {
    long i,j,k;
    long place;
    double diff, cutMin, cutMax;
    place = 0;
    for (i=0; i<nTrace; i++) {
        while(x[place]<xcen[i]) place++;
        if (round(place-xcen[i])>0.5) place--;
        xInd[i] = place;
    }
    xmin[0] = xInd[0]-maxHalfWidth>0?xInd[0]-maxHalfWidth:0;
    diff = (xInd[1] - xInd[0]);
    cutMin = (long)round(diff/(sigma[0]+sigma[1])*sigma[0]);
    cutMin = cutMin>maxHalfWidth?maxHalfWidth:cutMin;
    cutMax = (long)(diff-cutMin);
    cutMax = cutMax>maxHalfWidth?maxHalfWidth:cutMax;
    xmax[0] = xInd[0]+cutMin;
    xmin[1] = xInd[1]-cutMax;
    for (i=1; i<nTrace-1; i++) {
        diff = (xInd[i+1] - xInd[i]);
        cutMin = (long)round(diff/(sigma[i]+sigma[i+1])*sigma[i]);
        cutMin = cutMin>maxHalfWidth?maxHalfWidth:cutMin;
        cutMax = (long)(diff-cutMin);
        cutMax = cutMax>maxHalfWidth?maxHalfWidth:cutMax;
        xmax[i] = xInd[i]+cutMin;
        xmin[i+1] = xInd[i+1]-cutMax;
    }
    xmax[nTrace-1] = xInd[nTrace-1]+maxHalfWidth>nx-1?nx-1:xInd[nTrace-1]+maxHalfWidth;
}
void subtractProfile(double *y, long nx, long *xmin, long *xmax, long nTrace, long nCoeff,
        double **aprofile, long *ia, double *a) {
    long i,j,k,l,coeff;

    for (i=0,l=0;i<nTrace;i++) 
        for (coeff=0; coeff<nCoeff; coeff++,l++) 
            if (!ia[l])  
                for (j=xmin[i],k=0;j<=xmax[i];j++,k++)  
                    y[j] -= a[l] * aprofile[l][k];
}

void subtractPoly(double *y, long nx, long nPoly, double **apoly, long *ia, double *a) {
    long i,j;

    for (i=0;i<nPoly;i++) 
        if (!ia[i]) 
            for (j=0;j<nx;j++) 
                y[i] -= a[i] * apoly[i][j];
}

//   CheckRowFibers(aprofile, xmin, xmax, nTrace, nCoeff, ans, ia, invvar,
//                 reject1, reject2, rejectflux, partialreject, fullreject);
void CheckRowFibers(double **abig, long *xmin, long *xmax,
        long nTrace, long nCoeff, double *a, long *ia, double *invvar,
        double reject1, double reject2, double rejectflux, long *partial, long *full) {
    long i,j,k,l,m;
    double total;
    for (i=0;i<nTrace;i++) {
        total = 0.0;
        // coeff 0 total flux, base is a whole fixed profile
        for (k=xmin[i],m=0; k<=xmax[i]; k++,m++)
            if (invvar[k] > 0.0) total += abig[i*nCoeff][m];
        // information fraction
        if (total < reject1) {
            partial[i] = 1;
            for(j=nCoeff-1,l=j+i*nCoeff;j>=1;j--,l--) 
                if (ia[l]) {
                    ia[l] = 0;
                    a[l] = 0.0;
                    /*fprintf(stderr,"Fiber %d, dropped term %d, total: %g %g %d %d\n,
                      (int) i, (int) j,total, reject1, 
                      (int) xmin[i], (int) xmax[i]);   */
                }
        }
        if (total < rejectflux) full[i] = 1;
        if (total < reject2) {
            j = 0;
            l=i*nCoeff;
            if (ia[l]) {
                ia[l] = 0;
                a[l] = 0.0;
                /*fprintf(stderr,"Fiber %d, dropped term %d, total: %g %g\n",
                  (int) i, (int) j,total, reject2);   */
            }
        }
    }   
}
void CheckMultiRowFibers(double **abig, long *xmin, long *xmax,
        long nTrace, long nCoeff, double *a, long *ia, double *invvar, 
        long nsingle, long multirow) {
    long i,j,k,l,m;
    long row;
    double total;

    for (i=0;i<nTrace;i++) {
        total = 0.0;
        for (row = 0; row < multirow; row++)
            for (k=xmin[i+row*nTrace],m=0; k<=xmax[i+row*nTrace]; k++,m++)
                if (invvar[k+row*nsingle] > 0.0) total += abig[(i+row*nTrace)*nCoeff][m];
        if (total < 0.8) {
            for(j=nCoeff-1,l=j+i*nCoeff;j>=1;j--,l--) 
                if (ia[l]) {
                    ia[l] = 0;
                    a[l] = 0.0;
                    /*          fprintf(stderr,"Fiber %d, dropped term %d, total: %g\n",
                                (int) i, (int) j,total);   */
                }
        }
        if (total < 0.4) {
            j = 0;
            l=i*nCoeff;
            if (ia[l]) {
                ia[l] = 0;
                a[l] = 0.0;
                /*          fprintf(stderr,"Fiber %d, dropped term %d, total: %g\n",
                            (int) i, (int) j,total);   */
            }
        }
    }
}


void fillProfile(double **y, double *x, double *xcen, long *xmin,
        long *xmax, double *sigma, long nx, long nCoeff,
        long nTrace, long proftype, double sigmal,
        double contribution) {
    int i, j, m, length;
    long tCoeff;
    for (i=0,j=0; i<nTrace; i++,j+=nCoeff) {
        length = xmax[i] - xmin[i] + 1;
        /*printf("i: %d, j: %d, length: %d, sigma: %g, proftype: %ld\n", i, j, length, sigma[i], proftype);*/
        if (length > 0 && sigma[i] > 0.0) {
            if (proftype == -232) {
                ProfileAbs232Jin(x, nx, &y[j], xcen[i], xmin[i], xmax[i],
                               sigma[i], nCoeff);
            } else if (proftype == -233) {
                ProfileAbs233Jin(x, nx, &y[j], xcen[i], xmin[i], xmax[i],
                               sigma[i], nCoeff);
            } else if (proftype == -23) {
                ProfileAbs23Jin(x, nx, &y[j], xcen[i], xmin[i], xmax[i],
                               sigma[i], nCoeff);
            }
            else if (proftype == -3) {
                ProfileAbs3Jin(x, nx, &y[j], xcen[i], xmin[i], xmax[i],
                        sigma[i], nCoeff);
            }
            else if (proftype == 0) {
                /*printf("Using Jin's Profile at %g with xmin %ld and xmax %ld\n", xcen[i], xmin[i], xmax[i]);*/
                ProfileGaussJin(x, nx, &y[j], xcen[i], xmin[i], xmax[i],
                        sigma[i], nCoeff);
            }
            else if (proftype == 1) {
                ProfileGauss(x, nx, &y[j], xcen[i], xmin[i], xmax[i], 
                        sigma[i], nCoeff);
            }
            else if (proftype == 2)  {
                ProfileAbs3(x, nx, &y[j], xcen[i], xmin[i], xmax[i], 
                        sigma[i], nCoeff);
            }
            else if (proftype == 3)  {
                ProfileAbs25(x, nx, &y[j], xcen[i], xmin[i],xmax[i],
                        sigma[i], nCoeff);
            }
            else if (proftype == 4)  {
                ProfileDoubleGauss(x, nx, &y[j], xcen[i], xmin[i],xmax[i],
                        sigma[i], nCoeff);
            }
            else if (proftype == 5)  {
                ProfileAbs3WideGauss(x, nx, &y[j], xcen[i], xmin[i],xmax[i],
                        sigma[i], nCoeff);
            }
            else if (proftype == 6)  {
                ProfileAbs3HalfLorentz(x, nx, &y[j], xcen[i], xmin[i],xmax[i],
                        sigma[i], nCoeff);
            }
            else if (proftype == 7)  {
                ProfileGaussHalfLorentz(x, nx, &y[j], xcen[i], xmin[i],xmax[i],
                        sigma[i], nCoeff);
            }
            else if (proftype == 10)  {
                ProfileGauss(x, nx, &y[j], xcen[i], xmin[i],xmax[i],
                        sigma[i], nCoeff);
                AddGaussWings(x, nx, &y[j], xcen[i], xmin[i],xmax[i],
                        sigma[i], nCoeff, 0.025);
            }
            else if (proftype == 11)  {
                tCoeff = 1;
                ProfileGauss(x, nx, &y[j], xcen[i], xmin[i],xmax[i],
                        sigma[i], tCoeff);
                if (nCoeff > 1)  {
                    for (m=0; m<length; m++) { y[j+1][m] = 0.0; }
                    AddGaussWings(x, nx, &y[j+1], xcen[i], xmin[i],xmax[i],
                            sigma[i], nCoeff, 1.0);
                }
            }
            /*      This is to put a sigma = 25.0 exponential profile    */
            else if (proftype == 12)  {
                ProfileGauss(x, nx, &y[j], xcen[i], xmin[i],xmax[i],
                        sigma[i], nCoeff);
                AddExponentialWings(x, nx, &y[j], xcen[i], xmin[i],
                        xmax[i], sigmal, contribution);
            }
            else {
                /*fprintf(stderr,"Using Gaussian, proftype: %ld\n", proftype);*/
                /*ProfileGauss(x, nx, &y[j], xcen[i], xmin[i], xmax[i], */
                        /*sigma[i], nCoeff);*/
                fprintf(stderr,"unknown proftype, exit\n %ld\n", proftype);
                exit(1);
            }
        }
    }
}
void fillPoly(double **y, double *x, long nx, long nPoly, double x1, double x2) {
    int i,j;
    double norm;
    double *atemp = (double *)malloc(nPoly * sizeof(double));   
    for (i=0;i<nx;i++) {
        norm = (2.0 * x[i] - (x1 + x2))/(x2 - x1);
        /*    These next two lines make a step function to account for
              uncertainties in the amplifier's gains matching 
              if (norm < 0.0) y[0][i] = 0.0; 
              else y[0][i] = 1.0; 
              The rest of the params are just nPoly-1 chebyshev coefficients 
              chebyshevRow(norm, atemp, nPoly-1);
              for(j=1; j<nPoly; j++) {
              y[j][i] = atemp[j-1];
              }   
              I don't think the step function is helping, it seems
              to be giving wrong answers when traces fall off one side */
        chebyshevRow(norm, atemp, nPoly);
        for(j=0; j<nPoly; j++) {
            y[j][i] = atemp[j];
            if (fabs(atemp[j])>1e7)
                printf("stop here!\n");
        }
    }
    free(atemp);
}
// model is B = (X^T W X)^{-1} X^T W Y
// covar should be ( X^T W X )
// beta  should be ( X^T W Y )
// X     is profile column vector
// X^T   is profile row    vector
// nBand is the range of influence of each fiber
//   fillCovar(ysub, invvar, nx, aprofile, apoly, nTrace, nCoeff, nBand,
//             wPoly, beta, ia, covar, xmin, xmax);
void fillCovar(double *ysub, double *invvar, long nx, double **aprofile,
        double **apoly, long nTrace, long nCoeff, long nBand, 
        long nPoly, double *beta, long *ia, double **covar, 
        long *xmin, long *xmax) { 
    long i,j,k,l,m,n;
    long tTrace = nCoeff*nTrace;
    long coeff;
    long mStop, jStop;
    long skip, check;
    long iMin, iMax;
    // Fill Profile part of covar first

    // mStop is (nBand+1)*nCoeff ahead of j
    for (l=0,j=0,jStop = nCoeff,mStop=(nBand+1)*nCoeff;
            l<nTrace;l++,mStop+=nCoeff,jStop+=nCoeff) {
        if(mStop > tTrace) mStop = tTrace;
        // fill beta
        // j is index in the 3n coeff, and also the column index
        for (coeff=0; coeff < nCoeff; coeff++, j++)
            if (ia[j]) { // if fit
                // beta  should be ( X^T W Y )
                for (i=xmin[l],k=0;i<=xmax[l];i++,k++) // sum at j column
                    beta[j] += ysub[i] * aprofile[j][k] * invvar[i];
                // fill the upper covar with same fiber
                // covar should be ( X^T W X )
                for (m=j; m<jStop; m++) {
                    if (ia[m]) {
                        for (i=xmin[l],k=0;i<=xmax[l];i++,k++) {
                            covar[j][m] += aprofile[m][k] * aprofile[j][k] * invvar[i];
                        }}}
                // Overlap with polynomial
                for (k=0,m=tTrace; k<nPoly; k++, m++) {
                    if (ia[m]) {
                        for (i=xmin[l],n=0;i<=xmax[l];i++,n++) {
                            covar[j][m] += apoly[k][i] * aprofile[j][n] * invvar[i];
                        }}}
                // Overlap with next fiber, influence on nBand more fiber
                for (m=jStop, skip=1; m<mStop; skip++) {
                    for (check=0;check<nCoeff;check++,m++) {
                        if (ia[m]) {
                            iMin = xmin[l+skip]>xmin[l]?xmin[l+skip]:xmin[l];
                            iMax = xmax[l+skip]<xmax[l]?xmax[l+skip]:xmax[l];
                            // for (i=xmin[l+skip],k=0;i<=xmax[l];i++,k++) {
                            for (i=iMin,k=iMin-xmin[l+skip];i<=iMax;i++,k++) {
                                covar[j][m] += aprofile[m][k] * aprofile[j][i-xmin[l]] * invvar[i];
                                /*if (fabs(covar[j][m])>1e7){
                                    printf("next fiber\n");
                                    printf("covar[%3ld][%3ld]   =%g\n", j, m, covar[j][m]);
                                    for (i=xmin[l+skip],k=0;i<=xmax[l];i++,k++)
                                        printf("aprofile[%3ld][%3ld]%4ld=%g\n", m, k, i,aprofile[m][k]);
                                    for (i=xmin[l+skip],k=0;i<=xmax[l];i++,k++)
                                        printf("aprofile[%3ld][%3ld]%4ld=%g\n", j, i-xmin[l],i, aprofile[j][i-xmin[l]]);
                                    for (i=xmin[l+skip],k=0;i<=xmax[l];i++,k++)
                                        printf("invvar[%3ld]        =%g\n", i, invvar[i]);
                                    printf("\n\n\n\n\n\n\n");
                                }*/
                            }}}}
            }
    }
    // File Profile part of poly
    for (l=0, j=tTrace; l <nPoly; l++, j++)
        if (ia[j]) {
            for (i=0; i<nx; i++) 
                beta[j] += ysub[i] * apoly[l][i] * invvar[i];
            for (k=l,m=j; k<nPoly; k++, m++)
                if (ia[m]) 
                    for (i=0; i<nx; i++) 
                        covar[j][m] += apoly[k][i] * apoly[l][i] * invvar[i];
        }
}

/* cholslRow replaces lower triangle of a with sqrt(covar)
   And we only loop over enough to work on nCoeff parameters for
   adjacent fibers, check that, now looping over nBand elements
   a is modified and cannot be used again for subsequent calls */

void cholslRow(double **a, long *ia, long nTrace, long nCoeff,
        long nBand, long nPoly, double *p, double *b, double *x) {
    long i,j,k;
    long kStart;
    long coeff;
    long tTrace=nTrace*nCoeff;
    double sum;
    // solve for lower triangle part
    for(j=0,i=0;j<nTrace;j++) {
        kStart = i - (nBand + 1)*nCoeff;
        if (kStart < -1) kStart = -1;
        for(coeff=0;coeff<nCoeff;coeff++,i++)
            if (ia[i]) {
                sum = b[i];

                for (sum=b[i],k=i-1; k > kStart; k--) 
                    if (ia[k]) sum -= a[i][k]*x[k];
                x[i] = sum/p[i];
            }
    }
    for(i=tTrace;i<tTrace+nPoly;i++) 
        if (ia[i]) {
            for (sum=b[i],k=i-1;k>=0;k--) 
                if (ia[k])  sum -= a[i][k]*x[k];
            x[i] = sum/p[i];
        }

    // solve for upper triangle part
    for(i=tTrace + nPoly - 1;i>=tTrace;i--) 
        if (ia[i]) {
            for (sum=x[i],k=i+1;k<tTrace+nPoly;k++) 
                if (ia[k])  sum -= a[k][i]*x[k];
            x[i] = sum/p[i];
        }
    for (j=nTrace-1,i=tTrace-1,kStart=tTrace;j>=0;j--) {
        if (j < nTrace-(nBand+1)) kStart -= nCoeff;
        for(coeff=0;coeff<nCoeff;coeff++,i--)
            if (ia[i]) {
                for (sum=x[i],k=i+1; k<kStart; k++) 
                    if (ia[k])  sum -= a[k][i]*x[k];
                for (k=tTrace;k<tTrace + nPoly;k++) 
                    if (ia[k]) sum -= a[k][i]*x[k];
                x[i] = sum/p[i];
            }
    }
}

/*   !!!!!!   This routine needs to be tested to ensure nBand > 1 works  */
void cholslRowCovar(double **a, long *ia, long nTrace, long nCoeff,
        long nBand, long nPoly, double *p) {
    long i,k;
    long kStart = (nBand+1)*nCoeff;
    long tTrace=nTrace*nCoeff;
    double sum;
    double *x;
    long pl;

    for(pl=0;pl < tTrace+nPoly; pl++)  
        if (ia[pl]) {
            x = a[pl];

            for(i=pl,sum=1.0;i<tTrace;i++,sum=0.0) 
                if (ia[i]) {
                    for (k=i-1;k>=pl && k>=i-kStart;k--) 
                        if (ia[k]) sum-=a[i][k]*x[k];
                    x[i] = sum/p[i];
                }
            for(;i<tTrace+nPoly;i++,sum=0.0) 
                if (ia[i]) {
                    for (k=i-1;k>=pl;k--) 
                        if (ia[k]) sum -= a[i][k]*x[k];
                    x[i] = sum/p[i];
                }

            for (i=tTrace+nPoly-1; i>=tTrace && i >= pl;i--) 
                if (ia[i]) {
                    sum = x[i];
                    for (k=i+1;k<tTrace+nPoly;k++) 
                        if (ia[k]) sum -= a[k][i]*x[k];
                    x[i] = sum/p[i];
                }

            for(i=tTrace-1;i>=0 && i >= pl;i--) 
                if (ia[i]) {
                    sum = x[i];
                    for (k=i+1;k<tTrace && k<i+kStart;k++) 
                        if (ia[k]) sum -= a[k][i]*x[k];
                    for (k=tTrace;k<tTrace + nPoly;k++) 
                        if (ia[k]) sum -= a[k][i]*x[k];
                    x[i] = sum/p[i];
                }
        }
}

// Row requires nCoeff parameters per Trace in upper triangle of covar
// do LLT decomposition using The Cholesky–Banachiewicz and Cholesky–Crout algorithms
// bad = choldcRow(covar, ia, nTrace, nCoeff, nBand, wPoly, p);
// p is the L_ii,
// if have error output in this function, may be nBand is too small....
int choldcRow(double **a, long *ia, long nTrace, long nCoeff,
        long nBand, long nPoly, double *p) {
    long i,j,k,l;
    long coeff;
    long tTrace = nCoeff*nTrace;
    long kStart;
    long jStop;
    double sum;

    for (l=0,i=0,jStop=(nBand+1)*nCoeff;l<nTrace;l++,jStop+=nCoeff) {
        if (jStop > tTrace) jStop = tTrace;
        kStart = i - (nBand+1)*nCoeff;
        if(kStart < -1) kStart = -1;
        // l index of trace
        // i: index of all coeff, coeff: index of coeff of a fiber
        for (coeff=0; coeff < nCoeff; coeff++, i++)
            if (ia[i]) {
                for (j=i; j<jStop; j++) 
                    if (ia[j]) {
                        //printf("sum[%3ld][%3ld]: %20.10g\n", i,j,a[i][j]);
                        for (sum = a[i][j], k=i-1; k > kStart; k--) {
                            if(ia[k]) {
                                //printf("%20.10g -= %10.5g[%3ld][%3ld] * %10.5g[%3ld][%3ld]\n", sum, a[i][k],i,k, a[j][k],j,k);
                                sum -= a[i][k]*a[j][k];
                            }
                        }
                        if (i==j) {
                            if (sum <= 0.0 || !isfinite(sum)) {
                                fprintf(stderr,"choldc failed %d %g\n",(int) i,sum);
                                sum = a[i][j];
                                fprintf(stderr,"%g\n",sum);
                                for (k=i-1; k > kStart; k--) {
                                    if(ia[k]) sum -= a[i][k]*a[j][k];
                                    fprintf(stderr,"%g %g %g\n",sum, a[i][k], a[j][k]);
                                }
                                FILE *fp;
                                fp=fopen("fillCovar.dat","wt");
                                for (i=0;i<tTrace+nPoly;i++){
                                    for (j=0;j<tTrace+nPoly;j++){
                                        fprintf(fp, "%g ", a[i][j]); } fprintf(fp, "\n"); }
                                fclose(fp);
                                return (long)(-i);
                            }
                            p[i] = sqrt(sum);
                        }
                        else a[j][i] = sum/p[i];
                    }
                for(j=tTrace; j<tTrace +nPoly; j++) 
                    if(ia[j]) {
                        for(sum = a[i][j], k=i-1; k > kStart; k--) 
                            if(ia[k]) sum -= a[i][k]*a[j][k];
                        a[j][i] = sum/p[i];
                    }
            }
    }

    for(i=tTrace;i<tTrace+nPoly;i++) 
        if(ia[i]) {
            for(j=i;j<tTrace+nPoly;j++) 
                if(ia[j]) {
                    for (sum=a[i][j],k=i-1;k>=0;k--) 
                        if(ia[k]) sum -= a[i][k]*a[j][k];
                    if(i==j) {
                        if (sum <= 0.0)  {
                            fprintf(stderr,"choldc failed %d\n", (int) i);
                            return -i;
                        }
                        p[i] = sqrt(sum);
                    } else a[j][i] = sum/p[i];
                }
        }
    return 1;
}

void chebyshevRow(double x, double *coeff, long nCoeff) {
    /* Return Terms of Chebyshev Polynomial */
    long i;
    double twox;
    if(nCoeff > 0) coeff[0] = 1.0;
    if(nCoeff > 1) coeff[1] = x;
    twox = 2.0*x;
    for(i=2;i<nCoeff;i++)
        coeff[i] = twox * coeff[i-1] - coeff[i-2];
    return;
}
//#endregion: key functions

//#region: function about 2d image process
//!! using map2d file to build a model image
void map2DSlice(long nx, double result[], double bmod[], double map2D[], double flux[], double invvar[]) {
    long i,j,k,fibNum, nb, start, end, mid, ws, we;
    double xcen, sum, reject;
    double fluxs, fluxsb, vars;
    long nData = (long)map2D[0];
    long nTrace = (long)map2D[1];
    double* map=map2D+2;
    double *tempResult;
    tempResult = (double *)malloc(nx * sizeof(double));
//    printf("nx: %ld\n", nData);
    for (i=0;i<nx;i++)
        result[i] = bmod[i];
    for (fibNum=0; fibNum<nTrace; fibNum++) {
        nb    = (long)*(map++);
        xcen  = *(map++);//!! of no use
        start = (long)*(map++);
        end   = (long)*(map++);
        reject= *(map++);
        fluxs = *(map++);
        fluxsb= *(map++);
        vars  = *(map++);

        /*if (width<=0){*/
            sum = 0;
            if (invvar[fibNum]>0) {
                for (j=start,k=0; j<=end; j++,k++) {
                    sum += *map;
                    tempResult[k] = flux[fibNum] * *(map++);
                }
                if (sum==0)
                    sum=1;// make bad fiber profile not nan
                for (j=start,k=0; j<=end; j++,k++)
                    result[j] += tempResult[k]/sum;
            } else {
                map += nb;
            }
        /*}*/
    }
    free(tempResult);
//    printf("max: %ld  min:%ld", max, min);
}

void map2DSliceFixNaNWidth(long nx, double result[], double bmod[], double map2D[], double flux[],
                           double invvar[], double yset[], double badValue, int halfwidth) {
    long i,j,k,fibNum, nb, start, end, mid, ws, we;
    double xcen, sum, reject;
    double fluxs, fluxsb, vars;
    long nData = (long)map2D[0];
    long nTrace = (long)map2D[1];
    double* map=map2D+2;
    double *tempResult;
    tempResult = (double *)malloc(nx * sizeof(double));
//    printf("nx: %ld\n", nData);
    for (i=0;i<nx;i++)
        result[i] = bmod[i];
    for (fibNum=0; fibNum<nTrace; fibNum++) {
        nb    = (long)*(map++);
        xcen  = *(map++);//!! of no use
        start = (long)*(map++);
        end   = (long)*(map++);
        reject= *(map++);
        fluxs = *(map++);
        fluxsb= *(map++);
        vars  = *(map++);

        if (halfwidth<=0){// no fix halfwidth
            sum = 0;
            if (invvar[fibNum]>0) {
                for (j=start,k=0; j<=end; j++,k++) {
                    sum += *map;
                    tempResult[k] = flux[fibNum] * *(map++);
                }
                if (sum==0)
                    sum=1;// make bad fiber profile not nan
                for (j=start,k=0; j<=end; j++,k++)
                    result[j] += tempResult[k]/sum;
            } else {
                map += nb;
            }
        } else { // have fix halfwidth
            sum = 0;
            if (invvar[fibNum]>0) {
                for (j=start,k=0; j<=end; j++,k++) {
                    sum += *map;
                    tempResult[k] = flux[fibNum] * *(map++);
                }
                if (sum==0)
                    sum=1;// make bad fiber profile not nan
                for (j=start,k=0; j<=end; j++,k++)
                    result[j] += tempResult[k]/sum;
            } else {
                map += nb;
                xcen = (int)(yset[fibNum]);
                start = xcen - halfwidth;
                end = xcen + halfwidth;
                for (j=start; j<=end; j++)
                    result[j] += badValue;
            }
        }
    }
    free(tempResult);
//    printf("max: %ld  min:%ld", max, min);
}

void map2DCutFixWidth(long nx, long ny, long nfiber, long width,
                       double fimage[], double fstd[],
                       double yset[],
                       double cutFlux[], double cutFstd[], double cutYset[]) {

    long xIndex, yIndex, fiberIndex, widthIndex, mid, cutYIndex;
    long halfwidth = (long)((width-1)/2);
    long ystart, yend, imageFullIndex, fiberFullIndex;
    long fiberIndexStart, cutFullIndex;
    double thisInvvar, thisYset;

    for (fiberIndex=0; fiberIndex<nfiber; fiberIndex++) {
        fiberIndexStart = fiberIndex*nx;
        for (xIndex=0; xIndex<nx; xIndex++) {
            fiberFullIndex = fiberIndexStart + xIndex;
            thisYset = (long)(yset[fiberFullIndex]);
            ystart = thisYset-halfwidth<0? 0 : thisYset-halfwidth;
            yend = thisYset+halfwidth>ny-1? ny-1 : thisYset+halfwidth;
            for (yIndex=ystart, cutYIndex=width*fiberIndex; yIndex<=yend; yIndex++, cutYIndex++) {
                imageFullIndex = yIndex * nx + xIndex;
                cutFullIndex = cutYIndex * nx + xIndex;
                cutFlux[cutFullIndex] = fimage[imageFullIndex];
                cutFstd[cutFullIndex] = fstd[imageFullIndex];
                cutYset[cutFullIndex] = yIndex - yset[fiberFullIndex];
            }
        }
    }
}

//!! sN, sS, sW, sE are of only one strip
void remapFlux(double *cutFlux, double *cutFstd,
               double *bN, double *bS, double *bW, double *bE, long bnx, long bny, long bWidth,
               double *sN, double *sS, double *sW, double *sE, long snx, long sny,
               double *resultFlux, double *resultVar, double *sFillingFactor) {
    long byIndex, bxIndex, syIndex, sxIndex;
    long xstart, xend, ystart, yend;
    double thisbN, thisbS, thisbW, thisbE;
    double thissN, thissS, thissW, thissE;
    double thisBigArea, thisFlux, thisStd, thisSmallArea, thisFactor, thisOriginSmallArea;
    long thisxstart, thisxend, thisystart, thisyend;
    long nFiber = (long)(bny/bWidth);
    long fiberNum, i;
    double thisResultFlux, thisResultVar;
    for (fiberNum=0; fiberNum<nFiber; fiberNum++) {
        for (byIndex=fiberNum*bWidth; byIndex<(fiberNum+1)*bWidth; byIndex++) {
            thisxstart = 0;
            thisxend = 0;
            for (bxIndex=0; bxIndex<bnx; bxIndex++) {
                thisystart = 0;
                thisyend = 0;
                thisbN = bN[byIndex*bnx + bxIndex];
                thisbS = bS[byIndex*bnx + bxIndex];
                thisbW = bW[byIndex*bnx + bxIndex];
                thisbE = bE[byIndex*bnx + bxIndex];
                thisBigArea = (thisbN-thisbS) * (thisbE-thisbW);
                thisFlux = cutFlux[byIndex*bnx + bxIndex];
                thisStd = cutFstd[byIndex*bnx + bxIndex];
                //!! assume that xset and yset grow in column and in lines in parallel
                while (sW[thisxstart]<thisbW)
                    thisxstart++;
                thisxstart--;
                thisxstart = thisxstart<0 ? 0 : thisxstart;
                thisxstart = thisxstart<snx ? thisxstart : snx-1;

                while (sE[thisxend]<thisbE)
                    thisxend++;
                thisxend = thisxend<snx ? thisxend : snx-1;

                while (sS[thisystart*snx+thisxstart]<thisbS)
                    thisystart++;
                thisystart--;
                thisystart = thisystart<0 ? 0 : thisystart;
                thisystart = thisystart<sny ? thisystart : sny-1;

                while (sN[thisyend*snx+thisxstart]<thisbN)
                    thisyend++;
                thisyend = thisyend<sny ? thisyend : sny-1;

                /*printf("fiberNum: %3ld, bS:%f bN:%f bW:%f bE:%f big y:%4ld, big x:%4ld, BigArea: %f, (%3ld, %3ld) (%3ld, %3ld)\n", fiberNum, thisbS, thisbN, thisbW, thisbE, byIndex, bxIndex, thisBigArea, thisxstart, thisxend, thisystart, thisyend);*/
                if (thisBigArea<0) exit(-1);
                if ((thisxstart>=thisxend) || (thisystart>=thisyend))
                    continue;// big pixel not in small pixel range
                for (syIndex=thisystart; syIndex<=thisyend; syIndex++) {
                    for (sxIndex=thisxstart; sxIndex<=thisxend; sxIndex++) {
                        thissN = sN[syIndex*snx + sxIndex];
                        thissS = sS[syIndex*snx + sxIndex];
                        thissW = sW[syIndex*snx + sxIndex];
                        thissE = sE[syIndex*snx + sxIndex];
                        thisOriginSmallArea = (thissN-thissS) * (thissE-thissW);

                        thissN = thissN<thisbN ? thissN : thisbN;
                        thissS = thissS<thisbS ? thisbS : thissS;
                        thissW = thissW<thisbW ? thisbW : thissW;
                        thissE = thissE<thisbE ? thissE : thisbE;
                        thisSmallArea = (thissN-thissS) * (thissE-thissW);
                        thisFactor = thisSmallArea/thisBigArea;
                        thisResultFlux = thisFlux * thisFactor;
                        /*thisResultInvvar = 1/(thisStd*thisStd*thisFactor);*/
                        thisResultVar = (thisStd*thisStd*thisFactor);
                        resultFlux[fiberNum*sny*snx + syIndex*snx + sxIndex] += thisResultFlux;
                        resultVar[fiberNum*sny*snx + syIndex*snx + sxIndex] += thisResultVar;
                        sFillingFactor[fiberNum*sny*snx + syIndex*snx + sxIndex] += thisSmallArea/thisOriginSmallArea;
                    }
                }
            }
        }
    }

}

//#endregion: function about 2d image process

//#region: profiles used in python
//!! not in use
void profile23(int nx, double* x,
               double A, double mu, double sigma,
               double A1, double sigma1,
               double* result) {
    double diff, diffabs, epow1, epow, ddiff, diff1;
    int N = 5;
    int half = (N-1)/2;
    double eachResult = 0;
    double step = 1/N;
    double r,r1;
    double c=sqrt(2*3.1415926);
    int i,j;

    for (i=0; i<nx; i++){
        r=0;
        r1=0;
        for (j=-half; j<=half; j++) {
            ddiff = j*step;
            diff = (x[i]-mu+ddiff)/sigma;
            diff1= (x[i]-mu+ddiff)/sigma1;
            diffabs = fabs(diff1);
            epow = diff*diff;
            epow1 = diff1*diff1*diffabs;
            r += A/(c*sigma) * exp(-epow/2);
            r1+= A/(2.575798634 * sigma) * exp(-epow1/3);
        }
        result[i] = (r+r1)/N;
    }
}

//!! in use
void profile233(int nx, double* x,
                double* mu, long nLine,
                double* A, double* sigma,
                double* A3, double* sigma3,
                double* result) {
    double diff, diffabs, epow3, epow, ddiff, diff3;
    int N = 5;
    int half = (N-1)/2;
    double eachResult = 0;
    double step = 1/N;
    double r,r3;
    double c=sqrt(2*3.1415926);
    double thisA, thisA3, thismu, thissigma, thissigma3;
    int i,j, line;

    for (line=0; line<nLine; line++) {
        for (i=0; i<nx; i++){
            if (fabs((x[i] - mu[line])/sigma[line])>10)
                continue;
            else {
                thismu = mu[line];
                thissigma = sigma[line];
                thissigma3 = sigma3[line];
                thisA = A[line];
                thisA3 = A3[line];
            }
            r=0;
            r3=0;
            for (j=-half; j<=half; j++) {
                ddiff = j*step;
                diff = (x[i]-thismu+ddiff)/thissigma;
                diff3= (x[i]-thismu+ddiff)/thissigma3;
                diffabs = fabs(diff3);
                epow = diff*diff;
                epow3 = diff3*diff3*diffabs;
                r += thisA/(c*thissigma) * exp(-epow/2);
                r3+= thisA3/(2.575798634 * thissigma3) * exp(-epow3/3);
            }
            result[i] += (r+r3)/N;
        }
    }
}

void profileGaussian(int nx, double* x,
                double* mu, long nLine,
                double* A, double* sigma,
                double* result) {
    double diff, diffabs, epow3, epow, ddiff, diff3;
    int N = 5;
    int half = (N-1)/2;
    double eachResult = 0;
    double step = 1/N;
    double r;
    double c=sqrt(2*3.1415926);
    double thisA, thismu, thissigma;
    int i,j, line;

    for (line=0; line<nLine; line++) {
        for (i=0; i<nx; i++){
            if (fabs((x[i] - mu[line])/sigma[line])>10)
                continue;
            else {
                thismu = mu[line];
                thissigma = sigma[line];
                thisA = A[line];
            }
            r=0;
            for (j=-half; j<=half; j++) {
                ddiff = j*step;
                diff = (x[i]-thismu+ddiff)/thissigma;
                epow = diff*diff;
                r += thisA/(c*thissigma) * exp(-epow/2);
            }
            result[i] += r/N;
        }
    }
}

//#endregion: profiles used in python


#undef MAXEXP
#undef NSUBSAMP


//#region: backup codes
/*

void map2DFixWidthRead(int nTrace, double y[], double result[], double yset[], int width) {
    // len(y) = ny
    // len(result) = nTrace * (2*width+1)
    long i,j,k, fibNum, start, end, mid;
    long nB = 2*width + 1;
    for (fibNum=0; fibNum<nTrace; fibNum++) {
        mid = yset[fibNum];
        start = mid-width;
        end   = mid+width;
        for (j=start, k=0; j<=end; j++, k++) {
            result[fibNum*nB + k] = y[j];
        }
    }
}

void map2DFixWidthWrite(int nTrace, double flux[], double result[], double yset[], int width) {
    // len(flux) = nTrace*(2*width+1)
    // len(result) = ny
    long i,j,k, fibNum, start, end, mid;
    long nB = 2*width + 1;
    for (fibNum=0; fibNum<nTrace; fibNum++) {
        mid = yset[fibNum];
        start = mid-width;
        end   = mid+width;
        for (j=start, k=0; j<=end; j++, k++) {
            result[j] = flux[fibNum*nB + k];
        }
    }
}

// data struct of origin : nx, fiberNum, [xcen, start, end, data...], [...]...
// data struct of result : nx, fiberNum, [N, x0...x?, y0....y?]
//                     pixelNumer             max width of all the map2d files
void buildMasterProfile(long nx, long nFile, long nLine,
                        // nFile * nx * nLine       nx * (3*nLine*nFile)
                        double *origin,           double *result) {
    int i,j,k,l,m,n;
    double *eachIndexs, *eachProfiles, *eachSNR;// nx
    int *index; //nx
    double *resultPointer;
    double **fiberPointers;
    double *thisFiberPointer;
    long fileSize = nx*nLine;
    long longLineSize = 3*nLine*nFile;
    int fiberNumber = origin[1];
    int pixelNum, fileNum, fiberNum;
    int count, countStart, lineCount;
    double xcen, norm, reject;
    double fluxs, fluxsb, vars, SNR;
    int start, end, len;

    for (i=0; i<nFile; i++)
        assert(origin[1+i*fileSize]==fiberNumber);

    eachIndexs = (double*)malloc(nx * sizeof(double));
    eachProfiles = (double*)malloc(nx * sizeof(double));
    eachSNR = (double*)malloc(nx * sizeof(double));
    index = (int*)malloc(nx * sizeof(int));
    fiberPointers = (double**)malloc(nFile * sizeof(double *));

    for (pixelNum=0; pixelNum<nx; pixelNum++) {
        // make all fiberPointers at the front of each line
        for (fileNum=0; fileNum<nFile; fileNum++)
            fiberPointers[fileNum] = origin + fileNum*fileSize+pixelNum*nLine + 2;

        resultPointer = result + pixelNum*longLineSize + 1;
        *resultPointer = fiberNumber; resultPointer++;
        lineCount = 2;
        for (fiberNum=0; fiberNum<fiberNumber; fiberNum++) {
            // merge nFile to one, then sort
            for (fileNum=0, count=0; fileNum<nFile; fileNum++) {
                len = *(fiberPointers[fileNum]++);
                xcen = *(fiberPointers[fileNum]++);
                start = *(fiberPointers[fileNum]++);
                end = *(fiberPointers[fileNum]++);
                reject = *(fiberPointers[fileNum]++);
                fluxs = *(fiberPointers[fileNum]++);
                fluxsb = *(fiberPointers[fileNum]++);
                vars = *(fiberPointers[fileNum]++);
                norm = 0;
                countStart = count;
                SNR = fluxs/sqrt(vars);
                for (i=(int)start; i<=(int)end; i++, count++) {
                    eachIndexs[count] = i-xcen;
                    eachSNR[count] = SNR;
                    norm += *(fiberPointers[fileNum]);
                    eachProfiles[count] = *(fiberPointers[fileNum]++);
                }
                if (norm==0)
                    norm=1;
                norm /= reject;
                for (i=countStart; i<count; i++)
                    eachProfiles[i] = eachProfiles[i]/norm;
            // now to sort for this profile
            }
            *resultPointer = count;resultPointer++;
            doubleIndexSort(eachIndexs, count, index);
            for (i=0; i<count; i++, resultPointer++)
                *resultPointer = eachIndexs[index[i]];
            for (i=0; i<count; i++, resultPointer++)
                *resultPointer = eachProfiles[index[i]];
            for (i=0; i<count; i++, resultPointer++)
                *resultPointer = eachSNR[index[i]];
            lineCount += count*3+1;
        }
        result[pixelNum*longLineSize] = lineCount;
    }

    free(eachIndexs);
    free(eachProfiles);
    free(eachSNR);
    free(index);
    free(fiberPointers);
}


    if (map2d) {
        double *tempFlux, *tempBack, *tempVar;
        double eachOtherFlux, eachFlux, eachRatio;
        //!! temp flux all
        tempFlux = (double*)malloc(nx * sizeof(double));
        tempBack = (double*)malloc(nx * sizeof(double));
        tempVar = (double*)malloc(nx * sizeof(double));
        var = (double*)malloc(nx * sizeof(double));
        long lowF, highF, lowX, highX;
        for (i=0; i<nx; i++)
            var[i] = 1/invvar[i];

        map2dFirst=map2d;
        *(map2d++) = map2dCount;// right map2dCount num in the end of the if
        *(map2d++) = nTrace;
        //!! for each fiber
        for (i=0; i<nTrace; i++) {
            length = xmax[i] - xmin[i] + 1;
            if (length < 0) length = 0;
            map2dCount += length+7; // count for profile data
            *(map2d++) = length;
            *(map2d++) = xcen[i];
            *(map2d++) = (double)xmin[i];
            *(map2d++) = (double)xmax[i];

            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                tempFlux[j] = fimage[j];
                tempBack[j] = bmod[j];
                tempVar[j] = var[j];
            }

            lowF = i-1>=0 ? i-1 :0;
            highF = i+1<=nTrace-1 ? i+1 : nTrace-1;
            //!! for each adjecent fiber
            for (j=lowF; j<=highF; j++) {
                if (j==i||xmax[j]<xmin[i]||xmin[j]>xmax[i])
                    continue;
                lowX = xmin[j] > xmin[i] ? xmin[j] : xmin[i];
                highX = xmax[j] < xmax[i] ? xmax[j] : xmax[i];

                //!! for each coresponsding pixel
                for (l=lowX, k=lowX-xmin[j]; l<=highX ; l++, k++) {
                    //!! only subtract the first profile
                    eachFlux = ans[i*nCoeff]*aprofile[i*nCoeff][l-xmin[i]];
                    eachOtherFlux = ans[j*nCoeff]*aprofile[j*nCoeff][k];
                    //!! we overestimate the error here... it's safe but not good...
                    if (eachFlux > eachOtherFlux) {
                        tempFlux[l] = tempFlux[l] - bmod[l] - eachOtherFlux>0?tempFlux[l]-bmod[l]-eachOtherFlux:0;
                    } else {
                        //!! the algorithm need to be improved here...
                        tempFlux[l] = eachFlux;
                        eachRatio = eachFlux/fimage[l];
                        tempVar[l] = var[l] * eachRatio;
                    }
                }
            }
            sumFlux[i] = 0;
            sumFluxBack[i] = 0;
            sumVar[i] = 0;
            for (j=xmin[i], k=0; j <= xmax[i]; j++, k++) {
                sumFlux[i] += tempFlux[j];
                sumFluxBack[i] += bmod[j];
                sumVar[i] += tempVar[j];
            }
            if (isfinite(sumFlux[i])&&isfinite(sumVar[i])){
                *(map2d++) = sumFlux[i];
                *(map2d++) = sumFluxBack[i];
                *(map2d++) = sumVar[i];
                for(k=xmin[i]; k<=xmax[i]; k++)
                    *(map2d++) = tempFlux[k];// only get map for the first conponent
            } else {//!! have bad pixel in the profile
                *(map2d++) = ans[i*nCoeff];//!! use the direct extract result..
                sumFlux[i] = ans[i*nCoeff];
                *(map2d++) = sumFluxBack[i];
                *(map2d++) = 1/p[i*nCoeff];//!! use the direct extract result..
                sumVar[i]  = 1/p[i*nCoeff];
                for(k=xmin[i], l=0; k<=xmax[i]; k++,l++)
                    *(map2d++) = aprofile[i*nCoeff][l];
            }
        }
        map2dFirst[0] = map2dCount;
        free(tempFlux);
        free(tempBack);
        free(tempVar);
        free(var);
    }


*/
//#endregion: backup codes
