#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

void correlate(double* v1, int n1, double* v2, int n2, int* lag, int nlag, double* value) {
    int i, i1, i2, iStart;
    for (i=0; i<nlag; i++) {
        value[i]=0;
        if (lag[i]<0) {
            iStart = -lag[i];
            for (i1=0, i2=iStart; i2<n2 && i1<n1; i1++, i2++)
                value[i] += v1[i1] * v2[i2];
        } else {
            iStart = lag[i];
            for (i1=iStart, i2=0; i2<n2 && i1<n1; i1++, i2++)
                value[i] += v1[i1] * v2[i2];
        }
    }
}

void buildLamp(double* lam, int nLam, double* modLam, int nx, int npad, double* intensity, double* model) {
    int iline, iloc;
    double dpix, dx;
    if (modLam[1] > modLam[0]) {
        for (iline=0; iline<nLam; iline++) {
            iloc = 0;
            while (lam[iline]>modLam[iloc]) {
                if (iloc>=nx) {
                    iloc=0;
                    break;
                } else {
                    iloc++;
                }}
            if (iloc>0) {
                iloc = iloc-1;
                dx = lam[iline] - modLam[iloc];
                if (iloc>0 && iloc<nx-2) {
                    dpix = modLam[iloc+1] - modLam[iloc];
                    model[npad+iloc] += intensity[iline] * (1-dx/dpix);
                    model[npad+iloc+1] += intensity[iline] * dx/dpix;
                }
            }
            //printf("iLine: %5d  loc: %5d  dx:%8g  dpix:%8g\n", iline, iloc, dx, dpix);
        }
    } else {
        for (iline=0; iline<nLam; iline++) {
            iloc = 0;
            while (lam[iline]<modLam[iloc]) {
                if (iloc>=nx) {
                    iloc=0;
                    break;
                } else {
                    iloc++;
                }}
            if (iloc>0) {
                iloc = iloc-1;
                dx = modLam[iloc] - lam[iline];
                if (iloc>0 && iloc<nx-2) {
                    dpix = modLam[iloc] - modLam[iloc+1];
                    model[npad+iloc] += intensity[iline] * (1-dx/dpix);
                    model[npad+iloc+1] += intensity[iline] * dx/dpix;
                }
            }
            //printf("iLine: %5d  loc: %5d  dx:%8g  dpix:%8g\n", iline, iloc, dx, dpix);
        }
    }
}

double gaussian(double x, double A, double mu, double sigma) {
    return A/(sqrt(2*3.1415926)*sigma) * exp(-((x-mu)*(x-mu))/(2*sigma*sigma));
}

//#!! assume lam is increasing
//!! step should be even
//!! mu +- sigma*sigmaLength should be larger than 1 modLam step
//!! subPixel should be odd
void buildSmoothLamp(double* lam, int nLam, double* modLam, int nx, int subPixel,
                     double* intensity, double *sigmas, int sigmaLength, double* model) {
    double low, high, step, thisIntensity;
    int indexLow, indexHigh;
    int i,j,k,l;
    step = modLam[1] - modLam[0];
    double subPixelLength = step/(subPixel+1);
    long subIndex = (long)((subPixel-1)/2);
    double result, xValue, allResult, *tempResult;
    tempResult = (double*)malloc(nx*sizeof(double));
    for (i=0; i<nLam; i++) {
        low = lam[i] - sigmas[i]*sigmaLength;
        high = lam[i] + sigmas[i]*sigmaLength;
        thisIntensity = intensity[i];
        if (modLam[0]>high||modLam[nx-1]<low)
            continue;
        indexLow = 0;
        indexHigh = nx-1;
        while (modLam[indexLow]<low)
            indexLow++;
        while (modLam[indexHigh]>high)
            indexHigh--;
        if (indexHigh<indexLow)
            printf("some flux may be missing due to large step in modLam..\n");
        allResult = 0;
        for (j=indexLow, l=0; j<=indexHigh; j++,l++) {
            result = 0;
            for (k=-subIndex; k<=subIndex; k++) {
                result += gaussian(modLam[j] + k*subPixelLength,
                                   1, lam[i], sigmas[i])/subPixel;
            }
            allResult += result;
            tempResult[l] = result;
        }
        for (j=indexLow, l=0; j<=indexHigh; j++,l++) {
            model[j] += tempResult[l]/allResult*thisIntensity;
        }
    }
    free(tempResult);
}

void getSubIndex(long ii, long* nsteps, long n, long* result) {
    long *toMod, *toDiv;
    long i;
    toMod = (long *)malloc(sizeof(long)*n);
    toDiv = (long *)malloc(sizeof(long)*n);
    toMod[0] = nsteps[0];
    toDiv[0] = 1;
    for (i=0; i<n-1; i++) {
        toMod[i+1] = toMod[i] * nsteps[i+1];
        toDiv[i+1] = toDiv[i] * nsteps[i];
    }
    for (i=n-1; i>=0; i--) {
        result[i] = (ii%toMod[i])/toDiv[i];
    }
}
