# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function, unicode_literals)
# Copyright 2016 Wu Jin <wujin_astro@pku.edu.cn>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
''' Notes
    vim:set foldmethod=indent
    use vimrc config 'customFolding.vim' from https://gist.github.com/Fmajor to
        make reading the code much more easier!
'''
''' Update history
    2016-08-09 add the license
'''
''' Bugs or Cautions:
    None
'''

import os
import sys
import time
import warnings
import ipdb
from ..utils import messager

#!! configs
codePath = os.path.dirname(__file__)
outputPath = "outputs"


#!! m2fs macros
#!! i just find that it also work for multi file....
# before this you should have
#    outputFileName
#    cofig2showList
# gen
#    toReturn : bool
#    outFile
singleFileMacro=r"""
global outputPath, outFile, rerun, toReturn
outputPath = configs["outputPath"]
outFile= os.path.join(outputPath, outputFileName)
config2showList = configs.get("config2show", "all")
rerun = int(configs.get('rerun', 0))
funcName = sys._getframe().f_back.f_code.co_name
fileName = globals().get("__file__","")
if fileName:
    fileName = os.path.basename(fileName)
    fileName = fileName.split(".py")[0]
    fileName += "."
if not rerun and os.path.exists(outFile):
    logger.p("in {}{}: {} already exists, skip".\
            format(fileName, funcName, outFile))
    toReturn = True
else:
    if isinstance(config2showList, list):
        config2show={}
        for eachKey in config2showList:
            config2show[eachKey]=configs.get(eachKey)
    elif config2showList.lower()=="all":
        config2show = configs
    else:
        raise logger.Exception("unknown config2show config: {}".format(config2showList))
    logger.p("{}{}({}, {})".format( fileName,
        funcName, outFile, config2show))
    toReturn = False
"""


def safe_mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)

class m2fsError(Exception):
    pass
class m2fsWarning(UserWarning):
    pass
warnings.simplefilter("always", m2fsWarning)

def init(logPath=None, configs={}):
    inParallel = configs.get("inParallel", False)
    name = configs.get("name")
    index = configs.get("index")

    if logPath is None:
        logPath = "logs"
    logFile = os.path.join(logPath, "log.txt")
    warningFile = os.path.join(logPath, "warnings.txt")
    errorFile = os.path.join(logPath, "errors.txt")
    defaultConfigFile = os.path.join(codePath,"configs","config.default.txt")
    defaultTraceTemplateFile = os.path.join(codePath,"configs","traceTemplate.default.txt")

    defaultConfigs = {
        "codePath":codePath,
        "outputPath":outputPath,
        "logPath":logPath,
        "logFile":logFile,
        "warningFile":warningFile,
        "errorFile":errorFile,
    }
    if sys.platform=="darwin":
        defaultConfigs["plotBackend"] = "MacOsX"
    elif sys.platform=="linux":
        defaultConfigs["plotBackend"] = "TkAgg"
    else:
        raise Exception("unsupported platform {}".format(sys.platform))


    safe_mkdir(logPath)
    safe_mkdir(os.path.join(logPath, "parallel-log"))
    safe_mkdir(outputPath)
    #!! init parallel or nonparallel logger
    if not inParallel:
        loggerName = "m2fsLogger-{}".format(time.strftime("%Y-%m-%dT%H-%M-%S", time.gmtime()))
        # logger is root logger
        logger = messager.Messager(name=loggerName, logFile=logFile,
                    warningFile=warningFile, errorFile=errorFile,
                    initLog=True, initErrorLog=True,
                    renew=False, warningClass=m2fsWarning, exceptionClass=m2fsError)
        logger.p("="*10 + "Init m2fs data reduction pipeline" + "="*10, "ygb")
    else:
        loggerName = "{}-Logger-{}".format(name, time.strftime("%Y-%m-%dT%H-%M-%S", time.gmtime()))
        thisLogger = {"warningFile":"warningFile", "errorFile":"errorFile",
                   "name":loggerName, "initErrorLog":False,
                   "logFile":[os.path.join(logPath, "parallel-log", name+".log"),
                              logFile],
                   "initWarningLog":False,"initLog":False,
                   "renewLog":[False,False], "renewError":False, "renewWarning":False,
                   }
        logger = messager.Messager(**thisLogger)
    # t is TeeStderr, which direct the stderr to a special logger
    t = messager.TeeStderr(logFile=warningFile)
    defaultConfigs["logger"] = logger
    defaultConfigs["teeStderr"] = t
    result = {"defaultConfigs":defaultConfigs,
              "defaultConfigFile":defaultConfigFile,
              "defaultTraceTemplateFile":defaultTraceTemplateFile
              }
    return result
