# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function, unicode_literals)
# Copyright 2016 Wu Jin <wujin_astro@pku.edu.cn>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
''' Notes
    vim:set foldmethod=indent
    use vimrc config 'customFolding.vim' from https://gist.github.com/Fmajor to
        make reading the code much more easier!
'''
''' Update history
    2016-08-11 add the license
'''
''' Bugs or Cautions:
    None
'''
"""Main file for the m2fs config 22 reduction pipeline

"""
import os
import stat
import ipdb
import sys
import shutil
import numpy as np
import astropy.io.fits as pyfits
import multiprocessing
import copy
from . import de
from . import amp
from . import cosmic
from . import dark
from . import extract
from . import wavecalib
from . import flat
from . import skysub
from . import stack
from ..DS10.bin import ds10_bin
from subprocess import Popen

def main(configs):
    jobName = configs["jobName"]
    scis = configs["data"]["scis"]
    arcs = configs["data"]["arcs"]
    darks = configs["data"]["darks"]
    twilights = configs["data"]["twilights"]
    scis.sort()
    arcs.sort()
    darks.sort()
    twilights.sort()
    alls = scis + arcs + darks + twilights
    alls.sort()
    endStep = configs["global"]["end"]
    logger = configs["global"]["logger"]
    rerunNum = configs["global"]["rerunNum"]
    outputPath = configs["global"]["outputPath"]
    dataPath = configs["global"]["dataPath"]
    codePath = configs["global"]["codePath"]
    start = configs["global"]["start"]
    end = configs["global"]["end"]
    inParallel = configs["global"].get("inParallel", False)
    sessionName = jobName
    plotDebug = configs.get("debug", {}).get("plotDebug",False)
    if not plotDebug:
        de.plt.switch_backend("Agg")
    with open(os.path.join(outputPath, "_cleanDebug"), "w") as f:
        f.write("tmux kill-session -t {}".format(sessionName))
    os.chmod(os.path.join(outputPath, "_cleanDebug"),stat.S_IRWXU)
    def getType(name):
        for each in scis:
            if each in name:
                return "sci"
        for each in arcs:
            if each in name:
                return "arc"
        for each in twilights:
            if each in name:
                return "twilight"
        for each in darks:
            if each in name:
                return "dark"
    def updateConfigsByType(filename, configs):
        keys = configs.keys()
        tnames = filter(lambda x:"sci-" in x or "dark-" in x or "arc-" in x or "twilight-" in x, keys)
        names = [each.split("-")[-1] for each in tnames]
        thistype = getType(filename)
        for eachname in names:
            todo = configs.get("{}-{}".format(thistype, eachname))
            if todo is not None:
                logger.p("update config {} with {} to {}".\
                    format(eachname, "{}-{}".format(thistype, eachname), todo))
                configs[eachname] = todo
    if not inParallel:
        logger.p("scis: {}".format(scis))
        logger.p("darks: {}".format(darks))
        logger.p("arcs: {}".format(arcs))
        logger.p("twilights: {}".format(twilights))
    #region: ==========gen aux files==========
    for eachType in configs["data"]:
        eachTypeDict = configs["data"][eachType]
        for eachName in eachTypeDict:
            todo = os.path.join(outputPath, eachType+"_"+eachName)
            if not os.path.exists(todo):
                open(todo, "w").close()
    tempStr = "\n".join(["c"+each+".fits" for each in scis]) + "\n"
    f = open(os.path.join(outputPath, "c_scis.txt"), "w")
    f.write(tempStr)
    f.close()
    tempStr = "\n".join(["c"+each+".fits" for each in arcs]) + "\n"
    f = open(os.path.join(outputPath, "c_arcs.txt"), "w")
    f.write(tempStr)
    f.close()
    tempStr = "\n".join(["c"+each+".fits" for each in darks]) + "\n"
    f = open(os.path.join(outputPath, "c_darks.txt"), "w")
    f.write(tempStr)
    f.close()
    tempStr = "\n".join(["c"+each+".fits" for each in twilights]) + "\n"
    f = open(os.path.join(outputPath, "c_twilights.txt"), "w")
    f.write(tempStr)
    f.close()
    tempStr = "\n".join(["c"+each+".fits" for each in scis+arcs] +\
                      ["cdc"+each+".fits" for each in twilights] +\
                      ["cc"+each+".fits" for each in darks]) + "\n"
    f = open(os.path.join(outputPath, "c_all.txt"), "w")
    f.write(tempStr)
    f.close()
    tempStr = "\n".join(["dc"+each+"Flux.fits" for each in scis+arcs] +\
                       ["cdc"+each+"Flux.fits" for each in twilights]) + "\n"
    f = open(os.path.join(outputPath, "c_all_flux.txt"), "w")
    f.write(tempStr)
    f.close()
    tempStr = "\n".join(["cdc"+each+"-flatfield.fits" for each in twilights]) + "\n"
    f = open(os.path.join(outputPath, "c_all_flatfield.txt"), "w")
    f.write(tempStr)
    f.close()
    #endregion: aux
    #region: ==========step: 1 amp the data files==============
    if start<2 and end>0:
        thisConfig = configs["parameters"].get("amp", {})
        fileNameDatePair = []
        ds10Debug = thisConfig.get("ds10Debug", False)
        #ds10Debug = True
        for eachType in ["scis", "arcs", "darks", "twilights"]:
            thisConfig.update({"object":eachType[:-1]})
            for eachRaw in configs["data"][eachType]:
                todo = os.path.join(dataPath, eachRaw)
                result = amp.amp(todo, configs=thisConfig)
                #region: config debug with ds10
                if result is None and ds10Debug:# in case of not rerun
                    path = os.path.join(outputPath, eachRaw+".fits")
                    result = (path, pyfits.open(path)[0].header["object"].split("@")[-1])
                fileNameDatePair.append(result)
                #endregion: config debug with ds10

        #!! view all output fits using ds10
        if ds10Debug:
            dtype = np.dtype([(str("filename"), "U100"), (str("datetime"), "U100")])
            fdPair = np.array(fileNameDatePair, dtype=dtype)
            index = np.argsort(fdPair["datetime"])
            allOutputPath = fdPair["filename"][index].tolist()
            #allOutputPath = [os.path.join(outputPath, each+".fits") for each in alls]
            sessionName = jobName
            windowName = "amp-debug"
            todo = "ds10 {} -f1 --name={} --width=200 --height=200".\
                    format(" ".join(allOutputPath), sessionName+":"+windowName)
            os.system('tmux new-session -s {sessionName} -d >/dev/null 2>&1'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('tmux kill-window -t {sessionName}:{windowName} >/dev/null 2>&1'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('tmux new-window -t {sessionName} -n {windowName}'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('sleep 1')
            os.system('tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                    format(sessionName=sessionName, windowName=windowName, todo=todo))
    #endregion: amp
    #region: ==========step: 2 clean cosmic ray for all files==============
    todoList = darks + scis + twilights + arcs
    if start<3 and end>1:
        #!! first do cosmic for all files
        parallel = configs["parameters"].get("laCosmic", {}).get("parallel", False)
        fileNameDatePair = []
        if not parallel:
            thisConfig = configs["parameters"].get("laCosmic", {})
            ds10Debug = thisConfig.get("ds10Debug", False)
            for eachFile in todoList:
                updateConfigsByType(eachFile, thisConfig)
                result = cosmic.lacosmic(os.path.join(outputPath, eachFile)+'.fits', thisConfig)
                #region: config debug with ds10
                if result is None and ds10Debug:# in case of not rerun
                    path = os.path.join(outputPath, "c"+eachFile+".fits")
                    result = (path, pyfits.open(path)[0].header["object"].split("@")[-1])
                fileNameDatePair.append(result)
                #endregion: config debug with ds10
        else:
            count = multiprocessing.cpu_count()-2
            pool = multiprocessing.Pool(processes=count)
            cosmicConfig = configs["parameters"].get("laCosmic", {})
            ds10Debug = cosmicConfig.get("ds10Debug", False)
            cosmicConfig.pop("logger")
            cosmicConfig.pop("teeStderr")
            jobList=[]
            logger.p("Do cosmic in parallel: {}".format(todoList))
            for eachFile in todoList:
                thisConfig = copy.deepcopy(cosmicConfig)
                updateConfigsByType(eachFile, thisConfig)
                thisLogger = {"warningFile":thisConfig["warningFile"],
                              "errorFile":thisConfig["errorFile"],
                              "name":eachFile, "initErrorLog":False,
                              "logFile":[os.path.join(thisConfig["logPath"], "parallel-log",
                                                     eachFile+"-cosmic.log"),
                                         thisConfig["logFile"]],
                              "initWarningLog":False,"initLog":False,
                              "renewLog":[True,False], "renewError":False, "renewWarning":False,
                              }
                    #logger = messager.Messager(name=loggerName, logFile=logFile,
                    #               warningFile=warningFile, errorFile=errorFile,
                    #               initLog=True, initErrorLog=True,
                    #               renew=True, warningClass=m2fsWarning, exceptionClass=m2fsError)
                thisConfig["logger"] = thisLogger
                #cosmic.lacosmic(os.path.join(outputPath, eachFile)+'.fits', thisConfig)
                jobList.append(
                        pool.apply_async(
                            cosmic.lacosmic,
                            (os.path.join(outputPath, eachFile)+'.fits',
                             thisConfig)))
            pool.close(); pool.join()
            result = [each.get() for each in jobList]
            temp = []
            for eachResult, eachFile in zip(result, todoList):
                if eachResult is None and ds10Debug:# in case of not rerun
                    path = os.path.join(outputPath, "c"+eachFile+".fits")
                    eachResult = (path, pyfits.open(path)[0].header["object"].split("@")[-1])
                temp.append(eachResult)
            fileNameDatePair = temp

        #!! make ds10 debug pair...
        #!! view all output fits using ds10
        if ds10Debug:
            N = len(fileNameDatePair)
            temp = ["?"] * N *2
            for i in range(N): # make d* and * pair
                todo = fileNameDatePair[i]
                filename = os.path.basename(todo[0])[1:]
                oldfilename = os.path.join(outputPath, filename)
                temp[2*i+0] = (todo[0], todo[1]+"..")
                temp[2*i+1] = (oldfilename, todo[1])
            fileNameDatePair = temp

            dtype = np.dtype([(str("filename"), "U100"), (str("datetime"), "U100")])
            fdPair = np.array(fileNameDatePair, dtype=dtype)
            index = np.argsort(fdPair["datetime"])
            allOutputPath = fdPair["filename"][index].tolist()
            sessionName = jobName
            windowName = "cosmic-debug"
            todo = "ds10 {} -f1 --name={} --width=200 --height=200".\
                     format(" ".join(allOutputPath), sessionName+"--"+windowName)
            os.system('tmux new-session -s {sessionName} -d >/dev/null 2>&1'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('tmux kill-window -t {sessionName}:{windowName} >/dev/null 2>&1'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('tmux new-window -t {sessionName} -n {windowName}'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('sleep 1')
            os.system('tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                    format(sessionName=sessionName, windowName=windowName, todo=todo))
    #endregion: do cosmic
    #region: ==========step: 3 make master dark and subdark================
    if start<4 and end>2:
        thisConfig = configs["parameters"].get("masterDark", {})
        aveCosmicConfig = copy.copy(configs["parameters"].get("aveCosmic", {}))
        if thisConfig["cmdChangeFile"]:
            thisConfig["rerun"] = True
            aveCosmicConfig["rerun"] = True
        stepName = "subdark"
        stopBeforeIt = thisConfig.get("stopBeforeIt", False)
        if stopBeforeIt:
            raise logger.Exception("stop before {}".format(stepName))
        #!! ave cosmic for dark
        updateConfigsByType(darks[0], aveCosmicConfig)
        cosmic.avecosmic([os.path.join(outputPath, "c"+each+".fits") for each in darks], aveCosmicConfig)
        ds10Debug = thisConfig.get("ds10Debug", False)
        #!! make master dark 
        dark.masterDark([os.path.join(outputPath, "cc"+each+".fits") for each in darks], thisConfig)
        #!! fix bad pixel
        dark.fixpixel(os.path.join(outputPath, "masterDark_notclean.fits"), thisConfig)
        #!! sub dark
        thisConfig = configs["parameters"].get("subDark", {})
        if thisConfig["cmdChangeFile"]:
            thisConfig["rerun"] = True
        ds10DebugSubDark = thisConfig.get("ds10Debug", False)
        todoList = scis + arcs + twilights
        fileNameDatePair = dark.subDark([os.path.join(outputPath, "c"+each+".fits") for each in todoList], thisConfig)

        if ds10Debug:
            allOutputPath = [os.path.join(outputPath, "masterDark_notclean.fits"),
                             os.path.join(outputPath, "masterDark.fits"),]
            sessionName = jobName
            windowName = "masterDark-debug"
            todo = "ds10 {} -a --name={} --width=200 --height=200".\
                     format(" ".join(allOutputPath), sessionName+"--"+windowName)
            os.system('tmux new-session -s {sessionName} -d >/dev/null 2>&1'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('tmux kill-window -t {sessionName}:{windowName} >/dev/null 2>&1'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('tmux new-window -t {sessionName} -n {windowName}'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('sleep 1')
            os.system('tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                    format(sessionName=sessionName, windowName=windowName, todo=todo))
        if ds10DebugSubDark:
            dtype = np.dtype([(str("filename"), "U100"), (str("datetime"), "U100")])
            fdPair = np.array(fileNameDatePair, dtype=dtype)
            index = np.argsort(fdPair["datetime"])
            allOutputPath = fdPair["filename"][index].tolist()
            temp = [(each,os.path.join(outputPath,
                     os.path.basename(each)[1:]))
                         for each in allOutputPath]
            allOutputPath = [eachSub for each in temp for eachSub in each]
            sessionName = jobName
            windowName = "subdark-debug"
            todo = "ds10 {} -f1 --name={} --width=200 --height=200".\
                    format(" ".join(allOutputPath), sessionName+"--"+windowName)
            os.system('tmux new-session -s {sessionName} -d >/dev/null 2>&1'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('tmux kill-window -t {sessionName}:{windowName} >/dev/null 2>&1'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('tmux new-window -t {sessionName} -n {windowName}'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('sleep 1')
            os.system('tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                    format(sessionName=sessionName, windowName=windowName, todo=todo))
    #endregion: make master dark and subdark
    #region: ==========step: 4 trace===================
    if start<5 and end>3:
        #!! make aveCosmic for flat
        thisConfig = configs["parameters"].get("aveCosmic", {})
        stepName = "trace"
        stopBeforeIt = thisConfig.get("stopBeforeIt", False)
        if stopBeforeIt:
            raise logger.Exception("stop before {}".format(stepName))
        ds10Debug = thisConfig.get("ds10Debug", False)
        thisConfig.update({"name":"twilights"})
        if thisConfig["cmdChangeFile"]:
            thisConfig["rerun"] = True
        fileNameDatePair,combineFile = cosmic.avecosmic([os.path.join(outputPath, "dc"+each+".fits") for each in twilights], thisConfig)
        cosmic.fixpixel(combineFile, thisConfig)
        if ds10Debug:
            dtype = np.dtype([(str("filename"), "U100"), (str("datetime"), "U100")])
            fdPair = np.array(fileNameDatePair, dtype=dtype)
            index = np.argsort(fdPair["datetime"])
            allOutputPath = fdPair["filename"][index].tolist()
            temp = [(os.path.join(outputPath,
                     os.path.basename(each)[1:]), each)
                         for each in allOutputPath]
            allOutputPath = [eachSub for each in temp for eachSub in each]
            sessionName = jobName
            windowName = "aveCosmic-debug"
            todo = "ds10 {} -f1 --name={} --width=200 --height=200".\
                    format(" ".join(allOutputPath), sessionName+"--"+windowName)
            os.system('tmux new-session -s {sessionName} -d >/dev/null 2>&1'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('tmux kill-window -t {sessionName}:{windowName} >/dev/null 2>&1'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('tmux new-window -t {sessionName} -n {windowName}'.\
                    format(sessionName=sessionName, windowName=windowName))
            os.system('sleep 1')
            os.system('tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                    format(sessionName=sessionName, windowName=windowName, todo=todo))
        #!! traceset
        thisConfig = configs["parameters"].get("traceset", {})
        thisConfig.update({"fiberInfo":configs["fibermap"]})
        if thisConfig["cmdChangeFile"]:
            thisConfig["rerun"] = True
        ds10Debug = thisConfig.get("ds10Debug", False)
        extract.makeTraceSet(os.path.join(outputPath, "masterFlat.fits"), configs=thisConfig)
        #!! debugOptions
        todoList = []
        sessionName = jobName
        windowName = "traceset-masterflat-debug"
        todo = "cd {};ds10 {} -f1 --name={} --width=200 --height=200 --regions={}".\
                format(outputPath, "c_all.txt", sessionName+"--"+windowName,
                       "masterFlat.reg")
        todoList.append('tmux new-session -s {sessionName} -d >/dev/null 2>&1'.\
                format(sessionName=sessionName, windowName=windowName))
        todoList.append('tmux kill-window -t {sessionName}:{windowName} >/dev/null 2>&1'.\
                format(sessionName=sessionName, windowName=windowName))
        todoList.append('tmux new-window -t {sessionName} -n {windowName}'.\
                format(sessionName=sessionName, windowName=windowName))
        todoList.append('sleep 1')
        todoList.append('tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                format(sessionName=sessionName, windowName=windowName, todo=todo))
        outfile = os.path.join(outputPath, "_cmdDebugTrace")
        with open(outfile, "w") as f:
            f.write("\n".join(todoList))
        os.chmod(outfile,stat.S_IRWXU)
        if ds10Debug or thisConfig.get("stopForDebug",False) or thisConfig.get("cmdDebugTrace", False):
            for each in todoList:
                os.system(each)
        if thisConfig.get("stopForDebug",False) or thisConfig.get("cmdDebugTrace", False):
            raise logger.Exception("Stop to look at {}/tracing.fig".format(outputPath))
    #endregion: step:4 trace
    #region: ==========step: 5 extract
    if start<6 and end>4:
        thisConfig = configs["parameters"].get("extract", {})
        stepName = "extract"
        stopBeforeIt = thisConfig.get("stopBeforeIt", False)
        if stopBeforeIt:
            raise logger.Exception("stop before {}".format(stepName))
        parallel = thisConfig.get("parallel", False)
        todoList = ["dc"+each+".fits" for each in scis+arcs]
        todoList += ["masterFlat.fits"]
        todoList += ["cdc"+each+".fits" for each in twilights]
        if not parallel:
            thisConfig.update({"fiberInfo":configs["fibermap"]})
            for eachFile in todoList:
                extract.imageExtract(os.path.join(outputPath, eachFile), configs=thisConfig)
        else:
            count = multiprocessing.cpu_count()-1
            pool = multiprocessing.Pool(processes=count)
            originConfig = thisConfig
            originConfig.pop("logger")
            originConfig.pop("teeStderr")
            jobList=[]
            logger.p("Do extractDirect in parallel: {}".format(todoList))
            for eachFile in todoList:
                thisConfig = copy.deepcopy(originConfig)
                thisLogger = {"warningFile":thisConfig["warningFile"],
                              "errorFile":thisConfig["errorFile"],
                              "name":eachFile, "initErrorLog":False,
                              "logFile":[os.path.join(thisConfig["logPath"], "parallel-log",
                                                     eachFile+"-ExtractDirect.log"),
                                         thisConfig["logFile"]],
                              "initWarningLog":False,"initLog":False,
                              "renewLog":[True,False], "renewError":False, "renewWarning":False,
                              }
                    #logger = messager.Messager(name=loggerName, logFile=logFile,
                    #               warningFile=warningFile, errorFile=errorFile,
                    #               initLog=True, initErrorLog=True,
                    #               renew=True, warningClass=m2fsWarning, exceptionClass=m2fsError)
                thisConfig["logger"] = thisLogger
                thisConfig.update({"fiberInfo":configs["fibermap"]})
                #cosmic.lacosmic(os.path.join(outputPath, eachFile)+'.fits', thisConfig)
                jobList.append(
                        pool.apply_async(
                            extract.imageExtract,
                            (os.path.join(outputPath, eachFile),
                             thisConfig)))
            pool.close(); pool.join()
            temp = [each.get() for each in jobList]
        ds10Debug = thisConfig.get("ds10Debug", False)
        #!! debug options
        sessionName = jobName
        windowName = "extract-all-debug"
        todo = "cd {};ds10 {} -f1 --name={} --width=200 --height=200 --selfRegions --regions={}".\
                format(outputPath, "c_all.txt", sessionName+"--"+windowName, "masterFlat.reg")
        todoList = []
        todoList.append('tmux new-session -s {sessionName} -d >/dev/null 2>&1'.\
                format(sessionName=sessionName, windowName=windowName))
        todoList.append('tmux kill-window -t {sessionName}:{windowName} >/dev/null 2>&1'.\
                format(sessionName=sessionName, windowName=windowName))
        todoList.append('tmux new-window -t {sessionName} -n {windowName}'.\
                format(sessionName=sessionName, windowName=windowName))
        todoList.append('sleep 1')
        todoList.append('tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                format(sessionName=sessionName, windowName=windowName, todo=todo))

        windowName = "flux-all-debug"
        todo = "cd {};ds10 {} -f1 --name={} --width=200 --height=200 --selfRegions --regions={}".\
                format(outputPath, "c_all_flux.txt", sessionName+"--"+windowName, "masterFlat.reg")
        todoList.append('tmux kill-window -t {sessionName}:{windowName} >/dev/null 2>&1'.\
                format(sessionName=sessionName, windowName=windowName))
        todoList.append('tmux new-window -t {sessionName} -n {windowName}'.\
                format(sessionName=sessionName, windowName=windowName))
        todoList.append('sleep 1')
        todoList.append('tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                format(sessionName=sessionName, windowName=windowName, todo=todo))
        outfile = os.path.join(outputPath, "_cmdDebugExtract")
        with open(outfile, "w") as f:
            f.write("\n".join(todoList))
        os.chmod(outfile,stat.S_IRWXU)
        if ds10Debug or thisConfig.get("stopForDebug",False) or thisConfig.get("cmdDebugExtract", False):
            for each in todoList:
                os.system(each)

        if thisConfig.get("stopForDebug",False) or thisConfig.get("cmdDebugExtract", False):
            raise logger.Exception("Stop to look at {}/tracing.fig".format(outputPath))
    #endregion: step 5: extract
    #region: ==========step: 6 wavecalib
    if start<7 and end>5:
        global ff
        thisConfig = configs["parameters"].get("findWave", {})
        stepName = "findWave"
        stopBeforeIt = thisConfig.get("stopBeforeIt", False)
        if stopBeforeIt:
            raise logger.Exception("stop before {}".format(stepName))
        thisConfig.update({"fiberInfo":configs["fibermap"]})
        wavecalib.findWave(os.path.join(outputPath, "dc"+arcs[0]+".fits"), configs=thisConfig)
    #endregion: step 6: wavecalib
    #region: ==========step: 7 make flatfield
    if start<8 and end>6:
        thisConfig = configs["parameters"].get("makeFlatField", {})
        stepName = "flatfield"
        stopBeforeIt = thisConfig.get("stopBeforeIt", False)
        if stopBeforeIt:
            raise logger.Exception("stop before {}".format(stepName))
        thisConfig.update({"fiberInfo":configs["fibermap"]})
        inParallel = thisConfig.get("inParallel", False)
        parallel = thisConfig.get("parallel", False)
        todoList = [os.path.join(outputPath, "cdc"+each+".fits") for each in twilights]
        if inParallel:
            inParallelIndex = thisConfig["inParallelIndex"]
            flat.makeFlatField(todoList[inParallelIndex], configs=thisConfig)
            return True
        else:
            if not parallel:
                for eachFile in todoList:
                    flat.makeFlatField(eachFile, configs=thisConfig)
            else:
                jobList = []
                for i, eachFile in enumerate(todoList):
                    eachName = os.path.basename(eachFile).split(".fits")[0]
                    jobList.append(Popen(["m2fsReduce", "--inParallel", "--inParallelIndex={}".format(i),
                                          "--start=7", "--end=7", "--inParallelName={}".format(eachName+"-flatfield")]))
                for p in jobList:
                    p.wait()

        sessionName = jobName
        windowName = "flux-all-debug"
        debugTodoList = []
        todo = "cd {};ds10 c_all_flux.txt -f1 --name={} --width=200 --height=200".\
                format(outputPath, sessionName+"--"+windowName)
        tt = 'tmux new-session -s {sessionName} -d'.\
                format(sessionName=sessionName, windowName=windowName)
        debugTodoList.append(tt)
        tt = 'tmux kill-window -t {sessionName}:{windowName}'.\
                format(sessionName=sessionName, windowName=windowName)
        debugTodoList.append(tt)
        tt = 'tmux new-window -t {sessionName} -n {windowName}'.\
                format(sessionName=sessionName, windowName=windowName)
        debugTodoList.append(tt)
        os.system('sleep 1')
        tt = 'tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                format(sessionName=sessionName, windowName=windowName, todo=todo)
        debugTodoList.append(tt)

        windowName = "flatfield-debug"
        todo = "cd {};ds10 c_all_flatfield.txt --name={} --width=200 --height=200".\
                format(outputPath, sessionName+"--"+windowName)
        tt = 'tmux new-session -s {sessionName} -d'.\
                format(sessionName=sessionName, windowName=windowName)
        debugTodoList.append(tt)
        tt = 'tmux kill-window -t {sessionName}:{windowName}'.\
                format(sessionName=sessionName, windowName=windowName)
        debugTodoList.append(tt)
        tt = 'tmux new-window -t {sessionName} -n {windowName}'.\
                format(sessionName=sessionName, windowName=windowName)
        debugTodoList.append(tt)
        os.system('sleep 1')
        tt = 'tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                format(sessionName=sessionName, windowName=windowName, todo=todo)
        debugTodoList.append(tt)
        #!! debug options
        if thisConfig["cmdChangeFile"]:
            thisConfig["rerun"] = True
        debugCommand = flat.selectBestFlats(todoList, configs=thisConfig)

        sessionName = jobName
        windowName = "best-flat-debug"
        todo = "cd {};{} --name={} --width=200 --height=200".\
                format(outputPath, debugCommand, sessionName+"--"+windowName)
        tt = 'tmux new-session -s {sessionName} -d'.\
                format(sessionName=sessionName, windowName=windowName)
        debugTodoList.append(tt)
        tt = 'tmux kill-window -t {sessionName}:{windowName}'.\
                format(sessionName=sessionName, windowName=windowName)
        debugTodoList.append(tt)
        tt = 'tmux new-window -t {sessionName} -n {windowName}'.\
                format(sessionName=sessionName, windowName=windowName)
        debugTodoList.append(tt)
        os.system('sleep 1')
        tt = 'tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
                format(sessionName=sessionName, windowName=windowName, todo=todo)
        debugTodoList.append(tt)
        outfile = os.path.join(outputPath, "_cmdDebugBestFlat")
        with open(outfile, "w") as f:
            f.write("\n".join(debugTodoList))
        os.chmod(outfile,stat.S_IRWXU)

        if thisConfig.get("stopForDebug",False) or thisConfig.get("cmdDebugBestFlat", False):
            for each in debugTodoList:
                os.system(each)
            raise logger.Exception("debug best flat")
    #endregion: step 6: waveTweak
    #region: ==========step: 8 waveTweak
    if start<9 and end>7:
        thisConfig = configs["parameters"].get("skyWaveTweak", {})
        stepName = "waveTweak"
        stopBeforeIt = thisConfig.get("stopBeforeIt", False)
        if stopBeforeIt:
            raise logger.Exception("stop before {}".format(stepName))
        thisConfig.update({"fiberInfo":configs["fibermap"]})
        parallel = thisConfig.get("parallel", False)
        inParallel = thisConfig.get("inParallel", False)
        todoList = [os.path.join(outputPath, "dc"+each+".fits") for each in scis]
        if inParallel:
            inParallelIndex = thisConfig["inParallelIndex"]
            wavecalib.skyWaveTweak(todoList[inParallelIndex], configs=thisConfig)
            return True
        else:
            if not parallel:
                for eachFile in todoList:
                    wavecalib.skyWaveTweak(eachFile, configs=thisConfig)
            else:
                jobList = []
                for i, eachFile in enumerate(todoList):
                    eachName = os.path.basename(eachFile).split(".fits")[0]
                    jobList.append(Popen(["m2fsReduce", "--inParallel", "--inParallelIndex={}".format(i),
                                          "--start=8", "--end=8", "--inParallelName={}".format(eachName+"-waveTweak")]))
                for p in jobList:
                    p.wait()
    #endregion: step 6: waveTweak
    #region: testForStack
    if 0:
        thisConfig = configs["parameters"].get("skySub", {})
        thisConfig.update({"fiberInfo":configs["fibermap"]})
        todoList = [os.path.join(outputPath, "dc"+each+".fits") for each in scis]
        stack.testRemap(todoList, configs=thisConfig)
    #endregion: testForStack
    #region: ==========step: 9 sky subtraction
    if start<10 and end>8:
        thisConfig = configs["parameters"].get("skySub", {})
        if not os.path.exists("data"):
            try:
                shutil.copytree(os.path.join(codePath, "data"), "data")
            except Exception as e:
                print(e)
        stepName = "subtraction"
        stopBeforeIt = thisConfig.get("stopBeforeIt", False)
        if stopBeforeIt:
            raise logger.Exception("stop before {}".format(stepName))
        thisConfig.update({"fiberInfo":configs["fibermap"]})
        parallel = thisConfig.get("parallel", False)
        inParallel = thisConfig.get("inParallel", False)
        fileStart = thisConfig.get("fileStart", 0)
        fileEnd = thisConfig.get("fileEnd",None)
        fileSlice = slice(fileStart, fileEnd)
        todoList = [os.path.join(outputPath, "dc"+each+".fits") for each in scis[fileSlice]]
        maxParallel = configs['parameters'].get('maxParallel', 4)
        if inParallel:
            inParallelIndex = thisConfig["inParallelIndex"]
            skysub.makeSkySub(todoList[inParallelIndex], configs=thisConfig)
            return True
        else:
            if not parallel:
                for eachFile in todoList[:]:
                    skysub.makeSkySub(eachFile, configs=thisConfig)
            else:
                piece = len(todoList)//maxParallel
                todoLists = [todoList[i*maxParallel:(i+1)*maxParallel] for i in range(piece)]
                _ = todoList[(i+1)*maxParallel:]
                if _:
                    todoLists.append(_)
                for todoList in todoLists:
                    jobList = []
                    for i, eachFile in enumerate(todoList):
                        eachName = os.path.basename(eachFile).split(".fits")[0]
                        if thisConfig.get("cmdDebugFlatResult", False):
                            jobList.append(Popen(["m2fsReduce", "--debugFlatResult", "--inParallel", "--inParallelIndex={}".format(i),
                                              "--start=9", "--end=9", "--inParallelName={}".format(eachName+"-skySub")]))
                        else:
                            jobList.append(Popen(["m2fsReduce", "--inParallel", "--inParallelIndex={}".format(i),
                                              "--start=9", "--end=9", "--inParallelName={}".format(eachName+"-skySub")]))
                    for p in jobList:
                        p.wait()
        with open(os.path.join(outputPath, "_cmdDebugFlatResult-gen"), "w") as f:
            f.write("cat _cmdDebugFlatResult_* >_cmdDebugFlatResult".format(outputPath))
        os.chmod(os.path.join(outputPath, "_cmdDebugFlatResult-gen"),stat.S_IRWXU)
        if thisConfig.get("stopForDebug"):
            raise logger.Exception("stopForDebug.")
        if thisConfig.get("cmdDebugFlatResult", False):
            raise logger.Exception("debug flat result")
    #endregion: step 6: waveTweak
    #region: step 10 stack
    if start<11 and end>9:
        thisConfig = configs["parameters"].get("stack", {})
        thisConfig.update({"fiberInfo":configs["fibermap"]})
        todoList = [os.path.join(outputPath, "dc"+each+".fits") for each in scis]
        stack.stackOneNight(todoList, configs=thisConfig)
    #endregion: stepL 10 stack
    #region: step 11 extract stack 1d
    if start<12 and end>10:
        thisConfig = configs["parameters"].get("extract1D", {})
        directStackExtractFile = os.path.join(outputPath, "final-directStackExtractModel.fits")
        if thisConfig.get("reExtract", False) or not os.path.exists(directStackExtractFile):
            directStackFile = os.path.join(outputPath, "final-directStack.fits")
            thisConfig = configs["parameters"].get("extract", {})
            thisConfig["rerun"] = True
            thisConfig.update({"fiberInfo":configs["fibermap"]})
            thisConfig["logger"] = logger

            extract.imageExtract(directStackFile, configs=thisConfig)
            thisConfig = configs["parameters"].get("skyWaveTweak", {})
            thisConfig.update({"fiberInfo":configs["fibermap"]})
            wavecalib.skyWaveTweak(directStackFile, configs=thisConfig)

        thisConfig = configs["parameters"].get("extract1D", {})
        thisConfig.update({"fiberInfo":configs["fibermap"]})
        thisConfig["logger"] = logger
        thisConfig["outputPath"] = outputPath
        thisConfig.update({"fiberInfo":configs["fibermap"]})
        stack.extract1D(configs=thisConfig)

    #endregion: extract stack 1d
    logger.p("start:{} end:{}, all finished!!!".format(start, end))