from __future__ import (absolute_import, division, print_function, unicode_literals)
import numpy as np
import collections
OD = collections.OrderedDict
import os
import sys
import matplotlib.pyplot as plt
from matplotlib import cm
import numpy
from numpy.polynomial import Legendre
from . import de
from .image import ccdproc
from .image import specLib
from ..utils import macro
from ..utils import messager
from ..utils import configTools
import scipy.interpolate as interp
import scipy
import pickle
from scipy.optimize import curve_fit
import glob
import multiprocessing
import copy
from ..DS10 import ds9
import ipdb
import stat

from . import m2fsConfigs
from . import extract
from . import wavecalib
from . import skysub
from .zscale import zscale
import shutil
import astropy.io.fits as pyfits
from astropy import wcs

def reshapeRemap(image, N):
    ny, nx = image.shape
    assert ny%N==0
    width = int(ny/N)
    xstart = np.zeros(N, dtype=np.int)
    xend = np.ones(N, dtype=np.int) * nx
    xlength = np.zeros(N, dtype=np.int)

    for i in range(N):
        cut = image[i*width:(i+1)*width].sum(axis=0)
        xstart[i] = np.where(cut>0)[0][0]
        xend[i] = np.where(cut>0)[0][-1]
        xlength[i] = xend[i] - xstart[i] + 1
    length = xlength.max()
    result = np.zeros((ny, length))
    for i in range(N):
        result[i*width:(i+1)*width, :xlength[i]] = image[i*width:(i+1)*width, xstart[i]:xend[i]+1]

    return result, xstart, xend, xlength

def reshapeRemapWithX(image, N, xstart, xend, xlength):
    ny, nx = image.shape
    assert ny%N==0
    width = int(ny/N)
    length = xlength.max()
    result = np.zeros((ny, length))
    for i in range(N):
        result[i*width:(i+1)*width, :xlength[i]] = image[i*width:(i+1)*width, xstart[i]:xend[i]+1]

    return result

def inv_reshapeRemap(image, N, xstart, xend, xlength, nx):
    ny, nxNew = image.shape
    assert ny%N==0
    width = int(ny/N)
    length = xlength.max()
    result = np.zeros((ny, nx))
    for i in range(N):
        result[i*width:(i+1)*width, xstart[i]:xend[i]+1]= image[i*width:(i+1)*width, :xlength[i]]
    return result

def stackExtract(cut, cutStd, profile, halfwidth=5):
    width = halfwidth*2+1
    ny, nx = cut.shape
    assert cut.shape==profile.shape==cutStd.shape
    assert ny%width==0, "{}%{}={}".format(ny, width, ny%width)
    nFiber = int(ny/width)
    fitExtract = np.zeros((nFiber, nx))
    sumExtract = np.zeros((nFiber, nx))
    for i in range(nFiber):
        todoImage = cut[i*width:(i+1)*width]
        todoStd = cutStd[i*width:(i+1)*width]
        todoProfile = profile[i*width:(i+1)*width]
        badPixel = ~np.isfinite(todoImage)
        todoImage[badPixel]=0
        todoProfile[badPixel]=0
        zeroStdMask = (todoStd==0)|(~np.isfinite(todoStd))
        todoStd[zeroStdMask] = np.inf
        todoImage[zeroStdMask] = 0
        fitExtract[i,:] =\
            ((todoImage*todoProfile/todoStd**2).sum(axis=0)/(todoProfile**2/todoStd**2).sum(axis=0))[:]
        sumExtract[i,:] = (todoImage.sum(axis=0)/todoProfile.sum(axis=0))[:]
    return {"fit":fitExtract, "sum":sumExtract}

def extract1D(configs={}):
    #region: setups
    outputFileName= 'stack-final-1d-remap.fits'
    configs.update({"config2show":["rerun"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return
    #endregion: setups

    reExtract = configs.get("reExtract", False)
    filename = os.path.join(outputPath, "final-stack-info.dat")
    with open(filename, "rb") as f:
        stackInfo = pickle.load(f)
    localSkynames = stackInfo["localSkynames"]
    #region: direct stack
    #!! first deal with direct stack data
    filename = os.path.join(outputPath, "final-directStack.fits")
    obj = pyfits.open(filename)
    fimage = obj[0].data
    fstd = obj[-1].data
    todoData = [obj[index].data for index in range(3, 3+len(localSkynames)+1)]
    todoHeaders = [obj[index].header for index in range(3, 3+len(localSkynames)+1)]
    todoNames = localSkynames+["smoothSky"]
    objImage = obj
    objHead = obj[0:3]
    objFoot = obj[3+len(localSkynames)+1:]

    filename = os.path.join(outputPath, "final-directStackMap2D.fits")
    obj = pyfits.open(filename)
    map2D = obj[0].data

    filename = os.path.join(outputPath, "final-directStackExtractModel.fits")
    obj = pyfits.open(filename)
    fExtractResidual = obj[3].data
    fExtractProfile = obj[-1].data

    filename = os.path.join(outputPath, "final-directStackFlux.fits")
    obj = pyfits.open(filename)
    flux = obj[0].data
    fluxStd = obj[1].data
    yset = obj["xcensmooth"].data

    waveset, coeffs, ysetSmooth, badFiber, debugInfo = wavecalib.readWaveSet(outputPath,
            name="final-directStack", debugInfo=2)
    rms = debugInfo["rms"].flatten()#!! all fiber that have been traceseted
    finalGoodFiber = np.isfinite(rms)#!! good for wavelength
    goodFiber = badFiber<1
    fiberIndex = np.where(goodFiber)[0]
    fiberIndex = fiberIndex[finalGoodFiber]#!! origin index in 0~127
    allSkyLines = skysub.readAllSkyData(".")
    skys = allSkyLines["lambda"]

    finalGoodFiberIndex = np.where(finalGoodFiber)[0]
    waveset = waveset[finalGoodFiber]
    flux = flux[finalGoodFiber]
    fluxStd = fluxStd[finalGoodFiber]
    yset = yset[finalGoodFiber]
    fiberInfo = configs["fiberInfo"]
    fiberInfo = fiberInfo[fiberIndex]#!! origin index in 0~127
    nFiber = finalGoodFiber.sum()
    names = ["_".join([str(fiberIndex[i]),
                       fiberInfo[i]["fiber"].decode('UTF-8'),
                       fiberInfo[i]["id"].decode('UTF-8')])
                for i in range(nFiber)]

    fimageCut = specLib.map2DCutFixWidthSingle(fimage, yset)["cutFlux"]
    fimageResidualCut = specLib.map2DCutFixWidthSingle(fExtractResidual, yset)["cutFlux"]
    fstdCut = specLib.map2DCutFixWidthSingle(fstd, yset)["cutFlux"]
    fExtractProfileCut = specLib.map2DCutFixWidthSingle(fExtractProfile, yset)["cutFlux"]
    rawResult = stackExtract(fimageCut, fstdCut, fExtractProfileCut)
    residualResult = stackExtract(fimageResidualCut, fstdCut, fExtractProfileCut)
    #!! debug outputs
    debugDir = os.path.join(outputPath, "final-directStack-debug-raw-flux")
    if not os.path.exists(debugDir):
        os.mkdir(debugDir)
    if rerun or not os.path.exists(os.path.join(debugDir, names[0]+".fig")):
        print("output ", debugDir)
        for i in range(nFiber):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.plot(waveset[i], flux[i], c="green", label="extract")
            ax.plot(waveset[i], rawResult["fit"][i], c="red", label="fit")
            ax.plot(waveset[i], rawResult["sum"][i], c="blue", label="sum")
            ax.plot(waveset[i], residualResult["fit"][i], c="red",  label="res fit")
            ax.plot(waveset[i], residualResult["sum"][i], c="blue", label="res sum", lw=0.4)
            ax.plot(waveset[i], fluxStd[i], "--", c="pink", label="flux std")
            xmin = waveset[i][0]
            xmax = waveset[i][-1]
            thisSkys = skys[(skys>xmin)&(skys<xmax)]
            thisYmax = flux[i].max()
            for eachsky in thisSkys:
                ax.plot([eachsky, eachsky], [0, thisYmax], "--", c="black", lw=0.3)
            ax.set_title(names[i])
            ax.legend()
            de.d(fig, os.path.join(debugDir, names[i]+".fig"))
            plt.close(fig)
    #!! do it for all smooth sky result

    blocks = skysub.readSkyBlock(".")["start"]
    c = cm.get_cmap("rainbow")
    c = c(np.linspace(0,1,len(todoNames)))

    plotResult = {}
    plotResult["waveset"] = waveset
    plotResult["todoNames"] = todoNames
    plotResult["names"] = names
    plotResult["flux"] = flux
    plotResult["std"] = fluxStd
    plotResult["skys"] = skys
    for eachdata, eachheader, eachname in zip(todoData, todoHeaders, todoNames):
        thisFluxContinuum = np.zeros(yset.shape)
        thisFluxContinuumMedian = np.zeros(yset.shape)
        plotResult[eachname]={}
        fimageCut = specLib.map2DCutFixWidthSingle(eachdata, yset)["cutFlux"]
        fitFlux = stackExtract(fimageCut, fstdCut, fExtractProfileCut)["fit"]
        plotResult[eachname]["continuum"] = thisFluxContinuum
        plotResult[eachname]["continuumMedian"] = thisFluxContinuumMedian
        plotResult[eachname]["flux"] = fitFlux
        for i in range(nFiber):
            thisy = fitFlux[i]
            thisstd = fluxStd[i]
            badmask = ~np.isfinite(thisy)
            thisy[badmask] = 0
            thisstd[badmask]=np.inf
            fitObj = Legendre.fit(waveset[i], thisy, 5, w=1/thisstd**2)
            thisContinuum = fitObj(waveset[i])
            thisFluxContinuum[i,:] = thisContinuum[:]
            thisFluxContinuumMedian[i,:] = scipy.ndimage.filters.median_filter(fitFlux[i], 100)[:]

    debugDir = os.path.join(outputPath, "final-directStack-debug-localskys")
    if not os.path.exists(debugDir):
        os.mkdir(debugDir)
    #!! plot example here
    if rerun or not os.path.exists(os.path.join(debugDir, names[0]+".fig")):
        print("output ", debugDir)
        with open(os.path.join(outputPath, "final-directStack.dat"), "wb") as f:
            pickle.dump(plotResult, f)
        for i in range(nFiber):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            xmin = waveset[i][0]
            xmax = waveset[i][-1]
            thisSkys = skys[(skys>xmin)&(skys<xmax)]
            thisYmax=0
            ax.plot(waveset[i], fluxStd[i], "--", c="pink", label="flux std")
            fig = plt.figure()
            ax = fig.add_subplot(111)
            for j,eachname in enumerate(todoNames):
                if j==len(todoNames)-1:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c="black", label=eachname+" flux", lw=0.7)
                    ax.plot(waveset[i], plotResult[eachname]["continuum"][i], "--",
                            c="black", lw=0.7)
                else:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c=c[j], label=eachname+" flux")
                    ax.plot(waveset[i], plotResult[eachname]["continuum"][i], "--", c=c[j])
                thisYmax = np.max([plotResult[eachname]["flux"][i].max(), thisYmax])
            for eachsky in thisSkys:
                ax.plot([eachsky, eachsky], [0, thisYmax], "--", c="black", lw=0.3)
            ax.set_title(names[i])
            ax.legend()
            de.d(fig, os.path.join(debugDir, names[i]+".fig"))
            plt.close(fig)
        for i in range(nFiber):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            xmin = waveset[i][0]
            xmax = waveset[i][-1]
            thisSkys = skys[(skys>xmin)&(skys<xmax)]
            thisYmax=0
            ax.plot(waveset[i], fluxStd[i], "--", c="pink", label="flux std")
            fig = plt.figure()
            ax = fig.add_subplot(111)
            for j,eachname in enumerate(todoNames):
                if j==len(todoNames)-1:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c="black", label=eachname+" flux", lw=0.7)
                    ax.plot(waveset[i], plotResult[eachname]["continuumMedian"][i], "--",
                            c="black", lw=0.7)
                else:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c=c[j], label=eachname+" flux")
                    ax.plot(waveset[i], plotResult[eachname]["continuumMedian"][i], "--", c=c[j])
                thisYmax = np.max([plotResult[eachname]["flux"][i].max(), thisYmax])
            for eachsky in thisSkys:
                ax.plot([eachsky, eachsky], [0, thisYmax], "--", c="black", lw=0.3)
            ax.set_title(names[i])
            ax.legend()
            de.d(fig, os.path.join(debugDir, names[i]+"_median_filters.fig"))
            plt.close(fig)
    cleanFile = os.path.join(outputPath, "final-directStack-clean.fits")
    if rerun or not os.path.exists(cleanFile):
        print("output ", cleanFile)
        bzeros = np.zeros(fimage.shape)
        middleExts = []
        for i, (eachdata, eachheader, eachname) in enumerate(zip(todoData, todoHeaders, todoNames)):
            ext0 = pyfits.ImageHDU(eachdata, header=eachheader)
            fluxZeros = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxZeros[finalGoodFiberIndex, :] = plotResult[eachname]["continuum"][:,:]
            fluxOnes = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxOnes[finalGoodFiberIndex, :] = 1
            thisContinuumImage = extract.map2DModel(bzeros, map2D, fluxZeros, fluxOnes)
            resultImage = eachdata - thisContinuumImage
            ext1 = pyfits.ImageHDU(resultImage,
                    header=pyfits.Header({"OBJECT":eachheader["OBJECT"] + " clean poly"}))
            fluxZeros = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxZeros[finalGoodFiberIndex, :] = plotResult[eachname]["continuumMedian"][:,:]
            fluxOnes = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxOnes[finalGoodFiberIndex, :] = 1
            thisContinuumImage = extract.map2DModel(bzeros, map2D, fluxZeros, fluxOnes)
            resultImage = eachdata - thisContinuumImage
            ext2 = pyfits.ImageHDU(resultImage,
                    header=pyfits.Header({"OBJECT":eachheader["OBJECT"] + " clean median"}))
            middleExts = middleExts + [ext0, ext1, ext2]
        outputHDUList = objHead + middleExts + objFoot
        fits = pyfits.HDUList(outputHDUList)
        fits.writeto(cleanFile, clobber='True')
        #!! make cut file for single object
        print("output final-directStack-clean-cut")
        cleanCutFilePath = os.path.join(outputPath, "final-directStack-clean-cut")
        os.makedirs(cleanCutFilePath, exist_ok=True)
        toCutFrames = outputHDUList[:32]
        for i in range(nFiber):
            thisname = names[i]
            thispath = os.path.join(cleanCutFilePath, thisname+".fits")
            thisSingleCutResult = np.empty((0,yset.shape[1]))
            thisHeader = pyfits.Header()
            for j,eachExt in enumerate(toCutFrames):
                eachdata = eachExt.data
                thisStartIndex = thisSingleCutResult.shape[0]
                thiscut = specLib.map2DCutFixWidthSingle(eachdata, yset[i:i+1,:])["cutFlux"]
                thiscut[~np.isfinite(thiscut)]=0
                thisSingleCutResult = np.concatenate((thisSingleCutResult, thiscut))
                thisEndIndex = thisSingleCutResult.shape[0]
                thisHeader["start{}".format(j)] = str(thisStartIndex)
                thisHeader["end{}".format(j)] = str(thisEndIndex)
                thisHeader["object{}".format(j)] = str(eachExt.header.get("object", "???"))
            thisHDU = pyfits.PrimaryHDU(thisSingleCutResult, header = thisHeader)
            waveHDU = pyfits.ImageHDU(yset[i])
            thisfits = pyfits.HDUList([thisHDU, waveHDU])
            thisfits.writeto(thispath, clobber='True')

    #endregion: direct stack

    #!! deal with remap
    #region: remap
    remapStackFile  = os.path.join(outputPath, "final-all-remap.fits")
    yset1D = stackInfo["remapYsetAll"]
    xwidth = stackInfo["xset1DLength"]
    badFiber = np.zeros(yset1D.shape)
    N = len(yset1D)
    yset = np.tile(yset1D, (xwidth, 1)).T
    obj = pyfits.open(remapStackFile)
    fimageOrigin = obj["sci"].data
    fstdOrigin = obj["std"].data
    ny,nx = fimageOrigin.shape
    width = ny/N
    assert ny%N==0
    assert int(width)%2==1
    halfwidth = int((width-1)/2)
    fstd, xstart, xend, xlength = reshapeRemap(fstdOrigin, N)
    fimage = reshapeRemapWithX(fimageOrigin, N, xstart, xend, xlength)

    imageNew = inv_reshapeRemap(fimage, N, xstart, xend, xlength, nx)
    assert fimageOrigin.shape==imageNew.shape
    yset = np.tile(yset1D, (fimage.shape[1], 1)).T

    waveset1D = np.linspace(stackInfo['xmin'], stackInfo['xmax'], stackInfo['xset1DLength'])
    waveset = np.ones(yset.shape)*np.nan
    for i in range(N):
        waveset[i,:xlength[i]] = waveset1D[xstart[i]:xend[i]+1]

    thisConfig = {"maxMuDiff":0.1, "offset":0, "outputPath":outputPath, "logger":logger,
                  "yset":yset, "badFiber":badFiber, "onlyUseSci":True, "genReg":False,
                  "sci":pyfits.PrimaryHDU(fimage, header=pyfits.Header({"OBJECT":"finalStackRemap"})),
                  "std":pyfits.PrimaryHDU(fstd, header=pyfits.Header({"OBJECT":"finalStackRemap"})),
                  "rerun":reExtract,
                 }
    extract.imageExtract(remapStackFile, configs=thisConfig)

    todoData = [reshapeRemapWithX(obj[index].data, N, xstart, xend, xlength)
                    for index in range(4, 4+len(localSkynames)+1)]
    todoHeaders = [obj[index].header for index in range(4, 4+len(localSkynames)+1)]
    todoNames = localSkynames+["smoothSky"]
    objHead = obj[0:4]
    objFoot = obj[4+len(localSkynames)+1:]

    filename = os.path.join(outputPath, "final-all-remapMap2D.fits")
    obj = pyfits.open(filename)
    map2D = obj[0].data

    filename = os.path.join(outputPath, "final-all-remapExtractModel.fits")
    obj = pyfits.open(filename)
    fimageResidual = obj[3].data
    fProfile = obj[-1].data

    filename = os.path.join(outputPath, "final-all-remapFlux.fits")
    obj = pyfits.open(filename)
    flux = obj[0].data
    fluxStd = obj[1].data
    yset = obj["xcensmooth"].data

    rawResult = stackExtract(fimage, fstd, fProfile, halfwidth=halfwidth)
    residualResult = stackExtract(fimageResidual, fstd, fProfile, halfwidth=halfwidth)

    #!! debug outputs
    debugDir = os.path.join(outputPath, "final-all-remap-debug-raw-flux")
    if not os.path.exists(debugDir):
        os.mkdir(debugDir)
    if rerun or not os.path.exists(os.path.join(debugDir, names[0]+".fig")):
        print("output ", debugDir)
        for i in range(nFiber):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.plot(waveset[i], flux[i], c="green", label="extract")
            ax.plot(waveset[i], rawResult["fit"][i], c="red", label="fit")
            ax.plot(waveset[i], rawResult["sum"][i], c="blue", label="sum")
            ax.plot(waveset[i], residualResult["fit"][i], c="red",  label="res fit")
            ax.plot(waveset[i], residualResult["sum"][i], c="blue", label="res sum", lw=0.4)
            ax.plot(waveset[i], fluxStd[i], "--", c="pink", label="flux std")
            xmin = waveset[i][0]
            xmax = waveset[i][-1]
            thisSkys = skys[(skys>xmin)&(skys<xmax)]
            thisYmax = flux[i].max()
            for eachsky in thisSkys:
                ax.plot([eachsky, eachsky], [0, thisYmax], "--", c="black", lw=0.3)
            ax.set_title(names[i])
            ax.legend()
            de.d(fig, os.path.join(debugDir, names[i]+".fig"))
            plt.close(fig)

    plotResult = {}
    plotResult["waveset"] = waveset
    plotResult["todoNames"] = todoNames
    plotResult["names"] = names
    plotResult["flux"] = flux
    plotResult["std"] = fluxStd
    plotResult["skys"] = skys
    for eachdata, eachheader, eachname in zip(todoData, todoHeaders, todoNames):
        thisFluxContinuum = np.zeros(yset.shape)
        thisFluxContinuumMedian = np.zeros(yset.shape)
        plotResult[eachname]={}
        fitFlux = stackExtract(eachdata, fstd, fProfile, halfwidth=halfwidth)["fit"]
        plotResult[eachname]["continuum"] = thisFluxContinuum
        plotResult[eachname]["continuumMedian"] = thisFluxContinuumMedian
        plotResult[eachname]["flux"] = fitFlux
        for i in range(nFiber):
            #xmin = waveset[i][0]
            #xmax = waveset[i][-1]
            #thisKnots=blocks[(blocks>xmin)&(blocks<xmax)]
            #fitObj=scipy.interpolate.LSQUnivariateSpline(waveset[i],fitFlux[i],thisKnots,w=1/fluxStd[i])
            goodWave = np.isfinite(waveset[i])
            thiswave = waveset[i][goodWave]
            thisy = fitFlux[i][goodWave]
            thisstd = fluxStd[i][goodWave]
            badmask = (~np.isfinite(thisy))
            thisy[badmask] = 0
            thisstd[badmask]=np.inf
            try:
                fitObj = Legendre.fit(thiswave, thisy, 5, w=1/thisstd**2)
            except Exception as e:
                print(e)
                ipdb.set_trace()
            thisContinuum = fitObj(waveset[i])
            thisFluxContinuum[i,:] = thisContinuum[:]
            thisFluxContinuumMedian[i,:] = scipy.ndimage.filters.median_filter(fitFlux[i], 100)[:]

    debugDir = os.path.join(outputPath, "final-all-remap-debug-localskys")
    if not os.path.exists(debugDir):
        os.mkdir(debugDir)
    if rerun or not os.path.exists(os.path.join(debugDir, names[0]+".fig")):
        print("output ", debugDir)
        with open(os.path.join(outputPath, "final-all-remap.dat"), "wb") as f:
            pickle.dump(plotResult, f)
        for i in range(nFiber):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            xmin = waveset[i][0]
            xmax = waveset[i][-1]
            thisSkys = skys[(skys>xmin)&(skys<xmax)]
            thisYmax=0
            ax.plot(waveset[i], fluxStd[i], "--", c="pink", label="flux std")
            fig = plt.figure()
            ax = fig.add_subplot(111)
            for j,eachname in enumerate(todoNames):
                if j==len(todoNames)-1:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c="black", label=eachname+" flux", lw=0.7)
                    ax.plot(waveset[i], plotResult[eachname]["continuum"][i], "--",
                            c="black", lw=0.7)
                else:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c=c[j], label=eachname+" flux")
                    ax.plot(waveset[i], plotResult[eachname]["continuum"][i], "--", c=c[j])
                thisYmax = np.max([plotResult[eachname]["flux"][i].max(), thisYmax])
            for eachsky in thisSkys:
                ax.plot([eachsky, eachsky], [0, thisYmax], "--", c="black", lw=0.3)
            ax.set_title(names[i])
            ax.legend()
            de.d(fig, os.path.join(debugDir, names[i]+".fig"))
            plt.close(fig)
        for i in range(nFiber):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            xmin = waveset[i][0]
            xmax = waveset[i][-1]
            thisSkys = skys[(skys>xmin)&(skys<xmax)]
            thisYmax=0
            ax.plot(waveset[i], fluxStd[i], "--", c="pink", label="flux std")
            fig = plt.figure()
            ax = fig.add_subplot(111)
            for j,eachname in enumerate(todoNames):
                if j==len(todoNames)-1:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c="black", label=eachname+" flux", lw=0.7)
                    ax.plot(waveset[i], plotResult[eachname]["continuumMedian"][i], "--",
                            c="black", lw=0.7)
                else:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c=c[j], label=eachname+" flux")
                    ax.plot(waveset[i], plotResult[eachname]["continuumMedian"][i], "--", c=c[j])
                thisYmax = np.max([plotResult[eachname]["flux"][i].max(), thisYmax])
            for eachsky in thisSkys:
                ax.plot([eachsky, eachsky], [0, thisYmax], "--", c="black", lw=0.3)
            ax.set_title(names[i])
            ax.legend()
            de.d(fig, os.path.join(debugDir, names[i]+"_median_filters.fig"))
            plt.close(fig)

    cleanFile = os.path.join(outputPath, "final-all-remap-clean.fits")
    if rerun or not os.path.exists(cleanFile):
        print("output ", cleanFile)
        bzeros = np.zeros(fimage.shape)
        middleExts = []
        for i, (eachdata, eachheader, eachname) in enumerate(zip(todoData, todoHeaders, todoNames)):
            ext0 = pyfits.ImageHDU(
                    inv_reshapeRemap(eachdata, N, xstart, xend, xlength, nx),
                    header=eachheader)
            fluxZeros = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxZeros[finalGoodFiberIndex, :] = plotResult[eachname]["continuum"][:,:]
            fluxOnes = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxOnes[finalGoodFiberIndex, :] = 1
            thisContinuumImage = extract.map2DModel(bzeros, map2D, fluxZeros, fluxOnes)
            resultImageOrigin = eachdata - thisContinuumImage
            resultImage = inv_reshapeRemap(resultImageOrigin, N, xstart, xend, xlength, nx)
            ext1 = pyfits.ImageHDU(resultImage,
                    header=pyfits.Header({"OBJECT":eachheader["OBJECT"] + " clean poly"}))
            fluxZeros = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxZeros[finalGoodFiberIndex, :] = plotResult[eachname]["continuumMedian"][:,:]
            fluxOnes = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxOnes[finalGoodFiberIndex, :] = 1
            thisContinuumImage = extract.map2DModel(bzeros, map2D, fluxZeros, fluxOnes)
            resultImageOrigin = eachdata - thisContinuumImage
            resultImage = inv_reshapeRemap(resultImageOrigin, N, xstart, xend, xlength, nx)
            ext2 = pyfits.ImageHDU(resultImage,
                    header=pyfits.Header({"OBJECT":eachheader["OBJECT"] + " clean median"}))
            middleExts = middleExts + [ext0, ext1, ext2]
        outputHDUList = objHead + middleExts + objFoot
        fits = pyfits.HDUList(outputHDUList)
        fits.writeto(cleanFile, clobber='True')
        #!! make cut file for single object
        cleanCutFilePath = os.path.join(outputPath, "final-all-remap-clean-cut")
        os.makedirs(cleanCutFilePath, exist_ok=True)
        toCutFrames = outputHDUList[:32]
        print("output final-all-remap-clean-cut")
        width = halfwidth*2+1
        for i in range(nFiber):
            thisname = names[i]
            thispath = os.path.join(cleanCutFilePath, thisname+".fits")
            thisSingleCutResult = np.empty((0,toCutFrames[0].data.shape[1]))
            thisHeader = pyfits.Header()
            for j,eachExt in enumerate(toCutFrames):
                eachdata = eachExt.data
                thisStartIndex = thisSingleCutResult.shape[0]
                #thiscut = specLib.map2DCutFixWidthSingle(eachdata, yset[i:i+1,:])["cutFlux"]
                thiscut = eachdata[i*width:(i+1)*width]
                thiscut[~np.isfinite(thiscut)]=0
                thisSingleCutResult = np.concatenate((thisSingleCutResult, thiscut))
                thisEndIndex = thisSingleCutResult.shape[0]
                thisHeader["start{}".format(j)] = str(thisStartIndex)
                thisHeader["end{}".format(j)] = str(thisEndIndex)
                thisHeader["object{}".format(j)] = str(eachExt.header.get("object", "???"))
            thisHDU = pyfits.PrimaryHDU(thisSingleCutResult, header = thisHeader)
            waveHDU = pyfits.ImageHDU(yset[i])
            thisfits = pyfits.HDUList([thisHDU, waveHDU])
            thisfits.writeto(thispath, clobber='True')
    #endregion: remoap

    remapStackFile  = os.path.join(outputPath, "final-all-remap-smooth.fits")
    if not os.path.exists(remapStackFile):
        filename  = os.path.join(outputPath, "final-all-remap.fits")
        thisobj = pyfits.open(filename)
        resultList = []
        smoothSize=5
        sigma = 2
        for i, eachExt in enumerate(thisobj):
            thisdata = eachExt.data
            thisheader = eachExt.header
            thisKernelSigma = np.ones(thisdata.shape[1]) * sigma
            xx = np.arange(thisdata.shape[1])
            badmask = thisdata==0
            for k in range(thisdata.shape[0]):
                if thisheader.get("EXTNAME", "") == "std":
                    thisdata[k,:] = np.sqrt(specLib.variateConvolve(
                        xx, (thisdata**2)[k,:],
                        thisKernelSigma, smoothSize)) * 3
                else:
                    thisdata[k,:] = specLib.variateConvolve(
                        xx, thisdata[k,:],
                        thisKernelSigma, smoothSize)
            if thisheader.get("EXTNAME", "") == "std":
                thisdata[badmask]=0

            if i==0:
                resultList.append(pyfits.PrimaryHDU(thisdata, thisheader))
            else:
                resultList.append(pyfits.ImageHDU(thisdata, thisheader))
        hdulist = pyfits.HDUList(resultList)
        hdulist.writeto(remapStackFile, clobber=True)
    #region: remap smooth
    remapStackFile  = os.path.join(outputPath, "final-all-remap-smooth.fits")
    yset1D = stackInfo["remapYsetAll"]
    xwidth = stackInfo["xset1DLength"]
    badFiber = np.zeros(yset1D.shape)
    N = len(yset1D)
    yset = np.tile(yset1D, (xwidth, 1)).T
    obj = pyfits.open(remapStackFile)
    fimageOrigin = obj["sci"].data
    fstdOrigin = obj["std"].data
    ny,nx = fimageOrigin.shape
    width = ny/N
    assert ny%N==0
    assert int(width)%2==1
    halfwidth = int((width-1)/2)
    fstd, xstart, xend, xlength = reshapeRemap(fstdOrigin, N)
    fimage = reshapeRemapWithX(fimageOrigin, N, xstart, xend, xlength)

    imageNew = inv_reshapeRemap(fimage, N, xstart, xend, xlength, nx)
    assert fimageOrigin.shape==imageNew.shape
    yset = np.tile(yset1D, (fimage.shape[1], 1)).T

    waveset1D = np.linspace(stackInfo['xmin'], stackInfo['xmax'], stackInfo['xset1DLength'])
    waveset = np.ones(yset.shape)*np.nan
    for i in range(N):
        waveset[i,:xlength[i]] = waveset1D[xstart[i]:xend[i]+1]

    thisConfig = {"maxMuDiff":0.1, "offset":0, "outputPath":outputPath, "logger":logger,
                  "yset":yset, "badFiber":badFiber, "onlyUseSci":True, "genReg":False,
                  "sci":pyfits.PrimaryHDU(fimage, header=pyfits.Header({"OBJECT":"finalStackRemap"})),
                  "std":pyfits.PrimaryHDU(fstd, header=pyfits.Header({"OBJECT":"finalStackRemap"})),
                  "rerun":reExtract,
                 }
    extract.imageExtract(remapStackFile, configs=thisConfig)

    todoData = [reshapeRemapWithX(obj[index].data, N, xstart, xend, xlength)
                    for index in range(4, 4+len(localSkynames)+1)]
    todoHeaders = [obj[index].header for index in range(4, 4+len(localSkynames)+1)]
    todoNames = localSkynames+["smoothSky"]
    objHead = obj[0:4]
    objFoot = obj[4+len(localSkynames)+1:]

    filename = os.path.join(outputPath, "final-all-remap-smoothMap2D.fits")
    obj = pyfits.open(filename)
    map2D = obj[0].data

    filename = os.path.join(outputPath, "final-all-remap-smoothExtractModel.fits")
    obj = pyfits.open(filename)
    fimageResidual = obj[3].data
    fProfile = obj[-1].data

    filename = os.path.join(outputPath, "final-all-remap-smoothFlux.fits")
    obj = pyfits.open(filename)
    flux = obj[0].data
    fluxStd = obj[1].data
    yset = obj["xcensmooth"].data

    rawResult = stackExtract(fimage, fstd, fProfile, halfwidth=halfwidth)
    residualResult = stackExtract(fimageResidual, fstd, fProfile, halfwidth=halfwidth)

    #!! debug outputs
    debugDir = os.path.join(outputPath, "final-all-remap-smooth-debug-raw-flux")
    if not os.path.exists(debugDir):
        os.mkdir(debugDir)
    if rerun or not os.path.exists(os.path.join(debugDir, names[0]+".fig")):
        print("output ", debugDir)
        for i in range(nFiber):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.plot(waveset[i], flux[i], c="green", label="extract")
            ax.plot(waveset[i], rawResult["fit"][i], c="red", label="fit")
            ax.plot(waveset[i], rawResult["sum"][i], c="blue", label="sum")
            ax.plot(waveset[i], residualResult["fit"][i], c="red",  label="res fit")
            ax.plot(waveset[i], residualResult["sum"][i], c="blue", label="res sum", lw=0.4)
            ax.plot(waveset[i], fluxStd[i], "--", c="pink", label="flux std")
            xmin = waveset[i][0]
            xmax = waveset[i][-1]
            thisSkys = skys[(skys>xmin)&(skys<xmax)]
            thisYmax = flux[i].max()
            for eachsky in thisSkys:
                ax.plot([eachsky, eachsky], [0, thisYmax], "--", c="black", lw=0.3)
            ax.set_title(names[i])
            ax.legend()
            de.d(fig, os.path.join(debugDir, names[i]+".fig"))
            plt.close(fig)

    plotResult = {}
    plotResult["waveset"] = waveset
    plotResult["todoNames"] = todoNames
    plotResult["names"] = names
    plotResult["flux"] = flux
    plotResult["std"] = fluxStd
    plotResult["skys"] = skys
    for eachdata, eachheader, eachname in zip(todoData, todoHeaders, todoNames):
        thisFluxContinuum = np.zeros(yset.shape)
        thisFluxContinuumMedian = np.zeros(yset.shape)
        plotResult[eachname]={}
        fitFlux = stackExtract(eachdata, fstd, fProfile, halfwidth=halfwidth)["fit"]
        plotResult[eachname]["continuum"] = thisFluxContinuum
        plotResult[eachname]["continuumMedian"] = thisFluxContinuumMedian
        plotResult[eachname]["flux"] = fitFlux
        for i in range(nFiber):
            #xmin = waveset[i][0]
            #xmax = waveset[i][-1]
            #thisKnots=blocks[(blocks>xmin)&(blocks<xmax)]
            #fitObj=scipy.interpolate.LSQUnivariateSpline(waveset[i],fitFlux[i],thisKnots,w=1/fluxStd[i])
            goodWave = np.isfinite(waveset[i])
            thiswave = waveset[i][goodWave]
            thisy = fitFlux[i][goodWave]
            thisstd = fluxStd[i][goodWave]
            badmask = (~np.isfinite(thisy))
            thisy[badmask] = 0
            thisstd[badmask]=np.inf
            try:
                fitObj = Legendre.fit(thiswave, thisy, 5, w=1/thisstd**2)
            except Exception as e:
                print(e)
                ipdb.set_trace()
            thisContinuum = fitObj(waveset[i])
            thisFluxContinuum[i,:] = thisContinuum[:]
            thisFluxContinuumMedian[i,:] = scipy.ndimage.filters.median_filter(fitFlux[i], 100)[:]

    debugDir = os.path.join(outputPath, "final-all-remap-smooth-debug-localskys")
    if not os.path.exists(debugDir):
        os.mkdir(debugDir)
    if rerun or not os.path.exists(os.path.join(debugDir, names[0]+".fig")):
        print("output ", debugDir)
        with open(os.path.join(outputPath, "final-all-remap.dat"), "wb") as f:
            pickle.dump(plotResult, f)
        for i in range(nFiber):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            xmin = waveset[i][0]
            xmax = waveset[i][-1]
            thisSkys = skys[(skys>xmin)&(skys<xmax)]
            thisYmax=0
            ax.plot(waveset[i], fluxStd[i], "--", c="pink", label="flux std")
            fig = plt.figure()
            ax = fig.add_subplot(111)
            for j,eachname in enumerate(todoNames):
                if j==len(todoNames)-1:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c="black", label=eachname+" flux", lw=0.7)
                    ax.plot(waveset[i], plotResult[eachname]["continuum"][i], "--",
                            c="black", lw=0.7)
                else:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c=c[j], label=eachname+" flux")
                    ax.plot(waveset[i], plotResult[eachname]["continuum"][i], "--", c=c[j])
                thisYmax = np.max([plotResult[eachname]["flux"][i].max(), thisYmax])
            for eachsky in thisSkys:
                ax.plot([eachsky, eachsky], [0, thisYmax], "--", c="black", lw=0.3)
            ax.set_title(names[i])
            ax.legend()
            de.d(fig, os.path.join(debugDir, names[i]+".fig"))
            plt.close(fig)
        for i in range(nFiber):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            xmin = waveset[i][0]
            xmax = waveset[i][-1]
            thisSkys = skys[(skys>xmin)&(skys<xmax)]
            thisYmax=0
            ax.plot(waveset[i], fluxStd[i], "--", c="pink", label="flux std")
            fig = plt.figure()
            ax = fig.add_subplot(111)
            for j,eachname in enumerate(todoNames):
                if j==len(todoNames)-1:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c="black", label=eachname+" flux", lw=0.7)
                    ax.plot(waveset[i], plotResult[eachname]["continuumMedian"][i], "--",
                            c="black", lw=0.7)
                else:
                    ax.plot(waveset[i], plotResult[eachname]["flux"][i],
                            c=c[j], label=eachname+" flux")
                    ax.plot(waveset[i], plotResult[eachname]["continuumMedian"][i], "--", c=c[j])
                thisYmax = np.max([plotResult[eachname]["flux"][i].max(), thisYmax])
            for eachsky in thisSkys:
                ax.plot([eachsky, eachsky], [0, thisYmax], "--", c="black", lw=0.3)
            ax.set_title(names[i])
            ax.legend()
            de.d(fig, os.path.join(debugDir, names[i]+"_median_filters.fig"))
            plt.close(fig)

    cleanFile = os.path.join(outputPath, "final-all-remap-smooth-clean.fits")
    if rerun or not os.path.exists(cleanFile):
        print("output ", cleanFile)
        bzeros = np.zeros(fimage.shape)
        middleExts = []
        for i, (eachdata, eachheader, eachname) in enumerate(zip(todoData, todoHeaders, todoNames)):
            ext0 = pyfits.ImageHDU(
                    inv_reshapeRemap(eachdata, N, xstart, xend, xlength, nx),
                    header=eachheader)
            fluxZeros = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxZeros[finalGoodFiberIndex, :] = plotResult[eachname]["continuum"][:,:]
            fluxOnes = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxOnes[finalGoodFiberIndex, :] = 1
            thisContinuumImage = extract.map2DModel(bzeros, map2D, fluxZeros, fluxOnes)
            resultImageOrigin = eachdata - thisContinuumImage
            resultImage = inv_reshapeRemap(resultImageOrigin, N, xstart, xend, xlength, nx)
            ext1 = pyfits.ImageHDU(resultImage,
                    header=pyfits.Header({"OBJECT":eachheader["OBJECT"] + " clean poly"}))
            fluxZeros = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxZeros[finalGoodFiberIndex, :] = plotResult[eachname]["continuumMedian"][:,:]
            fluxOnes = np.zeros((finalGoodFiber.shape[0], fimage.shape[1]))
            fluxOnes[finalGoodFiberIndex, :] = 1
            thisContinuumImage = extract.map2DModel(bzeros, map2D, fluxZeros, fluxOnes)
            resultImageOrigin = eachdata - thisContinuumImage
            resultImage = inv_reshapeRemap(resultImageOrigin, N, xstart, xend, xlength, nx)
            ext2 = pyfits.ImageHDU(resultImage,
                    header=pyfits.Header({"OBJECT":eachheader["OBJECT"] + " clean median"}))
            middleExts = middleExts + [ext0, ext1, ext2]
        outputHDUList = objHead + middleExts + objFoot
        fits = pyfits.HDUList(outputHDUList)
        fits.writeto(cleanFile, clobber='True')
        #!! make cut file for single object
        cleanCutFilePath = os.path.join(outputPath, "final-all-remap-smooth-clean-cut")
        os.makedirs(cleanCutFilePath, exist_ok=True)
        toCutFrames = outputHDUList[:32]
        print("output final-all-remap-smooth-clean-cut")
        width = halfwidth*2+1
        for i in range(nFiber):
            thisname = names[i]
            thispath = os.path.join(cleanCutFilePath, thisname+".fits")
            thisSingleCutResult = np.empty((0,toCutFrames[0].data.shape[1]))
            thisHeader = pyfits.Header()
            for j,eachExt in enumerate(toCutFrames):
                eachdata = eachExt.data
                thisStartIndex = thisSingleCutResult.shape[0]
                #thiscut = specLib.map2DCutFixWidthSingle(eachdata, yset[i:i+1,:])["cutFlux"]
                thiscut = eachdata[i*width:(i+1)*width]
                thiscut[~np.isfinite(thiscut)]=0
                thisSingleCutResult = np.concatenate((thisSingleCutResult, thiscut))
                thisEndIndex = thisSingleCutResult.shape[0]
                thisHeader["start{}".format(j)] = str(thisStartIndex)
                thisHeader["end{}".format(j)] = str(thisEndIndex)
                thisHeader["object{}".format(j)] = str(eachExt.header.get("object", "???"))
            thisHDU = pyfits.PrimaryHDU(thisSingleCutResult, header = thisHeader)
            waveHDU = pyfits.ImageHDU(yset[i])
            hisfits = pyfits.HDUList([thisHDU, waveHDU])
            thisfits.writeto(thispath, clobber='True')
    #endregion: remoap




#!! fimages should have been subtracted with equal width
#!! reXset and reYset are 1D
def stackAndRemap(fimageList, fstdList, ysetList, wavesetList, objNameList,
                  nameList, yset1D, xset1D=None, xset1DLength=3000,
                  halfwidth=5, badFlux=0, badVar=0, scaleList=None):
    #region: test data
    for i, (eachFlux, eachStd, eachyset, eachwaveset, eachNames) in \
            enumerate(zip(fimageList, fstdList, ysetList, wavesetList, nameList)):
        assert eachFlux.shape==eachStd.shape
        assert eachFlux.shape[1]==eachyset.shape[1]
        assert eachyset.shape==eachwaveset.shape
        assert eachyset.shape[0]==len(eachNames)
    if scaleList is None:
        scaleList = np.ones(len(fimageList))
    #endregion: test data
    singleFlux = []
    singleStd = []
    #!! finalStack result for each 
    finalVar = OD()
    finalFlux = OD()
    finalMapCount = OD()
    finalMapWeight = OD()
    if xset1D is None:
        raise
        wavesetListArray = np.array(wavesetList)
        xmax = wavesetListArray.max()+10
        xmin = np.min(wavesetListArray[wavesetListArray>0])-10
        xset1D = np.linspace(xmin, xmax, xset1DLength)
    reXset, reYset = np.meshgrid(xset1D, yset1D)
    nXset = xset1D.size
    nYset = yset1D.size
    width=2*halfwidth+1
    #!! cut and rectify all fibers and make sum
    for i, (eachFlux, eachStd, eachyset, eachwaveset, eachNames, eachObjName) in \
            enumerate(zip(fimageList, fstdList, ysetList, wavesetList, nameList, objNameList)):
        #print("stacking {}/{} {}".format(i+1, len(objNameList), eachObjName))
        eachFlux = eachFlux.copy()
        eachStd[~np.isfinite(eachFlux)] = np.inf #!! ++
        eachFlux[~np.isfinite(eachStd)] = 0
        thisResult = specLib.map2DCutFixWidth(eachFlux, eachStd, eachyset, eachwaveset, halfwidth)
        #!! smoothCenter+-width x nFiber
        thisCutFlux = thisResult["cutFlux"]
        thisCutFstd = thisResult["cutFstd"]
        thisCutYset = thisResult["cutYset"]
        thisCutWaveset = thisResult["cutWaveset"]
        thisResult = specLib.remapFlux(thisCutFlux, thisCutFstd, thisCutWaveset, thisCutYset, width,
                                    reXset, reYset)
        #!! fixed width x fixed wavelength
        thisFlux = thisResult["mapFlux"]
        thisVar = thisResult["mapVar"]
        thisCount = thisResult["mapCount"]
        thisTimeScale = scaleList[i]
        thisWeight = 1/thisTimeScale
        #!! use time scale as weight
        for i, eachName in enumerate(eachNames):
            tI = slice(i*nYset,(i+1)*nYset)#!! this index
            tC = thisCount[tI,:]
            if eachName in finalFlux.keys():
                finalFlux[eachName][tC]=finalFlux[eachName][tC]+thisFlux[tI,:][tC]*thisWeight
                finalVar[eachName][tC] = finalVar[eachName][tC]+thisVar[tI,:][tC]*thisWeight
                finalMapCount[eachName]=finalMapCount[eachName]+thisCount[tI, :]
                finalMapWeight[eachName]=finalMapWeight[eachName]+thisWeight
            else:
                finalFlux[eachName]=thisFlux[tI,:].copy()*thisWeight
                finalVar[eachName] =thisVar[tI,:].copy()*thisWeight
                finalVar[eachName][~np.isfinite(finalVar[eachName])]=0
                finalMapCount[eachName]=thisCount[tI, :].astype(np.int)
                finalMapWeight[eachName]=thisCount[tI, :]*thisWeight
        thisVar[~thisCount]=badVar
        thisFlux[~thisCount]=badFlux
        #!! for debug only
        singleFlux.append(thisFlux)
        singleStd.append(np.sqrt(thisVar))
    finalNames = list(finalFlux.keys())
    nFiber = len(finalNames)
    remapFluxImage = np.zeros((nFiber*nYset, nXset))
    remapStdImage = np.zeros((nFiber*nYset, nXset))
    remapCountImage = np.zeros((nFiber*nYset, nXset))
    for i, eachName in enumerate(finalNames):
        thisBadMask = finalMapCount[eachName]==0
        #thisFlux = finalFlux[eachName]/finalMapCount[eachName]
        thisFlux = finalFlux[eachName]/finalMapWeight[eachName]
        #thisVar = np.sqrt(finalVar[eachName])/finalMapCount[eachName]
        thisVar = np.sqrt(finalVar[eachName])/finalMapWeight[eachName]
        thisVar[thisBadMask]=badVar
        thisFlux[thisBadMask]=badFlux
        # methods to correct ds9 zscale calculation....
        #   thisFlux[thisBadMask] = (np.random.random(thisFlux.shape)*0.1)[thisBadMask]
        #   thisVar[thisBadMask] = (np.random.random(thisVar.shape)*0.1)[thisBadMask]

        allZeros = thisFlux.sum(axis=0)==0
        thisFlux[:,allZeros] = np.nanmedian(thisFlux)
        thisBadValues = np.nanmean(thisFlux, axis=0)
        badY, badX = np.where(thisBadMask)
        for eachy, eachx in zip(badY, badX):
            thisFlux[eachy, eachx] = thisBadValues[eachx]

        remapFluxImage[i*nYset:(i+1)*nYset, :] = thisFlux[:,:]
        remapStdImage[i*nYset:(i+1)*nYset, :] = thisVar[:,:]
        remapCountImage[i*nYset:(i+1)*nYset, :] = finalMapCount[eachName][:,:]
    return {"remapFluxImage": remapFluxImage, "remapStdImage":remapStdImage,
            "remapCountImage":remapCountImage,
            "fluxDict":finalFlux, "fvarDict":finalVar, "mapCountDict":finalMapCount,
            "singleFlux":singleFlux, "singleStd":singleStd,
            "nXset":nXset, "nYset":nYset, "finalNames":finalNames,
            "reXset":reXset}

def stackAndRemapConvolved(fimageList, fimageSmoothSkyList, fstdList, ysetList, wavesetList, objNameList,
                  nameList, yset1D, smoothGaussianFittingList, sigmaLegList,
                  xset1D=None, xset1DLength=3000,
                  halfwidth=5, badFlux=0, badVar=0, scaleList=None):
    #region: test data
    for i, (eachFlux, eachSkyFlux, eachStd, eachyset, eachwaveset, eachNames) in \
            enumerate(zip(fimageList, fimageSmoothSkyList, fstdList, ysetList, wavesetList, nameList)):
        assert eachFlux.shape==eachStd.shape
        assert eachFlux.shape==eachSkyFlux.shape
        assert eachFlux.shape[1]==eachyset.shape[1]
        assert eachyset.shape==eachwaveset.shape
        assert eachyset.shape[0]==len(eachNames)
    if scaleList is None:
        scaleList = np.ones(len(fimageList))
    #endregion: test data
    singleFlux = []
    singleStd = []
    #!! finalStack result for each 
    finalVar = OD()
    finalFlux = OD()
    finalMapCount = OD()
    finalMapWeight = OD()
    if xset1D is None:
        raise
        wavesetListArray = np.array(wavesetList)
        xmax = wavesetListArray.max()+10
        xmin = np.min(wavesetListArray[wavesetListArray>0])-10
        xset1D = np.linspace(xmin, xmax, xset1DLength)
    reXset, reYset = np.meshgrid(xset1D, yset1D)
    nXset = xset1D.size
    nYset = yset1D.size
    width=2*halfwidth+1
    #!! cut and rectify all fibers and make sum
    for i, (eachFlux, eachSkyFlux, eachStd, eachyset, eachwaveset, eachNames, eachObjName,
            smoothGaussianFitting, sigmaLeg) in \
            enumerate(zip(fimageList, fimageSmoothSkyList, fstdList,
                          ysetList, wavesetList, nameList, objNameList,
                          smoothGaussianFittingList, sigmaLegList)):
        #print("stacking {}/{} {}".format(i+1, len(objNameList), eachObjName))
        eachFlux = eachFlux.copy()
        eachSkyFlux = eachSkyFlux.copy()
        eachStd[~np.isfinite(eachFlux)] = np.inf #!! ++
        eachFlux[~np.isfinite(eachStd)] = 0
        eachSkyFlux[~np.isfinite(eachStd)] = 0
        thisResult = specLib.map2DCutFixWidth(eachFlux, eachStd, eachyset, eachwaveset, halfwidth)
        thisSkyResult = specLib.map2DCutFixWidth(eachSkyFlux, eachStd, eachyset, eachwaveset, halfwidth)
        #!! smoothCenter+-width x nFiber
        thisCutFlux = thisResult["cutFlux"]
        thisCutFstd = thisResult["cutFstd"]
        thisCutSkyFlux = thisSkyResult["cutFlux"]
        thisCutSkyFstd = thisSkyResult["cutFstd"]
        thisCutYset = thisResult["cutYset"]
        thisCutWaveset = thisResult["cutWaveset"]
        thisResult = specLib.remapFlux(thisCutFlux, thisCutFstd, thisCutWaveset, thisCutYset, width,
                                    reXset, reYset)
        thisSkyResult = specLib.remapFlux(thisCutSkyFlux, thisCutSkyFstd, thisCutWaveset,
                                       thisCutYset, width, reXset, reYset)
        #!! fixed width x fixed wavelength
        thisFlux = thisResult["mapFlux"]
        thisSkyFlux = thisSkyResult["mapFlux"]
        eachMinSigma = smoothGaussianFitting.min(axis=1).reshape((smoothGaussianFitting.shape[0],1))
        thisVar = thisResult["mapVar"]
        thisCount = thisResult["mapCount"]
        thisTimeScale = scaleList[i]
        thisWeight = 1/thisTimeScale

        smoothSize=5
        for i, eachName in enumerate(eachNames):
            inds = range(i*nYset,(i+1)*nYset)
            for k in inds:
                thisSigObj = sigmaLeg[i]
                thisKernelSigma = thisSigObj(reXset[0])
                thisFlux[k,:] = specLib.variateConvolve(
                        reXset[0], thisFlux[k,:],
                        thisKernelSigma, smoothSize)
                thisSkyFlux[k, :] = specLib.variateConvolve(
                        reXset[0], thisSkyFlux[k],
                        thisKernelSigma, smoothSize)
        thisFlux = thisFlux-thisSkyFlux
        #!! use time scale as weight
        for i, eachName in enumerate(eachNames):
            tI = slice(i*nYset,(i+1)*nYset)#!! this index
            tC = thisCount[tI,:]
            if eachName in finalFlux.keys():
                finalFlux[eachName][tC]=finalFlux[eachName][tC]+thisFlux[tI,:][tC]*thisWeight
                finalVar[eachName][tC] = finalVar[eachName][tC]+thisVar[tI,:][tC]*thisWeight
                finalMapCount[eachName]=finalMapCount[eachName]+thisCount[tI, :]
                finalMapWeight[eachName]=finalMapWeight[eachName]+thisWeight
            else:
                finalFlux[eachName]=thisFlux[tI,:].copy()*thisWeight
                finalVar[eachName] =thisVar[tI,:].copy()*thisWeight
                finalVar[eachName][~np.isfinite(finalVar[eachName])]=0
                finalMapCount[eachName]=thisCount[tI, :].astype(np.int)
                finalMapWeight[eachName]=thisCount[tI, :]*thisWeight
        thisVar[~thisCount]=badVar
        thisFlux[~thisCount]=badFlux
        #!! for debug only
        singleFlux.append(thisFlux)
        singleStd.append(np.sqrt(thisVar))
    finalNames = list(finalFlux.keys())
    nFiber = len(finalNames)
    remapFluxImage = np.zeros((nFiber*nYset, nXset))
    remapStdImage = np.zeros((nFiber*nYset, nXset))
    remapCountImage = np.zeros((nFiber*nYset, nXset))
    for i, eachName in enumerate(finalNames):
        thisBadMask = finalMapCount[eachName]==0
        #thisFlux = finalFlux[eachName]/finalMapCount[eachName]
        thisFlux = finalFlux[eachName]/finalMapWeight[eachName]
        #thisVar = np.sqrt(finalVar[eachName])/finalMapCount[eachName]
        thisVar = np.sqrt(finalVar[eachName])/finalMapWeight[eachName]
        thisVar[thisBadMask]=badVar
        thisFlux[thisBadMask]=badFlux
        # methods to correct ds9 zscale calculation....
        #   thisFlux[thisBadMask] = (np.random.random(thisFlux.shape)*0.1)[thisBadMask]
        #   thisVar[thisBadMask] = (np.random.random(thisVar.shape)*0.1)[thisBadMask]

        allZeros = thisFlux.sum(axis=0)==0
        thisFlux[:,allZeros] = np.nanmedian(thisFlux)
        thisBadValues = np.nanmean(thisFlux, axis=0)
        badY, badX = np.where(thisBadMask)
        for eachy, eachx in zip(badY, badX):
            thisFlux[eachy, eachx] = thisBadValues[eachx]

        remapFluxImage[i*nYset:(i+1)*nYset, :] = thisFlux[:,:]
        remapStdImage[i*nYset:(i+1)*nYset, :] = thisVar[:,:]
        remapCountImage[i*nYset:(i+1)*nYset, :] = finalMapCount[eachName][:,:]
    return {"remapFluxImage": remapFluxImage, "remapStdImage":remapStdImage,
            "remapCountImage":remapCountImage,
            "fluxDict":finalFlux, "fvarDict":finalVar, "mapCountDict":finalMapCount,
            "singleFlux":singleFlux, "singleStd":singleStd,
            "nXset":nXset, "nYset":nYset, "finalNames":finalNames,
            "reXset":reXset}

def directStack(fimageList, fstdList, scaleList=None):
    fluxShape = fimageList[0].shape
    if scaleList is not None:
        assert len(fimageList)==len(scaleList), "{} {}".format(len(fimageList), len(scaleList))
    else:
        scaleList = np.ones(len(scaleList))
    for eachFlux, eachStd in zip(fimageList, fstdList):
        assert fluxShape==eachFlux.shape==eachStd.shape
    resultFlux = np.zeros(fluxShape)
    resultVar = np.zeros(fluxShape)
    resultCount = np.zeros(fluxShape, dtype=np.float64)
    resultWeight = np.zeros(fluxShape, dtype=np.float64)
    for eachFlux, eachStd, eachScale in zip(fimageList, fstdList, scaleList):
        eachCount = np.isfinite(eachStd)&np.isfinite(eachFlux)#!! important
        resultCount += eachCount
        resultWeight += eachCount / eachScale
        resultFlux[eachCount] = resultFlux[eachCount] + eachFlux[eachCount] / eachScale
        resultVar[eachCount] = resultVar[eachCount] + eachStd[eachCount]**2 / eachScale**2
    badMask = ~np.isfinite(resultVar)
    resultFlux[badMask] = np.nan
    resultStd = np.sqrt(resultVar)
    resultStd[badMask] = np.inf
    #resultFlux = resultFlux/resultCount
    #resultStd  = resultStd/resultCount
    resultFlux = resultFlux / resultWeight
    resultStd  = resultStd / resultWeight
    return {"flux":resultFlux, "std":resultStd, "count":resultCount}

def stackOneNight(inputFileList, configs={}):
    #region: setups
    outputFileName= 'final-stack-info.dat'
    configs.update({"config2show":["rerun"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return
    #endregion: setups

    fimageCleanList = []
    fimageSmoothSkyList = []
    fimageResidualList = []
    fimageResidualContinuumList = []
    fimageExtractResidualList = []
    fimageSmoothSkyResidualList = []
    #fimageSmoothSkyConvolvedResidualList = []
    #fimageSmoothSkyConvolvedResidualCountList = []
    timeScaleList = []

    fiberInfo = configs["fiberInfo"]

    fstdList = []

    ysetSmoothList = []
    wavesetList = []
    nameList = []
    ysetSmoothOriginList = []
    wavesetOriginList = []
    nameOriginList = []
    smoothGaussianFittingList = []
    sigmaLegList = []

    objNameList = []
    scaleTime = 3600.0
    xset1DLength = 3000
    #xstart = 7500
    #xend = 9800
    #xset1D = np.linspace(xstart,xend,xset1DLength)
    yset1DLength = 25
    yset1DMiddle = int((yset1DLength-1)/2)
    yset1D = np.linspace(-7, 7, yset1DLength)
    finalGoodMaskList = []
    modelFittingResultsList = []
    blocksList = []

    stackInfo = {"images":[],
            "yset1DLength":yset1DLength, "xset1DLength":xset1DLength}
    logger.p("make stack list")
    #!! load localSkynames from (any) the first data
    eachFile = inputFileList[0]
    thisFinalResultData = eachFile.split(".fits")[0] + "-final.dat"
    with open(thisFinalResultData, "rb") as f:
        thisResult = pickle.load(f)
    localSkynames = thisResult["localSkynames"]
    fimageLocalSkyResidualListList = [list() for each in localSkynames]
    maxFiberNumber = 0
    maxFiberIndex = 0

    for ii, eachFile in enumerate(inputFileList):
        #region: read in data
        name = os.path.basename(eachFile).split(".fits")[0]
        thisWaveFile = name + "-waveset.fits"
        thisMap2DFile = eachFile.split(".fits")[0] + "Map2D.fits"
        thisFinalResultData = eachFile.split(".fits")[0] + "-final.dat"
        thisFinalResultFits = eachFile.split(".fits")[0] + "-final.fits"
        with open(thisFinalResultData, "rb") as f:
            thisResult = pickle.load(f)
        thisObj = pyfits.open(thisFinalResultFits)
        fimageClean = thisObj["DATA"].data
        fimageStd = thisObj["STD"].data
        fimageFittingStd = fimageStd.copy()
        fimageSmoothSky = thisObj["SSKY"].data
        fimageResidual = thisObj["RESIDUAL"].data
        fimageResidualContinuum = thisObj["CONTINUUMRES"].data
        fimageExtractResidual = thisObj["ERESIDUAL"].data
        fimageSmoothSkyResidual = thisObj["SSKYRESIDUAL"].data
        fimageFixedWidthMask = thisObj["MASK"].data.astype(np.bool)# nan is taken as true
        fimageFittingFixedWidthMask = thisObj["FITTINGMASK"].data.astype(np.bool)
        fiberNames = thisResult["names"]
        #!! TDDO: here to scale for factors
        fluxFactors = thisResult["fluxFactors"]
        #fluxFittingGoodMask = thisResult[""]
        ny, nx = fimageClean.shape
        exptime = thisResult["exptime"]
        stackInfo["images"].append({"name":name, "exptime":exptime})
        timeScale = scaleTime/exptime
        ysetSmooth = thisResult["ysetSmooth"]
        waveset = thisResult["waveset"]
        modelFittingResults = thisResult["modelFittingResults"]
        blocks = thisResult["blocks"]
        smoothGaussianFitting = thisResult["smoothGaussianFitting"]
        sigmaLeg = thisResult["gaussianSigmaLeg"]
        fiberIndex = thisResult["fiberIndex"].astype(np.int)
        finalGoodMask = thisResult["finalGoodMask"]
        #!! plot for gaussianSmooth
        nf= fiberInfo.shape[0]
        c = cm.get_cmap("rainbow")
        c = c(np.linspace(0,1,nf))
        #for i, index in enumerate(fiberIndex[finalGoodMask]):
            #de.p(smoothGaussianFitting[i], c=c[index])
        #endregion: read in data
        fimageClean[fimageFixedWidthMask] = np.nan
        fimageStd[fimageFixedWidthMask] = np.inf
        fimageFittingStd[fimageFittingFixedWidthMask] = np.inf

        fimageResidual[fimageFittingFixedWidthMask] = np.nan
        fimageResidualContinuum[fimageFittingFixedWidthMask] = np.nan

        fimageExtractResidual[fimageFixedWidthMask] = np.nan
        fimageSmoothSkyResidual[fimageFixedWidthMask] = np.nan

        fimageClean *= timeScale
        fimageStd *= timeScale
        fimageFittingStd *= timeScale
        fimageSmoothSky *= timeScale
        fimageResidual *= timeScale
        fimageResidualContinuum *= timeScale
        fimageExtractResidual *= timeScale
        fimageSmoothSkyResidual *= timeScale
        fimageGlobalScale = timeScale
        for i, eachlist in enumerate(fimageLocalSkyResidualListList):
            eachlist.append(thisObj[localSkynames[i]].data*timeScale)

        #names = [fiberInfo[i]["id"].decode('UTF-8') for i in range(nFiber)]
        namesOrigin = ["_".join([fiberInfo[i]["fiber"].decode('UTF-8'),
                           fiberInfo[i]["id"].decode('UTF-8')])
                    for i in fiberIndex]
        timeScaleList.append(timeScale)
        ysetSmoothOriginList.append(ysetSmooth)
        wavesetOriginList.append(waveset)
        nameOriginList.append(namesOrigin)

        #!! only use good data (for rect stack)
        ysetSmooth = ysetSmooth[finalGoodMask]
        waveset = waveset[finalGoodMask]
        names = [namesOrigin[i] for i in range(len(namesOrigin)) if finalGoodMask[i]]
        if finalGoodMask.sum()>maxFiberNumber:
            maxFiberNumber = finalGoodMask.sum()
            maxFiberIndex = ii
        ysetSmoothList.append(ysetSmooth)
        wavesetList.append(waveset)
        nameList.append(names)
        smoothGaussianFittingList.append(smoothGaussianFitting)
        sigmaLegList.append(sigmaLeg)
        #names = namesOrigin
        #assert len(names)==len(fiberIndex)==finalGoodMask.sum(),\
                    #"{} {} {}".format(len(names), len(fiberIndex), finalGoodMask.sum())
        ###!! gen convolved data
        ##logger.p("get single remap for {}".format(name))
        ##info = stackAndRemap([fimageClean], [fimageStd], [ysetSmooth],
        #                       #[waveset], [name], [names], yset1D,
        #                       #xset1D=None, xset1DLength = xset1DLength,
        #                       #badFlux = np.nan)
        ##singleMapClean = info["remapFluxImage"]
        ##reXset = info["reXset"]
        ##singleMapSkySmooth = stackAndRemap([fimageSmoothSky], [fimageStd], [ysetSmooth],
        #                       #[waveset], [name], [names], yset1D,
        #                       #xset1D=None, xset1DLength = xset1DLength,
        #                       #badFlux = np.nan)["remapFluxImage"]
        ##eachMinSigma = smoothGaussianFitting.min(axis=1).reshape((smoothGaussianFitting.shape[0],1))
        ##singleMapCleanConvolved = np.zeros(singleMapClean.shape)
        ##singleMapSkySmoothConvolved = np.zeros(singleMapClean.shape)
        ##finalFiberNum = int(singleMapClean.shape[0]/yset1DLength)
        ##assert finalFiberNum==waveset[finalGoodMask].shape[0]
        ###assert finalFiberNum==waveset.shape[0]
        ##smoothSize=5
        ##for j in range(finalFiberNum):
        #    #inds = range(j*yset1DLength, (j+1)*yset1DLength)
        #    #for k in inds:
        #        #thisSigObj = sigmaLeg[j]
        #        #thisKernelSigma = thisSigObj(reXset[0])
        #        #singleMapCleanConvolved[k,:] = specLib.variateConvolve(
        #                #reXset[0], singleMapClean[k],
        #                #thisKernelSigma, smoothSize)
        #        #singleMapSkySmoothConvolved[k, :] = specLib.variateConvolve(
        #                #reXset[0], singleMapSkySmooth[k],
        #                #thisKernelSigma, smoothSize)
        ##fimageSmoothSkyConvolvedResidual = singleMapCleanConvolved - singleMapSkySmoothConvolved
        ##fimageSmoothSkyConvolvedResidualCount = np.isfinite(fimageSmoothSkyConvolvedResidual)
        fstdList.append(fimageStd)
        fimageCleanList.append(fimageClean)
        fimageSmoothSkyList.append(fimageSmoothSky)

        fimageResidualList.append(fimageResidual)
        fimageResidualContinuumList.append(fimageResidualContinuum)

        fimageExtractResidualList.append(fimageExtractResidual)
        fimageSmoothSkyResidualList.append(fimageSmoothSkyResidual)
        ##fimageSmoothSkyConvolvedResidualList.append(fimageSmoothSkyConvolvedResidual)
        ##fimageSmoothSkyConvolvedResidualCountList.append(fimageSmoothSkyConvolvedResidualCount)

        #!! todo add cov here
        #!! fimageClean, fimageSmoothSky, smoothGaussianFitting
        objNameList.append(name)
        finalGoodMaskList.append(finalGoodMask)
        modelFittingResultsList.append(modelFittingResults)
        blocksList.append(blocks)

    #!! now, we have convolved smoothSky data for each imge
    #!! 
    wavesetListArray = np.array(wavesetOriginList)
    xmax = wavesetListArray.max()+10
    xmin = np.min(wavesetListArray[wavesetListArray>0])-10
    xset1D = np.linspace(xmin, xmax, xset1DLength)
    stackInfo["xmax"]=xmax;stackInfo["xmin"]=xmin;


    def thisHeader():
        w = wcs.WCS(naxis=2)
        w.wcs.crpix = [0,0];
        w.wcs.cdelt = [(xmax-xmin)/xset1DLength,1];
        w.wcs.crval = [xmin,0];
        w.wcs.ctype = ["x--LINEAR", "y--LINEAR"]
        return w.to_header(relax=True)
    #!! we have get all the data, now make stack
    #region: first do direct stack
    result = directStack(fimageCleanList, fstdList, scaleList = timeScaleList)
    direct_fimageClean = result["flux"]
    direct_fstdClean = result["std"]
    result = directStack(fimageExtractResidualList, fstdList, scaleList = timeScaleList)
    direct_fimageExtractResidual = result["flux"]
    result = directStack(fimageResidualList, fstdList, scaleList = timeScaleList)
    direct_fimageResidual = result["flux"]
    result = directStack(fimageResidualContinuumList, fstdList, scaleList = timeScaleList)
    direct_fimageResidualContinuum = result["flux"]
    result = directStack(fimageSmoothSkyResidualList, fstdList, scaleList = timeScaleList)
    direct_fimageSmoothSkyResidual = result["flux"]
    direct_fstd = result["std"]#!! any one is ok
    direct_count = result["count"]

    direct_localSkyResidualList = [
        directStack(eachlist, fstdList, scaleList=timeScaleList)["flux"]
            for eachlist in fimageLocalSkyResidualListList
    ]

    h = pyfits.Header()
    h["OBJECT"] = "direct stack clean fimage"#!! without background
    h["EXTNAME"] = "sci"
    f0 = pyfits.PrimaryHDU(direct_fimageClean, header=h)

    h["OBJECT"] = "direct stack clean fimage std"
    h["EXTNAME"] = "std"
    fl = pyfits.ImageHDU(direct_fstdClean, header=h)

    h = pyfits.Header()
    h["OBJECT"] = "direct stack extract residual"
    f1 = pyfits.ImageHDU(direct_fimageExtractResidual, header=h)
    h = pyfits.Header()
    h["OBJECT"] = "direct stack residual"
    f2 = pyfits.ImageHDU(direct_fimageResidual, header=h)
    h = pyfits.Header()
    h["OBJECT"] = "direct stack smoothsky residual"
    f3 = pyfits.ImageHDU(direct_fimageSmoothSkyResidual, header=h)
    h = pyfits.Header()
    h["OBJECT"] = "direct stack residual continuum"
    f4 = pyfits.ImageHDU(direct_fimageResidualContinuum, header=h)

    fList = [pyfits.ImageHDU(eachData, header=pyfits.Header(
        {"OBJECT":"direct stack {} local sky".format(eachName)}))
            for eachData, eachName in zip(direct_localSkyResidualList, localSkynames)]

    hduList = [f0, f1, f2, f3] + fList + [f4]
    for each, eachName in zip(fimageResidualList, objNameList):
        h = pyfits.Header()
        h["OBJECT"] = eachName + " residual"
        hduList.append(pyfits.ImageHDU(each, header=h))
    for each, eachName in zip(fimageSmoothSkyResidualList, objNameList):
        h = pyfits.Header()
        h["OBJECT"] = eachName + " smooth sky residual"
        hduList.append(pyfits.ImageHDU(each, header=h))
    for each, eachName in zip(fimageResidualContinuumList, objNameList):
        h = pyfits.Header()
        h["OBJECT"] = eachName + " residual continuum"
        hduList.append(pyfits.ImageHDU(each, header=h))

    hduList.append(fl)
    fits = pyfits.HDUList(hduList)
    fits.writeto(os.path.join(outputPath, "final-directStack.fits"),
                 clobber='True')
    #endregion: first do direct stack
    # use names and ysetSmooth from the last
    extract.genTracingMarkersByNamesSimple(os.path.join(outputPath, "final-directStack.reg"),
            nameList[maxFiberIndex], ysetSmoothList[maxFiberIndex])
    skysub.getLineMarks(os.path.join(outputPath, "final-directStack.reg"),
                        wavesetOriginList[maxFiberIndex], ysetSmoothOriginList[maxFiberIndex],
                        finalGoodMaskList[maxFiberIndex],
                        modelFittingResultsList[maxFiberIndex],
                        blocksList[maxFiberIndex])
    #!! then do cut stack




    todoNames = ["fimageClean", "fimageSmoothSky", "fimageExtractResidual", "fimageResidual",
                 "fimageSmoothSkyResidual", "fimageResidualContinuum"]
    #remapHDUList = []
    remapHDUDataList = []
    remapHDUHeaderList = []
    #nFile = len(todoNames)+len(localSkynames); xx
    for i,eachtodo in enumerate(todoNames):
        thisList = locals()[eachtodo+"List"]
        logger.p("do stack for {}".format(eachtodo+"List"))
        result = stackAndRemap(thisList, fstdList, ysetSmoothList,
                               wavesetList, objNameList, nameList, yset1D,
                               xset1D=xset1D, xset1DLength = xset1DLength, scaleList=timeScaleList)
        remapFluxImage = result["remapFluxImage"]
        remapStdImage = result["remapStdImage"]
        remapCountImage = result["remapCountImage"]
        fluxDict = result["fluxDict"]
        fvarDict = result["fvarDict"]
        mapCountDict = result["mapCountDict"]
        nXset = result["nXset"]
        nYset = result["nYset"]
        #if i in range(len(result["singleFlux"])):
            #result["singleFlux"][i] = result["singleFlux"][i][:,edgeCut:-edgeCut]
            #result["singleStd"][i] = result["singleStd"][i][:,edgeCut:-edgeCut]
        singleFlux = result["singleFlux"]
        singleStd = result["singleStd"]
        finalNames = result["finalNames"]

        h = thisHeader()
        h["OBJECT"] = eachtodo
        f0 = pyfits.PrimaryHDU(remapFluxImage, header=h)
        if len(remapHDUDataList):
            #ff = pyfits.ImageHDU(remapFluxImage, header=h)
            remapHDUDataList.append(remapFluxImage)
            remapHDUHeaderList.append(h)
        else:
            h["EXTNAME"] = "sci" #!! for later image extraction for fimageClean
            #ff = pyfits.PrimaryHDU(remapFluxImage, header=h)
            #remapHDUList.append(ff)
            remapHDUDataList.append(remapFluxImage)
            remapHDUHeaderList.append(h)
            h = thisHeader()
            h["OBJECT"] = eachtodo+" remapFluxImage STD"
            h["EXTNAME"] = "std" #!! for later image extraction for fimageClean
            #remapStdExt = pyfits.ImageHDU(remapStdImage, header=h)
            remapStdExtData = remapStdImage
            remapStdExtHeader = h

        #region: do not output...
        #h = thisHeader()
        #h["OBJECT"] = eachtodo+" remapStdImage"
        #f1 = pyfits.ImageHDU(remapStdImage, header=h)
        #h = thisHeader()
        #h["OBJECT"] = eachtodo+" remapCountImage"
        #f2 = pyfits.ImageHDU(remapCountImage, header=h)
        #hduList = [f0, f1, f2]
        #hduList2 = []
        #for eachFlux, eachStd, eachRemapFlux, eachRemapStd, eachName in\
                #zip(thisList, fstdList, singleFlux, singleStd, objNameList):
            #h = pyfits.Header()
            #h["OBJECT"] = "{} fimage".format(eachName)
            #data = pyfits.ImageHDU(eachFlux, header=h)
            #hduList2.append(data)
            #h = pyfits.Header()
            #h["OBJECT"] = "{} fstd".format(eachName)
            #data = pyfits.ImageHDU(eachStd, header=h)
            #hduList2.append(data)
            #h = thisHeader()
            #h["OBJECT"] = "{} flux remap".format(eachName)
            #data = pyfits.ImageHDU(eachRemapFlux, header=h)
            #hduList.append(data)
            #h = thisHeader()
            #h["OBJECT"] = "{} std remap".format(eachName)
            #data = pyfits.ImageHDU(eachRemapStd, header=h)
            #hduList.append(data)
        #fits = pyfits.HDUList(hduList+hduList2)
        #filename = os.path.join(outputPath, "final-{}-remap".format(eachtodo))
        #fits.writeto(filename+".fits", clobber='True')
        ##!! gen three kinds of reg file
        #remapYsetAll = yset1DMiddle + np.arange(len(finalNames))*yset1DLength
        #reXset, reYset = np.meshgrid(xset1D, remapYsetAll)
        #extract.genTracingMarkersByNamesSimple(filename.split(".fits")[0]+"-remap.reg", finalNames, reYset)
        #skysub.getLineMarks(filename.split(".fits")[0]+"-remap.reg",
                        #reXset, reYset, np.ones(reXset.shape[0]).astype(np.bool),
                        #modelFittingResults, blocks, rect=True)
        #remapYset = yset1DMiddle + np.arange(ysetSmoothList[0].shape[0])*yset1DLength
        #reXset, reYset = np.meshgrid(xset1D, remapYset)
        #extract.genTracingMarkersByNamesSimple(filename.split(".fits")[0]+"-single-remap.reg", names, reYset)
        #skysub.getLineMarks(filename.split(".fits")[0]+"-single-remap.reg",
                        #reXset, reYset, np.ones(reXset.shape[0]).astype(np.bool),
                        #modelFittingResults, blocks, rect=True)
        #extract.genTracingMarkersByNamesSimple(filename.split(".fits")[0]+"-image.reg", names, ysetSmooth)
        #skysub.getLineMarks(filename.split(".fits")[0]+"-image.reg",
                        #waveset, ysetSmooth, finalGoodMask, modelFittingResults, blocks)
        #with open(filename, "w") as f:
            #toWrite = {"python3Path":"/usr/local/bin/python3",
                       #"filename":filename+".fits",
                       #"list":repr([1,2,3]+[3+2+4*len(thisList)+each for each in range(7)]),
                       #"remap":repr([1,2,3]+[3+2+4*len(thisList)+each for each in range(7)]),
                       #"singleRemap":repr([3+1+each*2 for each in range(len(thisList))] +
                                              #[3+2+each*2 for each in range(len(thisList))])
                       #}
            #f.write(ds10Str.format(**toWrite))
            #os.chmod(filename,stat.S_IRWXU)
        #endregion: do not output...

    #!! for all local sky
    for i,eachtodo in enumerate(localSkynames):
        thisList = fimageLocalSkyResidualListList[i]
        logger.p("do stack for local sky: {}".format(eachtodo))
        result = stackAndRemap(thisList, fstdList, ysetSmoothList,
                               wavesetList, objNameList, nameList, yset1D,
                               xset1D=xset1D, xset1DLength = xset1DLength, scaleList=timeScaleList)
        remapFluxImage = result["remapFluxImage"]
        remapStdImage = result["remapStdImage"]
        remapCountImage = result["remapCountImage"]
        fluxDict = result["fluxDict"]
        fvarDict = result["fvarDict"]
        mapCountDict = result["mapCountDict"]
        nXset = result["nXset"]
        nYset = result["nYset"]
        #if i in range(len(result["singleFlux"])):
            #result["singleFlux"][i] = result["singleFlux"][i][:,edgeCut:-edgeCut]
            #result["singleStd"][i] = result["singleStd"][i][:,edgeCut:-edgeCut]
        singleFlux = result["singleFlux"]
        singleStd = result["singleStd"]
        finalNames = result["finalNames"]

        h = thisHeader()
        h["OBJECT"] = "local sky "+eachtodo
        f0 = pyfits.PrimaryHDU(remapFluxImage, header=h)
        ff = pyfits.ImageHDU(remapFluxImage, header=h)
        #remapHDUList.insert(-1,ff)
        remapHDUDataList.insert(-1,remapFluxImage)
        remapHDUHeaderList.insert(-1,h)

    #!! for convolved smooth sky !!!!!!!!!!!!!!!!!!!!!!!here
    if True:
        logger.p("do stack for {}".format("SkyConvolvedList"))
        result = stackAndRemapConvolved(fimageCleanList, fimageSmoothSkyList, fstdList, ysetSmoothList,
                               wavesetList, objNameList, nameList, yset1D,
                               smoothGaussianFittingList, sigmaLegList,
                               xset1D=xset1D, xset1DLength = xset1DLength, scaleList=timeScaleList)
        remapFluxImage = result["remapFluxImage"]
        remapStdImage = result["remapStdImage"]
        remapCountImage = result["remapCountImage"]
        fluxDict = result["fluxDict"]
        fvarDict = result["fvarDict"]
        mapCountDict = result["mapCountDict"]
        nXset = result["nXset"]
        nYset = result["nYset"]
        singleFlux = result["singleFlux"]
        singleStd = result["singleStd"]
        finalNames = result["finalNames"]

        h = thisHeader()
        h["OBJECT"] = "SmoothSkyConvolvedResidual"
        f0 = pyfits.PrimaryHDU(remapFluxImage, header=h)
        if len(remapHDUDataList):
            remapHDUDataList.insert(-1, remapFluxImage)
            remapHDUHeaderList.insert(-1, h)


    #logger.p("do stack for {}".format("fimageSmoothSkyConvolvedResidualList"))
    #result = np.zeros(fimageSmoothSkyConvolvedResidual.shape)
    #resultc = np.zeros(fimageSmoothSkyConvolvedResidual.shape)
    #convolvedHDUList = []
    #for i, (eachr, eachc, eachname) in\
    #        enumerate(zip(fimageSmoothSkyConvolvedResidualList, fimageSmoothSkyConvolvedResidualCountList, objNameList)):
    #    h = thisHeader()
    #    h["OBJECT"] = eachname + "smooth sky residual"
    #    convolvedHDUList.append(pyfits.ImageHDU(eachr, header=h))
    #    eachr[~eachc]=0
    #    result = result + eachr
    #    resultc = resultc + eachc
    #result = result/resultc
    #result[resultc==0]=0
    #h = thisHeader()
    #h["OBJECT"] = "stack convolved smooth sky residual"
    #convolvedHDUList.insert(0, pyfits.PrimaryHDU(result, header=h))

    #######remapHDUDataList.insert(-1, result)#!! before the continuum, (-2)==>(-1)
    #######remapHDUHeaderList.insert(-1, h)#!! before the continuum, (-2)==>(-1)

    remapHDUDataList.append(remapStdExtData)#!! add std to it
    remapHDUHeaderList.append(remapStdExtHeader)#!! add std to it
    #h = thisHeader()
    #h["OBJECT"] = "stack convolved smooth sky residual count"
    #convolvedHDUList.append(pyfits.ImageHDU(resultc, header=h))
    #fits = pyfits.HDUList(convolvedHDUList)
    #filename = os.path.join(outputPath, "final-smoothSkyConvolvedResidual-remap")
    #fits.writeto(filename+".fits", clobber='True')
    #extract.genTracingMarkersByNamesSimple(filename.split(".fits")[0]+"-remap.reg", finalNames, reYset)
    #skysub.getLineMarks(filename.split(".fits")[0]+"-remap.reg",
                        #reXset, reYset, np.ones(reXset.shape[0]).astype(np.bool),
                        #modelFittingResults, blocks, rect=True)
    #with open(filename, "w") as f:
        #toWrite = {"python3Path":"/usr/local/bin/python3",
                   #"filename":filename+".fits",
                   #"list":repr(list(range(1+len(convolvedHDUList)))),
                   #"remap":repr(list(range(1+len(convolvedHDUList)))),
                   #"singleRemap":repr([])
                   #}
        #f.write(ds10Str.format(**toWrite))
        #os.chmod(filename,stat.S_IRWXU)
    #!! add all result to single result
    #todoNames = ["fimageClean", "fimageSmoothSky", "fimageExtractResidual", "fimageResidual",
                 #"fimageSmoothSkyResidual", "smoothSkyConvolvedResidual", "fimageResidualContinuum"]
    #newFirstFrame = remapHDUList[0]
    #newFirstFrame = pyfits.ImageHDU(newFirstFrame.data, header=newFirstFrame.header)
    #newRemapList = [newFirstFrame] + remapHDUList[1:]
    #for i,eachtodo in enumerate(todoNames):
        #filename = os.path.join(outputPath, "final-{}-remap".format(eachtodo))+".fits"
        #obj = pyfits.open(filename)
        #fits = pyfits.HDUList(obj+newRemapList)
        #fits.writeto(filename, clobber='True')

    #!! for all-remap
    logger.p("make stack for all-remap")
    remapHDUList = [pyfits.PrimaryHDU(remapHDUDataList[0], remapHDUHeaderList[0])]
    for eachData, eachHeader in zip(remapHDUDataList[1:], remapHDUHeaderList[1:]):
        remapHDUList.append(pyfits.ImageHDU(eachData, eachHeader))
    fits = pyfits.HDUList(remapHDUList)

    filename = os.path.join(outputPath, "final-all-remap")
    fits.writeto(filename+".fits", clobber='True')
    with open(filename, "w") as f:
        toWrite = {"python3Path":"/usr/local/bin/python3",
                   "filename":filename+".fits",
                   #"list":repr([1,2,3,4,5,6,7]),
                   #"remap":repr([1,2,3,4,5,6,7]),
                   "list":repr(list(range(1,len(remapHDUList)+1))),
                   "remap":repr(list(range(1,len(remapHDUList)+1))),
                   "singleRemap":repr([])
                   }
        f.write(ds10Str.format(**toWrite))
        os.chmod(filename,stat.S_IRWXU)
    remapYsetAll = yset1DMiddle + np.arange(len(finalNames))*yset1DLength
    stackInfo["remapYsetAll"] = remapYsetAll
    stackInfo["finalNames"] = finalNames
    stackInfo["localSkynames"] = localSkynames
    filename_ = os.path.join(outputPath, "final-stack-info.txt")
    with open(filename_, "w") as f:
        totalExpTime = np.sum([float(each["exptime"]) for each in stackInfo["images"]])/3600.
        f.write("total exptime: {:3.1f}h\n".format(totalExpTime))
        for each in stackInfo["images"]:
            f.write("\t{:7}s {}\n".format(each['exptime'], each['name']))
        f.write("xmin:{}, xmax:{}\n".format(xmin, xmax))
        f.write("xwidth:{}, ywidth:{}\n".format(xset1DLength, yset1DLength))
        f.write("rectify fiber count:{}\n".format(len(finalNames)))
        f.write("rectify fiber names:{}\n".format(finalNames))
    filename_ = os.path.join(outputPath, "final-stack-info.dat")
    with open(filename_, "wb") as f:
        pickle.dump(stackInfo, f)

    reXset, reYset = np.meshgrid(xset1D, remapYsetAll)
    extract.genTracingMarkersByNamesSimple(filename.split(".fits")[0]+"-remap.reg", finalNames, reYset)
    skysub.getLineMarks(filename.split(".fits")[0]+"-remap.reg",
                        reXset, reYset, np.ones(reXset.shape[0]).astype(np.bool),
                        modelFittingResults, blocks, rect=True)

ds10Str="""#!{python3Path}
from fmajorAstroUtils.DS10 import ds9
import astropy.io.fits as pyfits
import os
import numpy as np
filepath = os.path.basename("{filename}")
todoFrames = {list}
remap = {remap}
singleRemap = {singleRemap}
obj=pyfits.open(filepath)
d = ds9.ds10(filepath)
allShape = [each.data.shape for each in obj]
for i,eachshape in enumerate(allShape):
    d.set("frame {{}}".format(i+1))
    d.set("fits {{}}[{{}}]".format(filepath, i))
    if i+1 in remap:
        d.set("regions {{}}".format(filepath.split(".fits")[0]+"-remap.reg"))
    elif i+1 in singleRemap:
        d.set("regions {{}}".format(filepath.split(".fits")[0]+"-single-remap.reg"))
    else:
        if os.path.exists(filepath.split(".fits")[0]+"-image.reg"):
            d.set("regions {{}}".format(filepath.split(".fits")[0]+"-image.reg"))
    if i+1 not in todoFrames:
        continue
    cropStrList = d.get("crop").split()
    x = float(cropStrList[0])
    y = float(cropStrList[1])
    xw = float(cropStrList[2])
    yw = float(cropStrList[3])
    for j in range(4):
        d.set("crop {{}} {{}} {{}} {{}}".format(x+250*(np.random.rand()-0.5), y+250*(np.random.rand()-0.5), xw-200, yw-200))
        #print('d.set("crop {{}} {{}} {{}} {{}}")'.format(x+100*(np.random.rand()-0.5), y+100*(np.random.rand()-0.5), xw-200, yw-200))
        d.set("crop {{}} {{}} {{}} {{}}".format(x, y, xw, yw))
        #print('d.set("crop {{}} {{}} {{}} {{}}")'.format(x, y, xw, yw))
"""

#region: backup
def testRemap(inputFileList, configs={}):
    outputPath = configs["outputPath"]
    fiberInfo = configs["fiberInfo"]

    nameList = []
    fimageList = []
    fstdList = []
    ysetList = []
    ysetSmoothList = []
    wavesetList = []
    objNameList = []
    xset1DLength = 3000
    xset1D = np.linspace(7500,9800,xset1DLength)
    yset1D = np.linspace(-8, 8, 25)
    for eachFile in inputFileList:
        thisMap2DFile = eachFile.split(".fits")[0] + "Map2D.fits"
        name = os.path.basename(eachFile).split(".fits")[0]
        thisWaveFile = name
        waveset,coeffs,ysetSmooth,badFiber,debugInfo = wavecalib.readWaveSet(outputPath, name=thisWaveFile, debugInfo=1)
        rms = debugInfo["rms"].flatten()#!! all fiber that have been traceseted
        finalGoodFiber = np.isfinite(rms)#!! good for wavelength

        waveset = waveset
        goodFiber = badFiber<1#!! 128
        ysetSmooth = ysetSmooth[goodFiber]

        fiberIndex = np.where(goodFiber)[0]
        #!! read flux
        eachFluxFile = eachFile.split(".fits")[0] + "Flux.fits"
        thisData, thisStd, thisYset = extract.readFlux(eachFluxFile)
        thisStd[~finalGoodFiber,:] = np.inf
        thisInvvar = 1/thisStd**2
        nFiber, nx = thisData.shape #!! nFiber in length of  finalgood
        names = [fiberInfo[i]["id"].decode('UTF-8') for i in range(nFiber)]
        obj = pyfits.open(thisMap2DFile)
        bmod = obj["bmod"].data
        map2D = obj["MAP2DO"].data
        fimageModelFixWidth = extract.map2DModelFixNaNWidth(bmod, map2D, thisData, thisInvvar, thisYset)
        obj = pyfits.open(eachFile)
        fimage = obj["sci"].data
        fstd = obj["std"].data
        fstd[~np.isfinite(fimageModelFixWidth)] = np.inf
        residual = fimage - fimageModelFixWidth
        assert waveset.shape==thisYset.shape==ysetSmooth.shape
        fimageList.append(residual)
        fstdList.append(fstd)
        ysetList.append(thisYset)
        ysetSmoothList.append(ysetSmooth)
        wavesetList.append(waveset)
        nameList.append(names)
        objNameList.append(name)
    result = stackAndRemap(fimageList, fstdList, ysetSmoothList, wavesetList, objNameList, nameList, yset1D, xset1D)
    remapFluxImage = result["remapFluxImage"]
    remapStdImage = result["remapStdImage"]
    remapCountImage = result["remapCountImage"]
    fluxDict = result["fluxDict"]
    fvarDict = result["fvarDict"]
    mapCountDict = result["mapCountDict"]
    singleFlux = result["singleFlux"]
    singleStd = result["singleStd"]

    h = pyfits.Header()
    h["OBJECT"] = "remapFluxImage"
    f0 = pyfits.PrimaryHDU(remapFluxImage, header=h)
    h = pyfits.Header()
    h["OBJECT"] = "remapStdImage"
    f1 = pyfits.ImageHDU(remapStdImage, header=h)
    h = pyfits.Header()
    h["OBJECT"] = "remapCountImage"
    f2 = pyfits.ImageHDU(remapCountImage, header=h)
    hduList = [f0, f1, f2]
    for eachFlux, eachStd, eachRemapFlux, eachRemapStd, eachName in\
            zip(fimageList, fstdList, singleFlux, singleStd, objNameList):
        h = pyfits.Header()
        h["OBJECT"] = "{} fimage".format(eachName)
        data = pyfits.ImageHDU(eachFlux, header=h)
        hduList.append(data)
        h = pyfits.Header()
        h["OBJECT"] = "{} fstd".format(eachName)
        data = pyfits.ImageHDU(eachStd, header=h)
        hduList.append(data)
        h = pyfits.Header()
        h["OBJECT"] = "{} flux remap".format(eachName)
        data = pyfits.ImageHDU(eachRemapFlux, header=h)
        hduList.append(data)
        h = pyfits.Header()
        h["OBJECT"] = "{} std remap".format(eachName)
        data = pyfits.ImageHDU(eachRemapStd, header=h)
        hduList.append(data)
    fits = pyfits.HDUList(hduList)
    fits.writeto("testStackExtract.fits", clobber='True')
    raise
#endregion: backup




