#region: imports
from __future__ import (absolute_import, division, print_function, unicode_literals)
import numpy as np
import os
import pprint
import sys
import matplotlib.pyplot as plt
import numpy
from numpy.polynomial import Legendre
from . import de
from . import extract
from .image import ccdproc
from .image import specLib
from ..utils import macro
from ..utils import messager
from ..utils import configTools
from scipy import interpolate
from scipy.optimize import curve_fit
import glob
import multiprocessing
import copy
from ..DS10 import ds9
import scipy
import matplotlib as mpl

from . import m2fsConfigs
import shutil
import astropy.io.fits as pyfits
from astropy import wcs
import pdb
import ipdb

import warnings
from scipy import ndimage
from matplotlib import cm
warnings.filterwarnings("ignore", message="divide by zero encountered in divide")

np.set_printoptions(precision=4)
#endregion: imports

def findWave(inputFile, configs={}):
    #region: setups
    outputFileName= 'waveset.fits'
    configs.update({"config2show":["rerun", "ncoeff", "initRow", "debug"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return
    #endregion: setups
    logger.p("use file {}".format(inputFile))
    #de.plt.switch_backend("Agg")
    arcType = pyfits.open(inputFile)[0].header["OLDOBJ"]
    configs["arcType"] = arcType
    yset, badFiber, boundary = extract.readTraceSet(outputPath, configs=configs)
    configs["fiberInfo"] = configs["fiberInfo"][~badFiber]
    flux, std, yset = extract.readFlux(inputFile.split(".fits")[0]+"Flux.fits", configs=configs)
    allCoeff, xmap, allRMS = waveCal(flux, std, yset, configs=configs)
    #de.plt.switch_backend(configs["plotBackend"])
    #region: save file
    h1 = pyfits.Header()
    h2 = pyfits.Header()
    h3 = pyfits.Header()
    h1["OBJECT"] = "xmap"
    h1["EXTNAME"] = "xmap"
    h2["OBJECT"] = "coeff"
    h2["EXTNAME"] = "coeff"
    h3["OBJECT"] = "RMS"
    h3["EXTNAME"] = "RMS"
    data1 = pyfits.PrimaryHDU(xmap, header=h1)
    data2 = pyfits.ImageHDU(allCoeff, header=h2)
    data3 = pyfits.ImageHDU(allRMS, header=h3)
    fits  = pyfits.HDUList([data1, data2, data3])
    fits.writeto(outFile, clobber='True')
    #endregion: save file

def waveCal(flux, std, yset, configs={}):
    #region: setups
    assert flux.shape==yset.shape
    nFiber, nx = yset.shape
    #!! ny to be auto read latter....
    ny = configs.get("ny", 2056)
    ysetInd = configs.get("ysetInd", 1024)
    #!! set lampData file
    arcType = configs["arcType"].lower()
    codePath = configs["codePath"]

    lampDataFile = configs.get('lampDataFile')
    if not lampDataFile:
        if ("hg" in arcType) and ("ar" in arcType) and ("ne" in arcType):
            if ("xe" in arcType):
                lampName = "NistHgArNeXe.m2fs.dat"
            else:
                lampName = "NistHgArNe.m2fs.dat"
        else:
            raise logger.Exception("unknown lamp type:{}".format(arcType))
        lampDataFile = os.path.join(codePath, 'data', lampName)
    else:
        lampDataFile = os.path.join(codePath, 'data', lampDataFile)
    lamp = readLampData(lampDataFile)
    logger = configs["logger"]
    logger.p('use lampfile: {}'.format(lampDataFile))
    ncoeff = configs.get('ncoeff', 6)
    debug = configs.get("debug", 0)
    smoothGuess = np.zeros((nFiber,ncoeff), dtype=np.float)

    waveFitDebug = configs.get("waveFitDebug", -1)
    FWHMDebug = configs.get("FWHMDebug",0)
    FWHMType = configs.get("FWHMType", "sigma")

    dcoeff = np.zeros((ncoeff,2), dtype=np.float)
    initRow = configs.get("initRow", int(nFiber/2)) # from the middle one
    allCoeffs=np.zeros((nFiber, ncoeff))
    yFiber = yset[:,0]
    xmap=np.zeros((nFiber, nx))
    allRMS=np.ones((nFiber, 1)) * np.inf
    outputPath = configs["outputPath"]
    op = os.path.join(outputPath,"wavesetDebug")
    #op = configs.get("outputPath","./")
    codePath = configs["codePath"]
    useTempelate = configs.get("useTempelate",0)
    sigmaInfo=np.zeros(nFiber, dtype=np.object)
    guess = np.zeros(ncoeff, dtype=np.float)
    spectrograph = configs.get("spectrograph")
    if spectrograph not in ["r", "b"]:
        raise Exception("spectrograph {} not understood".format(spectrograph))
    elif spectrograph == "r":
        waveTemplate = configs.get("waveTemplate", os.path.join(codePath, "data", "waveTemplate_r.dat"))
    else:
        waveTemplate = configs.get("waveTemplate", os.path.join(codePath, "data", "waveTemplate_b.dat"))
    # yset[initRow]=1064
    #endregion: setups
    #region: setup init guess
    if not useTempelate or not os.path.exists(waveTemplate):
        logger.p("not use tempelate")
        if os.path.exists(waveTemplate):
            logger.p("\t but we have a wave Template!! you can use it...")
        else:
            logger.p("\t because we do not have the tempelate file...")
        useTempelate = False #!! can not use tempelate...
    else:
        logger.p("tracing use template:{}".format(waveTemplate))
        smoothGuessCoeff = np.load(waveTemplate)
        for i in range(ncoeff):
            smoothGuess[:,i] = xcoeff2y(yset[:,ysetInd].flatten(), smoothGuessCoeff[i,:], xlim=[0, ny-1])
    #!! init guess parameters, should be written to a template file
    dcoeff[0][0] = -0.5
    dcoeff[0][1] =  0.5
    dcoeff[1][0] = -2
    dcoeff[1][1] =  2
    dcoeff[2:,0] = -1
    dcoeff[2:,1] =  1
    maxLag1 = configs.get('maxLag1', 500)
    maxLag2 = configs.get('maxLag2', 50)
    #endregion: setup init guess
    #region: fit for the middle one
    if useTempelate:
        guess  = smoothGuess[initRow].copy()
    else:
        #!! be caution of these values..
        configGuess = configs.get('guess')
        if configGuess:
            guess[0], guess[1], guess[2], guess[3], guess[4] = configGuess
        else:
            guess[0] = 8597 #858
            guess[1] = 1007 #1007
            guess[2] = -2   #-2
            guess[3] = -2   #-2
            guess[4] = -0   # 2

        gs=np.ones(ncoeff)
        dcoeff[1][0] = -10
        dcoeff[1][1] =  10
        gs[1] = 10
    logger.p('use init guess:{}, maxlags:{}'.format(guess, [maxLag1, maxLag2]))

    thisConfig = {"fibInd":initRow, "yInd":yset[initRow][0],
                 "maxLag1":maxLag1, "maxLag2":maxLag2}
    thisConfig.update(configs)
    thisConfig["op"] = op

    thisConfig["waveFitDebug"] = waveFitDebug
    thisConfig["FWHMDebug"] = FWHMDebug
    thisConfig["FWHMType"] = FWHMType

    #!! fit coeff for this fiber
    thisCoeff, thisMap, thisDebugInfo =\
        waveFit(flux[initRow], std[initRow], lamp, guess, dcoeff.copy(), configs=thisConfig)
    if useTempelate:
        smoothDiff = thisCoeff[0]-smoothGuess[initRow][0]
        smoothGuess[:,0] = smoothGuess[:,0] + smoothDiff
    allCoeffs[initRow,:] = thisCoeff[:]
    xmap[initRow,:] = thisMap[:]
    allRMS[initRow,:] = thisDebugInfo[0]
    lastRMS = thisDebugInfo[0]
    sigmaInfo[initRow] = thisDebugInfo[1]
    #!! handle bad fiber in the initRow
    if not np.isfinite(lastRMS):
        raise logger.Exception("the first fiber must be good, you should change the initRow config in the [findWave] section to fibers number other than {} and less than {}".format(initRow, nFiber))

    #!! not good fit for the first fiber, guess more and try again
    if lastRMS>0.1:
        logger.warn("Caution! the first fiber may be bad and the whole map may be bad")
        #newDcoeff = dcoeff.copy()
        #newDcoeff[0][0] = -0.5
        #newDcoeff[0][1] =  0.5
        #newDcoeff[1][0] = -6
        #newDcoeff[1][1] =  6
        #maxLag1 = 300
        #maxLag2 = 10
        #gs=np.ones(ncoeff)
        #gs[0] = 3
        #gs[1] = 3
        #thisConfig = {"fibInd":initRow, "yInd":yset[initRow][0],
                      #"maxLag1":maxLag1, "maxLag2":maxLag2}
        #thisConfig.update(configs)
        #thisConfig["op"] = op
        #thisConfig["waveFitDebug"] = waveFitDebug
        #thisConfig["FWHMDebug"] = FWHMDebug
        #thisConfig["FWHMType"] = FWHMType
        #thisCoeff, thisMap, thisDebugInfo =\
            #waveFit(flux[initRow],std[initRow], lamp, guess, dcoeff.copy(),
                    #guessStepFactor=gs, configs=thisConfig)
        #allCoeffs[initRow,:] = thisCoeff[:]
        #xmap[initRow,:] = thisMap[:]
        #allRMS[initRow,:] = thisDebugInfo[0]
        #lastRMS = thisDebugInfo[0]
        #sigmaInfo[initRow] = thisDebugInfo[1]
        ##!! handle bad fiber in the initRow
        #if not np.isfinite(lastRMS):
            #raise logger.Exception("the first fiber must be good, you should change the initRow config in the [findWave] section to fibers number other than {} and less than {}".format(initRow, nFiber))
    #endregion: fit for the middle one
    maxLag1 = 20
    maxLag2 = 5
    #region: fit for lower fibers
    for thisRow in range(initRow-1, -1, -1):
        if useTempelate:
            newGuess  = smoothGuess[thisRow].copy()
            factor = 1
            gs=np.ones(ncoeff)
            thisdcoeff = dcoeff.copy()
        else:
            newGuess = allCoeffs[thisRow+1].copy() # must be a copy...
            #!! 20161127 not success to not use tempelate....
            if np.abs(thisRow-initRow)>4:
                maxLag1 = 20
                maxLag2 = 5
                dcoeff[1][0] = -2
                dcoeff[1][1] =  2
                if not np.isfinite(allRMS[thisRow+1,0]):
                    goodFibers = np.isfinite(allRMS.flatten())
                    obj = Legendre.fit(yFiber[goodFibers], allCoeffs[goodFibers,0], 3, w=1/allRMS[goodFibers,0])
                    thisGuess0 = obj(yFiber[thisRow])
                    newGuess[0] = thisGuess0
                    obj = Legendre.fit(yFiber[goodFibers], allCoeffs[goodFibers,1], 3, w=1/allRMS[goodFibers,0])
                    thisGuess1 = obj(yFiber[thisRow])
                    newGuess[1] = thisGuess1
                thisdcoeff, gs, factor = getGuessFactor(lastRMS, dcoeff)
            else:
                maxLag1 = 500
                maxLag2 = 50
                dcoeff[1][0] = -5
                dcoeff[1][1] =  5
                thisdcoeff, gs, factor = getGuessFactor(lastRMS, dcoeff)
                gs[1] = gs[1]*5

            newGuess[2] = -2   #-2
            newGuess[3] = -2   #-2
            newGuess[4] = -0   # 2
            thisdcoeff, gs, factor = getGuessFactor(lastRMS, dcoeff)
        #ipdb.set_trace()
        thisConfig = {"fibInd":thisRow, "yInd":yset[thisRow][0],
                 "maxLag1":int(maxLag1*factor), "maxLag2":int(maxLag2*factor)}
        thisConfig.update(configs)
        thisConfig["op"] = op
        thisConfig["waveFitDebug"] = waveFitDebug
        thisConfig["FWHMDebug"] = FWHMDebug
        thisConfig["FWHMType"] = FWHMType
        thisCoeff, thisMap, thisDebugInfo =\
            waveFit(flux[thisRow], std[thisRow], lamp, newGuess, thisdcoeff,
                    guessStepFactor=gs,configs=thisConfig)
        allCoeffs[thisRow,:] = thisCoeff[:]
        xmap[thisRow,:] = thisMap[:]
        allRMS[thisRow,:] = thisDebugInfo[0]
        lastRMS = thisDebugInfo[0]
        sigmaInfo[thisRow] = thisDebugInfo[1]
    #endregion: fit for lower fibers
    #region: fit for upper fibers
    for thisRow in range(initRow+1, nFiber):
        if useTempelate:
            newGuess  = smoothGuess[thisRow].copy()
            factor = 1
            gs=np.ones(ncoeff)
            thisdcoeff = dcoeff.copy()
        else:
            newGuess = allCoeffs[thisRow-1].copy() # must be a copy...
            #!! fit for guess0 when last lamp fitting is bad
            if not np.isfinite(allRMS[thisRow-1,0]):
                goodFibers = np.isfinite(allRMS.flatten())
                obj = Legendre.fit(yFiber[goodFibers], allCoeffs[goodFibers,0], 3, w=1/allRMS[goodFibers,0])
                thisGuess0 = obj(yFiber[thisRow])
                newGuess[0] = thisGuess0
                obj = Legendre.fit(yFiber[goodFibers], allCoeffs[goodFibers,1], 3, w=1/allRMS[goodFibers,0])
                thisGuess1 = obj(yFiber[thisRow])
                newGuess[1] = thisGuess1
            newGuess[2] = -2   #-2
            newGuess[3] = -2   #-2
            newGuess[4] = -0   # 2
            thisdcoeff, gs, factor = getGuessFactor(lastRMS, dcoeff)
        thisConfig = {"fibInd":thisRow, "yInd":yset[thisRow][0],
                      "maxLag1":int(maxLag1*factor), "maxLag2":int(maxLag2*factor)}
        thisConfig.update(configs)
        thisConfig["op"] = op
        thisConfig["waveFitDebug"] = waveFitDebug
        thisConfig["FWHMDebug"] = FWHMDebug
        thisConfig["FWHMType"] = FWHMType
        thisCoeff, thisMap, thisDebugInfo = \
            waveFit(flux[thisRow], std[thisRow], lamp, newGuess, thisdcoeff,
                    guessStepFactor=gs,configs=thisConfig)
        allCoeffs[thisRow,:] = thisCoeff[:]
        xmap[thisRow,:] = thisMap[:]
        allRMS[thisRow,:] = thisDebugInfo[0]
        lastRMS = thisDebugInfo[0]
        sigmaInfo[thisRow] = thisDebugInfo[1]
    #endregion: fit for upper fibers
    #region: finish
    sigmaInfo.dump(os.path.join(op,"FWHM.dat"))
    c = cm.get_cmap("rainbow")
    c = c(np.linspace(0,1,nFiber))
    fig=de.f()
    ax = fig.add_subplot(111)
    for i, eachInfo in enumerate(sigmaInfo):
        ax.plot(eachInfo[0], eachInfo[1], c=c[i])
        ax.scatter(eachInfo[0], eachInfo[1], c=c[i], s=8)

    de.d(fig, os.path.join(op,"-FWHM.fig"))
    fig.clear()
    plt.close(fig)

    xx = np.arange(nFiber)
    fig = de.f()
    for i in range(ncoeff):
        ax = fig.add_subplot((ncoeff+1)*100+10+i+1)
        ax.scatter(xx, allCoeffs[:,i])
    ax = fig.add_subplot((ncoeff+1)*100+10+ncoeff+1)
    ax.scatter(xx, allRMS)
    de.d(fig, os.path.join(op,"-coeffs.fig"))
    fig.clear()
    plt.close(fig)

    nCoeffCoeff = configs.get("nCoeffCoeff",4)
    smoothGuessCoeff = np.zeros((ncoeff, nCoeffCoeff))

    fig = de.f()
    for i in range(ncoeff):
        result, badmask = xy2coeff(yset[:,ysetInd].flatten(), allCoeffs[:,i].flatten(),
                                   xlim=[0, ny-1], reject=3, configs={"ncoeff":nCoeffCoeff})[:]
        smoothGuessCoeff[i,:] = result[:]
        smoothGuess[:,i] = xcoeff2y(yset[:,ysetInd].flatten(), smoothGuessCoeff[i,:], xlim=[0, ny-1])
        ax = fig.add_subplot((ncoeff+1)*100+10+i+1)
        ax.scatter(yset[:,ysetInd].flatten(), allCoeffs[:,i].flatten())
        ax.plot(yset[:,ysetInd].flatten(), smoothGuess[:,i].flatten())
    de.d(fig, os.path.join(op,"-coeffs_smooth.fig"))
    fig.clear()
    plt.close(fig)
    smoothGuessCoeff.dump(os.path.join(op, "waveTemplate.dat"))

    return allCoeffs, xmap, allRMS
    #endregion: finish

def getGuessFactor(rms, dcoeff):
    ncoeff = dcoeff.shape[0]
    gs = np.ones(ncoeff)
    thedcoeff = dcoeff.copy()
    if rms<0.04:
        return dcoeff, gs, 1
    elif rms<0.1:
        factor = (rms-0.04)/0.06*2 + 1
        thedcoeff[1] *= factor
        gs[1] = factor*2
        return thedcoeff, gs, factor
    else:
        factor = 3
        thedcoeff[1] *= factor
        gs[1] = factor*2
        return thedcoeff, gs, factor

def filterGoodMask(goodMask, badness, mu, N, filterRange=200, configs={}):
    assert goodMask.shape==badness.shape==mu.shape
    logger = configs["logger"]
    fibInd = configs["fibInd"]
    fiberInfo = configs["fiberInfo"]
    yInd = configs.get("yInd", "?")
    message = configs.get("message", "")
    toPrint = configs.get("toPrint", 1)
    resultMask = goodMask.copy()
    lowerIndex = np.where(mu<filterRange)[0]
    upperIndex = np.where(mu>N-filterRange)[0]
    lowerGoodMask = goodMask[lowerIndex]
    upperGoodMask = goodMask[upperIndex]
    if 0<lowerGoodMask.shape[0]<=2:
        resultMask[lowerIndex] = True
        if toPrint:
            logger.warn("\t\t{}_{:6.2f} {} force only {} lines in lowerRegion {}".\
                format(fibInd, yInd, fiberInfo[fibInd].tolist()[:3],
                       lowerGoodMask.shape[0], message))
    elif lowerGoodMask.shape[0]>2 and lowerGoodMask.sum()<2:#!! at lease have to lines in ..region
        badIndexSort = np.argsort(badness[lowerIndex])
        resultMask[lowerIndex[badIndexSort[:2]]] = True
        if toPrint:
            logger.warn("\t\t{}_{:6.2f} {} force 2/{} lines in lowerRegion {}".\
                format(fibInd, yInd, fiberInfo[fibInd].tolist()[:3],
                       lowerGoodMask.shape[0], message))
    if 0<upperGoodMask.shape[0]<=2:
        resultMask[upperIndex] = True
        if toPrint:
            logger.warn("\t\t{}_{:6.2f} {} force only {} lines in lowerRegion".\
                format(fibInd, yInd, fiberInfo[fibInd].tolist()[:3],
                       upperGoodMask.shape[0], message))
    elif upperGoodMask.shape[0]>2 and upperGoodMask.sum()<2:#!! at lease have to lines in ..region
        badIndexSort = np.argsort(badness[upperIndex])
        resultMask[upperIndex[badIndexSort[:2]]] = True
        if toPrint:
            logger.warn("\t\t{}_{:6.2f} {} force 2/{} lines in lowerRegion".\
                format(fibInd, yInd, fiberInfo[fibInd].tolist()[:3],
                       upperGoodMask.shape[0], message))
    return resultMask


#region: notes for wave Fit
"""
lam and intensity contains all arc lines
    then they are fillter by inRange
    then they are fillter by forFit = lamp["type"]>=2
    we also have forSigmaFit
oldMu is the origin fit result in pixel
mu is the line used to fit lamp in pixel
oldLam is the origin lamData in astron
lam is the origin lamData used to fit lamp in astron
newCen is (origin lamData to pixel) using fitting coeff
newLam is (origin fit result to astron) using fitting coeff
goodMaskFit is for process profile fit
goodMask is for process lamp fit

final good lines are mu[goodMaskFit][goodMask]
"""
#endregion: notes
def waveFit(im2d, im2dstd, lamp, initGuess, dcoeff, guessStepFactor=None, noGuess=None, configs={}):
    """2d or 1d wave Fit, just finish the 1d part"""
    #region: setups
    # leave these for coding 2d wave calibration
    #ny, nx = im2d.shape
    #nmed= configs.get('nmed', ny/2)
    #row = configs.get('row', ny/2)
    #fibInd = configs.get("fibInd", "?")
    #xx = np.arange(nx, dtype=np.float)
    #spec     = np.median(im2d[int(row-nmed/2):int(row+nmed/2),:], axis=0)
    #specivar = np.median(im2dstd[int(row-nmed/2):int(row+nmed/2),:], axis=0)
    # this is the 1d part, just ignore the names...
    fiberInfo = configs.get("fiberInfo", "???")
    fibInd = configs.get("fibInd", "???")
    yInd   = configs.get("yInd", "???")
    debug  = configs.get("waveFitDebug", 1)
    forceIncludeBoundaryLines = configs.get("forceIncludeBoundaryLines", False)

    FWHMDebug = configs.get("FWHMDebug", 0)
    FWHMType  = configs.get("FWHMType", 'sigma')
    if FWHMType not in ["FWHM", "sigma"]:
        raise Exception("unknown FWHM Type: {}".format(FWHMType))

    ncoeff = configs.get("ncoeff",6)
    waveRef = configs.get("waveRef", 8500)

    debugRMSMax = configs.get("debugRMSMax", 0.05/8500)
    debugRMSMin = configs.get("debugRMSMin", 0.007/8500) # all too large, all too small
    minDataNum = configs.get("minDataNum",20)

    oldRms = configs.get("oldRms", 0) # for second fit
    op = configs.get("op","./")

    if len(im2d.shape)!=1:
        raise Exception("2d wave calibration not finished...")

    #!! get basic info for spec and arc data
    nx = im2d.size
    xx = np.arange(nx)
    spec     = im2d
    specstd = im2dstd
    lam = lamp["lambda"]
    intensity = lamp["intensity"]
    forFit = lamp["type"]>=2
    #xxxxxxx  normalize both spec and intensity for cross match
    #specMax = spec.max()
    #spec /= specMax
    #specstd /= (specMax)
    intensity = intensity/intensity.max()*spec.max()
    skyLinesIndex = np.arange(intensity.size)
    silent = 0
    if oldRms:
        silent=1
    #endregion: setups
    if not np.isfinite(oldRms):
        logger.warn("\t not good fiber in arcfile fitting, skipping this!")
        return np.zeros((ncoeff,)), np.zeros((nx,)),\
               [np.inf, np.array([np.empty((0,)), np.empty((0,)),
                        np.empty((0,)),np.empty((0,)), np.empty((0,))])]
    #region: guess fit
    if noGuess is None:
        if guessStepFactor is not None:
            thisConfig = {"silent":silent}
            thisConfig.update(configs)
            guess, model = waveGuess(spec, lam, intensity, initGuess, dcoeff, stepFactor=guessStepFactor,
                                     configs=thisConfig)
        else:
            thisConfig = {"silent":silent}
            thisConfig.update(configs)
            guess, model = waveGuess(spec, lam, intensity, initGuess, dcoeff,
                                     configs=thisConfig)
    else:
        guess = initGuess.copy()
        # fit by arc
    #!! handle bad fiber guess
    if guess is None:
        logger.p("\t Bad Tracing ")
        return np.zeros((ncoeff,)), np.zeros((nx,)),\
               [np.inf, np.array([np.empty((0,)), np.empty((0,)),
                                  np.empty((0,)),np.empty((0,)), np.empty((0,))])]

    #endregion: guess fit
    #region: setup for fit
    xcenOffset = 3
    xLam = ycoeff2xInd(lam, guess, nx,configs={"ncoeff":ncoeff})
    wavesetGuess = xcoeff2y(np.arange(spec.size), guess)
    inRange = np.logical_and(0+xcenOffset<xLam, xLam<nx-1-xcenOffset)
    xLam = xLam[inRange]# xInd for each Lam
    lam = lam[inRange]# each lambda from the data file
    intensity = intensity[inRange]
    #!! sky mask ++
    skyLinesIndex = skyLinesIndex[inRange]
    forFit = forFit[inRange]
    forSigmaFit = (lamp["type"]>=3)[inRange][forFit]#!! len is forFit
    nLam = lam.size
    xcen = xLam
    #!! piece cut
    #nP = configs.get("nP", 4)
    #pieceMinNum = configs.get("pieceMinNum", 3)
    #logLam = np.log10(lam)
    #logWaveRange = np.log10(xcoeff2y(xx, guess, nx))
    #logmin = logWaveRange.min()
    #logmax = logWaveRange.max()
    #knots = np.linspace(logmin, logmax, nP+1)
    #kStart = knots[:-1]
    #kEnd   = knots[1:]
    #endregion: setup for fit
    #region: do fit profile
    ##!! xcen is in range
    xcenMin = xcen-xcenOffset
    xcenMax = xcen+xcenOffset
    sigma    = np.ones(xcen.shape, dtype=np.float)*2.0
    sigmaMin = np.ones(xcen.shape, dtype=np.float)*1.1
    sigmaMax = np.ones(xcen.shape, dtype=np.float)*2.8
    initA = intensity*np.sqrt(np.pi*2) *sigma
    #if fibInd==5:
        #cdebug=2
    #else:
        #cdebug=-1
    A, mu, sigma, reduceChi2, info =\
            specLib.gaussianFit(spec, specstd,
                     xcen, xcenMin, xcenMax,
                     sigma, sigmaMin, sigmaMax,
                     initA=initA,
                     configs={"cDebug":-1, "ignoreBadFit":1, "nPoly":2,
                               "debugStop":False})

    bmod = info["bmod"]
    A = A[forFit]
    mu = mu[forFit]
    #!! lam in \AA unit
    lam = lam[forFit]
    #!! sky mask ++
    skyLinesIndex = skyLinesIndex[forFit]

    sigma = sigma[forFit]
    #sigSig = info["sigSig"][forFit]
    #sigLam = lam.copy()

    Asig = info["Asig"][forFit]
    muSig = info["muSig"][forFit]
    peakSNR = info["peakSNR"][forFit]
    peakSigSNR = info["peakSigSNR"][forFit]
    fullreject = info["fullreject"][forFit]
    notRejected = (1-fullreject).astype(np.bool)

    lamForSigma = lam[notRejected&forSigmaFit]
    sigmaForSigma = sigma[notRejected&forSigmaFit]
    A = A[notRejected]
    #!! sky mask ++
    skyLinesIndex = skyLinesIndex[notRejected]
    mu = mu[notRejected]
    lam = lam[notRejected]
    Asig = Asig[notRejected]
    muSig = muSig[notRejected]
    peakSNR = peakSNR[notRejected]
    peakSigSNR = peakSigSNR[notRejected]

    #sigLam = sigLam[notRejected&forSigmaFit]
    #sigma  = sigma[notRejected&forSigmaFit]
    #sigSig = sigSig[notRejected&forSigmaFit]

    SNR = A/Asig
    goodMaskSNR = SNR>5
    goodMaskSNRBadness = 5/SNR
    goodMaskMu = muSig<0.25
    goodMaskMuBadness = muSig/0.25*3
    goodMaskPeak = peakSNR>1
    goodMaskPeakBadness = 1/peakSNR
    goodMaskPeakSig = peakSigSNR>5
    goodMaskPeakSigBadness = 5/peakSigSNR
    badness = goodMaskSNRBadness+goodMaskMuBadness+goodMaskPeakBadness+goodMaskPeakSigBadness
    goodMaskFit = goodMaskSNR&goodMaskMu&goodMaskPeak&goodMaskPeakSig
    thisConfigs = {"logger":logger, "fibInd":fibInd,
            "fiberInfo":fiberInfo, "yInd":yInd, "message":"before fitting"}
    if forceIncludeBoundaryLines:
        goodMaskFit = filterGoodMask(goodMaskFit, badness, mu, spec.shape[0], configs=thisConfigs)
    #endregion: do fit profile
    #region:  select good points in each piece
    #for i in range(kStart.size):
        #inRange = np.logical_and(kStart[i]<np.log10(lam),np.log10(lam)<kEnd[i])
        #if inRange.sum()<pieceMinNum:
            #print("too less line ({}) in range {} {}".format(inRange.sum(), kStart[i], kEnd[i]))
            #continue
            ##raise
        #pieceAll = np.where(inRange)[0]
        #pieceNum = goodMaskFit[inRange].sum()
        #if pieceNum>pieceMinNum:
            #continue
        #else:
            #thisMuSig = muSig[inRange]
            #dataToUse = np.argsort(thisMuSig)[:pieceMinNum]
            #goodMaskFit[pieceAll[dataToUse]] = True
    #endregion:  select good points in each piece
    reject = 3
    #region: fit lamp
    #!! filter by SNR
    oldMu = mu.copy()
    oldLam = lam.copy() # lam before SNR select
    mu = mu[goodMaskFit]
    lam = lam[goodMaskFit]
    SNR = SNR[goodMaskFit]
    A = A[goodMaskFit]
    #!! sky mask ++
    skyLinesIndex = skyLinesIndex[goodMaskFit]
    muSig = muSig[goodMaskFit]
    ww = (1/lam**2)
    # use mu as the center of each peak to fit the coeff
    # goodMask should be called goodMaskInd
    thisConfig = {"minDataNum":minDataNum, "ncoeff":ncoeff}
    if forceIncludeBoundaryLines:
        thisConfig = {"minDataNum":minDataNum, "ncoeff":ncoeff,
                   "arrayLength":spec.shape[0],
                   "logger":logger, "fibInd":fibInd,
                   "fiberInfo":fiberInfo,
                   "yInd":yInd}
    coeff, goodMask = xy2coeff(mu, lam, w=ww, xlim=nx, reject=reject,
                               configs=thisConfig)
    if coeff is None:
        logger.warn("\t too few lines for fitting, bad fiber!")
        return np.zeros((ncoeff,)), np.zeros((nx,)),\
               [np.inf, np.array([np.empty((0,)), np.empty((0,)),
                                  np.empty((0,)), np.empty((0,)), np.empty((0,))])]

    # filter by SNR
    mu = mu[goodMask]
    #sigma = sigma[goodMask]
    lam = lam[goodMask]
    A = A[goodMask]
    #!! sky mask ++
    skyLinesIndex = skyLinesIndex[goodMask]
    #sigSig = info["sigSig"][goodMask]
    #sigToUse = sigma[np.sort(np.argsort(sigSig))[:10]]
    lamMod = xcoeff2y(mu, coeff, xlim=nx)
    allLam = xcoeff2y(xx, coeff, xlim=nx)
    rms = np.sqrt(((lamMod-lam)**2 * (1/lam)**2).sum()/goodMask.size)
    #endregion: fit lamp
    #!! here fibInd start from 1
    #region: gen fitted FWHM for latter model
    if FWHMType=="FWHM":
        if FWHMDebug>0:
            FWHMFile = os.path.join(op,"FWHM_{}_{:_^10}_{:_^10.5g}_{:_^10.5g}.fig".\
                    format(fibInd, fiberInfo[fibInd].tolist()[:3], rms*waveRef, rms))
        else:
            FWHMFile = 0
        FWHMArray = FWXM(allLam, spec-bmod, lamForSigma, 5, configs={"debug":FWHMFile})
    else:
        FWHMArray = sigmaForSigma * 2.354820045 # sigma to FWHM for gaussian

    #!! if have oldRms, compile with it and select a good one
    coeffStr = "".join(str(coeff).split("\n"))
    if oldRms:
        if oldRms>rms:
            logger.p("\t{} Tweaking {}: {} lines, use {} rms={:9.8f} rmsOld={:10.8f} ({:10.8f}<{:10.8f} @{:6.2f}) coeff={}".\
                format(fibInd, fiberInfo[fibInd].tolist()[:3], nLam, goodMask.size,
                       rms, oldRms, rms*waveRef, oldRms*waveRef, waveRef, coeffStr))
        else:
            logger.p("\t{} Tweaking {}: {} lines, use {} rms={:9.8f} rmsOld={:10.8f} ({:10.8f}>{:10.8f} @{:6.2f}) coeff={}".\
                format(fibInd, fiberInfo[fibInd].tolist()[:3], nLam, goodMask.size,
                       rms, oldRms, rms*waveRef, oldRms*waveRef, waveRef, coeffStr))
    else:
        if rms<debugRMSMin:
            logger.p("\t rsm={:10.8f}<{:10.8f} ({:10.8f}>{:10.8f} @{:6.2f})".\
                format(rms, oldRms, rms*waveRef, oldRms*waveRef, waveRef))
            logger.p("\tmaybe the init xcen is too small or too large... set rms to 1 to make a larger search")
            rms = 1
        logger.p("\t Tracing {}: {} lines, use {} rsm={:10.8f} ({:10.5f} @{:6.2f}) coeff={}".\
                format(fiberInfo[fibInd].tolist()[:3], nLam, goodMask.size, rms, rms*waveRef, waveRef, coeffStr))
    #endregion: ??

    if rms>debugRMSMax or rms<debugRMSMin:
        if 0<=debug<1:
            debug=1
    if debug>0:
        newCen = ycoeff2xInd(oldLam, coeff, nx, configs={"ncoeff":ncoeff})
        newLam = xcoeff2y(oldMu, coeff, nx)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        xx = np.arange(spec.size)
        lamFit = xcoeff2y(xx, coeff, nx)
        #!! xInd profile
        ax.scatter(xx, spec, c="red")
        ax.plot(xx, spec, c="red")
        ax.plot(xx, info["gmod"])
        ax.plot(xx, info["bmod"])
        ax.plot(xx, info["gmod"]+info["bmod"], c="green")
        #!! wave profile
        ax.scatter(lamFit, spec, c="red")
        ax.plot(lamFit, spec, c="red")
        ax.plot(lamFit, info["gmod"])
        ax.plot(lamFit, info["bmod"])
        ax.plot(lamFit, info["gmod"]+info["bmod"], c="green")
        #!! all fitting profile
        #!! if the profile is used to fit lamp, it is blue, else is red
        for i in range(oldMu.size):
            if goodMaskFit[i]:
                ax.plot([oldMu[i], oldMu[i]], [0, spec[int(oldMu[i])]],"k-")
                ax.plot([newCen[i], newCen[i]], [0, spec.max()/1.3], "b--")
                ax.text(oldMu[i], spec[int(oldMu[i])]*1.3, str(i))
                ax.plot([newLam[i], newLam[i]], [0, spec[int(oldMu[i])]],"k-")
                ax.plot([oldLam[i], oldLam[i]], [0, spec.max()/1.3], "b--")
                ax.text(oldLam[i], spec[int(oldMu[i])]*1.3, str(i))
            else:
                ax.plot([oldMu[i], oldMu[i]], [0, spec[int(oldMu[i])]], "k-")
                ax.plot([newCen[i], newCen[i]], [0, spec.max()/1], "r--")
                ax.text(oldMu[i], spec[int(oldMu[i])]*1, str(i))
                ax.plot([newLam[i], newLam[i]], [0, spec[int(oldMu[i])]], "k-")
                ax.plot([oldLam[i], oldLam[i]], [0, spec.max()/1], "r--")
                ax.text(oldLam[i], spec[int(oldMu[i])]*1, str(i))

        newCen = ycoeff2xInd(lam, coeff, nx, configs={"ncoeff":ncoeff})
        fig.suptitle("{} {} rms={:10.5g} ({:10.5g}@{:6.2g})".format(fibInd, yInd, rms, rms*waveRef, waveRef))
        #!! all fitting lamp
        #!! if the profile is used to fit lamp, and was masked as goodMask, it is green
        for i in range(newCen.size):
            ax.plot([newCen[i], newCen[i]], [0, spec.max()/2.], "g--")
            ax.plot([lam[i], lam[i]], [0, spec.max()/2.], "g--")
        for i in range(mu.size):
            #ax.plot([mu[i], mu[i]], [0, spec[int(mu[i])]])
            ax.text(mu[i], spec[int(mu[i])]*1.5, str(i), horizontalalignment='right')
            #ax.plot([lam[i], lam[i]], [0, spec[int(mu[i])]])
            ax.text(lam[i], spec[int(mu[i])]*1.5, str(i), horizontalalignment='right')
        if debug>1:
            plt.show(0)
        else:
            if oldRms:
                de.d(fig, os.path.join(op,"{}_A_{:6.2f}_{:^3d}_{:_^10.5g}_{:_^10.8g}.fig". \
                                       format(fibInd,yInd,goodMask.size, rms*waveRef, rms)))
            else:
                de.d(fig, os.path.join(op,"{}_{:6.2f}_{:^3d}_{:_^10.5g}_{:_^10.8g}.fig". \
                                       format(fibInd,yInd,goodMask.size, rms*waveRef, rms)))
            fig.clear()
            plt.close()
    thisxmap = xcoeff2y(xx, coeff)
    #return coeff, thisxmap, [rms, np.array([sigLam, sigma, sigSig])]
    #                                                         lam for A, AArray
    return coeff, thisxmap, [rms, np.array([lamForSigma, FWHMArray, lam, A, skyLinesIndex])]

#region: aux functions
def safe_mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)

def readLampData(filename):
    lamData = []
    intensityData=[]
    typeData=[]
    with open(filename,"r") as f:
        for eachline in f:
            eachline = eachline.strip()
            if eachline and eachline[0] not in "#":
                try:
                    temp = eachline.split()
                    lamData.append(temp[0])
                    intensityData.append(temp[1])
                    typeData.append(temp[3])
                except:
                    pass
    N = len(lamData)
    dtype = np.dtype([("lambda", np.float64),("intensity", np.float64),("type", np.int32)])
    result = np.zeros(N, dtype=dtype)
    result["lambda"] = np.array(lamData, dtype=np.float64)
    result["intensity"] = np.array(intensityData, dtype=np.float64)
    result["type"] = np.array(typeData, dtype=np.int32)
    return result
    #result = np.loadtxt(filename,
    #                    dtype={'names':('lambda', 'intensity', 'name'),
    #                           'formats':('f8','f8','S10')})
    #lam = result['lambda']
    #intensity = result['intensity']
    #touse = lam>7000
    #lam = lam[touse]
    #intensity = intensity[touse]
    #notForFit = intensity<0
    #intensity[notForFit] = -intensity[notForFit]
    #return lam[touse], intensity[touse], np.logical_not(notForFit)

def readWaveSet(outputPath, name=None, debugInfo=None, configs={}):
    yset, badFiber, boundary = extract.readTraceSet(outputPath)
    if name is None:
        wavesetFile = os.path.join(outputPath, "waveset.fits")
        obj = pyfits.open(wavesetFile)
        waveset = obj[0].data
        coeffs = obj[1].data
    else:
        wavesetFile = os.path.join(outputPath, name+"-waveset.fits")
        traceFile = os.path.join(outputPath, name+"Flux.fits")
        obj = pyfits.open(traceFile)
        ysetSmooth = obj["xcenSmooth"].data
        yset[~badFiber,:] = ysetSmooth[:,:]
        obj = pyfits.open(wavesetFile)
        waveset = obj[0].data
        coeffs = obj[1].data

    if debugInfo is None:
        return waveset, coeffs, yset, badFiber
    elif debugInfo==1:
        debugInfo = {"rms":obj[2].data}
        return waveset, coeffs, yset, badFiber, debugInfo
    elif debugInfo==2: #!! ????????
        debugInfo = {"rms":obj[2].data, "smoothA":obj[3].data}
        return waveset, coeffs, yset, badFiber, debugInfo
    else:
        raise Exception("unknown debugInfo flag {]".format(debugInfo))

def readWaveExtract(outputPath, configs={}):
    infile = os.path.join(outputPath, "masterArcFlux.fits")
    obj = pyfits.open(infile)
    yset, badFiber = extract.readTraceSet(outputPath)
    flux = obj[0].data
    ivar = obj[1].data
    assert flux.shape[0]==len(badFiber) - badFiber.sum()
    return flux, ivar, yset, badFiber.astype(np.bool)
#endregion: aux functions

#region: fitting functions
def FWXM(x, y, xcen, halfPixelSearch, percent=50, configs={}):
    """search fro Full Width X percent Maximum, default of X is 50
       data should be subtracted by the continum
       xcen is the position where you want to search for the peak and the width
    """
    assert x.shape==y.shape
    debug = configs.get("debug", 0)
    maxIter = configs.get("maxIter", 50)
    epsFactor = configs.get("epsFactor", 0.00001)
    result = np.zeros(xcen.size)
    xlArray = np.zeros(xcen.size)
    xhArray = np.zeros(xcen.size)
    valueArray = np.zeros(xcen.size)
    nx = x.size
    allInds = np.arange(nx)
    #f = interpolate.interp1d(x, y, kind="slinear")
    f = interpolate.interp1d(x, y, kind="cubic")
    peakOffset = configs.get("peakOffset", 2)
    for i in range(xcen.size):
        xcenInd = np.argsort(np.abs(xcen[i]-x))[0]
        xpeakStart = [xcenInd-peakOffset, 0][xcenInd-peakOffset<0]
        xpeakEnd = [xcenInd+peakOffset, nx-1][xcenInd+peakOffset>nx-1]
        worky =       y[xpeakStart:xpeakEnd+1]
        workx = allInds[xpeakStart:xpeakEnd+1]
        maxInd = np.argsort(worky)[-1]
        maxValue = worky[maxInd]
        xcenInd = workx[maxInd]
        startInd = xcenInd - halfPixelSearch
        startInd = [startInd, 0][startInd<0]
        mono = np.diff(y[xcenInd:startInd-1:-1])>0
        if mono.sum()>0:
            startInd = xcenInd - np.where(mono)[0][0] - 1
        endInd = xcenInd + halfPixelSearch
        endInd = [endInd, nx - 1][endInd>nx-1]
        mono = np.diff(y[xcenInd:endInd+1])>0
        if mono.sum()>0:
            endInd = xcenInd + np.where(mono)[0][0] + 1
        workValue = maxValue * percent / 100.
        eps = epsFactor * maxValue
        xll = x[startInd]
        xlh = x[xcenInd]
        xlm = (xlh+xll)/2.0
        xhl = x[xcenInd]
        xhh = x[endInd]
        xhm = (xhh+xhl)/2.0
        ldiff = f(xlm) - workValue
        hdiff = f(xhm) - workValue
        niter = 0
        while niter<maxIter and (np.abs(ldiff)>eps or np.abs(hdiff)>eps):
            niter += 1
            if ldiff>eps:
                xlh = xlm
                xlm = (xlh+xll)/2.0
                ldiff = f(xlm) - workValue
            elif ldiff<-eps:
                xll = xlm
                xlm = (xlh+xll)/2.0
                ldiff = f(xlm) - workValue
            if hdiff>eps:
                xhl = xhm
                xhm = (xhh+xhl)/2.0
                hdiff = f(xhm) - workValue
            elif hdiff<-eps:
                xhh = xhm
                xhm = (xhh+xhl)/2.0
                hdiff = f(xhm) - workValue
        result[i]= xhm - xlm
        xlArray[i] = xlm
        xhArray[i] = xhm
        valueArray[i] = workValue
    if debug:
        fig = de.f()
        ax = fig.add_subplot(111)
        ax.plot(x, y, c="red")
        for i in range(xcen.size):
            ax.plot([xcen[i], xcen[i]], [0, y.max()], "b--")
            ax.plot([xlArray[i], xlArray[i]], [0, y.max()], "g--")
            ax.plot([xhArray[i], xhArray[i]], [0, y.max()], "r--")
            ax.plot([xlArray[i], xhArray[i]], [valueArray[i]] * 2, "k-")
        if isinstance(debug,str):
            de.d(fig, debug)
            fig.clear()
            plt.close(fig)

    return result

# all these three function assume x to be [0, nx-1]
# and y to be some mono function!!!
def xy2coeff(x, y, w=None, reject=0, xlim=None, sigmaOutput=None, configs={}):
    minDataNum = configs.get("minDataNum", 4)
    arrayLength = configs.get("arrayLength", 0)
    logger = configs.get("logger")
    fibInd = configs.get("logger")
    fiberInfo = configs.get("fiberInfo")
    yInd = configs.get("yInd")

    # setups
    assert x.shape==y.shape
    if xlim is None:
        xmin = x.min()
        xmax = x.max()
    elif type(xlim)==int:
        xmin = 0
        xmax = xlim - 1
    elif len(xlim)==2: # list of lenght 2
        xmin = xlim[0]
        xmax = xlim[1]
    else:
        raise Exception("unknown xlim input of {}").format(xlim)
    xWidth = float(xmax - xmin)
    #xArray = 2.0*x/(nx - 1.0) - 1.0
    xArray = (x - xmin)/xWidth*2.0-1.0

    maxiter = configs.get("maxiter", 20)
    ncoeff  = configs.get("ncoeff", 6)
    deg     = ncoeff - 1
    # begin fit of different type
    niter = 0
    goodMask = np.arange(x.size, dtype=np.int)
    #!! handle bad fiber
    if x.size<minDataNum:
        print("too few good points({}) for this fit!!\n".format(x.size))
        return None, None
    if reject > 0:
        if w is None :
            w = np.ones(x.size)
        if not np.all(w>=0):
            print(w)
            print(w>=0)
            raise Exception("?")
        lastxArray = xArray.copy()
        lasty = y.copy()
        lastgoodMask = goodMask.copy()
        lastw = w.copy()
        toPrint = 1
        while niter<=maxiter:
            niter+=1
            leg = Legendre.fit(xArray, y, deg, domain=[-1,1], w=w)# mind the domain!!!
            diff = (y-leg(xArray)) * w
            std = np.std(diff)
            good = (np.abs(diff)/std)<reject
            badness =(np.abs(diff)/std)
            if arrayLength:
                thisConfigs = {"logger":logger, "fibInd":fibInd, "toPrint":toPrint,
                            "fiberInfo":fiberInfo, "yInd":yInd, "message":"in fitting"}
                lastGood = good.copy()
                good = filterGoodMask(good, badness, xArray, arrayLength, configs=thisConfigs)
                if good.sum()!=lastGood.sum():
                    toPrint -= 1
            if good.sum()<minDataNum:
                #raise Exception("too few points({}) for this fit!!".format(good.sum()))
                print("too few good points({}) for this fit!!\n".format(good.sum()))
                xArray = lastxArray
                y = lasty
                w = lastw
                return None, None
                #leg = Legendre.fit(xArray, y, deg, domain=[-1,1], w=w)# mind the domain!!!
                #return leg.coef, lastgoodMask
            if (~good).sum()==0:
                break
            else:
                lastxArray = xArray.copy()
                lasty = y.copy()
                lastw = w.copy()
                lastgoodMask = goodMask.copy()
                goodMask = goodMask[good]
                xArray = xArray[good]
                y = y[good]
                w = w[good]
        if sigmaOutput is None:
            return leg.coef, goodMask
        else:
            #sigma2Leg = Legendre.fit(xArray, diff**2, deg, domain=[-1,1])
            sigma2Leg = Legendre.fit(xArray, ((y-leg(xArray))*w)**2, deg, domain=[-1,1])
            return leg.coef, goodMask, sigma2Leg.coef
    else:
        if w is None:
            leg = Legendre.fit(xArray, y, deg, domain=[-1,1])
        else:
            leg = Legendre.fit(xArray, y, deg, domin=[-1,1], w=w)
        return leg.coef

def xcoeff2y(x, coeff, xlim=None):
    if xlim is None:
        xmin = x.min()
        xmax = x.max()
    elif type(xlim)==int:
        xmin = 0
        xmax = xlim - 1
    elif len(xlim)==2: # list of lengh 2
        xmin = xlim[0]
        xmax = xlim[1]
    else:
        raise Exception("unknown xlim input of {}").format(xlim)
    xWidth = float(xmax - xmin)
    xArray = (x - xmin)/xWidth*2-1
    return Legendre(coeff)(xArray)

def ycoeff2xInd(y, coeff, nx, configs={}):
    ''' x~y should be mono to have a correct answer '''
    xInd = np.arange(nx)
    xArray = 2.0*xInd/(nx - 1.0) - 1.0
    yArray = Legendre(coeff)(xArray)
    invCoeff = xy2coeff(yArray, xInd, configs=configs)
    yWidth = yArray.max() - yArray.min()
    ymid   = (yArray.max() + yArray.min())/2.0
    y = (y-ymid)/yWidth * 2
    return Legendre(invCoeff)(y)
#endregion: fitting functions

def waveGuess(spec, lam, intensity, guess, dcoeff, stepFactor=None, configs={}):
    #region: setups
    loglam = np.log10(lam)
    nGuess = configs.get('nGuess', 5)
    silent = configs.get("silent", 0)
    nsteps = np.ones(dcoeff.shape[0])
    fibInd = configs.get("fibInd","?")
    yInd   = configs.get("yInd","?")
    maxLag1 = configs.get("maxLag1", 10)
    maxLag2 = configs.get("maxLag2", 5)
    logger = configs["logger"]
    # first guess
    nsteps[0] = 50
    nsteps[1] = 10
    nsteps[2] = 6
    if stepFactor is not None:
        nsteps = (nsteps*stepFactor).astype(np.int)
    thisConfig = configs
    thisConfig.update({'nsmooth': 3, 'maxLag':maxLag1})
    guess, bestcorr, bestmodel = waveGuessIter(spec, lam, intensity, guess, dcoeff, nsteps, configs=thisConfig)
    if guess is None:
        logger.warn('wavecal fiber {:3} @ {:.5}: {}'.format(fibInd, yInd, "bad guess!!"))
        return None, None
    # second guess
    dcoeff   /= 5.0
    nsteps[0] = 10
    nsteps[1] = 10
    nsteps[2] = 5
    nsteps[3] = 5
    if stepFactor is not None:
        nsteps = nsteps*stepFactor
    thisConfig = configs
    thisConfig.update({'nsmooth': 1, 'maxLag':maxLag2})
    #thisConfig = {'nsmooth': 1, 'maxLag':maxLag2}
    #endregion: setups
    guess, bestcorr, bestmodel = waveGuessIter(spec, lam, intensity, guess, dcoeff, nsteps, configs=thisConfig)
    if guess is None:
        logger.warn('wavecal fiber {:3} @ {:.5}: {}'.format(fibInd, yInd, "bad guess!!"))
        return None, None
    guessStr = "".join(str(guess).split("\n"))
    if not silent:
        logger.p('wavecal fiber {:3} @ {:.5}: bestcorr: {:10.5f} {}'.format(fibInd, yInd, bestcorr, guessStr))
    else:
        logger.log('wavecal fiber {:3} @ {:.5}: bestcorr: {:10.5f} {}'.format(fibInd, yInd, bestcorr, guessStr))
    return guess, bestmodel

def waveGuessIter(spec, lam, intensity, guess, dcoeff, nsteps, configs={}):
    #region: setups
    nsmooth = configs.get("nsmooth", 4)
    npix = spec.size
    medianFilterSize = configs.get('medianFilterSize', 101)
    maxLag = configs.get("maxLag", 20)
    dlag = configs.get("dlag",1)
    nlag = int(maxLag/dlag)
    lagRange = np.linspace(0, 2*maxLag, 2*nlag + 1) - maxLag
    logger = configs["logger"]
    ncoeff = dcoeff.shape[0]
    nKernel = 4*nsmooth + 1
    gausskern = np.exp(- ((nKernel-1)/2. - np.arange(nKernel))**2 / nsmooth**2)
    gausskern = gausskern / gausskern.sum()
    npad = int((nKernel+1)/2) # padding on left and right of spectra before cross-correlating
    lampBuildSmooth = float(configs.get("lampmBuildSmooth", 0.6))
    if maxLag>npad:
        npad = maxLag
    pixarray = np.arange(-npad, npix+npad) #!! use origin pixarray
    speccorr = spec.copy()
    medval1 = np.median(spec[0:[npix, 50][npix > 50]]) # median on the left side
    medval2 = np.median(spec[[npix-50, 0][npix < 50]:npix])# median on the right side
    speccorr = np.concatenate((np.zeros(npad)+medval1,
                               speccorr,
                               np.zeros(npad)+medval2))
    #speccorr = speccorr - scipy.ndimage.median_filter(speccorr, size=[101,npix][npix>101])
    speccorr = speccorr - scipy.ndimage.median_filter(speccorr, size=medianFilterSize)
    speccorr = np.convolve(speccorr, gausskern, 'same')
    speccorr[speccorr<0]=0
    speccorr = np.sqrt(speccorr)
    dcoeffList = [[np.array([0]),
                   np.linspace(dcoeff[i][0], dcoeff[i][1], n)][n>1]
                  for i, n in enumerate(nsteps)]
    #!! debug notes: de.p(speccorr[npad:-npad],"r");de.p(spec,"g")
    prodp = int(np.prod(nsteps))
    printFlag = int(prodp/100)
    #toMod = np.cumprod(nsteps).astype(np.int).tolist()
    #toDiv = [1]+toMod[:-1]
    # do correlation with lag for each guess model
    #endregion: setups
    bestcorr = 0
    if configs.get("waveGuessDebugBefore", False):
        #corrval, lag = specLib.maxCorrelate(speccorr, model, lagRange)
        i = 0
        thisStep = specLib.getSubIndex(i, nsteps)
        thisCoeff = np.array([guess[ii] + dcoeffList[ii][thisStep[ii]] for ii in range(ncoeff)])
        leg = Legendre(thisCoeff, domain=[0, npix])
        modLam = leg(pixarray)# lambda at each xInd
        model = specLib.buildSmoothLamp(lam, intensity, modLam, lampBuildSmooth)
        model = np.convolve(model, gausskern, 'same')
        model = np.sqrt(model)
        de.plt.switch_backend(configs["plotBackend"])
        print(prodp)
        print("doing waveGuessDebug")
        fig = plt.figure()
        plt.plot(spec)
        fig.subplots_adjust(bottom=0.07, top=1, left = 0.03, right=1)
        fig = plt.figure()
        fig.subplots_adjust(bottom=0.07, top=1, left = 0.03, right=1)
        ax1 = fig.add_subplot(111)
        ax1.plot(modLam, model, 'g')
        ax2 = ax1.twiny()
        ax2.plot(modLam, speccorr,'r')
        plt.show(0)
        ipdb.set_trace()
    for i in range(prodp):
        #thisStep = [int((i%toMod[ii])/toDiv[ii]) for ii in range(ncoeff-1,-1,-1)][::-1]
        thisStep = specLib.getSubIndex(i, nsteps)
        thisCoeff = np.array([guess[ii] + dcoeffList[ii][thisStep[ii]] for ii in range(ncoeff)])
        #!! construct the simulated arc spectrum
        leg = Legendre(thisCoeff, domain=[0, npix])
        modLam = leg(pixarray)# lambda at each xInd
        #!! this part has been reimplemented by C in specLib.buildLamp
        #region: build in c
        #model = np.zeros(npix+2*npad)
        #if logmodLam[1]>logmodLam[0]: #accending
        #    for iline in range(nline):
        #        iloc = np.argmax(loglam[iline] < logmodLam)-1
        #        dx = loglam[iline] - logmodLam[iloc]
        #        if iloc>0 and iloc<npix-2:
        #            dpix = logmodLam[iloc+1] - logmodLam[iloc]
        #            model[npad+iloc] += intensity[iline] * (1-dx/dpix)
        #            model[npad+iloc+1] += intensity[iline] * dx/dpix
        #else: #decending
        #    for iline in range(nline):
        #        iloc = np.argmax(loglam[iline] > logmodLam)-1
        #        dx = logmodLam[iloc] - loglam[iline]
        #        if iloc>0 and iloc<npix-2:
        #            dpix = logmodLam[iloc] - logmodLam[iloc+1]
        #            model[npad+iloc] += intensity[iline] * (1-dx/dpix)
        #            model[npad+iloc+1] += intensity[iline] * dx/dpix
        #model = np.convolve(model, gausskern, 'same')
        #model = np.sqrt(model)
        #endregion: build in c
        #!! allocate each flux to each (two) xInds
        #model = specLib.buildLamp(loglam, intensity, logmodLam, npad)
        model = specLib.buildSmoothLamp(lam, intensity, modLam, lampBuildSmooth)
        model = np.convolve(model, gausskern, 'same')
        model = np.sqrt(model)
        #corrval = np.correlate(speccorr, model)[0]
        #!! debug notes:de.p(modLam[npad:-npad],speccorr[npad:-npad],"r");de.p(modLam[npad:-npad],model[npad:-npad],"g");de.p(modLam[npad:-npad],spec)
        corrval, lag = specLib.maxCorrelate(speccorr, model, lagRange)
        if corrval>bestcorr:
            bestcorr = corrval
            bestcoeff = thisCoeff
            bestcoeff[0] -= lag
            bestlag = lag
            bestmodel = model
            #if i%printFlag==0:
            #    print i/printFlag, bestcorr, bestcoeff
    if configs.get("waveGuessDebugAfter", False):
        #corrval, lag = specLib.maxCorrelate(speccorr, model, lagRange)
        de.plt.switch_backend(configs["plotBackend"])
        print(prodp)
        print("doing waveGuessDebugAfter")
        print(bestcoeff)
        fig = plt.figure()
        plt.plot(spec)
        fig.subplots_adjust(bottom=0.07, top=1, left = 0.03, right=1)
        fig = plt.figure()
        fig.subplots_adjust(bottom=0.07, top=1, left = 0.03, right=1)
        plt.plot(modLam, bestmodel, 'g')
        plt.plot(modLam, speccorr,'r')
        plt.show(0)
        ipdb.set_trace()
    if bestcorr==0:
        return None, None, None

    return bestcoeff, bestcorr, bestmodel[npad-bestlag:-npad-bestlag]

def skyWaveTweak(inputFile, configs={}):
    #region: setups
    inputFileName = os.path.basename(inputFile).split(".fits")[0]
    outputFileName= inputFileName + '-waveset.fits'
    configs.update({"config2show":["rerun", "ncoeff", "initRow", "debug"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return
    configs["logger"] = logger
    #endregion: setups
    #de.plt.switch_backend("Agg")
    yset, badFiber, boundary = extract.readTraceSet(outputPath, configs=configs)
    if len(configs["fiberInfo"])!=(~badFiber).sum():
        configs["fiberInfo"] = configs["fiberInfo"][~badFiber]
    flux, std, yset = extract.readFlux(inputFile.split(".fits")[0]+"Flux.fits", configs=configs)

    obj = pyfits.open(os.path.join(outputPath, "waveset.fits"))
    xmap, coeffs, rms = obj[0].data, obj[1].data, obj[2].data
    #!! read flat data (do the same thing in skysub)
    masterFlatFieldFile = os.path.join(outputPath, "masterFlatField.fits")
    obj = pyfits.open(masterFlatFieldFile)
    fiberFlatSmooth, allFactor = obj[0].data, obj[1].data
    #assert fiberFlatSmooth.shape[0]==allFactor.shape[0]==flux.shape[0]
    factor = 1 / fiberFlatSmooth * allFactor.reshape((allFactor.size, 1))
    allFactors = np.zeros(flux.shape)
    allFactors[np.isfinite(rms.flatten()),:]=factor[:,:]
    flux = flux * allFactors
    std = std * allFactors
    std[~np.isfinite(rms.flatten()),:] = 0

    configs.update({"inputFileName":inputFileName})
    allCoeff, xmap, allRMS, smoothA = waveTweak(flux, std, yset, xmap, coeffs, rms, configs=configs)

    #de.plt.switch_backend(configs["plotBackend"])
    #region: save file
    h1 = pyfits.Header()
    h2 = pyfits.Header()
    h3 = pyfits.Header()
    h4 = pyfits.Header()
    h1["OBJECT"] = "xmap"
    h1["EXTNAME"] = "xmap"
    h2["OBJECT"] = "coeff"
    h2["EXTNAME"] = "coeff"
    h3["OBJECT"] = "RMS"
    h3["EXTNAME"] = "RMS"
    h4["OBJECT"] = "scale residual after flux"
    h4["EXTNAME"] = "smoothA"
    data1 = pyfits.PrimaryHDU(xmap, header=h1)
    data2 = pyfits.ImageHDU(allCoeff, header=h2)
    data3 = pyfits.ImageHDU(allRMS, header=h3)
    data4 = pyfits.ImageHDU(smoothA, header=h4)
    fits  = pyfits.HDUList([data1, data2, data3, data4])
    fits.writeto(outFile, clobber='True')
    #endregion: save file

def waveTweak(flux, std, yset, xmapOld, coeffs, rms, configs={}):
    #region: setups
    assert flux.shape==yset.shape
    nFiber, nx = yset.shape
    #!! ny to be auto read latter....
    ny = configs.get("ny", 2056)
    ysetInd = configs.get("ysetInd", 1024)
    #!! read sky data
    skytype="selected";
    thisConfig = {"skyDataFile":"m2fsSkyLines.dat",
               "skytype":skytype, "codePath":configs["codePath"]}
    skyData = readSkyData(configs=thisConfig)
    #lampDataFile = configs.get('lampDataFile', 'NistArNe.m2fs.dat')
    #lampDataFile = os.path.join(codePath, 'data', lampDataFile)
    #lamp = readLampData(lampDataFile)
    allSkylineA = np.zeros((nFiber, skyData.size))*np.nan
    allSkylineLam = np.zeros((nFiber, skyData.size))*np.nan

    codePath = configs["codePath"]
    logger = configs["logger"]
    ncoeff = configs.get('ncoeff', 6)
    debug = configs.get("debug", 0)

    waveFitDebug = configs.get("waveFitDebug", -1)
    FWHMDebug = configs.get("FWHMDebug",0)
    FWHMType = configs.get("FWHMType", "sigma")

    dcoeff = np.zeros((ncoeff,2), dtype=np.float)
    initRow = configs.get("initRow", int(nFiber/2)) # from the middle one
    allCoeffs=np.zeros((nFiber, ncoeff))
    yFiber = yset[:,0]
    xmap=np.zeros((nFiber, nx))
    allRMS=np.ones((nFiber, 1)) * np.inf
    outputPath = configs["outputPath"]
    inputFileName = configs["inputFileName"]
    op = os.path.join(outputPath, inputFileName+"-wavesetDebug")
    codePath = configs["codePath"]
    sigmaInfo=np.zeros(nFiber, dtype=np.object)
    guess = np.zeros(ncoeff, dtype=np.float)
    #endregion: setups
    #region: setup init guess
    #!! fit for smooth guess coeffs in case of bad fiber
    smoothGuess = np.zeros((nFiber,ncoeff), dtype=np.float)
    for i in range(ncoeff):
        obj = Legendre.fit(yFiber, coeffs[:,i], 4, w=1/rms.flatten())
        smoothGuess[:,i] = obj(yFiber)[:]

    #!! init guess parameters, should be written to a template file
    dcoeff[0][0] = -0.5
    dcoeff[0][1] =  0.5
    dcoeff[1][0] = -2
    dcoeff[1][1] =  2
    dcoeff[2:,0] = -1
    dcoeff[2:,1] =  1
    maxLag1 = 50
    maxLag2 = 5
    #endregion: setup init guess
    #region: fit for the middle one
    guess  = smoothGuess[initRow].copy()
    thisConfig = {"fibInd":initRow, "yInd":yset[initRow][0],
                 "maxLag1":maxLag1, "maxLag2":maxLag2}
    thisConfig.update(configs)
    thisConfig["op"] = op

    thisConfig["waveFitDebug"] = waveFitDebug
    thisConfig["FWHMDebug"] = FWHMDebug
    thisConfig["FWHMType"] = FWHMType
    thisConfig["oldRms"] = rms[initRow][0]

    #!! fit coeff for this fiber
    thisCoeff, thisMap, thisDebugInfo =\
        waveFit(flux[initRow], std[initRow], skyData, guess, dcoeff.copy(), configs=thisConfig)

    allCoeffs[initRow,:] = thisCoeff[:]
    xmap[initRow,:] = thisMap[:]
    allRMS[initRow,:] = thisDebugInfo[0]
    lastRMS = thisDebugInfo[0]
    sigmaInfo[initRow] = thisDebugInfo[1]
    if np.isfinite(thisDebugInfo[0]):
        thisIndex = thisDebugInfo[1][4].astype(np.int)
        allSkylineA[initRow, thisIndex] = thisDebugInfo[1][3][:]
        allSkylineLam[initRow, thisIndex] = thisDebugInfo[1][2][:]

    if (not np.isfinite(thisDebugInfo[0])) and (np.isfinite(thisConfig["oldRms"])):
        raise logger.Exception("arc good but tweak bad at init row {}, you should change to another init fiber".format(initRow))
        #allCoeffs[initRow,:] = coeffs[initRow,:]
        #xmap[initRow,:] = xmapOld[initRow,:]
        #allRMS[initRow,:] = rms[initRow,:]
        #lastRMS = rms[initRow,:]
        ##!! cution!!! here sigmaInfo may be bad....
        #sigmaInfo[initRow] = thisDebugInfo[1]
        #if np.isfinite(thisDebugInfo[0]):
            #thisIndex = thisDebugInfo[1][4].astype(np.int)
            #allSkylineA[initRow, thisIndex] = thisDebugInfo[1][3][:]
            #allSkylineLam[initRow, thisIndex] = thisDebugInfo[1][2][:]

    #!! handle bad fiber in the initRow
    if not np.isfinite(lastRMS):
        raise logger.Exception("the first fiber must be good, you should change the initRow config in the [findWave] section to fibers number other than {} and less than {}".format(initRow, nFiber))

    #!! not good fit for the first fiber, guess more and try again
    if lastRMS>0.1:
        newDcoeff = dcoeff.copy()
        newDcoeff[0][0] = -0.5
        newDcoeff[0][1] =  0.5
        newDcoeff[1][0] = -6
        newDcoeff[1][1] =  6
        maxLag1 = 300
        maxLag2 = 10
        gs=np.ones(ncoeff)
        gs[0] = 3
        gs[1] = 3
        thisConfig = {"fibInd":initRow, "yInd":yset[initRow][0],
                      "maxLag1":maxLag1, "maxLag2":maxLag2}
        thisConfig.update(configs)
        thisConfig["op"] = op
        thisConfig["waveFitDebug"] = waveFitDebug
        thisConfig["FWHMDebug"] = FWHMDebug
        thisConfig["FWHMType"] = FWHMType
        thisConfig["oldRms"] = rms[initRow][0]
        thisCoeff, thisMap, thisDebugInfo =\
            waveFit(flux[initRow],std[initRow], skyData, guess, dcoeff.copy(),
                    guessStepFactor=gs, configs=thisConfig)
        allCoeffs[initRow,:] = thisCoeff[:]
        xmap[initRow,:] = thisMap[:]
        allRMS[initRow,:] = thisDebugInfo[0]
        lastRMS = thisDebugInfo[0]
        sigmaInfo[initRow] = thisDebugInfo[1]
        if np.isfinite(thisDebugInfo[0]):
            thisIndex = thisDebugInfo[1][4].astype(np.int)
            allSkylineA[initRow, thisIndex] = thisDebugInfo[1][3][:]
            allSkylineLam[initRow, thisIndex] = thisDebugInfo[1][2][:]
        if (not np.isfinite(thisDebugInfo[0])) and (np.isfinite(thisConfig["oldRms"])):
            logger.warn("arc good but tweak bad at {}".format(initRow))
            allCoeffs[initRow,:] = coeffs[initRow,:]
            xmap[initRow,:] = xmapOld[initRow,:]
            allRMS[initRow,:] = rms[initRow,:]
            lastRMS = rms[initRow,:]
            #!! cution!!! here sigmaInfo may be bad....
            sigmaInfo[initRow] = thisDebugInfo[1]
            if np.isfinite(thisDebugInfo[0]):
                thisIndex = thisDebugInfo[1][4].astype(np.int)
                allSkylineA[initRow, thisIndex] = thisDebugInfo[1][3][:]
                allSkylineLam[initRow, thisIndex] = thisDebugInfo[1][2][:]
        #!! handle bad fiber in the initRow
        if not np.isfinite(lastRMS):
            raise logger.Exception("the first fiber must be good, you should change the initRow config in the [findWave] section to fibers number other than {} and less than {}".format(initRow, nTrace))
    #endregion: fit for the middle one
    maxLag1 = 10
    maxLag2 = 5
    #region: fit for lower fibers
    for thisRow in range(initRow-1, -1, -1):
        newGuess  = smoothGuess[thisRow].copy()
        factor = 1
        gs=np.ones(ncoeff)
        thisdcoeff = dcoeff.copy()

        thisConfig = {"fibInd":thisRow, "yInd":yset[thisRow][0], "forceIncludeBoundaryLines":True,
                 "maxLag1":int(maxLag1*factor), "maxLag2":int(maxLag2*factor)}
        thisConfig.update(configs)
        thisConfig["op"] = op
        thisConfig["waveFitDebug"] = waveFitDebug
        thisConfig["FWHMDebug"] = FWHMDebug
        thisConfig["FWHMType"] = FWHMType
        thisConfig["oldRms"] = rms[thisRow][0]
        thisCoeff, thisMap, thisDebugInfo =\
            waveFit(flux[thisRow], std[thisRow], skyData, newGuess, thisdcoeff,
                    guessStepFactor=gs,configs=thisConfig)
        allCoeffs[thisRow,:] = thisCoeff[:]
        xmap[thisRow,:] = thisMap[:]
        allRMS[thisRow,:] = thisDebugInfo[0]
        lastRMS = thisDebugInfo[0]
        sigmaInfo[thisRow] = thisDebugInfo[1]
        if np.isfinite(thisDebugInfo[0]):
            thisIndex = thisDebugInfo[1][4].astype(np.int)
            allSkylineA[thisRow, thisIndex] = thisDebugInfo[1][3][:]
            allSkylineLam[thisRow, thisIndex] = thisDebugInfo[1][2][:]
        if (not np.isfinite(thisDebugInfo[0])) and (np.isfinite(thisConfig["oldRms"])):
            logger.warn("arc good but tweak bad at {}".format(thisRow))
            #allCoeffs[thisRow,:] = coeffs[thisRow,:]
            #xmap[thisRow,:] = xmapOld[thisRow,:]
            #allRMS[thisRow,:] = rms[thisRow,:]
            #lastRMS = rms[thisRow,:]
            ##!! cution!!! here sigmaInfo may be bad....
            #sigmaInfo[thisRow] = thisDebugInfo[1]
            #if np.isfinite(thisDebugInfo[0]):
                #thisIndex = thisDebugInfo[1][4].astype(np.int)
                #allSkylineA[thisRow, thisIndex] = thisDebugInfo[1][3][:]
                #allSkylineLam[thisRow, thisIndex] = thisDebugInfo[1][2][:]
    #endregion: fit for lower fibers
    #region: fit for upper fibers
    for thisRow in range(initRow+1, nFiber):
        newGuess  = smoothGuess[thisRow].copy()
        factor = 1
        gs=np.ones(ncoeff)
        thisdcoeff = dcoeff.copy()

        thisConfig = {"fibInd":thisRow, "yInd":yset[thisRow][0], "forceIncludeBoundaryLines":True,
                      "maxLag1":int(maxLag1*factor), "maxLag2":int(maxLag2*factor)}
        thisConfig.update(configs)
        thisConfig["op"] = op
        thisConfig["waveFitDebug"] = waveFitDebug
        thisConfig["FWHMDebug"] = FWHMDebug
        thisConfig["FWHMType"] = FWHMType
        thisConfig["oldRms"] = rms[thisRow][0]
        thisCoeff, thisMap, thisDebugInfo = \
            waveFit(flux[thisRow], std[thisRow], skyData, newGuess, thisdcoeff,
                    guessStepFactor=gs,configs=thisConfig)
        allCoeffs[thisRow,:] = thisCoeff[:]
        xmap[thisRow,:] = thisMap[:]
        allRMS[thisRow,:] = thisDebugInfo[0]
        lastRMS = thisDebugInfo[0]
        sigmaInfo[thisRow] = thisDebugInfo[1]
        if np.isfinite(thisDebugInfo[0]):
            thisIndex = thisDebugInfo[1][4].astype(np.int)
            allSkylineA[thisRow, thisIndex] = thisDebugInfo[1][3][:]
            allSkylineLam[thisRow, thisIndex] = thisDebugInfo[1][2][:]
        if (not np.isfinite(thisDebugInfo[0])) and (np.isfinite(thisConfig["oldRms"])):
            logger.warn("arc good but tweak bad at {}".format(thisRow))
            #allCoeffs[thisRow,:] = coeffs[thisRow,:]
            #xmap[thisRow,:] = xmapOld[thisRow,:]
            #allRMS[thisRow,:] = rms[thisRow,:]
            #lastRMS = rms[thisRow,:]
            ##!! cution!!! here sigmaInfo may be bad....
            #sigmaInfo[thisRow] = thisDebugInfo[1]
            #if np.isfinite(thisDebugInfo[0]):
                #thisIndex = thisDebugInfo[1][4].astype(np.int)
                #allSkylineA[thisRow, thisIndex] = thisDebugInfo[1][3][:]
                #allSkylineLam[thisRow, thisIndex] = thisDebugInfo[1][2][:]
    #endregion: fit for upper fibers
    #region: finish
    sigmaInfo.dump(os.path.join(op,"skyLineModel.dat"))
    c = cm.get_cmap("rainbow")
    c = c(np.linspace(0,1,nFiber))
    #!! fwhm plot
    fig=de.f()
    ax = fig.add_subplot(111)
    for i, eachInfo in enumerate(sigmaInfo):
        ax.plot(eachInfo[0], eachInfo[1], c=c[i])
        ax.scatter(eachInfo[0], eachInfo[1], c=c[i], s=8)

    de.d(fig, os.path.join(op,"-FWHM.fig"))
    fig.clear()
    plt.close(fig)
    #!! skyLine flux plot
    fig=de.f()
    ax = fig.add_subplot(111)
    for i, eachInfo in enumerate(sigmaInfo):
        ax.plot(eachInfo[2], eachInfo[3], c=c[i])
        ax.scatter(eachInfo[2], eachInfo[3], c=c[i], s=8)

    de.d(fig, os.path.join(op,"-skyLineRaw.fig"))
    fig.clear()
    plt.close(fig)
    #!! fit for flux scale
    meanA = np.nanmean(allSkylineA, axis=0)
    scaleSkylineA = allSkylineA/meanA
    skylineMask = np.isfinite(scaleSkylineA)
    smoothA = np.ones(flux.shape)
    for i in range(flux.shape[0]):
        thismask = skylineMask[i]
        if thismask.sum()<3:
            continue
        thisLimit = [xmap[i].min(), xmap[i].max()]
        thisxx = allSkylineLam[i][thismask]
        thisyy = scaleSkylineA[i][thismask]
        try:
            thisCoeff, thisMask = xy2coeff(thisxx, thisyy, xlim=thisLimit, reject=3, configs={"ncoeff":3})
        except Exception as e:
            print(e)
            ipdb.set_trace()
        thisSmooth = xcoeff2y(xmap[i], thisCoeff, xlim=thisLimit)
        smoothA[i,:] = thisSmooth[:]
        # de.sc(thisxx, thisyy);de.p(xmap[i], thisSmooth)

    fig=de.f()
    ax = fig.add_subplot(111)
    for i in range(flux.shape[0]):
        ax.plot(xmap[i], smoothA[i], c=c[i])
        ax.scatter(allSkylineLam[i], scaleSkylineA[i], c=c[i], s=8)

    de.d(fig, os.path.join(op,"-skyLineFactorSmooth.fig"))
    fig.clear()
    plt.close(fig)

    #!! output coeffs
    xx = np.arange(nFiber)
    fig = de.f()
    for i in range(ncoeff):
        ax = fig.add_subplot((ncoeff+1)*100+10+i+1)
        ax.scatter(xx, allCoeffs[:,i])
    ax = fig.add_subplot((ncoeff+1)*100+10+ncoeff+1)
    ax.scatter(xx, allRMS)
    de.d(fig, os.path.join(op,"-coeffs.fig"))
    fig.clear()
    plt.close(fig)

    return allCoeffs, xmap, allRMS, smoothA
    #endregion: finish

def getNearMin(ab, i, n):
    N = ab.size
    indMin = np.max([0, i-n])
    indMax = np.min([N-1, i+n])
    theab = ab[:,indMin:indMax+1]
    min2d = theab.min(axis=1)
    min2dInd = np.argmin(theab, axis=1)
    theMin = min2d.min()
    theMinInd = np.argmin(min2d)
    return theMin, [theMinInd, min2dInd[theMinInd]+indMin]

#def readSkyData(configs={}):
#    skyDataFile = configs.get("skyDataFile", "Rousellot_OH_list.dat")
#    skyDataFile = os.path.join(codePath, "data", skyDataFile)
#    skyData = np.loadtxt(skyDataFile)
#    sky = skyData[:,0]
#    intensity = skyData[:,1]
#    mask = np.logical_and(sky>7000, sky<10000)
#    sky = sky[mask]
#    intensity = intensity[mask]
#    return [sky, intensity]
def readSkyData(configs={}):
    skytype = configs.get("skytype", "auto")
    skyDataFile = configs.get("skyDataFile", "Rousellot_OH_list.dat")
    codePath = configs["codePath"]
    skyDataFile = os.path.join(codePath, "data", skyDataFile)
    if skytype not in ["auto", "selected"]:
        raise Exception("unknown skytype:{}".format(skytype))

    if skytype=="auto":
        skyData = np.loadtxt(skyDataFile, dtype=[("sky", np.float64), ("intensity", np.float64)])
        return skyData
    elif skytype=="selected":
        skyData = []
        intensityData=[]
        typeData=[]
        with open(skyDataFile,"r") as f:
            for eachline in f:
                eachline = eachline.strip()
                if eachline and eachline[0] not in "#":
                    try:
                        temp = eachline.split()
                        #print(temp)
                        skyData.append(temp[0])
                        intensityData.append(temp[1])
                        typeData.append(temp[2])
                    except:
                        pass
        N = len(skyData)
        dtype = np.dtype([("lambda", np.float64),("intensity", np.float64),("type", np.int32)])
        result = np.zeros(N, dtype=dtype)
        result["lambda"] = np.array(skyData, dtype=np.float64)
        result["intensity"] = np.array(intensityData, dtype=np.float64)
        result["type"] = np.array(typeData, dtype=np.int32)
        return result

def makeSpecWithCenAndSig(npix, loglam, intensity, coeff, configs={}):
    # setups
    nsmooth = configs.get("nsmooth", 2)
    pixarray = 2.0*np.arange(npix)/(npix - 1.0) - 1.0 # change[0,nx-1] to [-1,1]
    nKernel = 4*nsmooth + 1
    gausskern = np.exp(- ((nKernel-1)/2. - np.arange(nKernel))**2 / nsmooth**2)
    gausskern = gausskern / gausskern.sum()
    leg = Legendre(coeff)
    logmodLam = np.log10(leg(pixarray))# lambda at each xInd
    model = specLib.buildLamp(loglam, intensity, logmodLam, 0)
    model = np.convolve(model, gausskern, 'same')
    return model

def makeSpecDirectly(npix, sigma, npad, lam, intensity, coeffs={}):
    """make spec with n pixel as sigma"""
    xx = np.arange(npix)
    npad = int(npad)
    xKernel = np.arange(-npad, npad+1)
    gausskern = np.exp(-(xKernel**2)/(2.*sigma**2))
    gausskern = gausskern/gausskern.sum()
    model = specLib.buildLamp(lam, intensity, xx, 0)
    model = np.convolve(model, gausskern, 'same')
    return model

def filterSky(xSky, sky, intensity, ivar):
    # drop lines in bad pixel block
    mask = np.ones(sky.shape, dtype=np.bool)
    bad = np.where(ivar==0)[0]
    xSkyInt = xSky.astype(np.int)
    for each in bad:
        mask[xSkyInt==each]=False
    xSky = xSky[mask]
    sky = sky[mask]
    intensity = intensity[mask]
    # replease lines too close to a average line
    xSky, sky, intensity = smoothSkyLines(xSky, sky, intensity, 1)
    return xSky, sky, intensity

def smoothSkyLines(node, yNode, intensity, minD, configs={}):
    assert np.all(np.isfinite(node))
    assert np.all(np.diff(node)>0)
    assert np.all(intensity>0)
    mask = np.ones(node.shape, dtype=np.bool)
    diff = np.diff(node)
    starts, ends = findBlock(diff<minD)
    for i in range(starts.size):
        start = node[starts[i]]
        end = node[ends[i]]
        assert ends[i]>=starts[i], "{}>={}".format(ends[i], starts[i])
        length = (end-start)
        if length<=minD:
            xcen =(node[starts[i]:ends[i]+1]*intensity[starts[i]:ends[i]+1]).sum() \
                  /intensity[starts[i]:ends[i]+1].sum()
            node[starts[i]] = xcen
            yxcen =(yNode[starts[i]:ends[i]+1]*intensity[starts[i]:ends[i]+1]).sum() \
                   /intensity[starts[i]:ends[i]+1].sum()
            yNode[starts[i]] = yxcen
            intensity[starts[i]] = intensity[starts[i]:ends[i]+1].sum()
            mask[starts[i]+1:ends[i]+1] = 0
        else:
            sInd = starts[i]
            eInd = sInd+1
            wInd = sInd
            #eInd = ends[i]
            while True:
                assert eInd<ends[-1]+3
                #print(starts[i], ends[i], sInd, eInd, wInd)
                #print(node[starts[i]:ends[i]+1])
                #print(diff[starts[i]:ends[i]+1])
                if sInd>=ends[i]+1:
                    break

                temp = diff[sInd:eInd+1].sum()
                if temp>minD:
                    xcen = (node[sInd:eInd+1]*intensity[sInd:eInd+1]).sum() \
                           /intensity[sInd:eInd+1].sum()
                    node[wInd] = xcen
                    yxcen = (yNode[sInd:eInd+1]*intensity[sInd:eInd+1]).sum() \
                            /intensity[sInd:eInd+1].sum()
                    yNode[wInd] = yxcen
                    intensity[wInd] = intensity[sInd:eInd+1].sum()
                    wInd += 1
                    sInd = eInd+1
                    eInd = sInd+1
                    continue
                else:
                    eInd += 1
            mask[wInd:ends[i]+1] = 0
    node = node[mask]
    intensity = intensity[mask]
    yNode = yNode[mask]

    return node, yNode, intensity

def findBlock(diff):
    N = diff.size
    ddiff = np.diff(diff.astype(np.int))
    diffNonZero = ddiff[(ddiff!=0)]
    starts = np.where(ddiff>0)[0]+1
    ends = np.where(ddiff<0)[0]+1
    if diffNonZero[0]==-1:
        starts = np.insert(starts, 0, 0)
    if diffNonZero[-1]==1:
        ends = np.append(ends, N-1)
    assert starts.size==ends.size
    return starts, ends

def testWaveset(datafile, wavesetfile, outputPath, configs={}):
    #setups
    datafile = os.path.join(outputPath, datafile)
    wavesetfile = os.path.join(outputPath, wavesetfile)
    # not show the longlong fiber info
    fiberInfo = configs.pop("fiberInfo")
    print("testWaveset({},{},{})".format(datafile, wavesetfile, configs))
    debug = configs.get("debug", 1)
    # get all auxiliary data
    waveset, coeffs, yset, badFiber,debugInfo = \
        readWaveSet(outputPath, name=wavesetfile, debugInfo=1)
    goodFiber = badFiber<1
    yset = yset[goodFiber]
    fiberInfo = fiberInfo[goodFiber]
    logWave = np.log10(waveset)
    nFib = yset.shape[0]
    # gen flattened data
    obj=pyfits.open(datafile)
    spec = obj[0].data
    ivar = obj[1].data

    nmed = configs.get("nmend", 200)
    specmed = spec.copy()
    for i in range(nFib):
        specmed[i,:] = ndimage.median_filter(specmed[i,:], nmed)

    #spec = spec-specmed
    #spec = spec - specmed

    indToScale = np.argmin(debugInfo["rms"][:,0].flatten())
    this1dData = spec[indToScale]
    # scale all the fibers to the first
    med = np.nanmedian(this1dData)
    first1dData = spec[indToScale]/ med
    first1dIvar = ivar[indToScale]* med**2
    firstLogWave = logWave[indToScale]
    #allFlux, allIvar, allFactor =\
    #    scaleAllToTheFirst(logWave, thisData, thisIvar, 1./med)
    allFlux, allIvar, allFactor = \
        flat.scaleAll(logWave, spec, ivar, firstLogWave, first1dData, first1dIvar)
    allFlux[allIvar==0]=np.nan

    print("stop")

def fitForContinum(wave, spec, ivar, lines, offsets, skip):
    low = lines - offsets
    high = lines + offsets
    lineMask = findMaskInRange(wave, low, high)
    workwave = wave.copy()[~lineMask]
    workspec = spec.copy()[~lineMask]
    workivar = ivar.copy()[~lineMask]
    fit = scipy.interpolate.LSQUnivariateSpline( \
        workwave, workspec, workwave[1::skip], w=workivar)
    result = fit(wave)
    return result

def findMaskInRange(wave, low, high):
    assert low.size==high.size
    assert np.all(high-low>=0)
    N = low.size
    goodMask = np.zeros(wave.size, dtype=np.bool)
    for i in range(N):
        goodMask[low[i]<=wave & wave<=high[i]] = True
    return goodMask


#region: backup codes
#def waveSkyFit(spec, ivar, skyData, lampCoeff, configs={}):
#    # setups
#    nx = spec.size
#    xx = np.arange(spec.size)
#    fiberInfo = configs.get("fiberInfo", "?")
#    fibInd = configs.get("fibInd", "?")
#    yInd   = configs.get("yInd", "?")
#    debug  = configs.get("waveSkyFitDebug", 0)
#    FWHMDebug = configs.get("FWHMDebug", 0)
#    FWHMType  = configs.get("FWHMType", 'sigma')
#    if FWHMType not in ["FWHM", "sigma"]:
#        raise Exception("unknown FWHM Type: {}".format(FWHMType))
#    outpath   = configs.get("outpath", "./")
#    minDataNum = configs.get("minDataNum",15)
#    oldRms = configs["oldRms"][0]
#    reGuess = configs.get("reGuess", 1)
#    waveRef = configs.get("waveRef", 8500)
#    oldfspec = configs.get("oldfspec", None)
#    sky = skyData["sky"]
#    intensity = skyData["intensity"]
#    # scale intensity and spec to 1 to get fit quickly
#    specMax = spec.max()
#    #spec = spec/specMax
#    #ivar = ivar*specMax**2
#   # sky configs1
#    skytype = configs.get("skytype", "auto")
#    assert np.all(np.diff(sky)>0)
#    assert skytype in ["selected", "auto"]
#   # reguess config
#    if reGuess:
#        ncoeff = lampCoeff.size
#        dcoeff = np.zeros((ncoeff,2), dtype=np.float)
#        nsteps = np.ones(dcoeff.shape[0])
#        dcoeff[0][0] = -15
#        dcoeff[0][1] =  15
#        dcoeff[1][0] = -3
#        dcoeff[1][1] = 3
#        nsteps[0] = 200
#
#        thisConfig = {'nsmooth': 1, 'maxLag':2}
#        guessIntensity = intensity.copy()
#        tmask = guessIntensity<0
#        guessIntensity[tmask] = -guessIntensity[tmask]
#        guess, bestcorr, bestmodel = waveGuessIter(spec, np.log10(sky), guessIntensity,
#                                                   lampCoeff, dcoeff, nsteps, configs=thisConfig)
#    else:
#        guess = lampCoeff
#        # get sky, convert it to xInd and make filter
#    xcenOffset = 2.0
#    waveset = xcoeff2y(np.arange(nx), guess, nx)
#    xSky = ycoeff2xInd(sky, guess, nx,configs={"minDataNum":minDataNum, "ncoeff":ncoeff})
#    allSky = sky.copy()
#    inRange = np.logical_and(0+xcenOffset<xSky, xSky<nx-1-xcenOffset)
#    xSky      = xSky[inRange]# xInd for each Lam
#    sky       = sky[inRange]# each lambda from the data file
#    intensity = intensity[inRange]
#   # filter sky lines
#    if skytype=="auto":
#        assert np.all(intensity>0)
#        minFlux=5e-3
#        skyMax = intensity.max()
#        intensity = intensity/skyMax
#        mask = intensity>minFlux
#        oldxSky      = xSky      = xSky[mask]
#        oldsky       = sky       = sky[mask]
#        oldintensity = intensity = intensity[mask]
#        # filter sky use intensity and according to ivar
#        xSky, sky, intensity = filterSky(xSky, sky, intensity, ivar)
#        forFit = np.ones(sky.shape, dtype=np.bool)
#        forSigmaFit = np.ones(sky.shape, dtype=np.bool)
#        if debug>1: # debug filter sky
#            de.f()
#            pxsky = np.tile(oldxSky,(2,1)).T
#            p01   = np.zeros(pxsky.shape)
#            p01[:,1] = oldintensity[:]
#            de.p(pxsky.T, p01.T, c="red")
#            de.sc(pxsky.T, p01.T, c="red")
#            pxsky = np.tile(xSky,(2,1)).T
#            p01   = np.zeros(pxsky.shape)
#            p01[:,1] = intensity[:]
#            de.p(pxsky.T, p01.T, c="green")
#            de.p(spec/spec.max())
#            de.f()
#            pxsky = np.tile(oldsky,(2,1)).T
#            p01   = np.zeros(pxsky.shape)
#            p01[:,1] = oldintensity[:]
#            de.p(pxsky.T, p01.T, c="red")
#            de.sc(pxsky.T, p01.T, c="red")
#            pxsky = np.tile(sky,(2,1)).T
#            p01   = np.zeros(pxsky.shape)
#            p01[:,1] = intensity[:]
#            de.p(pxsky.T, p01.T, c="green")
#            de.p(waveset, spec/spec.max())
#    elif skytype=="selected":
#        forFit = skyData["type"]>=2
#        forFit = forFit[inRange]
#        forSigmaFit = (skyData["type"]>=3)[inRange][forFit]
#        allForSigmaFit = allSky[skyData["type"]>=3]
#        allForSigMask = skyData["type"]>=3
#        tCum = np.cumsum(allForSigMask)
#        tDiff = np.diff(tCum)
#        tMask = np.insert(tDiff, 0, allForSigMask[0]).astype(np.bool)
#        tCum[~tMask]=0
#        allForSigInd = tCum
#        allForSigInd = allForSigInd[inRange][forFit]
#        assert allForSigInd.size==forSigmaFit.size
#        pass
#    xcen  = xSky
#   # piece cut
#    nP = configs.get("nP", 4)
#    pieceMinNum = configs.get("pieceMinNum", 6)
#    logSky = np.log10(sky)
#    logWaveRange = np.log10(xcoeff2y(xx, guess, nx))
#    logmin = logWaveRange.min()
#    logmax = logWaveRange.max()
#    knots = np.linspace(logmin, logmax, nP+1)
#    kStart = knots[:-1]
#    kEnd   = knots[1:]
#
#    # setup initial conditions for sky fit
#    xcenMin = xcen-xcenOffset
#    xcenMax = xcen+xcenOffset
#    sigma    = np.ones(xcen.shape, dtype=np.float)*2.0
#    sigmaMin = np.ones(xcen.shape, dtype=np.float)*0.7
#    sigmaMax = np.ones(xcen.shape, dtype=np.float)*2.6
#    initA = intensity * np.sqrt(np.pi*2) * sigma * specMax
#    nSky = xcen.size
#
#    if debug>1:# debug sky
#        model = makeSpecWithCenAndSig(spec.size, np.log10(sky), intensity, guess)
#        #pxsky = np.tile(xSky,(2,1)).T
#        #p01   = np.zeros(pxsky.shape)
#        #p01[:,1] = intensity[:]
#        #de.f()
#        #de.p(pxsky.T, p01.T);de.p(spec);de.p(model)
#        pxsky = np.tile(sky,(2,1)).T
#        p01   = np.zeros(pxsky.shape)
#        p01[:,1] = intensity[:]
#        de.f()
#        de.p(pxsky.T, p01.T);de.p(waveset, spec);de.p(waveset, model)
#        #de.piecePlotMulti([
#        #    [pxsky, p01, "plot"],
#        #    [waveset, spec, "plot"],
#        #    [waveset, model/model.max(), "plot"]], outDir="debugSkyTweak", N=20)
#
#    theSig = 1/np.sqrt(ivar)
#    A, mu, sigma, reduceChi2, info = specLib.gaussianFit(spec,theSig,
#                                                         xcen, xcenMin, xcenMax,
#                                                         sigma, sigmaMin, sigmaMax,
#                                                         initA=initA,
#                                                         configs={"cDebug":0, "sigmal":3, "reject":0.85,
#                                                                  "mindchi2":0.001, "ignoreBadFit":1})
#   # filter unused fit
#    bmod = info["bmod"]
#    Asig = info["Asig"]
#    muSig = info["muSig"]
#    fullreject = info["fullreject"]
#    A = A[forFit]
#    mu = mu[forFit]
#    sky = sky[forFit]
#    Asig = Asig[forFit]
#    muSig = muSig[forFit]
#    fullreject = fullreject[forFit]
#    sigma = sigma[forFit]
#   # filter bad fit
#    goodMask = (1-fullreject).astype(np.bool) * (A/Asig>1e-5)
#    skyForSigma = sky[goodMask&forSigmaFit]
#    sigmaForSigma = sigma[goodMask&forSigmaFit]
#    AForSigma = A[goodMask&forSigmaFit]
#    allForSigInd = allForSigInd[goodMask&forSigmaFit]
#    allForSigInd = allForSigInd[allForSigInd>0]
#    allForSigInd -= 1
#
#    A = A[goodMask]
#    mu = mu[goodMask]
#    sky = sky[goodMask]
#    Asig = Asig[goodMask]
#    muSig = muSig[goodMask]
#    sigma = sigma[goodMask]
#
#    # select skylines to use by SNR
#    SNR = A/Asig
#    goodMaskSNR = SNR>3
#    goodMaskMu = muSig<0.1
#    goodMaskFit = goodMaskSNR&goodMaskMu
#    # select good points in each piece
#    for i in range(kStart.size):
#        inRange = np.logical_and(kStart[i]<np.log10(sky),np.log10(sky)<kEnd[i])
#        if inRange.sum()<pieceMinNum:
#            print("only {} fibers in sky range {:10.5f}~{:10.5f}".\
#                format(inRange.sum(), np.power(10,kStart[i]), np.power(10,kEnd[i])))
#            continue
#        pieceAll = np.where(inRange)[0]
#        pieceNum = goodMaskFit[inRange].sum()
#        if pieceNum>pieceMinNum:
#            continue
#        else:
#            thisMuSig = muSig[inRange]
#            dataToUse = np.argsort(thisMuSig)[:pieceMinNum]
#            goodMaskFit[pieceAll[dataToUse]] = True
#
#    #SNR = A/Asig
#    #goodMaskSNR = SNR>3
#    #if goodMaskSNR.sum()<=minDataNum:
#    #    goodMaskSNR = SNR>SNR[np.argsort(SNR)[-minDataNum-1]]
#    #mu = mu[goodMaskSNR]
#    #sky = sky[goodMaskSNR]
#    #SNR = SNR[goodMaskSNR]
#    mu = mu[goodMaskFit]
#    Asig = Asig[goodMaskFit]
#    muSig = muSig[goodMaskFit]
#    sky = sky[goodMaskFit]
#    SNR = SNR[goodMaskFit]
#    sigma = sigma[goodMaskFit]
#    oldMu = mu.copy() # all good mu get from gaussian fit
#    oldSky = sky.copy()
#    # use mu as the center of each peak to fit the coeff
#    #SNRFactor = np.log(SNR)
#    #SNRFactor = SNRFactor-SNRFactor.min() + 1
#    oldSkyMod = xcoeff2y(mu, guess, xlim=nx)
#    coeff, goodMask = xy2coeff(mu, sky, w=(1/sky)**2,
#                               xlim=nx, reject=3,
#                               configs={"minDataNum":minDataNum,"ncoeff":ncoeff})
#    # debug plot
#    #coeff = guess + coeff
#    newSkyMod = xcoeff2y(mu, coeff, xlim=nx)
#    oldDiff =  newSkyMod - oldSky
#    allInd = np.zeros(mu.size)
#    allInd[goodMask] = 1
#    if debug>1:
#        for i in range(newSkyMod.size):
#            print("{:3} {:2} {:10.5f} {:10.5f} {:10.5f} {:10.5f} {:10.5f} {:10.5f}".\
#            format(i, allInd[i], oldSky[i], oldDiff[i], SNR[i], muSig[i], Asig[i], sigma[i]))
#    mu = mu[goodMask]
#    #sigma = sigma[goodMask]
#    sky = sky[goodMask]
#    #sigSig = info["sigSig"][goodMask]
#    skyMod = xcoeff2y(mu, coeff, xlim=nx)
#    allSky = xcoeff2y(xx, coeff, xlim=nx)
#    diffs = skyMod-sky
#    rms = np.sqrt((diffs**2 * (1/sky)**2).sum()/goodMask.size)
#
#    if FWHMType=="FWHM":
#        if FWHMDebug>0:
#            FWHMFile = os.path.join(outpath,"FWHM_{}_{:_^10}_{:_^10.5g}_{:_^10.5g}.fig".\
#                format(fibInd, fiberInfo[1], rms*waveRef, rms))
#        else:
#            FWHMFile = 0
#        FWHMArray = FWXM(allSky, spec-bmod, skyForSigma, 5, configs={"debug":FWHMFile})
#    else:
#        FWHMArray = sigmaForSigma * 2.354820045
#
#    oldFlag=0
#    lfactor = configs.get("lfactor", 1.1)
#    if rms>lfactor*oldRms:
#        oldFlag=1
#        print("\t {} for {:<10}:Sky is no better than lamp ({:10.8f}>{:5.2f}*{:10.8f}) ({:10.8f}>{:5.2f}*{:10.8f} @{:5.2f}), use lamp result". \
#            format(fibInd, fiberInfo[1], rms, lfactor, oldRms,
#                   rms*waveRef, lfactor, oldRms*waveRef, waveRef))
#        coeff = guess
#        rms = oldRms
#    else:
#        print("\t {} for {:<10}: Sky Tweaking {} use {} rsm={:10.7f}<{:5.2f}*{:10.7f} ({:10.7f}<{:5.2f}*{:10.7f} @ {:5.2f}) coeff={}". \
#            format(fibInd, fiberInfo[1], nSky, goodMask.size,
#                   rms, lfactor, oldRms, rms*waveRef, lfactor, oldRms*waveRef, waveRef, coeff))
#        if -1<debug<1:
#            debug=1
#    if debug>0:
#        newCen = ycoeff2xInd(oldSky, guess, nx, configs={"ncoeff":ncoeff})# guess result
#        fig = plt.figure()
#        ax = fig.add_subplot(111)
#        if oldfspec is not None:
#            ax.plot(xx,    oldfspec, c="purple")
#            ax.scatter(xx, oldfspec, c="purple", s=6, edgecolors="none")
#            ax.plot(waveset,    oldfspec, c="purple")
#            ax.scatter(waveset, oldfspec, c="purple", s=6, edgecolors="none")
#        # pixel plot
#        ax.scatter(xx, spec, c="red")
#        ax.plot(xx, spec, c="red")
#        ax.plot(xx, info["gmod"], c="cyan")
#        ax.plot(xx, info["bmod"], c="magenta")
#        ax.plot(xx, info["gmod"]+info["bmod"], c="green")
#        ax.scatter(xx, info["gmod"]+info["bmod"], c="green", s=6, edgecolors="none")
#        # wave plot
#        ax.scatter(waveset, spec, c="red")
#        ax.plot(waveset, spec, c="red")
#        ax.plot(waveset, info["gmod"], c="cyan")
#        ax.plot(waveset, info["bmod"], c="magenta")
#        ax.plot(waveset, info["gmod"]+info["bmod"], c="green")
#        ax.scatter(waveset, info["gmod"]+info["bmod"], c="green", s=6, edgecolors="none")
#        # in pixel plot, solid line is real, dashed line should be close to solid line
#        # in wave plot,  dashed line is real, solid line should be close to dashed line
#        for i in range(oldMu.size):
#           # pixel plot
#            # all mu get from gaussian fit
#            ax.plot([oldMu[i], oldMu[i]], [0, spec[int(oldMu[i])]])
#            # all sky to xx use the best fit
#            ax.plot([newCen[i], newCen[i]], [0, spec.max()], "r--")
#            ax.text(oldMu[i], spec[int(oldMu[i])]*1.3, str(i))
#           #wave plot
#            # all mu to wave use the best fit
#            ax.plot([newSkyMod[i], newSkyMod[i]], [0, spec[int(oldMu[i])]])
#            # where the sky should be
#            ax.plot([oldSky[i], oldSky[i]], [0, spec.max()], "r--")
#            ax.text(oldSky[i], spec[int(oldMu[i])]*1.3, str(i))
#        #newCen = ycoeff2xInd(oldSky, coeff, nx)
#        doc = "\nsolid line is fit result, blue is the init, green is the polyfit result"
#        title = "{} {:8.3f} rms={:8.3g} ({:10.5g}@{:5.2f}".format(fibInd, yInd, rms, rms*waveRef, waveRef)
#        fig.suptitle(title+doc)
#        newCen = ycoeff2xInd(oldSky, coeff, nx,configs={"ncoeff":ncoeff})
#
#        for i in range(newCen.size):
#            if i in goodMask:
#                ax.plot([newCen[i], newCen[i]], [0, spec.max()], "g--")
#                ax.plot([newSkyMod[i], newSkyMod[i]], [0, spec.max()], "g--")
#            else:
#                ax.plot([newCen[i], newCen[i]], [0, spec.max()/2], "b--")
#                ax.plot([newSkyMod[i], newSkyMod[i]], [0, spec.max()/2], "b--")
#        for i in range(mu.size):
#            #ax.plot([mu[i], mu[i]], [0, spec[int(mu[i])]])
#            ax.text(mu[i], spec[int(mu[i])]*1.3, str(i), horizontalalignment='right')
#            ax.text(sky[i], spec[int(mu[i])]*1.3, str(i), horizontalalignment='right')
#        #ax.set_ylim(-0.1, 1.1)
#        if debug>1:
#            plt.show(0)
#        elif debug>0:
#            de.d(fig, os.path.join(outpath,"{}_{}_{:_^10}_{:_^10.5g}_{:_^10.5g}.fig". \
#                                   format(oldFlag, fibInd,fiberInfo[1], rms*waveRef, rms)))
#            ax.clear()
#            fig.clear()
#            plt.close(fig)
#            #plt.close('all')
#
#    thisxmap = xcoeff2y(xx, coeff)
#    return coeff, thisxmap, [rms, np.array([skyForSigma, FWHMArray, AForSigma, allForSigInd, allForSigmaFit])]
#
#
#def waveExtract(infile, outputPath, configs={}):
#    # setups
#    infile = os.path.join(outputPath, infile)
#    outfile= os.path.join(outputPath, 'masterArcFlux.fits')
#    rerun = int(configs.get('rerun', 0))
#    if not rerun and os.path.exists(outfile):
#        print("\tin waveExtract: {} already exists, skip".format(outfile))
#        return
#    print("\twaveExtract({},{})".format(outfile, configs))
#    extractSigma = configs.get("extractSigma", 2)
#    waveDirection = configs.get('waveDirection', 'row')
#    if waveDirection not in ['row', 'column']:
#        raise Exception('invalid waveDirection!')
#    yset, badFiber = extract.readTraceSet(outputPath)
#    badMask = (badFiber.astype(np.bool))
#    yset = yset[~badMask,:]
#    obj = pyfits.open(infile)
#    data = obj[0].data; dataHeader = obj[0].header
#    error = obj[1].data; errorHeader = obj[1].header
#    if waveDirection=='column':
#        data = data.T.copy() # copy is necessary because we call c function latter
#        error = error.T.copy()
#    thisConfig={"proftype":-3,
#                "map2dFileName":os.path.join(outputPath, infile.split(".fits")[0])};
#    thisConfig.update(configs)
#    ymod, flux, ivar= \
#        extract.extract(data, 1/(error**2), yset, extractSigma, configs=thisConfig)
#    #ymod, flux, ivar = extract.extract1(data, 1/(error**2), yset, extractSigma, configs=thisConfig)
#    data1 = pyfits.PrimaryHDU(flux)
#    data2 = pyfits.ImageHDU(ivar)
#    fits  = pyfits.HDUList([data1, data2])
#    fits.writeto(outfile, clobber='True')
#    data1 = pyfits.PrimaryHDU(ymod)
#    data2 = pyfits.ImageHDU(data)
#    data3 = pyfits.ImageHDU((ymod-data))
#    data4 = pyfits.ImageHDU((ymod-data)/error)
#    data5 = pyfits.ImageHDU((ymod-data)/ymod*100.)
#    fits  = pyfits.HDUList([data1, data2, data3, data4, data5])
#    fits.writeto(os.path.join(outputPath, "masterArcExtractModel.fits"), clobber='True')
#
#def tweakBySky(infile, outputPath, configs={}):
#    # setups
#    infile = os.path.join(outputPath, infile)
#    infileName = infile.split("mdFlux.fits")[0]
#    debug = configs.get("debug", 1)
#    doSigmaConvolve = configs.get("doSigmaConvolve", 0)
#    if doSigmaConvolve:
#        print("sky tweak")
#        outfile = os.path.join(outputPath, infileName+"SkyWaveset.fits")
#        debugPath = os.path.join(outputPath, infileName + "Waveset")
#        oldwave = os.path.join(outputPath, infileName+"SkyWaveset0.fits")
#        waveset, coeffs, yset, badFiber,debugInfo = readWaveSet(outputPath, name=oldwave, debugInfo=2)
#
#        goodFiber = badFiber<1
#        yset = yset[goodFiber]
#        nFiber, nx = yset.shape
#
#        sigma2FWHM = 2.354820045
#        FWHM = debugInfo["fiberModel"][2]
#        FWHMCoeff = debugInfo["fiberModel"][1]
#        FWHMLimits = debugInfo["fiberModel"][0]
#        sigma = FWHM/sigma2FWHM
#        # select smooth kernel size for all data
#        xmin = FWHMLimits[0]
#        xmax = FWHMLimits[1]
#        N = 1000
#        xx = np.linspace(xmin, xmax, N)
#        allSigma = np.zeros((nFiber, N))
#        for i in range(nFiber):
#            fit = Legendre(FWHMCoeff[i], domain=FWHMLimits)
#            allSigma[i][:] = fit(xx)
#        allSigma /= sigma2FWHM
#        allSigmaY = np.zeros(N)
#        percentile = scipy.interpolate.UnivariateSpline( \
#            np.linspace(xmin, xmax, 8),
#            np.array([90, 90, 90, 90, 90, 92, 94, 96]))(xx)
#        for i in range(N):
#            allSigmaY[i] = np.percentile(allSigma[:,i], percentile[i])
#        sigmaFit = scipy.interpolate.UnivariateSpline(xx, allSigmaY)
#        allSigmaMod = sigmaFit(xx)
#        kernelSig = np.zeros((nFiber, nx))
#        for i in range(nFiber):
#            value = sigmaFit(waveset[i])**2 - sigma[i]**2
#            value[value<=0]=0
#            kernelSig[i][:] = np.sqrt(value)[:]
#    else:
#        print("sky tweak0")
#        outfile = os.path.join(outputPath, infileName+"SkyWaveset0.fits")
#        debugPath = os.path.join(outputPath, infileName + "Waveset0")
#    rerun = configs.get("rerun", 0)
#    if not rerun and os.path.exists(outfile):
#        print("\tin tweakBySky: {} already exists, skip".format(outfile))
#        return 0
#    # not to show the longlong fiberInfo
#    fiberInfo = configs.pop("fiberInfo")["assignments"]
#    ncoeff = configs.get("ncoeff", 6)
#    waveSkyFitDebug = configs.get("waveSkyFitDebug", 0)
#    FWHMDebug = configs.get("FWHMDebug",0)
#    FWHMType = configs.get("FWHMType", "sigma")
#    print("tweakBySky({}, {})".format(infile, configs))
#    # readin all data needed
#    obj = pyfits.open(infile)
#    spec = obj[0].data
#    ivar = obj[1].data
#    nFib, nx = spec.shape
#    # do the same thing as in sky sub
#    fiberFlat, factors = flat.readFiberFlat(outputPath)
#    # one way to do scale
#    maxmin = np.nanmin(np.nanmedian(spec, axis=1))
#    scaleFactor = 1.0/maxmin*(factors/factors.min()).reshape(factors.size,1)/fiberFlat
#    fspec = spec * scaleFactor
#    fivar = ivar / scaleFactor **2
#    fsig2 = 1/fivar
#    oldfspec = fspec.copy()
#
#    waveset, coeffs, yset, badFiber,debugInfo = readWaveSet(outputPath, debugInfo=1)
#    goodFiber = badFiber<1
#    yset = yset[goodFiber]
#    fiberInfo = fiberInfo[goodFiber]
#   # convolve all spec
#    if doSigmaConvolve:
#        allfspecModel = np.zeros((nFiber, nx))
#        allfsigmaModel = np.zeros((nFiber, nx))
#        fspec[fivar==0]=np.nan
#        smoothSize = configs.get("smoothSize", 5)
#        for i in range(nFiber):
#            thisWave = waveset[i]
#            thisModelCov = specLib.variateConvolve(thisWave, fspec[i], kernelSig[i], smoothSize)
#            allfspecModel[i][:] = thisModelCov[:]
#            thisSigma2Cov = specLib.variateConvolve(thisWave, fsig2[i], kernelSig[i], smoothSize)
#            allfsigmaModel[i][:] = thisSigma2Cov[:]
#        allfsigmaModel[np.isnan(allfsigmaModel)]=np.inf
#        fspec = allfspecModel
#        fspec[~np.isfinite(allfsigmaModel)] = 0
#        fivar = 1/allfsigmaModel
#
#    assert goodFiber.sum()==nFib
#    #tskytype="auto";hisConfig = {"skyDataFile":"Rousellot_OH_list.dat", "skytype":skytype}
#    skytype="selected";thisConfig = {"skyDataFile":"m2fsSkyLines.dat", "skytype":skytype}
#    skyData = readSkyData(configs=thisConfig)
#
#    # init variables
#    allCoeffs = np.zeros((nFib, ncoeff))
#    xmap = np.zeros((nFib, nx))
#    allRMS=np.ones((nFib, 1)) * np.inf
#    fiberModelInfo=np.zeros(nFib, dtype=np.object)
#    for fibInd in range(0,nFib):
#        thisConfig = {"fibInd": fibInd+1, "yInd":yset[fibInd][0],
#                      "waveset":waveset[fibInd], "outpath":debugPath,
#                      "fiberInfo":fiberInfo[fibInd], "oldRms":debugInfo["rms"][fibInd],
#                      "skytype":skytype, "waveSkyFitDebug":waveSkyFitDebug,
#                      "FWHMDebug":FWHMDebug, "FWHMType":FWHMType}
#        if doSigmaConvolve:
#            thisConfig["oldfspec"] = oldfspec[fibInd]
#        thisCoeff, thisMap, thisDebugInfo= waveSkyFit(\
#            fspec[fibInd], fivar[fibInd], skyData, coeffs[fibInd], configs=thisConfig)
#        allCoeffs[fibInd,:] = thisCoeff[:]
#        xmap[fibInd,:] = thisMap[:]
#        allRMS[fibInd,0] = thisDebugInfo[0]
#        fiberModelInfo[fibInd] = thisDebugInfo[1]
#
#   # fit for smooth FWHM
#    xMin = np.min(np.array([each[0].min() for each in fiberModelInfo]))
#    xMax = np.max(np.array([each[0].max() for each in fiberModelInfo]))
#    limits = [xMin, xMax]
#    #smoothFWHM = np.empty((nFib,), dtype=np.object)
#    smoothFWHM = np.zeros((nFib, nx))
#    smoothA = np.zeros((nFib, nx))
#    thisConfig = {"ncoeff":4}
#    fwhmOut = np.empty((5,), dtype=np.object)
#    allSigmaCoeffs = np.zeros((nFib, thisConfig["ncoeff"]))
#    allACoeffs = np.zeros((nFib, thisConfig["ncoeff"]))
#
#    nAllForSigmaFit = fiberModelInfo[0][4].size
#    allForSigmaFitSky = np.zeros((nFib, nAllForSigmaFit))
#    allForSigmaFitA   = np.zeros((nFib, nAllForSigmaFit)) * np.nan
#    for i in range(nFib):
#        allForSigmaFitSky[i][:] = fiberModelInfo[i][4][:]
#        allForSigmaFitA[i][fiberModelInfo[i][3]] = fiberModelInfo[i][2][:]
#
#    allA = allForSigmaFitA/np.nanmedian(allForSigmaFitA, axis=0)
#
#    print("fit for smooth sigma...")
#    for i in range(nFib):
#        thisCoeff, thisMask = xy2coeff(fiberModelInfo[i][0], fiberModelInfo[i][1],
#                                       xlim=limits, reject=3, configs=thisConfig)
#        allSigmaCoeffs[i,:] = thisCoeff
#        thisXX = waveset[i]
#        thisSmooth = xcoeff2y(thisXX, thisCoeff, xlim=limits)
#        smoothFWHM[i][:] = thisSmooth[:]
#        eachx = allForSigmaFitSky[i]
#        eachy = allA[i]
#        eachmask = np.isfinite(eachy)
#        thisCoeff, thisMask = xy2coeff(eachx[eachmask], eachy[eachmask],
#                                       xlim=limits, reject=3, configs=thisConfig)
#        allACoeffs[i,:] = thisCoeff
#        thisXX = waveset[i]
#        thisSmooth = xcoeff2y(thisXX, thisCoeff, xlim=limits)
#        smoothA[i][:] = thisSmooth[:]
#
#    fwhmOut[0] = limits
#    fwhmOut[1] = allSigmaCoeffs
#    fwhmOut[2] = smoothFWHM
#    fwhmOut[3] = allACoeffs
#    fwhmOut[4] = smoothA
#    fwhmOut.dump(outfile+".fiberModel.dat")
#
#    c = cm.get_cmap("rainbow")
#    c = c(np.linspace(0,1,nFib))
#    sigma2FWHM = 2.354820045
#    print("output for skySigma debug fig")
#    fig=de.f()
#    ax = fig.add_subplot(111)
#    for i, eachInfo in enumerate(fiberModelInfo):
#        ax.plot(eachInfo[0], eachInfo[1]/sigma2FWHM, c=c[i])
#        ax.scatter(eachInfo[0], eachInfo[1]/sigma2FWHM, c=c[i], s=8)
#        for j in range(eachInfo[0].size):
#            ax.text(eachInfo[0][j],  eachInfo[1][j]/sigma2FWHM,  "{}".format(i), color=c[i])
#    de.d(fig, os.path.join(debugPath,"-skySigma.fig"))
#    fig.clear()
#    plt.close(fig)
#    print("output for skySigmaSmooth debug fig")
#    fig=de.f()
#    ax = fig.add_subplot(111)
#    for i in range(nFib):
#        ax.plot(waveset[i], smoothFWHM[i]/sigma2FWHM, c=c[i])
#        ax.scatter(fiberModelInfo[i][0], fiberModelInfo[i][1]/sigma2FWHM, c=c[i], s=8)
#        ax.text(waveset[i][0],   fiberModelInfo[i][0]/sigma2FWHM,  "{}".format(i), color=c[i])
#        ax.text(waveset[i][-1],  fiberModelInfo[i][-1]/sigma2FWHM,  "{}".format(i), color=c[i])
#
#    de.d(fig, os.path.join(debugPath,"-skySigmaSmooth.fig"))
#    fig.clear()
#    plt.close(fig)
#
#    print("output for skyA debug fig")
#    fig=de.f()
#    ax = fig.add_subplot(111)
#    for i in range(nFib):
#        ax.plot(allForSigmaFitSky[i], allA[i], c=c[i])
#        ax.scatter(allForSigmaFitSky[i], allA[i], c=c[i], s=8)
#        for j in range(nAllForSigmaFit):
#            if np.isfinite(allA[i][j]):
#                ax.text(allForSigmaFitSky[i][j],  allA[i][j],  "{}".format(i), color=c[i])
#    de.d(fig, os.path.join(debugPath,"-skyA.fig"))
#    fig.clear()
#    plt.close(fig)
#    print("output for skyASmooth debug fig")
#    fig=de.f()
#    ax = fig.add_subplot(111)
#    for i in range(nFib):
#        ax.plot(waveset[i], smoothA[i], c=c[i])
#        ax.scatter(allForSigmaFitSky[i], allA[i], c=c[i], s=8)
#        ax.text(waveset[i][0],   smoothA[i][0],  "{}".format(i), color=c[i])
#        ax.text(waveset[i][-1],  smoothA[i][-1],  "{}".format(i), color=c[i])
#
#    de.d(fig, os.path.join(debugPath,"-skyASmooth.fig"))
#    fig.clear()
#    plt.close(fig)
#
#    data1 = pyfits.PrimaryHDU(xmap, header=None)
#    data2 = pyfits.ImageHDU(allCoeffs, header=None)
#    data3 = pyfits.ImageHDU(allRMS, header=None)
#    fits  = pyfits.HDUList([data1, data2, data3])
#    fits.writeto(outfile, clobber='True')
#    # output debug file
#    xx = np.arange(nFib)
#    fig = de.f()
#    for i in range(ncoeff):
#        ax = fig.add_subplot((ncoeff+1)*100+10+i+1)
#        ax.scatter(xx, allCoeffs[:,i])
#    ax = fig.add_subplot((ncoeff+1)*100+10+ncoeff+1)
#    ax.scatter(xx, allRMS)
#    de.d(fig, os.path.join(debugPath,"-skyTweakCoeffs.fig"))
#    fig.clear()
#    plt.close(fig)
#
#    #smoothFWHM.dump(outfile+".FWHM.dat")
#    return 1
#
#endregion: backup codes
