#region: imports
from __future__ import (absolute_import, division, print_function, unicode_literals)
import numpy as np
import stat
import os
import sys
import matplotlib.pyplot as plt
from matplotlib import cm
import numpy
from numpy.polynomial import Legendre
from . import de
from . import extract
from . import wavecalib
from .image import ccdproc
from .image import specLib
from ..utils import macro
from ..utils import messager
from ..utils import configTools
import scipy.interpolate as interp
from scipy.optimize import curve_fit
import scipy.integrate as integrate
import glob
import multiprocessing
import copy
import scipy
from ..DS10 import ds9
from ..DS10 import goodFigure
from ..DS10.figureClass import DebugFigure
from ..math import astroMath

from . import m2fsConfigs
from . import flat
import shutil
import astropy.io.fits as pyfits
from astropy import wcs
import pdb
from numpy.polynomial import Legendre
from scipy import ndimage
import signal
import time
import pickle
import multiprocessing
import ipdb

#signal.signal(signal.SIGALRM,OriginalHandler)
#endregion: imports
np.set_printoptions(precision=4)
#region: aux functions
regStrH="""# Region file format: DS9 version 4.1
global color={color} dashlist=8 3 width=1 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1
physical
"""
regStrP = "circle {0} {1} {2}i #color={3} move=0\n"
regStrT = "#text({0},{1}) text={{{2}}} color={3}\n"
regStrE = "ellipse({0}, {1}, {2}, {3}, 0) #color={4}\n"
regStrB = "box({0}, {1}, {2}, {3}, 0) #color={4}\n"
def regStrX(x,y,s,color):
    s1 = "line({},{},{},{}) # line=0 0 color={}\n".format(x-s, y-s, x+s, y+s, color)
    s2 = "line({},{},{},{}) # line=0 0 color={}\n".format(x-s, y+s, x+s, y-s, color)
    return s1+s2
def regStrPlus(x,y,s,color):
    s1 = "line({},{},{},{}) # line=0 0 color={}\n".format(x, y-s, x, y+s, color)
    s2 = "line({},{},{},{}) # line=0 0 color={}\n".format(x-s, y, x+s, y, color)
    return s1+s2

def fixKnots(x,k):
    """ make knots to satisfiy blabla condition"""
    k = k[k>x[0]]
    k = k[k<x[-1]]
    mask = np.ones(k.shape, dtype=np.bool)
    assert k.size>3 and x.size>3
    start=0;end=1
    while end<k.size:
        if np.any((x>k[start])&(x<k[end])):
            start = end
            end +=1
        elif end==k.size-1:
            mask[start]=False
            break
        else:
            mask[end]=False
            end +=1
    return k[mask]

"""default sigma is 1.6, default A,A3 is 0.01, default AUpper is 10"""
def readAllSkyData(codepath, configs={}):
    filename = os.path.join(codepath, "data", "allSkyLines.dat")
    dtype = np.dtype([("group", "U10"),
                      ("lambda", np.float64),
                      ("flag", np.int64),
                      ("residualType", np.int64),
                      ("type", np.int64),
                      ("sigma", np.float64),
                      ("sigma3", np.float64),
                      ("mulower", np.float64),
                      ("muupper", np.float64),
                      ("sigmalower", np.float64),
                      ("sigmaupper", np.float64),
                      ("sigma3lower", np.float64),
                      ("sigma3upper", np.float64),
                      ("Alower", np.float64),
                      ("Aupper", np.float64),
                      ("A3lower", np.float64),
                      ("A3upper", np.float64)])
    defaultSigma = configs.get("defaultSigma", 1.6)
    defaultSigma3 = configs.get("defaultSigma3", 1.6)
    defaultA = configs.get("defaultA", 0.01)
    defaultA3 = configs.get("defaultA#", 0.01)
    defaultSigmaLow = 1
    defaultSigmaUp = 2
    badFlux = 10
    maxdiff = configs.get("maxdiff", 2.0)
    result = []
    count = 0
    allDefaults = [defaultSigma, defaultSigma3, '-', '-', defaultSigmaLow, defaultSigmaUp, defaultSigmaLow, defaultSigmaUp, 0, badFlux, 0, badFlux]
    lastWave=-np.inf
    with open(filename, "r") as f:
        for eachLine in f:
            count += 1
            eachClean = eachLine.strip()
            if not eachClean:
                continue
            elif eachClean[0]=="#":
                continue

            each = eachClean.split()
            if len(each)<5:
                continue
            if len(each)>17:
                raise Exception("bad format for line {}: {}".format(count, eachLine))
            else:
                N = len(each)
                group = each[0]
                flag = each[1]
                thisResidualType = each[2]
                thisType = each[3]
                mu = float(each[4])
                if mu>lastWave:
                    lastWave = mu
                else:
                    raise Exception("order error in sky data @ {}".format(mu))
                allDefaults[2] = mu - maxdiff
                allDefaults[3] = mu + maxdiff
                todo = [group, mu, flag, thisResidualType, thisType, '-', '-', '-',
                        '-', '-', '-', '-', '-', '-', '-', '-', '-']
                for i in range(5, N+1):
                    try:
                        todo[i] = float(each[i])
                    except:
                        todo[i] = allDefaults[i-5]
                for i in range(N+1, 17):
                    todo[i] = allDefaults[i-5]
            result.append(tuple(todo))
    result = np.array(result, dtype=dtype)
    return result

def readSkyRegion(codepath, configs={}):
    filename = os.path.join(codepath, "data", "skyRegions.dat")
    dtype = np.dtype([("start", np.float64),
                      ("end", np.float64),
                      ("type", np.int64),
                      ("density", np.float64)])
    result = []
    with open(filename, "r") as f:
        for eachLine in f:
            eachClean = eachLine.strip()
            if not eachClean:
                continue
            if eachClean[0]=="#":
                continue
            each = eachClean.split()
            if len(each)==3:
                each = (float(each[0]), float(each[1]), int(each[2]), 0)
            elif len(each)==4:
                each = (float(each[0]), float(each[1]), int(each[2]), float(each[3]))
            else:
                raise Exception("bad format for skyRegions: {}".format(eachLine))
            result.append(each)
    result = np.array(result, dtype=dtype)
    return result

def readSkyBlock(codepath, filename="skyRegionBlocks.dat"):
    filename = os.path.join(codepath, "data", filename)
    dtype = np.dtype([("start", np.float64),
                      ("end", np.float64),
                      ("fitStart", np.float64),
                      ("fitEnd", np.float64),
                      ("nPoly", np.int64),
                      ("factor", np.float64),#!! to control fitting
                      ("polyFactor", np.float64)#!! to control poly fitting
                      ])
    result = []
    with open(filename, "r") as f:
        for eachLine in f:
            eachClean = eachLine.strip()
            if not eachClean:
                continue
            if eachClean[0]=="#":
                continue
            each = eachClean.split()
            assert len(each)==7
            if int(each[4])>7:
                raise Exception("order must be less than 7")
            each = (float(each[0]), float(each[1]),
                 float(each[0])-float(each[2]), float(each[1])+float(each[3]),
                 int(each[4]), float(each[5]), float(each[6]))
            result.append(each)
    result = np.array(result, dtype=dtype)
    return result

def makeKnotes(skydata, waveset):
    min = waveset.min()
    max = waveset.max()
    meanDiff = 4
    minDiff  = 0.5
    length = int((max-min)/meanDiff)
    simpleKnots = np.linspace(min, max, length)
    #!! naive make knotes
    kernel = np.arange(-4,4+1./2.,1./2.)
    addedKnots = []
    for i in range(skydata.size):
        eachcen = skydata[i]["lambda"]
        eachsigma = skydata[i]["sigma"]
        toAppend = eachcen + eachsigma*kernel
        toAppend = toAppend[toAppend>min]
        toAppend = toAppend[toAppend<max]
        addedKnots.append(toAppend)
    addedKnots = np.concatenate(addedKnots)
    allKnots = np.concatenate([simpleKnots, addedKnots])
    allKnots.sort()
    #!! smooth knotes
    knots = flat.smoothNode(allKnots, minDiff)
    #!! reject bad region
    regionIndex = 0
    regionIndexEnd = 1
    dataIndex = 0
    data = waveset
    regionLen = knots.size
    dataLen = data.size
    badRegions = []
    #!! region points should be within data points
    #!! still have bug...
    while True:
        #!! start from beginning
        if data[dataIndex]>knots[regionIndex] and data[dataIndex]<knots[regionIndexEnd]:
            dataIndex += 1
            regionIndex = regionIndexEnd
            if regionIndex==regionLen-1:
                break
            while data[dataIndex]<knots[regionIndex]:
                dataIndex += 1
            if dataIndex==dataLen-1:
                break
            regionIndexEnd += 1
            while regionIndexEnd<regionLen-1 and knots[regionIndexEnd] <= data[dataIndex]:
                badRegions.append(regionIndexEnd)
                regionIndexEnd += 1
        elif data[dataIndex]>=knots[regionIndexEnd]:
            badRegions.append(regionIndexEnd)
            regionIndex += 1
            regionIndexEnd += 1
        elif data[dataIndex]<=knots[regionIndex]:
            dataIndex += 1
        else:
            raise Exception("should not be here..")
    goodMask = np.ones(knots.shape).astype(np.bool)
    if len(badRegions):
        goodMask[np.array(badRegions)] = False
    result = knots[goodMask]
    return result

def maskSkyLines(skydata, waveset):
    min = waveset.min()
    max = waveset.max()
    kernel = np.array([-3, 3])
    regions = []
    allMasks = np.ones(waveset.shape).astype(np.bool)
    for eachcen, eachsigma, eachint in skydata:
        toAppend = eachcen + eachsigma*kernel
        toAppend = toAppend[toAppend>min]
        toAppend = toAppend[toAppend<max]
        regions.append(toAppend)
    for eachmin, eachmax in regions:
        allMasks[(waveset>eachmin)&(waveset<eachmax)] = 0

    knots = np.linspace(min, max, 400)
    knotsMask = np.ones(knots.shape).astype(np.bool)
    for eachmin, eachmax in regions:
        knotsMask[(knots>eachmin)&(knots<eachmax)] = 0
    knots = knots[knotsMask]
    return allMasks, knots

def processRegion(skyRegions, waveset, flux):
    """return wave mask that are choosen to fit the sky continuum"""
    goodmask = np.zeros(flux.shape).astype(np.bool)
    for eachstart, eachend, eachtype, eachdensity in skyRegions:
        if eachtype==1:
            goodmask[(waveset>eachstart)&(waveset<eachend)] = 1
        elif eachtype==2:
            thismask = (waveset>eachstart)&(waveset<eachend)
            thisindex = np.where(thismask)[0]
            thiswaveset = waveset[thismask]
            thisflux  = flux[thismask]
            down = np.zeros(thisindex.shape).astype(np.bool)
            up = np.zeros(thisindex.shape).astype(np.bool)
            down[1:] = (np.diff(thisflux)<0)[:]
            up[0:-1] = (np.diff(thisflux)>0)[:]
            pointIndexs = thisindex[down&up]
            if pointIndexs.size:
                tempIndex = np.argsort(flux[pointIndexs])[0]
                goodmask[pointIndexs[tempIndex]] = 1
        elif eachtype==3:#!! define count in 1 \AA
            inRange = (waveset>eachstart)&(waveset<eachend)
            if inRange.sum():
                aveCount = inRange.sum()/(eachend-eachstart)
                factor = int(aveCount/eachdensity) #!! 3 count in 1 \AA is enough
                goodIndex = np.where(inRange)[0][::factor]
                goodmask[goodIndex] = 1
        else:
            raise Exception("should not be here...")
    return goodmask

def packParameters(nPoly, opts):
    N = int((len(opts)-nPoly)/5)
    mu = np.array(opts[:N], dtype=np.float64)
    A = np.array(opts[N:2*N], dtype=np.float64)
    sigma = np.array(opts[2*N:3*N], dtype=np.float64)
    A3 = np.array(opts[3*N:4*N], dtype=np.float64)
    sigma3 = np.array(opts[4*N:5*N], dtype=np.float64)
    poly = opts[5*N:]
    return list(zip(mu, A, sigma, A3, sigma3)), poly

def gaussian(x, mu, sigma):
    result = 1/(np.sqrt(2*np.pi)*sigma)* np.exp(-((x-mu)/sigma)**2/2)
    return result/result.sum()

def checkProfile(wave, std, skywave, sigma, sigma3, A, A3, configs={}):
    fullreject = configs.get("fullreject", 0.5)
    partialreject = configs.get("partialreject", 0.75)
    nSky = skywave.size
    goodmask = np.isfinite(std)
    result = np.zeros(nSky)
    totalA = A+A3
    factorA  =  A/totalA
    factorA3 = A3/totalA
    sig = sigma*factorA + sigma3*factorA3
    for i, (eachskywave, eachsigma) in enumerate(zip(skywave, sig)):
        inrange = (wave>eachskywave-5*eachsigma)&(wave<eachskywave+5*eachsigma)
        eachwave = wave[inrange]
        eachgoodmask = goodmask[inrange]
        profile = gaussian(eachwave, eachskywave, eachsigma)
        profile = profile/profile.sum()
        eachreject = (profile*eachgoodmask).sum()
        result[i] = eachreject
    return result<partialreject, result<fullreject

def getInBlocks(waveset, blocks):
    inrange = np.zeros(waveset.shape, dtype=np.bool)
    for i in range(blocks.shape[0]):
        eachstart = blocks[i][0]
        eachend = blocks[i][1]
        eachstartOffset = blocks[i][2]
        eachendOffset = blocks[i][3]
        thisinrange = (waveset>eachstart)&(waveset<eachend)
        inrange[thisinrange] = True
    return inrange

#def getContinueBlocks(array):
    #ind = np.where(array)[0]
    #if ind.size<4:
        #return [],[]
    #u = np.concatenate(([0], ind[1:]-ind[:-1]))
    #l = np.concatenate((ind[1:]-ind[:-1], [0]))
    #a = ind[np.logical_or(u==1, l==1)]
    #uu = np.concatenate(([0], a[1:]-a[:-1]))
    #ll = np.concatenate((a[1:]-a[:-1], [0]))
    #start = np.where(uu!=1)[0]
    #end = np.where(ll!=1)[0]
    #assert start.size==end.size
    #return start, end

def getContinueBlocks(mask):
    "get index of continue 1 mask, mask should be bool or only have 0 and 1"
    if not np.any(mask):
        return [],[]
    start = end = mask.tolist().index(1)
    startIndex = []
    endIndex = []
    N = mask.size
    while True:
        end += 1
        if not mask[start]:
            start += 1
        if end>=N:
            if end-start>1:
                startIndex.append(start)
                endIndex.append(end-1)
            break
        else:
            if not mask[end]:
                if end-start>1:
                    startIndex.append(start)
                    endIndex.append(end-1)
                start = end
    return startIndex, endIndex

def getContinueUp(data):
    if data.size<4:
        return 0
    up = (np.diff(data)>0).astype(np.int)
    starts, ends = getContinueBlocks(up)
    for i,each in enumerate(ends): ends[i]=ends[i]+1
    length = [eachend-eachstart for eachstart, eachend in zip(starts, ends)]
    #print(up, starts, ends, length)
    if not len(length):
        return 0
    maxLengthIndex = np.argmax(length)
    thisStart = starts[maxLengthIndex]
    thisEnd = ends[maxLengthIndex]
    thisLength = length[maxLengthIndex]
    if thisLength>1:
        return data[thisEnd]-data[thisStart]
    else:
        return 0
def getContinueDown(data):
    if data.size<4:
        return 0
    down = (np.diff(data)<0).astype(np.int)
    starts, ends = getContinueBlocks(down)
    for i,each in enumerate(ends): ends[i]=ends[i]+1
    length = [eachend-eachstart for eachstart, eachend in zip(starts, ends)]
    #print(down, starts, ends, length)
    if not len(length):
        return 0
    maxLengthIndex = np.argmax(length)
    thisStart = starts[maxLengthIndex]
    thisEnd = ends[maxLengthIndex]
    thisLength = length[maxLengthIndex]
    if thisLength>1:
        return data[thisStart]-data[thisEnd]
    else:
        return 0

#!! Caution: assuming MOST of the fibers are sky like!!!!
#!!     skylike fibers are fibers not (largely) rejected by a rejectSplineFitting
#!! also calculate localSmoothSky
def getSkyLikeFibers(waveset, flux, std, allSkyLines, skyRegions,
                     skyMask, objMask, fiberIndex, configs={}):
    debugDir = configs["debugDir"]
    skyLikeSigmaReject = configs["skyLikeSigmaReject"]
    logger = configs["logger"]
    fitFile = configs["fitFile"]
    refit = configs["refit"]
    name = configs["name"]
    names = configs["names"]
    prefix = configs.get("prefix", "")
    noOutput = configs.get("noOutput",False)
    goodFiberPixelRatio = configs.get("goodFiberPixelRatio", 0.60)
    m,n = waveset.shape
    goodFiberPixelNumber = n * goodFiberPixelRatio

    thisFitPath = os.path.join(debugDir, "data","_"+fitFile+"basic"+prefix)
    if refit or not os.path.exists(thisFitPath):
        #region: !! use all fiber and get skyLike fiber mask
        index = np.tile(np.arange(m).reshape((m,1)),(1,n))
        wavesetOrigin = waveset.copy()
        fluxOrigin = flux.copy()
        stdOrigin = std.copy()

        waveset_flatten = waveset.flatten()
        flux_flatten = flux.flatten()
        std_flatten = std.flatten()
        index_flatten = index.flatten()
        argIndex = np.argsort(waveset_flatten)
        waveset_flatten = waveset_flatten[argIndex]
        flux_flatten = flux_flatten[argIndex]
        std_flatten = std_flatten[argIndex]
        index_flatten = index_flatten[argIndex]
        max_wave = flux_flatten.max()
        min_wave = flux_flatten.min()

        knots = makeKnotes(allSkyLines, waveset_flatten)
        #!! make the knots more likely to 
        #knots = knots[(knots<max_wave-5)&(knots>min_wave+5)]
        goodMask,yy = rejectSpline(waveset_flatten, flux_flatten, knots[1:-1], std_flatten, sigmaRej=skyLikeSigmaReject)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        diffs = np.abs((yy-flux_flatten)/std_flatten)
        newSkyLikeSigmaReject = np.nanpercentile(diffs, 70)
        if newSkyLikeSigmaReject<skyLikeSigmaReject:
            newSkyLikeSigmaReject = skyLikeSigmaReject
        ax.hist(diffs, bins=np.linspace(0,50,1000))
        ax.set_title("new reject: {}".format(newSkyLikeSigmaReject))
        de.d(fig, os.path.join(debugDir, "skylikeHist.fig"))
        goodMask = diffs<newSkyLikeSigmaReject
        waveset_flatten = waveset_flatten[goodMask]
        flux_flatten = flux_flatten[goodMask]
        std_flatten = std_flatten[goodMask]
        index_flatten = index_flatten[goodMask]
        fiberGoodPixelCounts = np.zeros(m)
        for i in range(m):
            fiberGoodPixelCounts[i] = (index_flatten==i).sum()
        skyLikeFiberMask = fiberGoodPixelCounts>goodFiberPixelNumber
        skyLikeFiberMask[skyMask] = True #!! force all sky to be choosen
        fiberGoodPixelFactor = fiberGoodPixelCounts/n
        #endregion: use all fiber and get skyLike fiber mask
        #region: use good fibers and fit continuum
        waveset = waveset[skyLikeFiberMask]
        flux = flux[skyLikeFiberMask]
        std = std[skyLikeFiberMask]
        waveset_flatten = waveset.flatten()
        flux_flatten = flux.flatten()
        std_flatten = std.flatten()
        argIndex = np.argsort(waveset_flatten)
        waveset_flatten = waveset_flatten[argIndex]
        flux_flatten = flux_flatten[argIndex]
        std_flatten = std_flatten[argIndex]
        print("\tfitting sky...", end="")
        knotsOrigin = makeKnotes(allSkyLines, waveset_flatten)
        knots = fixKnots(waveset_flatten[np.isfinite(std_flatten)], knotsOrigin)
        #!! key variable: smooth sky fit from skylike fibers
        fluxSmoothObj = scipy.interpolate.LSQUnivariateSpline(waveset_flatten,
                                       flux_flatten,
                                       knots[1:-1],
                                       w=1/std_flatten)
        stdSmoothObj = scipy.interpolate.LSQUnivariateSpline(waveset_flatten,
                                       std_flatten,
                                       knots[1:-1],
                                       w=1/std_flatten)
        flux_smooth_flatten = fluxSmoothObj(waveset_flatten)
        std_smooth_flatten = stdSmoothObj(waveset_flatten)
        #!! important: getLocalSmoothObj
        #region: local sky
        if prefix=="_working":
            N = objMask.shape[0]
            skyLikeMaskIndex = np.where(skyLikeFiberMask)[0]
            objMaskIndex = np.where(objMask)[0]
            localSmooth5pmObjArray = np.empty((N,), dtype=np.object)
            localSmooth5pmStdObjArray = np.empty((N,), dtype=np.object)
            localSmooth5pmIndexArray = np.empty((N,), dtype=np.object)

            localSmooth5MaxpmObjArray = np.empty((N,), dtype=np.object)
            localSmooth5MaxpmStdObjArray = np.empty((N,), dtype=np.object)
            localSmooth5MaxpmIndexArray = np.empty((N,), dtype=np.object)

            localSmooth10tObjArray = np.empty((N,), dtype=np.object)
            localSmooth10tStdObjArray = np.empty((N,), dtype=np.object)
            localSmooth10tIndexArray = np.empty((N,), dtype=np.object)

            localSmooth10pmObjArray = np.empty((N,), dtype=np.object)
            localSmooth10pmStdObjArray = np.empty((N,), dtype=np.object)
            localSmooth10pmIndexArray = np.empty((N,), dtype=np.object)

            localSmooth10MaxpmObjArray = np.empty((N,), dtype=np.object)
            localSmooth10MaxpmStdObjArray = np.empty((N,), dtype=np.object)
            localSmooth10MaxpmIndexArray = np.empty((N,), dtype=np.object)

            localSmooth20tObjArray = np.empty((N,), dtype=np.object)
            localSmooth20tStdObjArray = np.empty((N,), dtype=np.object)
            localSmooth20tIndexArray = np.empty((N,), dtype=np.object)

            localSmooth15pmObjArray = np.empty((N,), dtype=np.object)
            localSmooth15pmStdObjArray = np.empty((N,), dtype=np.object)
            localSmooth15pmIndexArray = np.empty((N,), dtype=np.object)

            localSmooth15MaxpmObjArray = np.empty((N,), dtype=np.object)
            localSmooth15MaxpmStdObjArray = np.empty((N,), dtype=np.object)
            localSmooth15MaxpmIndexArray = np.empty((N,), dtype=np.object)

            localSmooth30tObjArray = np.empty((N,), dtype=np.object)
            localSmooth30tStdObjArray = np.empty((N,), dtype=np.object)
            localSmooth30tIndexArray = np.empty((N,), dtype=np.object)
            #!! do skysub for both obj and sky fibers
            skyLikeAbsIndex = fiberIndex[skyLikeMaskIndex]
            nSkyLike = skyLikeFiberMask.sum()
            for i in range(N):
                todoStr="\t\t getting local sky index {:3d}/{:3d} array for {}\r".\
                        format(i, N, name)
                sys.stdout.write(todoStr)
                sys.stdout.flush()
                thisAbsIndex = fiberIndex[i]
                thisWaveMin = wavesetOrigin[i].min()
                thisWaveMax = wavesetOrigin[i].max()
                #print(thisAbsIndex)
                #!! pn5
                diff = 5
                if skyLikeFiberMask[i]:#!! min and max is for skyLike
                    selfLocalIndex = skyLikeAbsIndex.tolist().index(thisAbsIndex)# index in skyLike fibers
                    #print(selfLocalIndex)
                    min = [0, selfLocalIndex-diff][selfLocalIndex-diff>=0]
                    max = [nSkyLike-1, selfLocalIndex+diff][selfLocalIndex+diff<nSkyLike]
                    thisLocalIndex = np.concatenate((np.arange(min, selfLocalIndex),# skip self
                                               np.arange(selfLocalIndex+1, max+1)))
                    if diff*2>nSkyLike:
                        logger.warn("too less skyLike fibers ({}) in {}".format(nSkyLike, name))
                    elif selfLocalIndex-diff<0:
                        max += np.abs(selfLocalIndex-diff)
                        assert max<=nSkyLike
                    elif selfLocalIndex+diff>nSkyLike:
                        min -= np.abs(selfLocalIndex+diff-nSkyLike)
                        assert min>=0
                    thisLocalTotalIndex = np.concatenate((np.arange(min, selfLocalIndex),# skip self
                                               np.arange(selfLocalIndex+1, max+1)))
                else:
                    #print(selfLocalIndex)
                    selfLocalIndex = skyLikeAbsIndex.searchsorted(thisAbsIndex)# index in skyLike fibers
                    min = [0, selfLocalIndex-diff][selfLocalIndex-diff>=0]
                    max = [nSkyLike-1, selfLocalIndex+diff][selfLocalIndex+diff<nSkyLike]
                    thisLocalIndex = np.concatenate((np.arange(min, selfLocalIndex),# do not have self
                                               np.arange(selfLocalIndex, max)))
                    if diff*2>nSkyLike:
                        logger.warn("too less skyLike fibers ({}) in {}".format(nSkyLike, name))
                    elif selfLocalIndex-diff<0:
                        max += np.abs(selfLocalIndex-diff)
                        assert max<=nSkyLike
                    elif selfLocalIndex+diff>nSkyLike:
                        min -= np.abs(selfLocalIndex+diff-nSkyLike)
                        assert min>=0
                    thisLocalTotalIndex = np.concatenate((np.arange(min, selfLocalIndex),# skip self
                                               np.arange(selfLocalIndex, max)))

                #print(thisLocalIndex)
                #print(thisLocalTotalIndex)

                thisLocalAbsIndex = fiberIndex[skyLikeMaskIndex[thisLocalIndex]]
                thisLocalMaxAbsIndexMask = np.logical_and(
                        thisLocalAbsIndex-thisAbsIndex<=diff,
                        thisLocalAbsIndex-thisAbsIndex>=-diff)
                thisLocalMaxIndex = thisLocalIndex[thisLocalMaxAbsIndexMask]

                localSmooth5pmIndexArray[i] = skyLikeMaskIndex[thisLocalIndex]
                localSmooth5MaxpmIndexArray[i] = skyLikeMaskIndex[thisLocalMaxIndex]
                localSmooth10tIndexArray[i] = skyLikeMaskIndex[thisLocalTotalIndex]
                #!! pn10
                diff = 10
                if skyLikeFiberMask[i]:#!! min and max is for skyLike
                    selfLocalIndex = skyLikeAbsIndex.tolist().index(thisAbsIndex)# index in skyLike fibers
                    #print(selfLocalIndex)
                    min = [0, selfLocalIndex-diff][selfLocalIndex-diff>=0]
                    max = [nSkyLike-1, selfLocalIndex+diff][selfLocalIndex+diff<nSkyLike]
                    thisLocalIndex = np.concatenate((np.arange(min, selfLocalIndex),# skip self
                                               np.arange(selfLocalIndex+1, max+1)))
                    if diff*2>nSkyLike:
                        logger.warn("too less skyLike fibers ({}) in {}".format(nSkyLike, name))
                    elif selfLocalIndex-diff<0:
                        max += np.abs(selfLocalIndex-diff)
                        assert max<=nSkyLike
                    elif selfLocalIndex+diff>nSkyLike:
                        min -= np.abs(selfLocalIndex+diff-nSkyLike)
                        assert min>=0
                    thisLocalTotalIndex = np.concatenate((np.arange(min, selfLocalIndex),# skip self
                                               np.arange(selfLocalIndex+1, max+1)))
                else:
                    #print(selfLocalIndex)
                    selfLocalIndex = skyLikeAbsIndex.searchsorted(thisAbsIndex)# index in skyLike fibers
                    min = [0, selfLocalIndex-diff][selfLocalIndex-diff>=0]
                    max = [nSkyLike-1, selfLocalIndex+diff][selfLocalIndex+diff<nSkyLike]
                    thisLocalIndex = np.concatenate((np.arange(min, selfLocalIndex),# do not have self
                                               np.arange(selfLocalIndex, max)))
                    if diff*2>nSkyLike:
                        logger.warn("too less skyLike fibers ({}) in {}".format(nSkyLike, name))
                    elif selfLocalIndex-diff<0:
                        max += np.abs(selfLocalIndex-diff)
                        assert max<=nSkyLike
                    elif selfLocalIndex+diff>nSkyLike:
                        min -= np.abs(selfLocalIndex+diff-nSkyLike)
                        assert min>=0
                    thisLocalTotalIndex = np.concatenate((np.arange(min, selfLocalIndex),# skip self
                                               np.arange(selfLocalIndex, max)))

                #print(thisLocalIndex)
                #print(thisLocalTotalIndex)

                thisLocalAbsIndex = fiberIndex[skyLikeMaskIndex[thisLocalIndex]]
                thisLocalMaxAbsIndexMask = np.logical_and(
                        thisLocalAbsIndex-thisAbsIndex<=diff,
                        thisLocalAbsIndex-thisAbsIndex>=-diff)
                thisLocalMaxIndex = thisLocalIndex[thisLocalMaxAbsIndexMask]

                localSmooth10pmIndexArray[i] = skyLikeMaskIndex[thisLocalIndex]
                localSmooth10MaxpmIndexArray[i] = skyLikeMaskIndex[thisLocalMaxIndex]
                localSmooth20tIndexArray[i] = skyLikeMaskIndex[thisLocalTotalIndex]
                #!! pn15
                diff = 15
                if skyLikeFiberMask[i]:#!! min and max is for skyLike
                    selfLocalIndex = skyLikeAbsIndex.tolist().index(thisAbsIndex)# index in skyLike fibers
                    #print(selfLocalIndex)
                    min = [0, selfLocalIndex-diff][selfLocalIndex-diff>=0]
                    max = [nSkyLike-1, selfLocalIndex+diff][selfLocalIndex+diff<nSkyLike]
                    thisLocalIndex = np.concatenate((np.arange(min, selfLocalIndex),# skip self
                                               np.arange(selfLocalIndex+1, max+1)))
                    if diff*2>nSkyLike:
                        logger.warn("too less skyLike fibers ({}) in {}".format(nSkyLike, name))
                    elif selfLocalIndex-diff<0:
                        max += np.abs(selfLocalIndex-diff)
                        assert max<=nSkyLike
                    elif selfLocalIndex+diff>nSkyLike:
                        min -= np.abs(selfLocalIndex+diff-nSkyLike)
                        assert min>=0
                    thisLocalTotalIndex = np.concatenate((np.arange(min, selfLocalIndex),# skip self
                                               np.arange(selfLocalIndex+1, max+1)))
                else:
                    selfLocalIndex = skyLikeAbsIndex.searchsorted(thisAbsIndex)# index in skyLike fibers
                    #print(selfLocalIndex)
                    min = [0, selfLocalIndex-diff][selfLocalIndex-diff>=0]
                    max = [nSkyLike-1, selfLocalIndex+diff][selfLocalIndex+diff<nSkyLike]
                    thisLocalIndex = np.concatenate((np.arange(min, selfLocalIndex),# do not have self
                                               np.arange(selfLocalIndex, max)))
                    if diff*2>nSkyLike:
                        logger.warn("too less skyLike fibers ({}) in {}".format(nSkyLike, name))
                    elif selfLocalIndex-diff<0:
                        max += np.abs(selfLocalIndex-diff)
                        assert max<=nSkyLike
                    elif selfLocalIndex+diff>nSkyLike:
                        min -= np.abs(selfLocalIndex+diff-nSkyLike)
                        assert min>=0
                    thisLocalTotalIndex = np.concatenate((np.arange(min, selfLocalIndex),# skip self
                                               np.arange(selfLocalIndex, max)))

                #print(thisLocalIndex)
                #print(thisLocalTotalIndex)

                thisLocalAbsIndex = fiberIndex[skyLikeMaskIndex[thisLocalIndex]]
                thisLocalMaxAbsIndexMask = np.logical_and(
                        thisLocalAbsIndex-thisAbsIndex<=diff,
                        thisLocalAbsIndex-thisAbsIndex>=-diff)
                thisLocalMaxIndex = thisLocalIndex[thisLocalMaxAbsIndexMask]

                localSmooth15pmIndexArray[i] = skyLikeMaskIndex[thisLocalIndex]
                localSmooth15MaxpmIndexArray[i] = skyLikeMaskIndex[thisLocalMaxIndex]
                localSmooth30tIndexArray[i] = skyLikeMaskIndex[thisLocalTotalIndex]
            #!! XX indexArray is of length N, not nSkyLike
            for i in range(N):
                todoStr="\t\t fitting local sky {:3d}/{:3d} array for {}\r".\
                        format(i, nSkyLike, name)
                sys.stdout.write(todoStr)
                sys.stdout.flush()
                inArrays = [localSmooth5pmIndexArray[i], localSmooth5MaxpmIndexArray[i],
                        localSmooth10tIndexArray[i],
                    localSmooth10pmIndexArray[i], localSmooth10MaxpmIndexArray[i],
                    localSmooth20tIndexArray[i], localSmooth15pmIndexArray[i],
                    localSmooth15MaxpmIndexArray[i], localSmooth30tIndexArray[i]]
                outObjs = [localSmooth5pmObjArray, localSmooth5MaxpmObjArray,
                        localSmooth10tObjArray,
                        localSmooth10pmObjArray, localSmooth10MaxpmObjArray,
                        localSmooth20tObjArray, localSmooth15pmObjArray,
                        localSmooth15MaxpmObjArray, localSmooth30tObjArray]
                outStdObjs = [localSmooth5pmStdObjArray, localSmooth5MaxpmStdObjArray,
                        localSmooth10tStdObjArray,
                    localSmooth10pmStdObjArray, localSmooth10MaxpmStdObjArray,
                    localSmooth20tStdObjArray, localSmooth15pmStdObjArray,
                    localSmooth15MaxpmStdObjArray, localSmooth30tStdObjArray]
                namesArray = ["5pm", "5Maxpm", "10t", "10pm","10Maxpm", "20t",
                              "15pm", "15Maxpm", "30t"]
                thisWaveMin = wavesetOrigin[i].min()
                thisWaveMax = wavesetOrigin[i].max()
                for eachInput, eachOutputObj, eachOutputStdObj, eachName in\
                        zip(inArrays, outObjs, outStdObjs, namesArray):
                    thisMask = np.zeros(skyLikeFiberMask.shape, dtype=np.bool)
                    thisMask[eachInput] = True
                    thiswaveset = wavesetOrigin[thisMask]
                    thisflux = fluxOrigin[thisMask]
                    thisstd = stdOrigin[thisMask]
                    thisNumber = thisMask.sum()
                    thisFileName = "_{}_{}_localSmoothSky.fig".format(names[i],eachName)
                    fig = plt.figure()
                    ax = fig.add_subplot(111)
                    c = cm.get_cmap("rainbow")
                    c = c(np.linspace(0,1,thisNumber+1))
                    for ii in range(thisNumber):
                        thisLabel = "{} {}".format(eachInput[ii],
                                            fiberIndex[eachInput[ii]])
                        ax.plot(thiswaveset[ii], thisflux[ii],
                                c=c[ii], lw=1, label=thisLabel)
                        ax.scatter(thiswaveset[ii], thisflux[ii], c=c[ii], edgecolor="None", s=10)
                    thiswaveset_flatten = thiswaveset.flatten()
                    thisflux_flatten = thisflux.flatten()
                    thisstd_flatten = thisstd.flatten()
                    if thiswaveset_flatten.min()>thisWaveMin:
                        extendMask = wavesetOrigin[i]<thiswaveset_flatten.min()
                        thiswaveset_flatten = np.concatenate((wavesetOrigin[i][extendMask],
                                                           thiswaveset_flatten))
                        thisflux_flatten = np.concatenate((fluxOrigin[i][extendMask], thisflux_flatten))
                        thisstd_flatten = np.concatenate((stdOrigin[i][extendMask], thisstd_flatten))
                        ax.scatter(wavesetOrigin[i][extendMask],
                                   fluxOrigin[i][extendMask], edgecolor="None",s=20, color="black")
                    if thiswaveset_flatten.max()<thisWaveMax:
                        extendMask = wavesetOrigin[i]>thiswaveset_flatten.max()
                        thiswaveset_flatten = np.concatenate((wavesetOrigin[i][extendMask],
                                                           thiswaveset_flatten))
                        thisflux_flatten = np.concatenate((fluxOrigin[i][extendMask], thisflux_flatten))
                        thisstd_flatten = np.concatenate((stdOrigin[i][extendMask], thisstd_flatten))
                        ax.scatter(wavesetOrigin[i][extendMask],
                                   fluxOrigin[i][extendMask], edgecolor="None", s=20, color="black")
                    ax.plot(wavesetOrigin[i], fluxOrigin[i], lw=2, color="r",
                                      label="center@{}".format(fiberIndex[i]))
                    ax.plot([wavesetOrigin[i].min(),
                             wavesetOrigin[i].max()], [0,0], lw=1, color="black")
                    ax.scatter(wavesetOrigin[i], fluxOrigin[i], s=15, edgecolor="None", color="r")
                    thisargIndex = np.argsort(thiswaveset_flatten)
                    thiswaveset_flatten = thiswaveset_flatten[thisargIndex]
                    thisflux_flatten = thisflux_flatten[thisargIndex]
                    thisstd_flatten = thisstd_flatten[thisargIndex]
                    knots = fixKnots(thiswaveset_flatten[np.isfinite(thisstd_flatten)], knotsOrigin)
                    eachOutputObj[i] = scipy.interpolate.LSQUnivariateSpline(thiswaveset_flatten,
                                                   thisflux_flatten,
                                                   knots[1:-1],
                                                   w=1/thisstd_flatten)
                    eachOutputStdObj[i] = scipy.interpolate.LSQUnivariateSpline(thiswaveset_flatten,
                                                   thisstd_flatten,
                                                   knots[1:-1],
                                                   w=1/thisstd_flatten)
                    ax.plot(thiswaveset_flatten, eachOutputObj[i](thiswaveset_flatten),
                            c="green", lw=1.5, label="result")
                    ax.plot(wavesetOrigin[i], eachOutputObj[i](wavesetOrigin[i])-fluxOrigin[i],
                            c="green", lw=1.5, ls="--", label="residual")
                    ax.plot(wavesetOrigin[i], fluxSmoothObj(wavesetOrigin[i])-fluxOrigin[i],
                            c="red", lw=1.5, ls="--", label="average residual")
                    ax.legend()
                    de.d(fig, os.path.join(debugDir, "_localSkyDebug", thisFileName))
                    plt.close(fig)

            todoStr="\t\t fitting local sky {:3d}/{:3d} array for {} Done!!!!\r".\
                    format(i, nSkyLike, name)
            sys.stdout.write(todoStr)
            sys.stdout.flush()
        #endregion: todo
        #!! not much points, easy to fit
        continumFitMask = processRegion(skyRegions, waveset_flatten, flux_smooth_flatten)
        continuumObj = scipy.interpolate.UnivariateSpline(waveset_flatten[continumFitMask],
                                       flux_smooth_flatten[continumFitMask], s=0)
        continuum = continuumObj(waveset_flatten)
        region_wave = waveset_flatten[continumFitMask]
        region_flux = flux_smooth_flatten[continumFitMask]
        #endregion: use good fibers and fit continuum
        #region: save result
        if not os.path.exists(thisFitPath):
            os.mkdir(thisFitPath)
        toSaves = ["flux_smooth_flatten", "std_smooth_flatten", "knots",
                   "continuumObj", "continuum", "region_wave", "region_flux",
                   "fluxSmoothObj", "stdSmoothObj", "waveset_flatten",
                   "flux_flatten","std_flatten", "skyLikeFiberMask",
                   "fiberGoodPixelFactor"]
        if prefix=="_working":
            toSaves = toSaves +\
                  ["localSmooth5pmIndexArray", "localSmooth5MaxpmIndexArray",
                   "localSmooth10tIndexArray", "localSmooth10pmIndexArray",
                   "localSmooth10MaxpmIndexArray", "localSmooth20tIndexArray",
                   "localSmooth15pmIndexArray", "localSmooth15MaxpmIndexArray",
                   "localSmooth30tIndexArray", "localSmooth5pmObjArray",
                   "localSmooth5MaxpmObjArray",
                   "localSmooth10tObjArray", "localSmooth10pmObjArray",
                   "localSmooth10MaxpmObjArray", "localSmooth20tObjArray",
                   "localSmooth15pmObjArray", "localSmooth15MaxpmObjArray",
                   "localSmooth30tObjArray", "localSmooth5pmStdObjArray",
                   "localSmooth5MaxpmStdObjArray",
                   "localSmooth10tStdObjArray", "localSmooth10pmStdObjArray",
                   "localSmooth10MaxpmStdObjArray", "localSmooth20tStdObjArray",
                   "localSmooth15pmStdObjArray", "localSmooth15MaxpmStdObjArray",
                   "localSmooth30tStdObjArray"]
        result = {}; l = locals()
        for each in toSaves: result[each] = l[each]
        print("saving data to {} ".format(thisFitPath), end="")
        for each in toSaves:
            thisFitFile = os.path.join(thisFitPath, each+".dat")
            with open(thisFitFile, "wb") as f:
                pickle.dump(result[each], f)
        logger.p("saving data done")
        #endregion: save result
    else:
        print("\tloading from {} ".format(thisFitPath), end="")
        toLoads = glob.glob(thisFitPath+"/*")
        result = {}
        for eachfile in toLoads:
            eachkey = os.path.basename(eachfile).split(".")[0]
            with open(eachfile, "rb") as f:
                result[eachkey] = pickle.load(f)
        logger.p("\tloading done")
    return result

def rejectSpline(x, y, k, sigma, sigmaRej=5, maxiter=10):
    w = 1/sigma
    for i in range(maxiter):
        k = k[k<x[w>0][-1]]
        k = k[k>x[w>0][0]]
        obj = scipy.interpolate.LSQUnivariateSpline(x, y, k, w)
        yMod = obj(x)
        diff = np.abs(yMod-y)/sigma
        w[diff>sigmaRej] = 0
        if np.isnan(yMod).sum()>0:
            k = fixKnots(x[w>0],k)
            obj = scipy.interpolate.LSQUnivariateSpline(x, y, k, w)
            yMod = obj(x)
            diff = np.abs(yMod-y)/sigma
            w[diff>sigmaRej] = 0
            if np.isnan(yMod).sum()>0:
                print("skylike detection error!!!")
                ipdb.set_trace()
    yy=obj(x)
    return w>0,yy

def getPlotStartEnd(mask):
    mm = np.concatenate(([0], mask, [0]))
    d = np.diff(mm)
    starts = np.where(d==1)[0]
    ends = np.where(d==-1)[0]-1
    return list(zip(starts, ends))

def profile2(A, sigma, mu, x):
    return A/(np.sqrt(2*np.pi)*sigma)*np.exp(-((x-mu)/sigma)**2)
def profile3(A3, sigma3, mu, x):
    diff = (x-mu)/sigma3
    return A3/(2.575798634*sigma3)*np.exp(-np.abs(diff)*diff**2)

def FWHM(A, sigma, A3, sigma3):
    muLow = np.zeros(A.shape)
    muHigh = (sigma+sigma3)*3
    mu = (muLow+muHigh)/2
    maxiter = 20
    halfMax = 0.5*(profile2(A, sigma, 0, 0)+profile3(A3, sigma3, 0, 0))
    for i in range(maxiter):
        values = profile2(A, sigma, 0, mu)+profile3(A3, sigma3, 0, mu)

        largeMask = values>halfMax
        muLow[largeMask] = mu[largeMask]
        mu[largeMask] = (mu[largeMask]+muHigh[largeMask])/2

        smallMask = np.logical_not(largeMask)
        muHigh[smallMask] = mu[smallMask]
        mu[smallMask] = (mu[smallMask]+muLow[smallMask])/2
    return mu

def getSigma(fittingResults, yset, waveset, result, skyFlag):
    A = fittingResults["gaussianA"]
    totalA = fittingResults["A"] + fittingResults["A3"]
    status = fittingResults["status"]
    goodLines = (status>0)&(np.bitwise_and(status, 0b110)==0)&(skyFlag>0)
    gf = fittingResults[goodLines]
    sigma = gf["gaussianSigma"]
    mus = gf["mu"]
    de.p(mus, sigma, "r.-");de.p(waveset, result);de.s()
    ipdb.set_trace()




""" notes about status
           1: in range
       0b010: partial reject
       0b100: full reject
     0b01000: small flux
     0b11000: small flux used (constrained to be small)
   0b0100000: select as residual constrain point
   0b1100000: use as residual constrain point
 0b010000000: flux too low, use default
 0b100000000: flux too high, use default
0b1000000000: do Residual Continuum Substraction
"""
dtype_m = np.dtype([("mu", np.float64),
                    ("A", np.float64),
                    ("sigma", np.float64),
                    ("A3", np.float64),
                    ("sigma3", np.float64),
                    ("status", np.int64),
                    ("sigmaError", np.float64),
                    ("sigmaErrorUpper", np.float64),
                    ("residualSigmaError", np.float64),
                    ("gaussianMu", np.float64),
                    ("gaussianA", np.float64),
                    ("gaussianSigma", np.float64)])

maxLegendre = 7
LegendreBoundary  = np.zeros((maxLegendre+1,2))
LegendreDBoundary = np.zeros((maxLegendre+1,2))
for i in range(maxLegendre+1):
    coeffs = np.zeros((maxLegendre+1,))
    coeffs[i] = 1
    obj = Legendre(coeffs)
    dobj = obj.deriv()
    LegendreBoundary[i][0] = obj(-1)
    LegendreBoundary[i][1] = obj(1)
    LegendreDBoundary[i][0] = dobj(-1)
    LegendreDBoundary[i][1] = dobj(1)

def plotMus(wave, flux, std, skywave, sky_lower, sky_upper):
    plt.plot(wave, flux, c="blue")
    plt.plot(wave, std, c="pink")
    for eachcen, eachlower, eachupper in zip(skywave, sky_lower, sky_upper):
        plt.plot([eachcen, eachcen], [0, 1], "--", c="black")
        plt.plot([eachlower, eachlower], [0, 1], "--", c="green")
        plt.plot([eachupper, eachupper], [0, 1], "--", c="red")

dtype_polyDebug = np.dtype([
    ("block", np.int64),
    ("start", np.float64),
    ("startTrue", np.float64),
    ("endTrue", np.float64),
    ("end", np.float64),
    ("nPoly", np.int64),
    ("lower", np.object),
    ("p0", np.object),
    ("upper", np.object)
])


def getLineMarks(filename, wavesetall, ysetSmooth, finalGoodMask,
                 modelFittingResults, blocks, rect=False):
    waveset = wavesetall[finalGoodMask]
    goodYsetSmooth = ysetSmooth[finalGoodMask]
    assert waveset.shape[0]==goodYsetSmooth.shape[0]
    if not rect:
        assert waveset.shape[0]==modelFittingResults.shape[0]
    with open(filename, "a") as f:
        f.write(regStrH.format(color="black"))
        for j in range(waveset.shape[0]):
            colorMod = j%2
            wave = waveset[j]
            yInd = goodYsetSmooth[j]
            ws = wave.size
            color = ["red", "green"][colorMod]
            ws = wave.size
            if rect:
                modelFittingResult = modelFittingResults[0]
            else:
                modelFittingResult = modelFittingResults[j]
            for i in range(modelFittingResult.shape[0]):
                eachstatus = modelFittingResult[i]["status"]
                if eachstatus<1 and not rect:
                    continue
                eachmu = modelFittingResult[i]["mu"]
                index = wave.searchsorted(eachmu)
                if index!=0 and index!=ws:
                    #!! index - 1 + 1
                    thisx = index + (eachmu-wave[index-2])/(wave[index]-wave[index-1])
                    f.write(regStrX(thisx, yInd[index]+1, 1, color))
            for i in range(blocks.shape[0]):
                eachstart = blocks[i][0]
                eachend = blocks[i][1]
                eachstartOffset = blocks[i][2]
                eachendOffset = blocks[i][3]

                ls = 0.5
                index = wave.searchsorted(eachstart)
                if index!=0 and index!=ws:
                    f.write(regStrE.format(index, yInd[index]+1, 2*ls, 1*ls, color))
                index = wave.searchsorted(eachend)
                if index!=0 and index!=ws:
                    f.write(regStrE.format(index, yInd[index]+1, 1*ls, 2*ls, color))
                index = wave.searchsorted(eachstartOffset)
                if index!=0 and index!=ws:
                    f.write(regStrB.format(index, yInd[index]+1, 2*ls, 1*ls, color))
                index = wave.searchsorted(eachendOffset)
                if index!=0 and index!=ws:
                    f.write(regStrB.format(index, yInd[index]+1, 1*ls, 2*ls, color))
#endregion: aux functions

def makeSkySub(inputFile, configs={}):
    #region: setups
    inputFileName = os.path.basename(inputFile).split(".fits")[0]
    outputFileName= inputFileName + '-final.dat'
    configs.update({"config2show":["rerun", "debug", "refit", "pngdebug",
                    "replot", "lastPlot","parallel", "fileStart", "fileEnd",
                    "cmdDebugFlatResult", "cmdDebugSkyStack"]})
    exec(messager.initLoggerMacro)
    exec(m2fsConfigs.singleFileMacro)
    if toReturn: return

    outfitsfile = os.path.join(outputPath, inputFileName+"-final.fits")
    #endregion: setups
    configs["logger"] = logger
    map2dFile = inputFile.split(".fits")[0] + "Map2D.fits"
    #de.plt.switch_backend("Agg")
    result = skysub(inputFile, configs=configs)
    #de.plt.switch_backend(configs["plotBackend"])
    if result is None:# for debug stop
        return None
    #region: save files
    obj = pyfits.open(map2dFile)
    bmod = obj["bmod"].data
    map2d = obj["MAP2DO"].data
    obj = pyfits.open(inputFile)
    exptime = obj[0].header["EXPTIME"]
    fimage = obj["sci"].data
    fstd   = obj["std"].data

    #!! of size fiberIndex
    fiberIndex = result["fiberIndex"]
    #!! flux = originData * factor
    fluxFactors = result["fluxFactors"]
    finalGoodMask = result["finalGoodMask"]
    names = result["names"]
    debugDir = result["debugDir"]

    waveset = result["waveset"]
    finalFittingGoodMask = result["finalFittingGoodMask"]
    flux  = result["flux"]/fluxFactors #!! of size fiberIndex(128>len(yset)>len(fierIndex))
    ivar  = result["invvar"]*(fluxFactors**2)
    yset  = result["yset"]
    ysetSmooth = result["ysetSmooth"]
    blocks = result["blocks"]
    #shutil.copy(inputFile.split(".fits")[0] + ".reg", inputFile.split(".fits")[0] + "-final.reg")
    names = result["names"]
    modelFittingResults = result["fittingResults"]
    localSkyArrays = result["localSkyArrays"]
    localSkynames = result["localSkynames"]

    goodYsetSmooth = ysetSmooth[finalGoodMask]
    with open(inputFile.split(".fits")[0] + "-final.reg", "w") as f:
        assert len(names) == waveset.shape[0]
        f.write(regStrH.format(color="black"))
        for j in range(waveset.shape[0]):
            colorMod = j%2
            wave = waveset[j]
            yInd = goodYsetSmooth[j]
            ws = wave.size
            name = names[j]
            modelFittingResult = modelFittingResults[j]

            color = ["red", "green"][colorMod]
            f.write(regStrT.format(-20,yInd[0], "{}".format(name), color))
            ws = wave.size
            for i in range(modelFittingResult.shape[0]):
                eachstatus = modelFittingResult[i]["status"]
                if eachstatus<1:
                    continue
                eachmu = modelFittingResult[i]["mu"]
                index = wave.searchsorted(eachmu)
                if index!=0 and index!=ws:
                    #!! index - 1 + 1
                    thisx = index + (eachmu-wave[index-2])/(wave[index]-wave[index-1])
                    f.write(regStrX(thisx, yInd[index]+1, 1, color))
            for i in range(blocks.shape[0]):
                eachstart = blocks[i][0]
                eachend = blocks[i][1]
                eachstartOffset = blocks[i][2]
                eachendOffset = blocks[i][3]

                ls = 0.5
                index = wave.searchsorted(eachstart)
                if index!=0 and index!=ws:
                    f.write(regStrE.format(index, yInd[index]+1, 2*ls, 1*ls, color))
                index = wave.searchsorted(eachend)
                if index!=0 and index!=ws:
                    f.write(regStrE.format(index, yInd[index]+1, 1*ls, 2*ls, color))
                index = wave.searchsorted(eachstartOffset)
                if index!=0 and index!=ws:
                    f.write(regStrB.format(index, yInd[index]+1, 2*ls, 1*ls, color))
                index = wave.searchsorted(eachendOffset)
                if index!=0 and index!=ws:
                    f.write(regStrB.format(index, yInd[index]+1, 1*ls, 2*ls, color))

    #region: get gaussian fitting result
    gaussianFitting = result["gaussianFitting"]
    fittingResults = result["fittingResults"]
    allSkyLines = result["allSkyLines"]
    skyFlag = allSkyLines["flag"]
    skyFlag = skyFlag.reshape((1,skyFlag.size))
    sky_flag = allSkyLines["flag"]
    status = fittingResults["status"]
    gaussianSigma = fittingResults["gaussianSigma"]
    gaussianMu = fittingResults["gaussianMu"]
    goodLines = (status>0)&(np.bitwise_and(status, 0b110)==0)&(skyFlag>0)&(np.isfinite(gaussianMu))
    c = cm.get_cmap("rainbow")
    c = c(np.linspace(0,1,fiberIndex.max()+1))
    smoothGaussianFitting = np.zeros(gaussianFitting.shape)
    fig1 = plt.figure()
    ax1 = fig1.add_subplot(111)
    fig2 = plt.figure()
    ax2 = fig2.add_subplot(111)
    gaussianSigmaLeg = []
    waveMin = gaussianMu[gaussianMu>0].min()
    waveMax = gaussianMu[gaussianMu>0].max()
    for i in range(fittingResults.shape[0]):
        xFit = gaussianMu[i][goodLines[i]]
        yFit = gaussianSigma[i][goodLines[i]]
        ax1.plot(xFit, yFit, "r.-", c=c[fiberIndex[i]])
        ax1.plot(waveset[i], gaussianFitting[i], c=c[fiberIndex[i]])
        if xFit.size>3:
            obj = Legendre.fit(xFit, yFit, 3)
        else:
            obj = Legendre.fit(np.linspace(waveMin, waveMax, 10), np.zeros(10), 3)
        gaussianSigmaLeg.append(obj)
        thisSmoothGaussianFitting = obj(waveset[i])
        smoothGaussianFitting[i,:] = thisSmoothGaussianFitting[:]
        ax2.plot(waveset[i], thisSmoothGaussianFitting, c=c[fiberIndex[i]])
        ax2.scatter(gaussianMu[i][goodLines[i]], gaussianSigma[i][goodLines[i]], c=c[fiberIndex[i]])
    de.d(fig1, os.path.join(debugDir, "gaussianSigmaModel.fig"))
    de.d(fig2, os.path.join(debugDir, "gaussianSigmaSmoothModel.fig"))
    plt.close(fig1)
    plt.close(fig2)
    #endregion: get gaussian fitting result

    #!! of size fiberIndex * nx
    skyLineModel = result["skyLineModel"]/fluxFactors
    continuumModel = result["continuumModel"]/fluxFactors
    continuumResidualModel = result["continuumResidualModel"]/fluxFactors
    objectContinuumModel = result["objectContinuumModel"]/fluxFactors
    smoothSky = result["smoothSky"]/fluxFactors
    localSmoothSkyList = [localSkyArrays[each]/fluxFactors for each in localSkynames]
    blocks = result["blocks"]
    inblocks = getInBlocks(waveset, blocks)
    continuumModel[~inblocks] = continuumModel[~inblocks] + objectContinuumModel[~inblocks]
    #!! origin flux(need this to generate 2d image)

    fluxAll = np.zeros((finalGoodMask.size, flux.shape[1]))
    finalLocalSmoothSkyList = [np.zeros((finalGoodMask.size, flux.shape[1])) for each in localSmoothSkyList]
    finalFittingGoodMaskAll = np.zeros((finalGoodMask.size, flux.shape[1]))
    fiberIndexAll = np.zeros((finalGoodMask.size, ))
    wavesetAll = np.zeros((finalGoodMask.size, flux.shape[1]))
    zerosAll = np.zeros((finalGoodMask.size, flux.shape[1]))
    factorAll = np.zeros((finalGoodMask.size, flux.shape[1]))
    ivarAll = np.zeros((finalGoodMask.size, flux.shape[1]))
    smoothSkyAll = np.zeros((finalGoodMask.size, flux.shape[1]))
    objectContinuumModelAll = np.zeros((finalGoodMask.size, flux.shape[1]))
    skyLineModelAll = np.zeros((finalGoodMask.size, flux.shape[1]))
    continuumModelAll = np.zeros((finalGoodMask.size, flux.shape[1]))
    continuumResidualModelAll = np.zeros((finalGoodMask.size, flux.shape[1]))

    fluxAll[finalGoodMask,:] = flux[:,:]
    for i in range(len(localSmoothSkyList)):
        finalLocalSmoothSkyList[i][finalGoodMask,:] = localSmoothSkyList[i][:,:]
    finalFittingGoodMaskAll[finalGoodMask,:] = finalFittingGoodMask[:,:]
    fiberIndexAll[finalGoodMask] = fiberIndex[:]
    wavesetAll[finalGoodMask,:] = waveset[:,:]
    factorAll[finalGoodMask,:] =fluxFactors[:,:]
    ivarAll[finalGoodMask,:] = ivar[:,:]
    smoothSkyAll[finalGoodMask,:] = smoothSky[:,:]
    objectContinuumModelAll[finalGoodMask,:] = objectContinuumModel[:,:]
    skyLineModelAll[finalGoodMask,:] = skyLineModel[:,:]
    continuumModelAll[finalGoodMask,:] = continuumModel[:,:]
    continuumResidualModelAll[finalGoodMask,:] = continuumResidualModel[:,:]

    ivarFittingAll = ivarAll*finalFittingGoodMaskAll
    bzeros = np.zeros(bmod.shape)
    bones  = np.ones(ivarAll.shape).astype(np.bool)
    fimageClean = fimage - bmod
    fimageExtract = extract.map2DModel(bzeros, map2d, fluxAll, ivarAll)
    fimageFixedWidthMask = extract.map2DModelFixNaNWidth(bzeros, map2d, zerosAll, ivarAll, ysetSmooth)
    fimageFittingFixedWidthMask =\
            extract.map2DModelFixNaNWidth(bzeros, map2d, zerosAll, ivarFittingAll, ysetSmooth)
    fimageSkyLineModel = extract.map2DModel(bzeros, map2d, skyLineModelAll, ivarFittingAll)
    fimageContinuum = extract.map2DModel(bzeros, map2d, continuumModelAll, ivarFittingAll)
    fimageContinuumResidual = extract.map2DModel(bzeros, map2d, continuumResidualModelAll, ivarFittingAll)
    fimageObjectContinuum = extract.map2DModel(bzeros, map2d, objectContinuumModelAll, ivarFittingAll)
    #fimageSkyModel = fimageSkyLineModel+fimageContinuum-fimageObjectContinuum
    fimageSkyModel = fimageSkyLineModel+fimageContinuum
    fimageResidual = fimageClean-fimageSkyModel
    fimageExtractResidual = fimageClean - fimageExtract
    fimageSmoothSkyModel = extract.map2DModel(bzeros, map2d, smoothSkyAll, bones)
    fimageSmoothSkyResidual = fimageClean-fimageSmoothSkyModel
    fimageLocalSkyModel = [extract.map2DModel(bzeros, map2d, each, bones)
                             for each in finalLocalSmoothSkyList]
    fimageLocalSkyResidual = [fimageClean-each for each in fimageLocalSkyModel]

    namePairs = [("fluxFactors", "fluxFactors"), ("fiberIndexAll", "fiberIndex"),
                 ("finalGoodMask", "finalGoodMask"), ("names", "names"),
                 ("flux","flux"), ("ivar", "ivar"), ("exptime", "exptime"),
                 ("yset", "yset"), ("ysetSmooth", "ysetSmooth"), ("wavesetAll", "waveset"),
                 ("gaussianSigmaLeg", "gaussianSigmaLeg"), ("blocks", "blocks"),
                 ("modelFittingResults", "modelFittingResults"),
                 ("localSkynames","localSkynames"),
                 #("continuumModel", "continuumModelAll"), ("skyLineModel", "skyLineModel"),
                 #("fimageClean", "fimageClean"),
                 #("fimageSkyLineModel", "fimageSkyLineModel"),
                 #("fimageContinuum", "fimageContinuum"),
                 ("smoothGaussianFitting", "smoothGaussianFitting")]
    finalResult = macro.genResults(namePairs, locals())
    with open(outFile, "wb") as f:
        pickle.dump(finalResult, f)
    header1 = pyfits.Header()
    header2 = pyfits.Header()
    header3 = pyfits.Header()
    header4 = pyfits.Header()
    header5 = pyfits.Header()
    header6 = pyfits.Header()
    hls = [pyfits.Header({"OBJECT":"data - {}".format(eachname), "EXTNAME":eachname})
                for eachname in localSkynames]
    header7 = pyfits.Header()
    header75 = pyfits.Header()
    header77 = pyfits.Header()
    header8 = pyfits.Header()
    header9 = pyfits.Header()
    header10 = pyfits.Header()

    header1["OBJECT"] = "extractModel"
    header2["OBJECT"] = "skymodel"
    header3["OBJECT"] = "smoothSky"
    header4["OBJECT"] = "data without background"
    header5["OBJECT"] = "data-extractModel"
    header6["OBJECT"] = "data-skymodel"
    header7["OBJECT"] = "data-smoothSky"
    header75["OBJECT"] = "fixedWidthMask"
    header77["OBJECT"] = "fixedFittingWidthMask"
    header8["OBJECT"] = "std"
    header9["OBJECT"] = "continuum"
    header10["OBJECT"] = "continuumResidual"

    header1["EXTNAME"] = "EXTRACT"
    header2["EXTNAME"] = "SKY"
    header3["EXTNAME"] = "SSKY" #!! in use
    header4["EXTNAME"] = "DATA" #!! in use
    header5["EXTNAME"] = "ERESIDUAL" #!! in use
    header6["EXTNAME"] = "RESIDUAL" #!! in use
    header7["EXTNAME"] = "SSKYRESIDUAL" #!! in use
    header75["EXTNAME"] = "MASK" #!! in use
    header77["EXTNAME"] = "FITTINGMASK" #!! in use
    header8["EXTNAME"] = "STD" #!! in use
    header9["EXTNAME"] = "CONTINUUM"
    header10["EXTNAME"] = "CONTINUUMRES"
    data1= pyfits.PrimaryHDU(fimageExtract,         header=header1)
    data2= pyfits.ImageHDU(fimageSkyModel,          header=header2)
    data3= pyfits.ImageHDU(fimageSmoothSkyModel,    header=header3)
    data4= pyfits.ImageHDU(fimageClean,             header=header4)
    data5= pyfits.ImageHDU(fimageExtractResidual,   header=header5)
    data6= pyfits.ImageHDU(fimageResidual,          header=header6)
    data7= pyfits.ImageHDU(fimageSmoothSkyResidual, header=header7)
    dls = [pyfits.ImageHDU(eachArray, header=eachheader)
             for eachArray, eachheader in zip(fimageLocalSkyResidual, hls)]
    data75= pyfits.ImageHDU(fimageFixedWidthMask, header=header75)
    data77= pyfits.ImageHDU(fimageFittingFixedWidthMask, header=header77)
    data8= pyfits.ImageHDU(fstd,                    header=header8)
    data9= pyfits.ImageHDU(fimageContinuum,         header=header9)
    data10= pyfits.ImageHDU(fimageContinuumResidual,header=header10)
    #fits  = pyfits.HDUList([data1, data2, data3, data31, data32, data4, data5])
    fits  = pyfits.HDUList([data1, data2, data3, data4, data5,
                            data6, data7] + dls + [data75, data77, data8, data9, data10])
    fits.writeto(outfitsfile, clobber='True')
    #endregion: save files

""" notes
    continuum is made by skylike fibers
"""
def skysub(infile, configs={}):
    #region: setups
    outputPath = configs["outputPath"]
    name = os.path.basename(infile).split(".fits")[0]
    debugDir = os.path.join(outputPath, name+"SkySubDebug")
    extract.safe_mkdir(os.path.join(debugDir, "data"))
    masterFlatFieldFile = os.path.join(outputPath, "masterFlatField.fits")
    debug = configs.get("debug", 1)
    rerun = configs.get("rerun", 0)
    refit = configs.get("refit", False)
    replot = configs.get("replot", False)
    pngdebug = configs.get("pngdebug",False)
    fitAllSky = configs.get("fitAllSky",False)
    lastPlot = configs.get("lastPlot", True)
    reGenSkyUpperModel = configs.get("reGenSkyUpperModel",False)
    pngdebugResidual = configs.get("pngdebugResidual", False)
    fitFile = name+"SkyFit"
    logger = configs["logger"]
    codePath = configs["codePath"]
    skylineFlat = configs.get("useSkylineFlat", False)
    useSkylike = configs.get("useSkylike", False)
    debugFlatResult = configs.get("cmdDebugFlatResult", False)
    cmdDebugSkyStack = configs.get("cmdDebugSkyStack", False)
    skyLikeSigmaReject = configs.get("skyLikeSigmaReject", 3)
    goodFiberPixelRatio = configs.get("goodFiberPixelRatio", 0.7)
    if debugFlatResult or cmdDebugSkyStack:
        configs["refit"] = True
        refit=True
    parallelN = 3
    time0 = time.time()
    #endregion: setups
    #region: read flux data
    #!! this yset if from traceset
    thisWaveFile = name
    waveset, coeffs, ysetSmooth, badFiber, debugInfo = wavecalib.readWaveSet(outputPath, name=thisWaveFile, debugInfo=2)
    #waveset, coeffs, ysetSmooth, badFiber, debugInfo = wavecalib.readWaveSet(outputPath, debugInfo=1)
    rms = debugInfo["rms"].flatten()#!! all fiber that have been traceseted
    skylineScaleFactorSmooth = debugInfo["smoothA"]
    finalGoodFiber = np.isfinite(rms)#!! good for wavelength
    waveset = waveset[finalGoodFiber]
    coeffs = coeffs[finalGoodFiber]

    goodFiber = badFiber<1
    ysetSmooth = ysetSmooth[goodFiber]

    fiberIndex = np.where(goodFiber)[0]
    goodFiberIndex = fiberIndex
    fiberIndex = fiberIndex[finalGoodFiber]#!! origin index in 0~127
    #!! read map2D and origin image
    thisMap2DFile = infile.split(".fits")[0] + "Map2D.fits"
    obj = pyfits.open(thisMap2DFile)
    bmod = obj["bmod"].data
    map2D = obj["MAP2DO"].data
    obj = pyfits.open(infile)
    fimage = obj["sci"].data
    fstd = obj["std"].data
    #!! read flux
    infile = infile.split(".fits")[0] + "Flux.fits"
    thisData, thisStd, thisYset = extract.readFlux(infile)
    thisData = thisData[finalGoodFiber]
    thisStd = thisStd[finalGoodFiber]
    #!! filter wield point(sigma tooooo small), some bug..
    smallStd = np.nanpercentile(thisStd, 1)
    bugPoints = thisStd<smallStd/10
    if bugPoints.sum():
        logger.warn("there are some bug points(sigma toooooo small), filter them\n{}\n{}".\
                format(np.where(bugPoints)[0], np.where(bugPoints)[1]))
    thisStd[bugPoints] = np.inf
    #!!
    thisYset = thisYset[finalGoodFiber]
    yset = thisYset
    rms = rms[finalGoodFiber]
    thisIvar = 1/thisStd**2
    thisBad   = thisIvar==0
    skylineScaleFactorSmooth = skylineScaleFactorSmooth[finalGoodFiber]
    nFiber, nx = thisData.shape#!! nFiber in length of  finalgood
    #!! read master flatfield
    obj = pyfits.open(masterFlatFieldFile)
    fiberFlatSmooth, allFactor, flatFiberIndex = obj[0].data, obj[1].data, obj[2].data
    flatGoodMask = np.array([each in fiberIndex for each in flatFiberIndex.flatten()], dtype=np.bool)
    fiberFlatSmooth = fiberFlatSmooth[flatGoodMask]
    allFactor = allFactor[flatGoodMask]
    #!! read fiberInfo
    fiberInfo = configs["fiberInfo"]
    fiberInfo = fiberInfo[fiberIndex]#!! origin index in 0~127
    #ipdb.set_trace()
    #ipdb.set_trace()
    #!! mask in size of finalgood (127>??>finalgood)
    skyMask = fiberInfo["type"]==b"S"
    objMask = fiberInfo["type"]==b"T"
    unplugMask = fiberInfo["type"]==b"I"
    assert skyMask.sum()+objMask.sum()+unplugMask.sum()==fiberInfo.size
    if not (nFiber==fiberIndex.size==allFactor.size==fiberInfo.size):
        print("{} {} {} {}".format(nFiber,fiberIndex.size,allFactor.size,fiberInfo.size))
        ipdb.set_trace()
        raise logger.Exception("{} {} {} {}".format(nFiber,fiberIndex.size,allFactor.size,fiberInfo.size))
    if skylineFlat:
        factor = 1 / fiberFlatSmooth * allFactor.reshape((allFactor.size, 1)) / skylineScaleFactorSmooth
        factorDebug = 1 / fiberFlatSmooth * allFactor.reshape((allFactor.size, 1))
    else:
        factor = 1 / fiberFlatSmooth * allFactor.reshape((allFactor.size, 1))
        factorDebug = 1 / fiberFlatSmooth * allFactor.reshape((allFactor.size, 1)) / skylineScaleFactorSmooth
    if not (np.all(thisStd>0) and np.all(skylineScaleFactorSmooth>0) and\
            np.all(factor>0) and np.all(factorDebug>0) and\
            np.all(fiberFlatSmooth>0) and np.all(allFactor>0)):
        logger.p("np.all(thisStd>0):{}".format(np.all(thisStd>0)))
        logger.p("np.all(skylineScaleFactorSmooth>0):{}".format(np.all(skylineScaleFactorSmooth>0)))
        logger.p("np.all(factor>0):{}".format(np.all(factor>0)))
        logger.p("np.all(factorDebug>0):{}".format(np.all(factorDebug>0)))
        logger.p("np.all(fiberFlatSmooth>0):{}".format(np.all(fiberFlatSmooth>0)))
        logger.p("np.all(allFactor>0):{}".format(np.all(allFactor>0)))
        ipdb.set_trace()
    aveFactors = factor.mean(axis=1)
    aveFactors = aveFactors/aveFactors.max()
    #!! flux = originData * factor
    flux = thisData * factor
    std  = thisStd * factor
    invvar = 1/std**2
    fluxDebug = thisData * factorDebug
    stdDebug  = thisStd * factorDebug
    invvarDebug = 1/stdDebug**2
    #!! setup plot cm
    c = cm.get_cmap("rainbow")
    c = c(np.linspace(0,1,nFiber))
    #!! generate figure name
    names = ["_".join([name, str(fiberIndex[i]),
                       fiberInfo[i]["fiber"].decode('UTF-8'),
                       fiberInfo[i]["id"].decode('UTF-8')])
                for i in range(nFiber)]
    #!! names in range of 0~127
    namesSky = [each for i,each in enumerate(names) if skyMask[i] ]
    namesSkyIndex = [each for i,each in enumerate(fiberIndex) if skyMask[i]]
    namesObj = [each for i,each in enumerate(names) if objMask[i] ]
    namesObjIndex = [each for i,each in enumerate(fiberIndex) if objMask[i]]
    #!! up to here, len(yset)=120 and len(ysetSmooth)=121

    allSkyLines = readAllSkyData(".")
    skyRegions = readSkyRegion(".")#!! regions for sky continuum fitting
    blocks = readSkyBlock(".")#!! blocks info for sky line fitting
    #endregion: read flux data
    #region: read sky data
    mindiff = 0.1 #!! the min distance of two center
    skywave = allSkyLines["lambda"]
    assert np.all(np.diff(skywave)>0)
    skysigma = allSkyLines["sigma"]
    skysigma3 = allSkyLines["sigma3"]
    skytype = allSkyLines["type"]
    skyGroup = allSkyLines["group"]
    skyResidualType = allSkyLines["residualType"]
    skySigmaLower = allSkyLines["sigmalower"]
    skySigmaUpper = allSkyLines["sigmaupper"]
    skySigma3Lower = allSkyLines["sigma3lower"]
    skySigma3Upper = allSkyLines["sigma3upper"]
    skyALower = allSkyLines["Alower"]
    skyAUpper = allSkyLines["Aupper"]
    skyA3Lower = allSkyLines["A3lower"]
    skyA3Upper = allSkyLines["A3upper"]
    sky_upper = allSkyLines["muupper"]
    sky_lower = allSkyLines["mulower"]
    sky_flag = allSkyLines["flag"]
    middle_fit = (skywave[:-1]+skywave[1:])/2.0
    cross = np.where((sky_lower[1:] - sky_upper[:-1])<mindiff)[0]
    sky_upper[cross] = middle_fit[cross] - mindiff/2
    sky_lower[cross+1] = middle_fit[cross] + mindiff/2
    #endregion: read sky data
    #region: fit for sky and continuum
    waveset_sky = waveset[skyMask]
    flux_sky = flux[skyMask]
    std_sky = std[skyMask]
    index_sky = np.where(skyMask)[0]
    waveset_obj = waveset[objMask]
    flux_obj = flux[objMask]
    std_obj = std[objMask]
    index_obj = np.where(objMask)[0]

    thisConfig = {"debugDir":debugDir, "logger":logger,"name":name,
            "fitFile":fitFile, "refit":refit, "prefix":"_debug",
            "skyLikeSigmaReject":skyLikeSigmaReject, "names":names,
            "goodFiberPixelRatio":goodFiberPixelRatio}
    #!! detect skylike fibers and fit continuum
    resultDebug = getSkyLikeFibers(waveset, fluxDebug, stdDebug,
                              allSkyLines, skyRegions, skyMask,
                              objMask, fiberIndex, configs=thisConfig)
    thisConfig = {"debugDir":debugDir, "logger":logger, "name":name,
            "fitFile":fitFile, "refit":refit, "prefix":"_working",
            "skyLikeSigmaReject":skyLikeSigmaReject, "names":names,
            "goodFiberPixelRatio":goodFiberPixelRatio,
            "noOutput":debugFlatResult}
    result = getSkyLikeFibers(waveset, flux, std,
                              allSkyLines, skyRegions, skyMask,
                              objMask, fiberIndex, configs=thisConfig)
    #!! smooth sky fit from skylike fibers
    flux_smooth_flatten = result["flux_smooth_flatten"]
    std_smooth_flatten = result["std_smooth_flatten"]
    #!! sky continuum fit from skylike data
    continuumObj = result["continuumObj"]
    continuum = result["continuum"]
    region_wave = result["region_wave"]
    region_flux = result["region_flux"]
    fluxSmoothObj = result["fluxSmoothObj"]
    stdSmoothObj = result["stdSmoothObj"]
    #!! local and ave smooth sky result
    smoothSky = fluxSmoothObj(waveset)
    localSmooth5pmObjArray        = result["localSmooth5pmObjArray"]
    localSmooth5pmStdObjArray     = result["localSmooth5pmStdObjArray"]
    localSmooth5pmIndexArray      = result["localSmooth5pmIndexArray"]
    localSmooth5MaxpmObjArray    = result["localSmooth5MaxpmObjArray"]
    localSmooth5MaxpmStdObjArray = result["localSmooth5MaxpmStdObjArray"]
    localSmooth5MaxpmIndexArray  = result["localSmooth5MaxpmIndexArray"]
    localSmooth10tObjArray        = result["localSmooth10tObjArray"]
    localSmooth10tStdObjArray     = result["localSmooth10tStdObjArray"]
    localSmooth10tIndexArray      = result["localSmooth10tIndexArray"]

    localSmooth10pmObjArray       = result["localSmooth10pmObjArray"]
    localSmooth10pmStdObjArray    = result["localSmooth10pmStdObjArray"]
    localSmooth10pmIndexArray     = result["localSmooth10pmIndexArray"]
    localSmooth10MaxpmObjArray    = result["localSmooth10MaxpmObjArray"]
    localSmooth10MaxpmStdObjArray = result["localSmooth10MaxpmStdObjArray"]
    localSmooth10MaxpmIndexArray  = result["localSmooth10MaxpmIndexArray"]
    localSmooth20tObjArray        = result["localSmooth20tObjArray"]
    localSmooth20tStdObjArray     = result["localSmooth20tStdObjArray"]
    localSmooth20tIndexArray      = result["localSmooth20tIndexArray"]

    localSmooth15pmObjArray       = result["localSmooth15pmObjArray"]
    localSmooth15pmStdObjArray    = result["localSmooth15pmStdObjArray"]
    localSmooth15pmIndexArray     = result["localSmooth15pmIndexArray"]
    localSmooth15MaxpmObjArray    = result["localSmooth15MaxpmObjArray"]
    localSmooth15MaxpmStdObjArray = result["localSmooth15MaxpmStdObjArray"]
    localSmooth15MaxpmIndexArray  = result["localSmooth15MaxpmIndexArray"]
    localSmooth30tObjArray        = result["localSmooth30tObjArray"]
    localSmooth30tStdObjArray     = result["localSmooth30tStdObjArray"]
    localSmooth30tIndexArray      = result["localSmooth30tIndexArray"]
    #!! 
    localSmooth5pmArray        = np.zeros(smoothSky.shape)
    localSmooth5pmStdArray     = np.zeros(smoothSky.shape)
    localSmooth5MaxpmArray     = np.zeros(smoothSky.shape)
    localSmooth10tArray        = np.zeros(smoothSky.shape)
    localSmooth10tStdArray     = np.zeros(smoothSky.shape)
    localSmooth10pmArray       = np.zeros(smoothSky.shape)
    localSmooth10pmStdArray    = np.zeros(smoothSky.shape)
    localSmooth10MaxpmArray    = np.zeros(smoothSky.shape)
    localSmooth10MaxpmStdArray = np.zeros(smoothSky.shape)
    localSmooth20tArray        = np.zeros(smoothSky.shape)
    localSmooth20tStdArray     = np.zeros(smoothSky.shape)
    localSmooth15pmArray       = np.zeros(smoothSky.shape)
    localSmooth15pmStdArray    = np.zeros(smoothSky.shape)
    localSmooth15MaxpmArray    = np.zeros(smoothSky.shape)
    localSmooth15MaxpmStdArray = np.zeros(smoothSky.shape)
    localSmooth30tArray        = np.zeros(smoothSky.shape)
    localSmooth30tStdArray     = np.zeros(smoothSky.shape)
    for i in range(smoothSky.shape[0]):
        localSmooth5pmArray[i]        = localSmooth5pmObjArray[i](waveset[i])
        localSmooth5pmStdArray[i]     = localSmooth5pmStdObjArray[i](waveset[i])
        localSmooth5MaxpmArray[i]     = localSmooth5MaxpmObjArray[i](waveset[i])
        localSmooth10tArray[i]        = localSmooth10tObjArray[i](waveset[i])
        localSmooth10tStdArray[i]     = localSmooth10tStdObjArray[i](waveset[i])
        localSmooth10pmArray[i]       = localSmooth10pmObjArray[i](waveset[i])
        localSmooth10pmStdArray[i]    = localSmooth10pmStdObjArray[i](waveset[i])
        localSmooth10MaxpmArray[i]    = localSmooth10MaxpmObjArray[i](waveset[i])
        localSmooth10MaxpmStdArray[i] = localSmooth10MaxpmStdObjArray[i](waveset[i])
        localSmooth20tArray[i]        = localSmooth20tObjArray[i](waveset[i])
        localSmooth20tStdArray[i]     = localSmooth20tStdObjArray[i](waveset[i])
        localSmooth15pmArray[i]       = localSmooth15pmObjArray[i](waveset[i])
        localSmooth15pmStdArray[i]    = localSmooth15pmStdObjArray[i](waveset[i])
        localSmooth15MaxpmArray[i]    = localSmooth15MaxpmObjArray[i](waveset[i])
        localSmooth15MaxpmStdArray[i] = localSmooth15MaxpmStdObjArray[i](waveset[i])
        localSmooth30tArray[i]        = localSmooth30tObjArray[i](waveset[i])
        localSmooth30tStdArray[i]     = localSmooth30tStdObjArray[i](waveset[i])
    localSkys = np.array([{"ls5pm":localSmooth5pmArray[i],
                  "ls5Maxpm":localSmooth5MaxpmArray[i],
                  "ls10t":localSmooth10tArray[i],
                  "ls10pm":localSmooth10pmArray[i],
                  "ls10Maxpm":localSmooth10MaxpmArray[i],
                  "ls20t":localSmooth20tArray[i],
                  "ls15pm":localSmooth15pmArray[i],
                  "ls15Maxpm":localSmooth15MaxpmArray[i],
                  "ls30t":localSmooth30tArray[i]} for i in range(smoothSky.shape[0])])
    localSkyArrays = {"ls5pm":localSmooth5pmArray,
                      "ls5Maxpm":localSmooth5MaxpmArray,
                      "ls10t":localSmooth10tArray,
                      "ls10pm":localSmooth10pmArray,
                      "ls10Maxpm":localSmooth10MaxpmArray,
                      "ls20t":localSmooth20tArray,
                      "ls15pm":localSmooth15pmArray,
                      "ls15Maxpm":localSmooth15MaxpmArray,
                      "ls30t":localSmooth30tArray}
    #!!  
    waveset_flatten = result["waveset_flatten"]
    flux_flatten = result["flux_flatten"]
    std_flatten = result["std_flatten"]
    skyLikeFiberMask = result["skyLikeFiberMask"]
    skyLikeFiberMaskDebug = resultDebug["skyLikeFiberMask"]
    fiberGoodPixelFactor = result["fiberGoodPixelFactor"]
    assert fiberGoodPixelFactor.size==len(names)

    #!! bad pixel bug fixed knots
    knots = result["knots"]
    assert np.all(np.isfinite(continuum))

    waveset_sky_like = waveset[skyLikeFiberMask]
    flux_sky_like = flux[skyLikeFiberMask]
    std_sky_like = std[skyLikeFiberMask]
    index_sky_like = np.where(skyLikeFiberMask)[0]
    namesSkyLike = [each for i,each in enumerate(names) if skyLikeFiberMask[i] ]
    namesSkyLikeIndex = [each for i,each in enumerate(fiberIndex) if skyLikeFiberMask[i]]
    namesNotSkyLike = [each for i,each in enumerate(names) if not skyLikeFiberMask[i] ]
    namesSkyLikeFactors = ["{}_{:.4f}".format(each,aveFactors[i]) for i,each in enumerate(names) if skyLikeFiberMask[i] ]
    namesNotSkyLikeFactors = ["{}_{:.4f}".format(each,aveFactors[i]) for i,each in enumerate(names) if not skyLikeFiberMask[i] ]
    namesLocalSmooth5pm = [[names[i] for i in eachIndexs] for eachIndexs in localSmooth5pmIndexArray]
    namesLocalSmooth10t = [[names[i] for i in eachIndexs] for eachIndexs in localSmooth10tIndexArray]
    namesLocalSmooth10pm = [[names[i] for i in eachIndexs] for eachIndexs in localSmooth10pmIndexArray]
    namesLocalSmooth10Maxpm = [[names[i] for i in eachIndexs] for eachIndexs in localSmooth10MaxpmIndexArray]
    namesLocalSmooth20t = [[names[i] for i in eachIndexs] for eachIndexs in localSmooth20tIndexArray]
    namesLocalSmooth15pm = [[names[i] for i in eachIndexs] for eachIndexs in localSmooth15pmIndexArray]
    namesLocalSmooth15Maxpm = [[names[i] for i in eachIndexs] for eachIndexs in localSmooth15MaxpmIndexArray]
    namesLocalSmooth30t = [[names[i] for i in eachIndexs] for eachIndexs in localSmooth30tIndexArray]

    with open(os.path.join(debugDir, "names.txt"),"w") as f: #!! for eye debug...
        f.write("namesSky:")
        f.write(str(namesSky))
        f.write("\n")
        f.write("namesObj:")
        f.write(str(namesObj))
        f.write("\n")
        f.write("namesSkyLike:")
        f.write(str(namesSkyLike))
        f.write("\n")
        f.write("namesNotSkyLike:")
        f.write(str(namesNotSkyLike))
        f.write("\n")
        f.write("namesSkyLikeFactors:")
        f.write(str(namesSkyLikeFactors))
        f.write("\n")
        f.write("namesNotSkyLikeFactors:")
        f.write(str(namesNotSkyLikeFactors))
        f.write("\n")
        f.write("                          name isSkylike factor\n")
        for eachName, eachSkylike, eachSkylikeFactor in zip(names, skyLikeFiberMask, fiberGoodPixelFactor):
            f.write("{:30s} {:9} {:6.2f}\n".format(eachName, eachSkylike, eachSkylikeFactor))
    with open(os.path.join(debugDir, "localSkyIndex.txt"),"w") as f: #!! for eye debug...
        for i in range(len(names)):
            f.write("{}:\n".format(names[i]))
            f.write("\t5pm:{}\n".format(namesLocalSmooth5pm[i]))
            f.write("\t10t:{}\n".format(namesLocalSmooth10t[i]))
            f.write("\t10pm:{}\n".format(namesLocalSmooth10pm[i]))
            f.write("\t10Maxpm:{}\n".format(namesLocalSmooth10Maxpm[i]))
            f.write("\t20t:{}\n".format(namesLocalSmooth20t[i]))
            f.write("\t15pm:{}\n".format(namesLocalSmooth15pm[i]))
            f.write("\t15Maxpm:{}\n".format(namesLocalSmooth15Maxpm[i]))
            f.write("\t30t:{}\n".format(namesLocalSmooth30t[i]))

    #endregion: fit for sky and continuum
    #region: debug plots
    saveConfig = {"xscale":0.1}
    if debug==1: #!! plot stack fig aboug all, objs and skys
        fig = de.f()
        ax = fig.add_subplot(111)
        for i in range(flux.shape[0]):
            ax.plot(waveset[i], flux[i], c=c[i])
            ax.plot(waveset[i], std[i], c=c[i], ls="--")
        ax.set_title("{}:useSkylineFlat:{} all fiber number {}".format(os.path.basename(infile), skylineFlat, flux.shape[0]))
        de.d(fig, os.path.join(debugDir,"all.fig"), configs=saveConfig)
        fig.clear()
        ax = fig.add_subplot(111)
        for i in np.where(objMask)[0]:
            ax.plot(waveset[i], flux[i], c=c[i])
            ax.plot(waveset[i], std[i], c=c[i], ls="--")
        ax.set_title("{}:useSkylineFlat:{} all objs fiber number {}".format(os.path.basename(infile), skylineFlat, objMask.sum()))
        de.d(fig, os.path.join(debugDir,"objs.fig"), configs=saveConfig)
        fig.clear()
        ax = fig.add_subplot(111)
        for i in np.where(skyMask)[0]:
            ax.plot(waveset[i], flux[i], c=c[i])
            ax.plot(waveset[i], std[i], c=c[i], ls="--")
        ax.set_title("{}:useSkylineFlat:{} all sky fiber number {}".format(os.path.basename(infile), skylineFlat, skyMask.sum()))
        de.d(fig, os.path.join(debugDir,"skys.fig"), configs=saveConfig)
        fig.clear()
        ax = fig.add_subplot(111)
        for i in np.where(skyLikeFiberMask)[0]:
            ax.plot(waveset[i], flux[i], c=c[i])
            ax.plot(waveset[i], std[i], c=c[i], ls="--")
        ax.set_title("{}:useSkylineFlat:{} all sky fiber number {} and all skylike fiber number {} and all fiber number {}".\
                format(os.path.basename(infile), skylineFlat, skyMask.sum(), skyLikeFiberMask.sum(), skyMask.size))
        de.d(fig, os.path.join(debugDir,"skyLikes.fig"), configs=saveConfig)
        fig.clear()
        ax = fig.add_subplot(111)
        for i in np.where(skyLikeFiberMaskDebug)[0]:
            ax.plot(waveset[i], fluxDebug[i], c=c[i])
            ax.plot(waveset[i], stdDebug[i], c=c[i], ls="--")
        ax.set_title("{}:useSkylineFlat:{} all sky fiber number {} and all skylike fiber number {} and all fiber number {}".\
                format(os.path.basename(infile), np.logical_not(skylineFlat), skyMask.sum(), skyLikeFiberMask.sum(), skyMask.size))
        de.d(fig, os.path.join(debugDir,"skyLikesOtherMode.fig"), configs=saveConfig)
        fig.clear()
        plt.close(fig)
        fig=de.f()
        ax0 = fig.add_subplot(121)
        ax1 = fig.add_subplot(122, sharex=ax0, sharey=ax0)
        for i in np.where(skyLikeFiberMask)[0]:
            ax0.plot(waveset[i], flux[i], c=c[i])
            ax0.plot(waveset[i], std[i], c=c[i], ls="--")
        ax0.set_title("{}:useSkylineFlat:{} all sky fiber number {} and all skylike fiber number {} and all fiber number {}".\
                format(os.path.basename(infile), skylineFlat, skyMask.sum(), skyLikeFiberMask.sum(), skyMask.size))
        for i in np.where(skyLikeFiberMaskDebug)[0]:
            ax1.plot(waveset[i], fluxDebug[i], c=c[i])
            ax1.plot(waveset[i], stdDebug[i], c=c[i], ls="--")
        ax1.set_title("{}:useSkylineFlat:{} all sky fiber number {} and all skylike fiber number {} and all fiber number {}".\
                format(os.path.basename(infile), np.logical_not(skylineFlat), skyMask.sum(), skyLikeFiberMask.sum(), skyMask.size))
        de.d(fig, os.path.join(debugDir,"skyLikesDebug.fig"), configs=saveConfig)
        fig.clear()
        plt.close(fig)
    if debug>0:#!! plot continuum fitting result (of skylike fiber)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        #fig = goodFigure.GoodFigure(name=name, show=False, ion=False,
                                    #figsize=(10,5), logfile="./log.txt")
        #ax = fig.ax
        for i in np.where(skyLikeFiberMask)[0]:
            ax.plot(waveset[i], flux[i], c=c[i])
            ax.plot(waveset[i], std[i], c=c[i], ls="--")
        #for each in knots:
            #ax.plot([each, each], [0,1])
        #fig.addPlot("simplePlot", "fitting",
                    #waveset_flatten, flux_smooth_flatten,
                    #c="black", lw=3)
        for eachstart, eachend, eachtype, eachdensity in skyRegions:
            if eachtype==1:
                ax.fill([eachstart, eachstart, eachend, eachend],
                    [0, 1, 1, 0], "r", alpha=0.3)
            elif eachtype==2:
                ax.fill([eachstart, eachstart, eachend, eachend],
                    [0, 1, 1, 0], "r", alpha=0.3)
        for i in range(blocks.shape[0]):
            eachstart = blocks[i][0]
            eachend = blocks[i][1]
            eachstartOffset = blocks[i][2]
            eachendOffset = blocks[i][3]
            ax.plot([eachstartOffset, eachstartOffset], [0, 2], "-.", c="green", lw=2)
            ax.plot([eachstart, eachstart], [0, 2], "-", c="green", lw=2)
            ax.plot([eachend, eachend], [0, 2], "--", c="red", lw=3)
            ax.plot([eachendOffset, eachendOffset], [0, 2], "-.", c="red", lw=3)

        ax.set_title(name)
        ax.plot(waveset_flatten, flux_smooth_flatten, c="black", lw=3)
        ax.plot(waveset_flatten, continuum, c="red", lw=2)
        ax.plot(region_wave, region_flux, ".", markersize=15, color="green")
        de.d(fig, os.path.join(debugDir,"continuum.fig"), configs=saveConfig)
        if debug>1:
            plt.show()
        else:
            plt.close(fig)
    if debug>0: #!! continuum with knots
        fig = plt.figure()
        ax = fig.add_subplot(111)
        #fig = goodFigure.GoodFigure(name=name, show=False, ion=False,
                                    #figsize=(10,5), logfile="./log.txt")
        #ax = fig.ax
        for i in np.where(skyLikeFiberMask)[0]:
            ax.plot(waveset[i], flux[i], c=c[i])
            ax.plot(waveset[i], std[i], c=c[i], ls="--")
        for each in knots:
            ax.plot([each, each], [0,1])
        #fig.addPlot("simplePlot", "fitting",
                    #waveset_flatten, flux_smooth_flatten,
                    #c="black", lw=3)
        for eachstart, eachend, eachtype, eachdensity in skyRegions:
            if eachtype==1:
                ax.fill([eachstart, eachstart, eachend, eachend],
                    [0, 1, 1, 0], "r", alpha=0.3)
            elif eachtype==2:
                ax.fill([eachstart, eachstart, eachend, eachend],
                    [0, 1, 1, 0], "r", alpha=0.3)
        for i in range(blocks.shape[0]):
            eachstart = blocks[i][0]
            eachend = blocks[i][1]
            eachstartOffset = blocks[i][2]
            eachendOffset = blocks[i][3]
            ax.plot([eachstartOffset, eachstartOffset], [0, 2], "-.", c="green", lw=2)
            ax.plot([eachstart, eachstart], [0, 2], "-", c="green", lw=2)
            ax.plot([eachend, eachend], [0, 2], "--", c="red", lw=3)
            ax.plot([eachendOffset, eachendOffset], [0, 2], "-.", c="red", lw=3)

        ax.set_title(name)
        ax.plot(waveset_flatten, flux_smooth_flatten, c="black", lw=3)
        ax.plot(waveset_flatten, continuum, c="red", lw=2)
        ax.plot(region_wave, region_flux, ".", markersize=15, color="green")
        de.d(fig, os.path.join(debugDir,"continuum_with_knots.fig"), configs=saveConfig)
        if debug>1:
            plt.show()
        else:
            plt.close(fig)

    debugData = {"colorMap":c, "waveset":waveset, "flux":flux, "std":std, "skyLikeFiberMask":skyLikeFiberMask}
    with open(os.path.join(debugDir, "allDebug.dat"), "wb") as f:
        pickle.dump(debugData, f)
    #!! important data
    flux2fit_smooth = flux_smooth_flatten - continuum
    flux2fit = flux_flatten - continuum
    if debug>0: #!! xcen_upper_lower
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(waveset_flatten, flux2fit_smooth, c="blue", lw=2)
        for i in np.where(skyMask)[0]:
            ax.plot(waveset[i], flux[i], c=c[i])
            ax.plot(waveset[i], std[i], c=c[i], ls="--")
        ax.plot(waveset_flatten, flux_smooth_flatten, c="black", lw=3)
        ax.plot(waveset_flatten, continuum, c="red", lw=2)
        ax.plot([waveset[0], waveset[-1]], [0,0], c="black")
        ax.set_title(name)
        for i in range(blocks.shape[0]):
            eachstart = blocks[i][0]
            eachend = blocks[i][1]
            eachstartOffset = blocks[i][2]
            eachendOffset = blocks[i][3]
            ax.plot([eachstartOffset, eachstartOffset], [0, 2], "-.", c="green", lw=3)
            ax.plot([eachstart, eachstart], [0, 2], "-", c="green", lw=3)
            ax.plot([eachend, eachend], [0, 2], "--", c="red", lw=4)
            ax.plot([eachendOffset, eachendOffset], [0, 2], "-.", c="red", lw=4)

        for eachcen, eachlower, eachupper in zip(skywave, sky_lower, sky_upper):
            ax.plot([eachcen, eachcen], [0, 1], "--", c="black")
            ax.plot([eachlower, eachlower], [0, 1], "--", c="green")
            ax.plot([eachupper, eachupper], [0, 1], "--", c="red")
        de.d(fig, os.path.join(debugDir,"xcen_upper_lower.fig"))
        if debug>1:
            plt.show()
        else:
            plt.close(fig)
    #for eachwave, eachsigma, eachintensity in allSkyLines: #!! plot xcen and 3 sigma
        #ax.plot([eachwave, eachwave], [0, 1], "--", c="black")
        #ax.plot([eachwave-3*eachsigma, eachwave-3*eachsigma], [0, 1], "--", c="green")
        #ax.plot([eachwave+3*eachsigma, eachwave+3*eachsigma], [0, 1], "--", c="red")

    #endregion: debug plots
    #region: debug flat result
    jobName = configs["name"]
    sessionName = jobName
    windowName = os.path.basename(infile)[:-5]
    todoList = []
    todo = "cd {};loadPyFigs skyLikes.fig".format(debugDir)
    todoList.append('tmux new-session -s {sessionName} -d >/dev/null 2>&1'.\
            format(sessionName=sessionName, windowName=windowName))
    todoList.append('tmux kill-window -t {sessionName}:{windowName} >/dev/null 2>&1'.\
            format(sessionName=sessionName, windowName=windowName))
    todoList.append('tmux new-window -t {sessionName} -n {windowName}'.\
            format(sessionName=sessionName, windowName=windowName))
    todoList.append('sleep 1')
    todoList.append('tmux send-keys -t {sessionName}:{windowName} "{todo}" ENTER'.\
            format(sessionName=sessionName, windowName=windowName, todo=todo))
    debugOut = os.path.join(outputPath, "_cmdDebugFlatResult_{}".format(windowName))
    with open(debugOut, "w") as f:
        f.write("cd ..;"+"\n".join(todoList)+"\n")
    os.chmod(debugOut,stat.S_IRWXU)

    if debugFlatResult:
        for each in todoList:
            os.system(each)
        return None

    #endregion: debug flat result
    thisFileName0 = "l_{}_localSmoothSkyResidual.fig".format(names[0])
    thisFileName0 = os.path.join(debugDir, "_localSkyDebug",thisFileName0)
    if debug>0 and (rerun or not os.path.exists(thisFileName0)): #!! localSky single debug
        for i in range(len(names)):
            fig = plt.figure()
            ax = fig.add_subplot(111)
            thisFileName = "l_{}_localSmoothSkyResidual.fig".format(names[i])
            todoStr="\t\t output local sky residual result {:3d}/{:3d} array for {}\r".\
                        format(i, len(names), names[i])
            sys.stdout.write(todoStr)
            sys.stdout.flush()
            smoothSkyResidual          = smoothSky[i] - flux[i]
            localSmoothResidual5pm     = localSmooth5pmArray[i]-flux[i]
            localSmoothResidual5Maxpm  = localSmooth5MaxpmArray[i]-flux[i]
            localSmoothResidual10t     = localSmooth10tArray[i]-flux[i]
            localSmoothResidual10pm    = localSmooth10pmArray[i]-flux[i]
            localSmoothResidual10Maxpm = localSmooth10MaxpmArray[i]-flux[i]
            localSmoothResidual20t     = localSmooth20tArray[i]-flux[i]
            localSmoothResidual15pm    = localSmooth15pmArray[i]-flux[i]
            localSmoothResidual15Maxpm = localSmooth15MaxpmArray[i]-flux[i]
            localSmoothResidual30t     = localSmooth30tArray[i]-flux[i]
            ax.set_title(thisFileName)
            ax.plot([waveset[i][0], waveset[i][-1]], [0,0], c="b", lw=1, ls='--', label="0")
            ax.plot(waveset[i], flux[i], c="g", lw=2, label="RawData")
            ax.plot(waveset[i], smoothSkyResidual, c="k", lw=3, label="Average sky")
            ax.plot(waveset[i], localSmoothResidual5pm, lw=2, c="red", ls="-.", label="5pm")
            ax.plot(waveset[i], localSmoothResidual5Maxpm, lw=2, c="red", ls="--", label="5Maxpm")
            ax.plot(waveset[i], localSmoothResidual10t, lw=2, c="red",ls="-", label="10t")
            ax.plot(waveset[i], localSmoothResidual10pm, lw=2, c="green", ls="-.", label="10pm")
            ax.plot(waveset[i], localSmoothResidual10Maxpm, c="green", lw=2, ls="--", label="10Maxpm")
            ax.plot(waveset[i], localSmoothResidual20t, lw=2, c="green", ls="-.", label="20t")
            ax.plot(waveset[i], localSmoothResidual15pm, lw=2, c="blue", ls="-.", label="15pm")
            ax.plot(waveset[i], localSmoothResidual15Maxpm, lw=2, c="blue", ls="--", label="15Maxpm")
            ax.plot(waveset[i], localSmoothResidual30t, lw=2, c="blue", ls="-", label="30t")
            ax.legend()
            de.d(fig, os.path.join(debugDir, "_localSkyDebug",thisFileName))
            plt.close(fig)
    #region: fitting and model sky
    #region: !! fit stack sky
    #!! first fitting for stack data
    eachyset = None
    thisConfig = {"name":"stack", "refit":refit, "doLimitsDubug":False,
               "debugDir":debugDir, "logger":logger,
               "debug":debug, "fiberCount":0, "fiberTotal":1,
               "pngdebug":True, "pngdebugResidual":False, "fibertype":"stack",
               "replot":replot, "markerSize":2,"kernelSize":31, "time0":time0,
               "FAFactor": 1, "FAFactorEnd":2}#!! magic numbers
    info = fitSky(waveset_flatten, flux2fit, std_flatten, flux2fit_smooth, skyGroup, continuum, eachyset,
                  skywave, sky_lower, sky_upper, skysigma, skysigma3, skytype, skyResidualType,
                            skySigmaLower, skySigmaUpper, skySigma3Lower, skySigma3Upper,
                            skyALower, skyAUpper, skyA3Lower, skyA3Upper, sky_flag,
                  waveset_flatten, flux_smooth_flatten, continuum,  blocks, configs=thisConfig)

    #!! stop for debug
    if cmdDebugSkyStack:
        raise logger.Exception("stop to look at stack.fig")
    #!! update p0 and continuum
    continuumNew = info["continuumNew"]#!! must have last plot....
    m = modelFittingResult = info["modelFittingResult"]
    continuumObj = interp.interp1d(waveset_flatten, continuumNew, fill_value="extrapolate")
    #skywave = m["mu"]
    #skysigma = m["sigma"]
    #skysigma3 = m["sigma3"]
    #endregion: fit stack sky
    #region: !! fitting for individual skys
    for i in range(modelFittingResult.shape[0]):
        thisA  = modelFittingResult["A"][i]
        thisA3 = modelFittingResult["A3"][i]
        thisStatus = modelFittingResult["status"][i]
        thisSigmaError = modelFittingResult["sigmaError"][i]
        thisSigmaErrorUpper = modelFittingResult["sigmaErrorUpper"][i]
        if (thisStatus&0b1111000)>0:
            pass #!! will be constrain in the first fit for skylike
        elif thisStatus>0:
            if (thisA+thisA3)<10*thisSigmaError:#!! ???
                skyAUpper[i] = (thisA+thisA3)+5*thisSigmaError
                skyA3Upper[i] = (thisA+thisA3)+5*thisSigmaError
                skyALower[i] = 0
                skyA3Lower[i] = 0
            else:
                skyAUpper[i] = (thisA+thisA3)*1.3+3*thisSigmaError
                skyA3Upper[i] = (thisA+thisA3)*1.3+3*thisSigmaError
                skyALower[i] = 0
                skyA3Lower[i] = 0
    assert np.all(skyAUpper>skyALower)
    assert np.all(skyA3Upper>skyA3Lower)
    if useSkylike:
        todotype = "SkyLike"
        todoNames = namesSkyLike
        todowaveset = waveset_sky_like
        todoflux = flux_sky_like
        todostd = std_sky_like
        todoIndexs = index_sky_like
        todoFiberIndex = namesSkyLikeIndex
        todoLocalSky = localSkys[skyLikeFiberMask]
    else:
        todotype = "Sky"
        todoNames = namesSky
        todowaveset = waveset_sky
        todoflux = flux_sky
        todostd = std_sky
        todoIndexs = index_sky
        todoFiberIndex = namesSkyIndex
        todoLocalSky = localSkys[skyMask]

    assert len(todoLocalSky)==len(todoIndexs)
    cacheFile = os.path.join(debugDir, "data", "all{}FittingResults.dat".format(todotype))

    if refit or not os.path.exists(cacheFile):
        allSkyFittingTime = np.zeros((len(todoNames),))
        allSkyFAFactor = np.zeros((len(todoNames),))
        allSkyFittingResults = np.empty((len(todoNames), allSkyLines.size), dtype=dtype_m)
        allSkyFittingLineModel = np.zeros((len(todoNames), flux.shape[1]))
        allSkyFittingGoodMask = np.ones((len(todoNames), flux.shape[1]), dtype=np.bool)
        allSkyFittingContinuumModel = np.zeros((len(todoNames), flux.shape[1]))
        allSkyFittingContinuumResidualModel = np.zeros((len(todoNames), flux.shape[1]))
        allSkyFittingObjectContinuumModel = np.zeros((len(todoNames), flux.shape[1]))
        allSkyGaussianFitting = np.zeros((len(todoNames), flux.shape[1]))
        #!! test fitting for all sky fibers
        thisFAFactor = 0.2
        for i,(eachname, eachwave, eachflux, eachstd) in \
            enumerate(zip(todoNames, todowaveset, todoflux, todostd)):
            thisLogger = logger
            #eachname = name+"_"+eachname
            eachFluxFactor = factor[fiberIndex.tolist().index(todoFiberIndex[i])]#!! switch to 128 and..
            eachyset = yset[fiberIndex.tolist().index(todoFiberIndex[i])]#!! switch to 128 and search
            thisConfig = {"name":eachname, "refit":0, "debugDir":debugDir, "time0":time0,
                       "logger":thisLogger, "debug":debug, "markerSize": 30, "lastPlot":lastPlot,
                       "polyConstrain":True, "fiberCount":i, "fiberTotal":len(todoNames),
                       "pngdebug":pngdebug, "isSkylike":2, #!! to direct constrain small flux skylines
                       "pngdebugResidual":pngdebugResidual, "doSecondFitting":True,#!! we need gaussian fitting..
                       "FAFactor": thisFAFactor, "tweakContinuum":True, "checkReject":True,
                       "doResidualSubtract":False, "doResidualPolyConstrain":False,#!! sky not do this
                       "onlyMark":True, "fibertype":todotype, "todoLocalSky":todoLocalSky[i],
                       "bmod":bmod, "map2D":map2D, "fimage":fimage, "fstd":fstd,
                       "goodFiberIndex":goodFiberIndex, "fiberIndex":todoFiberIndex[i],
                       "fluxFactor":eachFluxFactor, "ysetSmooth":ysetSmooth,}
                       #"firstPlot":True, "secondPlot":True}
            eachFlux2fit = eachflux - continuumObj(eachwave)
            eachFlux2fit_smooth = fluxSmoothObj(eachwave) - continuumObj(eachwave)
            info = fitSky(eachwave, eachFlux2fit, eachstd, eachFlux2fit_smooth, skyGroup, continuumObj(eachwave), eachyset,
                      skywave, sky_lower, sky_upper, skysigma, skysigma3, skytype, skyResidualType,
                           skySigmaLower, skySigmaUpper, skySigma3Lower, skySigma3Upper,
                           skyALower, skyAUpper, skyA3Lower, skyA3Upper, sky_flag,
                      waveset_flatten, flux_smooth_flatten, continuumNew,  blocks,
                      configs=thisConfig)
            allSkyFAFactor[i] = thisFAFactor
            allSkyFittingTime[i] = info["fittingTime"]
            medianTime = np.median(allSkyFittingTime[allSkyFittingTime>0])
            goodN = (allSkyFittingTime>0).sum()
            if medianTime>500 and goodN>=1:
                thisFAFactor *= 1.3
            elif medianTime>400 and goodN>1:
                thisFAFactor *= 1.25
            elif medianTime>300 and goodN>1:
                thisFAFactor *= 1.2
            elif medianTime<80 and goodN>=1:
                thisFAFactor *= 0.75
            elif medianTime<100 and goodN>=1:
                thisFAFactor *= 0.80
            elif medianTime<150 and goodN>=1:
                thisFAFactor *= 0.85
            elif medianTime<200 and goodN>=1:
                thisFAFactor *= 0.90
            elif medianTime<250 and goodN>=1:
                thisFAFactor *= 0.95
            if thisFAFactor>0.6:
                thisFAFactor=0.6
            elif thisFAFactor<0.2:
                thisFAFactor=0.2
            allSkyFittingResults[i, :] = info["modelFittingResult"][:]
            allSkyFittingLineModel[i,:] = info["lines"][:]
            allSkyFittingContinuumModel[i,:] = info["continuum"][:]
            allSkyFittingObjectContinuumModel[i,:] = info["objContinuum"][:]
            allSkyGaussianFitting[i,:] = info["gaussianFittingResult"][:]
            allSkyFittingContinuumResidualModel[i,:] = info["continuumResidual"][:]
            allSkyFittingGoodMask[i,:] = info["fittingGoodMask"][:]


        thisResult = {"allSkyFittingResults":allSkyFittingResults,
                   "allSkyFittingLineModel":allSkyFittingLineModel,
                   "allSkyFittingGoodMask":allSkyFittingGoodMask,
                   "allSkyFittingContinuumModel":allSkyFittingContinuumModel,
                   "allSkyFittingObjectContinuumModel":allSkyFittingObjectContinuumModel,
                   "allSkyGaussianFitting":allSkyGaussianFitting,
                   "allSkyFittingContinuumResidualModel":allSkyFittingContinuumResidualModel,
                   "allSkyFAFactor":allSkyFAFactor,
                   "allSkyFittingTime":allSkyFittingTime}
        with open(cacheFile, "wb") as f:
            pickle.dump(thisResult, f)
    else:
        with open(cacheFile, "rb") as f:
            thisResult = pickle.load(f)

    allSkyFAFactor = thisResult["allSkyFAFactor"]
    allSkyFittingTime = thisResult["allSkyFittingTime"]
    allSkyFittingResults        = thisResult["allSkyFittingResults"]
    allSkyFittingLineModel      = thisResult["allSkyFittingLineModel"]
    allSkyFittingGoodMask        = thisResult["allSkyFittingGoodMask"]
    allSkyFittingContinuumModel = thisResult["allSkyFittingContinuumModel"]
    allSkyFittingContinuumResidualModel = thisResult["allSkyFittingContinuumResidualModel"]
    allSkyFittingObjectContinuumModel = thisResult["allSkyFittingObjectContinuumModel"]
    allSkyGaussianFitting = thisResult["allSkyGaussianFitting"]

    #endregion: fitting for individual skys
    #endregion: fit and model sky
    #region: renew resultALimits
    #!! renew resultALimits!
    cacheFile = os.path.join(debugDir, "data", "skyUpperModelUsing{}.dat".format(todotype))
    minLimit = configs.get("minAUpperLimit", 0.0001)
    maxLimit = configs.get("maxAUpperLimit", 0.1)
    if reGenSkyUpperModel or not os.path.exists(cacheFile):
        resultALimits = {}
        ALimitsUpper = np.ones(modelFittingResult.shape[0])*10
        ALimitsUpperAlert = np.ones(modelFittingResult.shape[0])*10
        ALimitsLower = np.ones(modelFittingResult.shape[0])*0
        resultALimits["upper"] = ALimitsUpper
        resultALimits["lower"] = ALimitsLower
        resultALimits["upperAlert"] = ALimitsUpperAlert
        #!! allSkyFittingResults is allSkyFittingResult
        allFittingFlux = allSkyFittingResults[:]["A"] + allSkyFittingResults[:]["A3"]
        allFittingSigmaError = allSkyFittingResults[:]["sigmaError"]
        allFittingStatus = allSkyFittingResults[:]["status"]
        allFittingmu = allSkyFittingResults[:]["mu"]
        allGroup = allSkyLines["group"]
        allSkyLinesMu = allSkyLines["lambda"]
        #!! define limits for each sky
        def plotDebugSky(i, subcall=False):
            N = allSkyFittingResults.shape[0]
            x = np.arange(N)
            fig = plt.figure()
            ax = fig.add_subplot(111)
            fig.set_size_inches(12./26*N,7/26*N)
            ax.grid(True, axis="x")
            A0 = modelFittingResult[i]["A"]+modelFittingResult[i]["A3"]
            thisTotalA = allFittingFlux[:,i]
            thisUpper = ALimitsUpper[i]
            thisLower = ALimitsLower[i]
            thisStatus = allFittingStatus[:,i]
            thisGroupName = allGroup[i]
            thisGroupMask = allGroup==thisGroupName
            thisGroupMus = allSkyLinesMu[thisGroupMask]
            inRangeMask = thisStatus>0
            goodFluxMask = np.bitwise_and(thisStatus, 0b110)==0
            badFluxMask = np.logical_not(goodFluxMask)
            s = thisSigmaError = allFittingSigmaError[:,i ]
            info = astroMath.iterstat1dC(thisTotalA[goodFluxMask&inRangeMask], sigrej=2.5)
            mean = info["mean"]
            std = info["std"]
            if std is None:
                std=mean
            sm = astroMath.iterstat1d(s[inRangeMask&goodFluxMask])["mean"]
            #!! x limits and ticks
            ax.set_xlabel("{} from {} having {}".format(allFittingmu[:,i].mean(), thisGroupName, thisGroupMus))
            saveFileName = "{}_{:7.3f}_{}".format(i, allFittingmu[:,i].mean(), thisGroupName)
            ax.xaxis.tick_top()
            ax.xaxis.set_label_position('top')
            ax.set_xticks(x)
            ax.set_xlim([-1, N])
            ax.set_ylim(bottom=0.0)
            statusFormat = "{:1b}\n{:0>2b}\n{:0>2b}\n{:0>2b}"
            thisSubStatus = [(each&1, (each&0b110)>>1, (each&0b11000)>>3, (each&0b1100000)>>5)
                              for each in thisStatus]
            ax.set_xticklabels([("{}\n{}\n"+statusFormat).format(eachx, eachIndex, *eachStatus)
                        for eachx, eachIndex, eachStatus in zip(x, fiberIndex[todoIndexs], thisSubStatus)])

            ax.plot(x[inRangeMask&goodFluxMask], thisTotalA[inRangeMask&goodFluxMask], "g.-", mec=None, ms=8)
            ax.plot(x[inRangeMask&badFluxMask], thisTotalA[inRangeMask&badFluxMask], "r.-", mec=None, ms=8)
            ax.plot(x[inRangeMask], thisSigmaError[inRangeMask], color="pink", ls="--", lw=2)
            ax.plot([0, N-1], [thisLower, thisLower], "g--")
            ax.plot([0, N-1], [thisUpper, thisUpper], "r--")
            ax.plot([0, N-1], [A0, A0], "k", lw=2)
            ax.plot([0, N-1], [mean, mean], "k--", lw=1)

            smFactor = mean/sm
            if smFactor<3:
                upper = mean+2*std
                upperAlert = upper*3
                lower = 0
            elif smFactor<10:
                upper = mean+2*std
                upperAlert = upper*3
                lower = mean-3*std if mean-3*std>0 else 0
            else:
                upper = np.min([mean+3*sm, mean+2*sm+1*std, mean+2.0*std])
                upperAlert = upper + 20*sm + 1*std
                lower = 2*mean - upper
            allGoodFlux = thisTotalA[goodFluxMask&inRangeMask]
            if allGoodFlux.size:
                if upper>allGoodFlux.max():
                    upper = allGoodFlux.max()
            xx = x[inRangeMask]
            ax.plot(xx, (mean+3*sm )*np.ones(xx.shape), "k--", lw=1)
            ax.plot(xx, (mean+5*sm )*np.ones(xx.shape), "k--", lw=1)
            ax.plot(xx, (mean+5*sm+std )*np.ones(xx.shape), "k--", lw=1)
            ax.plot(xx, (mean+10*sm)*np.ones(xx.shape), "k--", lw=1)
            ax.plot(xx, (upper     )*np.ones(xx.shape), "k.--", lw=1)
            ax.plot(xx, (lower     )*np.ones(xx.shape), "g.--", lw=1)
            ax.set_ylim(top=np.max([upper, mean+10*sm])*1.3)
            #!! will have bug when a cosmic ray hit one of the line in a group and an object line also lay in this group... 
            if thisGroupMask.sum()>1:
                thisAllA0 = (modelFittingResult[thisGroupMask]["A"] + modelFittingResult[thisGroupMask]["A3"]).sum()
                thisAllLower = ALimitsLower[thisGroupMask].sum()
                thisAllUpper = ALimitsUpper[thisGroupMask].sum()
                thisAllTotalA = allFittingFlux[:,thisGroupMask].sum(axis=1)
                thisAllStatus = allFittingStatus[:,thisGroupMask]
                thisAllInrange = ((thisAllStatus==0).sum(axis=1))==0
                thisAnyBadMask = ((np.bitwise_and(thisAllStatus, 0b110)>0).sum(axis=1))>0
                thisAnyGoodMask = np.logical_not(thisAnyBadMask)
                #!! to handle some weird large-sigma-low-SNR lines
                thisAllSNR = (allFittingFlux[:,thisGroupMask]/allFittingSigmaError[:,thisGroupMask])[thisAllInrange&thisAnyGoodMask,:]
                SNRFactor = np.median(thisAllSNR, axis=0)
                thisSNRs = np.median(allFittingSigmaError[:,thisGroupMask][thisAllInrange&thisAnyGoodMask,:], axis=0)
                goodGroupLine = SNRFactor>3
                if not np.all(goodGroupLine):
                    badGroupNumberSigmaSum = np.sqrt(((thisSNRs*SNRFactor)[~goodGroupLine]**2).sum())
                else:
                    badGroupNumberSigmaSum = 0
                #!! cal sigma with out these weird large-sigma-low-SNR lines
                if np.any(goodGroupLine):
                    ss = np.sqrt(\
                            ((allFittingSigmaError[:,thisGroupMask][:,goodGroupLine])**2).sum(axis=1))
                    ssm = np.median(ss[thisAnyGoodMask&thisAllInrange])
                    if badGroupNumberSigmaSum>2*ssm:#!! some tweak
                        badGroupNumberSigmaSum = ssm
                else:#!! use default method
                    ss = np.sqrt(\
                            ((allFittingSigmaError[:,thisGroupMask])**2).sum(axis=1))
                    ssm = np.median(ss[thisAnyGoodMask&thisAllInrange])
                    badGroupNumberSigmaSum = 0
                thisInfo = astroMath.iterstat1dC(thisAllTotalA[thisAllInrange&thisAnyGoodMask], sigrej=2.5)
                thisAllMean = thisInfo["mean"]
                thisAllStd = thisInfo["std"]
                if thisAllStd is None:
                    thisAllStd = thisAllMean
                ax.plot(x[thisAllInrange&thisAnyGoodMask], thisAllTotalA[thisAllInrange&thisAnyGoodMask], "k*-",  mec=None, ms=10)
                ax.plot(x[thisAllInrange&thisAnyBadMask], thisAllTotalA[thisAllInrange&thisAnyBadMask], "r*-", mec=None, ms=10)
                ax.plot([0, N-1], [thisAllA0, thisAllA0], "k", lw=2)
                ax.plot([0, N-1], [thisAllLower, thisAllLower], ".--", color="cyan", lw=2)
                ax.plot([0, N-1], [thisAllUpper, thisAllUpper], ".--", color="pink", lw=2)
                ax.plot([0, N-1], [thisAllMean, thisAllMean], "b--", lw=1)

                ssmFactor = thisAllMean/ssm

                if ssmFactor<3:
                    supper = thisAllMean+2*thisAllStd
                    supperAlert = supper*3
                    slower = 0
                elif ssmFactor<10:
                    supper = thisAllMean+2*thisAllStd
                    supperAlert = supper*3
                    slower = thisAllMean-2*thisAllStd if thisAllMean-2*thisAllStd>0 else 0
                else:
                    #supper = thisAllMean+5*ssm+1*thisAllStd+badGroupNumberSigmaSum
                    supper = np.min([thisAllMean+3*ssm+badGroupNumberSigmaSum,
                                     thisAllMean+2*ssm+1*thisAllStd+badGroupNumberSigmaSum,
                                     thisAllMean+2*thisAllStd+badGroupNumberSigmaSum,
                                     thisAllMean+2.5*thisAllStd
                                    ])
                    supperAlert = supper + 20*ssm+1*thisAllStd
                    slower = 2*thisAllMean - supper if 2*thisAllMean - supper>0 else 0

                allGoodFlux = thisAllTotalA[thisAllInrange&thisAnyGoodMask]
                if allGoodFlux.size:
                    if supper>allGoodFlux.max():
                        supper = allGoodFlux.max()

                xx = x[thisAllInrange]
                ax.plot(xx, (thisAllMean+3*ssm )*np.ones(xx.shape), "b--", lw=1)
                ax.plot(xx, (thisAllMean+5*ssm )*np.ones(xx.shape), "b--", lw=1)
                ax.plot(xx, (thisAllMean+5*ssm+thisAllStd )*np.ones(xx.shape), "b--", lw=1)
                ax.plot(xx, (thisAllMean+10*ssm)*np.ones(xx.shape), "b--", lw=1)
                ax.plot(xx, supper*np.ones(xx.shape), "b.--", lw=1)
                ax.plot(xx, slower*np.ones(xx.shape), "g.--", lw=1)
                ax.set_xlabel("{} from {} having {} badSigmaSum:{:5.4f}".\
                        format(allFittingmu[:,i].mean(), thisGroupName, thisGroupMus, badGroupNumberSigmaSum))
                upper = supper/thisAllMean*mean
                upperAlert = supperAlert/thisAllMean*mean
                lower = 0 #!! can not decide lower for group lines
                ax.set_ylim(top=np.max([supper, thisAllMean+10*ssm])*1.3)
            fig.tight_layout()
            if upper<minLimit:
                upper = minLimit
                upperAlert = maxLimit
            return fig, saveFileName, upper, upperAlert, lower
        #for i in [0,10,40,60,80,100,150,200]:
            #plotDebugSky(i)
        allSkyDebugDir = os.path.join(debugDir, "allSkyDebug")
        extract.safe_mkdir(allSkyDebugDir)
        timet = time.time(); timett = time.time()
        for i in range(allFittingStatus.shape[1]):
            fig,filename,thisupper,thisupperAlert,thislower = plotDebugSky(i)
            ALimitsUpper[i] = thisupper
            ALimitsLower[i] = thislower
            ALimitsUpperAlert[i] = thisupperAlert
            sys.stdout.write("\toutput "+filename+"  time= {:.2f}s {:.2f}s {}/{}    \t\t".\
                    format(time.time()-timet, time.time()-timett,i, allFittingStatus.shape[1])+"\r");
            sys.stdout.flush()
            timett = time.time()
            #de.d(fig, os.path.join(allSkyDebugDir, filename+".fig"))
            fig.savefig(os.path.join(allSkyDebugDir, filename+".png"), dpi=100, format="png", bbox_inches="tight")
            plt.close(fig)
        with open(cacheFile, "wb") as f:
            pickle.dump(resultALimits, f)
    else:
        with open(cacheFile, "rb") as f:
            resultALimits = pickle.load(f)
    #endregion: renew resultALimits
    #region: fit object
    todotype = "Obj"
    todoNames = namesObj
    todowaveset = waveset_obj
    todoflux = flux_obj
    todostd = std_obj
    todoFiberIndex = namesObjIndex
    todoLocalSky = localSkys[objMask]
    cacheFile = os.path.join(debugDir, "data", "all{}FittingResults.dat".format(todotype))
    if refit or not os.path.exists(cacheFile):
        allObjFittingTime = np.zeros((len(todoNames),))
        allObjFAFactor = np.zeros((len(todoNames),))
        allObjFittingResults = np.empty((len(todoNames), allSkyLines.size), dtype=dtype_m)
        allObjFittingLineModel = np.zeros((len(todoNames), flux.shape[1]))
        allObjFittingGoodMask = np.ones((len(todoNames), flux.shape[1]), dtype=np.bool)
        allObjFittingContinuumModel = np.zeros((len(todoNames), flux.shape[1]))
        allObjFittingContinuumResidualModel = np.zeros((len(todoNames), flux.shape[1]))
        allObjFittingObjectContinuumModel = np.zeros((len(todoNames), flux.shape[1]))
        allObjGaussianFitting = np.zeros((len(todoNames), flux.shape[1]))
        #!! test fitting for all sky fibers
        thisFAFactor=0.2
        for i,(eachname, eachwave, eachflux, eachstd) in \
            enumerate(zip(todoNames, todowaveset, todoflux, todostd)):
            thisLogger = logger
            #eachname = name+"_"+eachname
            eachFluxFactor = factor[fiberIndex.tolist().index(todoFiberIndex[i])]#!! switch to 128 and..
            eachyset = yset[fiberIndex.tolist().index(todoFiberIndex[i])]#!! switch to 128 and search
            #!! doResidualPolyConstrain:#!! (name is not right) obj and sky be true... leave it here
            thisConfig = {"name":eachname, "refit":0, "debugDir":debugDir, "time0":time0,
                    "logger":thisLogger, "debug":debug, "markerSize": 30,"lastPlot":lastPlot,
                       "polyConstrain":False, "fiberCount":i, "fiberTotal":len(todoNames),
                       "modelFittingResultOld": modelFittingResult, "pngdebug":pngdebug,
                       "pngdebugResidual":pngdebugResidual, "doSecondFitting":True,
                       "FAFactor": thisFAFactor, "tweakContinuum":True, "checkReject":True,
                       "doResidualSubtract":True, "resultALimits":copy.deepcopy(resultALimits),
                       "doResidualPolyConstrain":False, "fibertype":todotype,#!! obj also not do residual poly constrain
                       "bmod":bmod, "map2D":map2D, "fimage":fimage, "fstd":fstd,
                       "goodFiberIndex":goodFiberIndex, "fiberIndex":todoFiberIndex[i],
                       "fluxFactor":eachFluxFactor, "ysetSmooth":ysetSmooth, "todoLocalSky":todoLocalSky[i],
                       "fittingTweak":0.8, "maxRecorrect":1, "isSkylike":(todoFiberIndex[i] in namesSkyLikeIndex),
                       "polyThirdConstrain":True}
            eachFlux2fit = eachflux - continuumObj(eachwave)
            eachFlux2fit_smooth = fluxSmoothObj(eachwave) - continuumObj(eachwave)
            info = fitSky(eachwave, eachFlux2fit, eachstd, eachFlux2fit_smooth, skyGroup, continuumObj(eachwave), eachyset,
                          skywave.copy(), sky_lower, sky_upper, skysigma, skysigma3, skytype, skyResidualType,
                               skySigmaLower, skySigmaUpper, skySigma3Lower, skySigma3Upper,
                               skyALower, skyAUpper, skyA3Lower, skyA3Upper, sky_flag,
                          waveset_flatten, flux_smooth_flatten, continuumNew,  blocks,
                          configs=thisConfig)

            allObjFAFactor[i] = thisFAFactor
            allObjFittingTime[i] = info["fittingTime"]
            medianTime = np.median(allObjFittingTime[allObjFittingTime>0])
            goodN = (allObjFittingTime>0).sum()
            if medianTime>1800 and goodN>=1:
                thisFAFactor *= 1.3
            elif medianTime>1500 and goodN>=1:
                thisFAFactor *= 1.25
            elif medianTime>1200 and goodN>=1:
                thisFAFactor *= 1.2
            elif medianTime>1000 and goodN>=1:
                thisFAFactor *= 1.15
            elif medianTime>900 and goodN>=1:
                thisFAFactor *= 1.1
            elif medianTime>800 and goodN>=1:
                thisFAFactor *= 1.003
            elif medianTime>700 and goodN>=1:
                thisFAFactor *= 1.002
            elif medianTime>600 and goodN>=1:
                thisFAFactor *= 1.001
            elif medianTime<100 and goodN>=1:
                thisFAFactor *= 0.6
            elif medianTime<200 and goodN>=1:
                thisFAFactor *= 0.7
            elif medianTime<300 and goodN>=1:
                thisFAFactor *= 0.8
            elif medianTime<400 and goodN>=1:
                thisFAFactor *= 0.9
            elif medianTime<500 and goodN>=1:
                thisFAFactor *= 0.92
            if thisFAFactor>0.6:
                thisFAFactor=0.6
            elif thisFAFactor<0.2:
                thisFAFactor=0.2
            allObjFittingResults[i, :] = info["modelFittingResult"][:]
            allObjFittingLineModel[i,:] = info["lines"][:]
            allObjFittingContinuumModel[i,:] = info["continuum"][:]
            allObjFittingContinuumResidualModel[i,:] = info["continuumResidual"][:]
            allObjFittingObjectContinuumModel[i,:] = info["objContinuum"][:]
            allObjGaussianFitting[i,:] = info["gaussianFittingResult"][:]
            allObjFittingGoodMask[i,:] = info["fittingGoodMask"][:]

        thisResult = {"allObjFittingResults":allObjFittingResults,
                   "allObjFittingLineModel":allObjFittingLineModel,
                   "allObjFittingContinuumModel":allObjFittingContinuumModel,
                   "allObjFittingObjectContinuumModel":allObjFittingObjectContinuumModel,
                   "allObjFittingGoodMask":allObjFittingGoodMask,
                   "allObjGaussianFitting":allObjGaussianFitting,
                   "allObjFittingContinuumResidualModel":allObjFittingContinuumResidualModel,
                   "allObjFAFactor":allObjFAFactor,
                   "allObjFittingTime":allObjFittingTime}
        with open(cacheFile, "wb") as f:
            pickle.dump(thisResult, f)
    else:
        with open(cacheFile, "rb") as f:
            thisResult = pickle.load(f)
    allObjFittingResults        = thisResult["allObjFittingResults"]
    allObjFittingLineModel      = thisResult["allObjFittingLineModel"]
    allObjFittingGoodMask      = thisResult["allObjFittingGoodMask"]
    allObjFittingContinuumModel = thisResult["allObjFittingContinuumModel"]
    allObjFittingContinuumResidualModel = thisResult["allObjFittingContinuumResidualModel"]
    allObjFittingObjectContinuumModel = thisResult["allObjFittingObjectContinuumModel"]
    allObjGaussianFitting = thisResult["allObjGaussianFitting"]
    #endregion: fit object!!
    #region: record final result
    finalSkyLineModel = np.zeros(flux.shape)
    finalContinuumModel = np.zeros(flux.shape)
    finalFittingGoodMask = np.ones(flux.shape, dtype=np.bool)
    finalContinuumResidualModel = np.zeros(flux.shape)
    finalObjectContinuumModel = np.zeros(flux.shape)
    finalSkyLinesResult = np.zeros(flux.shape)
    finalSkyContinuumResult = np.zeros(flux.shape)
    allFittingResults = np.empty((len(names), allSkyLines.size), dtype=dtype_m)
    allGaussianFitting = np.zeros(yset.shape)
    if not np.all(allSkyFittingGoodMask):
        bugfiberIndex = np.where(~np.all(allSkyFittingGoodMask, axis=1))[0]
        #!! will have bug when use sky instead like
        bugfiberName = [namesSkyLike[i] for i in bugfiberIndex]
        logger.warn("allSkyFittingGoodMask not all good in {}, bug fibers {}".\
                format(name, bugfiberName))
    if not np.all(allObjFittingGoodMask):
        bugfiberIndex = np.where(~np.all(allObjFittingGoodMask, axis=1))[0]
        #!! will have bug when use sky instead like
        bugfiberName = [namesObj[i] for i in bugfiberIndex]
        logger.warn("allObjFittingGoodMask not all good in {}, bug fibers {}".\
                format(name, bugfiberName))
    if useSkylike:
        allFittingResults[skyLikeFiberMask, :] = allSkyFittingResults[:,:]
        allGaussianFitting[skyLikeFiberMask, :] = allSkyGaussianFitting[:,:]
        finalSkyLineModel[skyLikeFiberMask, :] = allSkyFittingLineModel[:,:]
        finalFittingGoodMask[skyLikeFiberMask, :] = allSkyFittingGoodMask[:,:]
        finalContinuumModel[skyLikeFiberMask, :] = allSkyFittingContinuumModel[:,:]
        finalContinuumResidualModel[skyLikeFiberMask, :] = allSkyFittingContinuumResidualModel[:,:]
        finalObjectContinuumModel[skyLikeFiberMask, :] = allSkyFittingObjectContinuumModel[:,:]
    else:
        allFittingResults[skyMask, :] = allSkyFittingResults[:,:]
        allGaussianFitting[skyMask, :] = allSkyGaussianFitting[:,:]
        finalSkyLineModel[skyMask, :] = allSkyFittingLineModel[:,:]
        finalFittingGoodMask[skyMask, :] = allSkyFittingGoodMask[:,:]
        finalContinuumModel[skyMask, :] = allSkyFittingContinuumModel[:,:]
        finalContinuumResidualModel[skyMask, :] = allSkyFittingContinuumResidualModel[:,:]
        finalObjectContinuumModel[skyMask, :] = allSkyFittingObjectContinuumModel[:,:]
    allFittingResults[objMask, :] = allObjFittingResults[:,:]
    allGaussianFitting[objMask, :] = allObjGaussianFitting[:,:]
    finalSkyLineModel[objMask, :] = allObjFittingLineModel[:,:]
    finalFittingGoodMask[objMask, :] = allObjFittingGoodMask[:,:]
    finalContinuumModel[objMask, :] = allObjFittingContinuumModel[:,:]
    finalContinuumResidualModel[objMask, :] = allObjFittingContinuumResidualModel[:,:]
    finalObjectContinuumModel[objMask, :] = allObjFittingObjectContinuumModel[:,:]
    localSkynames = ["ls5pm","ls5Maxpm","ls10t","ls10pm","ls10Maxpm","ls20t",
                         "ls15pm","ls15Maxpm","ls30t"]

    namePairs = [("names", "names"), ("yset", "yset"), ("waveset", "waveset"),
                 ("flux", "flux"), ("invvar", "invvar"), ("factor", "fluxFactors"),
                 ("smoothSky", "smoothSky"), ("ysetSmooth","ysetSmooth"),
                 ("blocks", "blocks"), ("blocks", "blocks"),
                 ("allFittingResults", "fittingResults"),
                 ("finalGoodFiber", "finalGoodMask"),
                 ("allGaussianFitting", "gaussianFitting"),
                 ("fiberIndex", "fiberIndex"),
                 ("finalContinuumResidualModel", "continuumResidualModel"),
                 ("allSkyLines", "allSkyLines"),
                 ("finalSkyLineModel", "skyLineModel"),
                 ("finalContinuumModel", "continuumModel"),
                 ("finalObjectContinuumModel", "objectContinuumModel"),
                 ("finalFittingGoodMask","finalFittingGoodMask"),
                 ("localSkyArrays","localSkyArrays"),
                 ("localSkynames", "localSkynames"),
                 ("debugDir", "debugDir")]

    result = macro.genResults(namePairs, locals())
    #endregion: record final result
    return result

#region: notes for fitSky
"""!! flux and flux_smooth here has been subtracted the fake continuum(flux_smooth)
flux = fluxOrigin - comtinuumObj(wave)
flux_smooth = fluxSmoothObj(wave) - continuumObj(wave)
"""
#endregion: notes for fitSky
#!!        this fiber data
def fitSky(wave, flux, std, flux_smooth, skyGroup, flux_continuum, yset,#!! have the save length
           skywave, sky_lower, sky_upper, skysigma, skysigma3, skytype, skyResidualType,
                    skySigmaLower, skySigmaUpper, skySigma3Lower, skySigma3Upper,
                    skyALower, skyAUpper, skyA3Lower, skyA3Upper, skyFlag,
           #!! smooth sky data, just for debug plot
           waveset_flatten, flux_smooth_flatten, continuum, blocks, configs={}):
    #region: setups
    #region: basic
    globalDebugFits = configs.get("globalDebugFits",False)
    debugDir = configs["debugDir"]
    time0 = configs["time0"]
    time00 = time.time()
    doAlert = configs.get("doAlert", False)
    fibertype = configs.get("fibertype", "???")
    name = configs["name"]
    refit = configs.get("refit", 0)
    doSecondFitting = configs.get("doSecondFitting", False)
    cacheFileFinal = os.path.join(debugDir, "data", fibertype+"_"+name+"_allFitResult.dat")
    replot = configs.get("replot", False)
    firstPlot = configs.get("firstPlot", False)
    secondPlot = configs.get("secondPlot", False)
    lastPlot = configs.get("lastPlot", True)
    isSkylike = configs.get("isSkylike",False)#!! constrain small flux directly
    #!! extra sigma difference before alert
    alertFactor = configs.get("alertFactor", 10)
    if os.path.exists(cacheFileFinal) and not replot and not refit:
        with open(cacheFileFinal, "rb") as f:
            result = pickle.load(f)
        return result

    logger = configs["logger"]
    debug = configs["debug"]
    maxRecorrect = configs.get("maxRecorrect", 0)
    polyThirdConstrain = configs.get("polyThirdConstrain", False)
    doLimitsDubug = configs.get("doLimitsDubug", True)
    #endregion: basic
    #region: about convolve residual
    kernelSize = configs.get("kernelSize", 5)
    ks = int((kernelSize-1)/2)
    kernelSigma = configs.get("kernelSigma", 2)
    kernel = gaussian(np.arange(-ks, ks+1), 0, kernelSigma)
    goodmask = np.isfinite(std)
    xx = np.arange(goodmask.size)
    originResidual = flux-flux_smooth;
    obj = interp.interp1d(xx[goodmask], originResidual[goodmask], fill_value="extrapolate")
    originResidual = obj(xx)
    originResidualConvolved = np.convolve(originResidual, kernel, mode="same")
    doResidualSubtract = configs.get("doResidualSubtract", False)
    if isSkylike:
        doResidualSubtract = False
    doResidualPolyConstrain = configs.get("doResidualPolyConstrain", False)
    #endregion: about convolve residual
    #region: output figure config
    figsize = configs.get("figsize", (15,9))
    pngdebug = configs.get("pngdebug", False)
    pngdebugResidual = configs.get("pngdebugResidual", False)
    pngNumber = configs.get("pngNumber", 40)
    pngOverlap = configs.get("pngOverlap", 0.3)
    pngPlotYMax = configs.get("pngPlotYMax", 0.2)
    #endregion: output figure config
    #region: constrains in fitting
    smallFluxSigmaReject = configs.get("smallFluxSigmaReject", 5)
    polyConstrain = configs.get("polyConstrain", True)
    checkReject = configs.get("checkReject", False)
    markerSize = configs.get("markerSize", 5)
    boundaryProtectionPoints = configs.get("boundaryProtectionPoints", 5)
    defaultSigma = configs.get("defaultSigma", 1.6)
    FAFactor = configs.get("FAFactor", None)#!! relative fitting accuracy
    FAFactorLength = configs.get("FAFactorLength", 10)
    FAFactorEnd = configs.get("FAFactorEnd", 100)
    assert FAFactor<=FAFactorEnd
    #endregion: constrains in fitting
    #region: constrains in result
    resultALimits = configs.get("resultALimits",None)
    tweakContinuum = configs.get("tweakContinuum",1)#!! 1 means use median std, 2 means use smooth residual
    tweakContinuumFactor = configs.get("tweakContinuumFactor", 3)
    tweakResidualContinuumFactor = configs.get("tweakResidualContinuumFactor", 2)
    #endregion: constrains in result
    #region: for outputs
    fiberCount = configs.get("fiberCount", "?")
    fiberTotal = configs.get("fiberTotal", "?")
    modelFittingResultOld = configs.get("modelFittingResultOld", None)
    #endregion: for outputs
    #region: this para is to handle profiles at the right or left edge of each fiber
    profileProtect = configs.get("profileProtect", 2)
    modelFittingResult = np.empty(skywave.shape, dtype=dtype_m)
    modelFittingResultUpper = np.empty(skywave.shape, dtype=dtype_m)
    modelFittingResultLower = np.empty(skywave.shape, dtype=dtype_m)
    for i in range(skywave.size):
        #! mu, A, sigma, A3, sigma3, status, sigmaError, sigmaErrorUpper, residualSigmaError, gaussianMu, gaussianA, guassianSigma
        modelFittingResult[i] = (skywave[i], 0, 1, 0, 1, 0, -1, -1, -1, skywave[i], 0, 1)
        modelFittingResultUpper[i] = (skywave[i], 0, 1, 0, 1, 0, -1, -1, -1, skywave[i], 0, 1)
        modelFittingResultLower[i] = (skywave[i], 0, 1, 0, 1, 0, -1, -1, -1, skywave[i], 0, 1)
    #endregion: todo
    #region: about continuum tweak
    #!! result = result_lines + result_poly
    #!! if tweak flux, result_poly is filter by a upper and lower limit (goodContinumFactor)
    flux_fit_result = flux_smooth.copy()
    flux_fit_gaussian_result = np.zeros(flux_fit_result.shape)
    dtype_popt = np.dtype([("nSky", np.int), ("popt", np.object)])
    modelFittingPopts = np.empty(blocks.shape, dtype=dtype_popt)
    flux_fit_result_lines = np.zeros(wave.shape)
    flux_fit_result_poly  = flux_smooth.copy()
    flux_fit_goodMask = np.ones(wave.shape, dtype=bool)
    dtype_p = np.dtype([("start", np.float64),
                        ("end", np.float64),
                        ("startTrue", np.float64),
                        ("endTrue", np.float64),
                        ("coeffs", np.object)])
    modelPolyResult = np.empty(blocks.shape, dtype=dtype_p)
    modelPolyResultLower = np.empty(blocks.shape, dtype=dtype_p)
    modelPolyResultUpper = np.empty(blocks.shape, dtype=dtype_p)
    goodmask = np.isfinite(std)
    stdObj = interp.interp1d(wave[goodmask], std[goodmask], fill_value="extrapolate")
    #endregion: about continuum tweak
    #region:  aobut flux tweak
    if (resultALimits is not None) and (modelFittingResultOld is not None):
        tweakFlux = True
    else:
        tweakFlux = False
    boundaryCount = configs.get("boundaryCount", 30)
    hardBoundaryCount = configs.get("hardBoundaryCount", 10)
    onlyMark = configs.get("onlyMark", False)
    #!! if a serises of lines excess or under the average flux, then fit again
    maxEmissionWidth = configs.get("maxEmissionWidth", 3)
    maxRefit = 2
    #endregion: about flux tweak
    #region: about limits debug
    pass
    #endregion: about limits debug
    #region: some init values
    resultMask = np.zeros(wave.shape,dtype=np.int64)
    if modelFittingResultOld is None:
        skyA = np.ones(skywave.shape)
        skyA3 = np.ones(skywave.shape)
        skyA[skyA>skyAUpper] = skyAUpper[skyA>skyAUpper]*0.99
        skyA3[skyA3>skyA3Upper] = skyA3Upper[skyA3>skyA3Upper]*0.99
    else:
        skyA = modelFittingResultOld["A"].copy()
        skyA3 = modelFittingResultOld["A3"].copy()
    #!! each fiber tweak its own A and A3 limits
    skyAUpper = skyAUpper.copy()
    skyALower = skyALower.copy()
    skyA3Upper = skyA3Upper.copy()
    skyA3Lower = skyA3Lower.copy()
    #endregion: some init values
    #endregion: setups
    #!! fitting for the model for the first time
    blockSize = blocks.size
    alertRegions = []#!! to generate seperate fig for searching
    time000 = time.time()
    configList = ["profileProtect", "defaultSigma", "doSecondFitting",
                  "FAFactor", "FAFactorLength", "FAFactorEnd", "checkReject",
                  "tweakContinuum", "tweakContinuumFactor", "polyConstrain",
                  "name", "fiberCount", "fiberTotal", "fibertype",
                  "debugDir", "refit", "hardBoundaryCount", "doResidualSubtract",
                  "boundaryCount", "tweakFlux", "doResidualPolyConstrain",
                  "tweakResidualContinuumFactor", "isSkylike"]
    l=locals(); thisConfig = {each:l[each] for each in configList}
    assert np.all(np.isfinite(skyAUpper))
    #!! FIRST FIT
    #!!     update status for the small flux and residual constrain skylines(just mark them)
    #!!         constrain small flux sky lines directly
    #!!     mark full, partial reject and cosmic lines
    skyfitOnce("first", logger, time0, time00, time000,
        #!! from upper
        wave, flux, std, flux_smooth, skyGroup,
        waveset_flatten, flux_smooth_flatten, continuum, blocks,
        skywave, sky_lower, sky_upper, skysigma, skysigma3, skytype, skyResidualType,
                 skySigmaLower, skySigmaUpper, skySigma3Lower, skySigma3Upper,
                 skyALower, skyAUpper, skyA3Lower, skyA3Upper, skyFlag,
                 skyA, skyA3,
        #!! gen in fitSky
        stdObj, goodmask,
        #!! variable to modify
        modelFittingResult, modelFittingPopts, modelPolyResult,
        modelFittingResultUpper, modelFittingResultLower,
        modelPolyResultUpper, modelPolyResultLower,
        resultMask, alertRegions,
        flux_fit_result, flux_fit_gaussian_result,
        flux_fit_result_lines, flux_fit_result_poly,
        flux_fit_goodMask,
        thisConfig)

    result = {}
    if doSecondFitting:#!! do second and third fitting
        todoLocalSky = configs["todoLocalSky"]
        flux_fit_resultOld = flux_smooth.copy()
        flux_fit_result_linesOld = np.zeros(wave.shape)
        modelFittingPoptsOld = np.empty(blocks.shape, dtype=dtype_popt)
        #region: plot for the first time
        thisTitle = fibertype + ":" + name + " first fit " + ["", " with continuum tweak"][tweakContinuum] +\
                        ["", " with flux tweak"][tweakFlux]
        continuumNew = flux_smooth_flatten.copy()#!! full length
        profile0 = specLib.profileGen(0);
        if firstPlot:#!! plot fitting result
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_title(thisTitle)
            #region: highlightings
            mask = np.bitwise_and(resultMask, 0b11000)==0b01000#!! small flux line selected
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,2,2,0], "yellow", alpha=0.2)
            mask = np.bitwise_and(resultMask, 0b11000)==0b11000#!! small flux line used
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,2,2,0], "cyan", alpha=0.2)
            mask = np.bitwise_and(resultMask, 0b1100000)==0b0100000#!! not use to constrain continuum
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,3,3,0], "blue", alpha=0.1)
            mask = np.bitwise_and(resultMask, 0b1100000)==0b1100000#!! used to constrain continuum
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,3,3,0], "green", alpha=0.2)
            if doSecondFitting:#!! mark cosmic, fullreject, partial reject
                mask = np.logical_not(np.isfinite(std))
                se = getPlotStartEnd(mask)#!! cosmic
                for s, e in se:
                    ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "black", alpha=0.8)
                mask = np.bitwise_and(resultMask, 0b100)==0b100#!! full reject
                se = getPlotStartEnd(mask)
                for s, e in se:
                    ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "r", alpha=0.3)
                mask = np.bitwise_and(resultMask, 0b010)==0b010#!! partial reject
                se = getPlotStartEnd(mask)
                for s, e in se:
                    ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "pink", alpha=0.3)

                mask = np.bitwise_and(resultMask, 0b100000000)==0b100000000#!! too high (may be a line!!)
                se = getPlotStartEnd(mask)
                for s, e in se:
                    ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,4,4,0], "blue", alpha=0.4)
                mask = np.bitwise_and(resultMask, 0b010000000)==0b010000000#!! too low (?????)
                se = getPlotStartEnd(mask)
                for s, e in se:
                    ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,4,4,0], "darkorchid", alpha=0.1)
            #endregion: highlightings
            #region: vertical lines
            #!! init mu
            for eachcen, eachlower, eachupper in zip(skywave, sky_lower, sky_upper):
                ax.plot([eachcen, eachcen], [0, 1], "--", c="black")
                ax.plot([eachlower, eachlower], [0, 1], "--", c="green")
                ax.plot([eachupper, eachupper], [0, 1], "--", c="red")
            #!! fitting blocks
            for i in range(blocks.shape[0]):
                eachstart = blocks[i][0]
                eachend = blocks[i][1]
                eachstartOffset = blocks[i][2]
                eachendOffset = blocks[i][3]
                eestart = blocks[i][0]
                eeend = blocks[i][1]
                ax.plot([eachstartOffset, eachstartOffset], [0, 2], "-.", c="green", lw=3)
                ax.plot([eachstart, eachstart], [0, 2], "-", c="green", lw=3)
                ax.plot([eachend, eachend], [0, 2], "--", c="red", lw=4)
                ax.plot([eachendOffset, eachendOffset], [0, 2], "-.", c="red", lw=4)
            #endregion: vertical lines
            #region: mu fitting and profile (green profile)
            for i in range(modelFittingResult.shape[0]):
                eachmu = modelFittingResult["mu"][i]
                eachA = modelFittingResult["A"][i]
                eachsigma = modelFittingResult["sigma"][i]
                eachA3 = modelFittingResult["A3"][i]
                eachsigma3 = modelFittingResult["sigma3"][i]
                x,y = specLib.profile23(waveset_flatten, eachmu, eachA, eachsigma, eachA3, eachsigma3)
                if not y.size:
                    continue
                ymax = y.max()
                if not np.isfinite(ymax) or np.abs(eachmu)>1e5:#!! useless?
                    if modelFittingResultOld is not None:
                        muOrigin = modelFittingResultOld[i]["mu"]
                    else:
                        muOrigin = "?"
                    logger.warn("bat fitting at {}:{} mu:{} ymax:{} muOrigin:{}".\
                            format(fibertype, name, eachmu, ymax, muOrigin))
                ax.plot([eachmu, eachmu], [0, ymax], "b--")
                ax.plot(x,y, "g--")
            #endregion: mu fitting and profile (green profile)
            #region: all profile for one block (blue profile)
            for i, (eachstart, eachend, eachstartTrue, eachendTrue, eachcoeffs) in enumerate(modelPolyResult):
                xmask = (waveset_flatten>eachstart)&(waveset_flatten<eachend)
                xmaskTrue = xmask.copy()
                xmaskTrue[(waveset_flatten<eachstartTrue)|(waveset_flatten>eachendTrue)] = 0
                x = waveset_flatten[xmask]
                xmaskmask = (x>eachstartTrue)&(x<eachendTrue)
                obj = Legendre(eachcoeffs, domain=[eachstart, eachend])
                y = obj(x)
                continuumNew[xmaskTrue] = (continuum[xmaskTrue] + y[xmaskmask])[:]
                ax.plot(x,y, "r--", lw=2)
                nSky, popt = modelFittingPopts[i]
                if popt is not None:
                    popt = popt[:5*nSky]
                    yy = profile0(x, *popt)
                    ax.plot(x,yy, "b--", lw=2)
            #endregion: all profile for one block (blue profile)
            #region: continuum
            ax.plot(waveset_flatten, -continuum, "k")
            ax.plot(waveset_flatten, -continuumNew, "b")
            ax.plot(wave, std, color="pink")
            ax.plot(wave, flux_smooth, "g-", lw=2)
            ax.scatter(wave, flux, edgecolors=None, s=markerSize)
            ax.plot(wave, flux_fit_result, "r-", lw=2)
            #badmask = np.logical_not(np.isfinite(std))
            #toPlot = flux_fit_result-flux; toPlot[badmask]=0
            #ax.plot(wave, flux_smooth-flux, "g-", lw=2)
            #ax.scatter(wave, toPlot, edgecolors=None, s=markerSize)
            #ax.plot(wave, toPlot, "r-", lw=2)
            #endregion: continuum
            #region: output format
            #!! change xlim before dump
            xstart = waveset_flatten.min()
            xend   = waveset_flatten.max()
            ax.plot([xstart, xend], [0,0], color="black", lw=0.5)
            eachLength = (xend-xstart)/pngNumber
            ax.set_xlim([xstart, xstart + eachLength])
            fig.set_size_inches(12,3)
            de.d(fig, os.path.join(debugDir,"{}_{}_first_fit.fig".format(fibertype, name, eachstart, eachend)))
            #endregion: about png outputs
            if debug>1:
                plt.show()
            else:
                plt.close(fig)
        #endregion: plot for the first time
        if not isSkylike:
            #!! select good points and set flag 0b11000 or 0b01000(3sigma width)
            #!!     update A upper
            for i in range(modelFittingResult.shape[0]):
                eachType = skytype[i]
                eachstatus = modelFittingResult["status"][i]
                if eachType and eachstatus>0:
                    eachmu = modelFittingResult["mu"][i]
                    eachA = modelFittingResult["A"][i]
                    eachA3 = modelFittingResult["A3"][i]
                    eachsigma = modelFittingResult["sigma"][i]
                    eachsigma3 = modelFittingResult["sigma3"][i]
                    eachSigmaErrorUpper = modelFittingResult["sigmaErrorUpper"][i]
                    thisLimit = modelFittingResult["sigmaError"][i]/2
                    assert (np.isfinite(thisLimit) and thisLimit>0)
                    eachATotal = eachA+eachA3
                    eachAFactor = eachA/eachATotal
                    eachA3Factor = eachA3/eachATotal
                    eachFlagUpper = eachmu + 2*eachsigma*eachAFactor + 2*eachsigma3*eachA3Factor
                    eachFlagLower = eachmu - 2*eachsigma*eachAFactor - 2*eachsigma3*eachA3Factor
                    if eachATotal<eachSigmaErrorUpper:#!! no object lines here
                        modelFittingResult[i]["status"] = modelFittingResult[i]["status"]|0b11000
                        thisIndex = np.argmin(np.abs(eachmu-wave))
                        inrange = (wave>eachFlagLower)&(wave<eachFlagUpper)
                        resultMask[inrange] = modelFittingResult[i]["status"]
                        skyA[i] = thisLimit*0.99
                        skyA3[i] = thisLimit*0.99
                        skyAUpper[i] = thisLimit
                        skyA3Upper[i] = thisLimit
                    else:#!! maybe here is some obj lines!!
                        modelFittingResult[i]["status"] = modelFittingResult[i]["status"]|0b01000
                        inrange = (wave>eachFlagLower)&(wave<eachFlagUpper)
                        resultMask[inrange] = modelFittingResult[i]["status"]
            #!! do second fitting
            time000 = time.time()
            configList = ["profileProtect", "defaultSigma", "doSecondFitting",
                          "FAFactor", "FAFactorLength", "FAFactorEnd", "checkReject",
                          "tweakContinuum", "tweakContinuumFactor", "polyConstrain",
                          "name", "fiberCount", "fiberTotal", "fibertype",
                          "debugDir", "refit", "hardBoundaryCount", "doResidualSubtract",
                          "boundaryCount", "tweakFlux", "doResidualPolyConstrain",
                          "tweakResidualContinuumFactor",]
            l=locals(); thisConfig = {each:l[each] for each in configList}
            #!! SECOND FIT
            #!!     nothing spectial, use the selected small flux lines
            skyfitOnce("second", logger, time0, time00, time000,
                #!! from upper
                wave, flux, std, flux_smooth, skyGroup,
                waveset_flatten, flux_smooth_flatten, continuum, blocks,
                skywave, sky_lower, sky_upper, skysigma, skysigma3, skytype, skyResidualType,
                         skySigmaLower, skySigmaUpper, skySigma3Lower, skySigma3Upper,
                         skyALower, skyAUpper, skyA3Lower, skyA3Upper, skyFlag,
                         skyA, skyA3,
                #!! gen in fitSky
                stdObj, goodmask,
                #!! variable to modify
                modelFittingResult, modelFittingPopts, modelPolyResult,
                modelFittingResultUpper, modelFittingResultLower,
                modelPolyResultUpper, modelPolyResultLower,
                resultMask, alertRegions,
                flux_fit_result, flux_fit_gaussian_result,
                flux_fit_result_lines, flux_fit_result_poly,
                flux_fit_goodMask,
                thisConfig)

        #!! select good points and set flag 0b1100000 or 0b0100000(3sigma width)
        #!!     use these points to get object continuum
        xIndex = []#!! record which pixel to use to fit a residual continuum
        for i in range(modelFittingResult.shape[0]):
            eachResidualType = skyResidualType[i]
            eachstatus = modelFittingResult["status"][i]
            if eachResidualType and eachstatus>0:
                eachmu = modelFittingResult["mu"][i]
                eachA = modelFittingResult["A"][i]
                eachA3 = modelFittingResult["A3"][i]
                eachsigma = modelFittingResult["sigma"][i]
                eachsigma3 = modelFittingResult["sigma3"][i]
                eachResidualSigmaError = modelFittingResult["residualSigmaError"][i]
                eachATotal = eachA+eachA3
                eachAFactor = eachA/eachATotal
                eachA3Factor = eachA3/eachATotal
                eachFlagUpper = eachmu + 3*eachsigma*eachAFactor + 3*eachsigma3*eachA3Factor
                eachFlagLower = eachmu - 3*eachsigma*eachAFactor - 3*eachsigma3*eachA3Factor
                if eachATotal<eachResidualSigmaError:#!! use residual of this line to constrain continum
                    modelFittingResult[i]["status"] = modelFittingResult[i]["status"]|0b1100000
                    thisIndex = np.argmin(np.abs(eachmu-wave))
                    xIndex.append(thisIndex)
                    inrange = (wave>eachFlagLower)&(wave<eachFlagUpper)
                    resultMask[inrange] = modelFittingResult[i]["status"]
                else:#!! do not use residual of this line to constrain continum
                        modelFittingResult[i]["status"] = modelFittingResult[i]["status"]|0b0100000
                        inrange = (wave>eachFlagLower)&(wave<eachFlagUpper)
                        resultMask[inrange] = modelFittingResult[i]["status"]
        #!! constrain continuum using the second result
        xx = np.array(xIndex);xx.sort()
        residualObj = interp.interp1d(wave[xx], originResidualConvolved[xx],
                              kind="slinear", fill_value=np.nan, bounds_error=False)
        residualX = wave[xx]
        residualY = originResidualConvolved[xx]
        residualInterpolated = residualObj(wave)
        flux_fit_result_obj_continuum = residualInterpolated.copy()
        #region: plot for the second fitting
        thisTitle = fibertype + " " + name + " second fit " + ["", " with continuum tweak"][tweakContinuum] +\
                        ["", " with flux tweak"][tweakFlux]
        continuumNew = flux_smooth_flatten.copy()#!! full length
        profile0 = specLib.profileGen(0);
        if secondPlot:#!! plot fitting result
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.set_title(thisTitle)
            #region: highlightings
            mask = np.bitwise_and(resultMask, 0b11000)==0b01000#!! small flux line selected
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,2,2,0], "yellow", alpha=0.2)
            mask = np.bitwise_and(resultMask, 0b11000)==0b11000#!! small flux line used
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,2,2,0], "cyan", alpha=0.2)
            mask = np.bitwise_and(resultMask, 0b1100000)==0b0100000#!! not use to constrain continuum
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,3,3,0], "blue", alpha=0.1)
            mask = np.bitwise_and(resultMask, 0b1100000)==0b1100000#!! used to constrain continuum
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,3,3,0], "green", alpha=0.2)
            if doSecondFitting:
                mask = np.logical_not(np.isfinite(std))
                se = getPlotStartEnd(mask)#!! cosmic
                for s, e in se:
                    ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "black", alpha=0.8)
                mask = np.bitwise_and(resultMask, 0b100)==0b100#!! full reject
                se = getPlotStartEnd(mask)
                for s, e in se:
                    ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "r", alpha=0.3)
                mask = np.bitwise_and(resultMask, 0b010)==0b010#!! partial reject
                se = getPlotStartEnd(mask)
                for s, e in se:
                    ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "pink", alpha=0.3)

                mask = np.bitwise_and(resultMask, 0b100000000)==0b100000000#!! too high (may be a line!!)
                se = getPlotStartEnd(mask)
                for s, e in se:
                    ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,4,4,0], "blue", alpha=0.4)
                mask = np.bitwise_and(resultMask, 0b010000000)==0b010000000#!! too low (?????)
                se = getPlotStartEnd(mask)
                for s, e in se:
                    ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,4,4,0], "darkorchid", alpha=0.1)
            #endregion: highlightings
            #region: vertical lines
            #!! init mu
            for eachcen, eachlower, eachupper in zip(skywave, sky_lower, sky_upper):
                ax.plot([eachcen, eachcen], [0, 1], "--", c="black")
                ax.plot([eachlower, eachlower], [0, 1], "--", c="green")
                ax.plot([eachupper, eachupper], [0, 1], "--", c="red")
            #!! fitting blocks
            for i in range(blocks.shape[0]):
                eachstart = blocks[i][0]
                eachend = blocks[i][1]
                eachstartOffset = blocks[i][2]
                eachendOffset = blocks[i][3]
                eestart = blocks[i][0]
                eeend = blocks[i][1]
                ax.plot([eachstartOffset, eachstartOffset], [0, 2], "-.", c="green", lw=3)
                ax.plot([eachstart, eachstart], [0, 2], "-", c="green", lw=3)
                ax.plot([eachend, eachend], [0, 2], "--", c="red", lw=4)
                ax.plot([eachendOffset, eachendOffset], [0, 2], "-.", c="red", lw=4)
            #endregion: vertical lines
            #region: mu fitting and profile (green profile)
            for i in range(modelFittingResult.shape[0]):
                eachmu = modelFittingResult["mu"][i]
                eachA = modelFittingResult["A"][i]
                eachsigma = modelFittingResult["sigma"][i]
                eachA3 = modelFittingResult["A3"][i]
                eachsigma3 = modelFittingResult["sigma3"][i]
                x,y = specLib.profile23(waveset_flatten, eachmu, eachA, eachsigma, eachA3, eachsigma3)
                if not y.size:
                    continue
                ymax = y.max()
                if not np.isfinite(ymax) or np.abs(eachmu)>1e5:#!! useless?
                    if modelFittingResultOld is not None:
                        muOrigin = modelFittingResultOld[i]["mu"]
                    else:
                        muOrigin = "?"
                    logger.warn("bat fitting at {}:{} mu:{} ymax:{} muOrigin:{}".\
                            format(fibertype, name, eachmu, ymax, muOrigin))
                ax.plot([eachmu, eachmu], [0, ymax], "b--")
                ax.plot(x,y, "g--")
            #endregion: mu fitting and profile (green profile)
            #region: all profile for one block (blue profile)
            for i, (eachstart, eachend, eachstartTrue, eachendTrue, eachcoeffs) in enumerate(modelPolyResult):
                xmask = (waveset_flatten>eachstart)&(waveset_flatten<eachend)
                xmaskTrue = xmask.copy()
                xmaskTrue[(waveset_flatten<eachstartTrue)|(waveset_flatten>eachendTrue)] = 0
                x = waveset_flatten[xmask]
                xmaskmask = (x>eachstartTrue)&(x<eachendTrue)
                obj = Legendre(eachcoeffs, domain=[eachstart, eachend])
                y = obj(x)
                continuumNew[xmaskTrue] = (continuum[xmaskTrue] + y[xmaskmask])[:]
                ax.plot(x,y, "r--", lw=2)
                nSky, popt = modelFittingPopts[i]
                if popt is not None:
                    popt = popt[:5*nSky]
                    yy = profile0(x, *popt)
                    ax.plot(x,yy, "b--", lw=2)
            #endregion: all profile for one block (blue profile)
            #region: continuum
            ax.plot(waveset_flatten, -continuum, "k")
            ax.plot(waveset_flatten, -continuumNew, "b")
            ax.plot(wave, std, color="pink")
            ax.plot(wave, flux_smooth, "g-", lw=2)
            ax.scatter(wave, flux, edgecolors=None, s=markerSize)
            ax.plot(wave, flux_fit_result, "r-", lw=2)

            ax.plot(wave, originResidualConvolved, color="blue", lw=2)
            ax.plot(wave, residualInterpolated, color="black", lw=3)
            ax.scatter(residualX, residualY, color="black", s=50)
            ax.plot([residualX[0], residualX[0]], [0,3], "k--", lw=4)
            ax.plot([residualX[-1], residualX[-1]], [0,3], "k--", lw=4)
            #badmask = np.logical_not(np.isfinite(std))
            #toPlot = flux_fit_result-flux; toPlot[badmask]=0
            #ax.plot(wave, flux_smooth-flux, "g-", lw=2)
            #ax.scatter(wave, toPlot, edgecolors=None, s=markerSize)
            #ax.plot(wave, toPlot, "r-", lw=2)
            #endregion: continuum
            #region: output format
            #!! change xlim before dump
            xstart = waveset_flatten.min()
            xend   = waveset_flatten.max()
            ax.plot([xstart, xend], [0,0], color="black", lw=0.5)
            eachLength = (xend-xstart)/pngNumber
            ax.set_xlim([xstart, xstart + eachLength])
            fig.set_size_inches(12,3)
            de.d(fig, os.path.join(debugDir,"{}_{}_second_fit.fig".format(fibertype, name, eachstart, eachend)))
            #endregion: about png outputs
            if debug>1:
                plt.show()
            else:
                plt.close(fig)
        #endregion: plot for the second fitting
        #!! do third fitting
        alertRegions = []#!! to generate seperate fig for searching
        time000 = time.time()
        configList = ["profileProtect", "defaultSigma", "doSecondFitting", "maxEmissionWidth",
                      "FAFactor", "FAFactorLength", "FAFactorEnd", "checkReject",
                      "tweakContinuum", "tweakContinuumFactor", "polyConstrain",
                      "name", "fiberCount", "fiberTotal", "fibertype", "onlyMark", "doAlert",
                      "debugDir", "refit", "hardBoundaryCount", "doResidualSubtract",
                      "boundaryCount", "residualObj", "tweakFlux", "doResidualPolyConstrain",
                      "resultALimits", "modelFittingPoptsOld", "modelFittingResultOld",
                      "tweakResidualContinuumFactor", "flux_fit_result_linesOld","polyThirdConstrain",
                      "flux_fit_resultOld", "flux_fit_result_obj_continuum", "maxRefit", "maxRecorrect"]
        l=locals(); thisConfig = {each:l[each] for each in configList}
        #!! THIRD FIT
        #!!     clean hard boundary sky lines (within 10 pixel)
        #!!     free polyConstrain[0] and leave others
        #!!     fitting gaussian use 0.9~1.1 poly factor of the profile
        #!!         if not good fitting, leave the gpopt as nan and goon(so we should check the warning file)
        #!!     use resultALimits to tweak the final flux (as groups)
        #!!         if not onlyMark, tweak popt to be upper limit (and also mark them)
        #!!     check series lower or upper, change poly boundarys and fit again
        #!!     record regions that have object continuum subtract
        #!!     flux subtract residual_continuum(obj continuum before fitting) before fitting
        polyDebugInfo = skyfitOnce("third", logger, time0, time00, time000,
            #!! from upper
            wave, flux, std, flux_smooth, skyGroup,
            waveset_flatten, flux_smooth_flatten, continuum, blocks,
            skywave, sky_lower, sky_upper, skysigma, skysigma3, skytype, skyResidualType,
                     skySigmaLower, skySigmaUpper, skySigma3Lower, skySigma3Upper,
                     skyALower, skyAUpper, skyA3Lower, skyA3Upper, skyFlag,
                     skyA, skyA3,
            #!! gen in fitSky
            stdObj, goodmask,
            #!! variable to modify
            modelFittingResult, modelFittingPopts, modelPolyResult,
            modelFittingResultUpper, modelFittingResultLower,
            modelPolyResultUpper, modelPolyResultLower,
            resultMask, alertRegions,
            flux_fit_result, flux_fit_gaussian_result,
            flux_fit_result_lines, flux_fit_result_poly,
            flux_fit_goodMask,
            thisConfig)

        #!! get continuumResidual
        continuumResidual = np.zeros(flux.shape)
        for i, (eachstart, eachend, eachstartTrue, eachendTrue, eachcoeffs) in enumerate(modelPolyResult):
            xmask = (wave>eachstart)&(wave<eachend)
            xmaskTrue = xmask.copy()
            xmaskTrue[(wave<eachstartTrue)|(wave>eachendTrue)] = 0
            x = wave[xmask]
            xmaskmask = (x>eachstartTrue)&(x<eachendTrue)
            obj = Legendre(eachcoeffs, domain=[eachstart, eachend])
            y = obj(x)
            continuumResidual[xmaskTrue] = y[xmaskmask]
        result["continuumResidual"] = continuumResidual
        #!! plot debug fits
        bmod = configs["bmod"]
        map2D = configs["map2D"]
        fimage = configs["fimage"].copy()
        fiberIndex = configs["fiberIndex"]#!! in range of 128
        goodFiberIndex = configs["goodFiberIndex"]
        bs = flux_continuum
        ysetSmooth = configs["ysetSmooth"]
        fiberInd = goodFiberIndex.tolist().index(fiberIndex)
        thisFactor = configs["fluxFactor"]
        thisFlux = np.zeros(ysetSmooth.shape)
        thisIvar = np.ones(ysetSmooth.shape)
        thisOnes = np.ones(ysetSmooth.shape)
        thisZeros = np.zeros(ysetSmooth.shape)
        bzeros = np.zeros(fimage.shape)
        ysetmin = int(ysetSmooth[fiberInd].min())-7
        ysetmax = int(ysetSmooth[fiberInd].max())+7
        ywidth = ysetmax-ysetmin
        #print("({},{}), fiberInd:{} ".format(ysetmin, ysetmax, fiberIndex))
        ind = slice(ysetmin, ysetmax+1)
        ivar = 1/std**2
        thisOnes[fiberInd, :] = 0
        extractMask = extract.map2DModelFixNaNWidth(bzeros, map2D, thisZeros, thisOnes, ysetSmooth)
        badMask = np.isfinite(extractMask)
        thisIvar[fiberInd, :] = ivar[:]
        thisFlux[fiberInd, :] = (flux+bs)[:]/thisFactor
        fimageModelExtract = extract.map2DModel(bmod, map2D, thisFlux, thisIvar)
        thisFlux[fiberInd, :] = (flux_fit_result_lines+flux_fit_result_poly+bs)[:]/thisFactor

        fimageModel = extract.map2DModel(bmod, map2D, thisFlux, thisIvar)
        thisFlux[fiberInd, :] = (flux_smooth+bs)[:]/thisFactor
        fimageModelSmoothSky = extract.map2DModel(bmod, map2D, thisFlux, thisIvar)
        thisFlux[fiberInd, :] = (continuumResidual)[:]/thisFactor
        fimageModelContinuumResidual = extract.map2DModel(bmod, map2D, thisFlux, thisIvar)
        fimageExtractResidual = fimage-fimageModelExtract
        fimageResidual = fimage-fimageModel
        fimageSmoothSkyResidual = fimage-fimageModelSmoothSky
        fimage[badMask]=np.nan
        fimageModel[badMask]=np.nan
        fimageExtractResidual[badMask]=np.nan
        fimageResidual[badMask]=np.nan
        fimageSmoothSkyResidual[badMask]=np.nan
        fimageModelContinuumResidual[badMask] = np.nan
        localSkynames = ["ls5pm","ls5Maxpm","ls10t","ls10pm","ls10Maxpm","ls20t",
                         "ls15pm","ls15Maxpm","ls30t"]
        localskyResidualImages = []
        for eachlsname in localSkynames:
            thisFlux = np.zeros(ysetSmooth.shape)
            thisFlux[fiberInd, :] = todoLocalSky[eachlsname][:]/thisFactor
            thisLocalSkyImage = extract.map2DModel(bmod, map2D, thisFlux, thisIvar)
            thisLocalSkyResidualImage = fimage-thisLocalSkyImage
            localskyResidualImages.append(thisLocalSkyResidualImage)
        #fimage[badMask]=(np.random.rand(badMask.sum())-0.5)[:]
        #fimageModel[badMask]=(np.random.rand(badMask.sum())-0.5)[:]
        #fimageExtractResidual[badMask]=(np.random.rand(badMask.sum())-0.5)[:]
        #fimageResidual[badMask]=(np.random.rand(badMask.sum())-0.5)[:]
        #fimageSmoothSkyResidual[badMask]=(np.random.rand(badMask.sum())-0.5)[:]
        nx = fimage.shape[1]
        xset1DLength = wave.size
        xstart = wave.min()
        xend = wave.max()
        def thisHeader(name=None):
            w = wcs.WCS(naxis=2)
            w.wcs.crpix = [0,0];
            w.wcs.cdelt = [(xend-xstart)/xset1DLength,1];
            w.wcs.crval = [xstart,0];
            w.wcs.ctype = ["x--LINEAR", "y--LINEAR"]
            if name is not None:
                result = w.to_header(relax=True)
                result["OBJECT"] = name
                return result
            else:
                return w.to_header(relax=True)
        h1 = thisHeader("fimage of "+name)
        h2 = thisHeader("extract residual of "+name)
        h3 = thisHeader("residual of "+name)
        h4 = thisHeader("smooth sky residual of "+name)
        hls = [thisHeader("{} residual {}".format(name, todoname)) for todoname in localSkynames]
        assert len(hls)==len(localskyResidualImages)
        dls = [pyfits.ImageHDU(thisimage[ind], header=thisheader)
                  for thisimage,thisheader in zip(localskyResidualImages, hls)]
        h55 = thisHeader()
        h55["OBJECT"] = "continuum residual of "+name
        d1 = pyfits.PrimaryHDU(fimage[ind], header=h1)
        d2 = pyfits.ImageHDU(fimageExtractResidual[ind], header=h2)
        d3 = pyfits.ImageHDU(fimageResidual[ind], header=h3)
        d4 = pyfits.ImageHDU(fimageSmoothSkyResidual[ind], header=h4)
        d55 = pyfits.ImageHDU(fimageModelContinuumResidual[ind], header=h55)
        fits = pyfits.HDUList([d1, d2, d3, d4]+dls+[d55])
        fits.writeto(os.path.join(debugDir,"{}_{}_block_fit.fits".format(fibertype, name)), clobber='True')
        #!! global debug file
        if globalDebugFits:
            allDebugFile = os.path.join(debugDir, "{}_all_block_fit.fits".format(fibertype))
            allDebugReg = os.path.join(debugDir, "{}_all_block_fit.reg".format(fibertype))
            if not os.path.exists(allDebugFile):
                h1 = thisHeader("fimage of {} to {}".format(fibertype, name))
                h2 = thisHeader("extract residual of {} to {}".format(fibertype, name))
                h3 = thisHeader("residual of {} to {}".format(fibertype, name))
                h4 = thisHeader("smooth sky residual of {} to {}".format(fibertype, name))
                h55 = thisHeader("continuum residual of {} to {}".format(fibertype, name))
                hls = [thisHeader("{} local sky residual {}".\
                        format(name, todoname)) for todoname in localSkynames]
                d1 = np.empty((0, nx), dtype=np.float64)
                d2 = np.empty((0, nx), dtype=np.float64)
                d3 = np.empty((0, nx), dtype=np.float64)
                d4 = np.empty((0, nx), dtype=np.float64)
                dls = [np.empty((0, nx),dtype=np.float64) for todoname in localSkynames]
                d55 = np.empty((0, nx), dtype=np.float64)
                with open(allDebugReg, "w") as f:
                    f.write(regStrH.format(color="black"))
                yInd = ywidth/2
            else:
                obj = pyfits.open(allDebugFile)
                h1, h2, h3, h4, h55 = obj[0].header, obj[1].header, obj[2].header, obj[3].header, obj[-1].header
                d1, d2, d3, d4, d55 = obj[0].data, obj[1].data, obj[2].data, obj[3].data, obj[-1].data
                hls = [obj[i].header for i in range(4, len(localSkynames)+4)]
                dls = [obj[i].data for i in range(4, len(localSkynames)+4)]
                assert d1.size==d2.size==d3.size==d4.size==d55.size
                yInd = ywidth/2 + d1.shape[0]
            ysetSmoothCut = ysetSmooth[fiberInd]-ysetSmooth[fiberInd].min()+yInd-ywidth/4
            #!! goon
            colorMod = int(d1.shape[0]/(ysetmax-ysetmin))%2
            dd1 = np.concatenate((d1, fimage[ind]))
            dd2 = np.concatenate((d2, fimageExtractResidual[ind]))
            dd3 = np.concatenate((d3, fimageResidual[ind]))
            dd4 = np.concatenate((d4, fimageSmoothSkyResidual[ind]))
            ddls = [np.concatenate((eachd, eachimage[ind]))
                      for eachd,eachimage in zip(dls, localskyResidualImages)]
            dd55 = np.concatenate((d55, fimageModelContinuumResidual[ind]))
            d1 = pyfits.PrimaryHDU(dd1, header=h1)
            d2 = pyfits.ImageHDU(dd2, header=h2)
            d3 = pyfits.ImageHDU(dd3, header=h3)
            d4 = pyfits.ImageHDU(dd4, header=h4)
            dls = [pyfits.ImageHDU(thisimage, header=thisheader)
                      for thisimage,thisheader in zip(ddls, hls)]
            d55 = pyfits.ImageHDU(dd55, header=h55)
            fits = pyfits.HDUList([d1, d2, d3, d4]+dls+[d55])
            fits.writeto(allDebugFile, clobber='True')

            with open(allDebugReg, "a") as f:
                color = ["green", "blue"][colorMod]
                f.write(regStrT.format(-10,yInd, "{}_{}".format(fibertype, name), color))
                ws = wave.size
                for i in range(modelFittingResult.shape[0]):
                    eachstatus = modelFittingResult[i]["status"]
                    if eachstatus<1:
                        continue
                    eachmu = modelFittingResult[i]["mu"]
                    index = wave.searchsorted(eachmu)
                    if index!=0 and index!=ws:
                        #!! index - 1 + 1
                        thisx = index + (eachmu-wave[index-2])/(wave[index]-wave[index-1])
                        if eachstatus&0b110000000==0:
                            f.write(regStrX(thisx, ysetSmoothCut[index], 1, color))
                        elif eachstatus&0b010000000>0:
                            f.write(regStrX(thisx, ysetSmoothCut[index], 0.5, color))
                        elif eachstatus&0b100000000>0:
                            f.write(regStrX(thisx, ysetSmoothCut[index], 2, color))
                for i in range(blocks.shape[0]):
                    eachstart = blocks[i][0]
                    eachend = blocks[i][1]
                    eachstartOffset = blocks[i][2]
                    eachendOffset = blocks[i][3]

                    ls = 0.5
                    index = wave.searchsorted(eachstart)
                    if index!=0 and index!=ws:
                        f.write(regStrE.format(index, ysetSmoothCut[index]+1, 2*ls, 1*ls, color))
                    index = wave.searchsorted(eachend)
                    if index!=0 and index!=ws:
                        f.write(regStrE.format(index, ysetSmoothCut[index]+1, 1*ls, 2*ls, color))
                    index = wave.searchsorted(eachstartOffset)
                    if index!=0 and index!=ws:
                        f.write(regStrB.format(index, ysetSmoothCut[index]+1, 2*ls, 1*ls, color))
                    index = wave.searchsorted(eachendOffset)
                    if index!=0 and index!=ws:
                        f.write(regStrB.format(index, ysetSmoothCut[index]+1, 1*ls, 2*ls, color))

    #region: output plot
    #!! if not doSecondFit, only have cyan and blue for two type of lines
    continuumNew = flux_smooth_flatten.copy()#!! full length
    profile0 = specLib.profileGen(0)

    thisTitle = fibertype + " " + name+" residual"+["", " with continuum tweak"][tweakContinuum] +\
            ["", " with flux tweak"][tweakFlux] + ["  notSkylike", "  isSkylike", "  isSkylike in sky fitting"][isSkylike] +\
            "fa:{:.2f}".format(FAFactor)
    if lastPlot:#!! residual (all default is true)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_title(thisTitle)
        #region: highlightings
        mask = np.bitwise_and(resultMask, 0b11000)==0b01000#!! small flux line selected
        se = getPlotStartEnd(mask)
        for s, e in se:
            ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,2,2,0], "yellow", alpha=0.2)
        mask = np.bitwise_and(resultMask, 0b11000)==0b11000#!! small flux line used
        se = getPlotStartEnd(mask)
        for s, e in se:
            ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,2,2,0], "cyan", alpha=0.2)
        mask = np.bitwise_and(resultMask, 0b1100000)==0b0100000#!! not use to constrain continuum
        se = getPlotStartEnd(mask)
        for s, e in se:
            ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,3,3,0], "blue", alpha=0.1)
        mask = np.bitwise_and(resultMask, 0b1100000)==0b1100000#!! used to constrain continuum
        se = getPlotStartEnd(mask)
        for s, e in se:
            ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,3,3,0], "green", alpha=0.2)
        if doSecondFitting:
            mask = np.logical_not(np.isfinite(std))
            se = getPlotStartEnd(mask)#!! cosmic
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "black", alpha=0.8)
            mask = np.bitwise_and(resultMask, 0b100)==0b100#!! full reject
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "r", alpha=0.3)
            mask = np.bitwise_and(resultMask, 0b010)==0b010#!! partial reject
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "pink", alpha=0.3)

            mask = np.bitwise_and(resultMask, 0b100000000)==0b100000000#!! too high (may be a line!!)
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,4,4,0], "blue", alpha=0.4)
            mask = np.bitwise_and(resultMask, 0b010000000)==0b010000000#!! too low (?????)
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,4,4,0], "darkorchid", alpha=0.1)

            mask = np.bitwise_and(resultMask, 0b1000000000)==0b1000000000#!! do residual subtraction
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,4,4,0], "green", alpha=0.05)
        #endregion: highlightings
        #region: vertical lines
        #!! init mu
        for eachcen, eachlower, eachupper in zip(skywave, sky_lower, sky_upper):
            ax.plot([eachcen, eachcen], [0, 1], "--", c="black")
            ax.plot([eachlower, eachlower], [0, 1], "--", c="green")
            ax.plot([eachupper, eachupper], [0, 1], "--", c="red")
        #!! fitting blocks
        for i in range(blocks.shape[0]):
            eachstart = blocks[i][0]
            eachend = blocks[i][1]
            eachstartOffset = blocks[i][2]
            eachendOffset = blocks[i][3]
            eestart = blocks[i][0]
            eeend = blocks[i][1]
            ax.plot([eachstartOffset, eachstartOffset], [0, 2], "-.", c="green", lw=3)
            ax.plot([eachstart, eachstart], [0, 2], "-", c="green", lw=3)
            ax.plot([eachend, eachend], [0, 2], "--", c="red", lw=4)
            ax.plot([eachendOffset, eachendOffset], [0, 2], "-.", c="red", lw=4)
        #endregion: vertical lines
        #region: debug limit profile
        if doLimitsDubug:
            #!! upper limit
            for i in range(modelFittingResultUpper.shape[0]):
                eachmu = modelFittingResult["mu"][i]
                eachA = modelFittingResultUpper["A"][i]
                eachsigma = modelFittingResult["sigma"][i]
                eachA3 = modelFittingResultUpper["A3"][i]
                eachsigma3 = modelFittingResult["sigma3"][i]
                x,y = specLib.profile23(waveset_flatten, eachmu, eachA, eachsigma, eachA3, eachsigma3)
                if not y.size:
                    continue
                ymax = y.max()
                ax.plot([eachmu, eachmu], [0, ymax], "r-", lw=1)
            for i, (eachstart, eachend, eachstartTrue, eachendTrue, eachcoeffs) in enumerate(modelPolyResultLower):
                xmask = (waveset_flatten>eachstart)&(waveset_flatten<eachend)
                xmaskTrue = xmask.copy()
                xmaskTrue[(waveset_flatten<eachstartTrue)|(waveset_flatten>eachendTrue)] = 0
                x = waveset_flatten[xmask]
                xmaskmask = (x>eachstartTrue)&(x<eachendTrue)
                obj = Legendre(eachcoeffs, domain=[eachstart, eachend])
                y = obj(x)
                ax.plot(x,y, "k-.", lw=2)
            for i, (eachstart, eachend, eachstartTrue, eachendTrue, eachcoeffs) in enumerate(modelPolyResultUpper):
                xmask = (waveset_flatten>eachstart)&(waveset_flatten<eachend)
                xmaskTrue = xmask.copy()
                xmaskTrue[(waveset_flatten<eachstartTrue)|(waveset_flatten>eachendTrue)] = 0
                x = waveset_flatten[xmask]
                xmaskmask = (x>eachstartTrue)&(x<eachendTrue)
                obj = Legendre(eachcoeffs, domain=[eachstart, eachend])
                y = obj(x)
                ax.plot(x,y, "k-.", lw=2)
        #endregion: debug limit profile
        #region: mu fitting and profile (green profile)
        for i in range(modelFittingResult.shape[0]):
            eachmu = modelFittingResult["mu"][i]
            eachA = modelFittingResult["A"][i]
            eachsigma = modelFittingResult["sigma"][i]
            eachA3 = modelFittingResult["A3"][i]
            eachsigma3 = modelFittingResult["sigma3"][i]
            x,y = specLib.profile23(waveset_flatten, eachmu, eachA, eachsigma, eachA3, eachsigma3)
            if not y.size:
                continue
            ymax = y.max()
            if not np.isfinite(ymax) or np.abs(eachmu)>1e5:#!! useless?
                if modelFittingResultOld is not None:
                    muOrigin = modelFittingResultOld[i]["mu"]
                else:
                    muOrigin = "?"
                logger.warn("bat fitting at {}:{} mu:{} ymax:{} muOrigin:{}".\
                        format(fibertype, name, eachmu, ymax, muOrigin))
            ax.plot([eachmu, eachmu], [0, ymax], "b--")
            ax.plot(x,y, "g--")
        #endregion: mu fitting and profile (green profile)
        #region: debug limit profile
        if doLimitsDubug:
            #!! lower limit
            for i in range(modelFittingResultLower.shape[0]):
                eachmu = modelFittingResult["mu"][i]
                eachA = modelFittingResultLower["A"][i]
                eachsigma = modelFittingResult["sigma"][i]
                eachA3 = modelFittingResultLower["A3"][i]
                eachsigma3 = modelFittingResult["sigma3"][i]
                x,y = specLib.profile23(waveset_flatten, eachmu, eachA, eachsigma, eachA3, eachsigma3)
                if not y.size:
                    continue
                ymax = y.max()
                if ymax>0:
                    ax.plot([eachmu, eachmu], [0, ymax], "g-", lw=3)
        #endregion: debug limit profile
        #region: all profile for one block (blue profile) and continuum !!
        if tweakFlux:
            for i in range(modelPolyResult.shape[0]):
                nSky, popt = modelFittingPoptsOld[i]
                if popt is not None:
                    popt = popt[:5*nSky]
                    yy = profile0(x, *popt)
                    ax.plot(x,yy, "k--", lw=2)
        for i, (eachstart, eachend, eachstartTrue, eachendTrue, eachcoeffs) in enumerate(modelPolyResult):
            xmask = (waveset_flatten>eachstart)&(waveset_flatten<eachend)
            xmaskTrue = xmask.copy()
            xmaskTrue[(waveset_flatten<eachstartTrue)|(waveset_flatten>eachendTrue)] = 0
            x = waveset_flatten[xmask]
            xmaskmask = (x>eachstartTrue)&(x<eachendTrue)
            obj = Legendre(eachcoeffs, domain=[eachstart, eachend])
            y = obj(x)
            continuumNew[xmaskTrue] = (continuum[xmaskTrue] + y[xmaskmask])[:]
            ax.plot(x,y, "r--", lw=2)
            nSky, popt = modelFittingPopts[i]
            if popt is not None:
                popt = popt[:5*nSky]
                yy = profile0(x, *popt)
                ax.plot(x,yy, "b--", lw=2)
        #endregion: all profile for one block (blue profile) and continuum !!
        #region: continuum
        ax.plot(waveset_flatten, -continuum, "k")
        ax.plot(waveset_flatten, -continuumNew, "b")
        badmask = np.logical_not(np.isfinite(std))

        if tweakFlux:
            toPlot = flux-flux_fit_result_linesOld; toPlot[badmask]=0
            ax.plot(wave, toPlot, "c--", lw=2)
            ax.scatter(wave, toPlot, edgecolors=None, s=markerSize/2, color="c")

        toPlot = flux-flux_fit_result; toPlot[badmask]=0
        ax.scatter(wave, toPlot, edgecolors=None, s=markerSize)
        ax.plot(wave, toPlot, "r-", lw=2)

        toPlot = flux-flux_smooth; toPlot[badmask]=0
        ax.plot(wave, toPlot, "g-", lw=2)

        toPlot = flux-flux_fit_result_lines; toPlot[badmask]=0
        ax.plot(wave, toPlot, "r--", lw=1)
        ax.scatter(wave, toPlot, edgecolors=None, s=markerSize/2, color="r")
        ax.plot(wave, std, color="pink")
        if doSecondFitting and doResidualSubtract:
            ax.plot(wave, originResidualConvolved, color="blue", lw=2)
            ax.plot(wave, residualInterpolated, color="black", lw=3)
            ax.scatter(residualX, residualY, color="black", s=50)
            ax.plot([residualX[0], residualX[0]], [3,3], "k--", lw=4)
            ax.plot([residualX[-1], residualX[-1]], [3,3], "k--", lw=4)
        #endregion: continuum
        #region: about png outputs
        #!! change xlim before dump
        xstart = waveset_flatten.min()
        xend   = waveset_flatten.max()
        ax.plot([xstart, xend], [0,0], color="black", lw=0.5)
        eachLength = (xend-xstart)/pngNumber
        ax.set_xlim([xstart, xstart + eachLength])
        fig.set_size_inches(12,3)
        figobj = {"figure":fig, "x":wave, "class":DebugFigure}
        de.d(figobj, os.path.join(debugDir,"{}_{}_block_fit_residual.fig".format(fibertype, name)))
        figdir = os.path.join(debugDir, fibertype+"_"+name+"FittingPngs")
        #!! output png
        gPlot = flux-flux_smooth; gPlot[badmask]=0
        rPlot = flux-flux_fit_result_lines; rPlot[badmask]=0
        if pngdebugResidual and (refit or replot or not os.path.exists(os.path.join(figdir, "{:0>4d}.png".format(0)))):
            fig.set_size_inches(*figsize)
            breakPoints = np.linspace(xstart, xend, pngNumber+1)
            allStarts=breakPoints.copy();allStarts=allStarts-pngOverlap*eachLength;allStarts=allStarts[:-1]
            allEnds=breakPoints.copy();allEnds=allEnds+pngOverlap*eachLength;allEnds=allEnds[1:]
            extract.safe_mkdir(figdir)
            for i, (eachStart, eachEnd) in enumerate(zip(allStarts, allEnds)):
                sys.stdout.write('\toutput png {}\t\t\t\r'.format(i)); sys.stdout.flush()
                inrange = (wave>eachStart)&(wave<eachEnd)
                if not inrange.size:
                    continue
                yMax = np.max([gPlot[inrange], rPlot[inrange]])
                yMax = yMax + yMax * [-0.15, 0.15][yMax>0]
                yMin = np.min([rPlot[inrange], rPlot[inrange]])
                yMin = yMin - yMin * [-0.15, 0.15][yMin>0]
                ax.set_xlim([eachStart, eachEnd])
                ax.set_ylim([yMin, yMax])
                figname = os.path.join(figdir, "r{:0>4d}.png".format(i))
                fig.savefig(figname, dpi=300, format="png", bbox_inches="tight")
        for i, (eachStart, eachEnd, eachYmax, eachFactor, eachStr) in enumerate(alertRegions):
            inrange = (wave>eachStart)&(wave<eachEnd)
            if not inrange.size:
                continue
            yMax = np.max([gPlot[inrange], rPlot[inrange]])
            yMax = yMax + yMax * [-0.15, 0.15][yMax>0]
            yMin = np.min([rPlot[inrange], rPlot[inrange]])
            yMin = yMin - yMin * [-0.15, 0.15][yMin>0]
            ax.set_xlim([eachStart, eachEnd])
            ax.set_ylim([yMin, yMax])
            ax.set_title(eachStr)
            figname = os.path.join(debugDir, "_{}_alert_{}_residual_{:_^7.3f}_{:_^7.3f}_{:^7.3f}.fig".\
                    format(fibertype, name, eachStart, eachEnd, eachFactor))
            de.d(fig, figname)
        #endregion: about png outputs
        if debug>1:
            plt.show()
        else:
            plt.close(fig)

    thisTitle = fibertype + " " + name + ["", " with continuum tweak"][tweakContinuum] +\
            ["", " with flux tweak"][tweakFlux] + ["  notSkylike", "  isSkylike", "  isSkylike in sky fitting"][isSkylike] +\
            "fa:{:.2f}".format(FAFactor)
    if lastPlot and False:#!! fitting full result
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_title(thisTitle)
        #region: highlightings
        mask = np.bitwise_and(resultMask, 0b11000)==0b01000#!! small flux line selected
        se = getPlotStartEnd(mask)
        for s, e in se:
            ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,2,2,0], "orange", alpha=0.2)
        mask = np.bitwise_and(resultMask, 0b11000)==0b11000#!! small flux line used
        se = getPlotStartEnd(mask)
        for s, e in se:
            ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,2,2,0], "cyan", alpha=0.2)
        mask = np.bitwise_and(resultMask, 0b1100000)==0b0100000#!! not use to constrain continuum
        se = getPlotStartEnd(mask)
        for s, e in se:
            ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,3,3,0], "yellow", alpha=0.2)
        mask = np.bitwise_and(resultMask, 0b1100000)==0b1100000#!! used to constrain continuum
        se = getPlotStartEnd(mask)
        for s, e in se:
            ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,3,3,0], "green", alpha=0.2)
        if doSecondFitting:
            mask = np.logical_not(np.isfinite(std))
            se = getPlotStartEnd(mask)#!! cosmic
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "black", alpha=0.8)
            mask = np.bitwise_and(resultMask, 0b100)==0b100#!! full reject
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "r", alpha=0.3)
            mask = np.bitwise_and(resultMask, 0b010)==0b010#!! partial reject
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,1,1,0], "pink", alpha=0.3)

            mask = np.bitwise_and(resultMask, 0b100000000)==0b100000000#!! too high (may be a line!!)
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,4,4,0], "blue", alpha=0.1)
            mask = np.bitwise_and(resultMask, 0b010000000)==0b010000000#!! too low (?????)
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,4,4,0], "darkorchid", alpha=0.1)

            mask = np.bitwise_and(resultMask, 0b1000000000)==0b1000000000#!! do residual subtraction
            se = getPlotStartEnd(mask)
            for s, e in se:
                ax.fill([wave[s],wave[s],wave[e],wave[e]], [0,4,4,0], "green", alpha=0.05)
        #endregion: highlightings
        #region: vertical lines
        #!! init mu
        for eachcen, eachlower, eachupper in zip(skywave, sky_lower, sky_upper):
            ax.plot([eachcen, eachcen], [0, 1], "--", c="black")
            ax.plot([eachlower, eachlower], [0, 1], "--", c="green")
            ax.plot([eachupper, eachupper], [0, 1], "--", c="red")
        #!! fitting blocks
        for i in range(blocks.shape[0]):
            eachstart = blocks[i][0]
            eachend = blocks[i][1]
            eachstartOffset = blocks[i][2]
            eachendOffset = blocks[i][3]
            eestart = blocks[i][0]
            eeend = blocks[i][1]
            ax.plot([eachstartOffset, eachstartOffset], [0, 2], "-.", c="green", lw=3)
            ax.plot([eachstart, eachstart], [0, 2], "-", c="green", lw=3)
            ax.plot([eachend, eachend], [0, 2], "--", c="red", lw=4)
            ax.plot([eachendOffset, eachendOffset], [0, 2], "-.", c="red", lw=4)
        #endregion: vertical lines
        #region: mu fitting and profile (green profile)
        if tweakFlux:#!! ??????
            for i in range(modelPolyResult.shape[0]):
                nSky, popt = modelFittingPoptsOld[i]
                if popt is not None:
                    popt = popt[:5*nSky]
                    yy = profile0(x, *popt)
                    ax.plot(x,yy, "k--", lw=2)
            ax.plot(wave, flux_fit_resultOld, "c-", lw=2)
        for i in range(modelFittingResult.shape[0]):
            eachmu = modelFittingResult["mu"][i]
            eachA = modelFittingResult["A"][i]
            eachsigma = modelFittingResult["sigma"][i]
            eachA3 = modelFittingResult["A3"][i]
            eachsigma3 = modelFittingResult["sigma3"][i]
            x,y = specLib.profile23(waveset_flatten, eachmu, eachA, eachsigma, eachA3, eachsigma3)
            if not y.size:
                continue
            ymax = y.max()
            if not np.isfinite(ymax) or np.abs(eachmu)>1e5:#!! useless?
                if modelFittingResultOld is not None:
                    muOrigin = modelFittingResultOld[i]["mu"]
                else:
                    muOrigin = "?"
                logger.warn("bat fitting at {}:{} mu:{} ymax:{} muOrigin:{}".\
                        format(fibertype, name, eachmu, ymax, muOrigin))
            ax.plot([eachmu, eachmu], [0, ymax], "b--")
            ax.plot(x,y, "g--")
        #endregion: mu fitting and profile (green profile)
        #region: all profile for one block (blue profile) and continuum !!
        for i, (eachstart, eachend, eachstartTrue, eachendTrue, eachcoeffs) in enumerate(modelPolyResult):
            xmask = (waveset_flatten>eachstart)&(waveset_flatten<eachend)
            xmaskTrue = xmask.copy()
            xmaskTrue[(waveset_flatten<eachstartTrue)|(waveset_flatten>eachendTrue)] = 0
            x = waveset_flatten[xmask]
            xmaskmask = (x>eachstartTrue)&(x<eachendTrue)
            obj = Legendre(eachcoeffs, domain=[eachstart, eachend])
            y = obj(x)
            continuumNew[xmaskTrue] = (continuum[xmaskTrue] + y[xmaskmask])[:]
            ax.plot(x,y, "r--", lw=2)
            nSky, popt = modelFittingPopts[i]
            if popt is not None:#!! ????????
                popt = popt[:5*nSky]
                yy = profile0(x, *popt)
                ax.plot(x,yy, "b--", lw=2)
        #!! continuum
        ax.plot(waveset_flatten, -continuum, "k")
        ax.plot(waveset_flatten, -continuumNew, "b")
        ax.plot(wave, std, color="pink")
        ax.plot(wave, flux_smooth, "g-", lw=2)
        ax.scatter(wave, flux, edgecolors=None, s=markerSize)
        ax.plot(wave, flux_fit_result, "r-", lw=2)
        if doSecondFitting:
            ax.plot(wave, originResidualConvolved, color="blue", lw=2)
            ax.plot(wave, residualInterpolated, color="black", lw=3)
            ax.scatter(residualX, residualY, color="black", s=50)
            ax.plot([residualX[0], residualX[0]], [0,3], "k--", lw=4)
            ax.plot([residualX[-1], residualX[-1]], [0,3], "k--", lw=4)
        #endregion: continuum
        #region: about png outputs
        #!! change xlim before dump
        xstart = waveset_flatten.min()
        xend   = waveset_flatten.max()
        ax.plot([xstart, xend], [0,0], color="black", lw=0.5)
        eachLength = (xend-xstart)/pngNumber
        ax.set_xlim([xstart, xstart + eachLength])
        fig.set_size_inches(12,3)
        figobj = {"figure":fig, "x":wave, "class":DebugFigure}
        de.d(figobj, os.path.join(debugDir,"{}_{}_block_fit.fig".format(fibertype, name)))
        figdir = os.path.join(debugDir, fibertype+"_"+name+"FittingPngs")
        #!! output png
        if pngdebug and (replot or refit or not os.path.exists(os.path.join(figdir, "{:0>4d}.png".format(0)))):
            fig.set_size_inches(*figsize)
            breakPoints = np.linspace(xstart, xend, pngNumber+1)
            allStarts=breakPoints.copy();allStarts=allStarts-pngOverlap*eachLength;allStarts=allStarts[:-1]
            allEnds=breakPoints.copy();allEnds=allEnds+pngOverlap*eachLength;allEnds=allEnds[1:]
            extract.safe_mkdir(figdir)
            for i, (eachStart, eachEnd) in enumerate(zip(allStarts, allEnds)):
                sys.stdout.write('\toutput png {}\t\t\t\r'.format(i)); sys.stdout.flush()
                inrange = (wave>eachStart)&(wave<eachEnd)
                if not inrange.sum():
                    continue
                try:
                    yMax = np.min([pngPlotYMax, flux_smooth[inrange].max()*1.1])
                    yMin = np.min([-continuum[inrange], -continuumNew[inrange]])
                except Exception as e:
                    print(e)
                    ipdb.set_trace()
                ax.set_xlim([eachStart, eachEnd])
                ax.set_ylim([yMin, yMax])
                figname = os.path.join(figdir, "{:0>4d}.png".format(i))
                fig.savefig(figname, dpi=300, format="png", bbox_inches="tight")
        for i, (eachStart, eachEnd, eachYmax, eachFactor, eachStr) in enumerate(alertRegions):
            inrange = (wave>eachStart)&(wave<eachEnd)
            if not inrange.size:
                continue
            yMax = eachYmax*1.2
            yMin = np.min([-continuum[inrange], -continuumNew[inrange]])
            ax.set_xlim([eachStart, eachEnd])
            ax.set_ylim([yMin, yMax])
            ax.set_title(eachStr)
            figname = os.path.join(debugDir, "_{}_alert_{}_{:_^10.7f}_{:_^10.7f}_{:^10.7f}.fig".\
                    format(fibertype, name, eachStart, eachEnd, eachFactor))
            de.d(fig, figname)

        #endregion: about png outputs
        if debug>1:
            plt.show()
        else:
            plt.close(fig)

    now = time.time()
    fittingTime = now-time00
    messager.cPrint("\nfinish fitting for {}:{} with time:{:.2f}s {:.2f}s factor:{}".format(fibertype, name, now-time0, now-time00, FAFactor), "gkb")
    logger.log("finish fitting for {}:{} with time:{:.2f}s {:.2f}s factor:{}".format(fibertype, name, now-time0, now-time00, FAFactor))
    #endregion: output plot
    result["continuumNew"] = continuumNew
    result["modelFittingResult"] = modelFittingResult
    result["lines"] = flux_fit_result_lines
    result["fittingGoodMask"] = flux_fit_goodMask
    result["continuum"] = flux_fit_result_poly + flux_continuum
    result["fittingTime"] = fittingTime
    if doSecondFitting:
        result["gaussianFittingResult"] = flux_fit_gaussian_result
        result["objContinuum"] = flux_fit_result_obj_continuum
        #getSigma(modelFittingResult, yset, wave, flux_fit_gaussian_result, skyFlag)
    #!! normal poly = poly + continuumTweak
    #!! normal line = line - continuumTweak
    with open(cacheFileFinal, "wb") as f:
        pickle.dump(result, f)
    #!! ouput debug txt data
    #thisLimitUpper = thisstd * (eachtype*1.2+20) * defaultSigma * 1.25
    #thisLimit = thisstd * eachResidualType * defaultSigma * 1.25#!! sqrt(2pi)~=2.575~=2.5
    #thisLimitUpper = thisstd * (eachtype*1.2+20) * defaultSigma * 1.25
    headerStr="   sg   lower   upper     sky      mu        A       A3  sig sig3     status rst  st   totalA       ff     fpf      sAL      sAU     sA3L     sA3U      ssL      ssU     ss3L     ss3U      gmu       gA       gsigma flag\n"
    formatStr="{:5s} {:7.2f} {:7.2f} {:7.2f} {:7.2f} {:8.2e} {:8.2e} {:4.2f} {:4.2f} {:10b} {:3d} {:3d} {:8.2e} {:8.2f} {:7.2f} {:8.2e} {:8.2e} {:8.2e} {:8.2e} {:8.2e} {:8.2e} {:8.2e} {:8.2e} {:8.2f} {:8.2e} {:8.2e} {:4d}\n"
    with open(os.path.join(debugDir,"{}_{}_block_fit.txt".format(fibertype, name)), "w") as f:
        f.write(headerStr)
        for i in range(skywave.shape[0]):
            thisTotalA = modelFittingResult[i]["A"]+modelFittingResult[i]["A3"]
            thismu = modelFittingResult[i]["mu"]
            thisSigmaError = modelFittingResult[i]["sigmaError"]
            thisSigmaErrorUpper = modelFittingResult[i]["sigmaErrorUpper"]
            thisResidualSigmaError = modelFittingResult[i]["residualSigmaError"]
            thisSigma = modelFittingResult[i]["sigma"]
            thisSigma3 = modelFittingResult[i]["sigma3"]
            thisA = modelFittingResult[i]["A"]
            thisA3 = modelFittingResult[i]["A3"]
            thisApeak = thisA/(np.sqrt(np.pi*2)*thisSigma)
            thisApeak3 = thisA3/(2.575798634*thisSigma3)
            thisPeak = thisApeak+thisApeak3
            thisStdPeak = np.median(stdObj(np.linspace(sky_lower[i], sky_upper[i], 100)))
            thisFluxFactor = thisTotalA/thisSigmaError
            thisFluxPeakFactor = thisPeak/thisStdPeak
            toWrite =\
                formatStr.format(skyGroup[i], sky_lower[i], sky_upper[i], skywave[i],
                             thismu, thisA, thisA3, thisSigma, thisSigma3, modelFittingResult[i]["status"],
                             skyResidualType[i], skytype[i],
                             thisTotalA, thisFluxFactor, thisFluxPeakFactor,
                             skyALower[i], skyAUpper[i], skyA3Lower[i], skyA3Upper[i],
                             skySigmaLower[i], skySigmaUpper[i], skySigma3Lower[i], skySigma3Upper[i],
                             modelFittingResult[i]["gaussianMu"],
                             modelFittingResult[i]["gaussianA"],
                             modelFittingResult[i]["gaussianSigma"],
                             skyFlag[i])
            f.write(toWrite)
        if doSecondFitting:
            polyStr="index   start startTrue  endTrue      end   nPoly\n"
            polyFormatStr="{:5d}  {:8.2f} {:8.2f} {:8.2f} {:8.2f} {:8d}\n"
            f.write(polyStr)
            for i in range(polyDebugInfo.shape[0]):
                to = polyDebugInfo[i]
                f.write(polyFormatStr.format(to["block"], to["start"], to["startTrue"],
                                             to["endTrue"], to["end"], to["nPoly"]))
                arrayStr = "\t\t"+"{:8.2e} "*polyDebugInfo[i]["nPoly"]+"\n"
                pl, p0, pu = polyDebugInfo[i]["lower"], polyDebugInfo[i]["p0"], polyDebugInfo[i]["upper"]
                if pl is not None:
                    f.write(arrayStr.format(*pl))
                else:
                    f.write("\n")
                if p0 is not None:
                    f.write(arrayStr.format(*p0))
                else:
                    f.write("\n")
                if pu is not None:
                    f.write(arrayStr.format(*pu))
                else:
                    f.write("\n")
    return result

"""fitname in [first second third]
"""
def skyfitOnce(fitname, logger, time0, time00, time000,
        #!! from upper
        wave, flux, std, flux_smooth, skyGroup,
        waveset_flatten, flux_smooth_flatten, continuum, blocks,
        skywave, sky_lower, sky_upper, skysigma, skysigma3, skytype, skyResidualType,
                 skySigmaLower, skySigmaUpper, skySigma3Lower, skySigma3Upper,
                 skyALower, skyAUpper, skyA3Lower, skyA3Upper, skyFlag,
                 skyA, skyA3,
        #!! gen in fitSky
        stdObj, goodmask,
        #!! variable to modify
        modelFittingResult, modelFittingPopts, modelPolyResult,
        modelFittingResultUpper, modelFittingResultLower,
        modelPolyResultUpper, modelPolyResultLower,
        resultMask, alertRegions,
        flux_fit_result, flux_fit_gaussian_result,
        flux_fit_result_lines, flux_fit_result_poly,
        flux_fit_goodMask,
        configs):
    #region: parameters
    refit = configs["refit"]
    refitOrigin = refit
    polyConstrain = configs["polyConstrain"]
    profileProtect = configs["profileProtect"]
    defaultSigma = configs["defaultSigma"]
    doSecondFitting = configs["doSecondFitting"]
    FAFactor = configs["FAFactor"]
    FAFactorLength = configs["FAFactorLength"]
    FAFactorEnd = configs["FAFactorEnd"]
    tweakContinuum = configs["tweakContinuum"]
    tweakContinuumFactor = configs["tweakContinuumFactor"]
    checkReject = configs["checkReject"]
    name = configs["name"]
    fiberCount = configs["fiberCount"]
    fiberTotal = configs["fiberTotal"]
    fibertype = configs["fibertype"]
    debugDir = configs["debugDir"]
    hardBoundaryCount = configs["hardBoundaryCount"]
    boundaryCount = configs["boundaryCount"]
    doResidualSubtract = configs["doResidualSubtract"]
    doResidualPolyConstrain = configs["doResidualPolyConstrain"]
    tweakResidualContinuumFactor = configs["tweakResidualContinuumFactor"]
    residualObj = configs.get("residualObj")
    upperFactor = configs.get("upperFactor", 1)
    lowerFactor = configs.get("lowerFactor", 1)
    modelFittingResultOld = configs.get("modelFittingResultOld")
    modelFittingPoptsOld = configs.get("modelFittingPoptsOld")
    flux_fit_result_linesOld = configs.get("flux_fit_result_linesOld")
    flux_fit_resultOld = configs.get("flux_fit_resultOld")
    flux_fit_result_obj_continuum = configs.get("flux_fit_result_obj_continuum")
    onlyMark = configs.get("onlyMark")
    doAlert = configs.get("doAlert")
    maxEmissionWidth = configs.get("maxEmissionWidth")
    tweakFlux   = configs["tweakFlux"]
    resultALimits = configs.get("resultALimits")
    blockSize = blocks.size
    maxRefit = configs.get("maxRefit")
    isSkylike = configs.get("isSkylike", False)
    fittingTweakOrigin = configs.get("fittingTweak", 1)
    maxRecorrect = configs.get("maxRecorrect", 3)
    polyThirdConstrain = configs.get("polyThirdConstrain", False)
    directDoThird = configs.get("directDoThird", False)
    #endregion: parameters
    #region: debugPlot
    def debugPlot(plotList, title=""):
        debugFile = os.path.join(debugDir, "_{}_{}.fig".format(name, title))
        fig = plt.figure()
        ax = fig.add_subplot(111)
        for each in plotList:
            x = each[0]
            y = each[1]
            if len(each)==3:
                kwargs = each[2]
            else:
                kwargs = {}
            if kwargs.get("type") is None:
                ax.plot(x,y,**kwargs)
            elif kwargs.get("type") =="plot":
                ax.plot(x,y,**kwargs)
            elif kwargs.get("type") =="scatter":
                ax.scatter(x,y,**kwargs)
            else:
                raise Exception("unknown plot type:{}".format(kwargs.get("type")))
        de.d(fig, debugFile)
        plt.close(fig)
    #endregion: debugPlot
    count=0
    reFitCount = 0
    reCorrect = 0
    fittingTweak = fittingTweakOrigin
    #for count, (eachstartTrue, eachendTrue, eachstart, eachend, eachnPoly, eachFitFactor, eachFitFactorPoly)\
        #in enumerate(blocks):
    if fitname=="third":
        polyDebugInfo = np.empty((blocks.shape[0],), dtype=dtype_polyDebug)
    while count<blocks.shape[0]:
        useFile = 0
        eachstartTrue = blocks[count][0]
        eachendTrue = blocks[count][1]
        eachstart = blocks[count][2]
        eachend = blocks[count][3]
        eachnPoly = blocks[count][4]
        eachFitFactor = blocks[count][5]
        eachFitFactorPoly = blocks[count][6]
        #region: setups for wavemask and skymask
        wave_flatten_mask = (waveset_flatten>eachstart)&(waveset_flatten<eachend)
        wave_fit_mask = (wave>eachstart)&(wave<eachend)
        wave_fit_mask_true = wave_fit_mask.copy()#!! actual region
        wave_fit_mask_true[(wave<eachstartTrue)|(wave>eachendTrue)]=0
        wave_fit = wave[wave_fit_mask]
        wave_fit_mask_mask = (wave_fit>eachstartTrue)&(wave_fit<eachendTrue)
        flux_fit = flux[wave_fit_mask]
        std_fit = std[wave_fit_mask]
        #factorBase = 1e-8
        factorBase = np.nanpercentile(std_fit,30)/1e5
        if not wave_fit.size:
            count+=1
            continue
        sky_fit_mask = (skywave>eachstart)&(skywave<eachend)&\
                       (skywave+profileProtect>wave_fit[0])&(skywave-profileProtect<wave_fit[-1])
        sky_fit_mask_true = (skywave>eachstartTrue)&(skywave<eachendTrue)&\
                       (skywave+profileProtect>wave_fit[0])&(skywave-profileProtect<wave_fit[-1])
        sky_fit = skywave[sky_fit_mask]
        sky_fit_mask_mask = (sky_fit>eachstartTrue)&(sky_fit<eachendTrue)
        skyIndexs = np.where(sky_fit_mask)[0]
        skyIndexsTrue = np.where(sky_fit_mask_true)[0]
        skyIndexsTrueIndex = np.where(sky_fit_mask_mask)[0]
        assert skyIndexsTrue.size==skyIndexsTrueIndex.size
        continuumInFit = continuum[wave_flatten_mask]
        nSky = sky_fit.size
        if nSky==0 or wave_fit.size==0:
            count+=1
            continue
        thisProfile = specLib.profileGen(eachnPoly)
        thisProfileSP = specLib.profileGenSP(eachnPoly)
        if fitname=="third":
            thisProfileGaussian = specLib.profileGaussianGen(eachnPoly)
        #endregion: setups
        #region: setup sky data
        sky_type_fit  = skytype[sky_fit_mask]
        sky_residualType_fit  = skyResidualType[sky_fit_mask]
        sky_group_fit  = skyGroup[sky_fit_mask]
        skysigma_fit       = skysigma[sky_fit_mask]
        skysigma3_fit      = skysigma3[sky_fit_mask]
        skySigmaLower_fit  = skySigmaLower[sky_fit_mask]
        skySigmaUpper_fit  = skySigmaUpper[sky_fit_mask]
        skySigma3Lower_fit = skySigma3Lower[sky_fit_mask]
        skySigma3Upper_fit = skySigma3Upper[sky_fit_mask]
        skyALower_fit  = skyALower[sky_fit_mask]
        skyAUpper_fit  = skyAUpper[sky_fit_mask]
        skyA3Lower_fit = skyA3Lower[sky_fit_mask]
        skyA3Upper_fit = skyA3Upper[sky_fit_mask]
        sky_upper_fit  = sky_upper[sky_fit_mask]
        sky_lower_fit  = sky_lower[sky_fit_mask]
        sky_A_fit = skyA[sky_fit_mask]
        sky_A3_fit = skyA3[sky_fit_mask]
        #!! make sure sky_A is good to use
        for eachIndex in skyIndexs: #!! all mask to be in ra|1
            modelFittingResult[eachIndex]["status"] = modelFittingResult[eachIndex]["status"]|1
        #endregion: setup sky data
        #region: setup p0 and boundarys for the {fitname} fitting
        if fitname=="third":
            if checkReject:#!! should be after constrain small flux
                lastmu = modelFittingResult[skyIndexs]["mu"]
                lastsigma  = modelFittingResult[skyIndexs]["sigma"]
                lastsigma3 = modelFittingResult[skyIndexs]["sigma3"]
                lastA  = modelFittingResult[skyIndexs]["A"]
                lastA3 = modelFittingResult[skyIndexs]["A3"]
                partialBadMask, fullBadMask = checkProfile(\
                        wave_fit, std_fit, lastmu, lastsigma, lastsigma3, lastA, lastA3)
                #sky_A[fullBadMask] = 0
                #skyAUpper_fit[fullBadMask] = 1e-10
                #sky_A3[fullBadMask] = 0
                #skyA3Upper_fit[fullBadMask] = 1e-10
                partialBadIndex = skyIndexs[partialBadMask]
                fullBadIndex = skyIndexs[fullBadMask]
                #print("partial:{} full:{}".format(partialBadIndex, fullBadIndex))
                for eachIndex in partialBadIndex:
                    modelFittingResult[eachIndex]["status"] = modelFittingResult[eachIndex]["status"]|0b010
                for eachIndex in fullBadIndex:
                    modelFittingResult[eachIndex]["status"] = modelFittingResult[eachIndex]["status"]|0b100
            #!! set lines in the first and end hardBoundaryCount pixels not in range (about 10)
            for eachmu, eachSkyIndex in zip(sky_fit, skyIndexs):
                if eachmu<wave[hardBoundaryCount] or eachmu>wave[-hardBoundaryCount]:
                    modelFittingResult[eachSkyIndex]["status"] = modelFittingResult[eachSkyIndex]["status"]&0xFFFFFFFE
            #!! sub residual continuum (or not)
            if doResidualSubtract:
                residual_continuum_fit = residualObj(wave_fit)
                #!! fit failed???
                if np.any(np.logical_not(np.isfinite(residual_continuum_fit))):
                    residual_continuum_fit = np.zeros(wave_fit.shape)
                    haveResidualSubtract = False
                else:
                    haveResidualSubtract = True
            else:
                residual_continuum_fit = np.zeros(wave_fit.shape)
                haveResidualSubtract = False
            #!! flux subtract residual_continuum(obj continuum before fitting) before fitting
            flux_fit = flux_fit - residual_continuum_fit
            polyValue = np.abs(np.nanpercentile(stdObj(wave_fit), 10)) * 2
            #!! use std constrain instead of continuum flux constran for the third fitting
            #!!     5 * median std
            if polyConstrain or doResidualPolyConstrain or polyThirdConstrain:
                polyUpper = np.ones(eachnPoly)*polyValue*eachFitFactorPoly#*upperFactor
                polyLower = -np.ones(eachnPoly)*polyValue*eachFitFactorPoly#*lowerFactor
                polyUpper[0] = polyUpper[0]*upperFactor
                polyLower[0] = polyLower[0]*lowerFactor
                if eachnPoly>1:
                    polyUpper[1] = polyUpper[1]*np.min([lowerFactor, upperFactor])
                    polyLower[1] = polyLower[1]*np.min([lowerFactor, upperFactor])
                if eachnPoly>2:
                    polyUpper[2] = polyUpper[2]*np.min([lowerFactor, upperFactor])/3
                    polyLower[2] = polyLower[2]*np.min([lowerFactor, upperFactor])/3
                if eachnPoly>3:
                    polyUpper[3] = polyUpper[3]*np.min([lowerFactor, upperFactor])/3
                    polyLower[3] = polyLower[3]*np.min([lowerFactor, upperFactor])/3
                if eachnPoly>4:
                    polyUpper[4] = polyUpper[4]*np.min([lowerFactor, upperFactor])/4
                    polyLower[4] = polyLower[4]*np.min([lowerFactor, upperFactor])/4
                if eachnPoly>5:
                    polyUpper[5] = polyUpper[5]*np.min([lowerFactor, upperFactor])/5
                    polyLower[5] = polyLower[5]*np.min([lowerFactor, upperFactor])/5
                #if doResidualPolyConstrain and not reFitCount:#!! (name is not right) obj and sky be true... leave it here
            else:
                polyUpper = np.ones(eachnPoly)*np.inf
                polyLower = -np.ones(eachnPoly)*np.inf
        else:
            if polyConstrain:#!! default is True, sky is True, and obj is False
                #polyValue = np.abs(np.median(continuumInFit))
                polyValue = np.abs(np.nanpercentile(stdObj(wave_fit), 10)) * 2
                polyUpper = np.ones(eachnPoly)*polyValue*eachFitFactorPoly
                polyLower = -np.ones(eachnPoly)*polyValue*eachFitFactorPoly
                if eachnPoly>1:
                    polyUpper[1] = polyUpper[1]/2
                    polyLower[1] = polyLower[1]/2
                if eachnPoly>2:
                    polyUpper[2] = polyUpper[2]/3
                    polyLower[2] = polyLower[2]/3
                if eachnPoly>3:
                    polyUpper[3] = polyUpper[3]/3
                    polyLower[3] = polyLower[3]/3
                if eachnPoly>4:
                    polyUpper[4] = polyUpper[4]/4
                    polyLower[4] = polyLower[4]/4
                if eachnPoly>5:
                    polyUpper[5] = polyUpper[5]/5
                    polyLower[5] = polyLower[5]/5
            else:
                polyUpper = np.ones(eachnPoly)*np.inf
                polyLower = -np.ones(eachnPoly)*np.inf
        if fitname=="first":#!! calculate flux for the small flux and residual constrain skylines and update status
            #!! constrain small flux lines (only in the first fitting)
            for i, (eachtype, eachResidualType, eachIndex) in\
                    enumerate(zip(sky_type_fit, sky_residualType_fit, skyIndexs)):
                thismu = sky_fit[i]
                searchFactor=2
                while True:
                    inrange = (wave[goodmask]>thismu-defaultSigma*searchFactor)&(wave[goodmask]<thismu+defaultSigma*searchFactor)
                    thisstds = stdObj(wave[goodmask][inrange])
                    if len(thisstds):
                        thisstd = np.median(thisstds)
                        break
                    else:
                        searchFactor+=1
                if eachResidualType:#!! each type is estimated by hight
                    thisLimit = thisstd * eachResidualType * defaultSigma * 1.25#!! sqrt(2pi)~=2.575~=2.5
                    assert np.isfinite(thisLimit) and thisLimit>0
                    modelFittingResult[eachIndex]["residualSigmaError"] = thisLimit*2
                    modelFittingResultLower[eachIndex]["residualSigmaError"] = thisLimit*2
                    modelFittingResultUpper[eachIndex]["residualSigmaError"] = thisLimit*2
                    modelFittingResult[eachIndex]["status"] = modelFittingResult[eachIndex]["status"]|0b0100000
                if eachtype:#!! each type is estimated by hight
                    thisLimit = thisstd * eachtype * defaultSigma * 1.25#!! sqrt(2pi)~=2.575~=2.5
                    assert np.isfinite(thisLimit) and thisLimit>0
                    #!! useful or not ?
                    #thisLimitUpper = thisstd * (eachtype*1.2+20) * defaultSigma * 1.25
                    thisLimitUpper = thisLimit*2
                    #!! when do skyline fitting, isSkyLike=2
                    #!! when do obj fitting, isSkyLike=1
                    #!! skyA is updated in the first fit!!
                    if not doSecondFitting or isSkylike>1:#!! constrain the small flux lines directly
                        sky_A_fit[i] = thisLimit*0.99
                        sky_A3_fit[i] = thisLimit*0.99
                        skyAUpper_fit[i] = thisLimit
                        skyA3Upper_fit[i] = thisLimit
                        skyAUpper[eachIndex] = thisLimit
                        skyA3Upper[eachIndex] = thisLimit
                        skyA[eachIndex] = thisLimit*0.99
                        skyA3[eachIndex] = thisLimit*0.99
                        modelFittingResult[eachIndex]["status"] = modelFittingResult[eachIndex]["status"]|0b11000
                    else:
                        modelFittingResult[eachIndex]["status"] = modelFittingResult[eachIndex]["status"]|0b01000
                    modelFittingResult[eachIndex]["sigmaError"] = thisLimit*2
                    modelFittingResult[eachIndex]["sigmaErrorUpper"] = thisLimitUpper*2
                    modelFittingResultUpper[eachIndex]["sigmaError"] = thisLimit*2
                    modelFittingResultUpper[eachIndex]["sigmaErrorUpper"] = thisLimitUpper*2
                    modelFittingResultLower[eachIndex]["sigmaError"] = thisLimit*2
                    modelFittingResultLower[eachIndex]["sigmaErrorUpper"] = thisLimitUpper*2
                else:
                    thisLimit = thisstd * 4 * defaultSigma * 1.25#!! sqrt(2pi)~=2.575~=2.5
                    assert np.isfinite(thisLimit) and thisLimit>0
                    modelFittingResult[eachIndex]["sigmaError"] = thisLimit*2
                    modelFittingResult[eachIndex]["sigmaErrorUpper"] = thisLimit*4
                    modelFittingResultUpper[eachIndex]["sigmaError"] = thisLimit*2
                    modelFittingResultUpper[eachIndex]["sigmaErrorUpper"] = thisLimit*4
                    modelFittingResultLower[eachIndex]["sigmaError"] = thisLimit*2
                    modelFittingResultLower[eachIndex]["sigmaErrorUpper"] = thisLimit*4
            if checkReject:#!! should be after constrain small flux, default False, sky and obj True
                partialBadMask, fullBadMask = checkProfile(\
                        wave_fit, std_fit, sky_fit, skysigma_fit, skysigma3_fit, sky_A_fit, sky_A3_fit)
                #sky_A[fullBadMask] = 0
                #skyAUpper_fit[fullBadMask] = 1e-10
                #sky_A3[fullBadMask] = 0
                #skyA3Upper_fit[fullBadMask] = 1e-10
                partialBadIndex = skyIndexs[partialBadMask]
                fullBadIndex = skyIndexs[fullBadMask]
                #print("partial:{} full:{}".format(partialBadIndex, fullBadIndex))
                for eachIndex in partialBadIndex:
                    modelFittingResult[eachIndex]["status"] = modelFittingResult[eachIndex]["status"]|0b010
                for eachIndex in fullBadIndex:
                    modelFittingResult[eachIndex]["status"] = modelFittingResult[eachIndex]["status"]|0b100
        p0 = np.concatenate((sky_fit,
                       sky_A_fit, skysigma_fit,
                       sky_A3_fit, skysigma3_fit,
                       np.zeros(eachnPoly))).tolist()
        bounds_upper = np.concatenate((sky_upper_fit,
                           skyAUpper_fit, skySigmaUpper_fit,
                           skyA3Upper_fit, skySigma3Upper_fit,
                           polyUpper)).tolist()
        bounds_lower = np.concatenate((sky_lower_fit,
                           skyALower_fit, skySigmaLower_fit,
                           skyA3Lower_fit, skySigma3Lower_fit,
                           polyLower)).tolist()
        #endregion: setup p0 and boundarys
        #region: printings for the first fitting
        #y = thisProfile(wave_fit, *p0)
        now = time.time()
        toPrintStr = '\t {}/{} {}:{} {:3}/{:3} {} fitting: block {:>10.3f}~{:<10.3f} with {:3} profiles t={:.2f}s {:.2f}s {:.2f}s fa:{:.2f}'.\
                      format(fiberCount, fiberTotal, fibertype, name, count, blockSize, fitname,
                             eachstart, eachend, nSky, now-time0, now-time00, now-time000, FAFactor)
        logger.log(toPrintStr)
        sys.stdout.write(toPrintStr+"\r"); sys.stdout.flush()
        #print(toPrintStr)
        if isinstance(name, int):
            name = "{:0>3d}".format(name)
        cacheFile = os.path.join(debugDir, "data", "_{}_{}_{}_fit_{:_^3}_{:_^10.3f}_{:_^10.3f}_{:_^3}.dat".\
                format(fibertype, name, fitname, count, eachstart, eachend, nSky))
        if fitname=="third":
            cacheFileg = os.path.join(debugDir, "data", "_{}_{}_third_gaussian_{:_^3}_{:_^10.3f}_{:_^10.3f}_{:_^3}.dat".\
                    format(fibertype, name, count, eachstart, eachend, nSky))
        #endregion: printings
        #region: do {fitname} fitting
        N = nSky
        lastException = Exception("????????")
        if fitname=="third":
            if refit or not os.path.exists(cacheFile):
                if FAFactor is not None:
                    #!! fit for profile233
                    goodFit = False
                    haveWarningInThisBlock=False
                    for thisi, eachFA in enumerate(np.linspace(FAFactor*eachFitFactor,
                                         FAFactorEnd*eachFitFactor, FAFactorLength)):
                        try:
                            thisConfig = {"ftol":factorBase*eachFA*fittingTweak,
                                          "xtol":factorBase*eachFA*fittingTweak,
                                          "gtol":factorBase*eachFA*fittingTweak,
                                       "max_nfev":wave_fit.size*150*(int(thisi/3)+1)}
                            popt, pcov = curve_fit(thisProfile,
                                           wave_fit, flux_fit, sigma=std_fit,
                                           p0=p0,
                                           bounds = [tuple(bounds_lower), tuple(bounds_upper)], **thisConfig)
                            goodFit = True
                            break
                        except Exception as e:
                            if not haveWarningInThisBlock:
                                print()
                                haveWarningInThisBlock=True
                            lastException = e
                            print("\n\tthird fitting failed with {} in {}:{} in block {:>10.3f}~{:<10.3f} with {:3} profiles".format(eachFA, fibertype, name, eachstart, eachend, nSky))
                            messager.cPrint("\t\t with error:"+str(e), "ykb")
                            logger.log("\tthird fitting failed with {} in {}:{} in block {:>10.3f}~{:<10.3f} with {:3} profiles".format(eachFA, fibertype, name, eachstart, eachend, nSky))
                            logger.log("\t with error:"+str(e))
                            for ii in range(N,2*N):
                                p0[ii] = p0[ii]/2.0
                            for ii in range(3*N,4*N):
                                p0[ii] = p0[ii]/2.0
                    if not goodFit:
                        logger.warn("third fitting failed with {} in {}:{} in block {:>10.3f}~{:<10.3f} with {:3} profiles\n\t{} fitting failed for {}:{} with Error {}".\
                                format(eachFA, fibertype, name, eachstart, eachend, nSky, fitname, fibertype, name, lastException))
                        flux_fit_goodMask[wave_fit_mask_true] = False
                        popt = np.ones(len(p0)) * np.nan
                        #!! temp debug
                    #!! fit for profileGaussian
                    thisPoly = np.array(popt[5*N:], dtype=np.float64)
                    thisA = np.array(popt[N:2*N], dtype=np.float64)
                    thisA3 = np.array(popt[3*N:4*N], dtype=np.float64)
                    thisgp0A = thisA+thisA3
                    gp0 = np.concatenate((sky_fit,
                                   thisgp0A, skysigma_fit,
                                   thisPoly)).tolist()
                    gbounds_upper = np.concatenate((sky_upper_fit,
                                       skyAUpper_fit+skyA3Upper_fit, skySigmaUpper_fit,
                                       thisPoly+np.abs(thisPoly)*0.1+1e-10)).tolist()
                    gbounds_lower = np.concatenate((sky_lower_fit,
                                       skyALower_fit+skyA3Lower_fit, skySigmaLower_fit,
                                       thisPoly-np.abs(thisPoly)*0.1-1e-10)).tolist()
                    goodFit = False
                    for thisi, eachFA in enumerate(np.linspace(FAFactor*eachFitFactor, FAFactorEnd*eachFitFactor, FAFactorLength)):
                        try:
                            thisConfig = {"ftol":factorBase*eachFA, "xtol":factorBase*eachFA, "gtol":factorBase*eachFA,
                                       "max_nfev":wave_fit.size*100*(int(thisi/3.0)+1)}
                            gpopt, gpcov = curve_fit(thisProfileGaussian,
                                           wave_fit, flux_fit, sigma=std_fit,
                                           p0=gp0,
                                           bounds = [tuple(gbounds_lower), tuple(gbounds_upper)], **thisConfig)
                            goodFit = True
                            break
                        except Exception as e:
                            if not haveWarningInThisBlock:
                                print()
                                haveWarningInThisBlock=True
                            lastException = e
                            print("\n\tthird fitting gaussian failed with {} in {}:{} in block {:>10.3f}~{:<10.3f} with {:3} profiles".format(eachFA, fibertype, name, eachstart, eachend, nSky))
                            messager.cPrint("\t\t with error:"+str(e), "ykb")
                            logger.log("\tthird fitting gaussian failed with {} in {}:{} in block {:>10.3f}~{:<10.3f} with {:3} profiles".format(eachFA, fibertype, name, eachstart, eachend, nSky))
                            logger.log("\t\t with error:"+str(e))
                            for ii in range(N,2*N):
                                gp0[ii] = gp0[ii]/2.0
                    if not goodFit:
                        logger.warn("third gaussian fitting failed with {} in {}:{} in block {:>10.3f}~{:<10.3f} with {:3} profiles\n\t{} fitting failed for {}:{} with Error {}".\
                                format(eachFA, fibertype, name, eachstart, eachend, nSky, fitname, fibertype, name, lastException))
                        gpopt = np.ones(len(gp0)) * np.nan
                        #!! temp debug
                else:
                    popt, pcov = curve_fit(thisProfile,
                                           wave_fit, flux_fit, sigma=std_fit,
                                           p0=p0,
                                           bounds = [tuple(bounds_lower), tuple(bounds_upper)])
                    gpopt, gpcov = curve_fit(thisProfileGaussian,
                                           wave_fit, flux_fit, sigma=std_fit,
                                           p0=gp0,
                                           bounds = [tuple(gbounds_lower), tuple(gbounds_upper)])
                with open(cacheFile, "wb") as f:
                    pickle.dump(popt, f)
                with open(cacheFileg, "wb") as f:
                    pickle.dump(gpopt, f)
            else:
                useFile=1
                with open(cacheFile, "rb") as f:
                    popt = pickle.load(f)
                with open(cacheFileg, "rb") as f:
                    gpopt = pickle.load(f)
        else:
            if refit or not os.path.exists(cacheFile):
                if FAFactor is not None: #!! default is 1, sky and obj is 0.2(to get good fitting)
                    goodFit = False
                    for thisi, eachFA in enumerate(np.linspace(FAFactor*eachFitFactor,
                                  #!! eachFitFactor is to tune accuracy for each fitting
                                  FAFactorEnd*eachFitFactor, FAFactorLength)):
                        try:
                            thisConfig = {"ftol":factorBase*eachFA, "xtol":factorBase*eachFA, "gtol":factorBase*eachFA,
                                       "max_nfev":wave_fit.size*100*(thisi+1)}
                            popt, pcov = curve_fit(thisProfile,
                                           wave_fit, flux_fit, sigma=std_fit,
                                           p0=p0,
                                           bounds = [tuple(bounds_lower), tuple(bounds_upper)], **thisConfig)
                            goodFit = True
                            break
                        except Exception as e:
                            print("\n\t{} fitting failed with {} in {}:{} in block {:>10.3f}~{:<10.3f} with {:3} profiles".\
                                    format(fitname, eachFA, fibertype, name, eachstart, eachend, nSky))
                            lastException = e
                            messager.cPrint("\t\t with error:"+str(e), "ykb")
                            logger.log("\t{} fitting failed with {} in {}:{} in block {:>10.3f}~{:<10.3f} with {:3} profiles".\
                                    format(fitname, eachFA, fibertype, name, eachstart, eachend, nSky))
                            logger.log("\twith error:"+str(e))
                            for ii in range(N,2*N):
                                p0[ii] = p0[ii]/2.0
                            for ii in range(3*N,4*N):
                                p0[ii] = p0[ii]/2.0
                    if not goodFit:
                        #plotMus(wave_fit, flux_fit, std_fit, sky_fit, sky_lower_fit, sky_upper_fit)
                        logger.warn("{} fitting failed with {} in {}:{} in block {:>10.3f}~{:<10.3f} with {:3} profiles\n\t{} fitting failed for {}:{}".\
                                    format(fitname, eachFA, fibertype, name, eachstart, eachend, nSky, fitname, fibertype, name))
                        flux_fit_goodMask[wave_fit_mask_true] = False
                        popt = np.ones(len(p0)) * np.nan
                        #!! temp debug
                else:
                    popt, pcov = curve_fit(thisProfile,
                                           wave_fit, flux_fit, sigma=std_fit,
                                           p0=p0,
                                           bounds = [tuple(bounds_lower), tuple(bounds_upper)])
                with open(cacheFile, "wb") as f:
                    pickle.dump(popt, f)
            else:
                useFile=1
                with open(cacheFile, "rb") as f:
                    popt = pickle.load(f)
        #endregion: do first fitting
        #region: estimate error and update resultMask
        poptOld = copy.copy(popt)
        thismu = np.array(popt[:N], dtype=np.float64)
        thisA = np.array(popt[N:2*N], dtype=np.float64)
        thisA3 = np.array(popt[3*N:4*N], dtype=np.float64)
        thissigma = np.array(popt[2*N:3*N], dtype=np.float64)
        thissigma3 = np.array(popt[4*N:5*N], dtype=np.float64)
        thisPoly = np.array(popt[5*N:], dtype=np.float64)
        if fitname=="third":
            thisgmu = np.array(gpopt[:N], dtype=np.float64)
            thisgA = np.array(gpopt[N:2*N], dtype=np.float64)
            thisgsigma = np.array(gpopt[2*N:3*N], dtype=np.float64)
        thisTotalA = thisA+thisA3
        thisAFactor = thisA/thisTotalA
        thisA3Factor = thisA3/thisTotalA
        thisAveSigma = thissigma*thisAFactor + thissigma3*thisA3Factor
        thisFlagUpper = thismu + thissigma*thisAFactor + thissigma3*thisA3Factor
        thisFlagLower = thismu - thissigma*thisAFactor - thissigma3*thisA3Factor
        thisFlagUpper3 = thismu + 3*thissigma*thisAFactor + 3*thissigma3*thisA3Factor
        thisFlagLower3 = thismu - 3*thissigma*thisAFactor - 3*thissigma3*thisA3Factor
        thisFlagLower3[thisFlagLower3<wave_fit[0]] = wave_fit[0]
        thisFlagLower3[thisFlagLower3>wave_fit[-1]] = wave_fit[-1]
        thisFlagUpper2 = thismu + 2*thissigma*thisAFactor + 2*thissigma3*thisA3Factor
        thisFlagLower2 = thismu - 2*thissigma*thisAFactor - 2*thissigma3*thisA3Factor
        thisFlagLower2[thisFlagLower2<wave_fit[0]] = wave_fit[0]
        thisFlagLower2[thisFlagLower2>wave_fit[-1]] = wave_fit[-1]
        if fitname in ["first", "third"]:
            #!! update status to the result mask(cosmic, small flux select, small flux residual select)
            thisStatus = modelFittingResult[skyIndexs]["status"]
            for i, (eachIndex, eachStatus) in enumerate(zip(skyIndexs, thisStatus)):
                if eachStatus>1:
                    inrange = (wave>thisFlagLower[i])&(wave<thisFlagUpper[i])
                    resultMask[inrange] = eachStatus
        #endregion: estimate error
        #!! tweak Flux for third fitting
        if fitname=="third":
            #!! detect bad fitting
            lineDeltaFitAgain = False
            if not useFile:
                if reCorrect<maxRecorrect:
                    thisLines, thisContinuum = thisProfileSP(wave_fit, *popt)
                    thisResidual = flux_fit-thisLines-thisContinuum
                    for i, eachSkyIndex in zip(range(sky_fit.size), skyIndexs):
                        thisSkyMu = thismu[i]
                        thisSkyA = thisA[i]
                        thisSkyA3 = thisA3[i]
                        thisSkyUpper = sky_upper_fit[i]
                        thisSkyLower = sky_lower_fit[i]
                        thisSkyALower = skyALower_fit[i]
                        thisSkyAUpper = skyAUpper_fit[i]
                        thisSkyA3Lower = skyA3Lower_fit[i]
                        thisSkyA3Upper = skyA3Upper_fit[i]
                        thisTestWave = wave_fit[(wave_fit>thisFlagLower2[i])&(wave_fit<thisFlagUpper2[i])]
                        thisTestWaveDiff = thisTestWave-thisSkyMu
                        thisTestResidual = flux_fit[(wave_fit>thisFlagLower2[i])&(wave_fit<thisFlagUpper2[i])]
                        thisTestStd = std_fit[(wave_fit>thisFlagLower2[i])&(wave_fit<thisFlagUpper2[i])]
                        thisTestStatus = modelFittingResult[eachSkyIndex]["status"]
                        if not thisTestWave.size:
                            continue
                        thisdiff = (thisTestResidual * thisTestWaveDiff).sum()/np.abs(thisTestResidual).sum()
                        thisL = thisFlagUpper2[i]-thisFlagLower2[i]
                        goodDiff = np.abs(thisdiff)>0.2#!! magic number
                        goodMu = thisSkyMu + thisdiff
                        thisStdFactor = np.std(thisTestResidual)/thisTestStd.max()
                        thisTestUp = getContinueUp(thisTestResidual)
                        thisTestDown = getContinueDown(thisTestResidual)
                        thisTestUpFactor = thisTestUp/thisTestStd.max()
                        thisTestDownFactor = thisTestDown/thisTestStd.max()
                        #print("\t\t\t\tbefore continue mu: {:.2f} thisdiff: {:.2f}~{:.2f}~{:.2f} status:{:b} stdF:{:.2f} up:{:.2f} down:{:.2f}".\
                                #format(thisSkyMu, thisdiff, thisL, thisdiff/thisL, thisTestStatus, thisStdFactor, thisTestUpFactor, thisTestDownFactor,))
                        #print("\t\t\t\t\t{},{}".format(thisTestResidual, thisTestWaveDiff))
                        if (thisTestStatus&0b1111000)>0:#!! skip small flux things
                            continue
                        #print("\t\t\t\tafter continue mu: {:.2f} thisdiff: {:.2f}".format(thisSkyMu, thisdiff))
                        if thisStdFactor>8 or thisTestUpFactor>7 or thisTestDownFactor>7:
                            if not goodDiff:
                                continue
                            if goodMu>thisSkyUpper: goodMu = thisSkyUpper-1e-3
                            if goodMu<thisSkyLower: goodMu = thisSkyLower+1e-3
                            skywave[eachSkyIndex] = goodMu
                            toPrintStr = '\t\t correct fitting offset: {}:delta:{:.2f},stdF:{:.2f},up:{:.2f},down:{:.2f},tweak:{:.2f} @{:.2f}~{:.2f}~{:.2f}  {}/{} {}:{} {:3}/{:3} third fitting: block {:>10.3f}~{:<10.3f} with {:3} profiles.. t={:.2f}s {:.2f}s {:.2f}s'.\
                                format(reCorrect, thisdiff, thisStdFactor, thisTestUpFactor, thisTestDownFactor, fittingTweak,
                                       thisFlagLower2[i],thisSkyMu,thisFlagUpper2[i], fiberCount, fiberTotal,
                                       fibertype, name, count, blockSize, eachstart, eachend, nSky,
                                       now-time0, now-time00, now-time000)
                            logger.log(toPrintStr)
                            if lineDeltaFitAgain:
                                print(toPrintStr)
                            else:
                                print("\n"+toPrintStr)
                            lineDeltaFitAgain = True
                        #elif np.abs(ddotr).sum()==2:#!! "^" or "v"
                            #if np.all(ddotr>0):#!! "v"
                                #skyA[eachSkyIndex] = (thisSkyA+thisSkyALower)/2
                                #skyA3[eachSkyIndex] = (thisSkyA3+thisSkyA3Lower)/2
                                #toPrintStr = '\t\t correct fitting for "v" {},{:.2f} @{:.2f}~{:.2f}~{:.2f}  {}/{} {}:{} {:3}/{:3} third fitting: block {:>10.3f}~{:<10.3f} with {:3} profiles.. t={:.2f}s {:.2f}s {:.2f}s'.\
                                #format(reCorrect, fittingTweak, thisFlagLower2[i],thisSkyMu,thisFlagUpper2[i], fiberCount, fiberTotal, fibertype, name, count, blockSize, eachstart, eachend, nSky, now-time0, now-time00, now-time000)
                                #logger.log(toPrintStr)
                                #if lineDeltaFitAgain:
                                    #messager.cPrint(toPrintStr, "ykb")
                                #else:
                                    #messager.cPrint("\n"+toPrintStr, "ykb")
                                #lineDeltaFitAgain = True
                            #elif np.all(ddotr<0):#!! "^"
                                #skyA[eachSkyIndex] = (thisSkyA+thisSkyAUpper)/2
                                #skyA3[eachSkyIndex] = (thisSkyA3+thisSkyA3Upper)/2
                                #toPrintStr = '\t\t correct fitting for "^" {},{:.2f} @{:.2f}~{:.2f}~{:.2f}  {}/{} {}:{} {:3}/{:3} third fitting: block {:>10.3f}~{:<10.3f} with {:3} profiles.. t={:.2f}s {:.2f}s {:.2f}s'.\
                                #format(reCorrect, fittingTweak, thisFlagLower2[i],thisSkyMu,thisFlagUpper2[i], fiberCount, fiberTotal, fibertype, name, count, blockSize, eachstart, eachend, nSky, now-time0, now-time00, now-time000)
                                #logger.log(toPrintStr)
                                #if lineDeltaFitAgain:
                                    #messager.cPrint(toPrintStr, "ykb")
                                #else:
                                    #messager.cPrint("\n"+toPrintStr, "ykb")
                                #lineDeltaFitAgain = True
                if lineDeltaFitAgain:
                    reCorrect+=1
                    fittingTweak *= 0.8
                    refit = True
                else:
                    refit = refitOrigin
                    fittingTweak = fittingTweakOrigin
                    reCorrect=0
            if tweakFlux: #!! tweakFlux
                haveWarningInThisBlock=False
                unique_sky_group_fit = np.unique(sky_group_fit)
                sortIndex = np.argsort(unique_sky_group_fit.astype(np.float))
                unique_sky_group_fit = unique_sky_group_fit[sortIndex]
                thisLastResults = modelFittingResultOld[sky_fit_mask]
                thisOldA  = thisLastResults["A"]
                thisOldA3 = thisLastResults["A3"]
                thisOldTotalA = thisOldA+thisOldA3
                thisAUpper = resultALimits["upper"][sky_fit_mask]
                thisALower = resultALimits["lower"][sky_fit_mask]
                thisAUpperAlert = resultALimits["upperAlert"][sky_fit_mask]
                for eachGroupName in unique_sky_group_fit:
                    groupMask = sky_group_fit==eachGroupName
                    thisGroupStatus = thisStatus[groupMask]
                    groupIndexs = np.where(groupMask)[0]
                    thisGroupmu = thismu[groupMask]
                    #!! skip starting and ending points
                    if thisGroupmu.min()<wave[boundaryCount] or thisGroupmu.max()>wave[-boundaryCount]:
                        continue
                    thisSkyIndex = skyIndexs[groupMask]
                    thisGroupLastResults = thisLastResults[groupMask]
                    thisGroupsigma = thissigma[groupMask]
                    thisGroupsigma3 = thissigma3[groupMask]

                    thisGroupA = thisA[groupMask]
                    thisGroupA3 = thisA3[groupMask]
                    thisGroupTotalA = thisGroupA+thisGroupA3

                    thisGroupOldA  = thisOldA[groupMask]
                    thisGroupOldA3 = thisOldA3[groupMask]
                    thisGroupOldTotalA = thisGroupOldA+thisGroupOldA3

                    #!! take lines effected by cosmic as average flux
                    badLines = np.bitwise_and(thisGroupStatus, 0b10)==0b10
                    thisGroupTotalA[badLines] = thisGroupOldTotalA[badLines]
                    thisGroupTotalSum = thisGroupTotalA.sum()
                    thisGroupOldTotalSum = thisGroupOldTotalA.sum()

                    thisGroupAUpper = thisAUpper[groupMask].sum()
                    thisGroupALower = thisALower[groupMask].sum()
                    thisGroupAUpperAlert = thisAUpperAlert[groupMask].sum()
                    originGroupA = thisGroupA.copy()
                    originGroupA3 = thisGroupA3.copy()
                    #thisFactor = thisGroupTotalSum / thisGroupOldTotalSum#!! only correct for the overflow
                    thisFactor = thisGroupTotalSum / thisGroupAUpper
                    thisGroupA  = originGroupA / thisFactor
                    thisGroupA3 = originGroupA3 / thisFactor
                    if thisGroupTotalSum>thisGroupAUpper:
                        thisGroupYmax = np.max(thisGroupA/(np.sqrt(2*np.pi)*thisGroupsigma) + thisGroupA3/(2.575798*thisGroupsigma3))
                        thisGroupLower = thisFlagLower[groupMask][0]
                        thisGroupUpper = thisFlagUpper[groupMask][-1]
                        for i, (eachIndex, eachGroupSkyIndex) in enumerate(zip(groupIndexs, thisSkyIndex)):
                            if not onlyMark:
                                popt[N+eachIndex] = thisGroupA[i]
                                popt[3*N+eachIndex] = thisGroupA3[i]
                            inrange = (wave>thisFlagLower[groupMask][i])&(wave<thisFlagUpper[groupMask][i])
                            modelFittingResult[eachGroupSkyIndex]["status"] = modelFittingResult[eachGroupSkyIndex]["status"]|0b100000000
                            resultMask[inrange] = np.bitwise_or(resultMask[inrange], 0b100000000)
                        if thisGroupTotalSum>thisGroupAUpperAlert and doAlert:
                            alertRegions.append([thisGroupLower-30, thisGroupUpper+30, thisGroupYmax, thisFactor,
                            "in {}:{}: too much flux for this group:  thisFactor: {:10.7f} with {}\npoly:{}\nmean_std:{:10.7f} Lower:{:10.7f} oldTotalA:{:10.7f} totalA:{:10.7f} Upper:{:10.7f}".\
                                    format(fibertype, name, thisFactor, thisGroupmu,
                                           list(zip(polyLower, thisPoly, polyUpper)),
                                           mean_std, thisGroupALower, thisGroupOldTotalSum,
                                           thisGroupTotalSum, thisGroupAUpper)])
                            toPrintStr = "find possible emission line in range {:7.3f}~{:7.3f} with factor {:7.3f} with {}".\
                                    format(thisGroupLower, thisGroupUpper, thisFactor, thisGroupmu)
                            if not haveWarningInThisBlock:
                                print("\n"+toPrintStr)
                                haveWarningInThisBlock = True
                            else:
                                print(toPrintStr)
                            logger.log(toPrintStr)
                            continue#!! to skip latter

                    thisFactor = thisGroupTotalSum / thisGroupALower
                    thisGroupA  = originGroupA / thisFactor
                    thisGroupA3 = originGroupA3 / thisFactor
                    if thisGroupTotalSum<thisGroupALower:
                        thisGroupYmax = np.max(thisGroupA/(np.sqrt(2*np.pi)*thisGroupsigma) + thisGroupA3/(2.575798*thisGroupsigma3))
                        thisGroupLower = thisFlagLower[groupMask][0]
                        thisGroupUpper = thisFlagUpper[groupMask][-1]
                        for i, (eachIndex, eachGroupSkyIndex) in enumerate(zip(groupIndexs, thisSkyIndex)):
                            if not onlyMark:
                                popt[N+eachIndex] = thisGroupA[i]
                                popt[3*N+eachIndex] = thisGroupA3[i]
                            inrange = (wave>thisFlagLower[groupMask][i])&(wave<thisFlagUpper[groupMask][i])
                            modelFittingResult[eachGroupSkyIndex]["status"] = modelFittingResult[eachGroupSkyIndex]["status"]|0b010000000
                            resultMask[inrange] = np.bitwise_or(resultMask[inrange], 0b010000000)
                        #!! do not warning for low flux, only show in plot
                        #alertRegions.append([thisGroupLower-30, thisGroupUpper+30, thisGroupYmax, thisFactor,
                            #"in {}: too little flux for this group:  thisFactor: {:10.7f} with {}".format(name, thisFactor, thisGroupmu)])
                        if doAlert:
                            toPrintStr = "find low emission line in range {:10.7f}~{:10.7f} with factor {:10.7f} with {}, may be bad continnum???".\
                                    format(thisGroupLower, thisGroupUpper, thisFactor, thisGroupmu)
                            if not haveWarningInThisBlock:
                                print("\n"+toPrintStr)
                                haveWarningInThisBlock = True
                            else:
                                print(toPrintStr)
                        logger.log(toPrintStr)
            #!! check series lower or upper, change poly boundarys and fit again
            thisStatus = modelFittingResult[skyIndexs]["status"]
            fitAgain = False
            excessUpper = ((thisStatus&0b100000000)>0)&((thisStatus&0b110)==0)
            excessLower = ((thisStatus&0b010000000)>0)&((thisStatus&0b110)==0)
            excessUpperSmallFlux = (thisStatus&0b100000000)>0
            if not useFile:
                #!! for too much flux
                thisStart, thisEnd = getContinueBlocks(excessUpper)
                for eachthisStart, eachthisEnd in zip(thisStart, thisEnd):
                    if (thismu[eachthisEnd]-thismu[eachthisStart])>maxEmissionWidth:
                        lowerFactor *= 0.6
                        upperFactor *= 1.4
                        fitAgain = True
                #if excessUpperSmallFlux.sum()>=3:
                    #lowerFactor *= 0.8
                    #upperFactor *= 1.2
                    #fitAgain = True
                #!! for too less flux
                thisStart, thisEnd = getContinueBlocks(excessLower)
                for eachthisStart, eachthisEnd in zip(thisStart, thisEnd):
                    if (thismu[eachthisEnd]-thismu[eachthisStart])>maxEmissionWidth:
                        lowerFactor *= 1.4
                        upperFactor *= 0.6
                        fitAgain = True
            goon = False
            if fitAgain:#!! yes
                fittingTweak = fittingTweakOrigin
                reCorrect=0
            fitAgain = fitAgain or lineDeltaFitAgain
            if not fitAgain or reFitCount>maxRefit:#!! good fit, or too much and give up
                if reFitCount>maxRefit:
                    goon = True
                upperFactor = 1
                lowerFactor = 1
                reFitCount = 0
                refit = refitOrigin
            else: #!! one more refit error
                refit = True
                reFitCount += 1
                now = time.time()
                toPrintStr = '\t\t bad continuum, Again:{} l:{:.2f} u:{:.2f} {}/{} {}:{} {:3}/{:3} third fitting: block {:>10.3f}~{:<10.3f} with {:3} profiles.. t={:.2f}s {:.2f}s {:.2f}s'.\
                      format(reFitCount, lowerFactor, upperFactor, fiberCount, fiberTotal, fibertype, name, count, blockSize, eachstart, eachend, nSky, now-time0, now-time00, now-time000)
                logger.log(toPrintStr)
                messager.cPrint("\n"+toPrintStr, "ykb")
                continue
            if haveResidualSubtract:#!! record regions that have object continuum subtract
                resultMask[wave_fit_mask] = np.bitwise_or(resultMask[wave_fit_mask], 0b1000000000)
        #region: record results
        tempTuple, thisPoly = packParameters(eachnPoly, popt)
        tempTupleUpper, thisPolyUpper = packParameters(eachnPoly, bounds_upper)
        tempTupleLower, thisPolyLower = packParameters(eachnPoly, bounds_lower)
        thisLines, thisContinuum = thisProfileSP(wave_fit, *popt)
        modelPolyResult[count] = (eachstart, eachend, eachstartTrue, eachendTrue, thisPoly)
        modelPolyResultUpper[count] = (eachstart, eachend, eachstartTrue, eachendTrue, thisPolyUpper)
        modelPolyResultLower[count] = (eachstart, eachend, eachstartTrue, eachendTrue, thisPolyLower)
        modelFittingPopts[count] = (nSky, popt)
        if fitname=="third":
            thisGaussianFLux = thisProfileGaussian(wave_fit, *gpopt)
            modelFittingPoptsOld[count] = (nSky, poptOld)
            tempTupleOld, thisPolyOld = packParameters(eachnPoly, poptOld)
            #!! thisContinuumOld should be equal to thisContinuum
            thisLinesOld, thisContinuumOld = thisProfileSP(wave_fit, *poptOld)
            if tweakContinuum:
                std_obj_fit = stdObj(wave_fit)

                lineContinuumMask = np.abs(thisContinuum)>tweakContinuumFactor * std_obj_fit
                lineContinuumSign = ((thisContinuum>0)-0.5)*2
                thisTweak = thisContinuum.copy()
                thisTweak[lineContinuumMask] = tweakContinuumFactor * std_obj_fit[lineContinuumMask] * lineContinuumSign[lineContinuumMask]
                thisLines = thisLines+thisTweak
                thisContinuum = thisContinuum-thisTweak

                lineContinuumMaskOld = np.abs(thisContinuumOld)>tweakContinuumFactor * std_obj_fit
                lineContinuumSignOld = ((thisContinuumOld>0)-0.5)*2
                thisTweakOld = thisContinuumOld.copy()
                thisTweakOld[lineContinuumMaskOld] = tweakContinuumFactor * std_obj_fit[lineContinuumMaskOld] * lineContinuumSignOld[lineContinuumMaskOld]
                thisLinesOld = thisLinesOld+thisTweakOld
                thisContinuumOld = thisContinuumOld-thisTweakOld


                #!! tweak residual continuum
                #!!     small objContinuum also count in sky line....
                lineContinuumMask = np.abs(residual_continuum_fit)>tweakResidualContinuumFactor * std_obj_fit
                lineContinuumSign = ((residual_continuum_fit>0)-0.5)*2
                thisTweak = residual_continuum_fit.copy()
                thisTweak[lineContinuumMask] = tweakResidualContinuumFactor * std_obj_fit[lineContinuumMask] * lineContinuumSign[lineContinuumMask]
                thisLines = thisLines+thisTweak
                thisLinesOld = thisLinesOld+thisTweak
                residual_continuum_fit = residual_continuum_fit-thisTweak

            thisContinuum    = thisContinuum    + residual_continuum_fit#!! checked, it should be placed here
            thisContinuumOld = thisContinuumOld + residual_continuum_fit#!! checked, it should be placed here
        else:
            #!! make the poly flux within 3 sigma to be skyline itself
            #!!      take the residual as continuum
            if tweakContinuum: #!! default and sky and obj is true
                std_obj_fit = stdObj(wave_fit)
                lineContinuumMask = np.abs(thisContinuum)>tweakContinuumFactor * std_obj_fit
                lineContinuumSign = (thisContinuum>0-0.5)*2
                thisTweak = thisContinuum.copy()
                thisTweak[lineContinuumMask] =\
                        tweakContinuumFactor * std_obj_fit[lineContinuumMask] *\
                            lineContinuumSign[lineContinuumMask]
                thisLines = thisLines+thisTweak
                thisContinuum = thisContinuum-thisTweak
        #!! thisLines is the skyline fitting result
        #!! thisContinuum is the skyline continuum fitting result
        #std2Obj = lambda x: stdObj(x)**2
        #!! update skyAUpper and skyA3Upper and sigmaError data using fitting sigma range
        for eachIndex, eachLower, eachUpper, eachType, eachResType, eachAveSig in\
                zip(skyIndexs, thisFlagLower3, thisFlagUpper3,
                    sky_type_fit, sky_residualType_fit, thisAveSigma):
            if eachType:
                thisStds = stdObj(np.linspace(eachLower, eachUpper, 20))
                thisstd = np.nanmedian(thisStds[thisStds>0])
                if not (np.isfinite(thisstd) and thisstd>0):
                    thisstd = np.nanmedian(std_fit)
                if not (np.isfinite(thisstd) and thisstd>0):
                    thisstd = np.nanmedian(std)
                #thisSigmaError, aux = integrate.quad(std2Obj, eachLower, eachUpper)
                #thisSigmaError = np.sqrt(thisSigmaError)
                thisSigmaError = thisstd * eachType * eachAveSig * 2.5
                modelFittingResult[eachIndex]["sigmaError"] = thisSigmaError
                modelFittingResultUpper[eachIndex]["sigmaError"] = thisSigmaError
                modelFittingResultLower[eachIndex]["sigmaError"] = thisSigmaError
                modelFittingResult[eachIndex]["sigmaErrorUpper"] = thisSigmaError*2
                modelFittingResultUpper[eachIndex]["sigmaErrorUpper"] = thisSigmaError*2
                modelFittingResultLower[eachIndex]["sigmaErrorUpper"] = thisSigmaError*2
                if np.isfinite(thisSigmaError):
                    skyAUpper[eachIndex] = thisSigmaError/2
                    skyA3Upper[eachIndex] = thisSigmaError/2
                    skyA[eachIndex] = skyAUpper[eachIndex]*0.99
                    skyA3[eachIndex] = skyA3Upper[eachIndex]*0.99
            else:
                thisStds = stdObj(np.linspace(eachLower, eachUpper, 20))
                thisstd = np.nanmedian(thisStds[thisStds>0])
                if not (np.isfinite(thisstd) and thisstd>0):
                    thisstd = np.nanmedian(std_fit)
                thisSigmaError = thisstd * 4 * eachAveSig * 2.5
                modelFittingResult[eachIndex]["sigmaError"] = thisSigmaError
                modelFittingResultUpper[eachIndex]["sigmaError"] = thisSigmaError
                modelFittingResultLower[eachIndex]["sigmaError"] = thisSigmaError
                modelFittingResult[eachIndex]["sigmaErrorUpper"] = thisSigmaError*2
                modelFittingResultUpper[eachIndex]["sigmaErrorUpper"] = thisSigmaError*2
                modelFittingResultLower[eachIndex]["sigmaErrorUpper"] = thisSigmaError*2
            if eachResType:
                thisStds = stdObj(np.linspace(eachLower, eachUpper, 20))
                thisstd = np.nanmedian(thisStds[thisStds>0])
                if not (np.isfinite(thisstd) and thisstd>0):
                    thisstd = np.nanmedian(std_fit)
                thisSigmaError = thisstd * eachResType * eachAveSig * 2.5
                modelFittingResult[eachIndex]["residualSigmaError"] = thisSigmaError
                modelFittingResultUpper[eachIndex]["residualSigmaError"] = thisSigmaError
                modelFittingResultLower[eachIndex]["residualSigmaError"] = thisSigmaError

        #!! fitting result for plot (fix offset)
        flux_fit_result_lines[wave_fit_mask_true] = thisLines[wave_fit_mask_mask]
        flux_fit_result_poly[wave_fit_mask_true] = thisContinuum[wave_fit_mask_mask]
        fit_result = thisLines + thisContinuum
        flux_fit_result[wave_fit_mask_true] = fit_result[wave_fit_mask_mask]
        tempResult = np.array(tempTuple)
        tempResultUpper = np.array(tempTupleUpper)
        tempResultLower = np.array(tempTupleLower)
        for i,allIndex in zip(skyIndexsTrueIndex, skyIndexsTrue):
            modelFittingResult[allIndex]["mu"] = tempResult[i,0]
            modelFittingResult[allIndex]["A"] = tempResult[i,1]
            modelFittingResult[allIndex]["sigma"] = tempResult[i,2]
            modelFittingResult[allIndex]["A3"] = tempResult[i,3]
            modelFittingResult[allIndex]["sigma3"] = tempResult[i,4]
            modelFittingResultUpper[allIndex]["mu"] = tempResultUpper[i,0]
            modelFittingResultUpper[allIndex]["A"] = tempResultUpper[i,1]
            modelFittingResultUpper[allIndex]["sigma"] = tempResultUpper[i,2]
            modelFittingResultUpper[allIndex]["A3"] = tempResultUpper[i,3]
            modelFittingResultUpper[allIndex]["sigma3"] = tempResultUpper[i,4]
            modelFittingResultLower[allIndex]["mu"] = tempResultLower[i,0]
            modelFittingResultLower[allIndex]["A"] = tempResultLower[i,1]
            modelFittingResultLower[allIndex]["sigma"] = tempResultLower[i,2]
            modelFittingResultLower[allIndex]["A3"] = tempResultLower[i,3]
            modelFittingResultLower[allIndex]["sigma3"] = tempResultLower[i,4]
        if fitname=="third":
            flux_fit_gaussian_result[wave_fit_mask_true] = thisGaussianFLux[wave_fit_mask_mask]
            flux_fit_result_linesOld[wave_fit_mask_true] = thisLinesOld[wave_fit_mask_mask]
            flux_fit_resultOld[wave_fit_mask_true] = (thisLinesOld + thisContinuumOld)[wave_fit_mask_mask]
            flux_fit_result_obj_continuum[wave_fit_mask_true] = residual_continuum_fit[wave_fit_mask_mask]
            polyDebugInfo[count] = (count, eachstart, eachstartTrue,eachendTrue, eachend, eachnPoly,
                        bounds_lower[5*N:], popt[5*N:], bounds_upper[5*N:])
            for i,allIndex in zip(skyIndexsTrueIndex, skyIndexsTrue):
                modelFittingResult[allIndex]["gaussianMu"] = thisgmu[i]
                modelFittingResult[allIndex]["gaussianA"] = thisgA[i]
                modelFittingResult[allIndex]["gaussianSigma"] = thisgsigma[i]
            if not fitAgain or reFitCount>maxRefit or goon:#!! good fit, or too much and give up
                count += 1
        else:
            count += 1
        #endregion: record results
    if fitname=="third":
        return polyDebugInfo


#region: backup snippet
"""
nPoly = 8
profile = specLib.profileGen(nPoly)
#!! mu, A, sigma, A3, sigma3
opts = np.concatenate((sky_fit,
                       np.ones(sky_fit.size), sky_std_fit,
                       np.ones(sky_fit.size), sky_std_fit/2,
                       np.ones(nPoly)))
y = profile(wave_fit, *(opts.tolist()))
fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(wave_fit, flux_fit)
ax.plot(wave_fit, y)
plt.show()

originskywave         =skywave.copy()
originskysigma        =skysigma.copy()
originskysigma3       =skysigma3.copy()
originskytype         =skytype.copy()
originskyGroup        =skyGroup.copy()
originskyResidualType =skyResidualType.copy()
originskySigmaLower   =skySigmaLower.copy()
originskySigmaUpper   =skySigmaUpper.copy()
originskySigma3Lower  =skySigma3Lower.copy()
originskySigma3Upper  =skySigma3Upper.copy()
originskyALower       =skyALower.copy()
originskyAUpper       =skyAUpper.copy()
originskyA3Lower      =skyA3Lower.copy()
originskyA3Upper      =skyA3Upper.copy()
originsky_upper       =sky_upper.copy()
originsky_lower       =sky_lower.copy()
assert np.all(originskywave==skywave)
assert np.all(originskysigma==skysigma)
assert np.all(originskysigma3==skysigma3)
assert np.all(originskytype==skytype)
assert np.all(originskyGroup==skyGroup)
assert np.all(originskyResidualType==skyResidualType)
assert np.all(originskySigmaLower==skySigmaLower)
assert np.all(originskySigmaUpper==skySigmaUpper)
assert np.all(originskySigma3Lower==skySigma3Lower)
assert np.all(originskySigma3Upper==skySigma3Upper)
assert np.all(originskyALower==skyALower)
assert np.all(originskyAUpper==skyAUpper)
assert np.all(originskyA3Lower==skyA3Lower)
assert np.all(originskyA3Upper==skyA3Upper)
assert np.all(originsky_upper==sky_upper)
assert np.all(originsky_lower==sky_lower)

dtype = np.dtype([
            ("mu", np.float64), ("mumin", np.float64), ("mumax", np.float64),
            ("min",np.float64), ("max",np.float64),
            ("mean",np.float64),("sigma",np.float64),
            ("goodmean",np.float64),("goodsigma",np.float64),
            ("sigmaError",np.float64),
            ("pmin",np.float64), ("pmax",np.float64),
            ("pmean",np.float64),("psigma",np.float64),
            ("pgoodmean",np.float64),("pgoodsigma",np.float64),
        ])
skyDebugResult = np.empty((skywave.size,), dtype=dtype)
for i in range(skywave.size):
    skyDebugResult[i]["mumin"] = allFittingmu[:,i].min()
    skyDebugResult[i]["mumax"] = allFittingmu[:,i].max()
    skyDebugResult[i]["mu"] = allFittingmu[:,i].mean()
    todoA = allFittingFlux[:,i][allFittingFlux[:,i]>0]
    skyDebugResult[i]["min"]  = todoA.min()
    skyDebugResult[i]["max"]  = todoA.max()
    skyDebugResult[i]["mean"] = todoA.mean()
    skyDebugResult[i]["sigma"] = todoA.std()


"""
#endregion: backup snippet
